local core = require("core")
local gmui = require("gmui")
local wave = require("WildCastle/WaveGenerator")
local data = require("WildCastle/GameData")
local levels = require("WildCastle/LevelData")
local ui = require("WildCastle/UI")
local shop = require("lazy_shop_lite")
local abtest = require("WildCastle/ABtest")
local bank = require("Bank")
local event_mode = require("EventMode")
local json = require("json")
local game = {}

-- version 1.13.1

-- set global tick rate to every two frames for (theoretically) better performance
local GLOBAL_TICK_RATE = 0.125
core.SetGlobalTimerTickRate(GLOBAL_TICK_RATE)

local GLOBAL_EVENT_MODE = false

-- Variables
local group_hero_slots = {}
local group_tower_slots = {}
local group_defense_slots = {}
local group_archers = {}
local group_castle_wall = {}
local group_summons = {}
local group_worker_dwarf = {}
local group_town_slots = {}
local gear_anchor

local CASTLE_CHUNK_NAME = DCEI.Unit("Player Castle Chunk")
local CASTLE_UNIT_NAME = DCEI.Unit("Player Castle Controller")
local CASTLE_UNIT
local CASTLE_UNIT_INVINC -- use to avoid damage and heal at the same frame and waste heal effect

local AIRSHIP_DOCK
local AIRSHIP
local AIRSHIP_LANDED = false

local GEAR_UNITY_HELPER_BUFF_INSTANCE = 0
local GEAR_UNITY_HELPER
local TOWN_HALL
local MINE_UNIT
local MINE_DEPOT
local VAULT
local MINE_UNIT_NAME = DCEI.Unit("Town Mechanic Mine")

local TOWN_SLOT_NAME = DCEI.Unit("Town Building Slot")

local bless_crystal

local DAILY_ADS_LIMIT = 10
local CRAFT_AUTO_REFRESH_TIME = 21600
local CRAFT_MAX_LEVEL = 5
local POINTS = {}
local SCENE = {}
local DEV = {
    airship_mode_tracking = 0,
    test_send_notification_tracking = 0,
}
local LEGENDARY_REFRESH_CHANCE = 0.4
-- bug tracking diagnostic data
local DIAGNOSTIC = {}

local BUILD = {
    max_castle = 120,
    max_town = 50,
}
local watch_ad_fail_count = 0
local ad_data = {
    gold = {
        name = "gold",
        is_available = true,
        cooldown = 180,
    },
    speed = {
        name = "speed",
        limit_count = 1,
        limit_duration = 3600,
        duration = 1260,
    },
}

local system = {
    locked = "locked",
    unlocked = "unlocked",
    active = "active",
}

local player = {
    remove_ad = false,
    player_xp = 0,
    paragon_xp = 0,
    speed = 1,
    heroes = {},
    towers = {},
    defenses = {},
    town = {},
    gear = {},
    inventory = {
        capacity = 160,
        items = {},
    },
    daily_ads_limit = DAILY_ADS_LIMIT,
    speed_up = false,
    speed_up_timer = nil,
    speed_ad_timer = nil,
    seasonpass_timer = nil,
    monster_bounty_timer = nil,
    gem_auto_time = 0,
    gem_auto_timer = nil,
    auto_save_timer = nil,
    targeting_data = nil,
    ads_watched_gold = 0,
    ads_watched_speed = 0,
    username = {},
    autobattle = {
        is_active = false,
        auto_stop_when_token_full = false,
        type = "none",
        priority_list = {},
    },
    system = {
        promote_tutorial = system.locked,
        town_upgrade_tutorial = system.locked,
        airship = system.locked,
        deal_popup = system.locked,
        autobattle_popup = system.locked,
        gear_equip_tutorial = system.locked,
        raid_boss_tutorial = system.locked,
        open_first_common_chest = system.locked,
        first_raid_boss = system.locked,
        vault = system.locked,
        first_open_monster_lab = system.locked,
        first_play_hell_mode = system.locked,
        popup_massive_bundle = system.locked,
        check_seceondary_menu = system.locked,
        colony_travel_tutorial = system.locked,
    },
    calendar = {
        day = -1,
        reward_index = 0,
        reward_available = false,
        total_rewards_claimed = 0,
        boost_index = 0,
    },
    hell_coin_shop = {
        start_rotate = false,
        rotate_time = 0,
        badge_index = 1,
        owned = {
            owned_badges = {},
        },
    },
    player_deck = {
        current_id = 1,
        decks = { "empty", "empty", "empty", "empty" },
        names = {
            DCEI.GetTranslationText("ui/deck_number", { x = 1 }),
            DCEI.GetTranslationText("ui/deck_number", { x = 2 }),
            DCEI.GetTranslationText("ui/deck_number", { x = 3 }),
            DCEI.GetTranslationText("ui/deck_number", { x = 4 }),
        },
    },
    shop_1st_time_bonus = {},
    shop_logging = {},
    spend_log = {},
    purchases = {},
    store_item_keys = {},
    tests = {},
    timers = {},
    raid_boss_log = {},
    wave_attempts = 0,
    replay_count = 0,
    wave_loss_count = 0,
    colony_loss_count = 0,
    equipment_upgrades = 0,
    enchant_non_epic = 0,
    enchant_epic = 0,
    claimed_mails = {},
    guarantee_epic = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    guarantee_legendary = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    non_epic_chest = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    non_legendary_chest = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    epic_interval = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    legendary_interval = {
        ["rare_chest_l"] = 0,
        ["rare_chest"] = 0,
        ["legendary_chest_l"] = 0,
        ["legendary_chest"] = 0,
    },
    craft_types = { 1, 5, 2 },
    craft_refresh = os.time() + CRAFT_AUTO_REFRESH_TIME,
    craft_legendary = false,
    craft_rarity = {},
    craft_refresh_currency = {},
    wave_win_increment = {},
    monster_data = {},
    monster_lab_token = 0,
    hell_mode = {},
    hell_mode_event_claim_list = {},
    raid_key_extra_purchase_count = 0,
    raid_key_extra_purchase_count_total = 0,
    raid_key_extra_purchase_count_spend = 0,
    winter_event = {},
    spring_event = {},
    royal_event = {},
    autumn_event = {},
    summer_event = {},
    winter_event2022 = {},
    egg_hunt_event = {
        raw = 0,
        egg = 0,
        gift = 0,
        open_gift = 0,
    },
    race_event = {},
    event = {},
}

local gear_upgrade_gold = 0
local gear_menu_click = 0

local raid_key_extra_purchase_cost = {
    1,
    2,
    4,
    6,
    8,
    10,
}

local wave_killed_monsters = {}
local wave_drop_resource = {}

player.attributes = {
    monster_gold = 0,
    hellmode_gold = 0,
    monster_xp = 0,
    token_tower_progress = 1,
    material_tower_progress = 1,
    max_bonus_token_wave = 8,
    max_bonus_material_wave = 8,
    monster_elixir = 0,
    monster_exp_trial = 0,
    mine_income = 0,
    colony_gold = 0,
    --colony_gold_flat = 0,
    elixir_production = 0,
    colony_cooldown = 0,
    all_gold = 0,
    hero_cooldown = 0,
    hero_cooldown_increase = 0,
    hero_mana_cost_increase = 0,
    hero_damage = 0,
    hero_damage_flat = 0,
    hero_life_steal = 0,
    tower_damage = 0,
    tower_damage_flat = 0,
    archer_damage = 0,
    archer_damage_flat = 0,
    summon_health = 0,
    summon_damage = 0,
    summon_human_health = 0,
    summon_human_damage = 0,
    summon_construct_health = 0,
    summon_construct_damage = 0,
    all_damage = 0,
    elemental_damage_fire = 0,
    elemental_damage_cold = 0,
    elemental_damage_lightning = 0,
    elemental_damage_poison = 0,
    elemental_damage_magic = 0,
    health_bonus_90 = 0,
    health_bonus_100 = 0,
    auto_battle_cost_reduction = 0,
    bless_element = nil,
    bless_wave = 0,
    bless_bonus = 1, -- 1 for gold bonus, 2 for exp bonus
    bless_bonus_count = 0,
    bless_bonus_units = {},
    token_cap = 0,
    skill_archer_cri_dmg = 0,
}

player.devil_trail = {
    key = nil,
    wave = 0,
    victory_wave = 0,
    fail = 0,
    seed = os.time(),
    accept = false,
    trigger_count = 0,
    accept_count = 0,
    success_count = 0,
    lost_count = 0,
}

levels.data = data
levels.monsters = wave.monsters
wave.data.player_attributes = player.attributes
wave.data.lab_monsters = data.lab_monsters
data.player_attributes = player.attributes
data.player = player
ui.item_data = data.items
ui.equipments = data.equipments
ui.player_attributes = player.attributes
ui.wave_data = wave.data
ui.maps = levels.maps
ui.lab_monsters = data.lab_monsters
ui.hell_mode_data = data.hell_mode
ui.hell_mode_seasonal_reward = data.hell_mode_seasonal_reward
ui.hell_mode_event_reward = data.hell_mode_event_reward
ui.event_mode = event_mode

local server = {
    day_fetched = false,
    is_connecting = false,
}

local RAID_KEY_DAILY_COUNT = 7

local HELL_MODE_SEASON_START = 1627995600 + 24 * 3600
local SEASON_DURATION = 7 * 24 * 3600
local SEASON_INTERVAL = 24 * 3600 -- Season from Wednesday to Tuesday
local NEW_RANK_REWARD_SEASON = 13
local HELL_COIN_SHOP_ROTATE_DURATION = 86400 * 30
local EGG_HUNT_DURATION = 8 * 24 * 3600
local EGG_UPGRADE_REQUIRE = 80
local RACE_EVENT_DURATION = 24 * 3600
local RACE_EVENT_COOLDOWN = 10 -- 10 seconds
local RACE_EVENT_TOTAL_PROGRESS = 50
local SEASONAL_COLONIES_RESET_TIME = 10 * 24 * 3600

local notifications = {}
local next_daily_reward_time
local notification_texts = {
    {
        title = DCEI.GetTranslationText("notification/rewards/title/a"),
        body = DCEI.GetTranslationText("notification/rewards/body/a"),
    },
    {
        title = DCEI.GetTranslationText("notification/rewards/title/b"),
        body = DCEI.GetTranslationText("notification/rewards/body/b"),
    },
}
local vault_full_notification_id

local cameras = {
    current = 1,
    castle = 1,
    town = 2,
    towers = 3,
    defenses = 4,
    mines = 5,
    buildings = 6,
    airship = 7,
    town_hall = 8,
    shop = 9,
    vault = 10,
    { x = 24, y = 24.4, distance = 46 }, -- castle
    { x = 24, y = 6.75, distance = 28 }, -- town
    { x = 24, y = 28.375, distance = 49 }, -- castle: towers
    { x = 24, y = 23, distance = 49 }, -- castle: defenses
    { x = 22, y = 13.5, distance = 26 }, -- town: mines
    { x = 24, y = 10.75, distance = 40 }, -- town: buildings
    { x = 24, y = 10, distance = 35 }, -- town: airship
    { x = 24, y = 9.15, distance = 40 }, -- town: town hall
    { x = 20, y = 24.4, distance = 46 }, -- shop
    { x = 23.2, y = 14, distance = 26 }, -- town: vault
}

local tutorial = {
    state = 0,
    message = {
        hero_skills01 = DCEI.GetTranslationText("campaign/tutorial/hero_skills/01"),
        hero_skills02 = DCEI.GetTranslationText("campaign/tutorial/hero_skills/02"),
        upgrade_archer = DCEI.GetTranslationText("campaign/tutorial/upgrade_archer"),
        upgrade_castle = DCEI.GetTranslationText("campaign/tutorial/upgrade_castle"),
        mine01 = DCEI.GetTranslationText("campaign/tutorial/mine/01"),
        mine02 = DCEI.GetTranslationText("campaign/tutorial/mine/02"),
        mine03 = DCEI.GetTranslationText("campaign/tutorial/mine/03"),
        mine04 = DCEI.GetTranslationText("campaign/tutorial/mine/04"),
        boss_retry_first = DCEI.GetTranslationText("campaign/tutorial/boss_retry/01"),
        boss_retry_again = {
            DCEI.GetTranslationText("campaign/tutorial/boss_retry/02"),
            DCEI.GetTranslationText("campaign/tutorial/boss_retry/03"),
            DCEI.GetTranslationText("campaign/tutorial/boss_retry/04"),
        },
        talent = DCEI.GetTranslationText("campaign/tutorial/talent"),
        talent_menu = DCEI.GetTranslationText("campaign/tutorial/open_menu"),
        gear = DCEI.GetTranslationText("campaign/tutorial/gear"),
        username = DCEI.GetTranslationText("campaign/tutorial/username"),
        challenge = DCEI.GetTranslationText("campaign/tutorial/challenge"),
        hero_level = DCEI.GetTranslationText("campaign/tutorial/hero_level"),
        hero_promote01 = DCEI.GetTranslationText("campaign/tutorial/hero_promote/01"),
        hero_promote02 = DCEI.GetTranslationText("campaign/tutorial/hero_promote/02"),
        hero_promote03 = DCEI.GetTranslationText("campaign/tutorial/hero_promote/03"),
        colony01 = DCEI.GetTranslationText("campaign/tutorial/colony/01"),
        colony02 = DCEI.GetTranslationText("campaign/tutorial/colony/02"),
        colony03 = DCEI.GetTranslationText("campaign/tutorial/colony/03"),
        colony04 = DCEI.GetTranslationText("campaign/tutorial/colony/04"),
        town01 = DCEI.GetTranslationText("campaign/tutorial/town/01"),
        town02 = DCEI.GetTranslationText("campaign/tutorial/town/02"),
        town03 = DCEI.GetTranslationText("campaign/tutorial/town/03"),
        town_building = DCEI.GetTranslationText("campaign/tutorial/town_building"),
        facebook = DCEI.GetTranslationText("campaign/tutorial/facebook"),
        equip_gear = DCEI.GetTranslationText("campaign/tutorial/equip_gear"),
        raid_boss = DCEI.GetTranslationText("campaign/tutorial/raid_boss"),
        equipment_system = DCEI.GetTranslationText("campaign/tutorial/equipment"),
    },
    talent = false,
    gear = false,
    town = false,
    username = false,
    challenge = false,
    colony = false,
}

local district_layers = {
    "Town_Camp01_Fire",
    "Town_Camp02_Fire",
    "Town_Camp03_Fire",
    "Town_Camp04_Fire",
}

local speed = {
    slow = 1 / 64,
    very_slow = 1 / 128,
}

local LAYER_DATA = {}
local LAYER_DATA_LEVEL = -1
local NEXT_PRELOAD_CHUNK = nil

function SortedPairs(t)
    local keys = {}
    for k, _ in pairs(t) do
        table.insert(keys, k)
    end
    if #keys == 0 then
        return pairs(t)
    end
    table.sort(keys, function(a, b)
        -- in alphabetical order
        local t1 = type(a)
        local t2 = type(b)
        if t1 ~= t2 then
            return t1 < t2
        elseif t1 == "table" then
            return false
        else
            return a < b
        end
    end)
    local index = 0
    local function next(t)
        index = index + 1
        if index <= #keys then
            local k = keys[index]
            return k, t[k]
        end
    end
    return next, t, nil
end

-- SECTION Load town and castle props
function GetPotentialTownLevel()
    local upgrade_level = GetTownUpgradeLevel()
    local contract_count = GetTownContracts()
    local potential_visuals = data.town_upgrade:GetVisualLevel(contract_count + upgrade_level)

    return potential_visuals
end

function DetermineTownChunksToLoad(layers_to_load)
    local preloadable_layers = data.TERRAIN_LAYERS.preloadable_town_chunks
    local unloadable_layers = data.TERRAIN_LAYERS.unloadable_town_chunks
    local potential_level = GetPotentialTownLevel()
    local current_level = data.town_upgrade:GetVisualLevel()

    for _, chunk in ipairs(preloadable_layers) do
        -- mark town layer as needed if potential town level is passed load requirement
        if potential_level >= chunk.load_at then
            for _, name in ipairs(chunk) do
                layers_to_load[name] = true
            end
        end
    end

    for _, chunk in ipairs(unloadable_layers) do
        -- mark town layers as not needed if town level is passed unload requirement
        if current_level >= chunk.unload_at then
            for _, name in ipairs(chunk) do
                layers_to_load[name] = false
            end
        end
    end

    return layers_to_load
end

function DetermineCastleChunksToLoad(layers_to_load)
    local castle_level = GetCastleLevel()
    local preloadable_layers = data.TERRAIN_LAYERS.preloadable_castle_chunks
    local unloadable_layers = data.TERRAIN_LAYERS.unloadable_castle_chunks

    for _, chunk in ipairs(preloadable_layers) do
        if castle_level >= chunk.load_at then
            DCEI.LogMessage("Preload level: " .. chunk.load_at)
            for _, name in ipairs(chunk) do
                layers_to_load[name] = true
            end
        else
            NEXT_PRELOAD_CHUNK = chunk.load_at
            break
        end
    end

    -- load next chunks on game start
    if NEXT_PRELOAD_CHUNK then
        DCEI.LogMessage("Preload next castle chunks level: " .. NEXT_PRELOAD_CHUNK)
        local chunk = data.TERRAIN_LAYERS.preload_by_castle_level[NEXT_PRELOAD_CHUNK]
        for _, name in ipairs(chunk) do
            layers_to_load[name] = true
        end
        NEXT_PRELOAD_CHUNK = NEXT_PRELOAD_CHUNK + 20
        if NEXT_PRELOAD_CHUNK > 101 then
            NEXT_PRELOAD_CHUNK = nil
        end
    end

    for _, chunk in ipairs(unloadable_layers) do
        -- mark castle layer as not needed if castle is progressed passed unload requirement
        if castle_level >= chunk.unload_at then
            for _, name in ipairs(chunk) do
                layers_to_load[name] = false
            end
        end
    end

    return layers_to_load
end

function PreloadPropLayers()
    local preloadable_layers = data.TERRAIN_LAYERS.preloadable
    local layers_to_load = {}
    local count = 0

    -- add all preloadable layers to layers_to_load
    for _, name in ipairs(preloadable_layers) do
        layers_to_load[name] = true
    end

    -- remove unecessary layers from layers_to_load
    DetermineCastleChunksToLoad(layers_to_load)
    DetermineTownChunksToLoad(layers_to_load)

    -- load layers_to_load
    for name, layer_state in pairs(layers_to_load) do
        if layer_state then
            local flag = false
            DCEI.LoadPropLayer(name, flag)

            count = count + 1
        end
    end

    DCEI.LogMessage("> LAYERS LOADED = " .. count .. " / " .. #preloadable_layers)
end

function PreloadNextTownChunks()
    local preloadable_layers = data.TERRAIN_LAYERS.preloadable_town_chunks
    local potential_level = GetPotentialTownLevel()
    local layers_to_load = {}

    for _, chunk in ipairs(preloadable_layers) do
        -- mark town layer as needed if potential town level is passed load requirement
        if potential_level >= chunk.load_at then
            for _, name in ipairs(chunk) do
                -- DCEI.LogMessage("> loaded " .. name)
                layers_to_load[name] = true
            end
        end
    end

    local count = 0
    for name, layer_state in pairs(layers_to_load) do
        if layer_state then
            local flag = false
            DCEI.LoadPropLayer(name, flag)

            count = count + 1
        end
    end
    DCEI.LogMessage("> LOADED " .. count .. " TOWN CHUNKS")
end

function PreloadNextCastleChunks(level)
    -- preload next chunks 5 levels ahead
    if NEXT_PRELOAD_CHUNK and level + 5 >= NEXT_PRELOAD_CHUNK then
        DCEI.LogMessage("Upgrade and preload castle chunks level: " .. NEXT_PRELOAD_CHUNK)
        local chunk = data.TERRAIN_LAYERS.preload_by_castle_level[NEXT_PRELOAD_CHUNK]
        for _, layer in ipairs(chunk) do
            DCEI.LoadPropLayer(layer, false)
        end
        NEXT_PRELOAD_CHUNK = NEXT_PRELOAD_CHUNK + 20
        if NEXT_PRELOAD_CHUNK > 101 then
            NEXT_PRELOAD_CHUNK = nil
        end
    end
end

-- preload more castle chunks on airship transition
function AttempToLoadMoreCastleChunks(chunk_count)
    local n = chunk_count or 1
    for i = 1, n do
        if NEXT_PRELOAD_CHUNK then
            DCEI.LogMessage("Preload castle chunks level: " .. NEXT_PRELOAD_CHUNK)
            local chunk = data.TERRAIN_LAYERS.preload_by_castle_level[NEXT_PRELOAD_CHUNK]
            for _, layer in ipairs(chunk) do
                DCEI.LoadPropLayer(layer, false)
            end
            NEXT_PRELOAD_CHUNK = NEXT_PRELOAD_CHUNK + 20
            if NEXT_PRELOAD_CHUNK > 101 then
                NEXT_PRELOAD_CHUNK = nil
            end
        else
            break
        end
    end
end

function AttemptQueueUnloadTownChunk(level)
    local chunk = data.TERRAIN_LAYERS.unloadable_by_town_level[level]

    if chunk then
        DCEI.LogMessage("> QUEUE UNLOAD TOWN CHUNK " .. level)
        table.insert(data.TERRAIN_LAYERS.unload_queue, chunk)
    end
end

function AttemptQueueUnloadCastleChunk(level)
    local chunk = data.TERRAIN_LAYERS.unloadable_by_castle_level[level]

    if chunk then
        DCEI.LogMessage("> QUEUE UNLOAD CASTLE CHUNK " .. level)
        table.insert(data.TERRAIN_LAYERS.unload_queue, chunk)
    end
end

function UnloadQueuedChunks()
    for _, segment in ipairs(data.TERRAIN_LAYERS.unload_queue) do
        local level = segment.unload_at or -1
        DCEI.LogMessage("> UNLOAD CHUNK from " .. level)

        for _, layer in ipairs(segment) do
            DCEI.UnloadPropLayer(layer)
        end
    end

    data.TERRAIN_LAYERS.unload_queue = {}
end

-- !SECTION

function InitData()
    local skip_commit = true
    local allow_overflow = true

    local key = bank.bank_data.sessions.key
    local sessions = bank.bank_data.sessions.value
    if sessions ~= 0 then
        DCEI.Save.Set(key, sessions + 1)
        bank.TriggerCommit()
    end

    local value = bank.bank_data.dcoin_spent.value
    SpendPlayerDCoin(value, skip_commit)

    local value = bank.bank_data.gold.value
    SetPlayerGold(value)

    local value = bank.bank_data.vault_level.value
    SetVaultLevel(value)

    local value = bank.bank_data.player_level.value
    SetPlayerLevel(value)

    local value = bank.bank_data.paragon_level.value
    SetParagonLevel(value)

    local value = bank.bank_data.archer_level.value
    SetArcherLevel(value)

    local value = bank.bank_data.castle_level.value
    SetCastleLevel(value)

    local value = bank.bank_data.talent_point.value
    SetTalentPoints(value)

    local value = bank.bank_data.paragon_point.value
    SetParagonPoints(value)

    local value = bank.bank_data.dwarf_level.value
    SetDwarfLevel(value)

    local value = bank.bank_data.town_level.value
    SetTownUpgradeLevel(value)

    -- elixir must be loaded after town level for proper elixir cap
    local value = bank.bank_data.elixir.value
    SetPlayerElixir(value, skip_commit)

    local value = bank.bank_data.winter_coin.value
    SetPlayerWinterCoin(value, skip_commit)

    local value = bank.bank_data.spring_coin.value
    SetPlayerSpringCoin(value, skip_commit)

    local value = bank.bank_data.royal_coin.value
    SetPlayerRoyalCoin(value, skip_commit)

    local value = bank.bank_data.summer_coin.value
    SetPlayerSummerCoin(value, skip_commit)

    local value = bank.bank_data.autumn_coin.value
    SetPlayerAutumnCoin(value, skip_commit)

    local value = bank.bank_data.winter_coin2022.value
    SetPlayerWinterCoin2022(value, skip_commit)

    local value = bank.bank_data.max_wave.value
    wave.data.SetMax(value)

    data.vault:UpdateVaultCap(wave.data.GetMax())
    local value = bank.bank_data.vault_gold.value
    SetVaultGold(value)

    local value = bank.bank_data.vault_record.value
    data.vault.time_record = value

    local value = bank.bank_data.dev_mode_enabled.value
    bank.player.dev_mode = value or bank.player.dev_mode

    local value = bank.bank_data.airship_test_mode.value
    bank.player.airship_test_mode = value

    local value = bank.bank_data.os_days.value
    bank.player.os_days = value

    local value = bank.bank_data.is_cheater.value
    bank.player.is_cheater = value

    local value = bank.bank_data.shop_logging.value
    local key = bank.bank_data.shop_logging.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.spend_log.value
    local key = bank.bank_data.spend_log.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.store_item_keys.value
    local key = bank.bank_data.store_item_keys.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.timers.value
    local key = bank.bank_data.timers.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.raid_boss_log.value
    local key = bank.bank_data.raid_boss_log.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.remove_ad.value
    local key = bank.bank_data.remove_ad.key

    bank.player[key] = value or false

    -- load raid key
    local value = bank.bank_data.raid_key_daily.value
    SetRaidKeyDailyCount(value or bank.RAID_KEY_DAILY_COUNT)
    local value = bank.bank_data.raid_key_extra.value
    SetRaidKeyExtraCount(value)
    -- local key = bank.bank_data.raid_key_extra_purchase_count
    -- bank.player[key] = DCEI.Save.Get(key) or 0
    -- local key = bank.bank_data.raid_key_extra_purchase_count_total
    -- bank.player[key] = DCEI.Save.Get(key) or 0
    -- local key = bank.bank_data.raid_key_extra_purchase_count_spend
    -- bank.player[key] = DCEI.Save.Get(key) or 0

    local value = bank.bank_data.claimed_mails.value
    local key = bank.bank_data.claimed_mails.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}
    bank.player.monster_data = {}

    if sessions == 0 then
        bank.Initialize()
        bank.InitializeTests()
    else
        -- load multi-step user data
        bank.LoadEventMode()
        bank.LoadBless()
        bank.LoadDevilTrail()
        bank.LoadTowerWaveProgress()
        bank.LoadGems()
        bank.LoadHeroes()
        bank.LoadTowers()
        bank.LoadDefenses()
        bank.LoadTown()
        bank.LoadGear()
        bank.LoadTalents()
        bank.LoadSystemState()
        bank.LoadChallengeState()
        bank.LoadColonyState()
        bank.LoadSeasonalColonyState()
        bank.LoadUsername()
        bank.LoadCalendarData()
        bank.LoadHellCoinShopData()
        bank.LoadAdData()
        bank.LoadTimers()
        bank.LoadPlayerDeckData()
        bank.LoadPurchaseData()
        bank.LoadFirstTimeBonus()
        bank.LoadInventory()
        bank.LoadTests()
        bank.LoadCraft()
        bank.LoadMonsterData()
        bank.LoadHellModeData()
        bank.LoadMonsterBounty()
        bank.LoadSeasonEvent("autumn_event")
        bank.LoadSeasonEvent("winter_event2022")
        bank.LoadEggHuntEvent()
        bank.LoadRaceEvent()

        -- when player updates to 0.0.118, check for retroactively earned contracts
        bank.RetroTownContractCheck()

        -- send spend log on app launch to reduce cases of "losing" spend logs
        local trigger = "App Launch"
        AttemptPushExistingSpendLog(trigger)
    end

    -- log username on load to capture usernames set before 0.0.109-patch
    local user_prop = bank.UserProps.username
    bank.LogUserProperty(user_prop)
end

function InitEventData()
    local skip_commit = true
    local allow_overflow = true

    local key = bank.bank_data.event_sessions.key
    local event_sessions = bank.bank_data.event_sessions.value
    if event_sessions ~= 0 then
        DCEI.Save.Set(key, event_sessions + 1)
        bank.TriggerCommit()
    end

    local value = bank.bank_data.dcoin_spent.value
    SpendPlayerDCoin(value, skip_commit)

    local value = bank.bank_data.event_gold.value
    SetPlayerGold(value)

    local value = bank.bank_data.vault_level.value
    SetVaultLevel(value)

    local value = bank.bank_data.event_player_xp.value
    bank.player.player_xp = value

    local value = bank.bank_data.event_player_level.value
    SetPlayerLevel(value)

    local value = bank.bank_data.event_paragon_level.value
    SetParagonLevel(value)

    local value = bank.bank_data.event_archer_level.value
    SetArcherLevel(value)

    local value = bank.bank_data.event_castle_level.value
    SetCastleLevel(value)

    local value = bank.bank_data.event_talent_point.value
    SetTalentPoints(value)

    local value = bank.bank_data.event_paragon_point.value
    SetParagonPoints(value)

    local value = bank.bank_data.event_dwarf_level.value
    SetDwarfLevel(value)

    local value = bank.bank_data.town_level.value
    SetTownUpgradeLevel(value)

    -- elixir must be loaded after town level for proper elixir cap
    local value = bank.bank_data.elixir.value
    SetPlayerElixir(value, skip_commit)

    local value = bank.bank_data.event_max_wave.value
    wave.data.SetMax(value)

    data.vault:UpdateVaultCap(wave.data.GetMax())
    local value = bank.bank_data.vault_gold.value
    SetVaultGold(value)

    local value = bank.bank_data.vault_record.value
    data.vault.time_record = value

    local value = bank.bank_data.dev_mode_enabled.value
    bank.player.dev_mode = value or bank.player.dev_mode

    local value = bank.bank_data.airship_test_mode.value
    bank.player.airship_test_mode = value

    local value = bank.bank_data.os_days.value
    bank.player.os_days = value

    local value = bank.bank_data.is_cheater.value
    bank.player.is_cheater = value

    local value = bank.bank_data.shop_logging.value
    local key = bank.bank_data.shop_logging.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.spend_log.value
    local key = bank.bank_data.spend_log.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.store_item_keys.value
    local key = bank.bank_data.store_item_keys.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.timers.value
    local key = bank.bank_data.timers.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.raid_boss_log.value
    local key = bank.bank_data.raid_boss_log.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    local value = bank.bank_data.remove_ad.value
    local key = bank.bank_data.remove_ad.key

    bank.player[key] = value or false

    -- load raid key
    local value = bank.bank_data.raid_key_daily.value
    SetRaidKeyDailyCount(value or bank.RAID_KEY_DAILY_COUNT)
    local value = bank.bank_data.raid_key_extra.value
    SetRaidKeyExtraCount(value)

    local value = bank.bank_data.claimed_mails.value
    local key = bank.bank_data.claimed_mails.key
    local encoded_value = value
    bank.player[key] = encoded_value and json.decode(encoded_value) or {}

    if event_sessions == 0 then
        bank.InitializeEvent()
    else
        -- load multi-step user data
        bank.LoadEventMode()
        bank.LoadTowerWaveProgress()
        bank.LoadEventHeroes() 
        bank.LoadEventTowers()
        bank.LoadEventTalents()
        bank.LoadSystemState()
        bank.LoadUsername()
        bank.LoadCalendarData()
        bank.LoadHellCoinShopData()
        bank.LoadAdData()
        bank.LoadTimers()
        bank.LoadPlayerDeckData()
        bank.LoadPurchaseData()
        bank.LoadFirstTimeBonus()
        bank.LoadTests()
        bank.LoadMonsterData()
        bank.LoadEventModeCards()
        bank.LoadEventItems()
       
        -- when player updates to 0.0.118, check for retroactively earned contracts
        bank.RetroTownContractCheck()

        -- send spend log on app launch to reduce cases of "losing" spend logs
        local trigger = "App Launch"
        AttemptPushExistingSpendLog(trigger)
    end
end

-- SECTION Log property and log event
-- Shared wave event properties
function WaveStartProps()
    local rating = wave.data.GetCurrent()
    local wave_mode = wave.data.mode
    local monster = wave.data.GetDifficulty(rating)
    local name = wave.data.name
    local seasonal_colony_index
    if wave.data.seasonal_colony then
        wave_mode = "seasonal colony"
        seasonal_colony_index = wave.data.seasonal_colony_index - #levels.colony
    end

    LogUserProperty_GearEquipped()

    return {
        Name = name,
        Max_wave = rating,
        Mode = wave_mode,
        Seasonal_colony = seasonal_colony_index,
        PlayerLevel = GetPlayerLevel(),
        CastleLevel = GetCastleLevel(),
        ArcherLevel = GetArcherLevel(),
        Deck = data.PlayerDeckString(player, GetCastleLevel()),
        AutoBattle = player.autobattle.is_active and 1 or 0,
        AutoBattleType = player.autobattle.type,
        Speed = player.speed,
        CurrentGold = GetPlayerGold(),
        PlayerGainGold = tostring(100 * (1 + player.attributes.monster_gold + player.attributes.all_gold)) .. "%",
        PlayerGainXP = tostring(100 * (1 + player.attributes.monster_xp)) .. "%",
        IsBossWave = wave.data.IsBossWave(rating),
        MonsterHealth = monster.health,
        MonsterAttack = monster.attack,
    }
end

function WaveEndProps()
    local props = WaveStartProps()

    props.ElapsedDuration = wave.data.elapsed
    props.SkillCasts = wave.data.skills
    return props
end

-- Log events and user props
local LogEvents = {
    logging = true,
    wave_cap = 400,
    ad_cap = 20,
    OpenGame = function(from_notification, from_change_focus)
        return {
            name = "Open Game",
            props = {
                OSDay = player.os_days,
                FromNotification = from_notification and 1 or 0,
                FromFocus = from_change_focus and 1 or 0,
            },
        }
    end,
    UnitUpgrade = function(unit_data)
        local root_data = unit_data:GetParentData() or unit_data
        local unit_type = root_data:GetType()
        local upgrade_type = (unit_data.level <= 1) and "Purchase" or "Upgrade"
        unit_type = unit_type:sub(1, 1):upper() .. unit_type:sub(2)

        return {
            name = unit_type .. " " .. upgrade_type,
            props = {
                Name = root_data.name,
                Class = unit_data.name,
                Level = math.max(root_data.level, 1),
                Wave = wave.data.GetMax(),
            },
        }
    end,
    HeroSwitchClass = function(old_unit, new_unit)
        local root_data = old_unit:GetParentData() or old_unit
        local type = ""
        local can_promote = false
        if old_unit.promote then
            for _, name in ipairs(old_unit.promote.unit_data) do
                if name == new_unit.name then
                    can_promote = true
                end
            end
        end
        if old_unit:CanDemoteTo(new_unit) then
            type = "reset"
        elseif can_promote then
            type = "promote"
        else
            type = "switch"
        end
        -- type: 1.promote, 2.switch, 3.reset
        return {
            name = "Hero Switch Class",
            props = {
                Name = root_data.name,
                Class = new_unit.name,
                type = type,
            },
        }
    end,
    new_user = {
        name = "Initialize New User",
        Props = function()
            return {
                LoadTime = DCEI.GetLevelDataLoadingTime(),
            }
        end,
    },
    wave_start = {
        name = "Stage Start",
        use_wave_cap = true,
        Props = WaveStartProps,
    },
    wave_replay = {
        name = "Stage Replay",
        use_wave_cap = true,
        Props = WaveStartProps,
    },
    wave_win = {
        name = "Stage Win",
        use_wave_cap = true,
        Props = function()
            local props = WaveEndProps()
            props.FailedWaveAttempts = player.wave_attempts
            props.DropContract = wave.data.drop_contract and 1 or 0

            local stats = DCEI.PerformanceTracking.CollectStats()
            for k, v in pairs(stats) do
                -- add prefix to performance stats to group them in amplitude so they're easier to read
                local key = "_" .. k
                props[key] = v
            end

            return props
        end,
    },
    wave_lose = {
        name = "Stage Lose",
        use_wave_cap = true,
        Props = WaveEndProps,
    },
    hell_mode_end = {
        name = "Hell Mode End",
        Props = function()
            local props = {
                PlayerLevel = GetPlayerLevel(),
                ParagonLevel = GetParagonLevel(),
                Deck = data.PlayerDeckString(player, GetCastleLevel()),
                Score = GetHellModeScore(),
                cum_count = tostring(player.hell_mode.count),
            }

            for _, v in ipairs(data.hell_mode.attributes) do
                props[v.name] = v.item.value
            end

            return props
        end,
    },
    wave_exit = {
        name = "Stage Exit",
        use_wave_cap = true,
        Props = WaveEndProps,
    },
    first_raid_boss_win = {
        name = "First Raid Boss Win",
        Props = function()
            local props = WaveEndProps()
            props.FailedAttempts = player.raid_boss_log.failed_attempts
            return props
        end,
    },
    dev_mode_enabled = {
        name = "Enable Developer Mode",
        props = {},
    },
    watch_ad_click = function(type, is_victory, gold_value)
        return {
            name = "Watch Ad Click",
            props = {
                Type = type,
                Max_wave = wave.data.GetCurrent(),
                Mode = wave.data.mode,
                IsVictory = is_victory and 1 or 0,
                GoldValue = gold_value or 0,
            },
        }
    end,
    watch_ad_failed = function(type, is_victory, gold_value)
        return {
            name = "Watch Ad Failed",
            props = {
                Type = type,
                Max_wave = wave.data.GetCurrent(),
                Mode = wave.data.mode,
                IsVictory = is_victory and 1 or 0,
                GoldValue = gold_value or 0,
            },
        }
    end,
    watch_ad = function(type, is_victory, gold_value)
        return {
            name = "Watch Ad",
            props = {
                Type = type,
                Max_wave = wave.data.GetCurrent(),
                Mode = wave.data.mode,
                IsVictory = is_victory and 1 or 0,
                GoldValue = gold_value or 0,
            },
        }
    end,
    review_accepted = {
        name = "Review Accepted",
        props = {},
    },
    social_menu_open = {
        name = "Social Menu Open",
        props = {},
    },
    wild_sky_clickthrough = {
        name = "WildSky Clickthrough",
        props = {},
    },
    -- seasonpass_clickthrough = function()
    --     local time_sec = player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() or 0
    --     return {
    --         name = "Season Pass Clickthrough",
    --         props = {
    --             SeasonPassTime = math.ceil( time_sec/(3600*24) )
    --         }
    --     }
    -- end,
    BugTracking_CreateHeroPromoteList = function(unit_data_name, list_show_time, list_close_time)
        return {
            name = "BugTracking CreateHeroPromoteList",
            props = {
                unit_data_name = unit_data_name or "invalid",
                list_show_time = list_show_time or "invalid",
                list_close_time = list_close_time or "invalid",
            },
        }
    end,
    BuyItem = function(item_data, items, start_spend_log)
        items = items or {}
        return {
            name = "Buy Item",
            items = items,
            props = {
                item_id = item_data.product_id,
                type = item_data.type,
                subtype = item_data.subtype or item_data.type,
                price = string.format("%.2f", item_data:GetPriceData().usd or 0),
                purchased_gold = items.gold,
                purchased_token = items.tokens,
                purchased_chest = items.chests,
                purchased_raid_turbo = items.raid_turbo,
                purchased_raid_key = items.raid_key,
                purchased_mine_contract = items.mine_contract,
                purchased_autobattle = items.autobattle and (items.autobattle .. " hours"),
                purchased_speedup = items.speedup and (items.speedup .. " days"),
                source = ui.shop.source or "none",
                max_wave = wave.data.GetMax(),
                wave_attempts = tostring(player.wave_attempts),
                purchase_count = tostring(player.purchase_count),
                purchase_spend = tostring(player.purchase_spend),
                daily_rewards_claimed = player.calendar and player.calendar.total_rewards_claimed or 0,
                first_time = items.first_time and 1 or 0,
            },
            start_spend_log = start_spend_log,
        }
    end,
    BuyHellCoinShop = function(item_name)
        return {
            name = "Buy Hell Coin Shop",
            props = {
                item_name = item_name,
            },
        }
    end,
    open_first_common_chest = {
        name = "Open 1st Common Chest",
        props = {
            max_wave = wave.data.GetMax(),
        },
    },
    FirstOpenMonsterLab = function()
        return {
            name = "Monster Lab 1st Open Menu",
            props = {
                max_wave = wave.data.GetMax(),
            },
        }
    end,
    UnlockMonsterInfo = function(monster)
        return {
            name = "Monster Lab 1st Unlock and Open Card",
            props = {
                max_wave = wave.data.GetMax(),
                monster_name = monster.name,
                token_owned = string.format("%d", GetPlayerGems()),
            },
        }
    end,
    PurchaseMonsterWeak = function(monster, wp_id, total_spend)
        return {
            name = "Monster Lab Buy Weakpoint",
            props = {
                max_wave = wave.data.GetMax(),
                monster_name = monster.name,
                id = wp_id,
                token_spend = total_spend,
                token_owned = string.format("%d", GetPlayerGems()),
            },
        }
    end,
    FirstPlayHellMode = function()
        return {
            name = "Hell Mode First Play",
            props = {
                max_wave = wave.data.GetMax(),
                score = GetHellModeScore(),
            },
        }
    end,
    SpinEggHuntGift = function(total_open)
        return {
            name = "Spin Wheel Claim Reward",
            props = {
                cum_spin_count = total_open,
            },
        }
    end,
    MonsterBountyStart = function(id)
        return {
            name = "MonsterBounty Start",
            props = {
                wave_trigger = id,
                wave = wave.data.GetMax(),
            },
        }
    end,
    MonsterBountyClaim = function(id)
        return {
            name = "MonsterBounty Claim",
            props = {
                wave_trigger = id,
                wave = wave.data.GetMax(),
            },
        }
    end,
    ClaimHellModeReward = function(season_name, reward_index, total_score)
        return {
            name = "Hell Mode Claim Reward",
            props = {
                season_name = season_name,
                milestone = reward_index,
                total_score = total_score,
            },
        }
    end,
    LogClaimSeasonEventReward = function(event_name, type, index)
        return {
            name = "Season Event Claim Reward",
            props = {
                event_name = event_name,
                type = type,
                milestone = index or 0,
            },
        }
    end,
    SeasonEventEnds = function(event_name, event_level, buy_pass)
        return {
            name = "Season Event Ends",
            props = {
                event_name = event_name,
                event_level = event_level,
                buy_pass = buy_pass and 1 or 0,
            },
        }
    end,
    ElementalTrailStart = {
        name = "Elemental Trial Start",
        props = {
            wave = wave.data.GetMax(),
            cum_count = player.element_trial_start,
        },
    },
    ElementalTrailComplete = {
        name = "Elemental Trial Complete",
        props = {
            event = event,
            wave = wave.data.GetMax(),
            cum_count = player.element_trial_complete,
        },
    },
    DevilTrailEvent = function(event_type, cum_count, nerf_name, buff_name)
        return {
            name = "Devil Trail " .. event_type,
            props = {
                max_wave = wave.data.GetMax(),
                cum_count = cum_count,
                nerf_name = nerf_name,
                buff_name = buff_name,
            },
        }
    end,
    CraftEquipment = function(rarity)
        return {
            name = "Craft Equipment",
            props = {
                rarity = rarity,
                max_wave = wave.data.GetMax(),
                cum_count = tostring(player.craft_rarity[rarity]),
            },
        }
    end,
    RefreshCraft = function(currency)
        return {
            name = "Refresh Craft",
            props = {
                currency = currency,
                max_wave = wave.data.GetMax(),
                cum_count = tostring(player.craft_refresh_currency[currency]),
            },
        }
    end,
    EnchantEpicAttribute = {
        name = "Enchant Epic Attribute",
        props = {
            wave = wave.data.GetMax(),
            normal_count = tostring(player.enchant_non_epic or 0),
            epic_count = tostring(player.enchant_epic or 0),
        },
    },
    PurchaseMonsterWeakPoint = {
        name = "Buy Monster Weak Point",
        props = {
            wave = wave.data.GetMax(),
            total_spend = tostring(player.monster_lab_token),
        },
    },
    UpgradeCraft = function(craft_level)
        return {
            name = "Upgrade Craft",
            props = {
                wave = wave.data.GetMax(),
                craft_level = craft_level,
            },
        }
    end,
    PurchaseRaidKey = {
        name = "Purchase Raid Key",
        props = {
            wave = wave.data.GetMax(),
            today_count = tostring(player.raid_key_extra_purchase_count),
            total_count = tostring(player.raid_key_extra_purchase_count_total),
            total_spend = tostring(player.raid_key_extra_purchase_count_spend),
        },
    },
    RaceEvent = function(event_type, rank)
        local props = {
            max_wave = wave.data.GetMax(),
            cum_count = player.race_event.cum_count or 1,
        }
        if event_type == "End" then
            props.rank = rank
        end
        return {
            name = "Race Event " .. event_type,
            props = props,
        }
    end,
}

local UserProps = {
    player_level = {
        name = "Player Level",
        Value = function()
            return GetPlayerLevel()
        end,
    },
    paragon_level = {
        name = "Paragon Level",
        Value = function()
            return GetParagonLevel()
        end,
    },
    Sessions = function(count)
        return {
            name = "Session Count",
            value = count,
        }
    end,
    LuaLoading = function(time)
        return {
            name = "Lua Loading Time",
            value = string.format("%.3f", time),
        }
    end,
    os_days = {
        name = "os_days",
        Value = function()
            return tostring(player.os_days)
        end,
    },
    wave_skip_attempt = {
        name = "Wave Skip Attempt",
        Value = function()
            return tostring(player.wave_skip_attempt)
        end,
    },
    wave_skip_success = {
        name = "Wave Skip Success Count",
        Value = function()
            return tostring(player.wave_skip_success)
        end,
    },
    wave_skip_failure = {
        name = "Wave Skip Failure Count",
        Value = function()
            return tostring(player.wave_skip_failure)
        end,
    },
    wave_skip_last = function(start_wave, skip_number, result)
        return {
            name = "Last Wave Skip",
            value = start_wave .. " + " .. skip_number .. " : " .. result,
        }
    end,
    last_loss = {
        name = "Last Wave Loss",
        Value = function()
            return wave.data.GetMax() + 1
        end,
    },
    last_exit = {
        name = "Last Wave Exit",
        Value = function()
            return wave.data.GetMax() + 1
        end,
    },
    raid_boss_attempts = {
        name = "RaidBoss Attempts",
        Value = function()
            return player.raid_boss_log.total_attempts
        end,
    },
    raid_boss_victories = {
        name = "RaidBoss Victories",
        Value = function()
            local total_victories = 0

            for _, boss in ipairs(levels.raid_boss) do
                local key = "Raid Boss " .. boss.index
                total_victories = total_victories + (player.raid_boss_log[key] or 0)
            end
            DCEI.LogMessage("total raid boss victories: " .. total_victories)
            return total_victories
        end,
    },
    raid_boss_victory = function(key, count)
        return {
            name = key .. " Victories",
            value = count,
        }
    end,
    ads_watched_gold = {
        name = "Gold Ads Watched",
        Value = function()
            return player.ads_watched_gold
        end,
    },
    ads_watched_speed = {
        name = "Speed Ads Watched",
        Value = function()
            return player.ads_watched_speed
        end,
    },
    player_xp = {
        name = "Total XP",
        Value = function()
            return math.floor(player.player_xp)
        end,
    },
    archer_level = {
        name = "Archer Level",
        Value = function()
            return GetArcherLevel()
        end,
    },
    castle_level = {
        name = "Castle Level",
        Value = function()
            return GetCastleLevel()
        end,
    },
    dwarf_level = {
        name = "Dwarf Worker Level",
        Value = function()
            return GetDwarfLevel()
        end,
    },
    mine_contract = {
        name = "Mine Contracts",
        Value = function()
            return GetMineContracts()
        end,
    },
    town_contract = {
        name = "Town Contracts",
        Value = function()
            return GetTownContracts()
        end,
    },
    town_level = {
        name = "Town Level",
        Value = function()
            return GetTownUpgradeLevel()
        end,
    },
    gold_owned = {
        name = "Currency Gold",
        Value = function()
            return GetPlayerGold()
        end,
    },
    vault_gold = {
        name = "Vault Gold",
        Value = function()
            return GetVaultGold()
        end,
    },
    vault_level = {
        name = "Vault Level",
        Value = function()
            return GetVaultLevel()
        end,
    },
    gem_owned = {
        name = "Currency Tokens",
        Value = function()
            return math.floor(DCEI.GetProtectedValue(data.safe.gems))
        end,
    },
    gem_overflow = {
        name = "Currency Tokens Overflow",
        Value = function()
            return math.floor(DCEI.GetProtectedValue(data.safe.gems_overflow))
        end,
    },
    elixir = {
        name = "Currency Elixir",
        Value = function()
            return GetPlayerElixir()
        end,
    },
    winter_coin = {
        name = "Winter Coin",
        Value = function()
            return GetPlayerWinterCoin()
        end,
    },
    spring_coin = {
        name = "Spring Coin",
        Value = function()
            return GetPlayerSpringCoin()
        end,
    },
    royal_coin = {
        name = "Royal Coin",
        Value = function()
            return GetPlayerRoyalCoin()
        end,
    },
    summer_coin = {
        name = "Summer Coin",
        Value = function()
            return GetPlayerSummerCoin()
        end,
    },
    autumn_coin = {
        name = "Autumn Coin",
        Value = function()
            return GetPlayerAutumnCoin()
        end,
    },
    winter_coin2022 = {
        name = "Winter Coin 2022",
        Value = function()
            return GetPlayerWinterCoin2022()
        end,
    },
    dcoin_spent = {
        name = "DCoin Spent",
        Value = function()
            return math.floor(DCEI.GetProtectedValue("dcoin_spent"))
        end,
    },
    dev_mode_enabled = {
        name = "Dev Mode Enabled",
        Value = function()
            return player.dev_mode and 1 or 0
        end,
    },
    max_wave = {
        name = "Max Wave",
        Value = function()
            return wave.data.GetMax()
        end,
    },
    heroes_owned = {
        name = "Heroes Owned",
        Value = data.GetOwnedHeroCount,
    },
    towers_owned = {
        name = "Towers Owned",
        Value = data.GetOwnedTowerCount,
    },
    defenses_owned = {
        name = "Defenses Owned",
        Value = data.GetOwnedDefenseCount,
    },
    gear_owned = {
        name = "Gear Owned",
        Value = data.GetOwnedGearCount,
    },
    gear_equipped = {
        name = "Gear Equipped",
        Value = function()
            local count = 0
            for _ in pairs(player.gear) do
                count = count + 1
            end
            return count
        end,
    },
    gear_upgrades = {
        name = "Gear Upgrades",
        Value = data.GetGearUpgradeCount,
    },
    gear_upgrade_gold = {
        name = "Gear Upgrades Gold",
        Value = function()
            return tostring(gear_upgrade_gold)
        end,
    },
    -- gear_menu_click = {
    --     name = "Gear Menu Click",
    --     Value = function() return tostring(gear_menu_click) end
    -- },
    equipment_equipped = {
        name = "Equipment Equipped Heroes",
        Value = function()
            local count = 0
            for k, unit_data in ipairs(data.heroes.gallery) do
                if next(unit_data.equipment) then
                    count = count + 1
                end
            end
            return count
        end,
    },
    town_owned = {
        name = "Town Upgrades Owned",
        Value = data.GetOwnedTownCount,
    },
    TownOwnedWave = function(count)
        return {
            name = "TownPurchased " .. tostring(count),
            value = wave.data.GetMax(),
        }
    end,
    challenges = {
        name = "Challenges",
        Value = function()
            return #levels.challenge.complete
        end,
    },
    colonies = {
        name = "Colonies",
        Value = function()
            return #levels.colony.complete
        end,
    },
    colony_income = {
        name = "Colony Income",
        Value = function()
            return levels.colony:Income()
        end,
    },
    deck = {
        name = "Deck",
        Value = function()
            return data.PlayerDeckString(player, GetCastleLevel())
        end,
    },
    is_cheater = {
        name = "Cheater",
        value = 1,
    },
    wave_attempts = {
        name = "Current Failed Wave Attempts",
        Value = function()
            return player.wave_attempts
        end,
    },
    daily_rewards_claimed = {
        name = "Daily Rewards Claimed",
        Value = function()
            return player.calendar.total_rewards_claimed or 0
        end,
    },
    username = {
        name = "Username",
        Value = function()
            return tostring(player.username.name)
        end,
    },
    gem_auto_time = {
        name = "Gem AutoBattle Time",
        Value = function()
            return ui.FormatTimeHours(tonumber(player.gem_auto_time))
        end,
    },
    element_trial_start = {
        name = "Element Trial Start Count",
        Value = function()
            return tostring(player.element_trial_start)
        end,
    },
    element_trial_complete = {
        name = "Element Trial Complete Count",
        Value = function()
            return tostring(player.element_trial_complete)
        end,
    },
    devil_trial_trigger = {
        name = "Devil Trial Trigger Count",
        Value = function()
            return tostring(player.devil_trail.trigger_count)
        end,
    },
    devil_trial_accept = {
        name = "Devil Trial Accept Count",
        Value = function()
            return tostring(player.devil_trail.accept_count)
        end,
    },
    devil_trial_win = {
        name = "Devil Trial Win Count",
        Value = function()
            return tostring(player.devil_trail.success_count)
        end,
    },
    devil_trial_lose = {
        name = "Devil Trial Lose Count",
        Value = function()
            return tostring(player.devil_trail.lost_count)
        end,
    },
    replay_count = {
        name = "Wave Replay Count",
        Value = function()
            return tostring(player.replay_count)
        end,
    },
    wave_loss_count = {
        name = "Wave Loss Count",
        Value = function()
            return tostring(player.wave_loss_count)
        end,
    },
    colony_loss_count = {
        name = "Colony Loss Count",
        Value = function()
            return tostring(player.colony_loss_count)
        end,
    },
    purchase_count = {
        name = "Buy Item Count",
        Value = function()
            return tostring(player.purchase_count)
        end,
    },
    purchase_spend = {
        name = "Buy Item Spend",
        Value = function()
            return tostring(player.purchase_spend)
        end,
    },
    OpenLeaderboardCount = function(source)
        return {
            -- convenient to use player.shop_logging to store leaderboard count
            name = "Open Leaderboard Count",
            Value = function()
                return player.shop_logging["leaderboard"]
            end,
        }
    end,
    ShopEntryFrom = function(source)
        return {
            name = "Enter Shop " .. "(" .. tostring(source) .. ")",
            Value = function()
                return player.shop_logging[source]
            end,
        }
    end,
    TalentLevel = function(talent_data)
        return {
            name = "Talent " .. talent_data.id,
            value = talent_data.level,
        }
    end,
    UnitLevel = function(unit_data, level)
        local root_data = unit_data:GetParentData() or unit_data
        return {
            name = "Unit " .. root_data.name .. " Level",
            value = level or root_data.level,
        }
    end,
    UnitClass = function(unit_data)
        local root_data = unit_data:GetParentData() or unit_data
        return {
            name = "Unit " .. root_data.name .. " Class",
            value = unit_data.name,
        }
    end,
    UnitSkin = function(unit_data)
        local root_data = unit_data:GetParentData() or unit_data
        local skin = ""
        if unit_data.skins then
            for name, value in pairs(unit_data.skins) do
                if value == 2 then
                    skin = name
                end
            end
        end
        return {
            name = "Unit " .. root_data.name .. " Skin",
            value = skin,
        }
    end,
    equipment_upgrades = {
        name = "Equipment Upgrades",
        Value = function()
            return tostring(player.equipment_upgrades)
        end,
    },
    normal_material_count = {
        name = "Normal Material",
        Value = function()
            return GetMaterialCount("normal_material")
        end,
    },
    epic_material_count = {
        name = "Epic Material",
        Value = function()
            return GetMaterialCount("epic_material")
        end,
    },
    -- autobattle_open = {
    --     name = "AutoBattle Open",
    --     Value = function() return tostring(player.autobattle_open) end
    -- },
    -- autobattle_gold = {
    --     name = "AutoBattle Gold",
    --     Value = function() return tostring(player.autobattle_gold) end
    -- },
    -- autobattle_gem = {
    --     name = "AutoBattle Gem",
    --     Value = function() return tostring(player.autobattle_gem) end
    -- },
    HellModeRecord = function(season_name, total_score)
        return {
            name = "Hell Mode " .. season_name,
            value = tostring(total_score),
        }
    end,
    SeasonEventLevel = function(name, level)
        return {
            name = name .. " Level",
            value = tostring(level),
        }
    end,
}
-- !SECTION
bank.LogEvents = LogEvents
bank.UserProps = UserProps

-- SECTION STORE DATA and help functions
local store_propers = {}
if event_mode.IsActive() then
    store_propers.collections = {
        token_pack = {
            "com.wildsky.castle.iap.bundle.token.01",
            "com.wildsky.castle.iap.bundle.token.02",
        },
    }
else
    store_propers.collections = {
        -- DEALS
        remove_ad = {
            "com.wildsky.castle.iap.remove_ad.01",
        },
        mine_pack = {
            "com.wildsky.castle.iap.bundle.mine.01",
            "com.wildsky.castle.iap.bundle.mine.02",
        },
        starter_pack = {
            "com.wildsky.castle.iap.bundle.starter.02",
            "com.wildsky.castle.iap.bundle.starter.03",
        },
        autobattle_pack = {
            "com.wildsky.castle.iap.bundle.autobattle.new.02",
            "com.wildsky.castle.iap.bundle.autobattle.new.03",
        },
        token_pack = {
            "com.wildsky.castle.iap.bundle.token.01",
            "com.wildsky.castle.iap.bundle.token.02",
        },
        raid_turbo_pack = {
            "com.wildsky.castle.iap.bundle.raid_turbo.01",
        },
        -- BOOSTS
        boost_gold = {
            "com.wildsky.castle.iap.boost.gold.30",
            "com.wildsky.castle.iap.boost.gold.90x2",
        },
        boost_token = {
            "com.wildsky.castle.iap.boost.token.30",
            "com.wildsky.castle.iap.boost.gold.60x2",
        },
    }
end


-- these items get red pips the first time they are added to the store
local RED_PIP_PRODUCTS = {}
local RED_PIP_PRODUCT_LIST = {
    "com.wildsky.castle.iap.bundle.mine.01",
    "com.wildsky.castle.iap.bundle.mine.02",
    "com.wildsky.castle.iap.bundle.autobattle.02",
    "com.wildsky.castle.iap.bundle.autobattle.03",
    "com.wildsky.castle.iap.bundle.autobattle.new.02",
    "com.wildsky.castle.iap.bundle.autobattle.new.03",
    "com.wildsky.castle.iap.bundle.token.01",
    "com.wildsky.castle.iap.bundle.token.02",
}
for _, v in pairs(RED_PIP_PRODUCT_LIST) do
    RED_PIP_PRODUCTS[v] = true
end

local SHOP_SECTIONS = {
    gold = {
        {
            "com.wildsky.castle.iap.gold.01",
            "com.wildsky.castle.iap.gold.02",
            "com.wildsky.castle.iap.gold.03",
        },
        {
            "com.wildsky.castle.iap.gold.04",
            "com.wildsky.castle.iap.gold.05",
            "com.wildsky.castle.iap.gold.06",
        },
    },
    tokens = {
        {
            "com.wildsky.castle.iap.token.new.01",
            "com.wildsky.castle.iap.token.01",
            "com.wildsky.castle.iap.token.02",
        },
        {
            "com.wildsky.castle.iap.token.03",
            "com.wildsky.castle.iap.token.new.04",
            "com.wildsky.castle.iap.token.04",
        },
    },
}

function GetStoreCollectionData(key)
    return store_propers.collections[key]
end

function CheckFirstTimeBonus(product_id)
    return not player.shop_1st_time_bonus[product_id]
end

function GetShopGoldValue(value_multiplier)
    if store_propers.max_wave ~= wave.data.GetMax() then
        store_propers.max_wave = wave.data.GetMax()
        local min_wave_considered = 25
        local wave_considered = math.max(wave.data.GetMax(), min_wave_considered)
        local base_value = wave.data.GetGoldValue(wave_considered)
        store_propers.gold_base_value = base_value
    end

    local gold_value = value_multiplier * store_propers.gold_base_value
    gold_value = gold_value - (gold_value % 1000)

    gold_value = math.floor(gold_value)
    gold_value = math.max(gold_value, 0)
    -- round up 9,8, round down 1,2 if not the first digit
    local divisor_large = 10
    while gold_value / divisor_large > 1 do
        divisor_large = divisor_large * 10
    end
    local divisor = divisor_large / 100
    local first_two = math.floor(gold_value / divisor)
    local second = first_two % 10
    local first = math.floor(first_two / 10)
    if second >= 8 then
        first = first + 1
        gold_value = first * 10 * divisor
    elseif second <= 2 then
        gold_value = first * 10 * divisor
    else
        gold_value = first_two * divisor
    end

    return gold_value
end

function GenerateGoldValue(self)
    local items_data = self.item_data:GetItemsData()

    local gold_value = GetShopGoldValue(items_data.value_multiplier)

    local max_underflow = 1000000000
    local underflow = gold_value
    local overflow = 0

    -- store values above/below 1b separately due to int overflow issues in protected values
    if gold_value >= max_underflow then
        underflow = gold_value % max_underflow
        overflow = math.floor(gold_value / max_underflow)
    end

    local gold_key = items_data._gold and "_gold" or "gold"
    items_data[gold_key] = underflow
    items_data["overflow"] = overflow
    self.item_data:SetItemsData(items_data)
    if self.Update then
        self:Update()
    end
end

function GetStoreItemGold(item_data)
    local gold_value = item_data.gold or item_data._gold
    if item_data.overflow then
        local max_underflow = 1000000000
        gold_value = item_data.overflow * max_underflow + gold_value
    end

    return gold_value
end

function StoreMineBundleOnshow(self)
    GenerateGoldValue(self)
    local function FormatNumberToThousands(value)
        local formatted_value = "+"
            .. "<size=125%%>"
            .. string.format("%.1f", value / 1000)
            .. "</size>"
            .. DCEI.GetTranslationText("general/cost/k")
        return formatted_value
    end

    local store_badge = self.item_data.store_badge
    local dwarf_level = data.dwarf_workers.GetLevel()
    local contracts = self.item_data:GetItemsData().mine_contract

    local old_income = data.dwarf_workers:GetIncomePerMinute(dwarf_level)
    local new_income = data.dwarf_workers:GetIncomePerMinute(dwarf_level + contracts)

    local text = FormatNumberToThousands(new_income - old_income)
    store_badge = store_badge:gsub("VAL", text)
    DCEI.SetTextFrameText(self.TitleLabel, store_badge)
end

function UpdateRemoveAdsGoldDisplay(self)
    local bundle_items = self.item_data:GetItemsData()
    local gold_value = GetStoreItemGold(bundle_items)
    local text = ui.FormatGoldCost(gold_value, 142)
    DCEI.SetTextFrameText(self.GoldCount, "<color=yellow>" .. text)
end

function UpdateStoreGoldDisplay(self)
    local gold_value = GetStoreItemGold(self.item_data:GetItemsData())
    local text = ui.FormatGoldCost(gold_value, 142)
    DCEI.SetTextFrameText(self.TitleLabel, "<color=yellow>" .. text)
    local is_first_time = CheckFirstTimeBonus(self.item_data.product_id)
    if is_first_time then
        DCEI.SetFrameActive(self.first_bonus_badge_frame.Frame, true)
        DCEI.SetTextFrameText(self.first_bonus_badge_frame.ValueLabel, "+" .. text)
        if self.item_data.badge_bonus then
            DCEI.SetFrameActive(self.badge_frame.Frame, false)
        end
    else
        DCEI.SetFrameActive(self.first_bonus_badge_frame.Frame, false)
        if self.item_data.badge_bonus then
            DCEI.SetFrameActive(self.badge_frame.Frame, true)
        end
    end
end

function UpdateStoreTokenDisplay(self)
    local is_first_time = CheckFirstTimeBonus(self.item_data.product_id)
    if is_first_time then
        DCEI.SetFrameActive(self.first_bonus_badge_frame.Frame, true)
        if self.item_data.badge_bonus then
            DCEI.SetFrameActive(self.badge_frame.Frame, false)
        end
    else
        DCEI.SetFrameActive(self.first_bonus_badge_frame.Frame, false)
        if self.item_data.badge_bonus then
            DCEI.SetFrameActive(self.badge_frame.Frame, true)
        end
    end
end

function UpdateStoreMineBundleInfo(self)
    local text = self.item_data.info.title
    DCEI.SetTextFrameText(self.TitleLabel, text)

    if not self.BackgroundGlow then
        self.BackgroundGlow = DCEI.GetChildFrameById(self.Frame, "BackgroundGlow")
    end
    UpdateStoreBundleDisplay(self)
end

function UpdateStoreBundleDisplay(self)
    local bundle_items = self.item_data:GetItemsData()
    local gold_value = GetStoreItemGold(bundle_items)
    local text = ui.FormatGoldCost(gold_value, 142)
    DCEI.SetTextFrameText(self.bundle_items[1].Label, "<color=yellow>" .. text)

    local auto_battle = bundle_items.autobattle
    if auto_battle then
        local auto_battle_section = ui.shop.sections["auto_battle"]
        if not auto_battle_section.info_frame then
            local layout = gmui.Layout:New({
                parent = auto_battle_section.SectionStack,
                name = "Shop/Section/InfoFrame",
            })

            auto_battle_section.info_frame = layout
        end

        DCEI.TriggerAddTimerEventElapsed(function()
            local value = player.gem_auto_time

            if value and value > 0 then
                -- display AB time remaining
                local text = DCEI.GetTranslationText(
                    "shop/info/auto_battle",
                    { x = "<color=green>" .. ui.FormatTimeHours(value) .. "</color>" }
                ) .. "\n" .. DCEI.GetTranslationText("shop/info/auto_battle/consumed")

                DCEI.SetFrameActive(auto_battle_section.info_frame.Info, true)
                DCEI.SetTextFrameText(auto_battle_section.info_frame.InfoLabel, text)
            end
        end, 0, false, true)
    end

    local product_id = self.item_data.product_id
    local stock_remaining = shop.GetItemStockAvailable(product_id)

    if self.StockFrame then
        local CYAN = "<color=#00ffff>"
        local DARK_GRAY = "<color=#707070>"
        local color = (stock_remaining > 0 and CYAN) or DARK_GRAY
        local stock_text = self.StockLabel
        local text = DCEI.GetTranslationText(
            "shop/item/in_stock",
            { x = "<size=120%>" .. color .. stock_remaining .. "</color>" .. "</size>" }
        )
        DCEI.SetTextFrameText(stock_text, text)
    end

    if stock_remaining <= 0 then
        SetStoreBundleItemDisabled(self)
    end
end

function SetStoreBundleItemDisabled(self)
    -- disable the button and gray out background
    DCEI.SetButtonFrameEnable(self.Button, false)

    if DCEI.FrameExists(self.Border) then
        DCEI.SetFrameImageColor(self.Border, { r = 0.88, g = 0.88, b = 0.88, a = 1 })
    end

    if DCEI.FrameExists(self.Background) then
        DCEI.SetFrameImageGrayScale(self.Background, true)
    end

    if DCEI.FrameExists(self.BackgroundGlow) then
        DCEI.SetFrameActive(self.BackgroundGlow, false)
    end

    if DCEI.FrameExists(self.CrossLine) then
        DCEI.SetFrameActive(self.CrossLine, false)
    end

    if DCEI.FrameExists(self.FullCostLabel) then
        DCEI.SetFrameActive(self.FullCostLabel, false)
    end

    if DCEI.FrameExists(self.TimerFrame) then
        DCEI.SetFrameActive(self.TimerFrame, false)
    end

    for _, item in ipairs(self.bundle_items) do
        DCEI.SetFrameImage(item.Background, "ui_ingame_hub")
        DCEI.SetFrameImageColor(item.Background, { r = 0, g = 0, b = 0, a = 0.4 })
    end

    -- change cost label to "SOLD OUT", hide full price label
    local text = DCEI.GetTranslationText("shop/item/sold_out")
    DCEI.SetTextFrameText(self.CostLabel, text)
end

function CustomizeStoreBundle(self)
    -- check if localized price is USD
    -- <s> tag doesn't work in Arabic, need to create a delete line manually
    local full_price_text = "¥" .. self.item_data.custom.full_price
    local product_id = self.item_data.product_id
    local localized_price = shop.GetProductPriceLocalized(product_id)
    local default_price = "¥" .. shop.GetProductPriceUSD(product_id)

    -- add "USD" suffix to full price display if localized price exists and isn't USD
    if localized_price and localized_price ~= default_price then
        full_price_text = full_price_text .. "<size=50%>" .. "RMB"
    end

    local text = full_price_text
    DCEI.SetTextFrameText(self.FullCostLabel, text)
end

function CustomizeStoreBundleHuge(self)
    CustomizeStoreBundle(self)
    local guarantee_banner = gmui.Layout:New({
        parent = self.Background,
        frame = self.chest_guarantee,
    })
    DCEI.SetFrameActive(guarantee_banner.Frame, true)
    local item_data = self.item_data:GetItemsData()
    local chest_guarantee_text = "<color=#ffbb00>"
        .. DCEI.GetTranslationText("ui/menu/guaranteed")
        .. "</color>\n"
        .. "<color=#ff8b4d>"
        .. item_data.guarantee_legendary
        .. "x "
        .. DCEI.GetTranslationText("data/equipment/rarity/legendary")
        .. "</color>"
    DCEI.SetTextFrameText(guarantee_banner.GuaranteeLabel, chest_guarantee_text)
end

function CustomizeStoreBundleSpringPass(self)
    DCEI.SetFrameActive(self.FullCostLabel, false)
    DCEI.SetFrameActive(self.CrossLine, false)
    UpdateItemStockLimit(self)
end

function UpdateStoreSpringPassInfo(self)
    DCEI.SetFrameVerticalOffsetInParent(self.InfoContainer, 160)
end

function CustomizeStoreItem_Boost(self)
    if self.section ~= ui.popup and self.item_data.custom.wide_frame then
        -- adjust width in shop
        DCEI.SetFrameSize(self.Background, 400, 370)
        DCEI.SetFrameSize(self.InfoButton, 400, 312)
    end
    -- add section timer display
    local text = DCEI.GetTranslationText("shop/section/daily_boosts/refresh")

    -- create new timer display for section if one doesn't exist
    if not self.section.timer_display then
        NewShopSectionTimerDisplay(self, text)
    end

    -- hookup stock limit
    UpdateItemStockLimit(self)

    -- banner title
    local text = self.item_data.label
    DCEI.SetTextFrameText(self.TitleLabel, "<line-height=88%>" .. text)

    -- set subtitle
    local text = self.item_data.custom.subtitle_func and self.item_data.custom.subtitle_func()
        or self.item_data.custom.subtitle
    text = text:gsub("%[%[", "<color=#f17aed>")
    text = text:gsub("%]%]", "</color>")
    text = text:gsub("%[", "<color=yellow>")
    text = text:gsub("%]", "</color>")
    DCEI.SetTextFrameText(self.DescriptionLabel, "<line-height=88%>" .. text)

    -- set boost display
    local Item = gmui.Layout:New({
        parent = self.Button,
        frame = self.boost_item.Frame,
    })

    HookupBoostItemDisplay(Item, self.item_data)
end

function CustomizeStoreItem_PercentBadge(self)
    DCEI.SetFrameActive(self.badge_frame.Frame, true)
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.badge_frame.Frame,
    })
    local value = self.item_data.badge_bonus * 100
    local text_value = string.format("%d", value) .. "%"
    local text = text_value .. "\n" .. "<size=75%>" .. DCEI.GetTranslationText("shop/item/more_badge")
    DCEI.SetTextFrameText(Item.BadgeLabel, "<line-height=66%>" .. text)
    self.badge_frame = Item
end

function CustomizeStoreItem_NumberBadge(self)
    DCEI.SetFrameActive(self.badge_frame.Frame, true)
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.badge_frame.Frame,
    })
    local value = self.item_data.badge_bonus
    local text_value = string.format("%d", value)
    local text = text_value .. "\n" .. "<size=75%>" .. DCEI.GetTranslationText("shop/item/bonus_badge")
    DCEI.SetTextFrameText(Item.BadgeLabel, "<line-height=66%>" .. text)
    self.badge_frame = Item
end

function CustomizeStoreItem_Gold(self)
    if self.item_data.badge_bonus then
        CustomizeStoreItem_PercentBadge(self)
    end
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.first_bonus_badge_frame.Frame,
    })
    DCEI.SetFrameImage(Item.ValueIcon, "icon_coin00")
    self.first_bonus_badge_frame = Item
end

function CustomizeStoreItem_Mithril(self)
    if self.item_data.badge_bonus then
        CustomizeStoreItem_PercentBadge(self)
    end
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.first_bonus_badge_frame.Frame,
    })
    DCEI.SetFrameImage(Item.ValueIcon, "icon_token_awaken01")
    self.first_bonus_badge_frame = Item
end

function CustomizeStoreItem_Token(self)
    if self.item_data.badge_bonus then
        CustomizeStoreItem_NumberBadge(self)
    end
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.first_bonus_badge_frame.Frame,
    })
    DCEI.SetFrameImage(Item.ValueIcon, "icon_coin_dungeon00")
    local value = self.item_data:GetItemsData().token
    DCEI.SetTextFrameText(Item.ValueLabel, "+" .. value)
    self.first_bonus_badge_frame = Item
    UpdateStoreTokenDisplay(self)
end

function CustomizeStoreItem_PowerPotion(self)
    if self.item_data.badge_bonus then
        CustomizeStoreItem_NumberBadge(self)
    end
    local Item = gmui.Layout:New({
        parent = self.Border,
        frame = self.first_bonus_badge_frame.Frame,
    })
    DCEI.SetFrameImage(Item.ValueIcon, "icon_catalyst_main_item5_orange_03")
    local value = self.item_data:GetItemsData().potion
    DCEI.SetTextFrameText(Item.ValueLabel, "+" .. value)
    self.first_bonus_badge_frame = Item
    UpdateStoreTokenDisplay(self)
end

function CustomizeStoreItem_Chest(self)
    DCEI.SetFrameMaxSize(self.TitleLabel, 200, 30)
    DCEI.SetTextFrameFontSizeMax(self.TitleLabel, 26)

    DCEI.SetFrameSize(self.BackgroundGlow, 260, 260)
    DCEI.SetFrameSize(self.ItemImage, 150, 150)
    DCEI.SetFrameVerticalOffsetInParent(self.ItemImage, 6)

    if self.item_data.display_guarantee then
        local guarantee_banner = gmui.Layout:New({
            parent = self.ItemContainer,
            name = "Shop/Section/Item/ChestGuarantee",
        })
        local text = ""
        local item_data = self.item_data:GetItemsData()
        if item_data.guarantee_legendary then
            text = "<color=#ffbb00>"
                .. DCEI.GetTranslationText("ui/menu/guaranteed")
                .. "</color>\n"
                .. "<color=#ff8b4d>"
                .. item_data.guarantee_legendary
                .. "x "
                .. DCEI.GetTranslationText("data/equipment/rarity/legendary")
                .. "</color>"
        elseif item_data.guarantee_epic then
            text = "<color=#ffbb00>"
                .. DCEI.GetTranslationText("ui/menu/guaranteed")
                .. "</color>\n"
                .. "<color=#be5cff>"
                .. item_data.guarantee_epic
                .. "x "
                .. DCEI.GetTranslationText("data/equipment/rarity/epic")
                .. "</color>"
        end
        DCEI.SetTextFrameText(guarantee_banner.GuaranteeLabel, text)
    end

    local number_text = DCEI.CreateTextFrame(self.Background)
    DCEI.SetFrameVerticalOffsetInParent(number_text, -50)
    DCEI.SetFrameHorizontalOffsetInParent(number_text, 70)
    DCEI.SetTextFrameFontSize(number_text, 36)
    local number = self.item_data:GetItemsData().chests
    DCEI.SetTextFrameText(number_text, "<color=#ffbb00>x" .. number)

    if self.item_data.badge_bonus then
        CustomizeStoreItem_PercentBadge(self)
    end
end

function CustomizeStoreItem_EventChest(self)
    DCEI.SetFrameMaxSize(self.TitleLabel, 200, 30)
    DCEI.SetTextFrameFontSizeMax(self.TitleLabel, 26)

    DCEI.SetFrameSize(self.BackgroundGlow, 260, 260)
    DCEI.SetFrameSize(self.ItemImage, 150, 150)
    DCEI.SetFrameVerticalOffsetInParent(self.ItemImage, 6)

    local number_text = DCEI.CreateTextFrame(self.Background)
    DCEI.SetFrameVerticalOffsetInParent(number_text, -50)
    DCEI.SetFrameHorizontalOffsetInParent(number_text, 70)
    DCEI.SetTextFrameFontSize(number_text, 36)
    local number = self.item_data:GetItemsData().chests
    DCEI.SetTextFrameText(number_text, "<color=#ffbb00>x" .. number)

    if self.item_data.badge_bonus then
        CustomizeStoreItem_PercentBadge(self)
    end
end

function NewShopSectionTimerDisplay(self, text)
    local parent = self.section.SubtitleFrame
    DCEI.SetFrameActive(parent, true)

    local layout = gmui.Layout:New({
        parent = parent,
        name = "Shop/Section/TimerDisplay",
    })

    -- hookup parent layout reference, used to determine if timer is in shop or popup
    layout.section = self.section
    layout.time = -1
    layout.remaining_time = 0

    if text then
        DCEI.SetTextFrameText(layout.DisplayLabel, text)
    end

    -- functions
    function layout:Update()
        local timer_time = self.timer:GetTimeElapsed()
        local timer_time_seconds = math.floor(timer_time)

        if DCEI.FrameExists(self.Frame) then
            if self.remaining_time == 0 then
                self.remaining_time = ui.calendar.timer and ui.calendar.timer:GetTimeRemaining() or 0
                local text = ui.FormatTimeHM(self.remaining_time)
                DCEI.SetTextFrameText(self.TimerLabel, text)
            end
            -- use floored time so layout only updates once per second
            if timer_time_seconds > self.time then
                self.time = timer_time_seconds

                -- update time display
                self.remaining_time = ui.calendar.timer and ui.calendar.timer:GetTimeRemaining() or 0
                local text = ui.FormatTimeHM(self.remaining_time)
                DCEI.SetTextFrameText(self.TimerLabel, text)

                -- check if seconds value has increased since last tick as it wraps around to 86400 when timer expires
                if self.time_to_refresh and self.remaining_time > self.time_to_refresh then
                    -- CheckServerTime()
                    shop.AttemptGetServerTime()
                end
                self.time_to_refresh = self.remaining_time

                -- animate clock hand
                self.clock_hand_rotation = self.clock_hand_rotation or 0
                self.clock_hand_rotation = (self.clock_hand_rotation - 90) % 360

                local duration = 0.5
                local delay = 1
                local k1 = self.clock_hand_rotation + 90
                local k2 = self.clock_hand_rotation
                local ease = "OutBack"

                DCEI.AnimateFrameRotation(self.ClockHand, { z = k1 }, { z = k2 }, duration, ease)
            end
        else
            -- if frame no longer exists, destroy the timer function
            ui.shop:RemoveTimerUpdateFunction(self.timer_id)
            ui.popup:RemoveTimerUpdateFunction(self.timer_id)
        end
    end

    -- hookup timer
    local timer_id = self.section.name .. "_timer_display"
    layout.timer_id = timer_id

    -- determine if update timer should be added to popup or store layout
    if self.section == ui.popup then
        layout.timer = ui.popup.timer
        ui.popup:AddTimerUpdateFunction(timer_id, function()
            layout:Update()
        end)
    else
        layout.timer = ui.shop.timer
        ui.shop:AddTimerUpdateFunction(timer_id, function()
            layout:Update()
        end)
    end

    -- initialize
    self.section.timer_display = layout
    self.section.timer_display:Update()
end

function AttemptRemoveShopSectionTimerDisplay(self)
    if self.section.timer_display then
        self.section.timer_display:Destroy()
        self.section.timer_display = nil
    end
end

function CustomizeStoreItemInfo_Boost(self)
    -- update label colors
    local text = self.item_data.info.label_func and self.item_data.info.label_func() or self.item_data.info.label
    text = text:gsub("%[", "<color=yellow>")
    text = text:gsub("%]", "</color>")
    DCEI.SetTextFrameText(self.InfoLabel, text)

    -- create boost display
    local layout = gmui.Layout:New({
        parent = self.ItemContainer,
        name = "Shop/Section/Item/Boost/BoostItem",
    })

    -- update boost display
    HookupBoostItemDisplay(layout, self.item_data)

    -- increase item size
    local k = { x = 1.4, y = 1.4, z = 1.4 }
    DCEI.AnimateFrameScale(layout.Frame, k, k, 0, "Linear")
    DCEI.SetFrameVerticalOffsetInParent(layout.Frame, 12)

    -- update stock availability
    UpdateItemStockLimit(self)

    -- initialize
    self.boost_display = layout
end

function HookupBoostItemDisplay(layout, item_data)
    -- use bigger font if multiplier is a whole number
    local value = item_data.custom.boost.multiplier
    if value % 1 <= 0 then
        DCEI.SetTextFrameFontSizeMax(layout.BoostLabel, 36)
    end

    local type = item_data.custom.boost.type
    if type == "gold" then
        -- displays "1.5x GOLD"
        DCEI.SetFrameImage(layout.BoostIcon, "icon_coin00")

        local text = tostring(value + 1) .. "<size=75%>" .. "x"
        DCEI.SetTextFrameText(layout.BoostLabel, text)
    elseif type == "token" then
        -- displays "3x TOKEN"
        local icon = "icon_coin_dungeon00"
        DCEI.SetFrameImage(layout.BoostIcon, icon)

        local text = tostring(value + 1) .. "<size=75%>" .. "x"
        DCEI.SetTextFrameText(layout.BoostLabel, text)
    elseif type == "mine" then
        -- displays "SCROLL 25%"
        DCEI.SetFrameImage(layout.BoostIcon_Mine, "icon_catalyst_main_item2_orange_03")
        DCEI.SetFrameActive(layout.BoostIcon_Mine, true)
        DCEI.SetFrameActive(layout.BoostIcon, false)

        local text = string.format("+%d", (value * 100)) .. "<size=75%>" .. "%"
        DCEI.SetTextFrameText(layout.BoostLabel, text)
    end
end

function ResetInfoPopup(self)
    DCEI.SetFrameSpacing(self.BundleStack, 8)
    DCEI.SetFrameSize(self.Background, 592, 592)
    DCEI.SetFrameVerticalOffsetInParent(self.InfoContainer, 130)
    -- reset boost display
    if self.boost_display then
        self.boost_display:Destroy()
    end

    if self.drop_rate_display then
        DCEI.DestroyFrame(self.drop_rate_display)
        self.drop_rate_display = nil
    end

    DCEI.SetFrameActive(self.ItemCount, false)
    DCEI.SetFrameVerticalOffsetInParent(self.ItemImage, 0)
    DCEI.SetFrameHorizontalOffsetInParent(self.ItemImage, 0)
    DCEI.SetFrameMaxSize(self.InfoLabel, 480, 150)
    if DCEI.FrameExists(self.Background) then
        DCEI.SetFrameImageGrayScale(self.Background, false)
    end
    if self.BackgroundGlow then
        DCEI.SetFrameActive(self.BackgroundGlow, true)
    end
end

function UpdateItemStockLimit(self)
    local stock_remaining = shop.GetItemStockAvailable(self.item_data.product_id)
    local CYAN = "<color=#00ffff>"
    local DARK_GRAY = "<color=#707070>"
    local color = (stock_remaining > 0 and CYAN) or DARK_GRAY

    -- check update stock label
    if self.StockFrame then
        local stock_text = self.StockLabel
        local text = DCEI.GetTranslationText(
            "shop/item/in_stock",
            { x = "<size=120%>" .. color .. stock_remaining .. "</color>" .. "</size>" }
        )
        DCEI.SetTextFrameText(stock_text, text)
    end

    -- update item if out of stock
    if stock_remaining == 0 then
        -- disable button
        DCEI.SetButtonFrameEnable(self.Button, false)

        -- replace cost text with "sold out"
        local text = DCEI.GetTranslationText("shop/item/sold_out")
        DCEI.SetTextFrameText(self.CostLabel, text)

        -- set cost button grayscale
        if self.CostFrame then
            DCEI.SetFrameImageGrayScale(self.CostFrame, true)
        end
    end

    if self.StockFrame then
        DCEI.SetFrameActive(self.StockFrame, true)
    end
end

function UpdateStoreAutoBattleTimeDisplay(self)
    if not self.section.info_frame then
        local layout = gmui.Layout:New({
            parent = self.section.SectionStack,
            name = "Shop/Section/InfoFrame",
        })

        self.section.info_frame = layout
    end

    DCEI.TriggerAddTimerEventElapsed(function()
        local value = player.gem_auto_time

        if value and value > 0 then
            -- display AB time remaining
            local text = DCEI.GetTranslationText(
                "shop/info/auto_battle",
                { x = "<color=green>" .. ui.FormatTimeHours(value) .. "</color>" }
            ) .. "\n" .. DCEI.GetTranslationText("shop/info/auto_battle/consumed")

            DCEI.SetFrameActive(self.section.info_frame.Info, true)
            DCEI.SetTextFrameText(self.section.info_frame.InfoLabel, text)
        else
            -- hide info frame if no AB time
            DCEI.SetFrameActive(self.section.info_frame.Info, false)
        end
    end, 0, false, true)
end

function UpdateStoreSpeedupTimeDisplay(self)
    if not self.section.info_frame then
        local layout = gmui.Layout:New({
            parent = self.section.SectionStack,
            name = "Shop/Section/InfoFrame",
        })
        self.section.info_frame = layout
    end

    DCEI.TriggerAddTimerEventElapsed(function()
        UpdateShopSpeedUpInfo(self.section.info_frame)
    end, 0, false, true)
end

function CustomizeStoreItemInfo_Chest(self)
    DCEI.SetFrameSize(self.Background, 592, 650)
    DCEI.SetTextFrameText(self.TitleLabel, self.item_data.popup_title)
    DCEI.SetFrameMaxSize(self.InfoLabel, 480, 100)
    DCEI.SetFrameVerticalOffsetInParent(self.ItemImage, 150)
    DCEI.SetFrameHorizontalOffsetInParent(self.ItemImage, -160)
    DCEI.SetFrameActive(self.ItemCount, true)
    DCEI.SetTextFrameText(self.ItemCount, "<color=#ffbb00>x" .. self.item_data:GetItemsData().chests)
    if not self.BackgroundGlow then
        self.BackgroundGlow = DCEI.GetChildFrameById(self.Frame, "BackgroundGlow")
    end
    DCEI.SetFrameActive(self.BackgroundGlow, false)
    if self.drop_rate_display then
        DCEI.SetFrameActive(self.drop_rate_display, true)
    else
        local chest_type = self.item_data.chest_type
        local rarity_info = ui.item_data[chest_type]
        self.drop_rate_display = DCEI.CreateFrame(self.InfoLabelStack)
        DCEI.SetFrameMinSize(self.drop_rate_display, 180, 200)
        local drop_rate = ui.NewRewardDropRateFrame(self.drop_rate_display, rarity_info)
        --DCEI.SetFrameVerticalOffsetInParent(drop_rate, -20)
        DCEI.SetFrameTopAlignmentInParent(drop_rate, true)
    end
end

function CustomizeStoreItemInfo_EventChest(self)
    DCEI.SetFrameSize(self.Background, 592, 650)
    DCEI.SetTextFrameText(self.TitleLabel, "Event Chest")
    DCEI.SetFrameMaxSize(self.InfoLabel, 480, 100)
    DCEI.SetFrameVerticalOffsetInParent(self.ItemImage, 150)
    DCEI.SetFrameHorizontalOffsetInParent(self.ItemImage, -160)
    DCEI.SetFrameActive(self.ItemCount, true)
    DCEI.SetTextFrameText(self.ItemCount, "<color=#ffbb00>x" .. self.item_data:GetItemsData().chests)
    if not self.BackgroundGlow then
        self.BackgroundGlow = DCEI.GetChildFrameById(self.Frame, "BackgroundGlow")
    end
    DCEI.SetFrameActive(self.BackgroundGlow, false)
    if self.drop_rate_display then
        DCEI.SetFrameActive(self.drop_rate_display, true)
    else
        local chest_type = self.item_data.chest_type
        local rarity_info = ui.item_data[chest_type]
        self.drop_rate_display = DCEI.CreateFrame(self.InfoLabelStack)
        DCEI.SetFrameMinSize(self.drop_rate_display, 180, 200)
        local drop_rate = event_mode.NewEventChestInfoFrame(self.drop_rate_display)
    end
end

function CreateHugeBundlePopup()
    local product_id = "com.wildsky.castle.iap.bundle.huge.01"
    if shop.GetItemStockAvailable(product_id) > 0 and not ui.shop.huge_bundle_popup then
        local popup = gmui.Layout:New({
            parent = ui.shop.Frame,
            name = "Portrait/Shop/BundleHugePopup",
        })
        DCEI.SetFrameVerticalOffsetInParent(popup.Background, -140)

        DCEI.SetTextFrameText(popup.TitleLabel, DCEI.GetTranslationText("shop/general/one_time"))
        local item = shop.GetProductData(product_id)
        local value = item.badge_bonus * 100
        local text_value = string.format("%d", value)
        DCEI.SetTextFrameText(popup.DiscountLabel, DCEI.GetTranslationText("shop/item/discount", { x = text_value }))

        local token_img = "shop_dungeontoken_lv6"
        DCEI.SetFrameImage(popup.TokenImage, token_img)
        local token_icon = "icon_coin_dungeon00"
        DCEI.SetFrameImage(popup.TokenIcon, token_icon)

        -- check if localized price is USD
        -- <s> tag doesn't work in Arabic, need to create a delete line manually
        local full_price_text = "¥" .. item.custom.full_price
        local localized_price = shop.GetProductPriceLocalized(product_id)
        local default_price = "¥" .. shop.GetProductPriceUSD(product_id)
        local price_text = localized_price or default_price

        -- add "USD" suffix to full price display if localized price exists and isn't USD
        if localized_price and localized_price ~= default_price then
            full_price_text = full_price_text .. "<size=50%>" .. "USD"
        end

        DCEI.SetTextFrameText(popup.FullCostLabel, full_price_text)
        DCEI.SetTextFrameText(popup.CostLabel, price_text)

        function popup:Close()
            if DCEI.FrameExists(self.Frame) then
                DCEI.DestroyFrame(self.Frame)
                ui.shop.huge_bundle_popup = nil
            end
        end

        DCEI.SetOnClickCallback(popup.CloseButton, function()
            popup:Close()
        end)

        DCEI.SetOnClickCallback(popup.Button, function()
            if shop.GetItemStockAvailable(product_id) > 0 then
                AttemptAnyItemPurchase(product_id)
                popup:Close()
            end
        end)

        local item_data = item:GetItemsData()
        local gold_value = GetStoreItemGold(item_data)
        local text = ui.FormatGoldCost(gold_value, 142)
        DCEI.SetTextFrameText(popup.GoldLabel, "<color=yellow>" .. text)

        local text = item_data.token
        DCEI.SetTextFrameText(popup.TokenLabel, "<color=#f17aed>" .. text)

        -- local text = item_data.autobattle.." "..DCEI.GetTranslationText("general/hours")
        -- DCEI.SetTextFrameText(popup.AutoBattleLabel, "<color=#ff6e6e>"..text)

        local text = item_data.legendary_chest
        DCEI.SetTextFrameText(popup.ChestLabel, "<color=#ffbb00>x" .. text)

        local guarantee_banner = gmui.Layout:New({
            frame = popup.chest_guarantee,
        })
        local chest_guarantee_text = "<color=#ffbb00>"
            .. DCEI.GetTranslationText("ui/menu/guaranteed")
            .. "</color>\n"
            .. "<color=#ff8b4d>"
            .. item_data.guarantee_legendary
            .. "x "
            .. DCEI.GetTranslationText("data/equipment/rarity/legendary")
            .. "</color>"
        DCEI.SetTextFrameText(guarantee_banner.GuaranteeLabel, chest_guarantee_text)

        ui.shop.huge_bundle_popup = popup
    end
    if ui.shop.huge_bundle_popup then
        DCEI.AnimateFrameScale(
            ui.shop.huge_bundle_popup.Background,
            { x = 0.5, y = 0.5 },
            { x = 1, y = 1 },
            0.5,
            "OutBack"
        )
    end
end

function CustomizeStoreItem_HellCoinShop(self)
    DCEI.SetFrameMinSize(self.Background, 185, 300)
    DCEI.SetFrameVerticalOffsetInParent(self.BackgroundGlow, 20)
    DCEI.SetFrameVerticalOffsetInParent(self.info_pip.Frame, -50)
    if self.item_data.stock_limit then
        UpdateItemStockLimit(self)
    end
end

function CustomizeStoreItem_HellCoinShop_Badge(self)
    CustomizeStoreItem_HellCoinShop(self)
    local limited_frame = DCEI.CreateFrame(self.Background)
    DCEI.SetFrameVerticalOffsetInParent(limited_frame, 80)
    DCEI.SetFrameImage(limited_frame, "ui_ingame_hub")
    DCEI.SetFrameImageColor(limited_frame, { r = 0, g = 0, b = 0, a = 0.6 })
    DCEI.SetFrameSize(limited_frame, 90, 32)
    local limited_text = DCEI.CreateTextFrame(limited_frame)
    DCEI.SetTextFrameFontSizeMax(limited_text, 18)
    DCEI.SetTextFrameFontSizeMin(limited_text, 8)
    DCEI.SetFrameMaxSize(limited_text, 80, 30)
    DCEI.SetTextFrameText(limited_text, DCEI.GetTranslationText("shop/item/limited", { x = "<color=yellow>1</color>" }))
end

function CustomizeStoreItem_HellCoinShop_RaidTurbo(self)
    CustomizeStoreItem_HellCoinShop(self)
    DCEI.SetFrameVerticalOffsetInParent(self.BackgroundGlow, 10)
end

function UpdateStoreHellCoinDisplay(self)
    DCEI.SetTextFrameText(self.section.hell_coin_frame.HellCoinCount, GetHellCoin())
    if self.item_data.stock_limit then
        UpdateItemStockLimit(self)
    end
end

function AnimateSparkle(frame)
    local k1, k2 = 0, 1
    local duration = core.Random(1, 2, 2)
    local ease = "OutQuart"
    local animation = DCEI.AnimateFrameAlpha(frame, k1, k2, duration, ease)
    DCEI.SetFrameAnimationLoops(animation, -1, "YoYo")

    local k1 = { x = 0.5, y = 0.5, z = 0.5 }
    local k2 = { x = 1.5, y = 1.5, z = 1.5 }
    local ease = "OutSine"
    local animation = DCEI.AnimateFrameScale(frame, k1, k2, duration, ease)
    DCEI.SetFrameAnimationLoops(animation, -1, "YoYo")

    local start_rotation = math.random(1, 360)
    local k1, k2 = { z = start_rotation }, { z = start_rotation - 90 }
    local ease = "Linear"
    local duration = 1
    local duration = core.Random(1.5, 2.5, 2)
    local animation = DCEI.AnimateFrameRotation(frame, k1, k2, duration, ease)
    DCEI.SetFrameAnimationLoops(animation, -1, "Incremental")
end

function CustomizeStoreSection_Deals(self)
    -- animate sparkles
    for n = 1, 6 do
        local key = "Sparkle0" .. n
        local sparkle = self[key]
        AnimateSparkle(sparkle)
    end
end

function CustomizeStoreSection_HellCoinShop(self)
    local parent = self.SubtitleFrame
    DCEI.SetFrameActive(parent, true)

    local layout = gmui.Layout:New({
        parent = parent,
        name = "Shop/Section/TimerDisplay",
    })

    local text = DCEI.GetTranslationText("shop/section/hell_coin_shop/rotation")
    DCEI.SetTextFrameText(layout.DisplayLabel, text)

    -- functions
    function layout:TimerUpdate(remaining_time)
        -- update time display
        local text = ui.FormatTimeDHM(remaining_time)
        DCEI.SetTextFrameText(self.TimerLabel, text)

        -- animate clock hand
        self.clock_hand_rotation = self.clock_hand_rotation or 0
        self.clock_hand_rotation = (self.clock_hand_rotation - 90) % 360

        local duration = 0.5
        local delay = 1
        local k1 = self.clock_hand_rotation + 90
        local k2 = self.clock_hand_rotation
        local ease = "OutBack"

        DCEI.AnimateFrameRotation(self.ClockHand, { z = k1 }, { z = k2 }, duration, ease)
    end

    -- initialize
    self.timer_display = layout
    -- self.timer_display:Update()

    local hell_coin_frame = gmui.Layout:New({
        parent = self.SubtitleFrame,
        name = "Shop/Section/HellCoinFrame",
    })
    DCEI.SetFrameMinSize(self.SubtitleFrame, 400, 60)
    DCEI.SetFrameVerticalOffsetInParent(self.timer_display.Frame, 40)
    -- DCEI.SetFrameVerticalOffsetInParent(hell_coin_frame.Frame, 0)
    self.hell_coin_frame = hell_coin_frame
end

local CURRENCIES = {
    {
        -- use "_" to make sure gold is always the first in bundles
        name = "_gold",
        icon = DCEI.Texture("icon_item_coin"),
    },
    {
        name = "token",
        icon = DCEI.Texture("icon_item_coin_dungeon"),
        bundle_label_format = "<color=#f17aed>{[x]}",
    },
    {
        name = "mine_contract",
        icon = DCEI.Texture("icon_catalyst_main_item2_orange_03"),
        bundle_label_format = "<color=green>x {[x]}",
    },
    {
        name = "autobattle",
        icon = DCEI.Texture("shop_autobattle_01"),
        bundle_label_format = "<color=green>" .. DCEI.GetTranslationText("general/hours_number"),
        bundle_layout = "Shop/Section/Item/BundleItemAutoBattle",
    },
    {
        name = "legendary_chest",
        icon = DCEI.Texture("premium_chest"),
        bundle_label_format = "<color=#ffbb00>x {[x]}",
    },
    {
        name = "raid_turbo",
        icon = DCEI.Texture("icon_guild_boss_key"),
    },
    {
        name = "raid_key",
        icon = DCEI.Texture("icon_mine_key"),
    },
    {
        name = "_remove_ad",
    },
    {
        name = "hell_coin",
        icon = DCEI.Texture("icon_shield_boss_gold"),
    },
    {
        name = "_1autumnpass",
        icon = DCEI.Texture("icon_castlepass_chest"),
        bundle_layout = "Portrait/Shop/Section/Item/AutumnPass",
    },
    {
        name = "_1winterpass",
        icon = DCEI.Texture("icon_castlepass_chest"),
        bundle_layout = "Portrait/Shop/Section/Item/WinterPass",
    },
}

local EVENT_CURRENCIES = {
    {
        -- use "_" to make sure gold is always the first in bundles
        name = "_gold",
        icon = DCEI.Texture("icon_token_awaken01"),
    },
    {
        name = "token",
        icon = DCEI.Texture("icon_item_coin_dungeon"),
        bundle_label_format = "<color=#f17aed>{[x]}",
    },
    {
        name = "potion",
        icon = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
        bundle_label_format = "<color=#f17aed>{[x]}",
    },
    {
        name = "mine_contract",
        icon = DCEI.Texture("icon_catalyst_main_item2_orange_03"),
        bundle_label_format = "<color=green>x {[x]}",
    },
    {
        name = "autobattle",
        icon = DCEI.Texture("shop_autobattle_01"),
        bundle_label_format = "<color=green>" .. DCEI.GetTranslationText("general/hours_number"),
        bundle_layout = "Shop/Section/Item/BundleItemAutoBattle",
    },
    {
        name = "legendary_chest",
        icon = DCEI.Texture("premium_chest"),
        bundle_label_format = "<color=#ffbb00>x {[x]}",
    },
    {
        name = "raid_turbo",
        icon = DCEI.Texture("icon_guild_boss_key"),
    },
    {
        name = "raid_key",
        icon = DCEI.Texture("icon_mine_key"),
    },
    {
        name = "_remove_ad",
    },
    {
        name = "hell_coin",
        icon = DCEI.Texture("icon_shield_boss_gold"),
    }
}

local store_product_list = {
    -- winter pass
    {
        product_id = "com.wildsky.castle.iap.winterpass.01",
        type = "winterpass",
        price = {
            usd = 9.99,
        },
        info = {
            title = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro"),
            label = DCEI.GetTranslationText(
                "ui/menu/spring_event/pass/description",
                { x = "<color=yellow>" .. DCEI.GetTranslationText("ui/menu/winter_event/pass/pro") .. "</color>" }
            ),
            update_func = UpdateStoreSpringPassInfo,
        },
        items = {
            _1winterpass = 1,
        },
        label = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro"),
        display = {
            use_item_bundle = true,
            use_stock_display = true,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/BundleHuge",
            layout_func = CustomizeStoreBundleSpringPass,
        },
    },
    -- autumn pass
    {
        product_id = "com.wildsky.castle.iap.autumnpass.01",
        type = "autumnpass",
        price = {
            usd = 9.99,
        },
        info = {
            title = DCEI.GetTranslationText("ui/menu/autumn_event/pass"),
            label = DCEI.GetTranslationText(
                "ui/menu/spring_event/pass/description",
                { x = "<color=yellow>" .. DCEI.GetTranslationText("ui/menu/autumn_event/pass") .. "</color>" }
            ),
            update_func = UpdateStoreSpringPassInfo,
        },
        items = {
            _1autumnpass = 1,
        },
        label = DCEI.GetTranslationText("ui/menu/autumn_event/pass"),
        display = {
            use_item_bundle = true,
            use_stock_display = true,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/BundleHuge",
            layout_func = CustomizeStoreBundleSpringPass,
        },
    },
    -- remove ads
    {
        product_id = "com.wildsky.castle.iap.remove_ad.01",
        type = "bundles",
        price = {
            usd = 9.99,
        },
        items = {
            _remove_ad = 1,
            _gold = 220,
            value_multiplier = 220,
        },
        display = {
            use_stock_display = true,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/BundleRemoveAds",
            layout_func = CustomizeStoreBundle,
            onshow_func = GenerateGoldValue,
            update_func = UpdateRemoveAdsGoldDisplay,
            full_price = 52,
        },
    },
    -- SEASON PASS
    {
        product_id = "com.wildsky.castle.iap.seasonpass.01",
        type = "seasonpass",
        price = {
            usd = 9.99,
        },
        items = {
            rewardx3 = 30 * 24 * 60 * 60, -- 30 days
        },
    },
    {
        product_id = "com.wildsky.castle.iap.seasonpass.02",
        type = "seasonpass",
        price = {
            usd = 4.99,
        },
        items = {
            rewardx3 = 14 * 24 * 60 * 60, -- 14 days
        },
    },
    -- MINE PACKS (MINE CONTRACT & GOLD) - WAVE 5
    {
        product_id = "com.wildsky.castle.iap.bundle.mine.01",
        type = "bundles",
        subtype = "mine_pack",
        display = {
            use_item_bundle = true,
        },
        info = {
            title = DCEI.GetTranslationText("shop/popup/mine_bundle"),
            label = DCEI.GetTranslationText("shop/mine_bundle/info/01") .. "\n\n" .. DCEI.GetTranslationText(
                "shop/mine_bundle/info/02"
            ),
            update_func = UpdateStoreMineBundleInfo,
        },
        store_badge = DCEI.GetTranslationText("shop/mine_bundle/description", { x = "VAL" }),
        items = {
            _gold = 80,
            value_multiplier = 80,
            mine_contract = 10,
        },
        price = {
            usd = 4.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = StoreMineBundleOnshow,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 52,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.mine.02",
        type = "bundles",
        subtype = "mine_pack",
        display = {
            use_item_bundle = true,
        },
        info = {
            title = DCEI.GetTranslationText("shop/popup/mine_bundle"),
            label = DCEI.GetTranslationText("shop/mine_bundle/info/01") .. "\n\n" .. DCEI.GetTranslationText(
                "shop/mine_bundle/info/02"
            ),
            update_func = UpdateStoreMineBundleInfo,
        },
        store_badge = DCEI.GetTranslationText("shop/mine_bundle/description", { x = "VAL" }),
        items = {
            _gold = 200,
            value_multiplier = 200,
            mine_contract = 30,
        },
        price = {
            usd = 9.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = StoreMineBundleOnshow,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 105,
        },
    },
    -- STARTER PACKS (GOLD & TOKEN) - WAVE 10
    {
        product_id = "com.wildsky.castle.iap.bundle.starter.02",
        type = "bundles",
        subtype = "starter_pack",
        label = DCEI.GetTranslationText("shop/general/most_popular"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 150,
            value_multiplier = 150,
            token = 20,
        },
        price = {
            usd = 4.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 52,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.starter.03",
        type = "bundles",
        subtype = "starter_pack",
        label = DCEI.GetTranslationText("shop/general/best_value"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 300,
            value_multiplier = 300,
            token = 100,
        },
        price = {
            usd = 9.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 87,
        },
    },
    -- AUTOBATTLE PACKS (GOLD & AUTO BATTLE) - WAVE 25
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.02",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>I</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 240,
            value_multiplier = 240,
            autobattle = 24,
        },
        price = {
            usd = 9.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 70,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.03",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>II</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 480,
            value_multiplier = 480,
            autobattle = 72,
        },
        price = {
            usd = 70,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 175,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.new.02",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>I</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 240,
            value_multiplier = 240,
            autobattle = 10,
        },
        price = {
            usd = 9.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 70,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.new.03",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>II</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 480,
            value_multiplier = 480,
            autobattle = 30,
        },
        price = {
            usd = 70,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 175,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.new.04",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>I</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 240,
            value_multiplier = 240,
            autobattle = 5,
        },
        price = {
            usd = 6.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 70,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.autobattle.new.05",
        type = "bundles",
        subtype = "autobattle_pack",
        label = DCEI.GetTranslationText("shop/popup/limited_time") .. " <size=120%>II</size>",
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 480,
            value_multiplier = 480,
            autobattle = 10,
        },
        price = {
            usd = 13.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 105,
        },
    },
    -- TOKEN PACKS (GOLD & TOKEN) - WAVE 100
    {
        product_id = "com.wildsky.castle.iap.bundle.token.01",
        type = "bundles",
        subtype = "token_pack",
        label = DCEI.GetTranslationText("shop/general/most_popular"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 60,
            value_multiplier = 60,
            token = 180,
        },
        price = {
            usd = 4.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 52,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.bundle.token.02",
        type = "bundles",
        subtype = "token_pack",
        label = DCEI.GetTranslationText("shop/general/best_value"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 150,
            value_multiplier = 150,
            token = 400,
        },
        price = {
            usd = 9.99,
        },
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/Bundle",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 87,
        },
    },
    -- One time offer big bundle
    {
        product_id = "com.wildsky.castle.iap.bundle.huge.01",
        type = "bundles",
        subtype = "big_pack",
        label = DCEI.GetTranslationText("shop/general/one_time"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 1680,
            value_multiplier = 1680,
            token = 2000,
            legendary_chest = 200,
            guarantee_legendary = 3,
        },
        price = {
            usd = 99.99,
        },
        badge_bonus = 0.3,
        stock_limit = 1,
        custom = {
            layout = "Portrait/Shop/Section/Item/BundleHuge",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundleHuge,
            update_func = UpdateStoreBundleDisplay,
            full_price = 455,
        },
    },
    -- raid turbo boost bundle
    {
        product_id = "com.wildsky.castle.iap.bundle.raid_turbo.01",
        type = "bundles",
        subtype = "raid_turbo_bundle",
        label = DCEI.GetTranslationText("shop/turbo_bundle/title"),
        display = {
            use_item_bundle = true,
        },
        items = {
            _gold = 2,
            value_multiplier = 2,
            raid_turbo = 150,
            raid_key = 5,
        },
        price = {
            usd = 6.99,
        },
        stock_limit = 2,
        custom = {
            layout = "Portrait/Shop/Section/Item/BundleHuge",
            onshow_func = GenerateGoldValue,
            layout_func = CustomizeStoreBundle,
            update_func = UpdateStoreBundleDisplay,
            full_price = 45,
        },
    },
    -- GOLD BOOSTS
    {
        product_id = "com.wildsky.castle.iap.boost.gold.30",
        type = "boosts",
        label = DCEI.GetTranslationText("shop/boost/title", { x = "<size=100%>" .. "30" .. "</size>" }),
        items = {
            gold_boost_200 = 1800,
        },
        price = {
            usd = 4.99,
        },
        stock_limit = 2,
        info = {
            label = DCEI.GetTranslationText("shop/boost/info/01", { x = "[200%] <sprite name=gold>" })
                .. "\n\n"
                .. DCEI.GetTranslationText("shop/boost/info/02"),
            update_func = CustomizeStoreItemInfo_Boost,
        },
        custom = {
            layout = "Shop/Section/Item/Boost",
            layout_func = CustomizeStoreItem_Boost,
            update_func = UpdateItemStockLimit,
            remove_func = AttemptRemoveShopSectionTimerDisplay,
            subtitle = DCEI.GetTranslationText(
                "shop/boost/description",
                { x = "[" .. DCEI.GetTranslationText("general/gold") .. "]" }
            ),
            boost = {
                type = "gold",
                multiplier = 1,
            },
        },
    },
    {
        product_id = "com.wildsky.castle.iap.boost.gold.60",
        type = "boosts",
        label = DCEI.GetTranslationText("shop/boost/title", { x = "<size=100%>" .. "60" .. "</size>" }),
        items = {
            gold_boost_150 = 3600,
        },
        price = {
            usd = 4.99,
        },
        stock_limit = 1,
        info = {
            label = DCEI.GetTranslationText("shop/boost/info/01", { x = "[150%] <sprite name=gold>" })
                .. "\n\n"
                .. DCEI.GetTranslationText("shop/boost/info/02"),
            update_func = CustomizeStoreItemInfo_Boost,
        },
        custom = {
            layout = "Shop/Section/Item/Boost",
            layout_func = CustomizeStoreItem_Boost,
            update_func = UpdateItemStockLimit,
            remove_func = AttemptRemoveShopSectionTimerDisplay,
            subtitle = DCEI.GetTranslationText(
                "shop/boost/description",
                { x = "[" .. DCEI.GetTranslationText("general/gold") .. "]" }
            ),
            boost = {
                type = "gold",
                multiplier = 0.5,
            },
        },
    },
    {
        product_id = "com.wildsky.castle.iap.boost.gold.90x2",
        type = "boosts",
        label = DCEI.GetTranslationText("shop/boost/title", { x = "<size=100%>" .. "90" .. "</size>" }),
        items = {
            gold_boost_200 = 5400,
        },
        price = {
            usd = 52,
        },
        stock_limit = 1,
        info = {
            label = DCEI.GetTranslationText("shop/boost/info/01", { x = "[200%] <sprite name=gold>" })
                .. "\n\n"
                .. DCEI.GetTranslationText("shop/boost/info/02"),
            update_func = CustomizeStoreItemInfo_Boost,
        },
        custom = {
            layout = "Shop/Section/Item/Boost",
            layout_func = CustomizeStoreItem_Boost,
            update_func = UpdateItemStockLimit,
            remove_func = AttemptRemoveShopSectionTimerDisplay,
            subtitle = DCEI.GetTranslationText(
                "shop/boost/description",
                { x = "[" .. DCEI.GetTranslationText("general/gold") .. "]" }
            ),
            boost = {
                type = "gold",
                multiplier = 1,
            },
        },
    },
    -- TOKEN BOOSTS
    {
        product_id = "com.wildsky.castle.iap.boost.token.30",
        type = "boosts",
        label = DCEI.GetTranslationText("shop/boost/title", { x = "<size=100%>" .. "30" .. "</size>" }),
        price = { usd = 4.99 },
        items = {
            token_boost_500 = 1800,
        },
        stock_limit = 2,
        info = {
            label_func = function()
                local icon = "[5x] <sprite name=ecoin>"
                return DCEI.GetTranslationText("shop/boost/info/01", { x = icon })
                    .. "\n\n"
                    .. DCEI.GetTranslationText("shop/boost/info/02")
            end,
            label = DCEI.GetTranslationText("shop/boost/info/01", { x = "[5x] <sprite name=ecoin>" })
                .. "\n\n"
                .. DCEI.GetTranslationText("shop/boost/info/02"),
            update_func = CustomizeStoreItemInfo_Boost,
        },
        custom = {
            layout = "Shop/Section/Item/Boost",
            layout_func = CustomizeStoreItem_Boost,
            update_func = UpdateItemStockLimit,
            remove_func = AttemptRemoveShopSectionTimerDisplay,
            subtitle_func = function()
                local text = "[" .. DCEI.GetTranslationText("general/tokens") .. "]"
                return DCEI.GetTranslationText("shop/boost/description", { x = text })
            end,
            subtitle = DCEI.GetTranslationText(
                "shop/boost/description",
                { x = "[" .. DCEI.GetTranslationText("general/tokens") .. "]" }
            ),
            boost = {
                type = "token",
                multiplier = 4,
            },
            wide_frame = true,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.boost.gold.60x2",
        type = "boosts",
        label = DCEI.GetTranslationText("shop/boost/title", { x = "<size=100%>" .. "60" .. "</size>" }),
        price = { usd = 9.99 },
        items = {
            gold_boost_200 = 3600,
        },
        stock_limit = 1,
        info = {
            label = DCEI.GetTranslationText("shop/boost/info/01", { x = "[200%] <sprite name=gold>" })
                .. "\n\n"
                .. DCEI.GetTranslationText("shop/boost/info/02"),
            update_func = CustomizeStoreItemInfo_Boost,
        },
        custom = {
            layout = "Shop/Section/Item/Boost",
            layout_func = CustomizeStoreItem_Boost,
            update_func = UpdateItemStockLimit,
            remove_func = AttemptRemoveShopSectionTimerDisplay,
            subtitle = DCEI.GetTranslationText(
                "shop/boost/description",
                { x = "[" .. DCEI.GetTranslationText("general/gold") .. "]" }
            ),
            boost = {
                type = "gold",
                multiplier = 1,
            },
            wide_frame = true,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.01",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv1"),
            scale = 0.9,
            offset = { x = 0, y = -16 },
        },
        price = {
            usd = 2.99,
        },
        items = {
            gold = 50,
            value_multiplier = 50,
        },
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.02",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv2"),
            offset = { x = 0, y = -16 },
        },
        price = {
            usd = 4.99,
        },
        items = {
            gold = 100,
            value_multiplier = 100,
        },
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.03",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv3"),
            offset = { x = 0, y = -14 },
        },
        price = {
            usd = 9.99,
        },
        items = {
            gold = 220,
            value_multiplier = 220,
        },
        badge_bonus = 0.1,
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.04",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv4"),
            scale = 0.96,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 70,
        },
        items = {
            gold = 480,
            value_multiplier = 480,
        },
        badge_bonus = 0.2,
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.05",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv5"),
            scale = 1.08,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 105,
        },
        items = {
            gold = 780,
            value_multiplier = 780,
        },
        badge_bonus = 0.3,
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.06",
        type = "gold",
        label = {
            icon = "icon_coin00",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv6"),
            scale = 1.25,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 175,
        },
        items = {
            gold = 1400,
            value_multiplier = 1400,
        },
        badge_bonus = 0.4,
        custom = {
            layout_func = CustomizeStoreItem_Gold,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.new.01",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv1"),
            scale = 0.96,
            offset = { x = 0, y = -20 },
        },
        price = {
            usd = 0.99,
        },
        items = {
            token = 30,
        },
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.01",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv2"),
            scale = 0.96,
            offset = { x = 0, y = -20 },
        },
        price = {
            usd = 2.99,
        },
        items = {
            token = 90,
        },
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.02",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv3"),
            scale = 1.08,
            offset = { x = 0, y = -14 },
        },
        price = {
            usd = 4.99,
        },
        items = {
            token = 180,
        },
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.03",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv4"),
            scale = 1.08,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 9.99,
        },
        items = {
            token = 400,
        },
        badge_bonus = 40,
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.new.04",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv5"),
            scale = 1.08,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 70,
        },
        items = {
            token = 1080,
        },
        badge_bonus = 360,
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.04",
        type = "token",
        label = {
            icon = "icon_coin_dungeon00",
            item_type = "token",
            color = "<color=#f17aed>",
        },
        display = {
            image = DCEI.Texture("shop_dungeontoken_lv6"),
            scale = 1.2,
            offset = { x = 0, y = -8 },
        },
        price = {
            usd = 105,
        },
        items = {
            token = 1620,
        },
        badge_bonus = 540,
        custom = {
            layout_func = CustomizeStoreItem_Token,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.01",
        type = "chests",
        label = {
            name = DCEI.GetTranslationText("data/equipment/rarity/rare"), --DCEI.GetTranslationText("data/item/chest/rare"),
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("water_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_Chest,
            label = DCEI.GetTranslationText("data/item/chest/description"),
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/rare"),
        price = {
            usd = 4.99,
        },
        chest_type = "rare_chest_l",
        items = {
            chests = 25,
            guarantee_epic = 1,
        },
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_Chest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.02",
        type = "chests",
        label = {
            name = DCEI.GetTranslationText("data/equipment/rarity/epic"),
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("dark_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_Chest,
            label = DCEI.GetTranslationText("data/item/chest/description")
                .. "\n<color=#ffbb00>"
                .. DCEI.GetTranslationText("ui/menu/guaranteed")
                .. "</color> "
                .. "<color=#be5cff>2x "
                .. DCEI.GetTranslationText("data/equipment/rarity/epic")
                .. "</color>",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/epic"),
        price = {
            usd = 9.99,
        },
        display_guarantee = true,
        chest_type = "rare_chest",
        items = {
            chests = 55,
            guarantee_epic = 2,
        },
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_Chest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.03",
        type = "chests",
        label = {
            name = DCEI.GetTranslationText("data/equipment/rarity/legendary") .. " I",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("fire_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_Chest,
            label = DCEI.GetTranslationText("data/item/chest/description")
                .. "\n<color=#ffbb00>"
                .. DCEI.GetTranslationText("ui/menu/guaranteed")
                .. "</color> "
                .. "<color=#ff8b4d>2x "
                .. DCEI.GetTranslationText("data/equipment/rarity/legendary")
                .. "</color>",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/legendary"),
        price = {
            usd = 105,
        },
        display_guarantee = true,
        chest_type = "legendary_chest_l",
        items = {
            chests = 200,
            guarantee_epic = 8,
            guarantee_legendary = 2,
        },
        badge_bonus = 0.2,
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_Chest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.04",
        type = "chests",
        label = {
            name = DCEI.GetTranslationText("data/equipment/rarity/legendary") .. " II",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("premium_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_Chest,
            label = DCEI.GetTranslationText("data/item/chest/description")
                .. "\n<color=#ffbb00>"
                .. DCEI.GetTranslationText("ui/menu/guaranteed")
                .. "</color> "
                .. "<color=#ff8b4d>5x "
                .. DCEI.GetTranslationText("data/equipment/rarity/legendary")
                .. "</color>",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/legendary") .. " II",
        price = {
            usd = 175,
        },
        display_guarantee = true,
        chest_type = "legendary_chest",
        items = {
            chests = 360,
            guarantee_epic = 15,
            guarantee_legendary = 5,
        },
        badge_bonus = 0.3,
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_Chest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.01",
        type = "autobattle",
        label = {
            store_value = 10,
            name = "10" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_01"),
            scale = 1,
        },
        price = {
            usd = 4.99,
        },
        items = {
            autobattle = 10 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
            -- only add update and onshow to this item to avoid duplicate update
            onshow_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.02",
        type = "autobattle",
        label = {
            store_value = 24,
            name = "24" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_02"),
            scale = 1.3,
        },
        price = {
            usd = 9.99,
        },
        items = {
            autobattle = 24 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.03",
        type = "autobattle",
        label = {
            store_value = 100,
            name = "100" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_03"),
            scale = 1.3,
        },
        price = {
            usd = 105,
        },
        items = {
            autobattle = 100 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.04",
        type = "autobattle",
        label = {
            store_value = 200,
            name = "200" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_04"),
            scale = 1.4,
        },
        price = {
            usd = 175,
        },
        items = {
            autobattle = 200 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.new.01",
        type = "autobattle",
        label = {
            store_value = 1,
            name = "1" .. " " .. DCEI.GetTranslationText("general/hour"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_01"),
            scale = 1.3,
        },
        price = {
            usd = 2.99,
        },
        items = {
            autobattle = 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
            -- only add update and onshow to this item to avoid duplicate update
            onshow_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.new.02",
        type = "autobattle",
        label = {
            store_value = 3,
            name = "3" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_02"),
            scale = 1.3,
        },
        price = {
            usd = 4.99,
        },
        items = {
            autobattle = 3 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.new.03",
        type = "autobattle",
        label = {
            store_value = 10,
            name = "10" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_03"),
            scale = 1.3,
        },
        price = {
            usd = 9.99,
        },
        items = {
            autobattle = 10 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.autobattle.new.04",
        type = "autobattle",
        label = {
            store_value = 50,
            name = "50" .. " " .. DCEI.GetTranslationText("general/hours"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_autobattle_04"),
            scale = 1.4,
        },
        price = {
            usd = 105,
        },
        items = {
            autobattle = 50 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreAutoBattleTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.speed.01",
        type = "speedup",
        label = {
            store_value = 1,
            name = "1" .. " " .. DCEI.GetTranslationText("general/day"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_speedup_01"),
            scale = 1.2,
        },
        price = {
            usd = 2.99,
        },
        items = {
            speedup = 1 * 24 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreSpeedupTimeDisplay,
            -- only add update and onshow to this item to avoid duplicate update
            onshow_func = UpdateStoreSpeedupTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.speed.02",
        type = "speedup",
        label = {
            store_value = 3,
            name = "3" .. " " .. DCEI.GetTranslationText("general/days"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_speedup_02"),
            scale = 1.2,
        },
        price = {
            usd = 4.99,
        },
        items = {
            speedup = 3 * 24 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreSpeedupTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.speed.03",
        type = "speedup",
        label = {
            store_value = 7,
            name = "7" .. " " .. DCEI.GetTranslationText("general/days"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_speedup_03"),
            scale = 1.2,
        },
        price = {
            usd = 9.99,
        },
        items = {
            speedup = 7 * 24 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreSpeedupTimeDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.speed.04",
        type = "speedup",
        label = {
            store_value = 30,
            name = "30" .. " " .. DCEI.GetTranslationText("general/days"),
            color = "<color=#ff6e6e>",
        },
        display = {
            image = DCEI.Texture("shop_speedup_04"),
            scale = 1.2,
        },
        price = {
            usd = 70,
        },
        items = {
            speedup = 30 * 24 * 60 * 60,
        },
        custom = {
            update_func = UpdateStoreSpeedupTimeDisplay,
        },
    },
    -- hell coin shop
    {
        product_id = "com.wildsky.castle.hell_coin.philosopher_stone.01",
        type = "hell_coin_refresher",
        label = DCEI.GetTranslationText("data/item/refresher"),
        display = {
            image = DCEI.Texture("icon_premium_token"),
            offset = { x = 0, y = 10 },
            use_stock_display = true,
        },
        price = {
            type = "hell_coin",
            amount = 100,
        },
        stock_limit = 1,
        items = {
            philosopher_stone = 1,
        },
        info = {
            title = DCEI.GetTranslationText("data/item/refresher"),
            label = DCEI.GetTranslationText("data/item/refresher/description"),
        },
        custom = {
            layout_func = CustomizeStoreItem_HellCoinShop,
            update_func = UpdateStoreHellCoinDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.hell_coin.raid_turbo.01",
        type = "hell_coin_raid_turbo",
        label = DCEI.GetTranslationText("data/item/raid_turbo"),
        display = {
            image = DCEI.Texture("icon_guild_boss_key"),
            scale = 0.8,
            offset = { x = 0, y = 18 },
        },
        price = {
            type = "hell_coin",
            amount = 10,
        },
        items = {
            raid_turbo = 10,
        },
        info = {
            title = DCEI.GetTranslationText("data/item/raid_turbo"),
            label = DCEI.GetTranslationText("data/item/raid_turbo/description"),
        },
        custom = {
            layout_func = CustomizeStoreItem_HellCoinShop_RaidTurbo,
            update_func = UpdateStoreHellCoinDisplay,
            -- only add update and onshow to this item to avoid duplicate update
            onshow_func = UpdateStoreHellCoinDisplay,
        },
    },
}

local store_product_list_event = {
    {
        product_id = "com.wildsky.castle.iap.gold.01",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv1"),
            scale = 0.9,
            offset = { x = 0, y = -16 },
        },
        price = {
            usd = 2.99,
        },
        items = {
            gold = 50,
            value_multiplier = 50,
        },
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.02",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv2"),
            offset = { x = 0, y = -16 },
        },
        price = {
            usd = 4.99,
        },
        items = {
            gold = 100,
            value_multiplier = 100,
        },
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.03",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv3"),
            offset = { x = 0, y = -14 },
        },
        price = {
            usd = 9.99,
        },
        items = {
            gold = 220,
            value_multiplier = 220,
        },
        badge_bonus = 0.1,
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.04",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv4"),
            scale = 0.96,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 70,
        },
        items = {
            gold = 480,
            value_multiplier = 480,
        },
        badge_bonus = 0.2,
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.05",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv5"),
            scale = 1.08,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 105,
        },
        items = {
            gold = 780,
            value_multiplier = 780,
        },
        badge_bonus = 0.3,
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.gold.06",
        type = "gold",
        label = {
            icon = "icon_token_awaken01",
            item_type = "gold",
            color = "<color=yellow>",
            format_with_commas = true,
            disable_rtl_support = true,
        },
        display = {
            image = DCEI.Texture("shop_gold_lv6"),
            scale = 1.25,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 175,
        },
        items = {
            gold = 1400,
            value_multiplier = 1400,
        },
        badge_bonus = 0.4,
        custom = {
            layout_func = CustomizeStoreItem_Mithril,
            onshow_func = GenerateGoldValue,
            update_func = UpdateStoreGoldDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.new.01",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -20 },
        },
        price = {
            usd = 0.99,
        },
        items = {
            potion = 30,
        },
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.01",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -20 },
        },
        price = {
            usd = 2.99,
        },
        items = {
            potion = 90,
        },
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.02",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -14 },
        },
        price = {
            usd = 4.99,
        },
        items = {
            potion = 180,
        },
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.03",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 9.99,
        },
        items = {
            potion = 400,
        },
        badge_bonus = 40,
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.new.04",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -12 },
        },
        price = {
            usd = 70,
        },
        items = {
            potion = 1080,
        },
        badge_bonus = 360,
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.token.04",
        type = "token",
        label = {
            icon = "icon_catalyst_main_item5_orange_03",
            item_type = "potion",
            color = "<color=yellow>",
        },
        display = {
            image = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            scale = 2,
            offset = { x = 0, y = -8 },
        },
        price = {
            usd = 105,
        },
        items = {
            potion = 1620,
        },
        badge_bonus = 540,
        custom = {
            layout_func = CustomizeStoreItem_PowerPotion,
            update_func = UpdateStoreTokenDisplay,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.01",
        type = "chests",
        label = {
            name = "Treasure Pile",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("nature_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_EventChest,
            label = "Open chests to get a random hero card",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/rare"),
        price = {
            usd = 4.99,
        },
        chest_type = "rare_chest_l",
        items = {
            chests = 25,
            guarantee_epic = 1,
        },
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_EventChest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.02",
        type = "chests",
        label = {
            name = "Treasure Trove",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("nature_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_EventChest,
            label = "Open chests to get a random hero card",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/epic"),
        price = {
            usd = 9.99,
        },
        display_guarantee = true,
        chest_type = "rare_chest",
        items = {
            chests = 55,
            guarantee_epic = 2,
        },
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_EventChest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.03",
        type = "chests",
        label = {
            name = "Treasure Hoard",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("nature_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_EventChest,
            label = "Open chests to get a random hero card",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/legendary"),
        price = {
            usd = 105,
        },
        display_guarantee = true,
        chest_type = "legendary_chest_l",
        items = {
            chests = 200,
            guarantee_epic = 8,
            guarantee_legendary = 2,
        },
        badge_bonus = 0.2,
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_EventChest,
        },
    },
    {
        product_id = "com.wildsky.castle.iap.chests.04",
        type = "chests",
        label = {
            name = "Treasure Jackpot",
            color = "<color=#ffbb00>",
        },
        display = DCEI.Texture("nature_chest"),
        info = {
            update_func = CustomizeStoreItemInfo_EventChest,
            label = "Open chests to get a random hero card",
        },
        popup_title = "<color=#ffbb00>" .. DCEI.GetTranslationText("data/item/chest/legendary") .. " II",
        price = {
            usd = 175,
        },
        display_guarantee = true,
        chest_type = "legendary_chest",
        items = {
            chests = 360,
            guarantee_epic = 15,
            guarantee_legendary = 5,
        },
        badge_bonus = 0.3,
        custom = {
            layout = "Portrait/Shop/Section/Item/Large",
            layout_func = CustomizeStoreItem_EventChest,
        },
    },
    
}

-- !SECTION

-- declare behavior functions
local BehaviorCallbacks
local InBattleBehaviorCallbacks
local ProcessSummonedUnitCallbacks

-- SECTION Functions
function OnMapStart()
    if event_mode.IsActive() then
        OnEventMapStart()
    else
        OnCampaignMapStart()
    end
end

function OnCampaignMapStart()
    local start_time = os.clock()
    bank.GetMods(data, ui, wave, levels, event_mode, abtest)
    bank.player = player
    bank.LoadTests()

    ui.Initialize()
    InitializeBoostGetters()
    GenerateArcherPoints()

    -- increase UI scaling as they appear smaller in portrait mode
    local scale = 1.77
    DCEI.SetUnitHealthBarUiScale(scale)
    DCEI.SetGoldRewardTextScale(scale)

    -- DCEI.SetBackButton(PressBackButton)
    DCEI.SetBackButton(function() end)
    DCEI.SuppressPauseMenuOnApplicationPause()
    DCEI.HideFastForwardButton()
    DCEI.DisableGameEndDetection()
    DCEI.SetEnabledCameraClamp(false)
    DCEI.HideCallWaveIndicators()
    DCEI.HideShipSpellMenu()

    CreateHeroSlots()
    CreateTowerSlots()
    CreateTownSlots()
    CreateDefenseSlots()
    CreateGearAnchor()

    bank.ad_data = ad_data
    bank.system = system
    bank.DAILY_ADS_LIMIT = DAILY_ADS_LIMIT
    bank.CRAFT_AUTO_REFRESH_TIME = CRAFT_AUTO_REFRESH_TIME
    bank.CRAFT_MAX_LEVEL = CRAFT_MAX_LEVEL
    bank.RAID_KEY_DAILY_COUNT = RAID_KEY_DAILY_COUNT
    bank.gear_upgrade_gold = gear_upgrade_gold
    bank.gear_menu_click = gear_menu_click
    bank.tutorial = tutorial

    bank.Load()
    InitData()

    event_mode.GetMods(data, ui, bank)
    event_mode.InitEventMode()
    

    DefineBuildOrder()


    PreloadPropLayers()
    EnforceUnlockedSlots()
    --DebugPlayerXP()

    local duration = 0
    MoveCamera(cameras.castle, duration)

    CreateCastleUnits()
    CreateTownUnits()
    CreateCastleUI(true)

    RestartAutoSaveTimer()
    Tutorial00_Battle_Check()


    Tutorial07_Colony_CheckOnMapStart()
    Tutorial08_TownUpgrade_PromptCheck()
    Tutorial_Vault_CheckOnMapStart()

    ApplyCurrentTests()

    InitializeStore()

    HookupUI()
    if not event_mode.IsActive() then
        InitializeSeasonalColonies(os.time())
    end
    CheckServerTimeOnGameStart()
    AttemptStartAirshipTimer()
    AttemptStartElixirTimer()
    AttemptStartVaultTimer()
    AttemptStartCraftTimer()

    if player.tests.spin_wheel and not event_mode.IsActive() then
        AttemptStartEggHuntTimer(os.time())
    end
    if player.race_event.expire_time and not player.race_event.event_end then
        AttemptStartRaceEventTimer(os.time())
    end

    DCEI.SetScreenSleep(false)

    local end_time = os.clock()
    local loading_time = end_time - start_time
    DCEI.LogMessage(">>>>> LUA LOADING TIME: " .. loading_time .. " seconds <<<<<")
    local user_prop = UserProps.LuaLoading(loading_time)
    LogUserProperty(user_prop)
end

function OnEventMapStart()
    local start_time = os.clock()
    bank.GetMods(data, ui, wave, levels, event_mode, abtest)
    bank.player = player
    bank.LoadTests()

    ui.Initialize()
    InitializeBoostGetters()
    GenerateArcherPoints()

    -- increase UI scaling as they appear smaller in portrait mode
    local scale = 1.77
    DCEI.SetUnitHealthBarUiScale(scale)
    DCEI.SetGoldRewardTextScale(scale)

    -- DCEI.SetBackButton(PressBackButton)
    DCEI.SetBackButton(function() end)
    DCEI.SuppressPauseMenuOnApplicationPause()
    DCEI.HideFastForwardButton()
    DCEI.DisableGameEndDetection()
    DCEI.SetEnabledCameraClamp(false)
    DCEI.HideCallWaveIndicators()
    DCEI.HideShipSpellMenu()

    CreateHeroSlots()
    CreateTowerSlots()
    CreateTownSlots()
    CreateDefenseSlots()
    CreateGearAnchor()

    bank.ad_data = ad_data
    bank.system = system
    bank.DAILY_ADS_LIMIT = DAILY_ADS_LIMIT
    bank.CRAFT_AUTO_REFRESH_TIME = CRAFT_AUTO_REFRESH_TIME
    bank.CRAFT_MAX_LEVEL = CRAFT_MAX_LEVEL
    bank.RAID_KEY_DAILY_COUNT = RAID_KEY_DAILY_COUNT
    bank.gear_upgrade_gold = gear_upgrade_gold
    bank.gear_menu_click = gear_menu_click
    bank.tutorial = tutorial

    bank.Load()
    InitEventData()


    event_mode.GetMods(data, ui, bank)
    event_mode.InitEventMode()
    event_mode.CreatePotionPayoutTimer()

    DefineBuildOrder()

    PreloadPropLayers()
    EnforceUnlockedSlots()
    --DebugPlayerXP()

    local duration = 0
    MoveCamera(cameras.castle, duration)

    CreateCastleUnits()
    CreateTownUnits()
    CreateCastleUI(true)

    RestartAutoSaveTimer()

    ApplyCurrentTests()

    InitializeEventStore()

    HookupUI()

    if player.tests.spin_wheel and not event_mode.IsActive() then
        AttemptStartEggHuntTimer(os.time())
    end
    if player.race_event.expire_time and not player.race_event.event_end then
        AttemptStartRaceEventTimer(os.time())
    end

    event_mode.AddOfflineMineGold(os.time())

    DCEI.SetScreenSleep(false)

    local end_time = os.clock()
    local loading_time = end_time - start_time
    DCEI.LogMessage(">>>>> LUA LOADING TIME: " .. loading_time .. " seconds <<<<<")
    local user_prop = UserProps.LuaLoading(loading_time)
    LogUserProperty(user_prop)
end

-- NOTE DEBUG settings
local DEBUG = {
    -- select_devil_modifier = true
    -- test_monsters = true
    -- run_tests = true,
    -- skip_popup = true,
    -- print_events = true,
    -- print_user_props = true,
    -- simulate_platform = "Android",
    -- simulate_platform = "IPhonePlayer",
    -- simulate_platform = "WebGLPlayer",
}

-- wave.data.FAST_BATTLE = true
-- ui.CINEMATIC_MODE = true
-- player.is_dev = DCEI.IsDev()

function RunTests()
    player.dev_mode = DCEI.IsDev()
    wave.data.SetCurrent(wave.data.GetMax())
    DCEI.SetPauseButtonFrame(ui.settings.frame)
    --StartTrainingBattle(15)
end

function RunLocalizationTest()
    -- -- TEST A
    -- AddAllItemsToStore()
    -- DCEI.TriggerAddTimerEventElapsed(
    --     function()
    --         Tutorial04_Boss_Debug()
    --     end,
    --     1.5
    -- )

    -- -- TEST B
    -- CreateReviewInvite()

    -- -- TEST C
    -- DCEI.SetFrameActive(ui.bottom_panel.left_inner.frame, true)
    -- Tutorial05_Mine_StepA()

    -- -- TEST D
    -- Tutorial06_Promote_StepA()

    -- -- TEST E
    -- DCEI.SetFrameActive(ui.bottom_panel.left_inner.frame, true)
    -- DCEI.SetProtectedValue("max_wave", 20)
    -- player.system.airship = system.active
    -- tutorial.colony = false
    -- Tutorial07_Colony_StepA()
end

function AddAllItemsToStore()
    for key, _ in pairs(store_propers.collections) do
        AddItemsToShopDealsFromCollectionKey(key)
    end
end

function TestTownBuild()
    DCEI.Wait(1)

    local start = 1
    local stop = 50

    for n = start, stop do
        BuildTown(n, true)
        DCEI.Wait(0.125)
    end
end

function AddItemsToShopHellCoinSections(list, options)
    local section_name = "hell_coin"
    ui.shop:RemoveAllItemsFromSection(section_name)
    ui.shop:AddItemsToSection(section_name, list, options)
end

function InitializeHellCoinPage()
    local list = {}
    local badge_index = player.hell_coin_shop.badge_index
    table.insert(list, data.badges[badge_index].product_id)
    table.insert(list, "com.wildsky.castle.hell_coin.philosopher_stone.01")
    table.insert(list, "com.wildsky.castle.hell_coin.raid_turbo.01")

    local options = { init = true }
    AddItemsToShopHellCoinSections(list, options)
end

function InitializeDealsPage()
    local bundle_keys = {}
    local current_wave = wave.data.GetMax()

    local mine_pack_unlock_wave = 5
    local max_dwarf_level = 50

    local key = "remove_ad"
    if AnyBundleItemsAreAvailable(key) then
        table.insert(bundle_keys, key)
    end

    -- show mine pack if unlocked mine, mine level < 50, and any mine pack is available
    local key = "mine_pack"
    if
        current_wave >= mine_pack_unlock_wave
        and AnyBundleItemsAreAvailable(key)
        and GetDwarfLevel() <= max_dwarf_level
    then
        local key = "mine_pack"
        table.insert(bundle_keys, key)
    end

    -- if any current deal pack is purchased, show autobattle packs
    local token_pack_unlock_wave = 100
    local key = current_wave >= token_pack_unlock_wave and "token_pack" or "starter_pack"
    if AnyBundleItemsAreAvailable(key) then
        table.insert(bundle_keys, key)
    end

    -- raid turbo boost bundle
    local key = "raid_turbo_pack"
    if current_wave >= token_pack_unlock_wave and AnyBundleItemsAreAvailable(key) then
        table.insert(bundle_keys, key)
    end

    local options = { init = true }
    AddItemsToShopDealsFromCollectionKey(bundle_keys, options)

    -- add auto battle bundle if time is valid
    local bundle_keys = "autobattle_pack"
    if AnyBundleItemsAreAvailable(bundle_keys) then
        local id = "com.wildsky.castle.iap.bundle.autobattle.new.02"
        local timer_data = shop.GetTimedOfferData(id)

        if timer_data then
            if shop.TimedOfferStillValid(id) then
                local options = { init = true, expires_at = timer_data.expires_at }
                AddItemsToShopDealsFromCollectionKey(bundle_keys, options)
            end
        end
    end

    if TestB("test010920_massive_bundle") then
        local product_id = "com.wildsky.castle.iap.bundle.huge.01"
        if shop.GetItemStockAvailable(product_id) > 0 then
            ui.shop:AddItemsToSection("deals", { product_id }, options)
        end
    end
end

function ValidStoreCollectionKey(key)
    return key and store_propers.collections[key]
end

function StoreCollectionKeyIsBoostItem(key)
    return key and (key == "boost_gold" or key == "boost_token")
end

function CheckForNewProductsThatNeedRedPips(items)
    local key = bank.bank_data.store_item_keys.key
    player[key] = player[key] or {}

    for _, product_id in pairs(items) do
        if RED_PIP_PRODUCTS[product_id] and not player[key][product_id] then
            --DCEI.LogError("message " ..product_id)
            local layout = shop.GetItemLayout(product_id)
            if layout then
                layout:ShowPip()
                ShowShopPip()

                player[key][product_id] = true
            end
        end
    end

    bank.CommitStoreItemKeys()
end

function AddItemsToShopDealsFromCollectionKey(keys, options)
    local boost_items = {}

    if type(keys) == "string" then
        -- convert single key to table
        keys = { keys }
    elseif not type(keys) == "table" then
        -- error if keys not table
        DCEI.LogError("> keys not represented by table: " .. tostring(keys))
        return
    end

    -- sets the content of the store deals page from listed key
    for _, key in ipairs(keys) do
        if ValidStoreCollectionKey(key) then
            local items = GetStoreCollectionData(key)

            if StoreCollectionKeyIsBoostItem(key) then
                DCEI.LogMessage("> Adding items to Boosts: " .. "[" .. key .. "] " .. #items)
                local section_name

                if key == "boost_gold" then
                    section_name = "gold"
                    -- remove existing items from section
                    ui.shop:RemoveAllItemsFromSection(section_name)
                    -- add new bundles to section
                    ui.shop:AddItemsToSection(section_name, items, options)
                    -- add previous items back to section
                    for _, row in ipairs(SHOP_SECTIONS[section_name]) do
                        ui.shop:AddItemsToSection(section_name, row, options)
                    end
                elseif key == "boost_token" then
                    -- new boost items, add 1 token boost and 1 gold boost
                    section_name = "gold"
                    -- remove existing items from section
                    ui.shop:RemoveAllItemsFromSection(section_name)
                    -- add new bundles to section
                    ui.shop:AddItemsToSection(section_name, { items[2] }, options)
                    -- add previous items back to section
                    for _, row in ipairs(SHOP_SECTIONS[section_name]) do
                        ui.shop:AddItemsToSection(section_name, row, options)
                    end
                    section_name = "tokens"
                    -- remove existing items from section
                    ui.shop:RemoveAllItemsFromSection(section_name)
                    -- add new bundles to section
                    ui.shop:AddItemsToSection(section_name, { items[1] }, options)
                    -- add previous items back to section
                    for _, row in ipairs(SHOP_SECTIONS[section_name]) do
                        ui.shop:AddItemsToSection(section_name, row, options)
                    end
                end
            else
                DCEI.LogMessage("> Adding items to Deals: " .. "[" .. key .. "] " .. #items)
                ui.shop:AddItemsToSection("deals", items, options)
            end

            CheckForNewProductsThatNeedRedPips(items)
        else
            DCEI.LogError("> invalid store deals key: " .. tostring(key))
        end
    end
end

function RemoveCurrentBoostsFromDealsPage()
    for _, boost in pairs(data.boosts.array) do
        local items = GetStoreCollectionData(boost.key)
        ui.shop:RemoveItems(items)
    end
end

function AnyBundleItemsArePurchased(key)
    local items = GetStoreCollectionData(key)

    for k, product_id in ipairs(items) do
        -- for now all bundle items have stock count 1
        if shop.GetItemStockAvailable(product_id) <= 0 then
            return true
        end
    end

    return false
end

function AnyBundleItemsAreAvailable(key)
    if event_mode.IsActive() then
        return false
    end

    local items = GetStoreCollectionData(key)

    for k, product_id in ipairs(items) do
        if shop.GetItemStockAvailable(product_id) > 0 then
            return true
        end
    end

    return false
end

function AttemptOnStartDealsPopup()
    local unlock_wave = 10
    local game_state = ui.bottom_panel.state

    if DEBUG.skip_popup then
        -- skip deals popup if using dev settings
        return
    end

    -- avoid opening popup if player is in combat or store (and only open if deal items are available)
    local key = GetCurrentPopupDealKey()
    if
        (wave.data.GetMax() >= unlock_wave)
        and game_state == "castle"
        and not ui.shop.is_active
        and AnyBundleItemsAreAvailable(key)
    then
        local from = "app launch offer"
        SetStoreEntrySource(from)

        OpenDealsPopup(key)
    end
end

function GetCurrentPopupDealKey()
    local token_deal_unlock_wave = 100
    local key = "starter_pack"

    -- show token pack post wave 100
    if wave.data.GetMax() >= token_deal_unlock_wave then
        key = "token_pack"
    end

    -- add auto battle bundle if time is valid
    local id = "com.wildsky.castle.iap.bundle.autobattle.new.02"
    local timer_data = shop.GetTimedOfferData(id)

    if timer_data and shop.TimedOfferStillValid(id) then
        key = "autobattle_pack"
    end

    -- show boosts if previous deal items were purchased
    if AnyBundleItemsArePurchased(key) then
        local index = player.calendar.boost_index
        key = data.boosts.array[index].key
    end

    return key
end

function OpenDealsPopup(key, title, on_close)
    -- only open popup if user online
    if DCEI.IsOnline() then
        -- get current deal based on player context
        key = key or GetCurrentPopupDealKey()
        local items_list = GetStoreCollectionData(key)
        local options = { display_name = title or DCEI.GetTranslationText("shop/popup/daily_deals") }
        if key == "autobattle_pack" then
            local id = "com.wildsky.castle.iap.bundle.autobattle.new.02"
            local timer_data = shop.GetTimedOfferData(id)
            if timer_data then
                options.expires_at = timer_data.expires_at
            else
                options.expires_in = 86340
            end
        end
        ui.popup:Show(items_list, options)

        -- restore current deals page when popup is closed
        ui.popup.on_close = on_close
    end
end

function TestB(key)
    return player.tests[key] and player.tests[key] == 2
end

function GetPlayerTest(key)
    return player.tests[key]
end

function ApplyCurrentTests()
    -- 8/31 adjust some hero prices for all new players
    local test_value = player.tests.cost_rebalance
    if test_value then
        -- hero adjustments
        data.heroes["Hero Chemist"].cost_adjust = 5000
        data.heroes["Hero Naga"].cost_adjust = 10000
        data.heroes["Hero KatMage"].cost_adjust = 10000
        data.heroes["Hero Angel"].cost_adjust = 25000

        -- tower adjustments
        data.towers["Tower GoldTower"].cost_adjust = 10000
        data.towers["Tower ExpTower"].cost_adjust = 10000
    end

    -- 8/31 add difficulty steps to wave progression every 100 waves for all new players
    local test_value = player.tests.wave_rebalance
    if test_value then
        wave.ApplyStepScaling()
    end

    -- 10/13 AB test deferring town hall system to wave 50
    if TestB("deferred_town_hall") and not player.system.town_upgrade_tutorial ~= system.active then
        local text = DCEI.GetTranslationText("ui/town/building/airship/req", { x = "50" })
        ui.loadout.town.hall:SetSubtitle(text)
    end

    -- 11/12 add town building upgrade
    if not player.tests.town_building_upgrade then
        DCEI.LogMessage("player prior to town building upgrade")
        -- upgrade these 3 buildings to Lv2 for players that purchased them prior to this change
        local convert_buildings = {
            DCEI.Unit("Town Building Brewery"),
            DCEI.Unit("Town Building MysticFount"),
            DCEI.Unit("Town Building Temple"),
        }
        for _, name in ipairs(convert_buildings) do
            local town_building = data.town[name]
            if town_building.is_purchased then
                DCEI.LogMessage("already purchased " .. name .. " set level to 2")
                -- for safety not use upgrade function
                ProcessEquipBonuses(town_building, false)
                town_building.level = 2
                ProcessEquipBonuses(town_building, true)
                bank.CommitUnitData(town_building, bank.bank_data.prop.level, town_building:GetLevel())
            end
        end
        player.tests.town_building_upgrade = true
        bank.CommitTests()
    end

    -- 12/1 add gear tutorial, check for old players have gear but don't equipped
    if player.system.gear_equip_tutorial == nil or player.system.gear_equip_tutorial == system.locked then
        -- check if player have mine
        if CheckAnyGearEquipped() then
            -- found equipped gear, skip tutorial
            player.system.gear_equip_tutorial = system.active
        elseif data.gear["Mining Pick"].is_purchased then
            player.system.gear_equip_tutorial = system.unlocked
        end
        bank.CommitSystemState()
    end

    -- 02/23 log open game event for os days and is from notification
    local from_focus = false
    CheckOSDaysElapsed(from_focus)

    -- 02/25 patch, apply 10% users to watch IronSource Ads
    -- if not player.tests.ironsource_ads then
    --     player.tests.ironsource_ads = 1
    --     local RollIronSource = core.RandomInteger(1, 10)
    --     if RollIronSource == 1 then
    --         player.tests.ironsource_ads = 2
    --     end
    --     -- log test result as user property
    --     LogUserProperty{
    --         name = "AB_ironsource_ads",
    --         value = player.tests.ironsource_ads
    --     }
    --     bank.CommitTests()
    -- end

    if not player.tests.craft_upgrade then
        if UnlockEquipment() then
            SetCraftLevel(CRAFT_MAX_LEVEL)
            bank.CommitCraft()
        end
        player.tests.craft_upgrade = true
        bank.CommitTests()
    end

    -- tests for all players
    local tests = abtest.ApplyTests(player.tests, wave.data.GetMax())
    if next(tests) then
        for k, v in pairs(tests) do
            player.tests[k] = v
            -- log test result as user property
            LogUserProperty({
                name = "AB_" .. k,
                value = v,
            })
        end
        bank.CommitTests()
    end

    -- adjust paragon levels
    if not player.tests.adjust_paragon then
        for _, talent in ipairs(data.talents) do
            if talent.paragon then
                ResetTalentAttributes(talent)
                ResetTalentBehaviors(talent)
                talent.level = 0
                if event_mode.IsActive() then
                    bank.CommitEventTalent(talent)
                else
                    bank.CommitTalent(talent)
                end
            end
        end
        SetParagonPoints(GetParagonLevel())
        ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())
        if event_mode.IsActive() then
            bank.CommitEventParagonPoints()
        else
            bank.CommitParagonPoints()
        end

        ui.glue.talent_menu:ShowPipAnimated()
        player.tests.adjust_paragon = true
        bank.CommitTests()
    end

    -- if not player.tests.new_gear_onboarding then
    --     if wave.data.GetMax() >= 15 then
    --         local unit_data = data.gear["Health Potion"]
    --         PurchaseGear(unit_data)
    --         local unit_data = data.gear["Mana Potion"]
    --         PurchaseGear(unit_data)
    --     end
    --     if wave.data.GetMax() >= 25 then
    --         local unit_data = data.gear["Squire's Helmet"]
    --         PurchaseGear(unit_data)
    --     end
    --     player.tests.new_gear_onboarding = true
    --     bank.CommitTests()
    -- end

    -- 02/02 split maps, check players progress, if complete first colony set,then set map entry and load intermediate map
    local on_map_start = true
    ChangeMapCheck(on_map_start)
end

function ChangeMapCheck(on_map_start)
    local current_map = DCEI.GetLevel()

    local colonies_completed = #levels.colony.complete
    
    local need_check = true
    if event_mode.IsActive() then
        need_check = false
    end
    if colonies_completed >= #levels.colony then
        levels.colony.all_clear = true
        if current_map == levels.maps[1] then
            need_check = true
        else
            need_check = false
            if not event_mode.IsActive() then
                if not player.system.colony_travel_tutorial or player.system.colony_travel_tutorial == system.locked then
                    BlockInput()
                    DCEI.TriggerAddTimerEventElapsed(function()
                        if ui.popup then
                            ui.popup:Hide()
                        end
                        ui.ColonyAllClearPrompt()
                        bank.CommitCurrentMap(#levels.maps)
                        player.system.colony_travel_tutorial = system.active
                        bank.CommitSystemState()
                        UnblockInput()
                    end, 2, false, true)
                end
            end
        end
    end

    if need_check then
        local target_map_index = 1
        for k, colony in ipairs(levels.colony) do
            if colony.scene.map_index > target_map_index then
                target_map_index = colony.scene.map_index
            end
            if not colony.complete then
                break
            end
        end
        levels.colony.current_map_id = target_map_index
        bank.CommitCurrentMap(target_map_index)
        local target_map = levels.maps[target_map_index]
        if current_map ~= target_map then
            DCEI.Save.SetMapEntry(target_map)
            if on_map_start then
                -- load map in next frame to save in this frame
                DCEI.TriggerAddTimerEventElapsed(function()
                    DCEI.PlayLevel(target_map, "Wild Castle")
                end, 0, false, true)
            else
                DCEI.TriggerAddTimerEventElapsed(function()
                    local prompt = ui.ChangeMapPrompt(target_map_index)
                    DCEI.SetOnClickCallback(prompt.button, function()
                        DCEI.PlayLevel(target_map, "Wild Castle")
                    end)
                end, 0, false, true)
            end
        end
    end
end

function ChangeEntryMap(map_index)
    local target_map = levels.maps[map_index]
    DCEI.Save.SetMapEntry(target_map)
    bank.CommitCurrentMap(map_index)
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.PlayLevel(target_map, "Wild Castle")
    end, 0, false, true)
end

ui.ChangeEntryMap = ChangeEntryMap

function ChangeEntryMapEvent()
    local target_map = "WildCastle/Maps/MapEvent"
    DCEI.Save.SetMapEntry(target_map)
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.PlayLevel(target_map, "Wild Castle")
    end, 0, false, true)
end

ui.ChangeEntryMapEvent = ChangeEntryMapEvent

-- TESTS
function RunTest_ColonyScenes()
    for k, colony in ipairs(levels.colony) do
        CreateColonyBattleSet(colony)
        DCEI.Wait(0.25)
        RestoreDefaultScene()
    end
end

function AdLimitReached(ad_type)
    local current_time = os.time()
    local oldest_delta = current_time - (ad_type[1] or 0)
    local ad_count = 0

    for _, v in ipairs(ad_type) do
        if v > 0 then
            ad_count = ad_count + 1
        end
    end

    if ad_type.limit_count and (ad_count < ad_type.limit_count) then
        -- less than ad limit has been played
        return false
    elseif ad_type.limit_duration and (oldest_delta >= ad_type.limit_duration) then
        -- oldest tracked ad older than reset duration
        return false
    end

    -- ad limit reached
    return true
end

function AdTrackLimit(ad_type)
    local current_time = os.time()

    if not ad_type.limit_count then
        -- only track ad limits for ads with count limits
        return
    end

    if #ad_type < ad_type.limit_count then
        -- store time in table
        table.insert(ad_type, current_time)
    else
        -- remove oldest time in table
        table.remove(ad_type, 1)
        table.insert(ad_type, current_time)
    end

    bank.CommitAdTracking(ad_type)
end

function WaveProgressIs(wave_id)
    return (wave.data.GetCurrent() == wave_id) and wave.data.mode == "standard"
end

function ReviewInviteCheck()
    if WaveProgressIs(11) then
        if PlayingMobileVersion() and not event_mode.IsActive() then
            -- create review invite on mobile
            SetAdOnCooldown(ad_data.gold, 2)
           -- CreateReviewInvite()
        elseif PlayingWebVersion() then
            -- prompt players to save progress on web
          --  local offset = 0
         --   local text = tutorial.message.facebook

            SetAdOnCooldown(ad_data.gold, 2)
           -- ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
          --  DCEI.SetFrameSize(ui.prompt.tutorial.background, 570, 228)
        end
    end
end

function CreateReviewInvite()
    local function ThankForFeedback(delay, gold)
        delay = delay or 0.375

        DCEI.TriggerAddTimerEventElapsed(function()
            ui.screen_mask.Hide()

            local text = DCEI.GetTranslationText("ui/review/thanks")
            ui.CreateFeedbackThanks(text)

            if gold then
                -- give gold to attract player to rating
                local value = 5000
                AddPlayerGold(value)
                bank.CommitGold()
                ui.GoldRewardFeedback(value)
            end
        end, delay, true, true)
    end

    local function OnAcceptInvite()
        -- culled Review Accept event awhile ago, replaced with user property for 0.0.117
        LogEvent(LogEvents.review_accepted)
        DCEI.Event.SetUserProperty("Review Accept", tostring(true))

        DCEI.OpenUrl("market://details?id=com.wildsky.wildcastle")
        local gold = true
        ThankForFeedback(0.125, gold)
    end

    local function OnAcceptInviteIOS()
        LogEvent(LogEvents.review_accepted)
        DCEI.Event.SetUserProperty("Review Accept", tostring(true))

        DCEI.RequestAppStoreReview()

        local gold = true
        ThankForFeedback(0.125, gold)
    end

    local function OnFiveStar()
        local callbacks = {
            on_accept = OnAcceptInvite,
            on_decline = ThankForFeedback,
        }

        local ios_callbacks = {
            on_accept = OnAcceptInviteIOS,
            on_decline = ThankForFeedback,
        }

        DCEI.Event.LogRating("Rating", rating)

        DCEI.TriggerAddTimerEventElapsed(function()
            if PlayingAndroidVersion() then
                -- for android, create a prompt asking if they wish to review
                ui.NewReviewInvite(callbacks)
            elseif PlayingiOSVersion() then
                -- for ios, request app store review
                ui.NewReviewInvite(ios_callbacks)
            end
        end, 0.375, true, true)
    end

    local callbacks = {
        on_five_star = OnFiveStar,
        on_less_star = ThankForFeedback,
    }

  --  ui.NewReviewAsk(callbacks)
end

function BlockInput()
    player.block_input = true
    DCEI.UnblockGameWorld()
    DCEI.BlockGameWorld()
end

function BlockInputExcept(frame)
    BlockInput()
    DCEI.WhitelistUi(frame)
end

function UnblockInput()
    player.block_input = false
    DCEI.UnblockGameWorld()
end

function AttemptUnblockInput()
    if not player.block_input then
        UnblockInput()
    end
end

-- !SECTION

-- ===================================================================================================================================================================================================================
-- SECTION TUTORIALS
-- ===================================================================================================================================================================================================================

-- TUTORIAL Battle Start
function Tutorial00_Battle_Check()
    if wave.data.GetMax() == 0 then
        Tutorial00_Battle_StepA()
    end
end

function Tutorial00_Battle_StepA()
    local options = {
        anchor_bottom = true,
        flip = true,
    }

    BlockInputExcept(ui.bottom_panel.right_outer.frame)
    ui.prompt.helper = ui.NewBigHand(160, 60, options)
end

-- TUTORIAL Hero Skill
function Tutorial01_Hero_Check(wave_id)
    if wave_id == 1 then
        local wave_instace = wave.data.instance

        ui.DestroyBigHand()
        UnblockInput()

        DCEI.TriggerAddTimerEventElapsed(function()
            local unit_data = data.heroes["Hero Lancer"]
            local mana = DCEI.GetMana(CASTLE_UNIT)
            local cost = unit_data:GetAbilityCost()

            -- only start the tutorial if lancer is equipped and can cast skill
            if
                unit_data.slot
                and unit_data.slot > 0
                and wave.data.is_active
                and (not HeroSkillOnCooldown(unit_data))
                and (mana > cost)
                and not player.targeting_data
                and (wave_instace == wave.data.instance)
            then
                Tutorial01_Hero_StepA()
            end
        end, 2, false, true)
    end
end

function Tutorial01_Hero_StepA()
    -- get lancer slot in case player moves lancer
    local unit_data = data.heroes["Hero Lancer"]
    local slot = unit_data.slot

    if slot and slot > 0 then
        local offset = -441
        local text = tutorial.message.hero_skills01
        local unit = player.heroes[slot]
        local options = {
            unit = unit,
            attach = {
                offset = { up = 2.5 },
                center_at_unit_origin = true,
            },
        }

        tutorial.state = 1
        SetGameSpeed(speed.slow)

        BlockInputExcept(ui.loadout.hero.slots[slot].button)
        ui.prompt.helper = ui.NewBigHand(96, 0, options)
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
    end
end

function Tutorial01_Hero_StepB(unit)
    -- get lancer slot in case player moves lancer
    local unit_data = data.heroes["Hero Lancer"]
    local slot = unit_data.slot

    if tutorial.state == 1 and unit == player.heroes[slot] then
        local offset = -441
        local unit = player.heroes[1]
        tutorial.state = 0

        UnblockInput()
        SetGameSpeed(player.speed)
        ui.DestroyBigHand()

        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end

        -- show follow up message
        local display_delay = 1
        local display_duration = 5

        local function FollowUpMessage()
            local text = tutorial.message.hero_skills02
            ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
            DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, false, true)
        end

        DCEI.TriggerAddTimerEventElapsed(FollowUpMessage, display_delay, false, true)
    end
end

-- TUTORIAL Upgrade Archer
function Tutorial02_Archer_Check()
    if WaveProgressIs(1) then
        DCEI.BlockGameWorld()

        local delay = 0.375
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            Tutorial02_Archer_StepA()
        end, delay, true, simple)
    end
end

function Tutorial02_Archer_StepA()
    local offset = -441
    local text = tutorial.message.upgrade_archer
    tutorial.state = 3

    BlockInputExcept(ui.glue.upgrade_archer.frame)

    local options = { anchor_left = true }
    ui.prompt.helper = ui.NewBigHand(235, 150, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial02_Archer_StepB()
    if tutorial.state == 3 then
        -- finish tutorial sequence
        tutorial.state = 0

        UnblockInput()
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()
    end
end

-- TUTORIAL Upgrade Castle
function Tutorial03_Castle_Check()
    if WaveProgressIs(2) then
        DCEI.BlockGameWorld()

        local delay = 0.375
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            Tutorial03_Castle_StepA()
        end, delay, true, simple)
    end
end

function Tutorial03_Castle_StepA()
    local offset = -441
    local text = tutorial.message.upgrade_castle
    tutorial.state = 4

    BlockInputExcept(ui.glue.upgrade_castle.frame)

    local options = { anchor_left = true }
    ui.prompt.helper = ui.NewBigHand(235, 285, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial03_Castle_StepB()
    if tutorial.state == 4 then
        tutorial.state = 0
        UnblockInput()
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()
    end
end

-- TUTORIAL Retry Boss
function Tutorial04_Boss_Check()
    if WaveProgressIs(3) then
        if data.talents.Spent() == 0 then
            Tutorial04_1_Talent_Tutorial()
        else
            Tutorial04_Boss_StepA()
        end
    end
end

function Tutorial04_Boss_StepA()
    local offset = 320
    local text = tutorial.message.boss_retry_first
    local display_delay = 0.25
    local display_duration = 5

    -- display random encouragement text each subsequent time player loses
    if player.wave_attempts > 1 then
        local index = math.random(1, #tutorial.message.boss_retry_again)
        text = tutorial.message.boss_retry_again[index]
    end

    local function Message()
        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
        DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, false, true)
    end

    DCEI.TriggerAddTimerEventElapsed(Message, display_delay, true, true)
end

function Tutorial04_1_Talent_Check()
    if WaveProgressIs(4) then
        if data.talents.Spent() == 0 then
            Tutorial04_1_Talent_Tutorial()
        end
    end
end

function Tutorial04_1_Talent_Tutorial()
    local offset = -440
    local text = tutorial.message.talent_menu

    -- don't let tutorial sequence get interrupted by ads
    SetAdOnCooldown(ad_data.gold, 2)
    BlockInputExcept(ui.glue.talent_menu.frame)

    tutorial.state = 4.1
    local options = {
        parent = ui.glue.talent_menu.frame,
        scale = 0.6,
        flip = true,
    }
    ui.prompt.helper = ui.NewBigHand(-71, 36, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial04_Boss_Debug()
    local offset = 320
    local display_duration = 1.5

    local text = tutorial.message.boss_retry_first
    ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, false, true)

    for index, text in ipairs(tutorial.message.boss_retry_again) do
        DCEI.Wait(display_duration)

        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
        DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, false, true)
    end
end

-- TUTORIAL Upgrade Mine
function Tutorial05_Mine_Check()
    if WaveProgressIs(5) then
        -- don't let tutorial sequence get interrupted by ads
        SetAdOnCooldown(ad_data.gold, 2)

        DCEI.BlockGameWorld()

        local delay = 0.375
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            Tutorial05_Mine_StepA()
        end, delay, true, simple)
    end
end

function Tutorial05_Mine_StepA()
    tutorial.state = 5

    local offset = -214
    local text = tutorial.message.mine01

    BlockInputExcept(ui.bottom_panel.left_inner.frame)

    local options = { anchor_bottom = true }
    ui.prompt.helper = ui.NewBigHand(-21, 71, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial05_Mine_StepB()
    if tutorial.state == 5 then
        tutorial.state = 6

        -- reset tutorial state
        DCEI.BlockGameWorld()
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        local offset = -214
        local text = tutorial.message.mine02
        local options = {
            unit = MINE_DEPOT,
            attach = {
                center_at_unit_origin = true,
            },
        }
        DCEI.TriggerAddTimerEventElapsed(function()
            BlockInputExcept(ui.loadout.town.dwarf.button)
            ui.prompt.helper = ui.NewBigHand(82, 36, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, 0.375, false, true)
    end
end

function Tutorial05_Mine_StepC()
    if tutorial.state == 6 then
        tutorial.state = 7

        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        local offset = -214
        local text = tutorial.message.mine03
        local options = {
            scale = 0.7,
        }

        BlockInputExcept(ui.worker_menu.button.button)
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.prompt.helper = ui.NewBigHand(242, 64, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, 0.25, false, true)
    end
end

function Tutorial05_Mine_StepD()
    if tutorial.state == 7 then
        tutorial.state = 8

        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        local offset = -214
        local text = tutorial.message.mine04
        local display_duration = 5

        UnblockInput()
        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
        DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, true, true)
    end
end

function Tutorial_Vault_CheckOnMapStart()
    if GetPlayerTest("pause_offline_gold") then
        return
    end
    local unlock_wave = 20

    if
        player.system.town_upgrade_tutorial == system.active
        and player.system.vault == system.locked
        and wave.data.GetMax() >= unlock_wave
    then
        ui.bottom_panel:ShowTownPipAnimated()
    end
end

function Tutorial08_1_Mine_Vault_Check()
    if GetPlayerTest("pause_offline_gold") then
        return
    end
    if player.system.town_upgrade_tutorial == system.active and player.system.vault == system.locked then
        tutorial.state = 8.1
        player.system.vault = system.active
        bank.CommitSystemState()
        data.vault:UpdateVaultCap(wave.data.GetMax())
        DCEI.SetFrameActive(ui.loadout.town.vault.frame, false)

        DCEI.BlockGameWorld()
        local simple = true
        local t1 = 0.625
        -- build vault animation
        DCEI.TriggerAddTimerEventElapsed(function()
            CreateEffect("Build FX 100%", 23.2, 11.3)
            CreateEffect("Build FX 100%", 23.4, 11.1)
            if VAULT and DCEI.UnitExists(VAULT) then
                DCEI.RemoveUnit(VAULT)
                VAULT = nil
            end
            local vault_unit_name = DCEI.Unit("Town Vault")
            local x, y = 23.3, 11.2
            VAULT = DCEI.CreateUnitAsync(1, 1, vault_unit_name, x, y, 0, -1)
            DCEI.TriggerAddTimerEventElapsed(function()
                local id = DCEI.CompileActorAction('{"playAnimation":{"clipId":5}}')
                DCEI.SendActorAction(VAULT, id)
            end, 0, true, true)
        end, t1, true, simple)

        local t2 = 1.5
        -- show tutorial text and big hand
        DCEI.TriggerAddTimerEventElapsed(function()
            local options = {
                offset = { up = -0.2 },
                center_at_unit_origin = true,
            }
            local use_current_root = true
            local select_box = ui.loadout.town.vault.frame
            DCEI.SetFrameActive(select_box, true)
            DCEI.AttachFrameToUnit(select_box, VAULT, options, use_current_root)

            BlockInputExcept(ui.loadout.town.vault.button)

            local offset = -200
            local text = DCEI.GetTranslationText("campaign/tutorial/vault")
            local options = {
                unit = VAULT,
                attach = {
                    center_at_unit_origin = true,
                },
            }
            ui.prompt.helper = ui.NewBigHand(82, 32, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, t2, true, simple)
    end
end

function Tutorial05_Mine_SurfaceMineBundle()
    if tutorial.state == 8 then
        local function on_close()
            -- add mine items to deals page on close
            local key = "mine_pack"
            if AnyBundleItemsAreAvailable(key) then
                AddItemsToShopDealsFromCollectionKey(key)
            end
        end

        tutorial.state = 0
        CloseActivePrompts()

        local from = "mine offer"
        SetStoreEntrySource(from)

        -- open popup with mine pack bundles
        local key = "mine_pack"
        local text = DCEI.GetTranslationText("shop/popup/mine_bundle")
        OpenDealsPopup(key, text, on_close)
    end
end

-- TUTORIAL Promote Hero
function Tutorial06_Promote_Check()
    local unit_data = data.heroes["Hero Lancer"]
    local unit_class = unit_data.current_class or unit_data.name
    local slot = unit_data.slot

    -- only start tutorial if player has enough tokens and hasn't promoted lancer
    if WaveProgressIs(10) and GetPlayerGems() >= 10 and unit_class == "Hero Lancer" and slot and slot > 0 then
        -- don't let tutorial sequence get interrupted by ads
        SetAdOnCooldown(ad_data.gold, 2)
        DCEI.BlockGameWorld()

        local delay = 0.375
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            Tutorial06_Promote_StepA()
        end, delay, true, simple)
    end
end

function Tutorial06_Promote_StepA()
    tutorial.state = 9

    local unit_data = data.heroes["Hero Lancer"]
    local slot = unit_data.slot

    local unit = player.heroes[slot]
    local options = {
        unit = unit,
        attach = {
            offset = { up = 2.5, right = 0.5 },
            center_at_unit_origin = true,
        },
    }

    -- create big hand on lancer
    ui.prompt.helper = ui.NewBigHand(54, 0, options)
    BlockInputExcept(ui.loadout.hero.slots[slot].button)

    -- tutorial text
    local offset = 0
    local token_icon = "<sprite name=ecoin>"
    local text = DCEI.GetTranslationText("campaign/tutorial/hero_promote/01", { x = token_icon })
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial06_Promote_StepB()
    if tutorial.state == 9 then
        tutorial.state = 10

        -- clear force input
        ui.prompt.tutorial.Close()
        ui.DestroyBigHand()

        local options = {
            parent = ui.list_hero.items[1].frame,
            scale = 0.6,
        }

        local delay = 0.125
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            -- force input select lancer in list
            ui.prompt.helper = ui.NewBigHand(-75, 0, options)
            BlockInputExcept(ui.list_hero.items[1].frame)
        end, delay, true, simple)
    end
end

function Tutorial06_Promote_StepC_Check()
    return tutorial.state == 10 and not event_mode.IsActive()
end

function Tutorial06_Promote_StepC()
    if tutorial.state == 10 and not event_mode.IsActive() then
        tutorial.state = 11
        ui.DestroyBigHand()

        local options = {
            parent = ui.unit_card.promote.frame,
            scale = 0.8,
        }

        local delay = 0.125
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            -- force input select promote button
            ui.prompt.helper = ui.NewBigHand(60, 60, options)
            BlockInputExcept(ui.unit_card.promote.button)
        end, delay, true, simple)
    end
end

function Tutorial06_Promote_StepD()
    if tutorial.state == 11 then
        tutorial.state = 12
        ui.DestroyBigHand()

        local options = {
            parent = ui.list_promote.items[2].frame,
            scale = 0.6,
        }

        local delay = 0.125
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            -- force input select lancer 2A
            ui.prompt.helper = ui.NewBigHand(71, 36, options)
            BlockInputExcept(ui.list_promote.items[2].frame)

            -- tutorial text
            local offset = -441
            local text = tutorial.message.hero_promote02
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, delay, true, simple)
    end
end

function Tutorial06_Promote_StepE()
    if tutorial.state == 12 then
        tutorial.state = 13

        -- clear force input
        ui.prompt.tutorial.Close()
        ui.DestroyBigHand()

        local options = {
            flip = true,
            scale = 0.8,
        }

        local delay = 0.125
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            -- force input select promote button
            ui.prompt.helper = ui.NewBigHand(90, -80, options)
            BlockInputExcept(ui.unit_card.promote.button)
        end, delay, true, simple)
    end
end

function Tutorial06_Promote_StepF()
    if tutorial.state == 13 then
        tutorial.state = 0

        -- clear force input
        ui.prompt.tutorial.Close()
        ui.DestroyBigHand()
        UnblockInput()

        -- close all hero UI
        ui.CloseUnitCard()
        AttemptCloseHeroList()

        -- tutorial text
        local offset = -441
        local text = tutorial.message.hero_promote03
        local display_duration = 5
        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
        DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, true, true)
    end
end

function TutorialNewGearOnboarding_Check()
    if WaveProgressIs(25) then
        local purchase_state = data.gear["Squire's Helmet"].is_purchased
        if not purchase_state then
            -- don't let tutorial sequence get interrupted by ads
            local unit_data = data.gear["Squire's Helmet"]
            PurchaseGear(unit_data)

            if not CheckAnyGearEquipped() then
                SetAdOnCooldown(ad_data.gold, 2)
                DCEI.BlockGameWorld()
                local delay = 0.375
                local simple = true
                DCEI.TriggerAddTimerEventElapsed(function()
                    CloseActivePrompts()

                    BlockInputExcept(ui.glue.gear_menu.frame)

                    local offset = -214
                    local text = DCEI.GetTranslationText("campaign/tutorial/gear_unlocked")
                    local options = {
                        parent = ui.glue.gear_menu.frame,
                        flip = true,
                        scale = 0.75,
                    }

                    ui.prompt.helper = ui.NewBigHand(-71, 36, options)
                    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
                    tutorial.state = 21.1
                end, delay, true, simple)
            else
                ui.glue.gear_menu:ShowPipAnimated()
            end
        end
    else
        for _, gear_data in ipairs(data.wave_unlock_gear) do
            local purchase_state = gear_data.is_purchased
            if
                not purchase_state
                and wave.data.GetCurrent() >= gear_data.unlock_wave
                and wave.data.mode == "standard"
            then
                PurchaseGear(gear_data)
                ui.glue.gear_menu:ShowPipAnimated()
            end
        end
    end
end

-- NEW COLONY TUTORIAL
function Tutorial07_Colony_CheckOnMapStart()
    local first_colony = levels.colony[1]
    local unlock_wave = 20

    if not tutorial.colony and not first_colony.complete and wave.data.GetMax() >= unlock_wave then
        ui.bottom_panel:ShowTownPip()
    end
end

function Tutorial07_Colony_Check()
    local first_colony = levels.colony[1]
    local unlock_wave = 20

    if not tutorial.colony and not first_colony.complete and WaveProgressIs(unlock_wave) then
        -- don't let tutorial sequence get interrupted by ads
        SetAdOnCooldown(ad_data.gold, 2)
        DCEI.BlockGameWorld()

        local delay = 0.375
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            Tutorial07_Colony_StepA()
        end, delay, true, simple)
    end
end

function Tutorial07_Colony_StepA()
    local offset = -214
    local text = tutorial.message.colony01

    BlockInputExcept(ui.bottom_panel.left_inner.frame)

    local options = { anchor_bottom = true }
    ui.prompt.helper = ui.NewBigHand(-21, 71, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial07_Colony_StepB()
    local unlock_wave = 20
    local first_colony = levels.colony[1]

    if not tutorial.colony and not first_colony.complete and wave.data.GetMax() >= unlock_wave then
        tutorial.state = 14

        -- destroy previous step UI
        ui.DestroyBigHand()
        CloseActivePrompts()
        DCEI.BlockGameWorld()

        local offset = 0
        local text = tutorial.message.colony02
        local options = {
            unit = AIRSHIP_DOCK,
            flip = true,
            attach = {
                center_at_unit_origin = true,
            },
        }

        -- hide airship label so we can reveal it in sequence
        DCEI.SetFrameActive(ui.loadout.town.airship.container, false)

        local simple = true
        local t1, t2 = 0.4375, 1.75

        -- adjust timings and skip airship sequence of airship already landed (rare case of player previously quit during colony tutorial)
        if AIRSHIP_LANDED then
            t1 = 0
            t2 = 0.5
        end

        -- build airship animation
        DCEI.TriggerAddTimerEventElapsed(function()
            BuildAirshipSequence()
            DCEI.SetFrameActive(ui.loadout.town.airship.container, true)
        end, t1, true)

        -- show tutorial text and big hand
        DCEI.TriggerAddTimerEventElapsed(function()
            BlockInputExcept(ui.loadout.town.airship.button)

            ui.prompt.helper = ui.NewBigHand(-100, 57, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, t2, true, simple)
    end
end

function Tutorial07_Colony_StepB_End()
    if tutorial.state == 14 then
        -- destroy previous step UI
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        ui.screen_mask.upper:Show()
    end
end

function Tutorial07_Colony_StepC()
    if tutorial.state == 14 then
        tutorial.state = 15

        -- show dummy colony item above screen mask
        local colony_data = levels.colony[1]
        local function on_click()
            ui.screen_mask.Hide()

            -- create new battle confirmation
            local item = { colony = colony_data, index = 1 }
            ui.colony.confirmation:Update(item)

            -- destroy dummy frame and reference
            DCEI.DestroyFrame(ui.colony.dummy.frame)
            ui.colony.dummy = nil

            -- continue tutorial
            Tutorial07_Colony_StepD()
        end

        ui.screen_mask.upper:Show()
        ui.colony.dummy = ui.NewColonyItem(ui.upper, colony_data, on_click)

        -- move dummy item on top of original item
        local original_colony_item = ui.colony.regions[1].items[1].frame
        DCEI.WaitForUiUpdateToComplete(function()
            local pos1 = DCEI.GetFramePosition3D(original_colony_item)
            local pos2 = DCEI.GetFramePosition3D(ui.colony.dummy.frame)
            local offset = pos1.y - pos2.y
            DCEI.SetFrameVerticalOffsetInParent(ui.colony.dummy.frame, offset)

            -- force selection of dummy colony
            BlockInputExcept(ui.colony.dummy.frame)

            local offset = -214
            local text = tutorial.message.colony03
            local options = {
                parent = ui.colony.dummy.anchor,
                scale = 0.75,
            }

            local delay = 0.125
            local simple = true
            DCEI.TriggerAddTimerEventElapsed(function()
                -- display message and big-hand first colony
                ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
                ui.prompt.helper = ui.NewBigHand(0, 100, options)
            end, delay, true, simple)
        end)
    end
end

function Tutorial07_Colony_StepD()
    if tutorial.state == 15 then
        tutorial.state = 16

        -- destroy previous step UI
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        -- big-hand battle button
        local options = {
            parent = ui.colony.confirmation.button,
            scale = 0.75,
        }

        local simple = true
        local delay = 0.25
        DCEI.TriggerAddTimerEventElapsed(function()
            BlockInputExcept(ui.colony.confirmation.button)
            ui.prompt.helper = ui.NewBigHand(57, 57, options)
        end, delay, true, simple)
    end
end

function Tutorial07_Colony_StepE()
    if tutorial.state == 16 then
        tutorial.state = 0

        -- destroy previous step UI
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()
        UnblockInput()

        -- log colony tutorial as completed
        if not tutorial.colony then
            tutorial.colony = true
            bank.CommitColonyState()
        end
    end
end

function Tutorial07_Colony_StepF(colony)
    if #levels.colony.complete == 1 and not colony.seasonal then
        ui.screen_mask.lower:Show()
        ui.back_dummy = ui.NewDummyBackButton(ui.upper)

        local original_back_button = ui.bottom_panel.left_outer.frame

        DCEI.WaitForUiUpdateToComplete(function()
            local pos1 = DCEI.GetFramePosition3D(original_back_button)
            local pos2 = DCEI.GetFramePosition3D(ui.back_dummy.anchor)
            local xoffset = pos1.x - pos2.x
            local yoffset = pos1.y - pos2.y

            DCEI.SetFrameVerticalOffsetInParent(ui.back_dummy.anchor, yoffset)
            DCEI.SetFrameHorizontalOffsetInParent(ui.back_dummy.anchor, xoffset)

            local offset = 107
            local text = tutorial.message.colony04
            local options = {
                parent = ui.back_dummy.anchor,
                scale = 0.75,
            }

            DCEI.SetOnClickCallback(ui.back_dummy.frame, function()
                Tutorial07_Colony_StepF_End()
                CloseAirshipMenu()

                -- create first shipment so player gets quick feedback
                CreateColonyAirshipDelivery()

                -- destroy dummy frame and reference
                DCEI.DestroyFrame(ui.back_dummy.frame)
                ui.back_dummy = nil
            end)

            local simple = true
            local delay = 0.125
            DCEI.TriggerAddTimerEventElapsed(function()
                -- explain colony rewards
                BlockInputExcept(ui.back_dummy.frame)
                ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
                ui.prompt.helper = ui.NewBigHand(57, 57, options)
            end, delay, true, simple)
        end)
    end
end

function Tutorial07_Colony_StepF_End()
    -- destroy previous step UI
    ui.DestroyBigHand()
    ui.prompt.tutorial.Close()

    ui.glue.inventory_menu.resource_pip[3] = true
end

-- TUTORIAL Town Hall
function Tutorial08_TownUpgrade_PromptCheck()
    local colonies_completed = #levels.colony.complete
    local unlock_check = colonies_completed >= 1

    if TestB("deferred_town_hall") then
        unlock_check = wave.data.GetMax() >= 50
    end

    if player.system.town_upgrade_tutorial ~= system.active and unlock_check then
        ui.bottom_panel:ShowTownPip()
    end
end

function Tutorial08_TownUpgrade_Check()
    local colonies_completed = #levels.colony.complete
    local town_contracts = GetTownContracts()
    local unlock_check = town_contracts >= 1 and colonies_completed >= 1

    if TestB("deferred_town_hall") then
        unlock_check = wave.data.GetMax() >= 50
    end

    if player.system.town_upgrade_tutorial ~= system.active and unlock_check then
        Tutorial08_TownUpgrade_StepA()
    end
end

function Tutorial08_TownUpgrade_StepA()
    tutorial.state = 17

    local offset = -470
    local text = tutorial.message.town01
    local options = {
        unit = TOWN_HALL,
        attach = {
            center_at_unit_origin = true,
        },
    }

    -- show tutorial text and big hand town hall
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.prompt.helper = ui.NewBigHand(78, 100, options)
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        BlockInputExcept(ui.loadout.town.hall.button)
    end, 0.5, false, true)
end

function Tutorial08_TownUpgrade_StepB()
    if tutorial.state == 17 then
        tutorial.state = 18

        -- destroy previous step UI
        ui.DestroyBigHand()
        CloseActivePrompts()

        local offset = -470
        local text = tutorial.message.town02
        local options = {
            scale = 0.7,
        }

        -- show tutorial text and big hand upgrade button
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.prompt.helper = ui.NewBigHand(242, 60, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
            BlockInputExcept(ui.town_menu.upgrade_button.button)
        end, 0.375, true, simple)
    end
end

function Tutorial08_TownUpgrade_StepC()
    if tutorial.state == 18 then
        tutorial.state = 0

        -- destroy previous step UI
        ui.DestroyBigHand()
        CloseActivePrompts()
        UnblockInput()

        local offset = -470
        local text = tutorial.message.town03
        local display_duration = 5

        -- show dismissable tutorial text and resume regular input
        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
        DCEI.TriggerAddTimerEventElapsed(ui.prompt.tutorial.Close, display_duration, true, true)

        -- set town ugprade tutorial as completed
        ui.bottom_panel:HideTownPip()
        player.system.town_upgrade_tutorial = system.active
        bank.CommitSystemState()
    end
end

-- TUTORIAL Challenge
function GetChallengeUnlockWave()
    return TestB("unlock_challenge_earlier") and 25 or 30
end

function CheckChanllengeShowPip()
    local unlock_wave = GetChallengeUnlockWave()
    if WaveProgressIs(unlock_wave) then
        ui.bottom_panel:ShowTownPip()
        ui.loadout.town.airship:ShowPipAnimated()
    end
end

function Tutorial09_Challenge_Unlock()
    local completed_challenges = #levels.challenge.complete
    local unlock_wave = GetChallengeUnlockWave()

    return not tutorial.challenge and (completed_challenges == 0) and (wave.data.GetMax() >= unlock_wave)
end

ui.UnlockChallengeBattle = Tutorial09_Challenge_Unlock

function Tutorial09_Challenge_Check()
    if Tutorial09_Challenge_Unlock() then
        DCEI.SetFrameActive(ui.bottom_panel.right_inner.pip, false)
        Tutorial09_Challenge_StepA()
    end
end

function Tutorial09_Challenge_StepA()
    tutorial.state = 19

    local challenge_data = levels.challenge[1]
    local function on_click()
        ui.screen_mask.Hide()

        -- create new battle confirmation
        local item = { challenge = challenge_data, reward_name = challenge_data:RewardName(), index = 1 }
        ui.challenge.confirmation:Update(item)

        -- destroy dummy frame and reference
        DCEI.DestroyFrame(ui.challenge.dummy.frame)
        ui.challenge.dummy = nil

        -- continue tutorial
        Tutorial09_Challenge_StepB()
    end

    DCEI.BlockGameWorld()
    ui.screen_mask.upper:Show()

    ui.challenge.dummy = ui.NewChallengeItem(ui.upper, challenge_data, on_click)
    -- a wait is required to get the correct position of the target challenge frame
    local simple = true
    DCEI.WaitForUiUpdateToComplete(function()
        -- force selection of dummy challenge
        BlockInputExcept(ui.challenge.dummy.frame)

        -- move dummy item on top of original item
        local original_challenge_item = ui.challenge.items[1].frame
        local pos1 = DCEI.GetFramePosition3D(original_challenge_item)
        local pos2 = DCEI.GetFramePosition3D(ui.challenge.dummy.frame)
        local offset = pos1.y - pos2.y
        DCEI.SetFrameVerticalOffsetInParent(ui.challenge.dummy.frame, offset)

        local offset = -214
        local text = tutorial.message.challenge
        local options = {
            parent = ui.challenge.dummy.frame,
            scale = 0.75,
        }

        local delay = 0.125
        DCEI.TriggerAddTimerEventElapsed(function()
            -- display message and big-hand first challenge
            ui.prompt.helper = ui.NewBigHand(-100, 57, options)
            ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        end, delay, true, simple)
    end)
end

function Tutorial09_Challenge_StepB()
    if tutorial.state == 19 then
        tutorial.state = 20

        -- destroy previous step UI
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()

        -- big-hand battle button
        local options = {
            parent = ui.challenge.confirmation.button,
            scale = 0.75,
        }

        local simple = true
        local delay = 0.25
        DCEI.TriggerAddTimerEventElapsed(function()
            BlockInputExcept(ui.challenge.confirmation.button)
            ui.prompt.helper = ui.NewBigHand(57, 57, options)
        end, delay, true, simple)
    end
end

function Tutorial09_Challenge_StepC()
    if tutorial.state == 20 then
        tutorial.state = 0

        -- destroy previous step UI
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()
        UnblockInput()

        -- log colony tutorial as completed
        if not tutorial.challenge then
            tutorial.challenge = true
            bank.CommitChallengeState()
        end
    end
end

function CheckAnyGearEquipped()
    for k, v in pairs(player.gear) do
        if v ~= nil then
            return true
        end
    end
    return false
end

function Tutorial10_EquipGear_Check()
    if not CheckAnyGearEquipped() then
        player.system.gear_equip_tutorial = system.unlocked
        bank.CommitSystemState()
    end
end

function Tutorial10_EquipGear_Castle_Check()
    if player.system.gear_equip_tutorial == system.unlocked then
        Tutorial10_EquipGear_StepA()
    end
end

function Tutorial10_EquipGear_StepA()
    -- open gear menu
    tutorial.state = 21
    -- destroy previous step UI
    ui.DestroyBigHand()
    CloseActivePrompts()

    BlockInputExcept(ui.glue.gear_menu.frame)

    local offset = -214
    local text = tutorial.message.equip_gear
    local options = {
        parent = ui.glue.gear_menu.frame,
        flip = true,
        scale = 0.75,
    }

    ui.prompt.helper = ui.NewBigHand(-71, 36, options)
    ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
end

function Tutorial10_EquipGear_StepB()
    -- open gear card
    if tutorial.state == 21 or tutorial.state == 21.1 then
        local gear_slot = 6
        local h_offset = -142
        local v_offset = 220
        if tutorial.state == 21.1 then
            gear_slot = 3
            h_offset = 60
            v_offset = 310
        end
        tutorial.state = 22
        ui.DestroyBigHand()

        DCEI.SetScrollFrameScrollPosition(ui.gear_menu.gear_container, 0)
        DCEI.SetScrollFrameScrollEnable(ui.gear_menu.gear_container, false)

        BlockInputExcept(ui.gear_menu.gear_slots[gear_slot].button)

        local options = {
            scale = 0.75,
        }
        ui.prompt.helper = ui.NewBigHand(h_offset, v_offset, options)
    end
end

function Tutorial10_EquipGear_StepC()
    -- equip gear
    if tutorial.state == 22 then
        tutorial.state = 23
        ui.DestroyBigHand()
        CloseActivePrompts()

        BlockInputExcept(ui.unit_card.equip.button)

        ui.prompt.helper = ui.NewBigHand(71, -71)
    end
end

function Tutorial10_EquipGear_StepD()
    -- end tutorial
    if tutorial.state == 23 then
        tutorial.state = 0

        DCEI.SetScrollFrameScrollEnable(ui.gear_menu.gear_container, true)
        -- clear force input
        ui.DestroyBigHand()
        UnblockInput()

        -- set gear equip tutorial as completed
        player.system.gear_equip_tutorial = system.active
        bank.CommitSystemState()
    end
end

function TutorialHellMode01()
    tutorial.state = 24

    local function OnClose()
        TutorialHellMode02()
    end

    local offset = -360
    local text = DCEI.GetTranslationText("campaign/tutorial/hell_mode1")
    ui.prompt.tutorial = ui.NewPromptMessageNext(text, offset, OnClose, ui.super)
    BlockInputExcept(ui.prompt.tutorial.close_button)
    DCEI.SetFrameActive(ui.screen_mask.super.frame, true)

    DCEI.AnimateFrameScale(ui.prompt.tutorial.frame, { x = 1, y = 0.2, z = 1 }, { x = 1, y = 1, z = 1 }, 0.5, "OutBack")
    DCEI.AnimateFrameVerticalOffset(ui.hell_mode.Panel, 0, 100, 0.5, "OutBack")
    DCEI.SetFrameActive(ui.hell_mode.LeaderboardShortcut, false)
end

function TutorialHellMode02()
    if tutorial.state == 24 then
        tutorial.state = 25
        DCEI.SetFrameActive(ui.screen_mask.super.frame, false)

        ui.NewHighlightStencilMask(300, 70, -180, 426, nil, true)

        local function OnClose()
            TutorialHellMode03()
        end

        local offset = -360
        local text =
            DCEI.GetTranslationText("campaign/tutorial/hell_mode2", { y = "<size=90%><sprite name=gears></size>" })
        ui.prompt.tutorial = ui.NewPromptMessageNext(text, offset, OnClose, ui.super)
        BlockInputExcept(ui.prompt.tutorial.close_button)
    end
end

function TutorialHellMode03()
    if tutorial.state == 25 then
        tutorial.state = 26

        if ui.highlight_stencil and ui.highlight_stencil.Frame then
            DCEI.DestroyFrame(ui.highlight_stencil.Frame)
        end

        ui.NewHighlightStencilMask(320, 60, -160, 336, nil, true)

        local function OnClose()
            ui.DestroyBigHand()
            UnblockInput()
            if ui.highlight_stencil and ui.highlight_stencil.Frame then
                DCEI.DestroyFrame(ui.highlight_stencil.Frame)
            end
            DCEI.AnimateFrameVerticalOffset(ui.hell_mode.Panel, 100, 0, 0.5, "InBack")
            DCEI.SetFrameActive(ui.hell_mode.LeaderboardShortcut, true)
        end

        local offset = -360
        local text =
            DCEI.GetTranslationText("campaign/tutorial/hell_mode3", { y = "<size=90%><sprite name=gears></size>" })
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset, OnClose, ui.super)
        BlockInputExcept(data.hell_mode.attributes["HeroLevel"].item.Inc)

        local options = {
            scale = 0.75,
        }
        ui.prompt.helper = ui.NewBigHand(10, 370, options)
    end
end

function TutorialHellMode04Battle()
    tutorial.state = 27

    SetGameSpeed(0)
    local v_offset = -144
    if PlayingiOSVersion() then
        v_offset = -180
    end
    ui.NewHighlightStencilMask(220, 70, 130, v_offset, nil, true)
    DCEI.SetFrameTopAlignmentInParent(ui.highlight_stencil.Stencil)

    local function OnClose()
        TutorialHellMode05Battle()
    end

    local offset = -200
    local text = DCEI.GetTranslationText("campaign/tutorial/hell_mode4")
    ui.prompt.tutorial = ui.NewPromptMessageNext(text, offset, OnClose, ui.super)
    BlockInputExcept(ui.prompt.tutorial.close_button)
end

function TutorialHellMode05Battle()
    if tutorial.state == 27 then
        tutorial.state = 28

        if ui.highlight_stencil and ui.highlight_stencil.Frame then
            DCEI.DestroyFrame(ui.highlight_stencil.Frame)
        end

        local function OnClose()
            DCEI.SetFrameActive(ui.HellModePanel.info_panel.Frame, false)
            TutorialHellMode06Battle()
        end

        DCEI.SetFrameActive(ui.HellModePanel.info_panel.Frame, true)
        DCEI.SetFrameVerticalOffsetInParent(ui.HellModePanel.info_panel.Frame, 120)

        local offset = -300
        local text = DCEI.GetTranslationText("campaign/tutorial/hell_mode5")
        ui.prompt.tutorial = ui.NewPromptMessageNext(text, offset, OnClose, ui.super)
        BlockInputExcept(ui.prompt.tutorial.close_button)
    end
end

function TutorialHellMode06Battle()
    if tutorial.state == 28 then
        tutorial.state = 29

        local offset = -200
        local text =
            DCEI.GetTranslationText("campaign/tutorial/hell_mode6", { y = "<size=90%><sprite name=gears></size>" })
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset, nil, ui.super)
        BlockInputExcept(ui.HellModePanel.gem_icon.Frame)

        local options = {
            anchor_right = true,
            flip = true,
            scale = 0.75,
        }
        ui.prompt.helper = ui.NewBigHand(-160, 160, options)
    end
end

function TutorialHellMode06BattleEnd()
    tutorial.state = 0
    if ui.prompt.tutorial then
        ui.prompt.tutorial.Close()
    end
    ui.DestroyBigHand()
    UnblockInput()
end

-- !SECTION

-- SECTION checks
function TipCheckEmptyHeroSlots(unlock)
    local slot_count = unlock.hero_slots
    local equip_count = 0
    for _, unit in pairs(player.heroes) do
        if unit and DCEI.UnitExists(unit) then
            equip_count = equip_count + 1
        end
    end

    if (slot_count - equip_count) >= 2 then
        return true
    end
    return false
end

function TipCheckEmptyGearSlots(wave_progress)
    local count = 0
    for _ in pairs(player.gear) do
        count = count + 1
    end

    if
        (wave_progress > 20 and count < 3)
        or (wave_progress > 100 and count < 4)
        or (wave_progress > 200 and count < 5)
    then
        return true
    end
    return false
end

function TipCheckMineUpgrade()
    local mine_contract = GetMineContracts()
    local mine_level = GetDwarfLevel()
    if mine_contract >= 5 and mine_level < data.dwarf_workers.max_level then
        return true
    end
    return false
end

function TipCheckColony(wave_progress)
    local next_colony
    for _, colony in ipairs(levels.colony) do
        if not colony.complete then
            next_colony = colony
            break
        end
    end
    if next_colony then
        if wave_progress - next_colony.rating > 50 then
            return true
        end
    end
    return false
end

function TipCheckTownHallUpgrade()
    local max_level = data.town_upgrade.max_upgrade_level
    local level = GetTownUpgradeLevel()
    local town_contract = GetTownContracts()
    if level < max_level and town_contract >= 10 then
        return true
    end
    return false
end

function TipCheckEmptyTownBuildingSlots(unlock)
    local slot_count = unlock.town_slots
    local count = 0
    for _, unit_data in ipairs(data.town) do
        if unit_data.slot and unit_data.slot ~= 0 then
            count = count + 1
        end
    end
    if count < slot_count then
        return true
    end

    return false
end

function TipCheckHeroEquipment(wave_progress, unlock)
    if wave_progress > 110 then
        local slot_count = unlock.hero_slots
        local count = 0
        for k, unit in pairs(player.heroes) do
            local name = DCEI.GetUnitType(unit)
            local unit_data = data.heroes[name]

            if unit_data.equipment.weapon or unit_data.equipment.armor then
                count = count + 1
            end
        end
        if count < (slot_count * 0.5) then
            return true
        end
    end
    return false
end

function CheckDisplayTipOnBattleStats()
    local unlock = data.GetUnlockedSlotCount(GetCastleLevel())
    local wave_progress = wave.data.GetMax()
    local tip_key
    if TipCheckEmptyHeroSlots(unlock) then
        tip_key = "ui/menu/battle_stats/tip/hero"
    elseif TipCheckMineUpgrade() then
        tip_key = "ui/menu/battle_stats/tip/mine"
    elseif TipCheckEmptyGearSlots(wave_progress) then
        tip_key = "ui/menu/battle_stats/tip/gear"
    elseif TipCheckHeroEquipment(wave_progress, unlock) then
        tip_key = "ui/menu/battle_stats/tip/equipment"
    elseif TipCheckEmptyTownBuildingSlots(unlock) then
        tip_key = "ui/menu/battle_stats/tip/townbuilding"
    elseif TipCheckColony(wave_progress) then
        tip_key = "ui/menu/battle_stats/tip/colony"
    elseif TipCheckTownHallUpgrade() then
        tip_key = "ui/menu/battle_stats/tip/townhall"
    end

    return tip_key
end

function GetRaidBossUnlockWave()
    return 100
end

function AnyHeroesAboveLevel1()
    for k, unit_data in ipairs(data.heroes) do
        if unit_data:GetLevel() > 1 then
            return true
        end
    end
end

function CheckHeroLevelTutorial()
    local unlock_wave = 10
    local unlock_check = wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()

    -- show hero level-up tip post wave 10 if no heroes have been leveled
    if not AnyHeroesAboveLevel1() and unlock_check then
        local text = tutorial.message.hero_level
        local offset = -356

        ui.prompt.tutorial = ui.NewPromptMessage(text, offset, nil, ui.super)
        return true
    end
end

-- !SECTION

-- SECTION NEW CASTLE BUILD SYSTEM
function BuildTown(level, animate)
    -- BuildTown needs to run in its own thread as it contains waits
    DCEI.TriggerAddTimerEventElapsed(function()
        BuildTownActual(level, animate)
    end, 0)
end

function BuildTownActual(level, animate)
    level = (level > BUILD.max_town and BUILD.max_town) or level
    local build_func = BUILD.town[level]

    BUILD.type = BUILD.town
    BUILD.animate = animate

    if build_func then
        build_func()

        -- prop layers respect town unit buildings
        UpdateTownDistricts()

        UpdatePropLayer()
    end
end

function UpdateTownDistricts()
    local slot_count = BUILD.town.slot_count

    -- enforce prop layers from town slots
    for n = 1, slot_count do
        local build_func = BUILD.town_slot[n]
        if build_func then
            build_func()
        end
    end
end

function BuildCastle(level, animate)
    local build_level = (level > BUILD.max_castle and BUILD.max_castle) or level
    local build_func = BUILD.castle[build_level]
    -- LAYER_DATA_LEVEL = build_level

    if CASTLE_TEST then
        -- for testing castle build ugprades
        build_level = math.min(build_level, CASTLE_TEST + 1)
        build_func = BUILD.castle[build_level]
    end

    BUILD.type = BUILD.castle
    BUILD.animate = animate

    if level > BUILD.max_castle and not CASTLE_TEST then
        BUILD.animate = false
    end

    -- local test = BUILD.test
    -- if test then
    --     DCEI.LogMessage("TEST: " .. test)
    --     build_func = BUILD.castle[test]
    -- end

    if build_func then
        build_func()
        UpdatePropLayer()
    end
end

function CreateBuilder(name, x, y, z, rotation, skip_birth)
    if BUILD.animate then
        rotation = rotation or 0
        x = x or 0
        y = y or 0
        z = z or 0

        local unit = DCEI.CreateUnitAsync(1, 1, name, x, y)
        DCEI.SetUnitRotationEuler(unit, 0, rotation, 0)
        DCEI.SetUnitPosition3DAsync(unit, x, z, y)

        if not skip_birth then
            DCEI.ApplyBehaviorAsync(unit, DCEI.Behavior("Builder Birth"))
        end
        return unit
    end
end

function CreateBuilderDeath(name, x, y, z, rotation)
    local skip_birth = true
    local unit = CreateBuilder(name, x, y, z, rotation, skip_birth)
    if unit then
        DCEI.ApplyBehaviorAsync(unit, DCEI.Behavior("Builder Death"))
    end
end

function CreateBuilderNoFX(name, x, y, z, rotation)
    local unit = CreateBuilder(name, x, y, z, rotation)
    if unit then
        DCEI.ApplyBehaviorAsync(unit, DCEI.Behavior("Builder NoFX"))
    end
end

function WriteLayerData(layer, flag)
    local level = LAYER_DATA_LEVEL

    if not LAYER_DATA[layer] then
        LAYER_DATA[layer] = {
            on = -1,
            off = -1,
        }
    end

    local group = LAYER_DATA[layer]

    if flag then
        LAYER_DATA[layer].on = level
    else
        if group.off == -1 then
            LAYER_DATA[layer].off = level
        elseif group.on > group.off and level > group.on then
            LAYER_DATA[layer].off = level
        end
    end
end

function PrintLayerData()
    local print = "==== LAYER DATA ====" .. "\n"

    for k, v in pairs(LAYER_DATA) do
        print = print .. k .. "," .. v.on .. ", " .. v.off .. "\n"
        if v.on and v.off then
        end
    end

    DCEI.LogMessage(print)
end

local PROP_LAYERS = {}
local TEST_LAYERS = {}
function SetPropLayer(layer, flag)
    PROP_LAYERS[layer] = flag
    -- WriteLayerData(layer, flag)

    if CASTLE_TEST and CASTLE_TEST ~= GetCastleLevel() then
        local delay = 1
        TEST_LAYERS[layer] = false

        RevertPropLayer(delay)
    end
end

function SetPropLayerNow(layer, flag)
    SetPropLayer(layer, flag)

    if BUILD.animate then
        UpdatePropLayer()
    end
end

function UpdatePropLayer_AndHandleLoading()
    for layer, flag in pairs(PROP_LAYERS) do
        if flag then
            DCEI.LoadPropLayer(layer, flag)
        else
            DCEI.UnloadPropLayer(layer)
        end

        DCEI.ChangePropLayerDisplay(layer, flag)
    end
    PROP_LAYERS = {}
end

-- simulate a frame drop
function DropFrames()
    for x = 1, 12345 do
        for y = 1, 12345 do
            local z = 2
        end
    end
end

function UpdatePropLayer()
    for layer, flag in pairs(PROP_LAYERS) do
        DCEI.ChangePropLayerDisplay(layer, flag)
    end
    PROP_LAYERS = {}
end

function RevertPropLayer(delay)
    if not TEST_LAYERS.revert then
        TEST_LAYERS.revert = true

        DCEI.TriggerAddTimerEventElapsed(function()
            -- don't reset default layers
            TEST_LAYERS["_Default_Castle_Set00_Fence"] = false
            TEST_LAYERS["_Default_Castle_Set00_Hero04"] = false
            TEST_LAYERS["_Default_Castle_Set00_Tower01"] = false
            TEST_LAYERS["_Default_Castle_Set00_Tower02"] = false
            TEST_LAYERS["_Default_Castle_Set00_Tower03"] = false
            TEST_LAYERS["_Default_Castle_Set00_Tower04"] = false
            TEST_LAYERS["_Default_Castle_Set00_Deco"] = false
            TEST_LAYERS["_Default_Castle_Set01_Back"] = false
            TEST_LAYERS["Castle_Set01_Barricade01"] = true
            TEST_LAYERS["Castle_Set01_Barricade02"] = true
            TEST_LAYERS["Castle_Set01_Barricade03"] = true
            TEST_LAYERS["Castle_Set01_Barricade04"] = true

            -- revert changed layers
            for layer, flag in pairs(TEST_LAYERS) do
                DCEI.ChangePropLayerDisplay(layer, flag)
            end
            BuildCastle(CASTLE_TEST)
            TEST_LAYERS = {}
        end, delay)
    end
end

function DefineBuildOrder()
    BUILD.castle = {}
    BUILD.town_slot = {}
    BUILD.town = {}
    BUILD.town.slot_count = 0

    local function Include(n)
        local was_animated = BUILD.animate
        local build_type = BUILD.type
        BUILD.animate = false

        local build = build_type[n]
        if build then
            build()
        end

        BUILD.animate = was_animated
    end

    local function Pause(duration)
        if BUILD.animate then
            DCEI.Wait(duration)
        end
    end

    -- CASTLE SEQUENCE
    BUILD.castle[1] = function()
        -- castle defaults
        SetPropLayer("_Default_Castle_Set00_Fence", true)
        SetPropLayer("_Default_Castle_Set00_Tower01", true)
        SetPropLayer("_Default_Castle_Set00_Hero04", true)
        SetPropLayer("_Default_Castle_Set00_Tower02", true)
        SetPropLayer("_Default_Castle_Set00_Tower03", true)
        SetPropLayer("_Default_Castle_Set00_Tower04", true)

        SetPropLayer("_Default_Castle_Set01_Back", true)
    end
    BUILD.castle[2] = function()
        Include(1)
        -- 1st tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.5, 20.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 20.65)

        SetPropLayer("_Default_Castle_Set00_Tower01", false)
    end
    BUILD.castle[3] = function()
        Include(2)
        -- wooden archer platform
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Side"), 21.75, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 23.25, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24.75, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Side"), 26.25, 14.05, 0.28, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop Side"), 21.85, 14.81, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 23.25, 14.81, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 24.75, 14.81, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop Side"), 26, 14.78, 0.2)
        -- 4th hero slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 25.8, 17.5, 1.25)

        SetPropLayerNow("_Default_Castle_Set00_Hero04", false)
        Pause(0.25)
        SetPropLayer("Castle_Set01_Platform", true)
    end
    BUILD.castle[4] = function()
        Include(3)
        -- 2nd tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.4, 20.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.4, 20.65)

        SetPropLayer("_Default_Castle_Set00_Tower02", false)
    end
    BUILD.castle[5] = function()
        Include(4)
        -- front row of wood
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 21, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 27, 17.5, 0.2, 180)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop SideB"), 21.2, 17.6, 0.1, 315)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 22.5, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 24, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 25.5, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop SideB"), 26.8, 17.6, 0.1, -135)

        SetPropLayerNow("_Default_Castle_Set00_Fence", false)
        Pause(0.25)
        SetPropLayer("Castle_Set01_Fence", true)
        SetPropLayer("Castle_Set01_Floor_Top", true)
    end
    BUILD.castle[6] = function()
        Include(5)
        -- 5th hero slot
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 22.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 24, 16, 0.2, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 14.5, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 24, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero05", true)
    end
    BUILD.castle[7] = function()
        Include(6)
        -- wooden barricade 1
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 20.7, 19.5, 0, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade01", true)
    end
    BUILD.castle[8] = function()
        Include(7)
        -- wooden barricade 2
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 27.3, 19.5, 0)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade02", true)
    end
    BUILD.castle[9] = function()
        Include(8)
        -- 6th hero slot
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 25.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24, 16, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero05", false)
        SetPropLayer("Castle_Set01_Hero06", true)
    end
    BUILD.castle[10] = function()
        Include(9)
        -- floor decoration
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 22.85, 16.375, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 25.15, 16.375, 1.13, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set01_FloorDeco01", true)
    end
    BUILD.castle[11] = function()
        Include(10)
        -- lower towers
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.3, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.3, 1.25)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower"), 27.35, 17.3, -0.25, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower"), 20.8, 17.3, -0.25, 135)

        Pause(0.25)
        SetPropLayerNow("Castle_Set01_Towers", true)
    end
    BUILD.castle[12] = function()
        Include(11)
        -- hero slot 7
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 21, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 21, 14.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 19.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 19.5, 14.5, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero06", false)
        SetPropLayer("Castle_Set01_Hero07", true)
    end
    BUILD.castle[13] = function()
        Include(12)
        -- barricade 3
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 20.4, 21.4, 0)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade03", true)
    end
    BUILD.castle[14] = function()
        Include(13)
        -- barricade 4
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 27.6, 21.4, 0, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade04", true)
    end
    BUILD.castle[15] = function()
        Include(14)
        -- 8th hero slot
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 27, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 27, 14.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 28.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 28.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero07", false)
        SetPropLayer("Castle_Set01_Floor_Main", true)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
    end
    BUILD.castle[16] = function()
        Include(15)
        -- change floor decoration
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 21.6, 16.2, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 24, 16.375, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 26.4, 16.2, 1.13, 90)

        SetPropLayerNow("Castle_Set01_FloorDeco01", false)
        Pause(0.25)
        SetPropLayer("Castle_Set01_FloorDeco02", true)
    end
    BUILD.castle[17] = function()
        Include(16)
        -- 3rd tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.45, 22.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.45, 22.65)

        SetPropLayer("_Default_Castle_Set00_Tower03", false)
    end
    BUILD.castle[18] = function()
        Include(17)
        -- upper towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.9, -0.5, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 14.05, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.9, 13.9, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.3, 14.05, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set01_TowersBig", true)
    end
    BUILD.castle[19] = function()
        Include(18)
        -- 4th tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.55, 22.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.55, 22.65)

        SetPropLayer("_Default_Castle_Set00_Tower04", false)
    end
    BUILD.castle[20] = function()
        Include(19)
        -- castle decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 16.45, 1.2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 16.45, 1.2)

        SetPropLayer("Castle_Set01_Deco01", true)
        SetPropLayer("Castle_Set01_Deco02", true)
    end
    -- CASTLE SET 02: Reinforced Wood
    BUILD.castle[21] = function()
        Include(20)
        -- reinforce archer platform
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21.75, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 23.25, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24.75, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 26.25, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 21.75, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 23.25, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 24.75, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 26.25, 14.875, 0.25, 180)

        Pause(0.125)
        SetPropLayerNow("_Default_Castle_Set01_Back", false)
        SetPropLayerNow("Castle_Set01_Platform", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Platform", true)
        SetPropLayer("Castle_Set02_Back", true)
    end
    BUILD.castle[22] = function()
        Include(21)
        -- reinforce floor
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 14.5, 0.2, 180)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_Floor_Top", false)
        SetPropLayerNow("Castle_Set01_Floor_Main", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Floor", true)
    end
    BUILD.castle[23] = function()
        Include(22)
        -- reinforce fence
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop Side"), 21, 18.2, 0.15, 90)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 22.5, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 24, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 25.5, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop Side"), 27, 18.2, 0.15, 180)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Fence", true)
    end
    BUILD.castle[24] = function()
        Include(23)
        -- reinforce big towers
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower Big"), 20.8, 14, 0.7, -90)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower Big"), 27.25, 14, 0.7, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set02_TowersBig", true)
    end
    BUILD.castle[25] = function()
        Include(24)
        -- reinforce small towers
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower"), 20.7, 17.4, 0.85, 135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower"), 27.25, 17.4, 0.85, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Towers", true)
    end
    BUILD.castle[26] = function()
        Include(25)
        -- reinforce floor decoration
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 15.9, 1.3)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_FloorDeco02", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_FloorDeco", true)
    end
    BUILD.castle[27] = function()
        Include(26)
        -- add left furnace
        CreateBuilder(DCEI.Unit("Builder Set02 Furnace"), 20.3, 16.8, 0.2, -45)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Furnace01", true)
    end
    BUILD.castle[28] = function()
        Include(27)
        -- add right furnace
        CreateBuilder(DCEI.Unit("Builder Set02 Furnace"), 27.7, 16.67, 0.2, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Furnace02", true)
    end
    BUILD.castle[29] = function()
        Include(28)
        -- more weapon racks (left)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.7, 16.2, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.5, 16.9, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.4, 17.7, 1.25)

        SetPropLayer("Castle_Set01_Deco01", false)
        SetPropLayer("Castle_Set02_Deco01", true)
    end
    BUILD.castle[30] = function()
        Include(29)
        -- more weapon racks (right)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.3, 16.1, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.6, 16.8, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.7, 17.7, 1.25)

        SetPropLayer("Castle_Set01_Deco02", false)
        SetPropLayer("Castle_Set02_Deco02", true)
    end
    -- CASTLE SET 03: Classic Stone
    BUILD.castle[31] = function()
        -- reset construction site (wood)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set01_Towers", true)
        SetPropLayer("Castle_Set01_TowersBig", true)
        SetPropLayer("Castle_Set02_Back", true)

        -- fence
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set02_Floor", false)
        SetPropLayerNow("Castle_Set03_Construction_FloorDeco", true)
        SetPropLayerNow("Castle_Set03_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 14.5, 0.2, 180)

        -- platform
        SetPropLayerNow("Castle_Set02_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21.75, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 23.25, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24.75, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 26.25, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 21.75, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 23.25, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 24.75, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 26.25, 14.875, 0.25, 180)

        -- towers
        SetPropLayerNow("Castle_Set02_Towers", false)
        SetPropLayerNow("Castle_Set02_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower"), 20.7, 17.4, 0.85, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower"), 27.25, 17.4, 0.85, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower Big"), 20.8, 14, 0.7, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower Big"), 27.25, 14, 0.7, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.9, 14.05, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.6, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.32, 17.22, -0.75, -135)

        Pause(0.125)
        -- previous deco
        SetPropLayerNow("Castle_Set02_FloorDeco", false)
        SetPropLayerNow("Castle_Set02_Deco01", false)
        SetPropLayerNow("Castle_Set02_Deco02", false)
        SetPropLayerNow("Castle_Set02_Furnace01", false)
        SetPropLayerNow("Castle_Set02_Furnace02", false)
        SetPropLayerNow("Castle_Set02_Fence", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set02_Back", false)
        SetPropLayer("Castle_Set03_Back", true)
        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set03_Construction_Towers", true)
        SetPropLayer("Castle_Set03_Construction_TowersBig", true)
    end
    BUILD.castle[32] = function()
        Include(31)
        -- stone platform
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 21.75, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 23.25, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 24.75, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 26.25, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 21.75, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 23.25, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 24.75, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 26.25, 14.125, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set03_Platform", true)
    end
    BUILD.castle[33] = function()
        Include(32)
        -- stone fence
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Side"), 20.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 21.75, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 23.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 24.75, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 26.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Side"), 27.75, 18.125, 0.2, 90)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set03_Fence", true)
    end
    BUILD.castle[34] = function()
        Include(33)
        -- stone floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor01", true)
    end
    BUILD.castle[35] = function()
        Include(34)
        -- stone floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor02", true)
    end
    BUILD.castle[36] = function()
        Include(35)
        -- stone floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor03", true)
    end
    BUILD.castle[37] = function()
        Include(36)
        -- stone floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor04", true)
    end
    BUILD.castle[38] = function()
        Include(37)
        -- stone floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 16.7, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 15.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 14.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 16.7, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 15.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 14.2, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Floor", false)
        SetPropLayer("Castle_Set03_Floor05", true)
    end
    BUILD.castle[39] = function()
        Include(38)
        -- swap to reinforced wood floor deco (looks better)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 15.9, 1.3)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_FloorDeco", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_FloorDeco", true)
    end
    BUILD.castle[40] = function()
        Include(39)
        -- small towers 1/2
        SetPropLayerNow("Castle_Set03_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.6, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.32, 17.22, -0.75, -135)
        SetPropLayerNow("Castle_Set01_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.35, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.8, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.3, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.3, 1.25)

        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 20.65, 17.89, -0.9, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 27.34, 17.84, -0.9, -45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Base"), 20.65, 18, 0.5, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Base"), 27.35, 18, 0.5, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Towers01", true)
        SetPropLayer("Castle_Set03_TowersC", true)
    end
    BUILD.castle[41] = function()
        Include(40)
        -- small towers 2/2
        SetPropLayerNow("Castle_Set03_TowersC", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 20.65, 17.89, -0.9, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 27.34, 17.84, -0.9, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Top"), 20.65, 18, 0.5, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Top"), 27.35, 18, 0.5, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Towers02", true)
    end
    BUILD.castle[42] = function()
        Include(41)
        -- big towers 1/2
        SetPropLayerNow("Castle_Set03_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.9, 14.05, 1.15, -25)

        SetPropLayerNow("Castle_Set01_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.9, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 14.05, -0.5, 90)

        CreateBuilderNoFX(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 20.82, 13.92, -1.85, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 27.2, 13.94, -1.75)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 20.8, 14.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 27.2, 14.1, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.8, 14, 2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.2, 14, 2)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBig01", true)
        SetPropLayer("Castle_Set03_TowersBigC", true)
    end
    BUILD.castle[43] = function()
        Include(42)
        -- big towers 2/2
        SetPropLayerNow("Castle_Set03_TowersBigC", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 20.82, 13.92, -1.85, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 27.2, 13.94, -1.75)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 20.8, 14.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 27.2, 14.1, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 14, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.2, 14, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBig02", true)
    end
    BUILD.castle[44] = function()
        Include(43)
        -- small tower flags
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 20.65, 18, 2.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 27.35, 18, 2.1)
        SetPropLayer("Castle_Set03_TowersDeco", true)
    end
    BUILD.castle[45] = function()
        Include(44)
        -- platform stairs
        CreateBuilder(DCEI.Unit("Builder Set03 StoneStairs"), 27.25, 15.75, 1.25, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneStairs"), 20.75, 15.5, 1.25, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Stairs", true)
    end
    BUILD.castle[46] = function()
        Include(45)
        -- left crates
        CreateBuilder(DCEI.Unit("Builder Set03 Crates"), 20, 16.9, 1.2, -140)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Deco01", true)
    end
    BUILD.castle[47] = function()
        Include(46)
        -- right crates
        CreateBuilder(DCEI.Unit("Builder Set03 Crates"), 28.25, 17, 1.2, -95)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Deco02", true)
    end
    BUILD.castle[48] = function()
        Include(47)
        -- sword decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.8, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.8, 1.25)

        SetPropLayer("Castle_Set03_Deco03", true)
    end
    BUILD.castle[49] = function()
        Include(48)
        -- extra big towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 19.75, 14, 0, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 28.25, 14.2, 0, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.4, 14, 2.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.6, 14.2, 2.5)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBigger", true)
    end
    BUILD.castle[50] = function()
        Include(49)
        -- shield decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.8, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.8, 1.25)
        SetPropLayer("Castle_Set03_Deco04", true)
    end
    BUILD.castle[51] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set03_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set03_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set02_FloorDeco", false)
        SetPropLayerNow("Castle_Set03_Floor01", false)
        SetPropLayerNow("Castle_Set03_Floor02", false)
        SetPropLayerNow("Castle_Set03_Floor03", false)
        SetPropLayerNow("Castle_Set03_Floor04", false)
        SetPropLayerNow("Castle_Set03_Floor05", false)
        SetPropLayerNow("Castle_Set04_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 16.7, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 15.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 14.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 16.7, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 15.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 14.2, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set03_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 21.75, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 23.25, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 24.75, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 26.25, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 23.25, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 21.75, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 24.75, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 26.25, 14.125, 0.35)

        -- towers
        SetPropLayerNow("Castle_Set03_Towers01", false)
        SetPropLayerNow("Castle_Set03_Towers02", false)
        SetPropLayerNow("Castle_Set03_TowersBig01", false)
        SetPropLayerNow("Castle_Set03_TowersBig02", false)
        SetPropLayerNow("Castle_Set03_TowersBigger", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Base"), 20.65, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Base"), 27.35, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Top"), 20.65, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Top"), 27.35, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 20.8, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 27.2, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 20.8, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 27.2, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 19.75, 14, 0, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 28.25, 14.2, 0, 45)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        Pause(0.125)
        -- previous deco
        SetPropLayerNow("Castle_Set03_TowersDeco", false)
        SetPropLayerNow("Castle_Set03_Stairs", false)
        SetPropLayerNow("Castle_Set03_Deco01", false)
        SetPropLayerNow("Castle_Set03_Deco02", false)
        SetPropLayerNow("Castle_Set03_Deco03", false)
        SetPropLayerNow("Castle_Set03_Deco04", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set03_Back", false)
        SetPropLayer("Castle_Set04_Back", true)

        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set03_Construction_Platform", true)
        SetPropLayer("Castle_Set04_Construction_Towers", true)
        SetPropLayer("Castle_Set04_Construction_TowersBig", true)
    end
    BUILD.castle[52] = function()
        Include(51)
        -- desert platform
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 21.75, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 23.25, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 24.75, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 26.25, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.25, 14.75, 0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.23, 13.3, 0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.65, 14.75, 0.5, 270)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.63, 13.3, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 20.75, 14, 0.23, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 27.25, 14, 0.23, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 21.75, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 23.25, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 24.75, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 26.25, 14, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set04_Platform", true)
    end
    BUILD.castle[53] = function()
        Include(52)
        -- desert fence
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 21, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 22.5, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 24, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 25.5, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 27, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Convex"), 20.25, 18.13, -0.35)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Convex"), 27.75, 18.13, -0.35, 90)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set04_Fence", true)
    end
    BUILD.castle[54] = function()
        Include(53)
        -- desert floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor01", true)
    end
    BUILD.castle[55] = function()
        Include(54)
        -- desert floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor02", true)
    end
    BUILD.castle[56] = function()
        Include(55)
        -- desert floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor03", true)
    end
    BUILD.castle[57] = function()
        Include(56)
        -- desert floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor04", true)
    end
    BUILD.castle[58] = function()
        Include(57)
        -- desert floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 19.85, 17, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 28.15, 17, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Alt"), 20.88, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Alt"), 27.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor05", true)
    end
    BUILD.castle[59] = function()
        Include(58)
        -- desert floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.5, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.5, 15.83, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 15.83, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.5, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.5, 15.83, 1.15, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set04_Construction_Floor", false)

        Pause(0.125)
        SetPropLayer("Castle_Set04_FloorDeco", true)
    end
    BUILD.castle[60] = function()
        Include(59)
        -- desert towers
        SetPropLayerNow("Castle_Set04_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTower"), 20.3, 17, 0.6, -135)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTower"), 27.7, 17, 0.6, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Towers", true)
    end
    BUILD.castle[61] = function()
        Include(60)
        -- desert towers big
        SetPropLayerNow("Castle_Set04_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertTower Big"), 20.75, 14, -0.25, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertTower Big"), 27.25, 14, -0.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 14.25, 1.8)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.2, 14.25, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set04_TowersBig", true)
    end
    BUILD.castle[62] = function()
        Include(61)
        -- upper braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 17.8, 0.95, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 17.8, 0.95, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.75, 18, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.25, 18, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoBrazier01", true)
    end
    BUILD.castle[63] = function()
        Include(62)
        -- left ring
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Ring"), 20.75, 15.8, 1.38)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.75, 15.8, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing01", true)
    end
    BUILD.castle[64] = function()
        Include(63)
        -- right ring
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Ring"), 27.25, 15.8, 1.38)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.25, 15.8, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing02", true)
    end
    BUILD.castle[65] = function()
        Include(64)
        -- totems
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Totem"), 21.5, 17.5, 1.25, 155)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Totem"), 26.5, 17.5, 1.25, -165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.5, 17.6, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.5, 17.6, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoTotems", true)
    end
    BUILD.castle[66] = function()
        Include(65)
        -- left ring big
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco RingBig"), 20.8, 14, 1.35)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 21, 14, 1.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing03", true)
    end
    BUILD.castle[67] = function()
        Include(66)
        -- right ring big
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco RingBig"), 27.2, 14, 1.4)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27, 14, 1.45)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing04", true)
    end
    BUILD.castle[68] = function()
        Include(67)
        -- lower braziers
        CreateBuilder(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 15.8, 0.8)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 15.8, 0.8)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.75, 15.8, 1.4)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.25, 15.8, 1.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoBrazier02", true)
    end
    BUILD.castle[69] = function()
        Include(68)
        -- spears
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.3, 17, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.7, 17, 1.3)

        Pause(0.125)
        SetPropLayer("Castle_Set04_DecoSpears01", true)
    end
    BUILD.castle[70] = function()
        Include(69)
        -- more spears
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 22, 18.3, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26, 18.3, 1.5)

        Pause(0.125)
        SetPropLayer("Castle_Set04_DecoSpears02", true)
    end
    BUILD.castle[71] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set04_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set04_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set04_FloorDeco", false)
        SetPropLayerNow("Castle_Set04_Floor01", false)
        SetPropLayerNow("Castle_Set04_Floor02", false)
        SetPropLayerNow("Castle_Set04_Floor03", false)
        SetPropLayerNow("Castle_Set04_Floor04", false)
        SetPropLayerNow("Castle_Set04_Floor05", false)
        SetPropLayerNow("Castle_Set05_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 19.85, 17, 0.17)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 28.15, 17, 0.17)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Alt"), 20.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Alt"), 27.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 14.95, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set04_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop"), 21.75, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 23.25, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 24.75, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop"), 26.25, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.25, 14.75, 0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.23, 13.3, 0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.65, 14.75, 0.5, 270)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.63, 13.3, 0.5)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 20.75, 14, 0.23, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 27.25, 14, 0.23, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 21.75, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 23.25, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 24.75, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 26.25, 14, 0.25)

        -- towers
        SetPropLayerNow("Castle_Set04_Towers", false)
        SetPropLayerNow("Castle_Set04_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower"), 20.3, 17, 0.6, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower"), 27.7, 17, 0.6, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower Big"), 20.75, 14, -0.25, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower Big"), 27.25, 14, -0.25)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        -- deco
        SetPropLayerNow("Castle_Set04_DecoTotems", false)
        SetPropLayerNow("Castle_Set04_DecoBrazier02", false)
        SetPropLayerNow("Castle_Set04_DecoBrazier01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 17.8, 0.95, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 17.8, 0.95, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 15.8, 0.8)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 15.8, 0.8)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Totem"), 21.5, 17.5, 1.25, 155)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Totem"), 26.5, 17.5, 1.25, -165)

        Pause(0.125)
        -- instant deco
        SetPropLayerNow("Castle_Set04_DecoSpears02", false)
        SetPropLayerNow("Castle_Set04_DecoSpears01", false)
        SetPropLayerNow("Castle_Set04_DecoRing04", false)
        SetPropLayerNow("Castle_Set04_DecoRing03", false)
        SetPropLayerNow("Castle_Set04_DecoRing02", false)
        SetPropLayerNow("Castle_Set04_DecoRing01", false)
        SetPropLayerNow("Castle_Set04_FloorDeco", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set04_Back", false)
        SetPropLayer("Castle_Set05_Back", true)

        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set04_Construction_Towers", true)
        SetPropLayer("Castle_Set04_Construction_TowersBig", true)
    end
    BUILD.castle[72] = function()
        Include(71)
        -- dark platform
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 21.75, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 23.25, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 24.75, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 26.25, 14.1, 0.25)

        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 22.5, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 24, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 25.5, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 13.5, 0.25, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 14.95, 0.25, 90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 13.5, 0.25, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set05_Platform", true)
    end
    BUILD.castle[73] = function()
        Include(72)
        -- dark fence
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 21.75, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 23.25, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 24.75, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 26.25, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Concave"), 20.25, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Concave"), 27.75, 18.3, 0.2, 90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop ConcaveAlt"), 20.25, 16.8, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop ConcaveAlt"), 27.75, 16.8, 0.2, 90)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set05_Fence", true)
        SetPropLayer("Castle_Set05_FenceDeco01", true)
    end
    BUILD.castle[74] = function()
        Include(73)
        -- dark floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor01", true)
    end
    BUILD.castle[75] = function()
        Include(74)
        -- dark floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor02", true)
    end
    BUILD.castle[76] = function()
        Include(75)
        -- dark floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor03", true)
    end
    BUILD.castle[77] = function()
        Include(76)
        -- dark floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor04", true)
    end
    BUILD.castle[78] = function()
        Include(77)
        -- dark floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor05", true)
    end
    BUILD.castle[79] = function()
        Include(78)
        -- dark floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 21.375, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 21.375, 15.85, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 23.75, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 23.75, 15.85, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 26.125, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 26.125, 15.85, 0.125, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set05_Construction_Floor", false)

        Pause(0.125)
        SetPropLayer("Castle_Set05_FloorDeco", true)
    end
    BUILD.castle[80] = function()
        Include(79)
        -- dark towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 17.41, 1.35, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 28.27, 16.85, 1.35, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 16.28, 1.35, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.13, 16.85, 1.35, -45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 17.33, 1.43, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 28.18, 16.85, 1.43, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 16.38, 1.43, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.22, 16.85, 1.43, -45)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Deco"), 27.7, 16.85, 1.31, 165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.7, 16.85, 1.3)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 17.41, 1.35, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.87, 16.85, 1.35, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 16.28, 1.35, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 19.73, 16.85, 1.35, -45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 17.33, 1.43, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.78, 16.85, 1.43, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 16.38, 1.43, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 19.82, 16.85, 1.43, -45)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Deco"), 20.3, 16.85, 1.31, 165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.3, 16.85, 1.3)

        SetPropLayerNow("Castle_Set04_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Towers_Left", true)
        SetPropLayer("Castle_Set05_Towers_Right", true)
    end
    BUILD.castle[81] = function()
        Include(80)
        -- dark towers big
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 14.65, 0.95)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 14.65, 0.95, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 13.36, 0.95, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 13.36, 0.95, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 14.55, 1.08)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 14.55, 1.08, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 13.45, 1.08, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 13.45, 1.08, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 20.9, 14.01, 0.88, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.9, 14.01, 1.5)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 14.65, 0.95)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 14.65, 0.95, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 13.36, 0.95, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 13.36, 0.95, -90)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 14.55, 1.08)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 14.55, 1.08, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 13.45, 1.08, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 13.45, 1.08, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 27.13, 14.01, 0.88, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.13, 14.01, 1.5)

        SetPropLayerNow("Castle_Set04_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)

        Pause(0.25)
        SetPropLayer("Castle_Set05_TowersBig", true)
    end
    BUILD.castle[82] = function()
        Include(81)
        -- dark grate 1/4
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 20.9, 15.85, -0.1, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.9, 15.85, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate01", true)
    end
    BUILD.castle[83] = function()
        Include(82)
        -- dark grate 2/4
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 27.1, 15.85, -0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.1, 15.85, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate02", true)
    end
    BUILD.castle[84] = function()
        Include(83)
        -- dark fence deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTop Alt"), 21.75, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Alt"), 21.75, 18.3, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTop Alt"), 26.25, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Alt"), 26.25, 18.3, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_FenceDeco01", false)
        SetPropLayer("Castle_Set05_FenceDeco02", true)
    end
    BUILD.castle[85] = function()
        Include(84)
        -- dark fence braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 21.75, 18.3, 0, 70)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 26.25, 18.3, 0, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.75, 18.3, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.25, 18.3, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoBraziers", true)
    end
    BUILD.castle[86] = function()
        Include(85)
        -- dark grate 3/4
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Grate"), 21, 17.5, 0.22, -45)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate03", true)
    end
    BUILD.castle[87] = function()
        Include(86)
        -- dark grate 4/4
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Grate"), 27, 17.5, 0.22, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate04", true)
    end
    BUILD.castle[88] = function()
        Include(87)
        -- dark left big brazier
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 20.85, 14.05, 0.65)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.85, 14.05, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal01", true)
    end
    BUILD.castle[89] = function()
        Include(88)
        -- dark right big brazier
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 27.15, 14.05, 0.65)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.15, 14.05, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal02", true)
    end
    BUILD.castle[90] = function()
        Include(89)
        -- deco fence final
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Mast"), 24, 18.29, 0.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal03", true)
    end
    BUILD.castle[91] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set05_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set05_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set05_FloorDeco", false)
        SetPropLayerNow("Castle_Set05_Floor01", false)
        SetPropLayerNow("Castle_Set05_Floor02", false)
        SetPropLayerNow("Castle_Set05_Floor03", false)
        SetPropLayerNow("Castle_Set05_Floor04", false)
        SetPropLayerNow("Castle_Set05_Floor05", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 14.95, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        SetPropLayerNow("Castle_Set05_Platform", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 21.75, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 23.25, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 24.75, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 26.25, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 22.5, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 24, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 25.5, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 13.5, 0.25, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 14.95, 0.25, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 13.5, 0.25, 180)

        -- towers
        SetPropLayerNow("Castle_Set05_Towers_Left", false)
        SetPropLayerNow("Castle_Set05_Towers_Right", false)
        SetPropLayerNow("Castle_Set05_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 17.41, 1.35, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 28.27, 16.85, 1.35, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 16.28, 1.35, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.13, 16.85, 1.35, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 17.33, 1.43, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 28.18, 16.85, 1.43, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 16.38, 1.43, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.22, 16.85, 1.43, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Deco"), 27.7, 16.85, 1.31, 165)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 17.41, 1.35, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.87, 16.85, 1.35, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 16.28, 1.35, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 19.73, 16.85, 1.35, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 17.33, 1.43, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.78, 16.85, 1.43, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 16.38, 1.43, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 19.82, 16.85, 1.43, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Deco"), 20.3, 16.85, 1.31, 165)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 14.65, 0.95)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 14.65, 0.95, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 13.36, 0.95, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 13.36, 0.95, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 14.55, 1.08)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 14.55, 1.08, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 13.45, 1.08, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 13.45, 1.08, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 20.9, 14.01, 0.88, 120)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 14.65, 0.95)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 14.65, 0.95, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 13.36, 0.95, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 13.36, 0.95, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 14.55, 1.08)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 14.55, 1.08, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 13.45, 1.08, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 13.45, 1.08, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 27.13, 14.01, 0.88, 120)

        -- deco
        SetPropLayerNow("Castle_Set05_DecoFinal01", false)
        SetPropLayerNow("Castle_Set05_DecoFinal02", false)
        SetPropLayerNow("Castle_Set05_DecoFinal03", false)
        SetPropLayerNow("Castle_Set05_DecoGrate01", false)
        SetPropLayerNow("Castle_Set05_DecoGrate02", false)
        SetPropLayerNow("Castle_Set05_DecoGrate03", false)
        SetPropLayerNow("Castle_Set05_DecoGrate04", false)
        SetPropLayerNow("Castle_Set05_FenceDeco02", false)
        SetPropLayerNow("Castle_Set05_DecoBraziers", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 20.85, 14.05, 0.65)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 27.15, 14.05, 0.65)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Mast"), 24, 18.29, 0.3)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 20.9, 15.85, -0.1, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 27.1, 15.85, -0.1)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Grate"), 21, 17.5, 0.22, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Grate"), 27, 17.5, 0.22, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Alt"), 26.25, 18.3, 0.2)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Alt"), 21.75, 18.3, 0.2)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 21.75, 18.3, 0, 70)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 26.25, 18.3, 0, 120)

        -- new rubble
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 21.1, 17.2, 1, -60)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.8, 15.65, 1.2, -120)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 19.7, 16, 1, 129)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 20.2, 15.2, 1, 172)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.6, 14.15, 0.85, -40)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble"), 20.1, 13.95, 1.2, 51)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 26.9, 17.05, 1, 130)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 28.5, 15.85, 1, -37)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.45, 15.7, 1.2, 150)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 27.5, 15.4, 1, 259)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.25, 14.35, 0.85, 121)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble"), 27.5, 13.5, 1.2, -71)

        Pause(0.125)
        SetPropLayerNow("Castle_Set06_Construction", true)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set06_Construction_TowersBig", true)

        SetPropLayer("Castle_Set05_Back", false)
        SetPropLayer("Castle_Set06_Back", true)
        SetPropLayer("Castle_Set06_Rubble01", true)
        SetPropLayer("Castle_Set06_Rubble02", true)
        SetPropLayer("Castle_Set06_Rubble03", true)
        SetPropLayer("Castle_Set06_Rubble04", true)
        SetPropLayer("Castle_Set06_RubbleLeft", true)
        SetPropLayer("Castle_Set06_RubbleRight", true)
        SetPropLayer("Castle_Set06_RubbleTowers", true)
    end
    BUILD.castle[92] = function()
        Include(91)
        -- ancient platform
        SetPropLayerNow("Castle_Set06_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.6, 14.15, 0.85, -40)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.25, 14.35, 0.85, 121)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 21.75, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 23.25, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 24.75, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 26.25, 14, 0.25, 180)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 20.95, 14.85, 0.5, 0)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 21.05, 13.42, 0.5, -90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 26.95, 14.95, 0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 27.05, 13.52, 0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 22.5, 14.9, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 24, 14.9, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 25.5, 14.9, 0.5)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Platform", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_Platform", true)
    end
    BUILD.castle[93] = function()
        Include(92)
        -- ancient fence
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower"), 20, 17.15, 0.5, 135)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower"), 28, 17.15, 0.5, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFence Corner"), 20.26, 18.27, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFence Corner"), 27.75, 18.27, 1.3, 90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 21.54, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 22.75, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 24, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 25.25, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 26.51, 18.24, 0.4, 180)

        Pause(0.125)
        SetPropLayer("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_Fence01", true)
    end
    BUILD.castle[94] = function()
        Include(93)
        -- ancient floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 17.47, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor01", true)
    end
    BUILD.castle[95] = function()
        Include(94)
        -- ancient floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 17.47, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor02", true)
    end
    BUILD.castle[96] = function()
        Include(95)
        -- ancient floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 16.23, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 14.98, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 14.98, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor03", true)
    end
    BUILD.castle[97] = function()
        Include(96)
        -- ancient floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 16.23, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 14.98, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 14.98, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor04", true)
    end
    BUILD.castle[98] = function()
        Include(97)
        -- ancient floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 13.73, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor05", true)
    end
    BUILD.castle[99] = function()
        Include(98)
        -- ancient floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.625, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.625, 15.75, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 15.75, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.375, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.375, 15.75, 1.2, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set06_Construction", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_FloorDeco", true)
    end
    BUILD.castle[100] = function()
        Include(99)
        -- unlock hero slot 9
        SetPropLayer("Castle_Set06_Rubble01", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 21.1, 17.2, 1, -60)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 21.1, 17.2, 1.3)
    end
    BUILD.castle[101] = function()
        Include(100)
        -- ancient towers big
        SetPropLayerNow("Castle_Set06_RubbleTowers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble"), 20.1, 13.95, 1.2, 51)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble"), 27.5, 13.5, 1.2, -71)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower Big"), 21.08, 14.13, 0.5, 170)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower Big"), 26.92, 14.21, 0.5, -38)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig", true)
    end
    BUILD.castle[102] = function()
        Include(101)
        -- ancient towers reinforce bottom
        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 20.28, 13.5, 0.3, -90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 27.73, 13.5, 0.3, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Side01", true)
    end
    BUILD.castle[103] = function()
        Include(102)
        -- ancient towers reinforce left
        SetPropLayerNow("Castle_Set06_RubbleLeft", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 20.2, 15.2, 1, 172)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 20.25, 14.9, 0.3, 0)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Side02", true)
    end
    BUILD.castle[104] = function()
        Include(103)
        -- ancient towers reinforce right
        SetPropLayerNow("Castle_Set06_RubbleRight", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 27.5, 15.4, 1, 259)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 27.7, 14.9, 0.3, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Side03", true)
    end
    BUILD.castle[105] = function()
        Include(104)
        -- unlock hero slot 10
        SetPropLayer("Castle_Set06_Rubble02", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 26.9, 17.05, 1, 130)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 26.9, 17.05, 1.3)
    end
    BUILD.castle[106] = function()
        Include(105)
        -- ancient stone barricades
        SetPropLayerNow("Castle_Set01_Barricade01", false)
        SetPropLayerNow("Castle_Set01_Barricade02", false)
        SetPropLayerNow("Castle_Set01_Barricade03", false)
        SetPropLayerNow("Castle_Set01_Barricade04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 20.7, 19.5, 0, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 27.3, 19.5, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 20.4, 21.4, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 27.6, 21.4, 0, 180)

        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade"), 20.5, 21.65, -0.7, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 19.88, 21.65, -1.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade"), 27.5, 21.63, -0.7, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 28.05, 21.63, -1.1, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.25, 21.65, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.75, 21.63, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 19.75, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.5, 19.75, 0.1)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Barricade", true)
    end
    BUILD.castle[107] = function()
        Include(106)
        -- reinforce fence
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 21.75, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 23.2, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 24.8, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 26.2, 18.23, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Fence02", true)
    end
    BUILD.castle[108] = function()
        Include(107)
        -- big tower deco left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower BigDeco"), 21.13, 14.23, -0.55, -90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 21.13, 14.43, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Deco01", true)
    end
    BUILD.castle[109] = function()
        Include(108)
        -- big tower deco right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower BigDeco"), 26.88, 14.18, -0.55, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 26.88, 14.45, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Deco02", true)
    end
    BUILD.castle[110] = function()
        Include(109)
        -- unlock hero slot 11
        SetPropLayer("Castle_Set06_Rubble03", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.8, 15.65, 1.2, -120)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 19.7, 16, 1, 129)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 16, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 19.7, 16.15, 1.3)
    end
    BUILD.castle[111] = function()
        Include(110)
        -- fence braziers
        CreateBuilder(DCEI.Unit("Builder Set06 AncientDeco Brazier"), 20.95, 18.3, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientDeco Brazier"), 27.1, 18.3, 0.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Fence03", true)
    end
    BUILD.castle[112] = function()
        Include(111)
        -- ancient towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower Top"), 20, 17.15, 0.5, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower Top"), 28, 17.15, 0.5, -45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20, 17.15, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 28, 17.15, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Towers", true)
    end
    BUILD.castle[113] = function()
        Include(112)
        -- big tower statue left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Statue"), 21.2, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 14.3, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Statue01", true)
    end
    BUILD.castle[114] = function()
        Include(113)
        -- big tower statue right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Statue"), 26.9, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 14.3, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Statue02", true)
    end
    BUILD.castle[115] = function()
        Include(114)
        -- unlock hero slot 12
        SetPropLayer("Castle_Set06_Rubble04", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 28.5, 15.85, 1, -37)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.45, 15.7, 1.2, 150)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 28.5, 15.85, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.25, 16.2, 1.3)
    end
    BUILD.castle[116] = function()
        Include(115)
        -- deco swords left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 20, 16.45, 0.75, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 20.5, 16.65, 0.75, 135)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20, 16.45, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 16.65, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoSwords01", true)
    end
    BUILD.castle[117] = function()
        Include(116)
        -- deco swords right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 27.5, 16.65, 0.75, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 28.1, 16.45, 0.75, 180)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.5, 16.65, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 28.1, 16.45, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoSwords02", true)
    end
    BUILD.castle[118] = function()
        Include(117)
        -- tower deco flags
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Flag"), 20.1, 17.1, 1.75, -30)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Flag"), 27.9, 17.1, 1.75, 30)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.1, 17.1, 2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.9, 17.1, 2)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoFlags", true)
    end
    BUILD.castle[119] = function()
        Include(118)
        -- deco braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Brazier Big"), 20.375, 15.25, 0.3, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Brazier Big"), 27.625, 15.25, 0.3, -90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.4, 15.6, 1.4)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.6, 15.6, 1.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoBraziers", true)
    end
    BUILD.castle[120] = function()
        Include(119)
        -- replace statues with final defenses
        SetPropLayer("Castle_Set06_TowersBig_Statue01", false)
        SetPropLayer("Castle_Set06_TowersBig_Statue02", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientDeco Statue"), 21.2, 14.15, 1.75)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientDeco Statue"), 26.9, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 14.3, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 14.3, 1.7)
    end

    local function SetTownDefaultLayers()
        -- camp A
        SetPropLayer("Town_Camp01_Fire", false)
        SetPropLayer("Town_Camp01_Tent01", false)
        SetPropLayer("Town_Camp01_Tent02", false)
        SetPropLayer("Town_Camp01_Tent04", false)

        -- camp B
        SetPropLayer("Town_Camp02_Fire", false)
        SetPropLayer("Town_Camp02_Tent01", false)
        SetPropLayer("Town_Camp02_Tent02", false)
        SetPropLayer("Town_Camp02_Tent03", false)
        SetPropLayer("Town_Camp02_Tent04", false)

        -- camp C
        SetPropLayer("Town_Camp03_Fire", false)
        SetPropLayer("Town_Camp03_Tent01", false)
        SetPropLayer("Town_Camp03_Tent02", false)
        SetPropLayer("Town_Camp03_Tent04", false)

        -- camp D
        SetPropLayer("Town_Camp04_Fire", false)
        SetPropLayer("Town_Camp04_Tent01", false)
        SetPropLayer("Town_Camp04_Tent02", false)
        SetPropLayer("Town_Camp04_Tent03", false)
        SetPropLayer("Town_Camp04_Tent04", false)

        -- block A
        SetPropLayer("Town_Block01_House01", false)
        SetPropLayer("Town_Block01_House03", false)
        SetPropLayer("Town_Block01_House04", false)
        SetPropLayer("Town_Block01_House05", false)
        SetPropLayer("Town_Block01_House06", false)
        SetPropLayer("Town_Block01_House09", false)
        SetPropLayer("Town_Block01_House10", false)
        SetPropLayer("Town_Block01_House12", false)
        SetPropLayer("Town_Block01_House13", false)

        -- block B
        SetPropLayer("Town_Block02_House01", false)
        SetPropLayer("Town_Block02_House02", false)
        SetPropLayer("Town_Block02_House03", false)
        SetPropLayer("Town_Block02_House04", false)
        SetPropLayer("Town_Block02_House05", false)
        SetPropLayer("Town_Block02_House06", false)
        SetPropLayer("Town_Block02_House07", false)
        SetPropLayer("Town_Block02_House08", false)
        SetPropLayer("Town_Block02_House09", false)
        SetPropLayer("Town_Block02_House10", false)
        SetPropLayer("Town_Block02_House11", false)
        SetPropLayer("Town_Block02_House12", false)

        -- block C
        SetPropLayer("Town_Block03_House01", false)
        SetPropLayer("Town_Block03_House02", false)
        SetPropLayer("Town_Block03_House03", false)
        SetPropLayer("Town_Block03_House04", false)
        SetPropLayer("Town_Block03_House05", false)
        SetPropLayer("Town_Block03_House06", false)
        SetPropLayer("Town_Block03_House07", false)
        SetPropLayer("Town_Block03_House08", false)

        -- block D
        SetPropLayer("Town_Block04_House01", false)
        SetPropLayer("Town_Block04_House02", false)
        SetPropLayer("Town_Block04_House03", false)
        SetPropLayer("Town_Block04_House04", false)
        SetPropLayer("Town_Block04_House05", false)
        SetPropLayer("Town_Block04_House06", false)
        SetPropLayer("Town_Block04_House07", false)

        -- town upgrades
        SetPropLayer("Town_Upgrade_SwordStatues", false)
        SetPropLayer("Town_Upgrade_Statue01_Fence", false)
        SetPropLayer("Town_Upgrade_Statue02", false)
        SetPropLayer("Town_Upgrade_Statue03", false)
        SetPropLayer("Town_Upgrade_TownHall", false)
    end

    -- TOWN DISTRICTS
    function BUILD.town.Districts()
        local slot_count = BUILD.town.slot_count

        for n = 1, slot_count do
            local unit_data = nil

            -- get unit data from player town list
            if player.town[n] then
                local name = DCEI.GetUnitType(player.town[n])
                unit_data = data.town[name]
            end

            -- BUILD.town.SetDistrict(n, unit_data)
        end
    end

    -- function BUILD.town.SetDistrict(index, unit_data)
    --     if unit_data then
    --         DCEI.LogMessage(">> SetDistrict " .. index .. " = " .. unit_data.name)
    --     else
    --         DCEI.LogMessage(">> SetDistrict " .. index .. " = " .. "EMPTY")
    --     end
    -- end

    BUILD.town_slot[1] = function()
        local slot = 1
        SetPropLayer(district_layers[slot], true)
    end
    BUILD.town_slot[2] = function()
        local slot = 2
        SetPropLayer(district_layers[slot], true)
    end
    BUILD.town_slot[3] = function()
        local slot = 3
        SetPropLayer(district_layers[slot], true)
        SetPropLayer("_Default_Town_Props_Camp03", false)
    end
    BUILD.town_slot[4] = function()
        local slot = 4
        SetPropLayer(district_layers[slot], true)
        SetPropLayer("_Default_Town_Props_Camp04", false)
    end

    -- TOWN SEQUENCE
    BUILD.town[1] = function()
        -- town defaults
        SetTownDefaultLayers()

        SetPropLayer("_Default_Town_Props_Rocks", true)
        SetPropLayer("_Default_Town_Props_TownHall", true)

        SetPropLayer("_Default_Town_Props_Camp03", true)
        SetPropLayer("_Default_Town_Props_Camp04", true)
        SetPropLayer("_Default_Town_Props_Camp04B", true)
        SetPropLayer("_Default_Town_Props_Block01A", true)
        SetPropLayer("_Default_Town_Props_Block01B", true)
        SetPropLayer("_Default_Town_Props_Block02A", true)
        SetPropLayer("_Default_Town_Props_Block02B", true)
        SetPropLayer("_Default_Town_Props_Block03A", true)
        SetPropLayer("_Default_Town_Props_Block04A", true)
        SetPropLayer("_Default_Town_Props_Block04B", true)
        SetPropLayer("_Default_Town_Props_Block04C", true)
        SetPropLayer("_Default_Town_Props_Block04D", true)
    end
    BUILD.town[2] = function()
        Include(1)
        local BuildSlot = BUILD.town_slot[1]
        local is_slot_locked = (BUILD.town.slot_count < 1)

        -- camp A1
        if is_slot_locked then
            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 22, 6.3, 0)
        else
            district_layers[1] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.75, 7, 0, -17)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 22.45, 6.95, 0, -53)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp01_Tent02", true)
    end
    BUILD.town[3] = function()
        Include(2)

        -- camp A2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.33, 6.3, 0, 103)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.65, 6.2, 0)

        Pause(0.125)
        SetPropLayer("Town_Camp01_Tent01", true)
        SetPropLayer("Town_Camp01_Tent04", true)
    end
    BUILD.town[4] = function()
        Include(3)
        local BuildSlot = BUILD.town_slot[2]
        local is_slot_locked = (BUILD.town.slot_count < 2)

        -- camp B1
        if is_slot_locked then
            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 26, 6.3, 0)
        else
            district_layers[2] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 25.7, 6.97, 0, -115)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.5, 6.94, 0, 127)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp02_Tent02", true)
        SetPropLayer("Town_Camp02_Tent03", true)
    end
    BUILD.town[5] = function()
        Include(4)

        -- camp B2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 25.26, 6.45, 0, -71)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.75, 6.33, 0, -1)

        Pause(0.25)
        SetPropLayer("Town_Camp02_Tent01", true)
        SetPropLayer("Town_Camp02_Tent04", true)
    end
    BUILD.town[6] = function()
        Include(5)
        local BuildSlot = BUILD.town_slot[3]
        local is_slot_locked = (BUILD.town.slot_count < 3)

        -- camp C1
        if is_slot_locked then
            SetPropLayerNow("_Default_Town_Props_Camp03", false)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 21.48, 3.25, 0, -76)
            CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 21.86, 3.03, 0, 23)

            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 22, 3.5, 0)
        else
            SetPropLayer("_Default_Town_Props_Camp03", false)
            district_layers[3] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.7, 4.27, 0, -110)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.25, 3.52, 0, -84)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp03_Tent01", true)
        SetPropLayer("Town_Camp03_Tent04", true)
    end
    BUILD.town[7] = function()
        Include(6)

        -- camp C2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.59, 2.75, 0, -62)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 22.25, 2.7, 0, -105)

        Pause(0.25)
        SetPropLayer("Town_Camp03_Tent02", true)
    end
    BUILD.town[8] = function()
        Include(7)
        local BuildSlot = BUILD.town_slot[4]
        local is_slot_locked = (BUILD.town.slot_count < 4)

        -- camp D1
        if is_slot_locked then
            SetPropLayerNow("_Default_Town_Props_Camp04", false)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.53, 2.81, 0, -8)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.3, 3.42, 0, 38)

            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 26, 3.5, 0)
        else
            SetPropLayer("_Default_Town_Props_Camp04", false)
            district_layers[4] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.29, 4.34, 0, 19)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 26.77, 3.83, 0, 160)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp04_Tent01", true)
        SetPropLayer("Town_Camp04_Tent04", true)
    end
    BUILD.town[9] = function()
        Include(8)

        -- camp D2
        SetPropLayer("_Default_Town_Props_Camp04B", false)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 25.69, 2.64, 0, -71)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 26.45, 2.76, 0, 60)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 26.66, 3.22, 0)

        Pause(0.125)
        SetPropLayer("Town_Camp04_Tent02", true)
        SetPropLayer("Town_Camp04_Tent03", true)
    end
    BUILD.town[10] = function()
        Include(9)

        -- town level 10 (sword statues)
        CreateBuilderNoFX(DCEI.Unit("Builder Town SwordStatue"), 23.25, 3.85, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Town SwordStatue"), 24.75, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 23.25, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 24.75, 3.85, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_SwordStatues", true)
    end
    BUILD.town[11] = function()
        Include(10)

        -- block A2
        SetPropLayerNow("Town_Camp01_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.75, 7, 0, -17)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 22.45, 6.95, 0, -53)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 21.66, 7.1, 0, 165)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 22.48, 7.06, 0, 98)
        CreateBuilderNoFX(DCEI.Unit("Builder Town HouseC 30%"), 22.61, 6.96, 0.3, 15)

        Pause(0.25)
        SetPropLayer("Town_Block01_House01", true)
    end
    BUILD.town[12] = function()
        Include(11)

        -- block B1
        SetPropLayerNow("Town_Camp02_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 25.7, 6.97, 0, -115)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 25.55, 7.2, 0, 175)

        Pause(0.25)
        SetPropLayer("Town_Block02_House01", true)
    end
    BUILD.town[13] = function()
        Include(12)

        -- block B2
        SetPropLayerNow("Town_Camp02_Tent03", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.5, 6.94, 0, 127)

        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 26.54, 7.04, 0, -144)

        Pause(0.25)
        SetPropLayer("Town_Block02_House02", true)
    end
    BUILD.town[14] = function()
        Include(13)

        -- block C1
        SetPropLayerNow("Town_Camp03_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.7, 4.27, 0, -110)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 21.27, 4.42, 0.1, -26)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50% Long"), 21.82, 4.4, 0, 91)

        Pause(0.25)
        SetPropLayer("Town_Block03_House01", true)
    end
    BUILD.town[15] = function()
        Include(14)

        -- block D1
        SetPropLayerNow("Town_Camp04_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.29, 4.34, 0, 19)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 26, 4.45, 0, -177)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.4, 4.3, 0, 107)

        Pause(0.25)
        SetPropLayer("Town_Block04_House01", true)
    end
    BUILD.town[16] = function()
        Include(15)

        -- block A3
        SetPropLayerNow("Town_Camp01_Tent04", false)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.65, 6.2, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.83, 6.25, 0, -174)

        Pause(0.25)
        SetPropLayer("Town_Block01_House03", true)
    end
    BUILD.town[17] = function()
        Include(16)

        -- block B3
        SetPropLayerNow("Town_Camp02_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.75, 6.33, 0, -1)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 26.81, 6.1, 0, 3)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 27.33, 6.33, 0, 20)

        Pause(0.25)
        SetPropLayer("Town_Block02_House03", true)
    end
    BUILD.town[18] = function()
        Include(17)

        -- block C2
        SetPropLayerNow("Town_Camp03_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.25, 3.52, 0, -84)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.71, 4.49, 0, -116)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.89, 3.82, 0, 100)

        Pause(0.25)
        SetPropLayer("Town_Block03_House02", true)
    end
    BUILD.town[19] = function()
        Include(18)

        -- block D2
        SetPropLayerNow("Town_Camp04_Tent02", false)
        SetPropLayerNow("_Default_Town_Props_Block04A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 25.69, 2.64, 0, -71)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 25.43, 2.15, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 25.47, 2.48, 0, 12)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 25.89, 2.25, -0.1, 150)

        Pause(0.25)
        SetPropLayer("Town_Block04_House02", true)
    end
    BUILD.town[20] = function()
        Include(19)

        -- town level 20 (park statue fence)
        SetPropLayerNow("_Default_Town_Props_Rocks", false)

        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 25.34, 0.84, -0.11, 42)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 25.08, 1.97, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.8, 1.93, 0)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        Pause(0.125)
        SetPropLayer("Town_Upgrade_Statue01_Fence", true)
    end
    BUILD.town[21] = function()
        Include(20)

        -- block A5
        SetPropLayerNow("Town_Camp01_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.33, 6.3, 0, 103)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 21.22, 6.27, 0, 2)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 21.09, 6.91, 0, -80)

        Pause(0.25)
        SetPropLayer("Town_Block01_House04", true)
    end
    BUILD.town[22] = function()
        Include(21)

        -- block B4
        SetPropLayerNow("Town_Camp02_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 25.26, 6.45, 0, -71)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 25.16, 6.14, 0, -91)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 24.81, 6.14, -0.05, -91)

        Pause(0.25)
        SetPropLayer("Town_Block02_House04", true)
    end
    BUILD.town[23] = function()
        Include(22)

        -- block B5
        SetPropLayerNow("_Default_Town_Props_Block02B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 26.1, 8, 0, 101)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 26.12, 7.3, 0, -95)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 25.22, 6.76, -0.13, -68)

        Pause(0.25)
        SetPropLayer("Town_Block02_House05", true)
    end
    BUILD.town[24] = function()
        Include(23)

        -- block C3
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 20.43, 3.51, -0.2, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 20.95, 3, 0, 110)

        Pause(0.25)
        SetPropLayer("Town_Block03_House03", true)
    end
    BUILD.town[25] = function()
        Include(24)

        -- block D3
        SetPropLayerNow("Town_Camp04_Tent03", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 26.45, 2.76, 0, 60)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 26.66, 3.22, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.35, 2.49, -0.1, 75)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 26.9, 2.86, 0, 133)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.19, 2.57, 0, 137)

        Pause(0.25)
        SetPropLayer("Town_Block04_House03", true)
    end
    BUILD.town[26] = function()
        Include(25)

        -- block A6
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 23.3, 6.17, 0, -80)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 50%"), 23.2, 6.93, 0.1, -2)

        Pause(0.25)
        SetPropLayer("Town_Block01_House06", true)
    end
    BUILD.town[27] = function()
        Include(26)

        -- block B6
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 24.7, 6.7, -0.15, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 24.94, 7.4, 0, -47)

        Pause(0.25)
        SetPropLayer("Town_Block02_House06", true)
    end
    BUILD.town[28] = function()
        Include(27)

        -- block C4
        SetPropLayerNow("Town_Camp03_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.59, 2.75, 0, -62)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 22.25, 2.7, 0, -105)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 21.38, 2.53, 0, -174)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 21.85, 2.55, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 22.42, 2.4, 0, -26)

        Pause(0.25)
        SetPropLayer("Town_Block03_House04", true)
    end
    BUILD.town[29] = function()
        Include(28)

        -- block D4
        SetPropLayerNow("Town_Camp04_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 26.77, 3.83, 0, 160)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 26.97, 4.14, 0, -141)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 27.26, 3.5, -0.1, 179)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 27.5, 2.98, -0.1, 5)

        Pause(0.25)
        SetPropLayer("Town_Block04_House04", true)
    end
    BUILD.town[30] = function()
        Include(29)

        -- town level 30 (park statue construction)
        SetPropLayerNow("_Default_Town_Props_Block04C", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.05, 1.28, 0, 97)

        CreateBuilder(DCEI.Unit("Builder Town KnightStatue Broken"), 22.77, 1.51, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town KnightStatue Broken"), 25.37, 1.51, 0, 18)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_Statue02", true)
    end
    BUILD.town[31] = function()
        Include(30)

        -- block A8
        SetPropLayerNow("_Default_Town_Props_Block01A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 21.3, 8, 0, 97)

        CreateBuilder(DCEI.Unit("Builder Town HouseF 50% Long"), 20.64, 7.35, 0, 10)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 21.38, 7.55, 0, 83)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 55%"), 22.02, 7.55, -0.1, 95)

        Pause(0.25)
        SetPropLayer("Town_Block01_House05", true)
    end
    BUILD.town[32] = function()
        Include(31)

        -- block B7
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 27.1, 7, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.81, 6.61, -0.05, -92)

        Pause(0.25)
        SetPropLayer("Town_Block02_House07", true)
    end
    BUILD.town[33] = function()
        Include(32)

        -- block B8
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 25.48, 7.87, 0, 100)

        Pause(0.25)
        SetPropLayer("Town_Block02_House08", true)
    end
    BUILD.town[34] = function()
        Include(33)

        -- block C5
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.85, 2.4, -0.05, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 20.73, 1.57, -0.1, 170)

        Pause(0.25)
        SetPropLayer("Town_Block03_House05", true)
    end
    BUILD.town[35] = function()
        Include(34)

        -- block D5
        SetPropLayerNow("_Default_Town_Props_Block04B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 26.07, 1.69, 0, 101)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 26.21, 1.82, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.61, 1.99, -0.15, -100)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 27.03, 2.12, 0, 90)

        Pause(0.25)
        SetPropLayer("Town_Block04_House05", true)
    end
    BUILD.town[36] = function()
        Include(35)

        -- block A9
        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.5, 7.7, 0, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 23.1, 7.77, 0, -91)

        Pause(0.25)
        SetPropLayer("Town_Block01_House09", true)
    end
    BUILD.town[37] = function()
        Include(36)

        -- block B9
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.02, 7.81, -0.1, 3)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.5, 7.83, -0.2, -85)

        Pause(0.25)
        SetPropLayer("Town_Block02_House09", true)
    end
    BUILD.town[38] = function()
        Include(37)

        -- block C6
        SetPropLayerNow("_Default_Town_Props_Block03A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 21.23, 1.17, -0.19, 19)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 21.68, 0.47, -0.33, 82)

        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 21.23, 1.68, -0.05, 79)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 21.78, 1.26, -0.15, -21)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 21.8, 2.01, 0, 133)

        Pause(0.25)
        SetPropLayer("Town_Block03_House06", true)
    end
    BUILD.town[39] = function()
        Include(38)

        -- block D6
        SetPropLayerNow("_Default_Town_Props_Block04D", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 26.08, 0.36, -0.16, 63)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.75, 1.06, -0.05, 19)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.41, 1.6, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 26.84, 1.58, 0, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 27.59, 2.26, 0, -160)

        Pause(0.25)
        SetPropLayer("Town_Block04_House06", true)
    end
    BUILD.town[40] = function()
        Include(39)

        -- town level 40 (park statue construction)
        SetPropLayerNow("Town_Upgrade_Statue01_Fence", false)
        SetPropLayerNow("Town_Upgrade_Statue02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town KnightStatue Broken"), 22.77, 1.51, 0, -19)
        CreateBuilderDeath(DCEI.Unit("Builder Town KnightStatue Broken"), 25.37, 1.51, 0, 18)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        CreateBuilder(DCEI.Unit("Builder Town KnightStatue"), 22.74, 1.6, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town KnightStatue"), 25.4, 1.6, 0, 18)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_Statue03", true)
    end
    BUILD.town[41] = function()
        Include(40)

        -- block A11
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 20.33, 8.6, -0.1, -174)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 20.33, 8.04, 0, -100)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 20.16, 7.2, -0.1, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.57, 6.68, 0, -176)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.62, 6.08, 0, 5)

        Pause(0.25)
        SetPropLayer("Town_Block01_House10", true)
    end
    BUILD.town[42] = function()
        Include(41)

        -- block B10
        SetPropLayerNow("_Default_Town_Props_Block02A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 27.05, 8.36, 0, -158)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 27.06, 7.7, 0, -4)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 27.77, 7.78, 0.05, 4)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 27.64, 7.15, 0, 175)

        Pause(0.25)
        SetPropLayer("Town_Block02_House10", true)
    end
    BUILD.town[43] = function()
        Include(42)

        -- block B11
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.11, 8.33, 0, 2)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.61, 8.33, -0.1, 11)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.06, 8.18, 0, 179)

        Pause(0.25)
        SetPropLayer("Town_Block02_House11", true)
    end
    BUILD.town[44] = function()
        Include(43)

        -- block C7
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 21.3, 1.04, 0, 180)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 21.85, 0.5, 0, 163)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 21.42, 0.3, -0.1, -98)

        Pause(0.25)
        SetPropLayer("Town_Block03_House07", true)
    end
    BUILD.town[45] = function()
        Include(44)

        -- block D7
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 26.4, 1.2, -0.22, 179)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.51, 0.68, 0, -89)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 26.98, 1.1, 0, 80)

        Pause(0.25)
        SetPropLayer("Town_Block04_House07", true)
    end
    BUILD.town[46] = function()
        Include(45)

        -- block A12
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.87, 8.2, 0, -88)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 50%"), 21.48, 8.2, 0, 87)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.03, 8.26, 0, -10)

        Pause(0.25)
        SetPropLayer("Town_Block01_House12", true)
    end
    BUILD.town[47] = function()
        Include(46)

        -- block B12
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 27.24, 8.75, 0, 120)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 27.77, 8.36, 0, 91)

        Pause(0.25)
        SetPropLayer("Town_Block02_House12", true)
    end
    BUILD.town[48] = function()
        Include(47)

        -- block C8 & D8
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 22.15, 0.88, 0, -42)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 22.49, 0.45, 0, -175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 23.14, 0.53, 0, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 24.68, 0.59, 0, 92)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 25.22, 0.49, -0.15, -177)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 25.82, 0.72, 0, -25)

        Pause(0.25)
        SetPropLayer("Town_Block03_House08", true)
    end
    BUILD.town[49] = function()
        Include(48)

        -- block A13 & B13
        SetPropLayerNow("_Default_Town_Props_Block01B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 22.5, 8.56, 0, -14)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 22.52, 8.32, 0, -90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 23.1, 8.5, 0, -91)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 24.88, 8.2, 0, -170)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 25.53, 8.42, 0, 90)

        Pause(0.25)
        SetPropLayer("Town_Block01_House13", true)
    end
    BUILD.town[50] = function()
        Include(49)

        -- town level 50 (town hall upgrade)
        SetPropLayerNow("Town_Upgrade_SwordStatues", false)
        SetPropLayerNow("_Default_Town_Props_TownHall", false)
        CreateBuilder(DCEI.Unit("Builder Town HallAltar"), 24, 4.31, -0.17, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 23.25, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 24.75, 3.85, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_TownHall", true)
    end
end

function DefineBuildOrderLevelCap()
    BUILD.castle = {}
    BUILD.town_slot = {}
    BUILD.town = {}
    BUILD.town.slot_count = 0

    local function Include(n)
        local was_animated = BUILD.animate
        local build_type = BUILD.type
        BUILD.animate = false

        local build = build_type[n]
        if build then
            build()
        end

        BUILD.animate = was_animated
    end

    local function Pause(duration)
        if BUILD.animate then
            DCEI.Wait(duration)
        end
    end

    -- CASTLE SEQUENCE
    BUILD.castle[1] = function()
        -- castle defaults
        SetPropLayer("_Default_Castle_Set00_Fence", true)
        SetPropLayer("_Default_Castle_Set00_Tower01", true)
        SetPropLayer("_Default_Castle_Set00_Hero04", true)
        SetPropLayer("_Default_Castle_Set00_Tower02", true)
        SetPropLayer("_Default_Castle_Set00_Tower03", true)
        SetPropLayer("_Default_Castle_Set00_Tower04", true)

        SetPropLayer("_Default_Castle_Set01_Back", true)
    end
    BUILD.castle[2] = function()
        Include(1)
        -- 1st tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.5, 20.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 20.65)
        -- 4th hero slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 25.8, 17.5, 1.25)

        SetPropLayerNow("_Default_Castle_Set00_Hero04", false)

        SetPropLayer("_Default_Castle_Set00_Tower01", false)
    end
    BUILD.castle[3] = function()
        Include(2)
        -- wooden archer platform
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Side"), 21.75, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 23.25, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24.75, 14.05, 0.28)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Side"), 26.25, 14.05, 0.28, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop Side"), 21.85, 14.81, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 23.25, 14.81, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 24.75, 14.81, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop Side"), 26, 14.78, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Platform", true)
    end
    BUILD.castle[4] = function()
        Include(3)
    end
    BUILD.castle[5] = function()
        Include(4)
        -- front row of wood
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 21, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 17.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 27, 17.5, 0.2, 180)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop SideB"), 21.2, 17.6, 0.1, 315)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 22.5, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 24, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop"), 25.5, 18.1, 0.1, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTop SideB"), 26.8, 17.6, 0.1, -135)

        SetPropLayerNow("_Default_Castle_Set00_Fence", false)
        Pause(0.25)
        SetPropLayer("Castle_Set01_Fence", true)
        SetPropLayer("Castle_Set01_Floor_Top", true)
    end
    BUILD.castle[6] = function()
        Include(5)
        -- 5th hero slot
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 22.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 24, 16, 0.2, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 14.5, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 24, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero05", true)
    end
    BUILD.castle[7] = function()
        Include(6)
        -- wooden barricade 1
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 20.7, 19.5, 0, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade01", true)
    end
    BUILD.castle[8] = function()
        Include(7)
        -- wooden barricade 2
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 27.3, 19.5, 0)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade02", true)
    end
    BUILD.castle[9] = function()
        Include(8)
        -- 6th hero slot
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor SideB"), 25.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 24, 16, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero05", false)
        SetPropLayer("Castle_Set01_Hero06", true)
    end
    BUILD.castle[10] = function()
        Include(9)
        -- floor decoration
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 22.85, 16.375, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 25.15, 16.375, 1.13, 90)
        -- 2nd tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.4, 20.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.4, 20.65)

        SetPropLayer("_Default_Castle_Set00_Tower02", false)

        Pause(0.25)
        SetPropLayer("Castle_Set01_FloorDeco01", true)
    end
    BUILD.castle[11] = function()
        Include(10)
        -- lower towers
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.3, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.3, 1.25)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower"), 27.35, 17.3, -0.25, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower"), 20.8, 17.3, -0.25, 135)

        Pause(0.25)
        SetPropLayerNow("Castle_Set01_Towers", true)
    end
    BUILD.castle[12] = function()
        Include(11)
        -- hero slot 7
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 21, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 21, 14.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 19.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 19.5, 14.5, 0.2)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 22.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero06", false)
        SetPropLayer("Castle_Set01_Hero07", true)
    end
    BUILD.castle[13] = function()
        Include(12)
        -- barricade 3
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 20.4, 21.4, 0)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade03", true)
    end
    BUILD.castle[14] = function()
        Include(13)
        -- barricade 4
        CreateBuilder(DCEI.Unit("Builder Set01 Barricade"), 27.6, 21.4, 0, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Barricade04", true)
    end
    BUILD.castle[15] = function()
        Include(14)
        -- 8th hero slot
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodFloor"), 25.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 27, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 27, 14.5, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 28.5, 16, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor"), 28.5, 14.5, 0.2)

        Pause(0.25)
        SetPropLayer("Castle_Set01_Hero07", false)
        SetPropLayer("Castle_Set01_Floor_Main", true)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
    end
    BUILD.castle[16] = function()
        Include(15)
        -- change floor decoration
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 21.6, 16.2, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 24, 16.375, 1.13, 90)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodFloor Deco"), 26.4, 16.2, 1.13, 90)

        SetPropLayerNow("Castle_Set01_FloorDeco01", false)
        Pause(0.25)
        SetPropLayer("Castle_Set01_FloorDeco02", true)
    end
    BUILD.castle[17] = function()
        Include(16)
    end
    BUILD.castle[18] = function()
        Include(17)
        -- upper towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.9, -0.5, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 14.05, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.9, 13.9, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.3, 14.05, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set01_TowersBig", true)
    end
    BUILD.castle[19] = function()
        Include(18)
    end
    BUILD.castle[20] = function()
        Include(19)
        -- castle decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 16.45, 1.2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 16.45, 1.2)

        -- 3rd tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.45, 22.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.45, 22.65)

        SetPropLayer("_Default_Castle_Set00_Tower03", false)

        SetPropLayer("Castle_Set01_Deco01", true)
        SetPropLayer("Castle_Set01_Deco02", true)
    end
    -- CASTLE SET 02: Reinforced Wood
    BUILD.castle[21] = function()
        Include(20)
        -- reinforce archer platform
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21.75, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 23.25, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24.75, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 26.25, 14.13, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 21.75, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 23.25, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 24.75, 14.875, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 26.25, 14.875, 0.25, 180)

        Pause(0.125)
        SetPropLayerNow("_Default_Castle_Set01_Back", false)
        SetPropLayerNow("Castle_Set01_Platform", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Platform", true)
        SetPropLayer("Castle_Set02_Back", true)
    end
    BUILD.castle[22] = function()
        Include(21)
        -- reinforce floor
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 17.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 16, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 21, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 24, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 27, 14.5, 0.2, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 14.5, 0.2, 180)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_Floor_Top", false)
        SetPropLayerNow("Castle_Set01_Floor_Main", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Floor", true)
    end
    BUILD.castle[23] = function()
        Include(22)
        -- reinforce fence
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop Side"), 21, 18.2, 0.15, 90)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 22.5, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 24, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop"), 25.5, 18.2, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTop Side"), 27, 18.2, 0.15, 180)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_Fence", true)
    end
    BUILD.castle[24] = function()
        Include(23)
        -- reinforce big towers
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower Big"), 20.8, 14, 0.7, -90)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower Big"), 27.25, 14, 0.7, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set02_TowersBig", true)
    end
    BUILD.castle[25] = function()
        Include(24)
        -- reinforce small towers
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower"), 20.7, 17.4, 0.85, 135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodTower"), 27.25, 17.4, 0.85, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Towers", true)
    end
    BUILD.castle[26] = function()
        Include(25)
        -- reinforce floor decoration
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 15.9, 1.3)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set01_FloorDeco02", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_FloorDeco", true)
    end
    BUILD.castle[27] = function()
        Include(26)
        -- add left furnace
        CreateBuilder(DCEI.Unit("Builder Set02 Furnace"), 20.3, 16.8, 0.2, -45)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Furnace01", true)
    end
    BUILD.castle[28] = function()
        Include(27)
        -- add right furnace
        CreateBuilder(DCEI.Unit("Builder Set02 Furnace"), 27.7, 16.67, 0.2, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set02_Furnace02", true)
    end
    BUILD.castle[29] = function()
        Include(28)
        -- more weapon racks (left)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.7, 16.2, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.5, 16.9, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.4, 17.7, 1.25)

        SetPropLayer("Castle_Set01_Deco01", false)
        SetPropLayer("Castle_Set02_Deco01", true)
    end

    -- CASTLE SET 03: Classic Stone
    BUILD.castle[30] = function()
        -- 4th tower slot
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.55, 22.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.55, 22.65)

        SetPropLayer("_Default_Castle_Set00_Tower04", false)

        -- reset construction site (wood)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set01_Towers", true)
        SetPropLayer("Castle_Set01_TowersBig", true)
        SetPropLayer("Castle_Set02_Back", true)

        -- fence
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set02_Floor", false)
        SetPropLayerNow("Castle_Set03_Construction_FloorDeco", true)
        SetPropLayerNow("Castle_Set03_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 17.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 16, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 19.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 22.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 25.5, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 27, 14.5, 0.2, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 28.5, 14.5, 0.2, 180)

        -- platform
        SetPropLayerNow("Castle_Set02_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 21.75, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 23.25, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 24.75, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodFloor"), 26.25, 14.13, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 21.75, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 23.25, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 24.75, 14.875, 0.25, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTop"), 26.25, 14.875, 0.25, 180)

        -- towers
        SetPropLayerNow("Castle_Set02_Towers", false)
        SetPropLayerNow("Castle_Set02_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower"), 20.7, 17.4, 0.85, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower"), 27.25, 17.4, 0.85, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower Big"), 20.8, 14, 0.7, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set02 WoodTower Big"), 27.25, 14, 0.7, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.9, 14.05, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.6, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.32, 17.22, -0.75, -135)

        Pause(0.125)
        -- previous deco
        SetPropLayerNow("Castle_Set02_FloorDeco", false)
        SetPropLayerNow("Castle_Set02_Deco01", false)
        SetPropLayerNow("Castle_Set02_Deco02", false)
        SetPropLayerNow("Castle_Set02_Furnace01", false)
        SetPropLayerNow("Castle_Set02_Furnace02", false)
        SetPropLayerNow("Castle_Set02_Fence", false)
        SetPropLayer("Castle_Set01_Deco02", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set02_Back", false)
        SetPropLayer("Castle_Set03_Back", true)
        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set03_Construction_Towers", true)
        SetPropLayer("Castle_Set03_Construction_TowersBig", true)
    end
    BUILD.castle[31] = function()
        Include(30)
        -- stone platform
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 21.75, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 23.25, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 24.75, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Alt"), 26.25, 14.875, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 21.75, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 23.25, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 24.75, 14.125, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor Big"), 26.25, 14.125, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set03_Platform", true)
    end
    BUILD.castle[32] = function()
        Include(31)
        -- stone fence
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Side"), 20.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 21.75, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 23.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 24.75, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop"), 26.25, 18.125, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTop Side"), 27.75, 18.125, 0.2, 90)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set03_Fence", true)
    end
    BUILD.castle[33] = function()
        Include(32)
        -- stone floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor01", true)
    end
    BUILD.castle[34] = function()
        Include(33)
        -- stone floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor02", true)
    end
    BUILD.castle[35] = function()
        Include(34)
        -- stone floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor03", true)
    end
    BUILD.castle[36] = function()
        Include(35)
        -- stone floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Floor04", true)
    end
    BUILD.castle[37] = function()
        Include(36)
        -- stone floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 16.7, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 15.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 14.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 16.7, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 15.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 14.2, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Floor", false)
        SetPropLayer("Castle_Set03_Floor05", true)
    end
    BUILD.castle[38] = function()
        Include(37)
        -- swap to reinforced wood floor deco (looks better)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 21.65, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 24, 15.9, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 17.4, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set02 WoodFloor Deco"), 26.35, 15.9, 1.3)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_FloorDeco", false)

        Pause(0.125)
        SetPropLayer("Castle_Set02_FloorDeco", true)
    end
    BUILD.castle[39] = function()
        Include(38)
        -- small towers 1/2
        SetPropLayerNow("Castle_Set03_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.6, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.32, 17.22, -0.75, -135)
        SetPropLayerNow("Castle_Set01_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.35, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.8, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.3, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.3, 1.25)

        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 20.65, 17.89, -0.9, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 27.34, 17.84, -0.9, -45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Base"), 20.65, 18, 0.5, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Base"), 27.35, 18, 0.5, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Towers01", true)
        SetPropLayer("Castle_Set03_TowersC", true)
    end
    BUILD.castle[40] = function()
        Include(39)
        -- small towers 2/2
        SetPropLayerNow("Castle_Set03_TowersC", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 20.65, 17.89, -0.9, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 240%"), 27.34, 17.84, -0.9, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Top"), 20.65, 18, 0.5, 45)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneTower Top"), 27.35, 18, 0.5, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Towers02", true)
    end
    BUILD.castle[41] = function()
        Include(40)
        -- big towers 1/2
        SetPropLayerNow("Castle_Set03_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.9, 14.05, 1.15, -25)

        SetPropLayerNow("Castle_Set01_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.9, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 14.05, -0.5, 90)

        CreateBuilderNoFX(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 20.82, 13.92, -1.85, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 27.2, 13.94, -1.75)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 20.8, 14.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 27.2, 14.1, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.8, 14, 2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.2, 14, 2)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBig01", true)
        SetPropLayer("Castle_Set03_TowersBigC", true)
    end
    BUILD.castle[42] = function()
        Include(41)
        -- big towers 2/2
        SetPropLayerNow("Castle_Set03_TowersBigC", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 20.82, 13.92, -1.85, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 330%"), 27.2, 13.94, -1.75)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 20.8, 14.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 27.2, 14.1, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 14, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.2, 14, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBig02", true)
    end
    BUILD.castle[43] = function()
        Include(42)
        -- small tower flags
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 20.65, 18, 2.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 27.35, 18, 2.1)
        SetPropLayer("Castle_Set03_TowersDeco", true)
    end
    BUILD.castle[44] = function()
        Include(43)
        -- platform stairs
        CreateBuilder(DCEI.Unit("Builder Set03 StoneStairs"), 27.25, 15.75, 1.25, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 StoneStairs"), 20.75, 15.5, 1.25, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Stairs", true)
    end
    BUILD.castle[45] = function()
        Include(44)
        -- left crates
        CreateBuilder(DCEI.Unit("Builder Set03 Crates"), 20, 16.9, 1.2, -140)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Deco01", true)
    end
    BUILD.castle[46] = function()
        Include(45)
        -- right crates
        CreateBuilder(DCEI.Unit("Builder Set03 Crates"), 28.25, 17, 1.2, -95)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Deco02", true)
    end
    BUILD.castle[47] = function()
        Include(46)
        -- sword decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.8, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.8, 1.25)

        SetPropLayer("Castle_Set03_Deco03", true)
    end
    BUILD.castle[48] = function()
        Include(47)
        -- extra big towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 19.75, 14, 0, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 28.25, 14.2, 0, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.4, 14, 2.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.6, 14.2, 2.5)

        Pause(0.25)
        SetPropLayer("Castle_Set03_TowersBigger", true)
    end
    BUILD.castle[49] = function()
        Include(48)
        -- shield decoration
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.35, 17.8, 1.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 17.8, 1.25)
        SetPropLayer("Castle_Set03_Deco04", true)
    end
    BUILD.castle[50] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set03_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set03_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set02_FloorDeco", false)
        SetPropLayerNow("Castle_Set03_Floor01", false)
        SetPropLayerNow("Castle_Set03_Floor02", false)
        SetPropLayerNow("Castle_Set03_Floor03", false)
        SetPropLayerNow("Castle_Set03_Floor04", false)
        SetPropLayerNow("Castle_Set03_Floor05", false)
        SetPropLayerNow("Castle_Set04_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 22.125, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 23.375, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 24.625, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 25.875, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 20.875, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 16.7, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 15.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 19.65, 14.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 27.125, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 16.7, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 15.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor"), 28.37, 14.2, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set03_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 21.75, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 23.25, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 24.75, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTop Alt"), 26.25, 14.875, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 23.25, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 21.75, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 24.75, 14.125, 0.35)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneFloor Big"), 26.25, 14.125, 0.35)

        -- towers
        SetPropLayerNow("Castle_Set03_Towers01", false)
        SetPropLayerNow("Castle_Set03_Towers02", false)
        SetPropLayerNow("Castle_Set03_TowersBig01", false)
        SetPropLayerNow("Castle_Set03_TowersBig02", false)
        SetPropLayerNow("Castle_Set03_TowersBigger", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Base"), 20.65, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Base"), 27.35, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Top"), 20.65, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTower Top"), 27.35, 18, 0.5, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 20.8, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Base"), 27.2, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 20.8, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBig Top"), 27.2, 14.1, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 19.75, 14, 0, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 StoneTowerBigger Top"), 28.25, 14.2, 0, 45)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        Pause(0.125)
        -- previous deco
        SetPropLayerNow("Castle_Set03_TowersDeco", false)
        SetPropLayerNow("Castle_Set03_Stairs", false)
        SetPropLayerNow("Castle_Set03_Deco01", false)
        SetPropLayerNow("Castle_Set03_Deco02", false)
        SetPropLayerNow("Castle_Set03_Deco03", false)
        SetPropLayerNow("Castle_Set03_Deco04", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set03_Back", false)
        SetPropLayer("Castle_Set04_Back", true)

        SetPropLayer("Castle_Set03_Construction_Fence", true)
        SetPropLayer("Castle_Set03_Construction_Platform", true)
        SetPropLayer("Castle_Set04_Construction_Towers", true)
        SetPropLayer("Castle_Set04_Construction_TowersBig", true)
    end
    BUILD.castle[51] = function()
        Include(50)
        -- desert platform
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 21.75, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 23.25, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 24.75, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 26.25, 14.75, 0.3, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.25, 14.75, 0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.23, 13.3, 0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.65, 14.75, 0.5, 270)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.63, 13.3, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 20.75, 14, 0.23, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 27.25, 14, 0.23, 90)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 21.75, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 23.25, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 24.75, 14, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Big"), 26.25, 14, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set04_Platform", true)
    end
    BUILD.castle[52] = function()
        Include(51)
        -- desert fence
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 21, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 22.5, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 24, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop"), 25.5, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 27, 18.25, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Convex"), 20.25, 18.13, -0.35)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Convex"), 27.75, 18.13, -0.35, 90)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set04_Fence", true)
    end
    BUILD.castle[53] = function()
        Include(52)
        -- desert floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor01", true)
    end
    BUILD.castle[54] = function()
        Include(53)
        -- desert floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor02", true)
    end
    BUILD.castle[55] = function()
        Include(54)
        -- desert floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor03", true)
    end
    BUILD.castle[56] = function()
        Include(55)
        -- desert floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor04", true)
    end
    BUILD.castle[57] = function()
        Include(56)
        -- desert floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 19.85, 17, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTop Alt"), 28.15, 17, 0.17)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Alt"), 20.88, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor Alt"), 27.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Floor05", true)
    end
    BUILD.castle[58] = function()
        Include(57)
        -- desert floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.5, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.5, 15.83, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 15.83, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.5, 17.33, 1.15, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.5, 15.83, 1.15, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set04_Construction_Floor", false)

        Pause(0.125)
        SetPropLayer("Castle_Set04_FloorDeco", true)
    end
    BUILD.castle[59] = function()
        Include(58)
        -- desert towers
        SetPropLayerNow("Castle_Set04_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTower"), 20.3, 17, 0.6, -135)
        CreateBuilder(DCEI.Unit("Builder Set04 DesertTower"), 27.7, 17, 0.6, 45)

        Pause(0.25)
        SetPropLayer("Castle_Set04_Towers", true)
    end
    BUILD.castle[60] = function()
        Include(59)
        -- desert towers big
        SetPropLayerNow("Castle_Set04_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertTower Big"), 20.75, 14, -0.25, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertTower Big"), 27.25, 14, -0.25)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 14.25, 1.8)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.2, 14.25, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set04_TowersBig", true)
    end
    BUILD.castle[61] = function()
        Include(60)
        -- upper braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 17.8, 0.95, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 17.8, 0.95, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.75, 18, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.25, 18, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoBrazier01", true)
    end
    BUILD.castle[62] = function()
        Include(61)
        -- left ring
        -- CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Ring"), 20.75, 15.8, 1.38)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.75, 15.8, 1.3)

        Pause(0.25)
        -- SetPropLayer("Castle_Set04_DecoRing01", true)
    end
    BUILD.castle[63] = function()
        Include(62)
        -- right ring
        -- CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Ring"), 27.25, 15.8, 1.38)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.25, 15.8, 1.3)

        Pause(0.25)
        -- SetPropLayer("Castle_Set04_DecoRing02", true)
    end
    BUILD.castle[64] = function()
        Include(63)
        -- totems
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Totem"), 21.5, 17.5, 1.25, 155)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco Totem"), 26.5, 17.5, 1.25, -165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.5, 17.6, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.5, 17.6, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoTotems", true)
    end
    BUILD.castle[65] = function()
        Include(64)
        -- left ring big
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco RingBig"), 20.8, 14, 1.35)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 21, 14, 1.4)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing03", true)
    end
    BUILD.castle[66] = function()
        Include(65)
        --right ring big
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertDeco RingBig"), 27.2, 14, 1.4)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27, 14, 1.45)

        Pause(0.25)
        SetPropLayer("Castle_Set04_DecoRing04", true)
    end
    BUILD.castle[67] = function()
        Include(66)
        --lower braziers
        -- CreateBuilder(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 15.8, 0.8)
        -- CreateBuilder(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 15.8, 0.8)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.75, 15.8, 1.4)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.25, 15.8, 1.4)

        -- Pause(0.25)
        -- SetPropLayer("Castle_Set04_DecoBrazier02", true)
    end
    BUILD.castle[68] = function()
        Include(67)
        -- spears
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.3, 17, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.7, 17, 1.3)

        Pause(0.125)
        SetPropLayer("Castle_Set04_DecoSpears01", true)
    end
    BUILD.castle[69] = function()
        Include(68)
        -- more spears
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 22, 18.3, 1.5)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26, 18.3, 1.5)

        Pause(0.125)
        SetPropLayer("Castle_Set04_DecoSpears02", true)
    end
    BUILD.castle[70] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set04_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set04_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set04_FloorDeco", false)
        SetPropLayerNow("Castle_Set04_Floor01", false)
        SetPropLayerNow("Castle_Set04_Floor02", false)
        SetPropLayerNow("Castle_Set04_Floor03", false)
        SetPropLayerNow("Castle_Set04_Floor04", false)
        SetPropLayerNow("Castle_Set04_Floor05", false)
        SetPropLayerNow("Castle_Set05_Construction_Floor", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 22.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 23.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 24.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 25.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 19.85, 17, 0.17)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 28.15, 17, 0.17)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Alt"), 20.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 20.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 19.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Alt"), 27.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 27.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor"), 28.38, 14.95, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set04_Platform", false)
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop"), 21.75, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 23.25, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Alt"), 24.75, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop"), 26.25, 14.75, 0.3, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.25, 14.75, 0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 20.23, 13.3, 0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.65, 14.75, 0.5, 270)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTop Concave"), 27.63, 13.3, 0.5)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 20.75, 14, 0.23, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor BigAlt"), 27.25, 14, 0.23, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 21.75, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 23.25, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 24.75, 14, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertFloor Big"), 26.25, 14, 0.25)

        -- towers
        SetPropLayerNow("Castle_Set04_Towers", false)
        SetPropLayerNow("Castle_Set04_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower"), 20.3, 17, 0.6, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower"), 27.7, 17, 0.6, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower Big"), 20.75, 14, -0.25, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertTower Big"), 27.25, 14, -0.25)

        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilder(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        -- deco
        SetPropLayerNow("Castle_Set04_DecoTotems", false)
        SetPropLayerNow("Castle_Set04_DecoBrazier02", false)
        SetPropLayerNow("Castle_Set04_DecoBrazier01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 17.8, 0.95, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 17.8, 0.95, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 20.75, 15.8, 0.8)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Brazier"), 27.25, 15.8, 0.8)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Totem"), 21.5, 17.5, 1.25, 155)
        CreateBuilderDeath(DCEI.Unit("Builder Set04 DesertDeco Totem"), 26.5, 17.5, 1.25, -165)

        Pause(0.125)
        -- instant deco
        SetPropLayerNow("Castle_Set04_DecoSpears02", false)
        SetPropLayerNow("Castle_Set04_DecoSpears01", false)
        SetPropLayerNow("Castle_Set04_DecoRing04", false)
        SetPropLayerNow("Castle_Set04_DecoRing03", false)
        SetPropLayerNow("Castle_Set04_DecoRing02", false)
        SetPropLayerNow("Castle_Set04_DecoRing01", false)
        SetPropLayerNow("Castle_Set04_FloorDeco", false)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set04_Back", false)
        SetPropLayer("Castle_Set05_Back", true)

        SetPropLayer("Castle_Set03_Fence", true)
        SetPropLayer("Castle_Set04_Construction_TowersBig", true)
    end
    BUILD.castle[71] = function()
        Include(70)
        -- dark platform
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 21.75, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 23.25, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 24.75, 14.1, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Big"), 26.25, 14.1, 0.25)

        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 22.5, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 24, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Platform"), 25.5, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 14.95, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 13.5, 0.25, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 14.95, 0.25, 90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 13.5, 0.25, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set03_Construction_Platform", false)
        SetPropLayer("Castle_Set05_Platform", true)
    end
    BUILD.castle[72] = function()
        Include(71)
        -- dark fence
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 21.75, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 23.25, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 24.75, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop"), 26.25, 18.3, 0.22)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Concave"), 20.25, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop Concave"), 27.75, 18.3, 0.2, 90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop ConcaveAlt"), 20.25, 16.8, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTop ConcaveAlt"), 27.75, 16.8, 0.2, 90)

        Pause(0.125)
        SetPropLayer("Castle_Set03_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set05_Fence", true)
        SetPropLayer("Castle_Set05_FenceDeco01", true)
    end
    BUILD.castle[73] = function()
        Include(72)
        -- dark floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor01", true)
    end
    BUILD.castle[74] = function()
        Include(73)
        -- dark floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 17.45, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor02", true)
    end
    BUILD.castle[75] = function()
        Include(74)
        -- dark floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor03", true)
    end
    BUILD.castle[76] = function()
        Include(75)
        -- dark floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 16.2, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 14.95, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor04", true)
    end
    BUILD.castle[77] = function()
        Include(76)
        -- dark floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 17.45, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 14.95, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 16.2, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 14.95, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Floor05", true)
    end
    BUILD.castle[78] = function()
        Include(77)
        -- dark floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 21.375, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 21.375, 15.85, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 23.75, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 23.75, 15.85, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 26.125, 17.35, 0.125, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkFloor Deco"), 26.125, 15.85, 0.125, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set05_Construction_Floor", false)

        Pause(0.125)
        SetPropLayer("Castle_Set05_FloorDeco", true)
    end
    BUILD.castle[79] = function()
        Include(78)
        -- deco fence final
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Mast"), 24, 18.29, 0.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal03", true)
    end
    BUILD.castle[80] = function()
        Include(79)
        -- dark towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 17.41, 1.35, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 28.27, 16.85, 1.35, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 16.28, 1.35, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.13, 16.85, 1.35, -45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 17.33, 1.43, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 28.18, 16.85, 1.43, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 16.38, 1.43, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.22, 16.85, 1.43, -45)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Deco"), 27.7, 16.85, 1.31, 165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.7, 16.85, 1.3)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 17.41, 1.35, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.87, 16.85, 1.35, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 16.28, 1.35, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Wall"), 19.73, 16.85, 1.35, -45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 17.33, 1.43, 45)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.78, 16.85, 1.43, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 16.38, 1.43, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Spike"), 19.82, 16.85, 1.43, -45)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Deco"), 20.3, 16.85, 1.31, 165)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.3, 16.85, 1.3)

        SetPropLayerNow("Castle_Set04_Construction_Towers", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 27.5, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower"), 20.65, 17.3, -0.25, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 20.45, 17.07, -0.75, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Scaffold 275%"), 27.47, 17.22, -0.75, -135)

        Pause(0.25)
        SetPropLayer("Castle_Set05_Towers_Left", true)
        SetPropLayer("Castle_Set05_Towers_Right", true)
    end
    BUILD.castle[81] = function()
        Include(80)
        -- dark towers big
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 14.65, 0.95)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 14.65, 0.95, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 13.36, 0.95, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 13.36, 0.95, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 14.55, 1.08)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 14.55, 1.08, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 13.45, 1.08, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 13.45, 1.08, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 20.9, 14.01, 0.88, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 20.9, 14.01, 1.5)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 14.65, 0.95)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 14.65, 0.95, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 13.36, 0.95, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 13.36, 0.95, -90)

        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 14.55, 1.08)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 14.55, 1.08, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 13.45, 1.08, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 13.45, 1.08, -90)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 27.13, 14.01, 0.88, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 200%"), 27.13, 14.01, 1.5)

        SetPropLayerNow("Castle_Set04_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 20.9, 13.8, -0.5, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 WoodTower Big"), 27.3, 13.95, -0.5, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set03 Construction Crane"), 20.8, 13.95, 1.15, -25)

        Pause(0.25)
        SetPropLayer("Castle_Set05_TowersBig", true)
    end
    BUILD.castle[82] = function()
        Include(81)
        -- dark grate 1/4
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 20.9, 15.85, -0.1, 45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.9, 15.85, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate01", true)
    end
    BUILD.castle[83] = function()
        Include(82)
        -- dark grate 2/4
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 27.1, 15.85, -0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.1, 15.85, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate02", true)
    end
    BUILD.castle[84] = function()
        Include(83)
        -- dark fence deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTop Alt"), 21.75, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Alt"), 21.75, 18.3, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkTop Alt"), 26.25, 18.3, 0.2)
        CreateBuilder(DCEI.Unit("Builder Set05 DarkFloor Alt"), 26.25, 18.3, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set05_FenceDeco01", false)
        SetPropLayer("Castle_Set05_FenceDeco02", true)
    end
    BUILD.castle[85] = function()
        Include(84)
        -- dark fence braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 21.75, 18.3, 0, 70)
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 26.25, 18.3, 0, 120)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.75, 18.3, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.25, 18.3, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoBraziers", true)
    end
    BUILD.castle[86] = function()
        Include(85)
        -- dark grate 3/4
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Grate"), 21, 17.5, 0.22, -45)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate03", true)
    end
    BUILD.castle[87] = function()
        Include(86)
        -- dark grate 4/4
        CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Grate"), 27, 17.5, 0.22, 90)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoGrate04", true)
    end
    BUILD.castle[88] = function()
        Include(87)
        -- dark left big brazier
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 20.85, 14.05, 0.65)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.85, 14.05, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal01", true)
    end
    BUILD.castle[89] = function()
        Include(88)
        -- dark right big brazier
        CreateBuilderNoFX(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 27.15, 14.05, 0.65)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.15, 14.05, 1.8)

        Pause(0.25)
        SetPropLayer("Castle_Set05_DecoFinal02", true)
    end
    -- BUILD.castle[90] = function()
    --     Include(89)
    --     -- deco fence final
    --     CreateBuilder(DCEI.Unit("Builder Set05 DarkDeco Mast"), 24, 18.29, 0.3)

    --     Pause(0.25)
    --     SetPropLayer("Castle_Set05_DecoFinal03", true)
    -- end
    BUILD.castle[90] = function()
        -- reset construction site (stone)
        SetPropLayer("_Default_Castle_Set00_Fence", false)
        SetPropLayer("_Default_Castle_Set00_Tower01", false)
        SetPropLayer("_Default_Castle_Set00_Hero04", false)
        SetPropLayer("_Default_Castle_Set00_Tower02", false)
        SetPropLayer("_Default_Castle_Set00_Tower03", false)
        SetPropLayer("_Default_Castle_Set00_Tower04", false)
        SetPropLayer("_Default_Castle_Set00_Deco", false)
        SetPropLayer("_Default_Castle_Set01_Back", false)

        -- base layer
        SetPropLayer("Castle_Set01_Barricade01", true)
        SetPropLayer("Castle_Set01_Barricade02", true)
        SetPropLayer("Castle_Set01_Barricade03", true)
        SetPropLayer("Castle_Set01_Barricade04", true)
        SetPropLayer("Castle_Set05_Back", true)

        -- fence
        SetPropLayerNow("Castle_Set05_Fence", false)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 21.45, 17.98, 0.12, -45)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence Side"), 26.6, 17.98, 0.11, -135)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 22.35, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 23.4, 18.13, 0.15)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 24.6, 18.13, 0.15, 180)
        CreateBuilder(DCEI.Unit("Builder Set02 WoodFence"), 25.65, 18.13, 0.15)

        -- floor
        SetPropLayerNow("Castle_Set05_FloorDeco", false)
        SetPropLayerNow("Castle_Set05_Floor01", false)
        SetPropLayerNow("Castle_Set05_Floor02", false)
        SetPropLayerNow("Castle_Set05_Floor03", false)
        SetPropLayerNow("Castle_Set05_Floor04", false)
        SetPropLayerNow("Castle_Set05_Floor05", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 17.45, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 19.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 16.2, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 20.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 22.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 23.38, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 24.63, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 25.88, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 27.13, 14.95, 0.4)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor"), 28.38, 14.95, 0.4)

        -- platform
        SetPropLayerNow("Castle_Set03_Construction_Platform", true)
        SetPropLayerNow("Castle_Set05_Platform", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 21.75, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 23.25, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 24.75, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkFloor Big"), 26.25, 14.1, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 22.5, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 24, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Platform"), 25.5, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 14.95, 0.25)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 21, 13.5, 0.25, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 14.95, 0.25, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Spike"), 27, 13.5, 0.25, 180)

        -- towers
        SetPropLayerNow("Castle_Set05_Towers_Left", false)
        SetPropLayerNow("Castle_Set05_Towers_Right", false)
        SetPropLayerNow("Castle_Set05_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 17.41, 1.35, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 28.27, 16.85, 1.35, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.7, 16.28, 1.35, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 27.13, 16.85, 1.35, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 17.33, 1.43, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 28.18, 16.85, 1.43, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.7, 16.38, 1.43, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 27.22, 16.85, 1.43, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Deco"), 27.7, 16.85, 1.31, 165)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 17.41, 1.35, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.87, 16.85, 1.35, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 20.3, 16.28, 1.35, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Wall"), 19.73, 16.85, 1.35, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 17.33, 1.43, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.78, 16.85, 1.43, 135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 20.3, 16.38, 1.43, -135)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Spike"), 19.82, 16.85, 1.43, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Deco"), 20.3, 16.85, 1.31, 165)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 14.65, 0.95)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 14.65, 0.95, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 21.54, 13.36, 0.95, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 20.25, 13.36, 0.95, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 14.55, 1.08)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 14.55, 1.08, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 21.45, 13.45, 1.08, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 20.35, 13.45, 1.08, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 20.9, 14.01, 0.88, 120)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 14.65, 0.95)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 14.65, 0.95, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 27.78, 13.36, 0.95, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Wall"), 26.49, 13.36, 0.95, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 14.55, 1.08)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 14.55, 1.08, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 27.68, 13.45, 1.08, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Spike"), 26.59, 13.45, 1.08, -90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTower Big Deco"), 27.13, 14.01, 0.88, 120)

        -- deco
        SetPropLayerNow("Castle_Set05_DecoFinal01", false)
        SetPropLayerNow("Castle_Set05_DecoFinal02", false)
        SetPropLayerNow("Castle_Set05_DecoFinal03", false)
        SetPropLayerNow("Castle_Set05_DecoGrate01", false)
        SetPropLayerNow("Castle_Set05_DecoGrate02", false)
        SetPropLayerNow("Castle_Set05_DecoGrate03", false)
        SetPropLayerNow("Castle_Set05_DecoGrate04", false)
        SetPropLayerNow("Castle_Set05_FenceDeco02", false)
        SetPropLayerNow("Castle_Set05_DecoBraziers", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 20.85, 14.05, 0.65)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco BrazierBig"), 27.15, 14.05, 0.65)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Mast"), 24, 18.29, 0.3)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 20.9, 15.85, -0.1, 45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco GrateBig"), 27.1, 15.85, -0.1)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Grate"), 21, 17.5, 0.22, -45)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Grate"), 27, 17.5, 0.22, 90)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Alt"), 26.25, 18.3, 0.2)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkTop Alt"), 21.75, 18.3, 0.2)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 21.75, 18.3, 0, 70)
        CreateBuilderDeath(DCEI.Unit("Builder Set05 DarkDeco Brazier"), 26.25, 18.3, 0, 120)

        -- new rubble
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 21.1, 17.2, 1, -60)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.8, 15.65, 1.2, -120)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 19.7, 16, 1, 129)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 20.2, 15.2, 1, 172)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.6, 14.15, 0.85, -40)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble"), 20.1, 13.95, 1.2, 51)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 26.9, 17.05, 1, 130)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 28.5, 15.85, 1, -37)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.45, 15.7, 1.2, 150)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Concave"), 27.5, 15.4, 1, 259)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.25, 14.35, 0.85, 121)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientRubble"), 27.5, 13.5, 1.2, -71)

        Pause(0.125)
        SetPropLayerNow("Castle_Set06_Construction", true)

        Pause(0.125)
        -- construction layer
        SetPropLayer("Castle_Set04_Fence", true)
        SetPropLayer("Castle_Set06_Construction_TowersBig", true)

        SetPropLayer("Castle_Set05_Back", false)
        SetPropLayer("Castle_Set06_Back", true)

        SetPropLayer("Castle_Set06_RubbleLeft", true)
        SetPropLayer("Castle_Set06_RubbleTowers", true)
    end
    BUILD.castle[91] = function()
        Include(90)
        -- ancient platform
        SetPropLayerNow("Castle_Set06_Construction_TowersBig", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.6, 14.15, 0.85, -40)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.25, 14.35, 0.85, 121)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 21.75, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 23.25, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 24.75, 14, 0.25, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor Big"), 26.25, 14, 0.25, 180)

        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 20.95, 14.85, 0.5, 0)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 21.05, 13.42, 0.5, -90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 26.95, 14.95, 0.5, 90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop Corner"), 27.05, 13.52, 0.5, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 22.5, 14.9, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 24, 14.9, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTop"), 25.5, 14.9, 0.5)

        Pause(0.125)
        SetPropLayerNow("Castle_Set03_Construction_Platform", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_Platform", true)
    end
    BUILD.castle[92] = function()
        Include(91)
        -- ancient fence
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower"), 20, 17.15, 0.5, 135)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower"), 28, 17.15, 0.5, 135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFence Corner"), 20.26, 18.27, 1.3)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFence Corner"), 27.75, 18.27, 1.3, 90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 21.54, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 22.75, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 24, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 25.25, 18.24, 0.4, 180)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence"), 26.51, 18.24, 0.4, 180)

        Pause(0.125)
        SetPropLayer("Castle_Set04_Fence", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_Fence01", true)
    end
    BUILD.castle[93] = function()
        Include(92)
        -- ancient floor 1/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 17.47, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor01", true)
    end
    BUILD.castle[94] = function()
        Include(93)
        -- ancient floor 2/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 17.47, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor02", true)
    end
    BUILD.castle[95] = function()
        Include(94)
        -- ancient floor 3/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 16.23, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 22.125, 14.98, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 23.375, 14.98, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor03", true)
    end
    BUILD.castle[96] = function()
        Include(95)
        -- ancient floor 4/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 16.23, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 24.625, 14.98, 0.4)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientFloor"), 25.875, 14.98, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor04", true)
    end
    BUILD.castle[97] = function()
        Include(96)
        -- ancient floor 5/5
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 20.875, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 19.625, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 17.47, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 27.125, 13.73, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 16.23, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 14.98, 0.4)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFloor"), 28.375, 13.73, 0.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Floor05", true)
    end
    BUILD.castle[98] = function()
        Include(97)
        -- ancient floor deco
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.625, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 21.625, 15.75, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 24, 15.75, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.375, 17.25, 1.2, 90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set04 DesertFloor Deco"), 26.375, 15.75, 1.2, 90)

        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.65, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 24, 15.9, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 17.4, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.35, 15.9, 1.3)

        Pause(0.125)
        SetPropLayerNow("Castle_Set06_Construction", false)

        Pause(0.125)
        SetPropLayer("Castle_Set06_FloorDeco", true)
    end
    BUILD.castle[99] = function()
        Include(98)
    end
    BUILD.castle[100] = function()
        Include(99)
        -- ancient towers big

        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower Big"), 21.08, 14.13, 0.5, 170)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientTower Big"), 26.92, 14.21, 0.5, -38)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig", true)
    end
    BUILD.castle[101] = function()
        Include(100)
        -- ancient towers reinforce bottom
        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 20.28, 13.5, 0.3, -90)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 27.73, 13.5, 0.3, 180)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Side01", true)
    end
    BUILD.castle[102] = function()
        Include(101)
        -- ancient towers reinforce left
        -- SetPropLayerNow("Castle_Set06_RubbleLeft", false)
        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 20.2, 15.2, 1, 172)

        -- CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 20.25, 14.9, 0.3, 0)

        Pause(0.25)
        --SetPropLayer("Castle_Set06_TowersBig_Side02", true)
    end
    BUILD.castle[103] = function()
        Include(102)
        -- ancient towers reinforce right
        -- SetPropLayerNow("Castle_Set06_RubbleRight", false)
        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 27.5, 15.4, 1, 259)

        -- CreateBuilder(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 27.7, 14.9, 0.3, 90)

        Pause(0.25)
        --SetPropLayer("Castle_Set06_TowersBig_Side03", true)
    end
    BUILD.castle[104] = function()
        Include(103)
        -- unlock hero slot 10
        -- SetPropLayer("Castle_Set06_Rubble02", false)

        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 26.9, 17.05, 1, 130)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 26.9, 17.05, 1.3)
    end
    BUILD.castle[105] = function()
        Include(104)
        -- ancient stone barricades
        SetPropLayerNow("Castle_Set01_Barricade01", false)
        SetPropLayerNow("Castle_Set01_Barricade02", false)
        SetPropLayerNow("Castle_Set01_Barricade03", false)
        SetPropLayerNow("Castle_Set01_Barricade04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 20.7, 19.5, 0, 180)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 27.3, 19.5, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 20.4, 21.4, 0)
        CreateBuilderDeath(DCEI.Unit("Builder Set01 Barricade"), 27.6, 21.4, 0, 180)

        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade"), 20.5, 21.65, -0.7, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 19.88, 21.65, -1.1, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade"), 27.5, 21.63, -0.7, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientBarricade Corner"), 28.05, 21.63, -1.1, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.25, 21.65, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.75, 21.63, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 19.75, 0.1)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.5, 19.75, 0.1)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Barricade", true)
    end
    BUILD.castle[106] = function()
        Include(105)
        -- reinforce fence
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 21.75, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 23.2, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 24.8, 18.23, 0.25)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientFence Deco"), 26.2, 18.23, 0.25)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Fence02", true)
    end
    BUILD.castle[107] = function()
        Include(106)
        -- big tower deco left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower BigDeco"), 21.13, 14.23, -0.55, -90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 21.13, 14.43, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Deco01", true)
    end
    BUILD.castle[108] = function()
        Include(107)
        -- big tower deco right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower BigDeco"), 26.88, 14.18, -0.55, 90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 26.88, 14.45, 1.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Deco02", true)
    end
    BUILD.castle[109] = function()
        Include(108)
        -- unlock hero slot 11
        -- SetPropLayer("Castle_Set06_Rubble03", false)

        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 20.8, 15.65, 1.2, -120)
        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 19.7, 16, 1, 129)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20.8, 16, 1.3)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 19.7, 16.15, 1.3)
    end
    BUILD.castle[110] = function()
        Include(109)
        -- fence braziers
        CreateBuilder(DCEI.Unit("Builder Set06 AncientDeco Brazier"), 20.95, 18.3, 0.5)
        CreateBuilder(DCEI.Unit("Builder Set06 AncientDeco Brazier"), 27.1, 18.3, 0.5)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Fence03", true)
    end
    BUILD.castle[111] = function()
        Include(110)
        -- ancient towers
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower Top"), 20, 17.15, 0.5, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientTower Top"), 28, 17.15, 0.5, -45)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 20, 17.15, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 28, 17.15, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_Towers", true)
    end
    BUILD.castle[112] = function()
        Include(111)
        -- big tower statue left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Statue"), 21.2, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 14.3, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Statue01", true)
    end
    BUILD.castle[113] = function()
        Include(112)
        -- big tower statue right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Statue"), 26.9, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 14.3, 1.7)

        Pause(0.25)
        SetPropLayer("Castle_Set06_TowersBig_Statue02", true)
    end
    BUILD.castle[114] = function()
        Include(113)
        -- unlock hero slot 12
        -- SetPropLayer("Castle_Set06_Rubble04", false)

        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Concave"), 28.5, 15.85, 1, -37)
        -- CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientRubble Convex"), 27.45, 15.7, 1.2, 150)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 28.5, 15.85, 1.3)
        -- CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 27.25, 16.2, 1.3)
    end
    BUILD.castle[115] = function()
        Include(114)
        -- deco swords left
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 20, 16.45, 0.75, 180)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 20.5, 16.65, 0.75, 135)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20, 16.45, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.5, 16.65, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoSwords01", true)
    end
    BUILD.castle[116] = function()
        Include(115)
        -- deco swords right
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 27.5, 16.65, 0.75, -135)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Sword"), 28.1, 16.45, 0.75, 180)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.5, 16.65, 1.3)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 28.1, 16.45, 1.3)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoSwords02", true)
    end
    BUILD.castle[117] = function()
        Include(116)
        -- tower deco flags
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Flag"), 20.1, 17.1, 1.75, -30)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Flag"), 27.9, 17.1, 1.75, 30)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.1, 17.1, 2)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.9, 17.1, 2)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoFlags", true)
    end
    BUILD.castle[118] = function()
        Include(117)
        -- deco braziers
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Brazier Big"), 20.375, 15.25, 0.3, -90)
        CreateBuilderNoFX(DCEI.Unit("Builder Set06 AncientDeco Brazier Big"), 27.625, 15.25, 0.3, -90)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 20.4, 15.6, 1.4)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 27.6, 15.6, 1.4)

        Pause(0.25)
        SetPropLayer("Castle_Set06_DecoBraziers", true)
    end
    BUILD.castle[119] = function()
        Include(118)
        -- replace statues with final defenses
        SetPropLayer("Castle_Set06_TowersBig_Statue01", false)
        SetPropLayer("Castle_Set06_TowersBig_Statue02", false)

        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientDeco Statue"), 21.2, 14.15, 1.75)
        CreateBuilderDeath(DCEI.Unit("Builder Set06 AncientDeco Statue"), 26.9, 14.15, 1.75)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 21.2, 14.3, 1.7)
        CreateBuilder(DCEI.Unit("Builder FX Dust 100%"), 26.9, 14.3, 1.7)
    end
    BUILD.castle[120] = function()
        Include(119)
    end

    local function SetTownDefaultLayers()
        -- camp A
        SetPropLayer("Town_Camp01_Fire", false)
        SetPropLayer("Town_Camp01_Tent01", false)
        SetPropLayer("Town_Camp01_Tent02", false)
        SetPropLayer("Town_Camp01_Tent04", false)

        -- camp B
        SetPropLayer("Town_Camp02_Fire", false)
        SetPropLayer("Town_Camp02_Tent01", false)
        SetPropLayer("Town_Camp02_Tent02", false)
        SetPropLayer("Town_Camp02_Tent03", false)
        SetPropLayer("Town_Camp02_Tent04", false)

        -- camp C
        SetPropLayer("Town_Camp03_Fire", false)
        SetPropLayer("Town_Camp03_Tent01", false)
        SetPropLayer("Town_Camp03_Tent02", false)
        SetPropLayer("Town_Camp03_Tent04", false)

        -- camp D
        SetPropLayer("Town_Camp04_Fire", false)
        SetPropLayer("Town_Camp04_Tent01", false)
        SetPropLayer("Town_Camp04_Tent02", false)
        SetPropLayer("Town_Camp04_Tent03", false)
        SetPropLayer("Town_Camp04_Tent04", false)

        -- block A
        SetPropLayer("Town_Block01_House01", false)
        SetPropLayer("Town_Block01_House03", false)
        SetPropLayer("Town_Block01_House04", false)
        SetPropLayer("Town_Block01_House05", false)
        SetPropLayer("Town_Block01_House06", false)
        SetPropLayer("Town_Block01_House09", false)
        SetPropLayer("Town_Block01_House10", false)
        SetPropLayer("Town_Block01_House12", false)
        SetPropLayer("Town_Block01_House13", false)

        -- block B
        SetPropLayer("Town_Block02_House01", false)
        SetPropLayer("Town_Block02_House02", false)
        SetPropLayer("Town_Block02_House03", false)
        SetPropLayer("Town_Block02_House04", false)
        SetPropLayer("Town_Block02_House05", false)
        SetPropLayer("Town_Block02_House06", false)
        SetPropLayer("Town_Block02_House07", false)
        SetPropLayer("Town_Block02_House08", false)
        SetPropLayer("Town_Block02_House09", false)
        SetPropLayer("Town_Block02_House10", false)
        SetPropLayer("Town_Block02_House11", false)
        SetPropLayer("Town_Block02_House12", false)

        -- block C
        SetPropLayer("Town_Block03_House01", false)
        SetPropLayer("Town_Block03_House02", false)
        SetPropLayer("Town_Block03_House03", false)
        SetPropLayer("Town_Block03_House04", false)
        SetPropLayer("Town_Block03_House05", false)
        SetPropLayer("Town_Block03_House06", false)
        SetPropLayer("Town_Block03_House07", false)
        SetPropLayer("Town_Block03_House08", false)

        -- block D
        SetPropLayer("Town_Block04_House01", false)
        SetPropLayer("Town_Block04_House02", false)
        SetPropLayer("Town_Block04_House03", false)
        SetPropLayer("Town_Block04_House04", false)
        SetPropLayer("Town_Block04_House05", false)
        SetPropLayer("Town_Block04_House06", false)
        SetPropLayer("Town_Block04_House07", false)

        -- town upgrades
        SetPropLayer("Town_Upgrade_SwordStatues", false)
        SetPropLayer("Town_Upgrade_Statue01_Fence", false)
        SetPropLayer("Town_Upgrade_Statue02", false)
        SetPropLayer("Town_Upgrade_Statue03", false)
        SetPropLayer("Town_Upgrade_TownHall", false)
    end

    -- TOWN DISTRICTS
    function BUILD.town.Districts()
        local slot_count = BUILD.town.slot_count

        for n = 1, slot_count do
            local unit_data = nil

            -- get unit data from player town list
            if player.town[n] then
                local name = DCEI.GetUnitType(player.town[n])
                unit_data = data.town[name]
            end

            -- BUILD.town.SetDistrict(n, unit_data)
        end
    end

    -- function BUILD.town.SetDistrict(index, unit_data)
    --     if unit_data then
    --         DCEI.LogMessage(">> SetDistrict " .. index .. " = " .. unit_data.name)
    --     else
    --         DCEI.LogMessage(">> SetDistrict " .. index .. " = " .. "EMPTY")
    --     end
    -- end

    BUILD.town_slot[1] = function()
        local slot = 1
        SetPropLayer(district_layers[slot], true)
    end
    BUILD.town_slot[2] = function()
        local slot = 2
        SetPropLayer(district_layers[slot], true)
    end
    BUILD.town_slot[3] = function()
        local slot = 3
        SetPropLayer(district_layers[slot], true)
        SetPropLayer("_Default_Town_Props_Camp03", false)
    end
    BUILD.town_slot[4] = function()
        local slot = 4
        SetPropLayer(district_layers[slot], true)
        SetPropLayer("_Default_Town_Props_Camp04", false)
    end

    -- TOWN SEQUENCE
    BUILD.town[1] = function()
        -- town defaults
        SetTownDefaultLayers()

        SetPropLayer("_Default_Town_Props_Rocks", true)
        SetPropLayer("_Default_Town_Props_TownHall", true)

        SetPropLayer("_Default_Town_Props_Camp03", true)
        SetPropLayer("_Default_Town_Props_Camp04", true)
        SetPropLayer("_Default_Town_Props_Camp04B", true)
        SetPropLayer("_Default_Town_Props_Block01A", true)
        SetPropLayer("_Default_Town_Props_Block01B", true)
        SetPropLayer("_Default_Town_Props_Block02A", true)
        SetPropLayer("_Default_Town_Props_Block02B", true)
        SetPropLayer("_Default_Town_Props_Block03A", true)
        SetPropLayer("_Default_Town_Props_Block04A", true)
        SetPropLayer("_Default_Town_Props_Block04B", true)
        SetPropLayer("_Default_Town_Props_Block04C", true)
        SetPropLayer("_Default_Town_Props_Block04D", true)
    end
    BUILD.town[2] = function()
        Include(1)
        local BuildSlot = BUILD.town_slot[1]
        local is_slot_locked = (BUILD.town.slot_count < 1)

        -- camp A1
        if is_slot_locked then
            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 22, 6.3, 0)
        else
            district_layers[1] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.75, 7, 0, -17)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 22.45, 6.95, 0, -53)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp01_Tent02", true)
    end
    BUILD.town[3] = function()
        Include(2)

        -- camp A2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.33, 6.3, 0, 103)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.65, 6.2, 0)

        Pause(0.125)
        SetPropLayer("Town_Camp01_Tent01", true)
        SetPropLayer("Town_Camp01_Tent04", true)
    end
    BUILD.town[4] = function()
        Include(3)
        local BuildSlot = BUILD.town_slot[2]
        local is_slot_locked = (BUILD.town.slot_count < 2)

        -- camp B1
        if is_slot_locked then
            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 26, 6.3, 0)
        else
            district_layers[2] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 25.7, 6.97, 0, -115)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.5, 6.94, 0, 127)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp02_Tent02", true)
        SetPropLayer("Town_Camp02_Tent03", true)
    end
    BUILD.town[5] = function()
        Include(4)

        -- camp B2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 25.26, 6.45, 0, -71)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.75, 6.33, 0, -1)

        Pause(0.25)
        SetPropLayer("Town_Camp02_Tent01", true)
        SetPropLayer("Town_Camp02_Tent04", true)
    end
    BUILD.town[6] = function()
        Include(5)
        local BuildSlot = BUILD.town_slot[3]
        local is_slot_locked = (BUILD.town.slot_count < 3)

        -- camp C1
        if is_slot_locked then
            SetPropLayerNow("_Default_Town_Props_Camp03", false)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 21.48, 3.25, 0, -76)
            CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 21.86, 3.03, 0, 23)

            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 22, 3.5, 0)
        else
            SetPropLayer("_Default_Town_Props_Camp03", false)
            district_layers[3] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.7, 4.27, 0, -110)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.25, 3.52, 0, -84)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp03_Tent01", true)
        SetPropLayer("Town_Camp03_Tent04", true)
    end
    BUILD.town[7] = function()
        Include(6)

        -- camp C2
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 21.59, 2.75, 0, -62)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 22.25, 2.7, 0, -105)

        Pause(0.25)
        SetPropLayer("Town_Camp03_Tent02", true)
    end
    BUILD.town[8] = function()
        Include(7)
        local BuildSlot = BUILD.town_slot[4]
        local is_slot_locked = (BUILD.town.slot_count < 4)

        -- camp D1
        if is_slot_locked then
            SetPropLayerNow("_Default_Town_Props_Camp04", false)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.53, 2.81, 0, -8)
            CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.3, 3.42, 0, 38)

            CreateBuilder(DCEI.Unit("Builder Town CampFire"), 26, 3.5, 0)
        else
            SetPropLayer("_Default_Town_Props_Camp04", false)
            district_layers[4] = ""
        end
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 26.29, 4.34, 0, 19)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 26.77, 3.83, 0, 160)

        Pause(0.25)
        if is_slot_locked then
            BuildSlot()
        end
        SetPropLayer("Town_Camp04_Tent01", true)
        SetPropLayer("Town_Camp04_Tent04", true)
    end
    BUILD.town[9] = function()
        Include(8)

        -- camp D2
        SetPropLayer("_Default_Town_Props_Camp04B", false)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 75%"), 25.69, 2.64, 0, -71)
        CreateBuilder(DCEI.Unit("Builder Town CampTent 65%"), 26.45, 2.76, 0, 60)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 26.66, 3.22, 0)

        Pause(0.125)
        SetPropLayer("Town_Camp04_Tent02", true)
        SetPropLayer("Town_Camp04_Tent03", true)
    end
    BUILD.town[10] = function()
        Include(9)

        -- town level 10 (sword statues)
        CreateBuilderNoFX(DCEI.Unit("Builder Town SwordStatue"), 23.25, 3.85, 0)
        CreateBuilderNoFX(DCEI.Unit("Builder Town SwordStatue"), 24.75, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 23.25, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 24.75, 3.85, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_SwordStatues", true)
    end
    BUILD.town[11] = function()
        Include(10)

        -- block A2
        SetPropLayerNow("Town_Camp01_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.75, 7, 0, -17)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 22.45, 6.95, 0, -53)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 21.66, 7.1, 0, 165)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 22.48, 7.06, 0, 98)
        CreateBuilderNoFX(DCEI.Unit("Builder Town HouseC 30%"), 22.61, 6.96, 0.3, 15)

        Pause(0.25)
        SetPropLayer("Town_Block01_House01", true)
    end
    BUILD.town[12] = function()
        Include(11)

        -- block B1
        SetPropLayerNow("Town_Camp02_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 25.7, 6.97, 0, -115)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 25.55, 7.2, 0, 175)

        Pause(0.25)
        SetPropLayer("Town_Block02_House01", true)
    end
    BUILD.town[13] = function()
        Include(12)

        -- block B2
        SetPropLayerNow("Town_Camp02_Tent03", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.5, 6.94, 0, 127)

        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 26.54, 7.04, 0, -144)

        Pause(0.25)
        SetPropLayer("Town_Block02_House02", true)
    end
    BUILD.town[14] = function()
        Include(13)

        -- block C1
        SetPropLayerNow("Town_Camp03_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.7, 4.27, 0, -110)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 21.27, 4.42, 0.1, -26)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50% Long"), 21.82, 4.4, 0, 91)

        Pause(0.25)
        SetPropLayer("Town_Block03_House01", true)
    end
    BUILD.town[15] = function()
        Include(14)

        -- block D1
        SetPropLayerNow("Town_Camp04_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.29, 4.34, 0, 19)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 26, 4.45, 0, -177)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.4, 4.3, 0, 107)

        Pause(0.25)
        SetPropLayer("Town_Block04_House01", true)
    end
    BUILD.town[16] = function()
        Include(15)

        -- block A3
        SetPropLayerNow("Town_Camp01_Tent04", false)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.65, 6.2, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.83, 6.25, 0, -174)

        Pause(0.25)
        SetPropLayer("Town_Block01_House03", true)
    end
    BUILD.town[17] = function()
        Include(16)

        -- block B3
        SetPropLayerNow("Town_Camp02_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 26.75, 6.33, 0, -1)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 26.81, 6.1, 0, 3)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 27.33, 6.33, 0, 20)

        Pause(0.25)
        SetPropLayer("Town_Block02_House03", true)
    end
    BUILD.town[18] = function()
        Include(17)

        -- block C2
        SetPropLayerNow("Town_Camp03_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.25, 3.52, 0, -84)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.71, 4.49, 0, -116)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.89, 3.82, 0, 100)

        Pause(0.25)
        SetPropLayer("Town_Block03_House02", true)
    end
    BUILD.town[19] = function()
        Include(18)

        -- block D2
        SetPropLayerNow("Town_Camp04_Tent02", false)
        SetPropLayerNow("_Default_Town_Props_Block04A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 25.69, 2.64, 0, -71)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 25.43, 2.15, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 25.47, 2.48, 0, 12)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 25.89, 2.25, -0.1, 150)

        Pause(0.25)
        SetPropLayer("Town_Block04_House02", true)
    end
    BUILD.town[20] = function()
        Include(19)

        -- town level 20 (park statue fence)
        SetPropLayerNow("_Default_Town_Props_Rocks", false)

        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 25.34, 0.84, -0.11, 42)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 25.08, 1.97, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 22.8, 1.93, 0)

        Pause(0.125)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        Pause(0.125)
        SetPropLayer("Town_Upgrade_Statue01_Fence", true)
    end
    BUILD.town[21] = function()
        Include(20)

        -- block A5
        SetPropLayerNow("Town_Camp01_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.33, 6.3, 0, 103)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 21.22, 6.27, 0, 2)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 21.09, 6.91, 0, -80)

        Pause(0.25)
        SetPropLayer("Town_Block01_House04", true)
    end
    BUILD.town[22] = function()
        Include(21)

        -- block B4
        SetPropLayerNow("Town_Camp02_Tent01", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 25.26, 6.45, 0, -71)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 25.16, 6.14, 0, -91)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 24.81, 6.14, -0.05, -91)

        Pause(0.25)
        SetPropLayer("Town_Block02_House04", true)
    end
    BUILD.town[23] = function()
        Include(22)

        -- block B5
        SetPropLayerNow("_Default_Town_Props_Block02B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 26.1, 8, 0, 101)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 26.12, 7.3, 0, -95)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 25.22, 6.76, -0.13, -68)

        Pause(0.25)
        SetPropLayer("Town_Block02_House05", true)
    end
    BUILD.town[24] = function()
        Include(23)

        -- block C3
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 20.43, 3.51, -0.2, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 20.95, 3, 0, 110)

        Pause(0.25)
        SetPropLayer("Town_Block03_House03", true)
    end
    BUILD.town[25] = function()
        Include(24)

        -- block D3
        SetPropLayerNow("Town_Camp04_Tent03", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 26.45, 2.76, 0, 60)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 26.66, 3.22, 0)

        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.35, 2.49, -0.1, 75)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 26.9, 2.86, 0, 133)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.19, 2.57, 0, 137)

        Pause(0.25)
        SetPropLayer("Town_Block04_House03", true)
    end
    BUILD.town[26] = function()
        Include(25)

        -- block A6
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 23.3, 6.17, 0, -80)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 50%"), 23.2, 6.93, 0.1, -2)

        Pause(0.25)
        SetPropLayer("Town_Block01_House06", true)
    end
    BUILD.town[27] = function()
        Include(26)

        -- block B6
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 24.7, 6.7, -0.15, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 24.94, 7.4, 0, -47)

        Pause(0.25)
        SetPropLayer("Town_Block02_House06", true)
    end
    BUILD.town[28] = function()
        Include(27)

        -- block C4
        SetPropLayerNow("Town_Camp03_Tent02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 75%"), 21.59, 2.75, 0, -62)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 22.25, 2.7, 0, -105)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 21.38, 2.53, 0, -174)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 21.85, 2.55, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 22.42, 2.4, 0, -26)

        Pause(0.25)
        SetPropLayer("Town_Block03_House04", true)
    end
    BUILD.town[29] = function()
        Include(28)

        -- block D4
        SetPropLayerNow("Town_Camp04_Tent04", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town CampTent 65%"), 26.77, 3.83, 0, 160)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 26.97, 4.14, 0, -141)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 27.26, 3.5, -0.1, 179)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 27.5, 2.98, -0.1, 5)

        Pause(0.25)
        SetPropLayer("Town_Block04_House04", true)
    end
    BUILD.town[30] = function()
        Include(29)

        -- town level 30 (park statue construction)
        SetPropLayerNow("_Default_Town_Props_Block04C", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.05, 1.28, 0, 97)

        CreateBuilder(DCEI.Unit("Builder Town KnightStatue Broken"), 22.77, 1.51, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town KnightStatue Broken"), 25.37, 1.51, 0, 18)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_Statue02", true)
    end
    BUILD.town[31] = function()
        Include(30)

        -- block A8
        SetPropLayerNow("_Default_Town_Props_Block01A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 21.3, 8, 0, 97)

        CreateBuilder(DCEI.Unit("Builder Town HouseF 50% Long"), 20.64, 7.35, 0, 10)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 21.38, 7.55, 0, 83)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 55%"), 22.02, 7.55, -0.1, 95)

        Pause(0.25)
        SetPropLayer("Town_Block01_House05", true)
    end
    BUILD.town[32] = function()
        Include(31)

        -- block B7
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 27.1, 7, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.81, 6.61, -0.05, -92)

        Pause(0.25)
        SetPropLayer("Town_Block02_House07", true)
    end
    BUILD.town[33] = function()
        Include(32)

        -- block B8
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 25.48, 7.87, 0, 100)

        Pause(0.25)
        SetPropLayer("Town_Block02_House08", true)
    end
    BUILD.town[34] = function()
        Include(33)

        -- block C5
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.85, 2.4, -0.05, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 20.73, 1.57, -0.1, 170)

        Pause(0.25)
        SetPropLayer("Town_Block03_House05", true)
    end
    BUILD.town[35] = function()
        Include(34)

        -- block D5
        SetPropLayerNow("_Default_Town_Props_Block04B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 26.07, 1.69, 0, 101)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 26.21, 1.82, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.61, 1.99, -0.15, -100)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 27.03, 2.12, 0, 90)

        Pause(0.25)
        SetPropLayer("Town_Block04_House05", true)
    end
    BUILD.town[36] = function()
        Include(35)

        -- block A9
        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.5, 7.7, 0, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 23.1, 7.77, 0, -91)

        Pause(0.25)
        SetPropLayer("Town_Block01_House09", true)
    end
    BUILD.town[37] = function()
        Include(36)

        -- block B9
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.02, 7.81, -0.1, 3)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 50%"), 26.5, 7.83, -0.2, -85)

        Pause(0.25)
        SetPropLayer("Town_Block02_House09", true)
    end
    BUILD.town[38] = function()
        Include(37)

        -- block C6
        SetPropLayerNow("_Default_Town_Props_Block03A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 21.23, 1.17, -0.19, 19)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 21.68, 0.47, -0.33, 82)

        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 21.23, 1.68, -0.05, 79)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 21.78, 1.26, -0.15, -21)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 21.8, 2.01, 0, 133)

        Pause(0.25)
        SetPropLayer("Town_Block03_House06", true)
    end
    BUILD.town[39] = function()
        Include(38)

        -- block D6
        SetPropLayerNow("_Default_Town_Props_Block04D", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 80%"), 26.08, 0.36, -0.16, 63)
        CreateBuilderDeath(DCEI.Unit("Builder Town Tree 65%"), 26.75, 1.06, -0.05, 19)

        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.41, 1.6, 0, 90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 40%"), 26.84, 1.58, 0, 175)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 27.59, 2.26, 0, -160)

        Pause(0.25)
        SetPropLayer("Town_Block04_House06", true)
    end
    BUILD.town[40] = function()
        Include(39)

        -- town level 40 (park statue construction)
        SetPropLayerNow("Town_Upgrade_Statue01_Fence", false)
        SetPropLayerNow("Town_Upgrade_Statue02", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town KnightStatue Broken"), 22.77, 1.51, 0, -19)
        CreateBuilderDeath(DCEI.Unit("Builder Town KnightStatue Broken"), 25.37, 1.51, 0, 18)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 25.2, 1.6, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 150%"), 22.6, 1.5, 0)

        CreateBuilder(DCEI.Unit("Builder Town KnightStatue"), 22.74, 1.6, 0, -19)
        CreateBuilder(DCEI.Unit("Builder Town KnightStatue"), 25.4, 1.6, 0, 18)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_Statue03", true)
    end
    BUILD.town[41] = function()
        Include(40)

        -- block A11
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 20.33, 8.6, -0.1, -174)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 20.33, 8.04, 0, -100)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 20.16, 7.2, -0.1, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.57, 6.68, 0, -176)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 20.62, 6.08, 0, 5)

        Pause(0.25)
        SetPropLayer("Town_Block01_House10", true)
    end
    BUILD.town[42] = function()
        Include(41)

        -- block B10
        SetPropLayerNow("_Default_Town_Props_Block02A", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 27.05, 8.36, 0, -158)

        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 27.06, 7.7, 0, -4)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 27.77, 7.78, 0.05, 4)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 27.64, 7.15, 0, 175)

        Pause(0.25)
        SetPropLayer("Town_Block02_House10", true)
    end
    BUILD.town[43] = function()
        Include(42)

        -- block B11
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.11, 8.33, 0, 2)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 50%"), 26.61, 8.33, -0.1, 11)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 27.06, 8.18, 0, 179)

        Pause(0.25)
        SetPropLayer("Town_Block02_House11", true)
    end
    BUILD.town[44] = function()
        Include(43)

        -- block C7
        CreateBuilder(DCEI.Unit("Builder Town HouseB 45%"), 21.3, 1.04, 0, 180)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 21.85, 0.5, 0, 163)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 21.42, 0.3, -0.1, -98)

        Pause(0.25)
        SetPropLayer("Town_Block03_House07", true)
    end
    BUILD.town[45] = function()
        Include(44)

        -- block D7
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 26.4, 1.2, -0.22, 179)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 26.51, 0.68, 0, -89)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 26.98, 1.1, 0, 80)

        Pause(0.25)
        SetPropLayer("Town_Block04_House07", true)
    end
    BUILD.town[46] = function()
        Include(45)

        -- block A12
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 20.87, 8.2, 0, -88)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 50%"), 21.48, 8.2, 0, 87)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 60%"), 22.03, 8.26, 0, -10)

        Pause(0.25)
        SetPropLayer("Town_Block01_House12", true)
    end
    BUILD.town[47] = function()
        Include(46)

        -- block B12
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 27.24, 8.75, 0, 120)
        CreateBuilder(DCEI.Unit("Builder Town HouseF 60%"), 27.77, 8.36, 0, 91)

        Pause(0.25)
        SetPropLayer("Town_Block02_House12", true)
    end
    BUILD.town[48] = function()
        Include(47)

        -- block C8 & D8
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 22.15, 0.88, 0, -42)
        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 22.49, 0.45, 0, -175)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 23.14, 0.53, 0, 80)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50%"), 24.68, 0.59, 0, 92)
        CreateBuilder(DCEI.Unit("Builder Town HouseD 55%"), 25.22, 0.49, -0.15, -177)
        CreateBuilder(DCEI.Unit("Builder Town HouseB 50%"), 25.82, 0.72, 0, -25)

        Pause(0.25)
        SetPropLayer("Town_Block03_House08", true)
    end
    BUILD.town[49] = function()
        Include(48)

        -- block A13 & B13
        SetPropLayerNow("_Default_Town_Props_Block01B", false)
        CreateBuilderDeath(DCEI.Unit("Builder Town Bush 80%"), 22.5, 8.56, 0, -14)

        CreateBuilder(DCEI.Unit("Builder Town HouseC 45%"), 22.52, 8.32, 0, -90)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 45% Long"), 23.1, 8.5, 0, -91)
        CreateBuilder(DCEI.Unit("Builder Town HouseE 50% Long"), 24.88, 8.2, 0, -170)
        CreateBuilder(DCEI.Unit("Builder Town HouseA 50%"), 25.53, 8.42, 0, 90)

        Pause(0.25)
        SetPropLayer("Town_Block01_House13", true)
    end
    BUILD.town[50] = function()
        Include(49)

        -- town level 50 (town hall upgrade)
        SetPropLayerNow("Town_Upgrade_SwordStatues", false)
        SetPropLayerNow("_Default_Town_Props_TownHall", false)
        CreateBuilder(DCEI.Unit("Builder Town HallAltar"), 24, 4.31, -0.17, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 23.25, 3.85, 0)
        CreateBuilder(DCEI.Unit("Builder FX Dust 050%"), 24.75, 3.85, 0)

        Pause(0.25)
        SetPropLayer("Town_Upgrade_TownHall", true)
    end
end

-- !SECTION

-- SECTION Functions
function EnforceUnlockedSlots()
    local unlock = data.GetUnlockedSlotCount(GetCastleLevel())

    -- unequip heroes in locked slots
    for n = 1, data.unlock.hero_slot_max do
        if player.heroes[n] and n > unlock.hero_slots then
            UnequipHeroSlot(n)
        end
    end

    -- unequip towers in locked slots
    for n = 1, data.unlock.tower_slot_max do
        if player.towers[n] and n > unlock.tower_slots then
            UnequipTowerSlot(n)
        end
    end

    -- unequip town buildings in locked slots
    for n = 1, data.unlock.town_slot_max do
        if player.town[n] and n > unlock.town_slots then
            UnequipTownSlot(n)
        end
    end

    -- unequip defenses in locked slots
    for n = 1, data.unlock.defense_slot_max do
        if player.defenses[n] and n > unlock.defense_slots then
            UnequipDefenseSlot(n)
        end
    end
end

function GetEmptyHeroSlot()
    local solution = 1
    for i = 1, #ui.loadout.hero.slots do
        local unit = ui.loadout.hero.slots[i]
        if not DCEI.UnitExists(unit) and not unit.locked then
            solution = i
        end
    end
    return solution
end

function GetEmptyTowerSlot()
    local solution = 1
    for i = 1, #ui.loadout.tower.slots do
        local unit = ui.loadout.tower.slots[i]
        if not DCEI.UnitExists(unit) and not unit.locked then
            solution = i
        end
    end
    return solution
end

-- DEBUG
local zoom_level = 0
function TestZoomLevel()
    local levels = 3
    zoom_level = (zoom_level % levels) + 1

    SetFormation(zoom_level)
end

function DebugPlayerXP()
    local xp_threshold = data.GetPlayerLevelXP(GetPlayerLevel())
    local new_xp = xp_threshold + player.player_xp

    if player.player_xp < xp_threshold then
        DCEI.LogMessage("==XP ADJUSTMENT==")
        SetPlayerLevel(data.GetLevelFromXPValue(player.player_xp))
        AddPlayerXP(new_xp)
    end
end

function OnUnitSpawn()
    local u = DCEI.TriggeringUnit
    local owner = DCEI.GetUnitPlayerId(u)
    local type = DCEI.GetUnitCategory(u)

    if type == "Unit" then
        -- only run these checks on actual units
        if owner == -1 then
            ProcessMonsterUnit(u)
        elseif owner == 1 then
            ProcessSummonedUnit(u)
        end
    end
end

function OnUnitRemoved()
    local u = DCEI.TriggeringUnit
    local id = DCEI.GetUnitPlayerId(u)

    if id == -1 then
        wave.UpdateWaveUnitList(u)
    end
end

function OnUnitDied(killing_unit)
    local u = DCEI.TriggeringUnit
    local id = DCEI.GetUnitPlayerId(u)

    local charmed_monster = false
    if id == 1 and DCEI.GetUnitBehaviorStackCount(u, "Equipment Epic Attribute Charm Debuff") > 0 then
        charmed_monster = true
    end

    if (id == -1 and DCEI.GetUnitCategory(u) ~= "Missile") or charmed_monster then
        local xp_value = DCEI.GetMaxMana(u)
        local bounty = DCEI.GetUnitBounty(u)
        if DCEI.GetUnitBehaviorStackCount(u, "Hero WindKnight Skill Golden Wind Drop Gold Trigger") > 0 then
            local additional_gold = math.floor(bounty * 0.05)
            bounty = bounty + additional_gold
            ui.CreateFloatingLootDisplay(u, "icon_coin00", additional_gold)
        end

        -- track gold earned per wave for gold reward ads on defeat
        wave.data.earned = wave.data.earned + bounty
        AddPlayerGold(bounty, true)
        AddPlayerXP(xp_value)

        if wave.data.ModeStandard() then
            local name = DCEI.GetUnitType(u)
            if wave_killed_monsters[name] then
                wave_killed_monsters[name] = wave_killed_monsters[name] + 1
            end
            if player.running_monster_bounty and name == player.running_monster_bounty.monster then
                if player.running_monster_bounty.progress < player.running_monster_bounty.count then
                    player.running_monster_bounty.progress = player.running_monster_bounty.progress + 1
                    --ui.monster_bounty_battle:UpdateProgress(player.running_monster_bounty.progress)
                end
            end
        end

        if UnitIsBoss(u) then
            if wave.data.ModeStandard() then
                RollForContractDrop()
            end
            if wave.data.mode == "raid_boss" then
                ProcessRaidBossDrop()
            end
            if wave.data.mode == "hell_mode" then
                ui.DestroyBossBar()
                wave.data.is_boss_spawned = false
                wave.data.boss = nil
                HellModeAddScore(49)
            end
        end

        if wave.data.mode == "hell_mode" then
            HellModeAddScore(1)
            HellModeKillGemCooldown()
        end

        -- check for season event drops
        ProcessSeasonEventDrops(u)

        -- check for loot drops
        ProcessLootDrops(u)

        -- check rare monster drops
        ProcessRareMonsterDrops(u)

        -- killing objective/colony immediately ends wave
        EndWaveIfUnitIsWaveObjective(u)

        wave.UpdateWaveUnitList(u, CheckIfWaveIsOver)
    end
end

function OnUnitBehaviorAdd()
    local u = DCEI.TriggeringUnit
    local b = DCEI.TriggeringBehaviorName

    if BehaviorCallbacks[b] then
        BehaviorCallbacks[b](u)
    end

    if wave.data.is_active and InBattleBehaviorCallbacks[b] then
        InBattleBehaviorCallbacks[b](u)
    end
end

function OnAbilityCast(target, pos)
    local u = DCEI.TriggeringUnit
    local ability = DCEI.TriggeringAbilityName

    if ability == "Shared ManualTarget" and pos then
        CastTargetModeAbility(pos)
    end
end

function OnAbilityCooldownChange(ready)
    local u = DCEI.TriggeringUnit
    local owner = DCEI.GetUnitPlayerId(u)

    if wave.data.is_active and owner == 1 then
        if ready then
            EnableHeroSkill(u)
        else
            StartHeroCooldownVisuals(u)
        end
    end
end

function ProcessHeroStun(unit, stun)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]
    if unit_data then
        local ability = data.GetHeroActiveSkillData(unit_data, "ability")

        if unit_data and ability then
            local slot = unit_data.slot
            local skill_bar = ui.loadout.hero.slots[slot].skill_bar

            unit_data.stun = stun
            if unit_data.skill.active then
                skill_bar:Disable(stun)
            end
        end
    end
end

function OnFocusChange(has_focus)
    if not has_focus and not wave.data.is_active then
        DCEI.LogMessage("> Commit Resources On Lost Focus")
        bank.CommitGold()
        bank.CommitElixir()
        bank.CommitVaultGold()
    end
    if has_focus then
        local function ResultCallbcak(time_stamp)
            ScheduleNotification(time_stamp)
            if time_stamp then
                if not GetPlayerTest("pause_offline_gold") and player.system.vault == system.active then
                    if time_stamp > data.vault.time_record then
                        if GetVaultGold() < data.vault.capacity then
                            local time_diff = time_stamp - data.vault.time_record
                            local offline_gold = math.floor(time_diff / data.vault.interval)
                                * data.vault:GetVaultIncome()

                            GenerateVaultGold(offline_gold)
                        else
                            ui.bottom_panel:ShowTownPip()
                        end
                    end

                    data.vault.time_record = time_stamp
                    bank.CommitVaultRecord()

                    ScheduleVaultNotification()
                end
            end
        end

        AttemptGetServerTime(ResultCallbcak)
        CheckOSDaysElapsed(true)
    else
        if wave.data.is_active then
            if PlayerInTargetMode() then
                ExitTargetMode()
            end
        end
    end
end

-- Debug Commands
function OnTextCommand(text_command)
    local print = ""

    if text_command == "wave" then
        print = wave.DebugLog()
    elseif text_command == "p" then
        print = LogPlayerAttributes()
    elseif text_command == "z" then
        TestZoomLevel()
    elseif text_command == "r" then
        DCEI.RestartApplication()
    elseif text_command == "b" then
        NextCastleBuild()
    end

    DCEI.ShowFeedbackMessage(print)
end

-- Functions
function CreateCastleUnits()
    InitializeCastle()
    DCEI.TriggerAddTimerEventElapsed(InitializeArcher, 0.25, true, true)

    if CASTLE_TEST then
        -- for testing castle build ugprades
        SetCastleLevel(CASTLE_TEST)
    end

    BuildCastle(GetCastleLevel())
end

-- Town/Castle Transition
function TownTransition()
    if wave.data.is_active then
        -- prevent town transition after wave has started
        return
    end

    if cameras.current == cameras.castle then
        -- show town
        CreateTownUI()
        CloseActivePrompts()

        MoveCamera(cameras.town)
        if not event_mode.IsActive() then
            Tutorial05_Mine_StepB()
            Tutorial07_Colony_StepB()
            Tutorial08_TownUpgrade_Check()
            Tutorial08_1_Mine_Vault_Check()
        end
    else
        -- show castle
        CreateCastleUI()
        CloseActivePrompts()

        MoveCamera(cameras.castle)

        Tutorial10_EquipGear_Castle_Check()
    end
end

function PlayWaveStartSound()
    DCEI.PlaySound(DCEI.Sound("call_wave_v1"))
end

function PlayGoldSpendSound()
    DCEI.PlaySound(DCEI.Sound("ui_item_purchased"))
end

function PlayCastleConstructionSound()
    DCEI.PlaySound(DCEI.Sound("tower_build_v2"))
    DCEI.PlaySound(DCEI.Sound("awaken_sucessed"))
end

function NextWave()
    return {
        mode = "standard",
        rating = wave.data.GetMax() + 1,
    }
end

function LastWave()
    return {
        mode = "replay",
        rating = wave.data.GetMax(),
    }
end

function WaveStartMessage(wave_data)
    if wave.data.ModeStandard() then
        ui.ShowFeedbackWaveStart(wave_data.rating, wave_data.skip)
    end
end

-- NOTE Wave Skip only affect some old players have abtest, can remove safely
function ShowWaveSkipWindow(nextWave)
    if cameras.current == cameras.town then
        MoveCamera(cameras.castle)
    end
    local function OnClose(frame_only)
        if not frame_only and not ui.shop.is_active then
            WaveStartActions(nextWave)
        end
        ui.skip_wave = nil
    end

    CloseActiveMenus()
    CloseActivePrompts()
    CreateBattleUI()
    local function Skip(skip_wave, cost)
        -- check gem
        local gems = GetPlayerGems()

        if cost <= gems then
            local spend_event = { type = "buy", target = "skip waves " .. skip_wave }
            SpendPlayerGems(cost, spend_event)
            PlayGoldSpendSound()

            local wave_after_skip = {
                mode = "standard",
                rating = wave.data.GetMax() + skip_wave + 1,
                skip = true,
            }
            AttemptCloseSkipWave()
            local current_wave = wave.data.GetMax()
            wave.data.start_wave = current_wave
            wave.data.skip_number = skip_wave
            IncrementWaveSkipAttempt()
            WaveStartActions(wave_after_skip)
        else
            local function OnStore()
                AttemptCloseSkipWave()
                if wave.data.is_active then
                    WaveEndActions()
                    CreateBattleUI()
                end
            end

            local function OnStoreClose()
                ShowWaveSkipWindow(nextWave)
            end

            local offset = -356
            ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore, OnStoreClose)
        end
    end

    local skip_settings = {
        Skip = Skip,
        {
            skip_wave = 5,
            cost = 7,
        },
    }
    if wave.data.GetMax() >= 50 then
        table.insert(skip_settings, {
            skip_wave = 10,
            cost = 12,
        })
    end
    if wave.data.GetMax() >= 100 then
        table.insert(skip_settings, {
            skip_wave = 20,
            cost = 20,
        })
    end

    ui.skip_wave = ui.CreateSkipWaveMenu(OnClose, skip_settings, ui.upper)
end

-- ANCHOR: Wave Start
function WaveStartActions(wave_data, is_auto, autobattle_type)
    DCEI.LogMessage("Start the wave")
    local wave_id = wave_data.rating
    wave.data.is_active = true
    wave.data.unfinished = nil
    -- prevent opening menus as wave starts
    CloseActiveMenus()
    CloseActivePrompts()
    PlayWaveStartSound()

    -- autobattle always enabled in event mode (per hero)
    if event_mode.IsActive() then
        EnableAutoBattle()
    end

    -- workers are removed and turned into a timer for optimization
    DwarfWorkers_UseTimers()

    -- for performance remove building when wave start
    RemoveBuildingsOnWaveStart()

    -- reset performance stats (performance stats is sent on "Stage Win" logging event)
    DCEI.PerformanceTracking.ResetStats()

    -- reset battle stats and autobattle priority
    ResetBattleData()
    GenerateAutoBattlePriorityList()

    for _, unit in pairs(group_archers) do
        DCEI.ApplyBehaviorAsync(unit, "System Hero InCombat")
        ColonyTargetingCheck(unit, wave_data)
    end
    for _, unit in pairs(player.towers) do
        ColonyTargetingCheck(unit, wave_data)
    end
    for _, unit in pairs(player.defenses) do
        ColonyTargetingCheck(unit, wave_data)
    end

    for _, unit in pairs(player.heroes) do
        DCEI.ApplyBehaviorAsync(unit, "System Hero InCombat")
        ColonyTargetingCheck(unit, wave_data)

        local battle_start = true
        ShowUnitSkillBar(unit, true, battle_start)
        EnableHeroSkill(unit)
        ApplyHeroInCombatBehavior(unit)
    end

    if cameras.current == cameras.town then
        MoveCamera(cameras.castle)
    end

    wave.SpawnWave(wave_data)
    if wave.data.ModeStandard() then
        ClearMonsterKillList()
        wave.data.monster_data = player.monster_data
    end
    -- devil's trail on affect in standard wave
    if wave.data.mode == "standard" and DevilTrailActive() then
        ApplyDevilTrailBehaviors()
    end
    WaveStartMessage(wave_data)
    AssignRareMonsterDrops(wave_id)
    -- Check token is full
    local max_gems = 50 + player.attributes.token_cap
    if DCEI.GetProtectedValue(data.safe.gems) >= max_gems then
        ui.ShowFeedbackTokenFull()
    end

    if not wave.data.objective and wave.data.mode ~= "hell_mode" then
        -- don't create wave bar if objective exists (ie colony)
        ui.CreateWaveBar(wave_data, wave.data.duration)
    end

    SetGameSpeed(player.speed)
    ui.bottom_panel.speed:Update(player.speed)

    CreateBattleUI(true)
    -- Hide Back Button in Wave 1
    if wave_id == 1 and wave.data.mode == "standard" then
        DCEI.SetFrameActive(ui.bottom_panel.left_outer.frame, false)
    end
    StartWaveTimer()
    SpawnSummons(wave_data)

    -- floating autobattle button
    if is_auto then
        EnableAutoBattle(autobattle_type)
    else
        -- unlock auto wave after wave 25
        local unlock_auto = 25
        if wave.data.mode == "standard" and (wave_id > unlock_auto or DevMode()) then
            -- auto battle only available in standard (not replay)
            local show_pip = wave_id == (unlock_auto + 1)
            ShowAutoBattleButton(show_pip)
        end
    end

    -- floating speedup button
    if player.speed_up then
        AttemptShowSpeedButton()
    end

    -- event logging
    if wave.data.mode == "replay" then
        -- LogEvent(LogEvents.wave_replay)
    else
        LogEvent(LogEvents.wave_start)
        if not event_mode.IsActive() then
            -- tutorial 01 hero skills
            Tutorial01_Hero_Check(wave_id)
        end
    end
    SetGameState("Wave Start " .. wave.data.mode)
end

function EndWaveIfUnitIsWaveObjective(u)
    if u == wave.data.objective then
        wave.data.is_spawning = false

        -- remove all wave data units except for objective, this way wave will end with objective death animation
        local units = {}
        for _, unit in pairs(wave.data.units) do
            table.insert(units, unit)
        end
        for k, unit in pairs(units) do
            if u ~= unit then
                DCEI.KillUnit(unit)
            end
        end
    end
end

function CheckIfWaveIsOver()
    if wave.data.is_active and not wave.data.is_spawning and #wave.data.units == 0 then
        WaveEndActions(true)
    end
end

-- NOTE Shop pop up
function CheckToQueueDealsPopup()
    local autobattle_unlock_wave = 25
    local system_autobattle_popup = not player.system.autobattle_popup
        or player.system.autobattle_popup == system.locked
    local unlock_wave = 30
    local token_pack_wave = 100
    local unlock_check = wave.data.GetMax() >= unlock_wave and wave.data.GetMax() < token_pack_wave

    -- queue popup on auto battle
    local key = "autobattle_pack"
    if system_autobattle_popup and wave.data.GetMax() > autobattle_unlock_wave and AnyBundleItemsAreAvailable(key) then
        player.queued_deal_source = "autobattle offer"
        player.queued_deal_popup = true
    end

    -- queue popup on loss-then win between waves 30 and 100
    local key = "starter_pack"
    if
        player.wave_attempts
        and player.wave_attempts > 0
        and unlock_check
        and AnyBundleItemsAreAvailable(key)
        and player.system.deal_popup == system.locked
    then
        player.queued_deal_source = "loss-win offer"
        player.queued_deal_popup = true
    end

    -- queue popup on beat wave 100
    if wave.data.GetMax() == token_pack_wave then
        player.queued_deal_source = "token pack offer"
        player.queued_deal_popup = true
    end
end

function CheckForQueuedDealPopup()
    if player.queued_deal_popup then
        local key
        local text
        local on_close

        -- prevent gold ad from displaying at the same time
        SetAdOnCooldown(ad_data.gold, 2)

        -- mark queued popup as seen
        player.queued_deal_popup = false

        if player.queued_deal_source == "loss-win offer" then
            -- only show win-loss popup once
            player.system.deal_popup = system.active
            bank.CommitSystemState()
            key = "starter_pack"
        elseif player.queued_deal_source == "token pack offer" then
            -- show token pack and add to deals page when dismissed
            key = "token_pack"

            -- remove starter pack
            local items = GetStoreCollectionData("starter_pack")
            ui.shop:RemoveItems(items)

            if AnyBundleItemsAreAvailable(key) then
                AddItemsToShopDealsFromCollectionKey(key)
            end
        elseif player.queued_deal_source == "autobattle offer" then
            -- show autobattle pack and add to deals page with limited time when dismissed
            key = "autobattle_pack"

            player.system.autobattle_popup = system.active
            bank.CommitSystemState()
            if AnyBundleItemsAreAvailable(key) then
                local options = { expires_in = 86340 }
                AddItemsToShopDealsFromCollectionKey(key, options)
            end
        end

        -- log entering shop source
        local from = tostring(player.queued_deal_source)
        SetStoreEntrySource(from)

        OpenDealsPopup(key, text, on_close)
    end
end

function IncrementWaveSkipAttempt()
    player.wave_skip_attempt = player.wave_skip_attempt or 0
    player.wave_skip_attempt = player.wave_skip_attempt + 1

    bank.CommitWaveSkipAttempCount()
end

function IncrementWaveSkipSuccess()
    player.wave_skip_success = player.wave_skip_success or 0
    player.wave_skip_success = player.wave_skip_success + 1

    bank.CommitWaveSkipSuccessCount()
end

function IncrementWaveSkipFailure()
    player.wave_skip_failure = player.wave_skip_failure or 0
    player.wave_skip_failure = player.wave_skip_failure + 1

    bank.CommitWaveSkipFailureCount()
end

function IncrementWaveLossCount()
    player.wave_loss_count = player.wave_loss_count or 0
    player.wave_loss_count = player.wave_loss_count + 1

    bank.CommitWaveLossCount()
end

function IncrementColonyLossCount()
    player.colony_loss_count = player.colony_loss_count or 0
    player.colony_loss_count = player.colony_loss_count + 1

    bank.CommitColonyLossCount()
end

function IncrementReplayCount()
    player.replay_count = player.replay_count or 0
    player.replay_count = player.replay_count + 1

    bank.CommitReplayCount()
end

function IncrementPurchaseCount()
    player.purchase_count = player.purchase_count or 0
    player.purchase_count = player.purchase_count + 1

    bank.CommitPurchaseCount()
end

function IncrementPurchaseSpend(price)
    player.purchase_spend = player.purchase_spend or 0
    player.purchase_spend = player.purchase_spend + price

    bank.CommitPurchaseSpend()
end

function IncrementEquipmentUpgradeCount()
    player.equipment_upgrades = player.equipment_upgrades or 0
    player.equipment_upgrades = player.equipment_upgrades + 1

    bank.CommitEquipmentUpgrade()
end

function IncrementEnchantNonEpicCount()
    player.enchant_non_epic = player.enchant_non_epic or 0
    player.enchant_non_epic = player.enchant_non_epic + 1

    bank.CommitEnchantNonEpicCount()
end

function IncrementEnchantEpicCount()
    player.enchant_epic = player.enchant_epic or 0
    player.enchant_epic = player.enchant_epic + 1

    bank.CommitEnchantEpicCount()
end

function IncrementCraftCount(rarity)
    player.craft_rarity[rarity] = player.craft_rarity[rarity] or 0
    player.craft_rarity[rarity] = player.craft_rarity[rarity] + 1

    local event = LogEvents.CraftEquipment(rarity)
    LogEvent(event)

    local key = "craft" .. "|" .. rarity
    DCEI.Save.Set(key, player.craft_rarity[rarity])
    bank.TriggerCommit()
end

function IncrementCraftRefreshCount(currency)
    player.craft_refresh_currency[currency] = player.craft_refresh_currency[currency] or 0
    player.craft_refresh_currency[currency] = player.craft_refresh_currency[currency] + 1

    local event = LogEvents.RefreshCraft(currency)
    LogEvent(event)

    local key = "refresh_craft|" .. currency
    DCEI.Save.Set(key, player.craft_refresh_currency[currency])
    bank.TriggerCommit()
end

function IncrementElementTrialStartCount()
    player.element_trial_start = player.element_trial_start or 0
    player.element_trial_start = player.element_trial_start + 1

    bank.CommitElementTrialStartCount()
end

function IncrementElementTrialCompleteCount()
    player.element_trial_complete = player.element_trial_complete or 0
    player.element_trial_complete = player.element_trial_complete + 1

    bank.CommitElementTrialCompleteCount()
end

function UpdateWaveAttempts(is_victory, is_defeat)
    -- wave attempts track how many times a player has attempted the current max wave
    if is_victory then
        -- reset to 0 attempt count on victory
        player.wave_attempts = 0

        bank.CommitWaveAttempts()
    elseif is_defeat then
        -- increment attempt count on defeat
        player.wave_attempts = player.wave_attempts + 1

        LogUserProperty(UserProps.last_loss)

        bank.CommitWaveAttempts()

        IncrementWaveLossCount()
    else
        -- log last wave exit
        LogUserProperty(UserProps.last_exit)
    end
end

function UpdateRaidBossLog(is_victory, is_defeat)
    if not is_victory and not is_defeat then
        -- exit battle, don't log
        return
    end
    local boss_index = levels.raid_boss.current.index
    if boss_index ~= player.raid_boss_log.last_attempt_boss then
        player.raid_boss_log.failed_attempts = 0
    end
    player.raid_boss_log.last_attempt_boss = boss_index
    player.raid_boss_log.total_attempts = player.raid_boss_log.total_attempts or 0
    player.raid_boss_log.total_attempts = player.raid_boss_log.total_attempts + 1

    LogUserProperty(UserProps.raid_boss_attempts)

    local key = "Raid Boss " .. boss_index
    if is_victory then
        local victory_count = player.raid_boss_log[key] or 0
        if victory_count == 0 then
            DCEI.LogMessage("first win " .. key)
            LogEvent(LogEvents.first_raid_boss_win)
        end
        player.raid_boss_log[key] = victory_count + 1
        player.raid_boss_log.failed_attempts = 0
        LogUserProperty(UserProps.raid_boss_victories)
        local user_prop = UserProps.raid_boss_victory(key, victory_count)
        LogUserProperty(user_prop)
    elseif is_defeat then
        player.raid_boss_log.failed_attempts = player.raid_boss_log.failed_attempts or 0
        player.raid_boss_log.failed_attempts = player.raid_boss_log.failed_attempts + 1
        DCEI.LogMessage("failed_attempts  " .. player.raid_boss_log.failed_attempts)
    end

    bank.CommitRaidBossLog()
end

-- !SECTION

-- SECTION: Elemental Trial
function HasElementTrail()
    return player.attributes.bless_element and player.attributes.bless_wave > 0
end

-- used crystal declarations
DCEI.Unit("Blessing Crystal Fire")
DCEI.Unit("Blessing Crystal Cold")
DCEI.Unit("Blessing Crystal Lightning")
DCEI.Unit("Blessing Crystal Poison")
DCEI.Unit("Blessing Crystal Magic")
DCEI.Behavior("Elemental Trial VFX Trigger Fire")
DCEI.Behavior("Elemental Trial VFX Trigger Cold")
DCEI.Behavior("Elemental Trial VFX Trigger Lightning")
DCEI.Behavior("Elemental Trial VFX Trigger Poison")
DCEI.Behavior("Elemental Trial VFX Trigger Magic")
function CreateBless()
    if bless_crystal and DCEI.UnitExists(bless_crystal) then
        DCEI.RemoveUnit(bless_crystal)
    end
    local crystal_name = "Blessing Crystal " .. player.attributes.bless_element
    bless_crystal = DCEI.CreateUnitAsync(1, 1, crystal_name, 24, 28)
    CreateBlessUI()
    ui.bless_ui:Update()
end

function CreateBlessUI()
    if not ui.bless_ui then
        ui.bless_ui = ui.CreateBlessUI()
        DCEI.SetOnClickCallback(ui.bless_ui.BonusIcon, CreateBlessUIPopup)
    end
end

function CreateBlessUIPopup()
    UnblockInput()
    ui.DestroyBigHand()
    if ui.bless_ui.has_bless then
        local layout = gmui.Layout:New({
            parent = ui.upper,
            name = "ElementalTrial/Popup",
        })
        if wave.data.is_active then
            SetGameSpeed(0)
        end
        -- update content
        DCEI.SetFrameImage(layout.ElementIcon, ui.element_icon[player.attributes.bless_element])
        local description_keys = {
            ["Fire"] = "data/bonus/damage_fire/all",
            ["Cold"] = "data/bonus/damage_ice/all",
            ["Lightning"] = "data/bonus/damage_lightning/all",
            ["Poison"] = "data/bonus/damage_poison/all",
            ["Magic"] = "data/bonus/damage_arcane/all",
        }
        local rich_text_element_icon = {
            ["Fire"] = "fire",
            ["Cold"] = "affinity_cyan",
            ["Lightning"] = "affinity_lemon",
            ["Poison"] = "affinity_teal",
            ["Magic"] = "affinity_magenta",
        }
        local element_description = DCEI.GetTranslationText(
            description_keys[player.attributes.bless_element],
            { x = "<color=yellow>+20%</color>" }
        )
        DCEI.SetTextFrameText(layout.ElementalDescription, element_description)
        local bonus_icon = player.attributes.bless_bonus == 2 and "talent_bonus_xp" or "talent_bonus_gold"
        DCEI.SetFrameImage(layout.BonusIcon, bonus_icon)
        local bonus_type = player.attributes.bless_bonus == 2 and "data/bonus/xp" or "data/bonus/gold"
        local bonus_number = player.attributes.bless_bonus_count and player.attributes.bless_bonus_count * 5 or 0
        local total_bonus = DCEI.GetTranslationText(bonus_type, { x = "<color=green>+" .. bonus_number .. "%</color>" })
        DCEI.SetTextFrameText(layout.TotalBonus, total_bonus)
        local rich_text_icon = rich_text_element_icon[player.attributes.bless_element]
        local bonus_description = "<color=yellow>+5%</color> "
            .. DCEI.GetTranslationText("ui/element_trail/explain", { x = "<sprite name=" .. rich_text_icon .. ">" })
        DCEI.SetTextFrameText(layout.BonusDescription, bonus_description)
        local last = " <color=green>"
            .. player.attributes.bless_wave
            .. "</color> "
            .. DCEI.GetTranslationText("ui/wave_skip/wave")
        DCEI.SetTextFrameText(layout.RemainingWave, DCEI.GetTranslationText("ui/calendar/seasonpass/remaining") .. last)
        DCEI.SetOnClickCallback(layout.CloseButton, function()
            if wave.data.is_active then
                SetGameSpeed(player.speed)
            end
            if ui.bless_description and DCEI.FrameExists(ui.bless_description.Frame) then
                DCEI.DestroyFrame(ui.bless_description.Frame)
                ui.bless_description = nil
            end
        end)
        ui.bless_description = layout
    end
end

function BlessCrystalSpawnCoolEffect()
    local unit = DCEI.CreateUnitAsync(1, 1, DCEI.Unit("Blessing Fall Meteror"), 24, 27)
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.RemoveUnit(unit)
    end, 1.25, false, true)
end

function HideBlessCrystal()
    if bless_crystal and DCEI.UnitExists(bless_crystal) then
        DCEI.RemoveUnit(bless_crystal)
        bless_crystal = nil
    end
    if ui.bless_ui then
        DCEI.SetFrameActive(ui.bless_ui.Frame, false)
    end
end

function ShowBlessCrystal()
    if player.attributes.bless_element and player.attributes.bless_wave > 0 and not bless_crystal then
        local crystal_name = "Blessing Crystal " .. player.attributes.bless_element
        bless_crystal = DCEI.CreateUnitAsync(1, 1, crystal_name, 24, 28)
    end
    if ui.bless_ui then
        ui.bless_ui:Update()
    end
end

function RollElementBless(wave_id)
    if HasDevilTrail() or HasElementTrail() or wave_id == 100 then
        return
    end

    if wave_id == 50 then
        -- hardcode give fire bonus from wave 51 ~ wave 60
        player.attributes.bless_element = "Fire"
        player.attributes.bless_wave = 10
        player.attributes.bless_bonus = 1
    elseif wave_id == 70 then
        -- hardcode give fire bonus from wave 71 ~ wave 79
        player.attributes.bless_element = "Cold"
        player.attributes.bless_wave = 9
        player.attributes.bless_bonus = 2
    elseif wave_id >= 85 then
        local chance = wave_id >= 200 and 0.2 or 0.4
        local roll = math.random()
        if roll < chance then
            local elements = { "Fire", "Cold", "Lightning", "Poison", "Magic" }
            player.attributes.bless_element = elements[math.random(1, 5)]
            player.attributes.bless_wave = math.random(3, 10)
            player.attributes.bless_bonus = math.random(1, 2)
        end
    end

    if HasElementTrail() then
        CloseActivePrompts()
        bank.CommitBless()

        IncrementElementTrialStartCount()

        local effect_duration = 0.25
        BlessCrystalSpawnCoolEffect()
        -- CreateBlessUI()

        DCEI.TriggerAddTimerEventElapsed(function()
            CreateBless()
            ui.bless_ui:AnimatedShowUp()
            -- update bless bonus
            for _, unit in pairs(player.towers) do
                local name = DCEI.GetUnitType(unit)
                UpdateBlessBonusUnits(data.towers[name], true)
            end
            for _, unit in pairs(player.heroes) do
                local name = DCEI.GetUnitType(unit)
                UpdateBlessBonusUnits(data.heroes[name], true)
            end
            for _, unit in pairs(player.defenses) do
                local name = DCEI.GetUnitType(unit)
                UpdateBlessBonusUnits(data.defenses[name], true)
            end
            BlessUnitPromptup()
        end, effect_duration, false, true)
    end
end

function TutorialElementalTrial()
    if WaveProgressIs(50) then
        BlockInput()
        player.queued_deal_popup = false
        SetAdOnCooldown(ad_data.gold, 2)
        DCEI.TriggerAddTimerEventElapsed(function()
            if ui.bless_ui then
                BlockInputExcept(ui.bless_ui.BonusIcon)

                local options = {
                    parent = ui.bless_ui.Frame,
                    flip = true,
                    scale = 0.75,
                }
                ui.prompt.helper = ui.NewBigHand(-60, 60, options)
            end
        end, 1, true, true)
    end
end

function ConsumeBless()
    if HasElementTrail() then
        local vfx_trigger_name = "Elemental Trial VFX Trigger " .. player.attributes.bless_element
        player.attributes.bless_wave = player.attributes.bless_wave - 1
        if player.attributes.bless_wave <= 0 then
            IncrementElementTrialCompleteCount()
            player.attributes.bless_wave = 0
            player.attributes.bless_element = nil
            -- remove bless bonus
            local bonus_type = player.attributes.bless_bonus == 2 and "monster_xp" or "monster_gold"
            local bonus = 0.05
            player.attributes[bonus_type] = player.attributes[bonus_type] - player.attributes.bless_bonus_count * bonus
            player.attributes.bless_bonus_count = 0
            player.attributes.bless_bonus_units = {}
            if bless_crystal and DCEI.UnitExists(bless_crystal) then
                DCEI.RemoveUnit(bless_crystal)
                bless_crystal = nil
            end
            if ui.bless_ui then
                DCEI.AnimateFrameScale(
                    ui.bless_ui.Frame,
                    { x = 1, y = 1, z = 1 },
                    { x = 0, y = 0, z = 1 },
                    0.5,
                    "InBack"
                )
            end
            for _, unit in pairs(player.towers) do
                DCEI.RemoveBehavior(unit, vfx_trigger_name)
            end
            for _, unit in pairs(player.heroes) do
                DCEI.RemoveBehavior(unit, vfx_trigger_name)
            end
            for _, unit in pairs(player.defenses) do
                DCEI.RemoveBehavior(unit, vfx_trigger_name)
            end
        end
        DCEI.TriggerAddTimerEventElapsed(function()
            if ui.bless_ui then
                ui.bless_ui:Update()
            end
        end, 1, false, true)
        bank.CommitBless()
    end
end

function UpdateBlessBonusUnits(unit_data, equip)
    if ui.bless_ui and unit_data.element and unit_data.element == player.attributes.bless_element then
        local sign = equip and 1 or -1
        player.attributes.bless_bonus_count = math.max(0, player.attributes.bless_bonus_count + sign)
        local bonus_type = player.attributes.bless_bonus == 2 and "monster_xp" or "monster_gold"
        local bonus = 0.05
        player.attributes[bonus_type] = player.attributes[bonus_type] + sign * bonus
        ui.bless_ui:Update()
        if equip then
            player.attributes.bless_bonus_units[unit_data.name] = unit_data
            if unit_data.unit then
                local vfx_trigger_name = "Elemental Trial VFX Trigger " .. player.attributes.bless_element
                DCEI.ApplyBehaviorAsync(unit_data.unit, vfx_trigger_name)
            end
        else
            player.attributes.bless_bonus_units[unit_data.name] = nil
        end
    end
end

function BlessUnitPromptup()
    if ui.bless_ui then
        local count = 0
        local duration1 = 1
        local duration2 = 0.5
        local promptup_offset = 80
        DCEI.SetFrameActive(ui.bless_ui.TotalBonus, false)
        DCEI.TriggerAddTimerEventElapsed(function()
            DCEI.SetFrameActive(ui.bless_ui.TotalBonus, true)
            DCEI.AnimateFrameScale(ui.bless_ui.TotalBonus, { x = 2, y = 2 }, { x = 1, y = 1 }, 0.5, "InBack")
        end, duration1 + duration2, true, true)
        for _, unit_data in pairs(player.attributes.bless_bonus_units) do
            if unit_data.unit then
                DCEI.TriggerAddTimerEventElapsed(function()
                    local anchor = DCEI.CreateFrame(ui.root)
                    local icon = DCEI.CreateFrame(anchor)
                    local bonus_icon = player.attributes.bless_bonus == 2 and "talent_bonus_xp" or "talent_bonus_gold"
                    DCEI.SetFrameImage(icon, bonus_icon)
                    DCEI.SetFrameSize(icon, 64, 64)
                    local use_current_root = true
                    DCEI.AttachFrameToUnit(anchor, unit_data.unit, {}, use_current_root)
                    DCEI.AnimateFrameVerticalOffset(icon, 0, promptup_offset, duration1, "OutBack")
                    DCEI.AnimateFrameScale(icon, { x = 0.2, y = 0.2 }, { x = 1, y = 1 }, 0.5, "OutSine")
                    DCEI.TriggerAddTimerEventElapsed(function()
                        local rootpos = DCEI.GetFramePosition3D(ui.root)
                        local anchorpos = DCEI.GetFramePosition3D(anchor)
                        local targetpos = DCEI.GetFramePosition3D(ui.bless_ui.Frame)

                        DCEI.AnimateFrameHorizontalOffset(
                            icon,
                            0,
                            (targetpos.x - rootpos.x) - (anchorpos.x - rootpos.x),
                            duration2,
                            "Linear"
                        )
                        DCEI.AnimateFrameVerticalOffset(
                            icon,
                            promptup_offset,
                            (targetpos.y - rootpos.y) - (anchorpos.y - rootpos.y),
                            duration2,
                            "OutSine"
                        )
                        DCEI.TriggerAddTimerEventElapsed(function()
                            if DCEI.FrameExists(icon) then
                                DCEI.DestroyFrame(icon)
                            end
                        end, duration2 * 0.8, true, true)
                    end, duration1 * 0.8, true, true)
                end, 0.125 * count, true, true)
                count = count + 1
            end
        end
    end
end

-- !SECTION

-- SECTION: Devil's Trial
function HasDevilTrail()
    return player.devil_trail.key and player.devil_trail.wave > 0
end
bank.HasDevilTrail = HasDevilTrail

function DevilTrailActive()
    return HasDevilTrail() and player.devil_trail.accept
end

function AcceptDevilTrail()
    local cost = player.attributes.devil_trail_modifiers.cost.token
    local gems = GetPlayerGems()

    if cost <= gems then
        local spend_event = { type = "Accept", target = "Devil Trail" }
        SpendPlayerGems(cost, spend_event)
        PlayGoldSpendSound()
    else
        local function OnStore()
            ui.deveil_trail_popup:Close()
            ui.loadout.Hide()
        end

        local offset = -520
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
        return
    end

    player.devil_trail.accept = true
    player.devil_trail.wave = player.devil_trail.victory_wave
    if ui.devil_trail_ui then
        ui.devil_trail_ui:Update(player.devil_trail)
        ui.devil_trail_ui:AnimatedEmphasize()
    end
    local negative_key = player.attributes.devil_trail_modifiers.negative.key
    local positive_key = player.attributes.devil_trail_modifiers.positive
            and player.attributes.devil_trail_modifiers.positive.key
        or ""
    local event = LogEvents.DevilTrailEvent("Accept", player.devil_trail.accept_count, negative_key, positive_key)
    LogEvent(event)
    player.devil_trail.accept_count = player.devil_trail.accept_count + 1
    LogUserProperty(UserProps.devil_trial_accept)
    bank.CommitDevilTrail()
    ui.deveil_trail_popup:Close()
    if wave.data.is_active then
        ApplyDevilTrailBehaviors()
    end
end

function DeclineDevilTrail()
    player.devil_trail.accept = false
    DevilTrailEnd()
    bank.CommitDevilTrail()
end

-- get random element in an weighted list
function GetRandomValueFromWeightedList(weighted_list)
    local total_weight = 0
    for value, weight in pairs(weighted_list) do
        total_weight = total_weight + weight
    end

    local random_weight = core.RandomInteger(0, total_weight - 1)
    local target_weight = 0
    for value, weight in SortedPairs(weighted_list) do
        target_weight = target_weight + weight
        if random_weight < target_weight then
            return value
        end
    end
end

function GetDevilTrailActualModifier(ModifierData)
    local modifier = {
        key = ModifierData.key,
        parent = ModifierData.parent,
        target = ModifierData.info.target,
        behavior = ModifierData.info.behavior,
        attribute = ModifierData.info.attribute,
        description = ModifierData.info.description,
        value_percent = ModifierData.info.value_percent,
    }
    if ModifierData.info.values then
        if ModifierData.info.values.value then
            modifier.value = ModifierData.info.values.value
        elseif ModifierData.info.values.min and ModifierData.info.values.max then
            modifier.value = core.RandomInteger(ModifierData.info.values.min, ModifierData.info.values.max)
        end
    end

    if modifier.value then
        local format = "%.f"
        modifier.description = modifier.description:gsub("VAL", string.format(format, modifier.value))
    end

    modifier.description = modifier.description:gsub("%[%[", "<color=green>")
    modifier.description = modifier.description:gsub("%]%]", "</color>")
    modifier.description = modifier.description:gsub("%[", "<color=red>")
    modifier.description = modifier.description:gsub("%]", "</color>")

    if ModifierData.info.icon then
        modifier.icon = ModifierData.info.icon
    elseif
        ModifierData.parent == "element_nerf"
        or ModifierData.parent == "element_buff"
        or ModifierData.parent == "element_restore"
        or ModifierData.parent == "element_double"
    then
        modifier.element = ModifierData.info.element
        modifier.icon = ui.element_icon[modifier.element]
    end

    return modifier
end

function GetModifierDataAccordingToKey(ModifierData, targetkey)
    local resultrecorder = nil
    local function GetModifierHelper(ModifierData)
        if resultrecorder then
            return
        end
        if ModifierData.key == targetkey then
            resultrecorder = ModifierData
        end
        if ModifierData.sub_modifier_list then
            for key, value in pairs(ModifierData.sub_modifier_list) do
                GetModifierHelper(value)
            end
        end
    end

    GetModifierHelper(ModifierData)
    return resultrecorder
end

--  Random pick [data] used to create modfiier according to its' weight value
function GetRandomModifier(ModifierData)
    if ModifierData.sub_modifier_list == nil then
        return ModifierData
    else
        local weightlist = {}
        for key, value in pairs(ModifierData.sub_modifier_list) do
            if value.start_wave and wave.data.GetMax() < value.start_wave then
                weightlist[key] = 0
            else
                weightlist[key] = value.weight
                if TestB("test010715_new_devil_trail") then
                    if
                        key == "castle_nerf_hp"
                        or key == "hero_nerf_cooldown"
                        or key == "battle_nerf_monster_speedup"
                    then
                        weightlist[key] = 0
                    end
                else
                    if key == "hero_nerf_mana_cost" or key == "tower_nerf" or key == "all_nerf" then
                        weightlist[key] = 0
                    end
                end
            end
            -- DCEI.LogMessage(key..": "..weightlist[key])
        end
        local randomKey = GetRandomValueFromWeightedList(weightlist)
        return GetRandomModifier(ModifierData.sub_modifier_list[randomKey])
    end
end

function GetDevilTrailModifier(key, seed)
    local modifiers = {}
    core.SetRandomSeed(seed)
    local negative_modifier = GetModifierDataAccordingToKey(data.devil_trail.negative_effects, key)
    DCEI.LogMessage("Devil Trail Negative ModifierData:  " .. negative_modifier.key)
    modifiers.negative = GetDevilTrailActualModifier(negative_modifier)

    local positive_target = negative_modifier.info.positive_target
    local temp = nil

    if positive_target then
        temp = GetModifierDataAccordingToKey(data.devil_trail.positive_effects, positive_target)
    else
        temp = data.devil_trail.positive_effects
    end

    if not negative_modifier.info.single then
        local positive_modifier = GetRandomModifier(temp)
        if negative_modifier.info.element then
            while positive_modifier.info.element == negative_modifier.info.element do
                positive_modifier = GetRandomModifier(temp)
            end
        end
        modifiers.positive = GetDevilTrailActualModifier(positive_modifier)
        DCEI.LogMessage("Devil Trail Positive ModifierData:  " .. positive_modifier.key)
    end
    modifiers.reward = GetRandomValueFromWeightedList(data.devil_trail.rewards.weight_list)
    local reward_item = data.devil_trail.rewards[modifiers.reward]
    modifiers.reward_value = core.RandomInteger(reward_item.values.min, reward_item.values.max)
    modifiers.reward_icon = reward_item.icon
    modifiers.cost = reward_item.cost

    player.attributes.devil_trail_modifiers = modifiers
end

function RandomChooseModifier()
    local negative_modifier = GetRandomModifier(data.devil_trail.negative_effects)
    player.devil_trail.key = negative_modifier.key
    player.devil_trail.wave = 3
    player.devil_trail.victory_wave = math.random(8, 14)
    player.devil_trail.seed = os.time()
end

function ApplyDevilTrailModifier(modifier, is_addition)
    local behavior_name = modifier.behavior
    local attribute = modifier.attribute
    local target = modifier.target
    local sign = is_addition and 1 or -1
    local value = 1
    if modifier.value then
        if modifier.value_percent then
            value = modifier.value * 0.01
        else
            value = modifier.value
        end
    end
    value = value * sign

    if attribute then
        -- DCEI.LogError("Devil Trail apply " .. value .. " " .. attribute .. " to " .. target)
        player.attributes[attribute] = player.attributes[attribute] + value
    end

    if behavior_name and target ~= "monster" then
        -- DCEI.LogError("Devil Trail apply " .. value .. " " .. behavior_name .. " to " .. target)
        AddBehaviorUpgrade(target, behavior_name, value)
    end
end

function ApplyDevilTrailBehaviors()
    local is_addition = true
    local negative_modifier = player.attributes.devil_trail_modifiers.negative
    ApplyDevilTrailModifier(negative_modifier, is_addition)

    local positive_modifier = player.attributes.devil_trail_modifiers.positive
    if positive_modifier then
        ApplyDevilTrailModifier(positive_modifier, is_addition)
    end
end

function RemoveDevilTrailBehaviors()
    local negative_modifier = player.attributes.devil_trail_modifiers.negative
    ApplyDevilTrailModifier(negative_modifier)

    local positive_modifier = player.attributes.devil_trail_modifiers.positive
    if positive_modifier then
        ApplyDevilTrailModifier(positive_modifier)
    end
end

function RollDevilTrade(wave_id)
    if HasDevilTrail() or HasElementTrail() or wave_id < 120 or event_mode.IsActive() then
        return
    end
    -- hard code first devil trail
    if not player.system.first_devil_trail or player.system.first_devil_trail == system.locked then
        player.devil_trail.key = "archer_nerf"
        player.devil_trail.wave = 3
        player.devil_trail.victory_wave = 8
        player.devil_trail.seed = 946254
        player.system.first_devil_trail = system.active
        bank.CommitSystemState()
    else
        if DEBUG.select_devil_modifier then
            player.devil_trail.key = "element_nerf_ice"
            player.devil_trail.wave = 3 
            player.devil_trail.victory_wave = 4
            player.devil_trail.seed = 54354
        else
            local chance = wave_id >= 250 and 0.1 or 0.18
            local roll = math.random()
            if roll < chance then
                RandomChooseModifier()
            end
        end
    end
    if HasDevilTrail() then
        CloseActivePrompts()
        CreateDevilTrail()
        ui.devil_trail_ui:AnimatedShowUp()
        local negative_key = player.attributes.devil_trail_modifiers.negative.key
        local positive_key = player.attributes.devil_trail_modifiers.positive
                and player.attributes.devil_trail_modifiers.positive.key
            or ""
        local event = LogEvents.DevilTrailEvent("Trigger", player.devil_trail.trigger_count, negative_key, positive_key)
        LogEvent(event)
        player.devil_trail.trigger_count = player.devil_trail.trigger_count + 1
        LogUserProperty(UserProps.devil_trial_trigger)
        bank.CommitDevilTrail()
    end
end

function CreateDevilTrail()
    GetDevilTrailModifier(player.devil_trail.key, player.devil_trail.seed)
    -- create devil trail UI
    if not ui.devil_trail_ui then
        ui.devil_trail_ui = ui.CreateDevilTrailUI()
        DCEI.SetOnClickCallback(ui.devil_trail_ui.Button, CreateDevilUIPopup)
    end
    ui.devil_trail_ui:Update(player.devil_trail)
end

function CreateDevilUIPopup()
    if HasDevilTrail() then
        local layout = gmui.Layout:New({
            parent = ui.upper,
            name = "DevilTrail/Popup",
        })
        layout.Condition = gmui.Layout:New({
            frame = layout.Condition,
        })
        layout.Burden = gmui.Layout:New({
            frame = layout.Burden,
        })
        layout.Prize = gmui.Layout:New({
            frame = layout.Prize,
        })
        layout.Prize.BoostItem = gmui.Layout:New({
            frame = layout.Prize.BoostItem,
        })
        if wave.data.is_active then
            SetGameSpeed(0)
        end
        local wave_remaining = player.devil_trail.wave
        local fail_wave = 3 - player.devil_trail.fail
        if player.devil_trail.accept then
            DCEI.SetFrameActive(layout.DecideButtons, false)
        else
            wave_remaining = player.devil_trail.victory_wave
        end

        local win_wave_text = '<align="left">'
            .. DCEI.GetTranslationText(
                "ui/menu/devil_trade/wave",
                { x = "<color=green><size=150%>" .. wave_remaining .. "</size></color>" }
            )
        DCEI.SetTextFrameText(layout.Condition.WinWave, win_wave_text)
        local wave_text = fail_wave > 1 and DCEI.GetTranslationText("ui/menu/devil_trade/wave/multiple")
            or DCEI.GetTranslationText("ui/menu/devil_trade/wave/single")
        local fail_wave_text = '<align="left">'
            .. DCEI.GetTranslationText(
                "ui/menu/devil_trade/fail_confition",
                { x = "<color=red><size=150%>" .. fail_wave .. "</size></color>", y = wave_text }
            )
        DCEI.SetTextFrameText(layout.Condition.FailWave, fail_wave_text)

        local negative_icon = player.attributes.devil_trail_modifiers.negative.icon
        DCEI.SetFrameImage(layout.Burden.NegativeIcon, negative_icon)
        local negative_description = player.attributes.devil_trail_modifiers.negative.description
        DCEI.SetTextFrameText(layout.Burden.NegativeDescription, negative_description)
        if player.attributes.devil_trail_modifiers.positive then
            DCEI.SetFrameActive(layout.Burden.PositiveEffect, true)
            local positive_icon = player.attributes.devil_trail_modifiers.positive.icon
            DCEI.SetFrameImage(layout.Burden.PositiveIcon, positive_icon)
            local positive_description = player.attributes.devil_trail_modifiers.positive.description
            DCEI.SetTextFrameText(layout.Burden.PositiveDescription, positive_description)
        else
            DCEI.SetFrameActive(layout.Burden.PositiveEffect, false)
        end

        local reward_name = player.attributes.devil_trail_modifiers.reward
        local reward_value = player.attributes.devil_trail_modifiers.reward_value
        local reward_icon = player.attributes.devil_trail_modifiers.reward_icon
        local cost = player.attributes.devil_trail_modifiers.cost.token
        if reward_name == "gold_boost_200" then
            DCEI.SetFrameActive(layout.Prize.BoostItem.Frame, true)
            DCEI.SetFrameActive(layout.Prize.IconItem, false)
            DCEI.SetFrameImage(layout.Prize.BoostItem.BoostIcon, reward_icon)
            DCEI.SetTextFrameText(layout.Prize.BoostItem.BoostLabel, "2<size=75%>x")
        elseif reward_name == "token_boost_500" then
            DCEI.SetFrameActive(layout.Prize.BoostItem.Frame, true)
            DCEI.SetFrameActive(layout.Prize.IconItem, false)
            DCEI.SetFrameImage(layout.Prize.BoostItem.BoostIcon, reward_icon)
            DCEI.SetTextFrameText(layout.Prize.BoostItem.BoostLabel, "5<size=75%>x")
        else
            DCEI.SetFrameActive(layout.Prize.BoostItem.Frame, false)
            DCEI.SetFrameActive(layout.Prize.IconItem, true)
            DCEI.SetFrameImage(layout.Prize.IconItem, reward_icon)
        end
        if reward_name == "raid_turbo" then
            DCEI.SetFrameSize(layout.Prize.IconItem, 80, 80)
        elseif reward_name == "auto_battle" then
            DCEI.SetFrameSize(layout.Prize.IconItem, 120, 120)
        else
            DCEI.SetFrameSize(layout.Prize.IconItem, 100, 100)
        end
        local reward_text = "  x" .. reward_value
        if reward_name == "gold_boost_200" or reward_name == "token_boost_500" or reward_name == "auto_battle" then
            reward_text = reward_value .. " " .. DCEI.GetTranslationText("general/min")
        end
        if reward_name == "token" then
            reward_text = "x" .. reward_value
            DCEI.SetFrameActive(layout.Prize.RewardIcon, true)
            DCEI.SetTextFrameColor(layout.Prize.RewardLabel, { r = 0.945, g = 0.478, b = 0.929 })
        else
            DCEI.SetFrameActive(layout.Prize.RewardIcon, false)
            DCEI.SetTextFrameColor(layout.Prize.RewardLabel, { r = 1, g = 0.92, b = 0.016 })
        end
        DCEI.SetTextFrameText(layout.Prize.RewardLabel, reward_text)
        -- cost
        DCEI.SetTextFrameText(layout.CostLabel, cost)
        function layout:Close()
            if wave.data.is_active then
                SetGameSpeed(player.speed)
            end
            if ui.deveil_trail_popup and DCEI.FrameExists(ui.deveil_trail_popup.Frame) then
                DCEI.DestroyFrame(ui.deveil_trail_popup.Frame)
                ui.deveil_trail_popup = nil
            end
        end

        DCEI.SetOnClickCallback(layout.CloseButton, function()
            layout:Close()
        end)
        DCEI.SetOnClickCallback(layout.DeclineButton, function()
            layout:Close()
            DeclineDevilTrail()
        end)
        DCEI.SetOnClickCallback(layout.AcceptButton, function()
            AcceptDevilTrail()
        end)
        ui.deveil_trail_popup = layout
    end
end

-- decline or success or fail
function DevilTrailEnd(success)
    -- save reward data
    local reward_name = player.attributes.devil_trail_modifiers.reward
    local reward_value = player.attributes.devil_trail_modifiers.reward_value
    local reward = {}
    reward[reward_name] = reward_value

    -- reset
    player.devil_trail.key = nil
    player.devil_trail.wave = 0
    player.devil_trail.victory_wave = 0
    player.devil_trail.fail = 0
    player.devil_trail.accept = false
    player.attributes.devil_trail_modifiers = nil
    if ui.devil_trail_ui then
        if success then
            ui.devil_trail_ui.ending = true
            DCEI.TriggerAddTimerEventElapsed(function()
                DCEI.AnimateFrameScale(
                    ui.devil_trail_ui.Frame,
                    { x = 1, y = 1, z = 1 },
                    { x = 1.5, y = 1.5, z = 1 },
                    1.5,
                    "InBack"
                )
                DCEI.AnimateFrameAlpha(ui.devil_trail_ui.Frame, 1, 0, 1.5, "InSine")
                ClaimDevilTrailReward(reward)
                DCEI.TriggerAddTimerEventElapsed(function()
                    ui.devil_trail_ui.ending = false
                    if ui.devil_trail_ui then
                        ui.devil_trail_ui:Update()
                    end
                end, 1.5, true, true)
            end, 1.5, true, true)
        else
            DCEI.AnimateFrameScale(
                ui.devil_trail_ui.Frame,
                { x = 1, y = 1, z = 1 },
                { x = 0, y = 0, z = 1 },
                1.5,
                "InBack"
            )
            DCEI.TriggerAddTimerEventElapsed(function()
                if ui.devil_trail_ui then
                    ui.devil_trail_ui:Update()
                end
            end, 1.5, true, true)
        end
    end
end

function ClaimDevilTrailReward(reward)
    -- use the same claim function for now
    ClaimSeasonEventReward(reward)
end

function ConsumeDevilTrail(is_victory)
    if HasDevilTrail() then
        if is_victory or not player.devil_trail.accept then
            player.devil_trail.wave = player.devil_trail.wave - 1
            if player.devil_trail.wave <= 0 then
                local success = is_victory and player.devil_trail.accept
                if success then
                    local negative_key = player.attributes.devil_trail_modifiers.negative.key
                    local positive_key = player.attributes.devil_trail_modifiers.positive
                            and player.attributes.devil_trail_modifiers.positive.key
                        or ""
                    local event =
                        LogEvents.DevilTrailEvent("Win", player.devil_trail.success_count, negative_key, positive_key)
                    LogEvent(event)
                    player.devil_trail.success_count = player.devil_trail.success_count + 1
                    LogUserProperty(UserProps.devil_trial_win)
                end
                DevilTrailEnd(success)
            end
        else
            player.devil_trail.fail = player.devil_trail.fail + 1
            if player.devil_trail.fail >= 3 then
                local negative_key = player.attributes.devil_trail_modifiers.negative.key
                local positive_key = player.attributes.devil_trail_modifiers.positive
                        and player.attributes.devil_trail_modifiers.positive.key
                    or ""
                player.devil_trail.lost_count = player.devil_trail.lost_count or 0
                local event =
                    LogEvents.DevilTrailEvent("Lose", player.devil_trail.lost_count, negative_key, positive_key)
                LogEvent(event)
                player.devil_trail.lost_count = player.devil_trail.lost_count + 1
                LogUserProperty(UserProps.devil_trial_lose)
                DevilTrailEnd()
            end
        end
        DCEI.TriggerAddTimerEventElapsed(function()
            if ui.devil_trail_ui then
                ui.devil_trail_ui:Update(player.devil_trail)
                ui.devil_trail_ui:AnimatedEmphasize()
            end
        end, 1, true, true)
        bank.CommitDevilTrail()
    end
end

-- !SECTION

-- SECTION: Wave End
function WaveEndActions(is_victory, is_defeat)
    local delay = (is_victory and 1.5) or (is_defeat and 1.25) or 0
    local result_message_options = {}

    if wave.data.skip then
        local start_wave = wave.data.start_wave or 0
        local skip_number = wave.data.skip_number or 0
        if is_victory then
            IncrementWaveSkipSuccess()
            local prop = UserProps.wave_skip_last(start_wave, skip_number, "Win")
            LogUserProperty(prop)
        end
        if is_defeat then
            IncrementWaveSkipFailure()
            local prop = UserProps.wave_skip_last(start_wave, skip_number, "Lose")
            LogUserProperty(prop)
        end
        wave.data.start_wave = nil
        wave.data.skip_number = nil
    end

    ui.loadout.hero.is_active = false
    wave.data.is_active = false
    wave.data.instance = wave.data.instance + 1

    -- calculate monsters kill
    if wave.data.ModeStandard() and not event_mode.IsActive() then
        SettleMonsterKills()
    end

    -- result message options
    if wave.data.mode == "challenge" or wave.data.mode == "colony" or wave.data.mode == "raid_boss" then
        result_message_options.mask_alpha = 0.6
        result_message_options.skip_mask_transition_out = true
    end

    -- interrupt auto battle for the first time bless introduces
    if wave.data.GetCurrent() == 50 then
        DisableAutoBattle()
    end

    if is_victory and player.autobattle.is_active and wave.data.mode == "standard" and not event_mode.IsActive() then
        local autobattle_delay = 0.25
        result_message_options.skip_mask_transition_out = true

        -- keeping showing battle UI if auto-battle prompt is queued
        DCEI.TriggerAddTimerEventElapsed(CreateAutoBattlePrompt, delay + autobattle_delay, false, true)
    else
        -- disable battle buttons during end wave message
        ui.bottom_panel:SetNavButtonsEnabled(false)

        DisableAutoBattle()

        local function ExitWave()
            DwarfWorkers_UseUnits()
            RestoreBuildingsOnWaveEnd()
            AttemptHandleElixirIncome()
            AttemptHandleVaultIncome()
            ui.bottom_panel:SetNavButtonsEnabled(true)

            if wave.data.ModeStandard() then
                CreateCastleUI()
                CreateWaveDropResourceFlyAnimation()
            elseif wave.data.mode == "colony" then
                ExitColonyBattle(is_victory)
            elseif wave.data.mode == "challenge" then
                ExitChallengeBattle(is_victory)
            elseif wave.data.mode == "raid_boss" then
                ExitRaidBossBattle()
            elseif wave.data.mode == "hell_mode" then
                ExitHellModeBattle()
                CreateCastleUI()
            elseif wave.data.mode == "training" then
                CreateCastleUI()
            end

            SetGameState(ui.bottom_panel.state)
            -- check for tutorial sequences
            if is_victory then
                if not event_mode.IsActive() then
                    Tutorial02_Archer_Check()
                    Tutorial03_Castle_Check()
                    Tutorial04_1_Talent_Check()
                    Tutorial05_Mine_Check()
                    Tutorial06_Promote_Check()
                    TutorialNewGearOnboarding_Check()
                    Tutorial07_Colony_Check()
                    Tutorial08_TownUpgrade_PromptCheck()
                    TutorialElementalTrial()
                end

                ReviewInviteCheck()

                if ui.delay_display_leaderboard_update then
                    ui.delay_display_leaderboard_update = false
                    ui.ShowLeaderBoardUpdate()
                end
            elseif is_defeat then
                if not event_mode.IsActive() then
                    Tutorial04_Boss_Check()
                end
            end

            if wave.data.mode == "standard" and (is_victory or is_defeat) then
                CheckForQueuedDealPopup()
                AttemptGoldAd(is_defeat)
            end
            CheckRaceEventUnlock()
            DCEI.TriggerAddTimerEventElapsed(ResetCastle, 0, true, true)
        end

        if wave.data.mode == "raid_boss" and is_victory then
            delay = 2.5
        end
        local unlock_auto = 25
        -- delay is 0 if wave ends via "back" button
        if delay == 0 and wave.data.mode ~= "training" then
            wave.data.unfinished = true
            ExitWave()
        elseif wave.data.mode == "challenge" and is_victory then
            local challenge_reward_delay = 0.5

            -- show challenge reward after brief delay to better match victory animation
            DCEI.TriggerAddTimerEventElapsed(function()
                local on_close = ExitWave
                ShowChallengeRewardFrame(on_close)
            end, challenge_reward_delay, false, true)
        elseif wave.data.mode == "standard" and (wave.data.GetCurrent() > unlock_auto) and is_defeat then
            DCEI.TriggerAddTimerEventElapsed(function()
                ui.screen_mask.lower:Show()
                ui.battlestats.on_hide = ExitWave
                ui.battlestats:Show()
            end, delay, false, true)
        else
            ui.battlestats.on_hide = nil
            DCEI.TriggerAddTimerEventElapsed(function()
                if ui.battlestats.is_active then
                    ui.battlestats.on_hide = ExitWave
                else
                    ExitWave()
                end
            end, delay, false, true)
        end
    end

    ui.DestroyBossBar()
    ui.DestroyWaveBar()

    CheckExitTargetMode()
    ClearGameSpeed()

    StopBoostTimer()

    if is_victory and wave.data.ModeStandard() then
        ConsumeBless()
    end

    if wave.data.mode == "standard" and DevilTrailActive() then
        RemoveDevilTrailBehaviors()
    end

    if wave.data.mode == "standard" then
        if is_victory then
            ConsumeDevilTrail(is_victory)
        end
        if is_defeat then
            ConsumeDevilTrail()
        end
    end
    
    bank.CommitGems()
    bank.CommitGold()
    bank.CommitVaultGold()
    bank.CommitElixir()
    bank.CommitWinterCoin2022()
    bank.CommitAutumnCoin()
    if event_mode.IsActive() then
        bank.CommitEventMineContract()
        bank.CommitEventMinetime()
        bank.CommitEventLabtime()
        bank.CommitEventPlayerXP()
        bank.CommitEventParagonXP()
    else
        bank.CommitMineContract()
        bank.CommitPlayerXP()
        bank.CommitParagonXP()
    end
    
    bank.CommitEventEggHunt()

    if wave.data.rare_monster_drop then
        bank.CommitChest(wave.data.rare_monster_drop.drops[1])
    end

    -- this timer delay is necessary in case units are summoned in the same frame as wave end
    local simple = true
    DCEI.TriggerAddTimerEventElapsed(RemoveSummonedUnits, 0.0625, false, simple)
    ResetUnits()

    if is_victory then
        local duration = (wave.data.mode == "challenge" and 4.4375) or 1.5
        if wave.data.mode == "raid_boss" then
            duration = 3
        end
        local message_delay = 0.25
        local reward_data = {
            -- mine contracts are rolled for standard boss waves (40% chance for x1 per boss)
            mine_contract = wave.data.drop_contract and 1 or 0,
            -- town contracts drop from colony battles (x10 per colony)
            town_contract = (wave.data.mode == "colony") and 10 or 0,
            -- medals/tokens drop from each standard wave (x1 per wave)
            medal = wave.data.ModeStandard() and 1 or 0,
            -- bonus gold is calculated based off bonus gold talents
            bonus_gold = GetWaveGoldBonus(),
            -- equipment add victory xp
            bonus_xp = GetWaveXPBonus(),
            -- chests from elite monster or raid boss
            -- { "chest name" = count, ... }
            chests = wave.data.drop_chests,
            drop_material = wave.data.drop_material,
            epic_material = 0,
            normal_material = 0,
            raid_turbo = 0,
            event_chests = 0
        }
        if event_mode.IsActive() then
            reward_data.event_chests = wave.data.ModeStandard() and 1 or 0
            reward_data.medal = 0
        else
            reward_data.potion = 0
        end

        ui.loadout.hero.Hide()

        if wave.data.mode == "standard" and (wave.data.GetCurrent() > wave.data.GetMax()) then
            -- new wave victory actions
            wave.data.SetMax(wave.data.GetCurrent())
            if event_mode.IsActive() then
                bank.CommitEventMaxWave()
            else
                bank.CommitMaxWave()
            end

            local show_update = true
            local score = WaveScoreWithTimeBonus()
            PostLeaderBoardScore(score, show_update)
            AttemptFetchRaceEventData()

            CheckCampaignEvent_Wave30()
            CheckCampaignEvent_Wave20()
            CheckChanllengeShowPip()

            local current_wave = wave.data.GetCurrent()
            DCEI.Event.LogLevelComplete("level " .. tostring(current_wave))

            -- roll element bless after boss wave
            if wave.data.GetCurrent() % 5 == 0 then
                -- wait until screen mask disappear
                DCEI.TriggerAddTimerEventElapsed(function()
                    RollElementBless(wave.data.GetCurrent())
                    RollDevilTrade(wave.data.GetCurrent())
                end, duration + 0.25, false, true)
            end
            -- wait until screen mask disappear
            DCEI.TriggerAddTimerEventElapsed(function()
                CheckStartMonsterBounty(wave.data.GetCurrent())
            end, duration + 0.25, false, true)

            if not GetPlayerTest("pause_offline_gold") and player.system.vault == system.active then
                data.vault:UpdateVaultCap(wave.data.GetCurrent())
                -- re schedule notification as capacity is changed
                ScheduleVaultNotification()
            end

            for k, v in pairs(player.wave_win_increment) do
                if player.attributes[k] >= v.max then
                    if k == "token_tower_progress" then
                        reward_data.medal = reward_data.medal + 1
                    elseif k == "material_tower_progress" then
                        reward_data.normal_material = reward_data.normal_material + 1
                    end
                    player.attributes[k] = 1
                else
                    local increment = math.random() < 0.6 and 2 or 1
                    player.attributes[k] = math.min(player.attributes[k] + increment, v.max)
                end

                local percentage = player.attributes[k] / v.max
                ui.loadout.tower.UpdateProgress(v.slot, percentage, true)
                bank.CommitTowerProgress(k)
            end
        end

        if
            reward_data.drop_material
            and reward_data.drop_material["epic_material"]
            and reward_data.drop_material["epic_material"] > 0
        then
            reward_data.epic_material = reward_data.epic_material + reward_data.drop_material["epic_material"]
        end

        -- first time raid boss give 5 raid turbo
        if wave.data.mode == "raid_boss" and is_victory then
            SpendPlayerRaidKey(-1)
            local boss_index = levels.raid_boss.current.index
            if boss_index == 1 then
                local key = "Raid Boss " .. boss_index
                local victory_count = player.raid_boss_log[key] or 0
                if victory_count == 0 then
                    reward_data.raid_turbo = 5
                    player.system.popup_raid_turbo_boost = system.unlocked
                    bank.CommitSystemState()
                end
            end
        end

        if
            is_victory
            and not player.autobattle.is_active
            and wave.data.mode == "standard"
            and wave.data.GetCurrent() > 26
        then
            DCEI.TriggerAddTimerEventElapsed(function()
                CreateVictoryScreenBattleStatsButton(duration)
            end, message_delay, false, true)
        end

        DCEI.TriggerAddTimerEventElapsed(function()
            ui.CreateVictoryMessage(duration, reward_data, result_message_options)
        end, message_delay, false, true)

        if reward_data.medal > 0 then
            AddPlayerGems(reward_data.medal)
        end
        if reward_data.town_contract > 0 then
            -- skip commit here and commit town contract together with colony state, to avoid game exit or crash during exit colony battle
            local skip_commit = true
            AddPlayerContract_Town(reward_data.town_contract, skip_commit)
        end
        if reward_data.mine_contract > 0 then
            AddPlayerContract_Mine(reward_data.mine_contract)
        end
        if reward_data.bonus_gold > 0 then
            AddPlayerGold(reward_data.bonus_gold)
            bank.CommitGold()
        end

        if reward_data.bonus_xp > 0 then
            AddPlayerXP(reward_data.bonus_xp)
        end

        if reward_data.chests then
            for name, count in pairs(reward_data.chests) do
                AddPlayerChest(name, count)
            end
        end

        if reward_data.event_chests > 0  then
            event_mode.AddChests(1)
        end

        if reward_data.normal_material > 0 then
            AddPlayerMaterial("normal_material", reward_data.normal_material)
        end

        if reward_data.epic_material > 0 then
            AddPlayerMaterial("epic_material", reward_data.epic_material)
        end

        if reward_data.raid_turbo > 0 then
            AddPlayerRaidTurbo(reward_data.raid_turbo)
        end

        -- check to queue deals popup on lose>win between wave 30-100, check to queue AB bundle once after wave 25, token bundle on wave 100
        CheckToQueueDealsPopup()

        if wave.data.mode ~= "replay" then
            LogEvent(LogEvents.wave_win)
        end
    elseif is_defeat then
        local duration = 1.25

        if wave.data.mode == "hell_mode" then
            result_message_options.hell_mode = true
            ui.CreateDefeatMessage(duration, result_message_options)
            DCEI.SetTextFrameText(ui.prompt.wave_result.Score, GetHellModeScore())
            CheckHellModeSeasonScore()
        else
            AttemptFetchRaceEventData()
            ui.CreateDefeatMessage(duration, result_message_options)
            LogEvent(LogEvents.wave_lose)
        end
        ui.loadout.hero.Hide()
    else
        -- log wave exit via "back" button
        LogEvent(LogEvents.wave_exit)
    end

    -- attempt to update wave attempt count
    if wave.data.mode == "standard" then
        UpdateWaveAttempts(is_victory, is_defeat)
    elseif wave.data.mode == "colony" and is_defeat then
        IncrementColonyLossCount()
    elseif wave.data.mode == "raid_boss" then
        UpdateRaidBossLog(is_victory, is_defeat)
    end

    -- run GC call during victory or defeat message in standard/reply mode, timed so it occurs between transition animations
    if (is_victory or is_defeat) and (wave.data.ModeStandard() or wave.data.mode == "hell_mode") then
        local simple = true
        DCEI.TriggerAddTimerEventElapsed(GarbageCollect, 0.75, false, simple)
    end

    -- reset castle needs to happen after GetWaveBonusGold
    --ResetCastle() move to ExitWave()
    wave.data:Reset()

    if TestB("autumn_event") then
        CheckAutumnEventUnlock(false, true)
    end
    if TestB("winter_event2022") then
        CheckWinterEventUnlock(false, true)
    end
    CheckEggHuntUnlock()
end

function GarbageCollect()
    -- DCEI.LogMessage("GC before: " .. collectgarbage("count"))
    DCEI.GarbageCollect()
    -- DCEI.LogMessage("GC after: " .. collectgarbage("count"))
end

function CreateVictoryScreenBattleStatsButton(duration)
    if ui.victory_screen_battle_stats_button then
        DCEI.DestroyFrame(ui.victory_screen_battle_stats_button.Frame)
    end

    local stats_button = gmui.Layout:New({
        parent = ui.super,
        name = "BattleStats/VictoryScreenButton",
    })
    local v_offset = -400
    local h_offset = 220
    DCEI.SetFrameHorizontalOffsetInParent(stats_button.Frame, h_offset)
    DCEI.SetFrameVerticalOffsetInParent(stats_button.Frame, v_offset)

    DCEI.SetOnClickCallback(stats_button.Button, function()
        DCEI.SetFrameActive(stats_button.Frame, false)
        ui.screen_mask.lower:Show()
        ui.battlestats:Show()
        ui.GlobalFadeIn(0, 0, 0)
        if ui.prompt.wave_result then
            ui.prompt.wave_result.transition_out = true
            DCEI.SetFrameActive(ui.prompt.wave_result.frame, false)
        end
    end)

    stats_button.alpha_animation = DCEI.AnimateFrameAlpha(stats_button.BackGlow, 1, 0, 1, "InSine")
    DCEI.SetFrameAnimationLoops(stats_button.alpha_animation, -1, "Restart")
    stats_button.scale_animation = DCEI.AnimateFrameScale(
        stats_button.BackGlow,
        { x = 1, y = 1 },
        { x = 1.3, y = 1.3 },
        1,
        "OutCubic"
    )
    DCEI.SetFrameAnimationLoops(stats_button.scale_animation, -1, "Restart")

    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.SetButtonFrameEnable(stats_button.Button, false)
        DCEI.StopFrameAnimationLoops(stats_button.alpha_animation, false)
        DCEI.StopFrameAnimationLoops(stats_button.scale_animation, false)
        DCEI.SetFrameActive(stats_button.BackGlow, false)
        local duration = 0.5
        DCEI.AnimateFrameScale(stats_button.Frame, { x = 1, y = 1 }, { x = 0.3, y = 0.3 }, duration * 2, "OutCubic")
        local rootpos = DCEI.GetFramePosition3D(ui.root)
        local target_pos = DCEI.GetFramePosition3D(ui.glue.other_buttons.frame)
        DCEI.AnimateFrameVerticalOffset(
            stats_button.Frame,
            v_offset,
            target_pos.y - rootpos.y + 20,
            duration,
            "OutSine"
        )
        DCEI.AnimateFrameHorizontalOffset(stats_button.Frame, h_offset, target_pos.x - rootpos.x, duration, "OutSine")
        DCEI.TriggerAddTimerEventElapsed(function()
            DCEI.DestroyFrame(stats_button.Frame)
        end, duration, false, true)
    end, duration, false, true)

    ui.victory_screen_battle_stats_button = stats_button
end

function GetWaveGoldBonus()
    local base_value = wave.data.gold or 0
    local health_max = DCEI.GetMaxHealth(CASTLE_UNIT)
    local health = DCEI.GetHealth(CASTLE_UNIT)
    local percent = health / health_max

    local equipment_vicoty_gold = 0
    for _, unit in pairs(player.heroes) do
        if unit and DCEI.UnitExists(unit) then
            local unit_type = DCEI.GetUnitType(unit)
            local unit_data = GetHeroDataFromUnitType(unit_type)
            local vicoty_gold = unit_data.equipment.equipment_vitory_gold or 0
            equipment_vicoty_gold = equipment_vicoty_gold + vicoty_gold
        end
    end

    local bonus90 = (percent >= 0.90) and player.attributes.health_bonus_90 or 0
    local bonus100 = (percent >= 0.9999) and player.attributes.health_bonus_100 or 0
    local bonus = equipment_vicoty_gold + bonus90 + bonus100

    return math.floor(base_value * bonus)
end

function GetWaveXPBonus()
    local base_xp = wave.data.xp or 0

    local equipment_vicoty_xp = 0
    for _, unit in pairs(player.heroes) do
        if unit and DCEI.UnitExists(unit) then
            local unit_type = DCEI.GetUnitType(unit)
            local unit_data = GetHeroDataFromUnitType(unit_type)
            local vicoty_xp = unit_data.equipment.equipment_vitory_xp or 0
            equipment_vicoty_xp = equipment_vicoty_xp + vicoty_xp
        end
    end

    return math.floor(base_xp * equipment_vicoty_xp)
end

-- campaign logging
function CheckCampaignEvent_Wave20()
    if WaveProgressIs(20) then
        DCEI.Event.LogView("Beat Level 20", "content_id")
    end
end

function CheckCampaignEvent_Wave30()
    if WaveProgressIs(30) then
        DCEI.Event.LogAchievement("Beat Level 30")
    end
end

function CheckCampaignEvent_GoldAdx3()
    if player.ads_watched_gold == 3 then
        DCEI.Event.LogTutorialComplete("Watch 3 Ad")
    end
end

function CheckCampaignEvent_GoldAdx5()
    if player.ads_watched_gold == 5 then
        DCEI.Event.LogRegistrationComplete("Watch 5 Ad")
    end
end

function ResetUnits()
    -- reset archers and heroes
    for _, unit in pairs(group_archers) do
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat"))
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat PeriodicRetarget"))

        for _, behavior in ipairs(data.combat_archer_behaviors) do
            DCEI.RemoveBehaviorAsync(unit, behavior)
        end
    end
    for _, unit in pairs(player.towers) do
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat PeriodicRetarget"))
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("Tower Barracks Summon Cooldown"))
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("Tower Monolith Summon Cooldown"))
    end
    for _, unit in pairs(player.defenses) do
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat PeriodicRetarget"))
    end
    for _, unit in pairs(player.heroes) do
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat"))
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat PeriodicRetarget"))
        ShowUnitSkillBar(unit, false)
        ResetHeroSkill(unit)
        RemoveHeroInCombatBehavior(unit)
    end

    -- remove enemy units
    local units = {}
    for _, unit in pairs(wave.data.units) do
        table.insert(units, unit)
    end
    -- RemoveUnitSync will trigger TriggerAddUnitRemovedEvent and edit wave.data.units table, so use another table to make remove safe
    for _, unit in pairs(units) do
        DCEI.RemoveUnit(unit)
    end
    wave.data.units = {}
    GEAR_UNITY_HELPER_BUFF_INSTANCE = 0
    ApplyGearUnityArcherCriticalDamage(-10)
end

function RemoveSummonedUnits()
    -- remove summoned units
    for k, unit in pairs(group_summons) do
        DCEI.RemoveUnit(unit)
    end
end

function ResetCastle()
    -- may want to use behaviors that regen over time, since damage may occur after this frame
    local full_health = DCEI.GetMaxHealth(CASTLE_UNIT)
    local full_mana = DCEI.GetMaxMana(CASTLE_UNIT)
    DCEI.AddHealth(CASTLE_UNIT, full_health)
    DCEI.AddMana(CASTLE_UNIT, full_mana)
end

-- !SECTION

-- SECTION AUTO BATTLE
function EnableAutoBattle(type)
    player.autobattle.is_active = true
    player.autobattle.type = type or "none"

    if not event_mode.IsActive() then
        ShowAutoBattleButton()
    end

    local simple = true
    DCEI.TriggerAddTimerEventElapsed(AutoBattleTick, 1, false, simple)
end

function GenerateAutoBattlePriorityList()
    player.autobattle.priority_list = {}
    local datatable = {}

    for k, unit in pairs(player.heroes) do
        local name = DCEI.GetUnitType(unit)
        local unit_data = data.heroes[name]

        local cooldown = unit_data:GetAbilityCooldown()
        local cost = unit_data:GetAbilityCost()

        if cooldown then
            table.insert(datatable, { name, cooldown, cost })
        end
    end

    function sort_by_cooldown(a, b)
        return a[2] > b[2]
    end

    table.sort(datatable, sort_by_cooldown)

    -- for k, v in ipairs(datatable) do
    --     DCEI.LogMessage(k .. ". " .. v[1] .. " (cooldown " .. v[2] .. ") (cost " .. v[3] .. ")")
    -- end

    player.autobattle.priority_list = datatable
end

function ClearMonsterKillList()
    for _, monster in ipairs(data.lab_monsters) do
        wave_killed_monsters[monster.name] = 0
    end
end

function SettleMonsterKills(skip_commit)
    for name, count in pairs(wave_killed_monsters) do
        -- make sure the monster is in monster lab
        if data.lab_monsters[name] and count > 0 then
            if not player.monster_data[name] then
                player.monster_data[name] = {
                    kill = 0,
                    weak = {},
                }
                for i = 1, #data.lab_monsters[name].weak_point do
                    local purchased = false
                    table.insert(player.monster_data[name].weak, purchased)
                end
            end
            local prev_kill = player.monster_data[name].kill
            player.monster_data[name].kill = player.monster_data[name].kill + count
            player.monster_data[name].kill = math.min(player.monster_data[name].kill, 9999)
            local current_kill = player.monster_data[name].kill
            local pip = false
            local unlock_kill = data.lab_monsters[name].unlock_kill
            if prev_kill < unlock_kill and current_kill >= unlock_kill then
                pip = true
            end
            for i = 1, #data.lab_monsters[name].weak_point do
                local required_kill = data.lab_monsters[name].weak_point[i].required_kill
                if prev_kill < required_kill and current_kill >= required_kill then
                    pip = true
                end
            end

            if pip and player.system.first_open_monster_lab == system.active then
                data.lab_monsters[name].show_pip = true
                ui.bottom_panel:ShowTownPip()
                ui.glue.monster_lab_menu:ShowPip()
            end
        end
    end
    if not skip_commit then
        bank.CommitMonsterData()
    end

    if player.running_monster_bounty and ui.monster_bounty then
        ui.monster_bounty:UpdateProgress(player.running_monster_bounty.progress)
        if not skip_commit then
            bank.CommitMonsterBounty()
        end
    end
end

function CheckCondition_EnemiesBelow(height, condition)
    local enemies = wave.data.units
    local boss_height = 32.5
    local count = 0
    local min_count = 4

    -- use cached result if possible
    local cached_result = player.autobattle.conditions[condition]
    if cached_result ~= nil then
        return cached_result
    end

    if #enemies == 0 then
        -- early out if there's no enemies
        return false
    else
        -- count how many enemies are below requested height
        for _, unit in pairs(enemies) do
            local pos = DCEI.GetUnitPosition2D(unit)

            if UnitIsBoss(unit) and pos.y <= boss_height then
                -- condition is success if boss is in range
                return true
            elseif pos.y <= height then
                count = count + 1
            end
        end

        -- cache result
        local result = count >= min_count
        player.autobattle.conditions[condition] = result

        -- DCEI.LogMessage("CHECK HEIGHT CONDITION: " .. count .. " below " .. height)
        return result
    end
end

function CheckCondition_SkillsOnCooldown(ratio)
    local list = player.autobattle.priority_list
    local count_total = #list
    local count_on_cd = 0

    for k, v in ipairs(list) do
        local name = v[1]
        local unit_data = data.heroes[name]

        if HeroSkillOnCooldown(unit_data) then
            count_on_cd = count_on_cd + 1
        end
    end

    local cd_percent = count_on_cd / count_total

    -- DCEI.LogMessage("CHECK CD CONDTION: " .. cd_percent .. " " .. tostring(cd_percent >= ratio))
    return cd_percent >= ratio
end

function CheckCondition_CastleHP(ratio)
    local health = DCEI.GetHealth(CASTLE_UNIT)
    local max_health = DCEI.GetMaxHealth(CASTLE_UNIT)
    return health / max_health < ratio
end

function AutoBattleSkillConditionMet(hero_data)
    local condition = hero_data:GetAbilityAutoCondition()

    if condition == data.abt.none then
        return true
    elseif condition == data.abt.enemy_below_32 then
        local height = 32.5
        return CheckCondition_EnemiesBelow(height, condition)
    elseif condition == data.abt.enemy_below_28 then
        local height = 28.5
        return CheckCondition_EnemiesBelow(height, condition)
    elseif condition == data.abt.enemy_below_26 then
        local height = 26.5
        return CheckCondition_EnemiesBelow(height, condition)
    elseif condition == data.abt.enemy_below_24 then
        local height = 24.5
        return CheckCondition_EnemiesBelow(height, condition)
    elseif condition == data.abt.skill_cooldown then
        local ratio = 0.75
        return CheckCondition_SkillsOnCooldown(ratio)
    elseif condition == data.abt.castle_hp then
        local ratio = 0.5
        return CheckCondition_CastleHP(ratio)
    end
end

function ClearAutoBattleConditionCache()
    player.autobattle.conditions = {}
end

function AutoBattleChoose(manual_cast_timer)
    local manual_cast = false
    for _, entry in ipairs(player.autobattle.priority_list) do
        local name = entry[1]
        local hero_data = data.heroes[name]

        -- No auto battle in raid boss battle, if add auto battle in the future need add check here
        if HeroCanUseSkill(hero_data) and AutoBattleSkillConditionMet(hero_data) then
            local target_manual = data.GetHeroActiveSkillData(hero_data, "target_manual")
            -- increase cast check interval for manual cast skills to avoid over kill damage to waste mana
            if not target_manual or (target_manual and manual_cast_timer == 0) then
                local slot = hero_data.slot
                local hero = player.heroes[slot]

                HeroCastSkill(hero, hero_data)
                manual_cast = target_manual
            end
            break
        end
    end
    return manual_cast
end

function AutoBattleTick(manual_cast_timer)
    local interval = (#wave.data.units < 20) and 1 or 0.375
    local instance = wave.data.instance

    local manual_cast_timer = manual_cast_timer or 0

    ClearAutoBattleConditionCache()
    local manual_cast = 0
    if event_mode.IsActive() then
        manual_cast = event_mode.AutoBattleChoose(manual_cast_timer)
    else
        manual_cast = AutoBattleChoose(manual_cast_timer)
    end
    if manual_cast or manual_cast_timer > 0 then
        manual_cast_timer = manual_cast_timer + interval
        if manual_cast_timer > interval * 3 then
            manual_cast_timer = 0
        end
    end

    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        if wave.data.is_active and player.autobattle.is_active and instance == wave.data.instance then
            AutoBattleTick(manual_cast_timer)
        end
    end, interval, false, simple)
end

function DisableAutoBattle()
    player.autobattle.is_active = false
    player.autobattle.type = "none"

    ui.battle.floating_autobattle:Deactivate()
end

function HeroSkillOnCooldown(unit_data)
    if unit_data then
        local ability = data.GetHeroActiveSkillData(unit_data, "ability")
        local slot = unit_data.slot
        -- check if hero is equipped
        if slot and slot > 0 then
            local unit = player.heroes[slot]
            return DCEI.GetUnitAbilityCooldown(unit, ability) > 0
        end
    end

    -- return unit_data and not unit_data.active_skill_ready
end

function HeroCanUseSkill(hero_data, report_error)
    if not hero_data.skill.active then
        return
    end

    local cost = hero_data:GetAbilityCost()
    local mana = DCEI.GetMana(CASTLE_UNIT)

    -- check skill on cooldown
    if HeroSkillOnCooldown(hero_data) then
        if report_error then
            ui.ShowFeedbackSkillCooldown()
        end
        return false
    end

    -- check skill mana cost
    if cost > mana then
        if report_error then
            ui.ShowFeedbackNeedMoreMana(cost)
        end
        return false
    end

    return wave.data.is_active
end

-- !SECTION

function ProcessMonsterUnit(unit)
    table.insert(wave.data.units, unit)
    local event_tag = DCEI.GetUnitTagCount(unit, "gt_season_event")
    if event_tag > 0 then
        event_tag = math.floor(event_tag)
        local tutorial_key = CheckSeasonEventExist("winter_event2022") and "winter_event_monster_"
            or "autumn_event_monster_"
        tutorial_key = tutorial_key .. event_tag
        if not player.system[tutorial_key] or player.system[tutorial_key] == system.locked then
            CreateSeasonEventMonsterIntroIcon(event_tag)
        end
        if event_tag > 1 then
            SeasonEventEnemyPassBy(unit, event_tag > 2)
        else
            SendEnemyTowardsCastle(unit)
        end
    else
        SendEnemyTowardsCastle(unit)
    end

    -- apply upgrades to monster units
    ApplyUpgradeDataToUnit(data.monster_upgrades, unit)

    if UnitIsBoss(unit) then
        local name
        if wave.data.mode == "raid_boss" then
            name = levels.raid_boss.current.boss_name
        end
        ui.CreateBossBar(unit, name)
    end
end

-- ===================================================================================================================================================================================================================
-- SECTION HERO SKILLS AND SUMMONS
-- ===================================================================================================================================================================================================================

function ProcessSummonedUnit(unit)
    local unit_type = DCEI.GetUnitType(unit)

    if ProcessSummonedUnitCallbacks[unit_type] then
        ProcessSummonedUnitCallbacks[unit_type](unit)

        -- apply upgrades to summoned units
        ApplyUpgradeDataToUnit(data.summon_upgrades, unit)
    end
end

function ApplySummonHealthScaling(unit_data, unit)
    local growth_level = unit_data:GetLevel() - 1
    local base_health = unit_data.base_health
    local count = math.floor(growth_level * base_health * 0.1)
    --local total_health = unit_data:GetSummonHealthValue()
    --DCEI.ApplyBehaviorCountAsync(unit, "Shared Summon ShieldScaling", total_health)

    if count > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Shared Summon HealthScaling"), count)
    end

    --DCEI.SetShieldValueAsync(unit, 0)
end

function ApplyEquipmentAttributesToSummon(unit_data, unit)
    if unit_data.equipment.equipment_summon_defense then
        local behavior_name = DCEI.Behavior("Equipment Attribute Summon Defense")
        local value = unit_data.equipment.equipment_summon_defense
        local step = 0.5
        local stack_count = value / step
        DCEI.ApplyBehaviorCountAsync(unit, behavior_name, stack_count)
    end

    if unit_data.equipment.equipment_summon_health then
        local behavior_name = DCEI.Behavior("Equipment Attribute Summon Health")
        local value = unit_data.equipment.equipment_summon_health
        local step = 0.005
        local stack_count = value / step
        DCEI.ApplyBehaviorCountAsync(unit, behavior_name, stack_count)
    end

    if unit_data.equipment.equipment_summon_damage then
        local behavior_name = DCEI.Behavior("Equipment Attribute Summon Damage")
        local value = unit_data.equipment.equipment_summon_damage
        local step = 0.001
        local stack_count = value / step
        DCEI.ApplyBehaviorCountAsync(unit, behavior_name, stack_count)
    end
end

function ProcessHeroSkillAngelicHost(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero Angel")
    local pos = DCEI.GetUnitPosition2D(unit)

    --ui.CreateSummonCreatureHealthBar(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, pos.x, pos.y + 16)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    ApplyEquipmentAttributesToSummon(unit_data, unit)
    CheckAndApplyTroopRevive(unit_data, unit)
end

function ProcessHeroSkillDefend(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero Knight")
    local pos = DCEI.GetUnitPosition2D(unit)

    --ui.CreateSummonCreatureHealthBar(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, pos.x, pos.y + 16)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    ApplyEquipmentAttributesToSummon(unit_data, unit)
    CheckAndApplyTroopRevive(unit_data, unit)
end

function ProcessHeroSkillReinforce(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero Lancer")
    local pos = DCEI.GetUnitPosition2D(unit)

    --ui.CreateSummonCreatureHealthBar(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, pos.x, pos.y + 16)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    ApplyEquipmentAttributesToSummon(unit_data, unit)
    CheckAndApplyTroopRevive(unit_data, unit)
end

function ProcessHeroSkillBabyDragon(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero BabyDragon")
    local pos = DCEI.GetUnitPosition2D(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, pos.x, pos.y + 16)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    ApplyEquipmentAttributesToSummon(unit_data, unit)
    CheckAndApplyTroopRevive(unit_data, unit)
end

function ProcessHeroSkillWolf(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero WolfHuntress")
    local pos = DCEI.GetUnitPosition2D(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, pos.x, pos.y + 16)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    ApplyEquipmentAttributesToSummon(unit_data, unit)
    CheckAndApplyTroopRevive(unit_data, unit)

    -- remove season event skill when season event is over
    if unit_data.event_lock and not CheckSeasonEventExist(unit_data.event_lock) then
        DCEI.RemoveBehavior(unit, "Hero Season Event Skill Damage")
        DCEI.RemoveBehavior(unit, "Hero Season Event Skill Coin Drop")
    end
end

function CheckAndApplyTroopRevive(unit_data, unit)
    for _, skill in ipairs(unit_data.skill) do
        if skill.name == "Troop Revive" then
            DCEI.ApplyBehaviorToSelf(unit, DCEI.Behavior("Troop Revive"))
            return
        end
    end
end

function ProcessHeroSkillGoblin3AmineMove(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero Goblin")
    if unit_data.name == "Hero Goblin 3A" or unit_data.name == "Hero Goblin 4A" then
        local pos = DCEI.GetUnitPosition2D(unit)
        DCEI.Move(unit, pos.x, pos.y + 18)
    end
end

function ProcessTowerBarracksSummon(unit)
    if not wave.data.is_active then
        DCEI.RemoveUnit(unit)
    end
    local unit_data = GetTowerCurrentClassDataFromBaseClassName("Tower Barracks")
    local pos = DCEI.GetUnitPosition2D(unit)

    --ui.CreateSummonCreatureHealthBar(unit)

    table.insert(group_summons, unit)
    DCEI.MoveAttack(unit, 24, pos.y)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)
    local r = DCEI.Random(-2, 2)
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.MoveAttack(unit, 24 + r, pos.y + 3)
    end, 1.0, false, true)
    if wave.data.mode == "raid_boss" or wave.data.mode == "colony" then
        DCEI.TriggerAddTimerEventElapsed(function()
            DCEI.MoveAttack(unit, 24 + r, pos.y + 16)
        end, 6.0, false, true)
    end
end

function ProcessTowerMonolithSummon(unit)
    if not wave.data.is_active then
        DCEI.RemoveUnit(unit)
    end
    local unit_data = GetTowerCurrentClassDataFromBaseClassName("Tower Monolith")
    local pos = DCEI.GetUnitPosition2D(unit)

    --ui.CreateSummonCreatureHealthBar(unit)

    table.insert(group_summons, unit)
    -- let troop face south and wait for scale animation
    DCEI.TurnUnitTowards(unit, 0, -1)
    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ApplySummonHealthScaling(unit_data, unit)

    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.MoveAttack(unit, 24, 19)
    end, 0.5, false, true)
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.MoveAttack(unit, 24, pos.y + 16)
    end, 1.0, false, true)
end

function ProcessLegendarySummonSkeleton(unit)
    if not wave.data.is_active then
        DCEI.RemoveUnit(unit)
    end

    table.insert(group_summons, unit)

    -- use the same attack and health stacks in this wave
    wave.ProcessPlayerSkeletonSummons(unit)
end

function ProcessHeroSkillLeech()
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero DarkScout")

    local multi = 0.25
    local base = unit_data:GetAttackValue()
    local heal_value = base * multi

    local health = DCEI.GetHealth(CASTLE_UNIT) + heal_value
    SetCastleHealth(health, heal_value)
    -- DCEI.LogMessage("heal for " .. heal_value)
end

function ProcessHeroSkillLeechSoul()
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero DarkScout")

    local base = unit_data:GetAttackValue()
    local heal_value = base * 0.25
    local mana_value = base * 0.1

    local health = DCEI.GetHealth(CASTLE_UNIT) + heal_value
    SetCastleHealth(health, heal_value)

    DCEI.AddMana(CASTLE_UNIT, mana_value)
end

function ProcessHeroLifesteal(unit)
    local hero_name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[hero_name]

    local base = unit_data:GetAttackValue()
    local heal_value = base * unit_data.equipment.equipment_lifesteal

    local health = DCEI.GetHealth(CASTLE_UNIT) + heal_value
    SetCastleHealth(health, heal_value)
end

function ProcessParagonHeroLifesteal(unit)
    local hero_name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[hero_name]

    local base = unit_data:GetAttackValue()
    local heal_value = base * player.attributes.hero_life_steal

    local health = DCEI.GetHealth(CASTLE_UNIT) + heal_value
    SetCastleHealth(health, heal_value)
end

function ProcessRoyalRoarLifesteal(unit)
    local hero_name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[hero_name]

    local base = unit_data:GetAttackValue()
    local heal_value = base * 0.05

    local health = DCEI.GetHealth(CASTLE_UNIT) + heal_value
    SetCastleHealth(health, heal_value)
end

function ProcessSummonArcaneGolemManaSteal()
    local name = "Primal Amethyst"
    local base = GetArcherDamageValue()
    local unit_data = data.gear[name]

    local growth_level = unit_data:GetLevel() - 1
    local multiplier = unit_data.values[1].base
    local steal_percent = 0.05

    if growth_level then
        local growth = unit_data.values[1].growth
        multiplier = multiplier + (growth_level * growth)
    end

    local mana_value = base * multiplier * steal_percent
    DCEI.AddMana(CASTLE_UNIT, mana_value)
end

function ProcessHeroSkillRepair()
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero Mason")
    local id = unit_data.skill.active
    local value = unit_data.skill[id].values[1].base

    if wave.data.mode == "hell_mode" then
        value = value * 0.3
    end

    local max_health = DCEI.GetMaxHealth(CASTLE_UNIT)
    DCEI.AddHealth(CASTLE_UNIT, max_health * value)

    CASTLE_UNIT_INVINC = true
    DCEI.TriggerAddTimerEventElapsed(function()
        CASTLE_UNIT_INVINC = false
    end, 0.0125, false, true)
end

function ProcessDevilTrailCastleHPLose()
    local max_health = DCEI.GetMaxHealth(CASTLE_UNIT)
    local stack = DCEI.GetUnitBehaviorStackCount(CASTLE_UNIT, "Devil Trail Negative Castle HP Lose")
    local damage = max_health * 0.01 * stack
    local castle_health = DCEI.GetHealth(CASTLE_UNIT) - damage

    SetCastleHealth(castle_health, -damage)
end

function ProcessHeroSkillChannel(unit)
    local unit_data = GetHeroCurrentClassDataFromBaseClassName("Hero WhiteMage")
    local level = unit_data:GetLevel()
    local id = unit_data.skill.active
    local value = unit_data.skill[id].values[1]

    local cooldown_reduction = value.base + ((level - 1) * value.growth)
    local unit_data = GetHeroDataFromUnitType(DCEI.GetUnitType(unit))

    ReduceHeroSkillCooldown(unit_data, cooldown_reduction)
end

function ProcessHeroSkillCurse(unit)
    local unit_data = GetHeroDataFromUnitType("Hero BlackMage")
    local level = unit_data:GetLevel()
    local behavior = DCEI.Behavior("Hero BlackMage Skill Curse Debuff Stack")
    local max = 40
    local stacks = math.min(level - 1, max)

    if stacks > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, behavior, stacks)
    end
end

function ProcessHeroSkillDecrepify(unit)
    local unit_data = GetHeroDataFromUnitType("Hero BlackMage")
    local level = unit_data:GetLevel()
    local behavior = DCEI.Behavior("Hero BlackMage Skill Decrepify Debuff Stack")
    local max = 40
    local stacks = math.min(level - 1, max)

    if stacks > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, behavior, stacks)
    end
end

function ProcessHeroSkillInspire(unit)
    local unit_data = GetHeroDataFromUnitType("Hero Preacher")
    local level = unit_data:GetLevel()
    local behavior = DCEI.Behavior("Hero Preacher Skill Inspire ArcherAttackDamage Buff Stack")
    local max = 20
    local stacks = math.min(level - 1, max)

    if stacks > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, behavior, stacks)
    end
end

function ProcessHeroSkillRousingWords(unit)
    local behavior = DCEI.Behavior("Hero Preacher Skill RousingWords HeroAttackSpeed Buff Stack")
    local unit_data = GetHeroDataFromUnitType("Hero Preacher")
    local level = unit_data:GetLevel()
    local max = 20
    local stacks = math.min(level - 1, max)

    if stacks > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, behavior, stacks)
    end
end

function ProcessHeroSkillVolley(unit)
    local upgrade_behavior = DCEI.Behavior("Hero Marksman Skill Volley ArcherAttackSpeed Buff Stack")
    local unit_data = GetHeroDataFromUnitType("Hero Marksman")
    local level = unit_data:GetLevel()
    local max = 20
    local stacks = math.min(level - 1, max)

    if stacks > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, upgrade_behavior, stacks)
    end
end

function ProcessIceFieldGenerator(unit)
    local name = "Defense IceFieldGenerator"
    local unit_data = data.defenses[name]
    local IFG_unit = player.defenses[unit_data.slot]
    if IFG_unit and DCEI.UnitExists(IFG_unit) then
        if unit_data.slot == 1 then
            local behavior_name = DCEI.Behavior("Defense IceFieldGenerator Aura Left")
            DCEI.ApplyBehaviorAsync(IFG_unit, behavior_name)
        elseif unit_data.slot == 2 then
            local behavior_name = DCEI.Behavior("Defense IceFieldGenerator Aura Right")
            DCEI.ApplyBehaviorAsync(IFG_unit, behavior_name)
        end
    end
end

function ProcessAcidArrowUpgrade(unit)
    local upgrade_behavior = DCEI.Behavior("Castle Archer Gear AcidArrow Debuff Upgrade")
    local growth_level = data.gear["Acid Arrow"]:GetLevel() - 1

    if growth_level > 0 then
        local current = DCEI.GetUnitBehaviorStackCount(unit, upgrade_behavior)
        if current < growth_level then
            DCEI.ApplyBehaviorCountAsync(unit, upgrade_behavior, growth_level)
        end
    end
end

function ProcessIceArrowUpgrade(unit)
    local upgrade_behavior = DCEI.Behavior("Castle Archer Gear IceArrow Slow Upgrade")
    local growth_level = data.gear["Ice Arrow"]:GetLevel() - 1

    if growth_level > 0 then
        local current = DCEI.GetUnitBehaviorStackCount(unit, upgrade_behavior)
        if current < growth_level then
            DCEI.ApplyBehaviorCountAsync(unit, upgrade_behavior, growth_level)
        end
    end
end

function ProcessUnityHeroCooldownReduce(unit)
    local cooldown_reduction = 0.5
    local unit_data = GetHeroDataFromUnitType(DCEI.GetUnitType(unit))

    ReduceHeroSkillCooldown(unit_data, cooldown_reduction)
end

-- !SECTION

-- ===================================================================================================================================================================================================================
-- SECTION COMBAT
-- ===================================================================================================================================================================================================================

function SetCastleHealth(health, change)
    if CASTLE_UNIT_INVINC then
        return
    end

    if health > 0 then
        DCEI.AddHealth(CASTLE_UNIT, change)
        -- DCEI.LogMessage("-- damage processed: " .. damage)
    elseif wave.data.is_active then
        -- Defeat & Reset Game
        DCEI.SetHealthValueAsync(CASTLE_UNIT, 0)
        WaveEndActions(false, true)
    end
end

function ProcessCastleDamage(castle_chunk)
    local name = DCEI.GetUnitType(castle_chunk)
    local health_max = DCEI.GetUnitTypeHealth(name)
    local health_cur = DCEI.GetHealth(castle_chunk)

    local damage = health_max - health_cur
    local castle_health = DCEI.GetHealth(CASTLE_UNIT) - damage
    local b = "Player Castle Chunk HitResponse Trigger"
    castle_health = math.max(0, castle_health)

    DCEI.RemoveBehavior(castle_chunk, b)
    DCEI.SetHealth(castle_chunk, health_max)

    SetCastleHealth(castle_health, -damage)
end

function ShowHeroSkillBars()
    for _, unit in pairs(player.heroes) do
        ShowUnitSkillBar(unit, true)
    end
end

function HideHeroSkillBars()
    for _, unit in pairs(player.heroes) do
        ShowUnitSkillBar(unit, false)
    end
end

function StartHeroSkillCooldown(unit, unit_data)
    local ability = data.GetHeroActiveSkillData(unit_data, "ability")
    local duration = unit_data:GetAbilityCooldown()
    local slot = unit_data.slot

    -- set skill progress bar to (empty) cast state
    ui.loadout.hero.slots[slot].skill_bar:Cast()

    -- adjust ability cooldown duration
    DCEI.SetUnitAbilityMaxCooldown(unit, ability, duration)

    -- for Nidhogg boss skill, add checking behavior when hero cast skill
    DCEI.ApplyBehaviorToSelf(unit, DCEI.Behavior("Hero Cast Skill"))

    -- add archer critical damage if equip gear unity
    if player.attributes.skill_archer_cri_dmg > 0 then
        -- refresh expire time when triggered
        GEAR_UNITY_HELPER_BUFF_INSTANCE = GEAR_UNITY_HELPER_BUFF_INSTANCE + 1
        ApplyGearUnityArcherCriticalDamage(1)
    end
end

function ApplyGearUnityArcherCriticalDamage(count)
    local instance = GEAR_UNITY_HELPER_BUFF_INSTANCE
    local behavior_name = DCEI.Behavior("Gear Unity Triggered ArcherCritDamage")
    local upgrade_behavior_name = DCEI.Behavior("Gear Unity Triggered ArcherCritDamage Upgrade")
    local growth_level = data.gear["Unity"]:GetLevel() - 1
    -- at least 1 archers, get first archer to check stacks
    local first_archer = group_archers[1]
    local stack = DCEI.GetUnitBehaviorStackCount(first_archer, behavior_name)
    local upgrade_behavior_stack = DCEI.GetUnitBehaviorStackCount(first_archer, upgrade_behavior_name)
    stack = stack + count
    stack = math.min(stack, 10)
    local upgrade_stack = stack * growth_level
    local duration = 2
    for _, unit in pairs(group_archers) do
        DCEI.RemoveBehavior(unit, behavior_name)
        DCEI.RemoveBehavior(unit, upgrade_behavior_name)
        if stack > 0 then
            DCEI.ApplyBehaviorToSelf(unit, behavior_name, stack)
        end
        if upgrade_stack > 0 then
            DCEI.ApplyBehaviorToSelf(unit, upgrade_behavior_name, upgrade_stack)
        end
    end

    if stack > 0 then
        local number = stack * 10 + upgrade_stack * 0.5
        local text = "<size=80%>"
            .. DCEI.GetTranslationText("data/bonus/critical_damage", { x = "<color=green>+" .. number .. "%</color>" })
            .. "</size>"

        DCEI.HideUnitLabel(GEAR_UNITY_HELPER)
        -- shows unit label
        DCEI.ShowUnitLabel(GEAR_UNITY_HELPER, text)

        DCEI.TriggerAddTimerEventElapsed(function()
            if instance == GEAR_UNITY_HELPER_BUFF_INSTANCE then
                ApplyGearUnityArcherCriticalDamage(-1)
            end
        end, duration, false, true)
    else
        DCEI.HideUnitLabel(GEAR_UNITY_HELPER)
    end
end

function StartHeroCooldownVisuals(unit)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]

    if unit_data then
        local ability = data.GetHeroActiveSkillData(unit_data, "ability")
        local slot = unit_data.slot

        ui.loadout.hero.slots[slot].skill_bar:Cooldown(ability)
    end
end

function ShowUnitSkillBar(unit, is_show, battle_start)
    local name = DCEI.GetUnitType(unit)
    local hero_data = data.heroes[name]
    local slot = hero_data.slot
    local unit = player.heroes[slot]
    local skill_bar = ui.loadout.hero.slots[slot].skill_bar

    if hero_data.skill.active then
        if is_show then
            skill_bar:Show()
        else
            skill_bar:Hide()
        end
    end
end

function ApplyHeroInCombatBehavior(unit)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]

    if unit_data and unit_data.in_combat_behavior then
        for _, behavior in ipairs(unit_data.in_combat_behavior) do
            DCEI.ApplyBehaviorAsync(unit, behavior)
        end
    end

    -- remove season event skill when season event is over
    if unit_data.event_lock and not CheckSeasonEventExist(unit_data.event_lock) then
        DCEI.RemoveBehavior(unit, "Hero Season Event Skill Damage")
        DCEI.RemoveBehavior(unit, "Hero Season Event Skill Coin Drop")
    end
end

function RemoveHeroInCombatBehavior(unit)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]

    if unit_data and unit_data.in_combat_behavior then
        for _, behavior in ipairs(unit_data.in_combat_behavior) do
            DCEI.RemoveBehaviorAsync(unit, behavior)
        end
    end

    -- remove combat behaviors
    for _, behavior in ipairs(data.combat_behaviors) do
        DCEI.RemoveBehaviorAsync(unit, behavior)
    end
end

function EnableHeroSkill(unit)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]
    local ability = data.GetHeroActiveSkillData(unit_data, "ability")

    if unit_data and ability then
        local slot = unit_data.slot
        local skill_bar = ui.loadout.hero.slots[slot].skill_bar

        if unit_data.skill.active then
            skill_bar:Bind(unit)
            skill_bar:Ready()
        end
    end
end

function ResetHeroSkill(unit)
    local name = DCEI.GetUnitType(unit)
    local unit_data = data.heroes[name]
    local ability = data.GetHeroActiveSkillData(unit_data, "ability")

    if unit_data and ability then
        local slot = unit_data.slot
        local unit = player.heroes[slot]

        -- removes cooldown of ability in case ability on cooldown at battle start
        DCEI.SetUnitAbilityCooldown(unit, ability, 0)

        local skill_bar = ui.loadout.hero.slots[slot].skill_bar
        skill_bar:Reset()
        unit_data.stun = false
    end
end

function ReduceHeroSkillCooldown(unit_data, duration)
    local ability = data.GetHeroActiveSkillData(unit_data, "ability")
    local slot = unit_data.slot
    local unit = player.heroes[slot]

    if unit_data and unit and ability then
        -- reduce current cooldown of ability
        local old_cooldown = DCEI.GetUnitAbilityCooldown(unit, ability)
        local new_cooldown = math.max(old_cooldown - duration, 0)

        if new_cooldown <= 0 then
            -- immediately set skill progress bar to ready
            ui.loadout.hero.slots[slot].skill_bar:Ready()
        end

        DCEI.SetUnitAbilityCooldown(unit, ability, new_cooldown)
    end
end

-- HELPERS
function UnitIsBoss(unit)
    -- return false without checking behaviors if boss hasn't spawned yet
    return wave.data.is_boss_spawned and DCEI.GetUnitBehaviorStackCount(unit, "Unit Tag Boss") > 0
end

function UnitIsHero(unit)
    for _, hero in pairs(player.heroes) do
        if unit == hero then
            return true
        end
    end
    return false
end

function GetUnitDataFromUnitType(unit_type, data_set)
    -- 12/15/2020 update, use dictionary to search unit data
    if data_set[unit_type] then
        return data_set[unit_type]
    end

    -- old implementation, use linear serach, keep it in case there is no such key in dictionary
    for id, unit_data in ipairs(data_set) do
        if unit_type == unit_data.name then
            return unit_data
        end
    end
end

function GetHeroCurrentClassDataFromBaseClassName(name)
    local root_data = data.heroes[name]
    local class = root_data.current_class or root_data.name
    return data.heroes[class]
end

function GetTowerCurrentClassDataFromBaseClassName(name)
    local root_data = data.towers[name]
    local class = root_data.current_class or root_data.name
    return data.towers[class]
end

function ApplyHeroAttackDamage(hero_data, unit, multiplier)
    multiplier = multiplier or 1
    local base_damage = hero_data:GetAttackValue()
    local damage = base_damage * multiplier
    core.DamageUnit(unit, damage)
end

function GetHeroDataFromUnitType(unit_type)
    return GetUnitDataFromUnitType(unit_type, data.heroes)
end

function GetHeroGalleryIndex(hero_data)
    if not ui.list_hero then
        InitHeroList()
    end

    for k, list_item in ipairs(ui.list_hero.items) do
        if list_item.unit_data == hero_data then
            return k
        end
    end

    return 1
end

function GetTowerGalleryIndex(tower_data)
    if not ui.list_tower then
        InitTowerList()
    end

    for k, list_item in ipairs(ui.list_tower.items) do
        if list_item.unit_data == tower_data then
            return k
        end
    end

    return 1
end

function GetUnitDataIndex(unit_data, data_set)
    for k, data in ipairs(data_set) do
        if data == unit_data then
            return k
        end
    end
end

function RollForContractDrop()
    local chance = 40
    local d100 = math.random(1, 100)

    local mine_boost = GetPlayerBoostModifier_Mine() * 100
    chance = chance + mine_boost

    if wave.data.GetMax() >= 25 or wave.data.mode == "replay" then
        -- normally roll for contract drop chance
        wave.data.drop_contract = (d100 <= chance)
    elseif wave.data.GetMax() < 4 then
        -- boss 1 doesn't drop contracts
        wave.data.drop_contract = false
    elseif wave.data.GetMax() < 20 then
        -- boss 2, 3, and 4 always drops contracts on first play
        wave.data.drop_contract = true
    end

    -- wave 100 boss must drop a common_chest
    if
        wave.data.GetMax() >= 99
        and wave.data.mode ~= "replay"
        and (not player.system.open_first_common_chest or player.system.open_first_common_chest == system.locked)
    then
        wave.data.drop_chests = {}
        wave.data.drop_chests.common_chest = 1

        player.system.open_first_common_chest = system.unlocked
        if not event_mode.IsActive() then
            ui.glue.inventory_menu:ShowPipAnimated()
        end

        if not player.system.raid_boss_tutorial or player.system.raid_boss_tutorial == system.locked then
            player.system.raid_boss_tutorial = system.unlocked
            -- only show pip on nav button once after raid boss unlock
            ui.bottom_panel:ShowTownPip()
            ui.loadout.town.airship:ShowPipAnimated()
        end
        bank.CommitSystemState()
    end
end

function ProcessRaidBossDrop()
    local raid_boss = levels.raid_boss.current
    wave.data.drop_chests = {}
    local reward_count = 1

    if raid_boss.reward.epic_material then
        wave.data.drop_material = {}
        local health = DCEI.GetHealth(CASTLE_UNIT)
        local max_health = DCEI.GetMaxHealth(CASTLE_UNIT)
        local epic_count = 1
        local hp_fraction = health / max_health
        if hp_fraction >= 0.7 then
            epic_count = 3
        elseif hp_fraction >= 0.4 then
            epic_count = 2
        end
        wave.data.drop_material["epic_material"] = epic_count
    end

    if raid_boss.turbo_on and raid_boss.raid_turbo_cost then
        reward_count = reward_count + 1
        if wave.data.drop_material then
            for k, v in pairs(wave.data.drop_material) do
                wave.data.drop_material[k] = v + 1
            end
        end
        AddPlayerRaidTurbo(-raid_boss.raid_turbo_cost)
    end
    wave.data.drop_chests[raid_boss.reward.name] = reward_count
end

-- !SECTION

-- ===================================================================================================================================================================================================================
-- SECTION RESOURCES
-- ===================================================================================================================================================================================================================

function GetPlayerWinterCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.winter_coin))
end

function AddPlayerWinterCoin(value, skip_commit)
    SetPlayerWinterCoin(value + GetPlayerWinterCoin(), skip_commit)
end

function SetPlayerWinterCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.winter_coin, value)

    if not skip_commit then
        bank.CommitWinterCoin()
    end
end

function GetPlayerSpringCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.spring_coin))
end

function AddPlayerSpringCoin(value, skip_commit)
    SetPlayerSpringCoin(value + GetPlayerSpringCoin(), skip_commit)
end

function SetPlayerSpringCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.spring_coin, value)

    if not skip_commit then
        bank.CommitSpringCoin()
    end
end

function GetPlayerRoyalCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.royal_coin))
end

function AddPlayerRoyalCoin(value, skip_commit)
    SetPlayerRoyalCoin(value + GetPlayerRoyalCoin(), skip_commit)
end

function SetPlayerRoyalCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.royal_coin, value)

    if not skip_commit then
        bank.CommitRoyalCoin()
    end
end

function GetPlayerSummerCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.summer_coin))
end

function AddPlayerSummerCoin(value, skip_commit)
    SetPlayerSummerCoin(value + GetPlayerSummerCoin(), skip_commit)
end

function SetPlayerSummerCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.summer_coin, value)

    if not skip_commit then
        bank.CommitSummerCoin()
    end
end

function GetPlayerAutumnCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.autumn_coin))
end

function AddPlayerAutumnCoin(value, skip_commit)
    SetPlayerAutumnCoin(value + GetPlayerAutumnCoin(), skip_commit)
end

function SetPlayerAutumnCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.autumn_coin, value)

    if not skip_commit then
        bank.CommitAutumnCoin()
    end
end

function GetPlayerWinterCoin2022()
    return math.floor(DCEI.GetProtectedValue(data.safe.winter_coin2022))
end

function AddPlayerWinterCoin2022(value, skip_commit)
    SetPlayerWinterCoin2022(value + GetPlayerWinterCoin2022(), skip_commit)
end

function SetPlayerWinterCoin2022(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.winter_coin2022, value)

    if not skip_commit then
        bank.CommitWinterCoin2022()
    end
end

function GetPlayerElixir()
    return math.floor(DCEI.GetProtectedValue(data.safe.elixir))
end

function AddPlayerElixir(value, skip_commit)
    SetPlayerElixir(value + GetPlayerElixir(), skip_commit)
end

function SetPlayerElixir(value, skip_commit)
    -- enforce elixir cap
    local cap = data.town_upgrade:GetElixirCap()
    value = math.min(value, cap)

    DCEI.SetProtectedValue(data.safe.elixir, value)

    -- update town hall display
    if ui.town_menu then
        ui.town_menu:UpdateElixirDisplay()
    end

    -- logging
    if not skip_commit then
        bank.CommitElixir()
    end
end

function SpendPlayerElixir(value)
    -- minimum cost of 0 to prevent negative values from cheaters
    local spend = math.max(value, 0)
    AddPlayerElixir(-1 * spend)

    -- logging
    bank.CommitElixir()
end

function AddPlayerContract_Town(value, skip_commit)
    AddTownContracts(value)

    if GetTownUpgradeLevel() <= 10 and GetTownUpgradeLevel() > 0 and value > 0 then
        if not event_mode.IsActive() then
            ui.glue.inventory_menu:ShowPipAnimated()
        end
        ui.glue.inventory_menu.resource_pip[2] = true
    end

    if not skip_commit then
        bank.CommitTownContract()
    end
end

function AddPlayerContract_Mine(value, skip_commit)
    AddMineContracts(value)

    if GetDwarfLevel() == 1 and value > 0 then

        ui.glue.inventory_menu:ShowPipAnimated()
        ui.glue.inventory_menu.resource_pip[1] = true
    end
    CheckMineUpgradable()

    if not skip_commit then
        if event_mode.IsActive() then
            bank.CommitEventMineContract()
            bank.CommitEventMinetime()
        else
            bank.CommitMineContract()
        end
    end
end

function AddPlayerChest(chest_type, value, skip_commit)
    AddChestCount(chest_type, value)
    if not skip_commit then
        bank.CommitChest(chest_type)
    end
end

function AddPlayerMaterial(material_type, value, skip_commit)
    AddMaterialCount(material_type, value)

    if not skip_commit then
        bank.CommitMaterial(material_type)
    end
end

function AddPlayerPhilosopherStone(value)
    AddPhilosopherStone(value)

    bank.CommitPhilosopherStone()
end

function AddPlayerRaidTurbo(value)
    AddRaidTurbo(value)

    bank.CommitRaidTurbo()
end

function AddPlayerRaidKeyDaily(value)
    AddRaidKeyDailyCount(value)
    bank.CommitRaidKeyDailyCount()
end

function AddPlayerRaidKeyExtra(value)
    AddRaidKeyExtraCount(value)

    bank.CommitRaidKeyExtraCount()
end

-- value < 0
function SpendPlayerRaidKey(value)
    if GetRaidKeyDailyCount() > 0 then
        AddPlayerRaidKeyDaily(value)
    elseif GetRaidKeyExtraCount() > 0 then
        AddPlayerRaidKeyExtra(value)
    end
end

function GetPlayerGems()
    return DCEI.GetProtectedValue(data.safe.gems) + DCEI.GetProtectedValue(data.safe.gems_overflow)
end

function AddPlayerOverflowGems(value, skip_commit)
    local overflow_value = math.max(0, value) + DCEI.GetProtectedValue(data.safe.gems_overflow)
    DCEI.SetProtectedValue(data.safe.gems_overflow, math.max(0, overflow_value))

    UpdateGemDisplay()

    if not skip_commit then
        bank.CommitGems()
    end
end

function AddPlayerGems(value, skip_commit)
    local max_gems = 50 + player.attributes.token_cap

    local underflow_value = DCEI.GetProtectedValue(data.safe.gems)
    local new_value = underflow_value + value
    if underflow_value < max_gems then
        -- clamp new gems within max gems
        local clamped_value = math.min(max_gems, new_value)
        DCEI.SetProtectedValue(data.safe.gems, math.max(0, clamped_value))
    end

    if DCEI.GetProtectedValue(data.safe.gems) >= max_gems then
        ui.ShowFeedbackTokenFull()
    end

    -- -- check overflow gems
    -- local overflow_value = math.max(0, new_value - max_gems) + DCEI.GetProtectedValue(data.safe.gems_overflow)

    -- if allow_overflow and overflow_value > 0 then
    --     DCEI.SetProtectedValue(data.safe.gems_overflow, math.max(0, overflow_value))
    -- end

    UpdateGemDisplay()

    if not skip_commit then
        bank.CommitGems()
    end
end

function SpendPlayerGems(cost, spend_event)
    local underflow_value = DCEI.GetProtectedValue(data.safe.gems)
    local overflow_value = DCEI.GetProtectedValue(data.safe.gems_overflow)

    -- spend from gems first
    underflow_value = underflow_value - cost

    -- spend gems from overflow only if gems is 0
    if underflow_value < 0 then
        overflow_value = overflow_value + underflow_value
    end

    DCEI.SetProtectedValue(data.safe.gems, math.max(0, underflow_value))
    DCEI.SetProtectedValue(data.safe.gems_overflow, math.max(0, overflow_value))

    UpdateGemDisplay()

    -- logging
    bank.CommitGems()
    SpendLog_Token(cost, spend_event)
end

function GetPlayerGold()
    local max_underflow = 1000000000

    -- player gold (under 1b) is stored as a protected value in data.safe.gold
    local underflow = DCEI.GetProtectedValue(data.safe.gold)

    -- player gold (billions value) is stored separately in data.save.billions
    local overflow = DCEI.GetProtectedValue(data.safe.billions) * max_underflow

    return underflow + overflow
end

function SetPlayerGold(value)
    value = math.floor(value)
    value = math.max(value, 0)

    local max_underflow = 1000000000
    local underflow = value
    local overflow = 0

    -- store values above/below 1b separately due to int overflow issues in protected values
    if value >= max_underflow then
        underflow = value % max_underflow
        overflow = math.floor(value / max_underflow)
    end

    DCEI.SetProtectedValue(data.safe.gold, underflow)
    DCEI.SetProtectedValue(data.safe.billions, overflow)

    UpdateGoldDisplay()
end

function AddPlayerGold(value, skip_town_upgrade_check)
    SetPlayerGold(value + GetPlayerGold())

    if not skip_town_upgrade_check then
        CheckTownBuildingUpgradable()
        CheckMineUpgradable()
    end
end

function SpendPlayerGold(cost, spend_event)
    -- spend player gold is used to combat memory hacking where players can set unit prices to negative values
    local spend = math.max(cost, 0)

    AddPlayerGold(-1 * cost)

    -- logging
    SpendLog_Gold(cost, spend_event)
    bank.CommitGold()
end

function GetVaultGold()
    local max_underflow = 1000000000

    local underflow = DCEI.GetProtectedValue(data.safe.vault_gold)

    local overflow = DCEI.GetProtectedValue(data.safe.vault_billions) * max_underflow

    return underflow + overflow
end

function SetVaultGold(value)
    value = math.floor(value)
    value = math.max(value, 0)
    value = math.min(value, data.vault.capacity)

    local max_underflow = 1000000000
    local underflow = value
    local overflow = 0

    -- store values above/below 1b separately due to int overflow issues in protected values
    if value >= max_underflow then
        underflow = value % max_underflow
        overflow = math.floor(value / max_underflow)
    end

    DCEI.SetProtectedValue(data.safe.vault_gold, underflow)
    DCEI.SetProtectedValue(data.safe.vault_billions, overflow)
end

function AddVaultGold(value)
    SetVaultGold(value + GetVaultGold())
end

-- ===================================================================================================================================================================================================================
-- NOTE PROTECTED VALUES
-- ===================================================================================================================================================================================================================

-- PLAYER LEVEL
function GetPlayerLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.player_level))
end

function SetPlayerLevel(value)
    DCEI.SetProtectedValue(data.safe.player_level, math.floor(value))
end

function AddPlayerLevel(value)
    SetPlayerLevel(math.floor(value) + GetPlayerLevel())
end

function GetParagonLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.paragon_level))
end

function SetParagonLevel(value)
    DCEI.SetProtectedValue(data.safe.paragon_level, math.floor(value))
end

function AddParagonLevel(value)
    SetParagonLevel(math.floor(value) + GetParagonLevel())
end

-- ARCHER LEVEL
function GetArcherLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.archer_level))
end

function SetArcherLevel(value)
    DCEI.SetProtectedValue(data.safe.archer_level, math.floor(value))
end

function AddArcherLevel(value)
    SetArcherLevel(math.floor(value) + GetArcherLevel())
end

-- CASTLE LEVEL
function GetCastleLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.castle_level))
end

function SetCastleLevel(value)
    DCEI.SetProtectedValue(data.safe.castle_level, math.floor(value))
end

function AddCastleLevel(value)
    SetCastleLevel(math.floor(value) + GetCastleLevel())
end

-- TALENT POINT
function GetTalentPoints()
    return math.floor(DCEI.GetProtectedValue(data.safe.talent_point))
end

function SetTalentPoints(value)
    DCEI.SetProtectedValue(data.safe.talent_point, math.floor(value))
end

function AddTalentPoints(value)
    SetTalentPoints(math.floor(value) + GetTalentPoints())
end

function GetParagonPoints()
    return math.floor(DCEI.GetProtectedValue(data.safe.paragon_point))
end

function SetParagonPoints(value)
    DCEI.SetProtectedValue(data.safe.paragon_point, math.floor(value))
end

function AddParagonPoints(value)
    SetParagonPoints(math.floor(value) + GetParagonPoints())
end

-- MINE CONTRACT
function GetMineContracts()
    return math.floor(DCEI.GetProtectedValue(data.safe.mine_contract))
end

function SetMineContracts(value)
    DCEI.SetProtectedValue(data.safe.mine_contract, math.floor(value))
end

function AddMineContracts(value)
    SetMineContracts(math.floor(value) + GetMineContracts())
end

-- DWARF LEVEL
function GetDwarfLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.dwarf_level))
end

function SetDwarfLevel(value)
    value = math.min(value, data.dwarf_workers.max_level)

    DCEI.SetProtectedValue(data.safe.dwarf_level, math.floor(value))
end

function AddDwarfLevel(value)
    SetDwarfLevel(math.floor(value) + GetDwarfLevel())
end

data.dwarf_workers.GetLevel = GetDwarfLevel

function GetVaultLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.vault_level))
end

function SetVaultLevel(value)
    value = math.min(value, data.vault.max_level)

    DCEI.SetProtectedValue(data.safe.vault_level, math.floor(value))
end

function AddVaultLevel(value)
    SetVaultLevel(math.floor(value) + GetVaultLevel())
end

data.vault.GetLevel = GetVaultLevel
data.vault.colony = levels.colony

-- TOWN CONTRACT
function GetTownContracts()
    return math.floor(DCEI.GetProtectedValue(data.safe.town_contract))
end

function SetTownContracts(value)
    DCEI.SetProtectedValue(data.safe.town_contract, math.floor(value))
end

function AddTownContracts(value)
    SetTownContracts(math.floor(value) + GetTownContracts())
end

-- CHEST COUNTS
function GetChestCount(chest_type)
    return math.floor(DCEI.GetProtectedValue(data.safe[chest_type]))
end

function SetChestCount(chest_type, value)
    DCEI.SetProtectedValue(data.safe[chest_type], math.floor(value))
end

function AddChestCount(chest_type, value)
    if value > 0 and not event_mode.IsActive() then
        ui.glue.inventory_menu:ShowPipAnimated()
    end
    SetChestCount(chest_type, math.floor(value) + GetChestCount(chest_type))
end

-- CRAFT LEVEL
function GetCraftLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.craft_level))
end

function SetCraftLevel(value)
    DCEI.SetProtectedValue(data.safe.craft_level, math.floor(value))
end

function AddCraftLevel(value)
    SetCraftLevel(math.floor(value) + GetCraftLevel())
end

-- MATERIAL COUNTS
function GetMaterialCount(material_type)
    return math.floor(DCEI.GetProtectedValue(data.safe[material_type]))
end

function SetMaterialCount(material_type, value)
    DCEI.SetProtectedValue(data.safe[material_type], math.floor(value))
end

function AddMaterialCount(material_type, value)
    SetMaterialCount(material_type, math.floor(value) + GetMaterialCount(material_type))
end

-- TOWN UPGRADE LEVEL
function GetTownUpgradeLevel()
    return math.floor(DCEI.GetProtectedValue(data.safe.town_upgrade))
end

function SetTownUpgradeLevel(value)
    -- enforce max town level
    local above_level_cap = value > data.town_upgrade.max_upgrade_level
    if above_level_cap then
        value = data.town_upgrade.max_upgrade_level
    end

    DCEI.SetProtectedValue(data.safe.town_upgrade, math.floor(value))

    if above_level_cap then
        bank.CommitTownLevel()
    end

    -- set shipment bonus attribute so it can be referenced in colony gold calculations
    -- player.attributes.colony_gold_flat = data.town_upgrade:GetShipmentBonus()
end

function AddTownUpgradeLevel(value)
    SetTownUpgradeLevel(value + GetTownUpgradeLevel())
end

data.town_upgrade.GetUpgradeLevel = GetTownUpgradeLevel

-- GEM VALUE WAVE (for gem > gold exchange)
function SetGemValueWave(value)
    DCEI.SetProtectedValue(data.safe.gem_value_wave, value)
end

function GetGemValueWave()
    return math.floor(DCEI.GetProtectedValue(data.safe.gem_value_wave) or 0)
end

function GetGoldExchangeValue()
    local gem_value_wave = GetGemValueWave()
    local multiplier = 10

    return math.floor(wave.data.GetGoldValue(gem_value_wave) * 10)
end

-- Hell Mode
function SetHellModeScore(value)
    DCEI.SetProtectedValue(data.safe.hell_mode_score, value)
end

function GetHellModeScore()
    return math.floor(DCEI.GetProtectedValue(data.safe.hell_mode_score))
end

function AddHellModeScore(value)
    SetHellModeScore(value + GetHellModeScore())
end

-- Philosopher stone
function SetPhilosopherStone(value)
    DCEI.SetProtectedValue(data.safe.philosopher_stone, value)
end

function GetPhilosopherStone()
    return math.floor(DCEI.GetProtectedValue(data.safe.philosopher_stone))
end

function AddPhilosopherStone(value)
    SetPhilosopherStone(value + GetPhilosopherStone())
end

-- raid turbo
function SetRaidTurbo(value)
    DCEI.SetProtectedValue(data.safe.raid_turbo, value)
end

function GetRaidTurbo()
    return math.floor(DCEI.GetProtectedValue(data.safe.raid_turbo))
end

function AddRaidTurbo(value)
    SetRaidTurbo(value + GetRaidTurbo())
end

-- hell coin
function SetHellCoin(value)
    DCEI.SetProtectedValue(data.safe.hell_coin, value)
end

function GetHellCoin()
    return math.floor(DCEI.GetProtectedValue(data.safe.hell_coin))
end

function AddHellCoin(value)
    SetHellCoin(value + GetHellCoin())
end

-- hell mode free count
function SetHellModeFreeCount(value)
    DCEI.SetProtectedValue(data.safe.hell_mode_free_count, value)
end

function GetHellModeFreeCount()
    return math.floor(DCEI.GetProtectedValue(data.safe.hell_mode_free_count))
end

function AddHellModeFreeCount(value)
    SetHellModeFreeCount(value + GetHellModeFreeCount())
end

-- raid key free count
function SetRaidKeyDailyCount(value)
    DCEI.SetProtectedValue(data.safe.raid_key_daily, math.min(value, RAID_KEY_DAILY_COUNT))
end

function GetRaidKeyDailyCount()
    return math.floor(DCEI.GetProtectedValue(data.safe.raid_key_daily))
end

function AddRaidKeyDailyCount(value)
    SetRaidKeyDailyCount(value + GetRaidKeyDailyCount())
end

-- raid key extra count
function SetRaidKeyExtraCount(value)
    DCEI.SetProtectedValue(data.safe.raid_key_extra, value)
end

function GetRaidKeyExtraCount()
    return math.floor(DCEI.GetProtectedValue(data.safe.raid_key_extra))
end

function AddRaidKeyExtraCount(value)
    SetRaidKeyExtraCount(value + GetRaidKeyExtraCount())
end

-- !SECTION

-- ===================================================================================================================================================================================================================
-- SECTION HERO SKILL SYSTEM
-- ===================================================================================================================================================================================================================

function AttemptHeroCastSkill(unit)
    local name = DCEI.GetUnitType(unit)
    local hero_data = data.heroes[name]
    local report_error = true

    if hero_data.stun then
        return
    end

    if HeroCanUseSkill(hero_data, report_error) then
        HeroCastSkill(unit, hero_data)
    end
end

function GetHeroSkillTargetPosition(unit, pos)
    if pos.min_x and pos.max_x then
        -- bounds hero x position between min and max
        local unit_pos = DCEI.GetUnitPosition2D(unit)
        local x = unit_pos.x
        x = math.min(x, pos.max_x)
        x = math.max(x, pos.min_x)
        pos.x = x
    end

    return pos
end

function HeroCastSkill(unit, hero_data)
    local ability = data.GetHeroActiveSkillData(hero_data, "ability")
    local cost = hero_data:GetAbilityCost()

    -- DCEI.LogMessage("CAST " .. DCEI.GetUnitType(unit) .. " ability: " .. ability .. "!")
    wave.data.skills = wave.data.skills + 1

    local target_manual = data.GetHeroActiveSkillData(hero_data, "target_manual")
    local target_monster = data.GetHeroActiveSkillData(hero_data, "target_monster")
    local target_position = data.GetHeroActiveSkillData(hero_data, "target_pos")
    local event_hero_autobattle = event_mode.HeroIsInAutoBattleList(hero_data)

    if event_mode.IsActive() then
        if target_manual and not event_hero_autobattle then
            -- casts spells with manual targeting if not in auto battle
            EnterTargetMode(unit, hero_data)
        else
            if target_monster or target_manual then
                -- cast spells based on closest monster position
                local pos = GetSpellTargetPosition()
                DCEI.CastAbilityAtPosition(ability, unit, pos.x, pos.y)
            elseif target_position then
                -- cast spells with explicit target position
                local pos = GetHeroSkillTargetPosition(unit, target_position)
                DCEI.CastAbilityAtPosition(ability, unit, pos.x, pos.y)
            else
                -- cast spells with no target
                DCEI.CastAbilityAtUnit(ability, unit, unit)
            end

            -- mana and cooldown are not triggered by entering target mode
            DCEI.AddManaAsync(CASTLE_UNIT, -1 * cost)
            StartHeroSkillCooldown(unit, hero_data)
        end
    else
        if target_manual and not player.autobattle.is_active then
            -- casts spells with manual targeting if not in auto battle
            EnterTargetMode(unit, hero_data)
        else
            if target_monster or target_manual then
                -- cast spells based on closest monster position
                local pos = GetSpellTargetPosition()
                DCEI.CastAbilityAtPosition(ability, unit, pos.x, pos.y)
            elseif target_position then
                -- cast spells with explicit target position
                local pos = GetHeroSkillTargetPosition(unit, target_position)
                DCEI.CastAbilityAtPosition(ability, unit, pos.x, pos.y)
            else
                -- cast spells with no target
                DCEI.CastAbilityAtUnit(ability, unit, unit)
            end

            -- mana and cooldown are not triggered by entering target mode
            DCEI.AddManaAsync(CASTLE_UNIT, -1 * cost)
            StartHeroSkillCooldown(unit, hero_data)
        end
    end
end

function EnterTargetMode(unit, hero_data)
    local skill_name = data.GetHeroActiveSkillData(hero_data, "ability")
    local target_skill = "Shared ManualTarget"

    ui.loadout.hero.Hide()
    ui.screen_mask.world:Show()
    ui.prompt.manual_cast:Show()

    -- show "cancel" battle command
    DCEI.SetFrameActive(ui.bottom_panel.middle.frame, true)

    player.targeting_data = hero_data
    DCEI.LogMessage("Enter Target: " .. skill_name)

    DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("System Hero ManualTarget"), 1)
    DCEI.EnterTargetingMode(CASTLE_UNIT, target_skill)
    SetGameSpeed(speed.slow)
end

function CastTargetModeAbility(pos)
    local unit_data = player.targeting_data

    if unit_data then
        local slot = unit_data.slot
        local unit = player.heroes[slot]
        local cost = unit_data:GetAbilityCost()
        local skill_name = data.GetHeroActiveSkillData(unit_data, "ability")

        -- adjust cast position within bounds
        pos = pos or { x = 24, y = 24 }
        pos.x = 24
        pos.y = math.max(pos.y, 21)
        pos.y = math.min(pos.y, 34)

        -- cast ability and exit targeting mode
        ExitTargetMode()
        DCEI.CastAbilityAtPosition(skill_name, unit, pos.x, pos.y)

        -- spend resources, start cooldown
        DCEI.AddManaAsync(CASTLE_UNIT, -1 * cost)
        StartHeroSkillCooldown(unit, unit_data)
    end
end

function PlayerInTargetMode()
    return player.targeting_data
end

function CheckExitTargetMode()
    if PlayerInTargetMode() then
        ExitTargetMode()
    end
end

function ExitTargetMode()
    local unit_data = player.targeting_data
    local slot = unit_data.slot
    local unit = player.heroes[slot]

    ui.screen_mask.Hide()
    ui.prompt.manual_cast:Hide()

    -- hide "cancel" battle command
    DCEI.SetFrameActive(ui.bottom_panel.middle.frame, false)

    SetGameSpeed(player.speed)
    player.targeting_data = nil
    DCEI.ExitTargetingMode()

    -- re-enable hero selection if wave is still active
    if wave.data.is_active then
        ui.loadout.hero.Show()
        ui.loadout.hero.HideBorder()
    end

    -- remove manual cast behavior on delay to ensure ability executes first
    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        if unit then
            DCEI.RemoveBehaviorCountAsync(unit, "System Hero ManualTarget", 1)
        end
    end, 0.125, false, simple)
end

function GetSpellTargetPosition()
    local sum = 0
    local x = 24
    local min, max = 18, 30

    if #wave.data.units == 0 then
        return { x = x, y = x }
    end

    for _, unit in pairs(wave.data.units) do
        local pos = DCEI.GetUnitPosition2D(unit)
        sum = sum + pos.y
    end

    local y = (sum / #wave.data.units) - 2
    y = math.min(y, max)
    y = math.max(y, min)

    -- DCEI.LogMessage("> best target is " .. x .. ", " .. y)
    return { x = x, y = y }
end

-- !SECTION

-- SECTION Purchase and equip hero
function ApplyBaseBehaviorsToUnit(unit_data, unit)
    local level = unit_data:GetLevel()

    -- apply unit attack damage upgrades
    if unit_data.base_damage then
        local attack_power = DCEI.Behavior("Attribute AttackPower")
        local base = unit_data.base_damage * 10
        local growth = 0.1

        local current_count = DCEI.GetUnitBehaviorStackCount(unit, attack_power)
        local count = base + (growth * base * (level - 1))

        DCEI.RemoveBehaviorCountAsync(unit, attack_power, current_count)
        DCEI.ApplyBehaviorCountAsync(unit, attack_power, count)
    end

    -- apply unit skill upgrades
    local upgrades = unit_data.upgrade
    if upgrades then
        for k, upgrade in ipairs(upgrades) do
            local current_count = DCEI.GetUnitBehaviorStackCount(unit, upgrade)
            local count = level - 1

            if current_count > 0 then
                DCEI.RemoveBehaviorCountAsync(unit, upgrade, current_count)
            end
            if count > 0 then
                DCEI.ApplyBehaviorCountAsync(unit, upgrade, count)
            end
        end
    end
end

function PurchaseHero(hero_data)
    hero_data.is_purchased = true
    if event_mode.IsActive() then
        bank.CommitEventUnitData(hero_data, bank.bank_data.prop.level, 1)
        bank.CommitEventUnitData(hero_data, bank.bank_data.prop.class, hero_data.name, true)
    else
        bank.CommitUnitData(hero_data, bank.bank_data.prop.level, 1)
        bank.CommitUnitData(hero_data, bank.bank_data.prop.class, hero_data.name, true)
        LogUserProperty(UserProps.heroes_owned)
    end 
end

function UpgradeHero(hero_data)
    local unit = EquippedHero(hero_data)

    if hero_data.slot > 0 then
        ProcessSkillEquipBonuses(hero_data, false)
    end
    hero_data:LevelUp()
    if event_mode.IsActive() then
        bank.CommitEventUnitData(hero_data, bank.bank_data.prop.level, hero_data:GetLevel())
    else
        bank.CommitUnitData(hero_data, bank.bank_data.prop.level, hero_data:GetLevel())
    end

    if hero_data.slot > 0 then
        ProcessSkillEquipBonuses(hero_data, true)
    end

    if unit and DCEI.UnitExists(unit) then
        ApplyBaseBehaviorsToUnit(hero_data, unit)
    end

    if ui.prompt.tutorial then
        -- close tutorial and show promote button on hero level-up tutorial completion
        ui.prompt.tutorial.Close()
        ui.GenerateUnitCardPromoteButton()
        event_mode.GenerateUnitCardAutoBattleButton()
        if not ui.list_promote and not ui.event_mode.IsActive() then
            ui.GenerateUnitCardEquipmentSlots(ui.unit_card, nil, UnlockEquipment())
        end
    end
end

function EquippedHero(hero_data)
    local slot = hero_data.slot
    if slot and slot > 0 then
        return player.heroes[slot]
    end
end

function EquipHero(hero_data, slot, skip_commit)
    slot = slot or ui.loadout.hero.select
    local unit_type = hero_data.name

    -- unequip hero currently in the slot, if any
    UnequipHeroSlot(slot)

    -- unequip hero if currently equipped in another slot
    UnequipHero(hero_data)

    DCEI.LogMessage("> equipping " .. unit_type .. " to slot " .. slot)
    local dx, dy = 0, 1
    local pos = DCEI.GetUnitPosition2D(group_hero_slots[slot])
    local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, dx, dy)
    player.heroes[slot] = unit
    hero_data.slot = slot
    hero_data.unit = unit

    if not skip_commit then
        if event_mode.IsActive() then
            bank.CommitEventUnitSlot(hero_data, slot)
        else
            bank.CommitUnitSlot(hero_data, slot)
        end
    end

    ApplyUpgradeDataToUnit(data.hero_upgrades, unit)
    ApplyBaseBehaviorsToUnit(hero_data, unit)
    ProcessSkillEquipBonuses(hero_data, unit)
    ProcessEquipmentBehaviorToHero(hero_data, unit, true)
    UpdateHeroList()
    RefreshHeroesEquipBehaviors()
    UpdateBlessBonusUnits(hero_data, true)

    if ui.list_hero then
        local unit_index = GetHeroGalleryIndex(hero_data)
        ui.list_hero.Highlight(unit_index)
    end

    -- wait after next frame for heroes behaviors refresh
    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        if ui.unit_card then
            ui.UpdateUnitCard()
        end
    end, 0.125, false, simple)
end

-- !SECTION

-- SECTION Talents
function ProcessTalentBehaviors(talent_data, from_bank)
    local count = from_bank and talent_data.level or 1

    if not talent_data.behavior or count == 0 then
        return
    end

    local behavior = talent_data.behavior.name
    local target = talent_data.behavior.target

    AddBehaviorUpgrade(target, behavior, count)
end

function ResetTalentBehaviors(talent_data)
    if not talent_data.behavior or talent_data.level == 0 then
        return
    end
    local behavior = talent_data.behavior.name
    local target = talent_data.behavior.target

    AddBehaviorUpgrade(target, behavior, -talent_data.level)
end

function AddBehaviorUpgrade(target, behavior, count)
    if target == "castle" then
        AddCastleUpgrade(behavior, count)
    elseif target == "castle_wall" then
        AddCastleWallUpgrade(behavior, count)
    elseif target == "archer" then
        AddArcherUpgrade(behavior, count)
    elseif target == "summon" then
        AddSummonUpgrade(behavior, count)
    elseif target == "hero" then
        AddHeroUpgrade(behavior, count)
    elseif target == "tower" then
        AddTowerUpgrade(behavior, count)
    elseif target == "monster" then
        AddMonsterUpgrade(behavior, count)
    elseif target == "all" then
        local source = target

        -- source is tracked so "all" upgrades aren't double applied to summon/archer upgrade units
        AddArcherUpgrade(behavior, count, source)
        AddSummonUpgrade(behavior, count, source)
        AddTowerUpgrade(behavior, count, source)
        AddHeroUpgrade(behavior, count, source)
    end
end

function ProcessTalentAttributes(talent_data, from_bank)
    local count = from_bank and talent_data.level or 1

    if not talent_data.attributes or count == 0 then
        return
    end

    for _, attribute in ipairs(talent_data.attributes) do
        local value = talent_data.value.growth * count
        local name = attribute.name

        player.attributes[name] = player.attributes[name] + value

        if name == "colony_cooldown" and not from_bank then
            UpdateColonyCooldown()
        end
    end
end

function ResetTalentAttributes(talent_data)
    if not talent_data.attributes or talent_data.level == 0 then
        return
    end

    for _, attribute in ipairs(talent_data.attributes) do
        local value = talent_data.value.growth * talent_data.level
        local name = attribute.name

        player.attributes[name] = player.attributes[name] - value

        if name == "colony_cooldown" then
            UpdateColonyCooldown()
        end
    end
end

-- !SECTION

-- SECTION Functions
function UpdateColonyCooldown()
    -- update colony list UI
    if ui.colony then
        ui.colony.Update()
    end

    -- update colony timer interval
    if levels.colony.timer then
        levels.colony.timer.interval = levels.colony:Cooldown()
    end
end

function UnequipHeroSlot(slot)
    local unit = player.heroes[slot]

    if not (unit and DCEI.UnitExists(unit)) then
        return
    end

    local unit_type = DCEI.GetUnitType(unit)
    local hero_data = GetHeroDataFromUnitType(unit_type)

    UnequipHero(hero_data)
end

function UnequipHero(hero_data)
    local unit_type = hero_data.name
    local slot = hero_data.slot
    local unit = player.heroes[slot]

    if unit and DCEI.UnitExists(unit) then
        DCEI.LogMessage("> unequipping " .. unit_type .. " from slot " .. slot)
        if event_mode.IsActive() then
            bank.ClearEventUnitSlot(bank.bank_data.units.hero, slot)
        else
            bank.ClearUnitSlot(bank.bank_data.units.hero, slot)
        end

        hero_data.slot = 0
        hero_data.unit = nil
        player.heroes[slot] = nil
        DCEI.RemoveUnit(unit)
        ProcessSkillEquipBonuses(hero_data, false)
        -- for legendary equipment gear effect, we need to unload the effect when unequip the hero
        ProcessEquipmentBehaviorToHero(hero_data, unit, false)
        PrcessHeroEquipBebaviors(unit, hero_data, false)
        RefreshHeroesEquipBehaviors()
        UpdateBlessBonusUnits(hero_data, false)
    end

    -- update UI
    UpdateHeroList()

    if ui.list_hero then
        ui.list_hero.Highlight(0)
    end

    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        if ui.unit_card then
            ui.UpdateUnitCard()
        end
    end, 0, false, simple)
end

-- HERO SLOTS
function CreateHeroSlots()
    local dx, dy = 1, -1
    local name = DCEI.Unit("Castle Hero Slot")

    local mid_x, mid_y = 24, 16.5
    local x_offset, y_offset = 1.1875, 1.625
    local x_count, y_count = 6, 2

    local height = y_offset * (y_count - 1)
    local width = x_offset * (x_count - 1)

    local start_x = mid_x - (width / 2)
    local start_y = mid_y + (height / 2)

    local points = {}
    for y = start_y, start_y - ((y_count - 1) * y_offset), -y_offset do
        for x = start_x, start_x + ((x_count - 1) * x_offset), x_offset do
            table.insert(points, { x = x, y = y })
        end
    end

    -- sort table based on unlock order
    points = ui.OrganizeHeroLoadoutTable(points)

    -- adjust points so the attached UI frames are aligned
    local adjustments = {
            { index = 11, x = 0.1 },
            { index = 7, x = 0.06 },
            { index = 5, x = 0.02 },
            { index = 6, x = -0.02 },
            { index = 8, x = -0.06 },
            { index = 12, x = -0.1 },
        }
    for k, v in ipairs(adjustments) do
        points[v.index] = AdjustPoint(points[v.index], v.x, 0.04)
    end

    -- actually create hero slot units in correct position and order
    for k, point in ipairs(points) do
        local slot = DCEI.CreateUnitAsync(1, 1, name, point.x, point.y, dx, dy)
        table.insert(group_hero_slots, slot)
    end
end

function AdjustPoint(point, x_offset, y_offset)
    point.x = point.x + x_offset
    point.y = point.y + y_offset
    return point
end

function CreateDefenseSlots()
    local name = DCEI.Unit("Castle Defense Slot")
    local points = {
        { x = 21.15, y = 14.1 },
        { x = 26.85, y = 14.1 },
    }

    for _, point in ipairs(points) do
        local slot = DCEI.CreateUnitAsync(1, 1, name, point.x, point.y, 0, -1)
        table.insert(group_defense_slots, slot)
    end
end

function CreateTowerSlots()
    local name = DCEI.Unit("Castle Tower Slot")
    local points = {
        { x = 20.72, y = 20.5 },
        { x = 27.28, y = 20.5 },
        { x = 20.6, y = 22.35 },
        { x = 27.4, y = 22.35 },
        -- 5th and 6th
        { x = 22, y = 22 },
        { x = 26.1, y = 22 },
    }

    for _, point in ipairs(points) do
        local slot = DCEI.CreateUnitAsync(1, 1, name, point.x, point.y, 0, 1)
        table.insert(group_tower_slots, slot)
    end
end

function CreateGearAnchor()
    local name = DCEI.Unit("Castle Gear Anchor")
    local point = { x = 24, y = 20 }
    gear_anchor = DCEI.CreateUnitAsync(1, 1, name, point.x, point.y, 0, 1)
end

-- TOWN SLOTS
function CreateTownSlots()
    local hide = DCEI.Behavior("System Hide")
    local dx, dy = 0, 1
    local x, y = { 22, 26, 21.9, 26.1 }, { 5.8, 5.8, 3.1, 3.1 }

    for n = 1, #x do
        local unit = DCEI.CreateUnitAsync(1, 1, TOWN_SLOT_NAME, x[n], y[n], dx, dy)
        table.insert(group_town_slots, unit)
        DCEI.ApplyBehaviorAsync(unit, hide)
    end
end

function GetUnitGroupIndex(unit, group)
    for index, group_unit in ipairs(group) do
        if unit == group_unit then
            return index
        end
    end
end

function AddPlayerXP(value)
    local max_level = data.GetPlayerLevelMax()

    -- don't check for level up if player is max level
    if GetPlayerLevel() == max_level then
        return AddParagonXP(value)
    end

    player.player_xp = player.player_xp + value
    -- DCEI.LogMessage("gain" .. value .. " xp")

    local xp_to_level = data.GetPlayerLevelXP(GetPlayerLevel() + 1)
    if player.player_xp >= xp_to_level then
        while player.player_xp >= xp_to_level do
            AddTalentPoints(1)
            AddPlayerLevel(1)
            ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())
            xp_to_level = data.GetPlayerLevelXP(GetPlayerLevel() + 1)
        end
        if event_mode.IsActive() then
            bank.CommitEventTalentPoints()
            bank.CommitEventPlayerLevel()
        else
            bank.CommitTalentPoints()
            bank.CommitPlayerLevel()
        end
        ui.CreateLevelUpMessage()
    end

    UpdateExperienceBar()
end

function AddParagonXP(value)
    player.paragon_xp = player.paragon_xp + value

    local paragon_level = GetParagonLevel()
    data.paragon_level_xp = data.paragon_level_xp or data.UpdateParagonLevelXP(paragon_level)
    if player.paragon_xp >= data.paragon_level_xp then
        while player.paragon_xp >= data.paragon_level_xp do
            player.paragon_xp = player.paragon_xp - data.paragon_level_xp
            AddParagonPoints(1)
            AddParagonLevel(1)
            ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())
            data.paragon_level_xp = data.UpdateParagonLevelXP(paragon_level)
            -- show red pip when first get paragon level
            if GetParagonLevel() == 1 then
                ui.glue.talent_menu:ShowPipAnimated()
            end
        end

        if event_mode.IsActive() then
            bank.CommitEventParagonPoints()
            bank.CommitEventParagonLevel()
        else
            bank.CommitParagonPoints()
            bank.CommitParagonLevel()
        end
        
        local paragon = true
        ui.CreateLevelUpMessage(paragon)
    end

    UpdateExperienceBar()
end

function AttemptResetTalent()
    local cost = 5
    local gems = GetPlayerGems()

    if gems < cost then
        local function OnStore()
            ui.talent_menu:Close()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)

        return false
    end

    return true
end

function GetParagonTalenUpgradeCost(talent)
    local upgrade_cost = 0
    if talent.cost_curve then
        local upper_threshold, upper_index
        for index, v in ipairs(talent.cost_curve) do
            if talent.level >= v.start_level then
                upper_index = index
                upper_threshold = v.start_level
                break
            end
        end
        upgrade_cost = upgrade_cost + (talent.level - upper_threshold) * talent.cost_curve[upper_index].cost
        for i = upper_index + 1, #talent.cost_curve do
            upgrade_cost = upgrade_cost
                + (talent.cost_curve[i - 1].start_level - talent.cost_curve[i].start_level)
                    * talent.cost_curve[i].cost
        end
    else
        upgrade_cost = talent.level
    end

    return upgrade_cost
end

function ResetNormalTalentPoints()
    for _, talent in ipairs(data.talents) do
        if not talent.paragon then
            ResetTalentAttributes(talent)
            ResetTalentBehaviors(talent)
            talent.level = 0
            if event_mode.IsActive() then
                bank.CommitEventTalent(talent)
            else
                bank.CommitTalent(talent)
            end
            
        end
    end
    SetTalentPoints(GetPlayerLevel() - 1)
    ui.talent_menu.UpdatePoints(GetTalentPoints(), GetParagonPoints())
    ui.talent_menu.DeselectAll()
    ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())

    for _, talent_layout in ipairs(ui.talent_menu.talents) do
        talent_layout.Update()
    end

    if event_mode.IsActive() then
        bank.CommitEventTalentPoints()
    else
        bank.CommitTalentPoints()
    end
    

    local cost = 5
    local spend_event = { type = "reset", target = "talent" }
    SpendPlayerGems(cost, spend_event)
    PlayGoldSpendSound()

    DCEI.SetFrameActive(ui.talent_menu.reset_paragon.frame, false)
end

function ResetParagonPoints()
    for _, talent in ipairs(data.talents) do
        if talent.paragon then
            ResetTalentAttributes(talent)
            ResetTalentBehaviors(talent)
            talent.level = 0
            
            if event_mode.IsActive() then
                bank.CommitEventTalent(talent)
            else
                bank.CommitTalent(talent)
            end
        end
    end
    SetParagonPoints(GetParagonLevel())
    ui.talent_menu.UpdatePoints(GetTalentPoints(), GetParagonPoints())
    ui.talent_menu.DeselectAll()
    ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())

    for _, talent_layout in ipairs(ui.talent_menu.talents) do
        talent_layout.Update()
    end

    if event_mode.IsActive() then
        bank.CommitEventParagonPoints()
    else
        bank.CommitParagonPoints()
    end

    local cost = 5
    local spend_event = { type = "reset", target = "paragon" }
    SpendPlayerGems(cost, spend_event)
    PlayGoldSpendSound()

    DCEI.SetFrameActive(ui.talent_menu.reset_paragon.frame, false)
end

-- !SECTION

-- SECTION STORE Functions
function InitializeStore()
    local store_data = {
        {
            name = "deals",
            display_name = DCEI.GetTranslationText("shop/section/deals"),
            custom = {
                layout = "Portrait/Shop/SectionDeals",
                layout_func = CustomizeStoreSection_Deals,
            },
        },
        {
            name = "winter_pass",
            display_name = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro"),
            rows = {
                {
                    "com.wildsky.castle.iap.winterpass.01",
                },
            },
        },
        {
            name = "autumn_pass",
            display_name = DCEI.GetTranslationText("ui/menu/autumn_event/title"),
            rows = {
                {
                    "com.wildsky.castle.iap.autumnpass.01",
                },
            },
        },
        {
            name = "gold",
            display_name = DCEI.GetTranslationText("shop/section/gold"),
            rows = {
                {
                    "com.wildsky.castle.iap.gold.01",
                    "com.wildsky.castle.iap.gold.02",
                    "com.wildsky.castle.iap.gold.03",
                },
                {
                    "com.wildsky.castle.iap.gold.04",
                    "com.wildsky.castle.iap.gold.05",
                    "com.wildsky.castle.iap.gold.06",
                },
            },
            navigation = {
                icon = "icon_coin00",
            },
        },
        {
            name = "tokens",
            display_name = DCEI.GetTranslationText("shop/section/tokens"),
            rows = {
                {
                    "com.wildsky.castle.iap.token.new.01",
                    "com.wildsky.castle.iap.token.01",
                    "com.wildsky.castle.iap.token.02",
                },
                {
                    "com.wildsky.castle.iap.token.03",
                    "com.wildsky.castle.iap.token.new.04",
                    "com.wildsky.castle.iap.token.04",
                },
            },
            navigation = {
                icon = "icon_coin_dungeon00",
            },
        },
        {
            name = "chests",
            display_name = DCEI.GetTranslationText("shop/section/equipment"),
            rows = {
                {
                    "com.wildsky.castle.iap.chests.01",
                    "com.wildsky.castle.iap.chests.02",
                },
                {
                    "com.wildsky.castle.iap.chests.03",
                    "com.wildsky.castle.iap.chests.04",
                },
            },
            navigation = {
                icon = "premium_chest",
            },
            custom = {
                layout = "Portrait/Shop/SectionChests",
            },
        },
        {
            name = "auto_battle",
            display_name = DCEI.GetTranslationText("shop/section/auto_battle"),
            rows = {
                {
                    "com.wildsky.castle.iap.autobattle.new.01",
                    "com.wildsky.castle.iap.autobattle.new.02",
                },
                {
                    "com.wildsky.castle.iap.autobattle.new.03",
                    "com.wildsky.castle.iap.autobattle.new.04",
                },
            },
            navigation = {
                icon = "icon_sword",
            },
        },
        {
            name = "speed_up",
            display_name = DCEI.GetTranslationText("shop/section/speed_up"),
            rows = {
                {
                    "com.wildsky.castle.iap.speed.01",
                    "com.wildsky.castle.iap.speed.02",
                },
                {
                    "com.wildsky.castle.iap.speed.03",
                    "com.wildsky.castle.iap.speed.04",
                },
            },
        },
        {
            name = "hell_coin",
            display_name = DCEI.GetTranslationText("shop/section/hell_coin_shop"),
            -- rows = {
            --     {
            --         "com.wildsky.castle.hell_coin.badge.01",
            --         "com.wildsky.castle.hell_coin.philosopher_stone.01",
            --         "com.wildsky.castle.hell_coin.raid_turbo.01",
            --     }
            -- },
            custom = {
                layout_func = CustomizeStoreSection_HellCoinShop,
            },
        },
    }

    -- define IAP List and initialize shop mod
    -- hook up badges
    for index, item in ipairs(data.badges) do
        local new_badge_product = {
            product_id = "com.wildsky.castle.hell_coin.badge.01",
            type = "hell_coin_badge",
            label = DCEI.GetTranslationText("data/item/badge"),
            display = {
                image = DCEI.Texture("icon_trophy_mini_1"),
                scale = 1.2,
                use_stock_display = true,
            },
            price = {
                type = "hell_coin",
                amount = 100,
            },
            stock_limit = 1,
            items = {
                badge_id = 1,
            },
            info = {
                title = DCEI.GetTranslationText("data/item/badge"),
                label = DCEI.GetTranslationText("data/item/badge/description"),
            },
            custom = {
                layout_func = CustomizeStoreItem_HellCoinShop_Badge,
                update_func = UpdateStoreHellCoinDisplay,
            },
        }
        new_badge_product.product_id = item.product_id
        new_badge_product.display.image = item.icon
        new_badge_product.items.badge_id = item.badge_id
        table.insert(store_product_list, new_badge_product)
    end

    shop.InitializeIAP({
        on_purchase_attempt_callback = AttemptAnyItemPurchase,
        on_purchase_success_callback = OnAnyItemPurchase,
        product_list = store_product_list,
        currency_list = CURRENCIES,
        is_game_localized = true,
    })
    -- initialize UI
    ui.InitializeShopUI(shop)
    ui.popup.OnShowAnimation = gmui.Animations.BackIn
    function ui.shop.info_popup:OnReset()
        ResetInfoPopup(self)
    end

    -- initialize store data & deals page
    ui.shop:Initialize(store_data)
    InitializeDealsPage()
    InitializeHellCoinPage()

    for _, section in pairs(ui.shop.sections) do
        local layout = section.nav
        if layout then
            layout.ButtonFill = DCEI.GetChildFrameById(layout.Frame, "ButtonFill")

            function layout:Highlight()
                DCEI.SetFrameImageColor(self.ButtonFill, { r = 1, g = 1, b = 1, a = 1 })
            end

            function layout:ClearHighlight()
                DCEI.SetFrameImageColor(self.ButtonFill, { r = 1, g = 1, b = 1, a = 0 })
            end
        end
    end

    local layout = ui.shop.sections["auto_battle"].nav
    layout.Div = DCEI.GetChildFrameById(layout.Frame, "Div")
    layout.Icon2 = DCEI.GetChildFrameById(layout.Frame, "Icon2")
    DCEI.SetFrameActive(layout.Div, true)
    DCEI.SetFrameActive(layout.Icon2, true)

    local layout = ui.shop.sections["chests"].nav
    layout.icon = DCEI.GetChildFrameById(layout.Frame, "Icon")
    layout.stack = DCEI.GetChildFrameById(layout.Frame, "HStack")
    DCEI.SetFrameSize(layout.icon, 72, 72)
    DCEI.SetFrameVerticalOffsetInParent(layout.stack, 10)

    DCEI.SetOnClickCallback(ui.shop.corner_nav_button, function()
        ui.shop.sections["deals"]:ScrollTo()
    end)

    ui.shop:AddTimerUpdateFunction("DisplayCornerNav", function()
        local offset = DCEI.GetScrollFrameScrollPosition(ui.shop.VScroll)
        local buffer = 276

        if ui.shop.is_scroll_clean then
            local section = ui.shop.sections["gold"]

            -- highlight section if scroll offset within section view
            if section.scroll_offset and offset >= (section.scroll_offset - buffer) then
                ui.shop:ShowCornerNavs()
            else
                ui.shop:HideCornerNavs()
            end
        end
    end)

    local options = { init = true }

    ui.shop.sections["winter_pass"]:Hide()
    ui.shop.sections["autumn_pass"]:Hide()
end

-- SECTION STORE Functions
function InitializeEventStore()
    local store_data = {
        {
            name = "deals",
            display_name = DCEI.GetTranslationText("shop/section/deals"),
            custom = {
                layout = "Portrait/Shop/SectionDeals",
                layout_func = CustomizeStoreSection_Deals,
            },
        },
        {
            name = "gold",
            display_name = "Mithril",
            rows = {
                {
                    "com.wildsky.castle.iap.gold.01",
                    "com.wildsky.castle.iap.gold.02",
                    "com.wildsky.castle.iap.gold.03",
                },
                {
                    "com.wildsky.castle.iap.gold.04",
                    "com.wildsky.castle.iap.gold.05",
                    "com.wildsky.castle.iap.gold.06",
                },
            },
            navigation = {
                icon = "icon_token_awaken01",
            },
        },
        {
            name = "potions",
            display_name = "Power Potions",
            rows = {
                {
                    "com.wildsky.castle.iap.token.new.01",
                    "com.wildsky.castle.iap.token.01",
                    "com.wildsky.castle.iap.token.02",
                },
                {
                    "com.wildsky.castle.iap.token.03",
                    "com.wildsky.castle.iap.token.new.04",
                    "com.wildsky.castle.iap.token.04",
                },
            },
            navigation = {
                icon = DCEI.Texture("icon_catalyst_main_item5_orange_03"),
            },
        },
        {
            name = "chests",
            display_name = "Hero Chests",
            rows = {
                {
                    "com.wildsky.castle.iap.chests.01",
                    "com.wildsky.castle.iap.chests.02",
                },
                {
                    "com.wildsky.castle.iap.chests.03",
                    "com.wildsky.castle.iap.chests.04",
                },
            },
            navigation = {
                icon = "nature_chest",
            },
        },
    }

    shop.InitializeIAP({
        on_purchase_attempt_callback = AttemptAnyItemPurchase,
        on_purchase_success_callback = OnAnyItemPurchase,
        product_list = store_product_list_event,
        currency_list = EVENT_CURRENCIES,
        is_game_localized = true,
    })
    -- initialize UI
    ui.InitializeShopUI(shop)
    ui.popup.OnShowAnimation = gmui.Animations.BackIn
    function ui.shop.info_popup:OnReset()
        ResetInfoPopup(self)
    end

    -- initialize store data & deals page
    ui.shop:Initialize(store_data)
    InitializeDealsPage()

    for _, section in pairs(ui.shop.sections) do
        local layout = section.nav
        if layout then
            layout.ButtonFill = DCEI.GetChildFrameById(layout.Frame, "ButtonFill")

            function layout:Highlight()
                DCEI.SetFrameImageColor(self.ButtonFill, { r = 1, g = 1, b = 1, a = 1 })
            end

            function layout:ClearHighlight()
                DCEI.SetFrameImageColor(self.ButtonFill, { r = 1, g = 1, b = 1, a = 0 })
            end
        end
    end

    DCEI.SetOnClickCallback(ui.shop.corner_nav_button, function()
        ui.shop.sections["deals"]:ScrollTo()
    end)

    ui.shop:AddTimerUpdateFunction("DisplayCornerNav", function()
        local offset = DCEI.GetScrollFrameScrollPosition(ui.shop.VScroll)
        local buffer = 276

        if ui.shop.is_scroll_clean then
            local section = ui.shop.sections["gold"]

            -- highlight section if scroll offset within section view
            if section.scroll_offset and offset >= (section.scroll_offset - buffer) then
                ui.shop:ShowCornerNavs()
            else
                ui.shop:HideCornerNavs()
            end
        end
    end)

    local options = { init = true }
end

function AttemptAnyItemPurchase(product_id)
    DCEI.LogMessage("> Attempt Purchase " .. tostring(product_id))

    local custom_price_data = shop.GetProductCustomPriceData(product_id)
    if custom_price_data then
        local item_data = shop.GetProductData(product_id)
        if custom_price_data.type == "hell_coin" then
            local cost = custom_price_data.amount
            if cost <= GetHellCoin() then
                AddHellCoin(-cost)
                bank.CommitHellCoin()
                PurchaseHellCoinShopItem(item_data)
                -- stock update
                shop.OnAnyItemPurchase(product_id)
                PlayGoldSpendSound()
            else
                ui.ShowFeedbackNotEnoughHellCoin()
            end
        end
    else
        if PlayingEditorMode() then
            -- simulate sucessful purchase in editor
            shop.OnAnyItemPurchase(product_id)
        else
            -- otherwise purchase normally
            -- NOTE: anything placed below DCEI.PurchaseIapProduct() will take a few seconds to execute..
            --DCEI.PurchaseIapProduct(product_id)
			DCEI.TriggerAddMappedKeyDownEvent(product_id,shop.OnAnyItemPurchase)   
        end
    end
end

function OnAnyItemPurchase(product_id)
    local item_data = shop.GetProductData(product_id)
    local item_price = item_data:GetPriceData().usd or 0
    local item_type = item_data.type

    DCEI.LogMessage("> purchased [" .. item_type .. "] " .. product_id .. " for " .. string.format("%.2f", item_price))
    PlayGoldSpendSound()

    if item_type == "speedup" then
        PurchaseStoreItem_SpeedUp(item_data)
    elseif item_type == "autobattle" then
        PurchaseStoreItem_AutoBattle(item_data)
    elseif item_type == "token" then
        if event_mode.IsActive() then
            PurchaseStoreItem_Potion(item_data)
        else
            PurchaseStoreItem_Token(item_data)
        end
    elseif item_type == "gold" then
        PurchaseStoreItem_Gold(item_data)
    elseif item_type == "chests" then
        if event_mode.IsActive() then
            PurchaseStoreItem_EventChests(item_data)
        else
            PurchaseStoreItem_Chests(item_data)
        end
    elseif item_type == "bundles" then
        PurchaseStoreItem_Bundle(item_data)
    elseif item_type == "boosts" then
        PurchaseStoreItem_Boost(item_data)
    elseif item_type == "seasonpass" then
        PurchaseSeasonPass(item_data)
    elseif item_type == "winterpass" then
        PurchaseWinterPass(item_data)
    elseif item_type == "autumnpass" then
        PurchaseAutumnPass(item_data)
    elseif item_type == "potion" then
        PurchaseStoreItem_Potion(item_data)
    end

    -- log total number of purchases
    IncrementPurchaseCount()
    -- log total IAP spend
    IncrementPurchaseSpend(string.format("%.2f", item_price or 0))

    -- close the popup when an item is purchased from the popup
    local popup_layout = shop.GetItemLayoutFromPopup(product_id)
    if popup_layout then
        ui.popup:Hide()
    end
end

function PurchaseHellCoinShopItem(item_data)
    local items = item_data:GetItemsData()
    local item_name = ""
    if items.badge_id then
        local key = "a" .. items.badge_id
        player.hell_coin_shop.owned.owned_badges[key] = true
        item_name = "Badge" .. items.badge_id
        bank.CommitHellCoinShopData()
    elseif items.philosopher_stone then
        AddPlayerPhilosopherStone(items.philosopher_stone)
        item_name = "philosopher stone"
    elseif items.raid_turbo then
        AddPlayerRaidTurbo(items.raid_turbo)
        item_name = "raid turbo"
        ui.ShowFeedbackAddRaidTurboAndKeys(items.raid_turbo)
    end
    local event = LogEvents.BuyHellCoinShop(item_name)
    LogEvent(event)
end

function PurchaseSeasonPass(item_data)
    local type, duration = next(item_data:GetItemsData())

    DCEI.AnimateFrameScale(ui.calendar.seasonpass_button, { x = 1, y = 0.5 }, { x = 1, y = 1 }, 0.5, "OutBack")
    ui.calendar.preview_seasonpass_reward = false

    AddSeasonPassTime(duration)

    local event = LogEvents.BuyItem(item_data, {})
    LogEvent(event)
end

function PurchaseWinterPass(item_data)
    local items = item_data:GetItemsData()
    if items._1winterpass then
        player.winter_event2022.purchase_pass = true
    end

    if player.winter_event2022.purchase_pass then
        ui.shop.sections["winter_pass"]:Hide()
    end

    CheckWinterEventUnlock()
    bank.CommitSeasonEvent("winter_event2022")

    local event = LogEvents.BuyItem(item_data, {})
    LogEvent(event)
end

function PurchaseAutumnPass(item_data)
    local items = item_data:GetItemsData()
    if items._1autumnpass then
        player.autumn_event.purchase_pass = true
    end

    if player.autumn_event.purchase_pass then
        ui.shop.sections["autumn_pass"]:Hide()
    end

    CheckAutumnEventUnlock()
    bank.CommitSeasonEvent("autumn_event")

    local event = LogEvents.BuyItem(item_data, {})
    LogEvent(event)
end

function PurchaseStoreItem_Boost(item_data)
    local type, duration = next(item_data:GetItemsData())

    if not type then
        DCEI.LogError("> invalid timer type: " .. tostring(type) .. " for product " .. tostring(item_data.product_id))
    end

    if not duration then
        DCEI.LogError(
            "> invalid timer duration: " .. tostring(duration) .. " for product " .. tostring(item_data.product_id)
        )
    else
        -- initialize timer if necessary and add duration to timer time
        player.timers[type] = player.timers[type] or 0
        player.timers[type] = player.timers[type] + duration
    end

    bank.CommitTimers()

    local start_spend_log = true
    local event = LogEvents.BuyItem(item_data, {}, start_spend_log)
    LogEvent(event)
end

function PurchaseStoreItem_Bundle(item_data)
    local log_items = {}
    local items = item_data:GetItemsData()
    if items then
        if items._gold then
            -- apply product gold
            local value = GetStoreItemGold(items)
            log_items.gold = value
            AddPlayerGold(value)
            bank.CommitGold()
        end

        if items.token then
            -- apply product tokens
            local value = items.token
            log_items.tokens = string.format("%d", value)
            AddPlayerOverflowGems(value)
        end

        if items.autobattle then
            -- apply product autobattle
            local value = items.autobattle
            log_items.autobattle = string.format("%d", value)
            AddGemAutoBattleTime(value * 3600)
        end

        if items.mine_contract then
            -- apply product mine contracts
            local value = items.mine_contract
            log_items.mine_contract = string.format("%d", value)
            AddPlayerContract_Mine(value)
        end

        if items.legendary_chest then
            -- apply product chests
            local value = items.legendary_chest
            log_items.chests = string.format("%d", value)
            local name = "legendary_chest"
            AddPlayerChest(name, value)

            local guarantee_legendary = item_data:GetItemsData().guarantee_legendary
            if guarantee_legendary then
                player.guarantee_legendary[name] = player.guarantee_legendary[name] + guarantee_legendary
                bank.CommitGuaranteeEpic()
                player.legendary_interval[name] = math.floor(GetChestCount(name) / player.guarantee_legendary[name])
            end
        end

        if items.raid_turbo then
            -- apply product tokens
            local value = items.raid_turbo
            log_items.raid_turbo = string.format("%d", value)
            AddPlayerRaidTurbo(value)
        end

        if items.raid_key then
            -- apply product tokens
            local value = items.raid_key
            log_items.raid_key = string.format("%d", value)
            AddPlayerRaidKeyExtra(value)
        end

        if items._remove_ad then
            local items_list = GetStoreCollectionData("remove_ad")
            ui.shop:RemoveItems(items_list)
            bank.CommitRemoveAd()
        end
    end

    -- bundled items have multiple feedback types
    DCEI.TriggerAddTimerEventElapsed(function()
        if log_items.gold then
            ui.GoldRewardFeedback(log_items.gold, 40)
        end
        if log_items.tokens then
            ui.TokenRewardFeedback(log_items.tokens, -40)
        end
        if log_items.autobattle then
            ui.ShowFeedbackAddAutoBattle(log_items.autobattle)
        end
        if items.legendary_chest then
            ui.ChestsRewardFeedback(log_items.chests, "legendary_chest", -120)
        end
        if log_items.raid_turbo and items.raid_key then
            ui.ShowFeedbackAddRaidTurboAndKeys(log_items.raid_turbo, items.raid_key)
        end
    end, 0.125, true, true)

    -- start spend logging
    local start_spend_log = true
    local event = LogEvents.BuyItem(item_data, log_items, start_spend_log)
    LogEvent(event)
end

function PurchaseStoreItem_Gold(item_data)
    -- apply product
    local value = GetStoreItemGold(item_data:GetItemsData())
    local first_time = CheckFirstTimeBonus(item_data.product_id)
    if first_time then
        value = value + value
        bank.CommitFirstTimeBonus(item_data.product_id)
        UpdateStoreGoldDisplay(shop.GetItemLayout(item_data.product_id))
    end
    AddPlayerGold(value)
    bank.CommitGold()

    -- display feedback
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.GoldRewardFeedback(value)
    end, 0.125, true, true)

    -- start spend logging
    local items = { gold = value, first_time = first_time }
    local start_spend_log = true
    local event = LogEvents.BuyItem(item_data, items, start_spend_log)
    LogEvent(event)
end

function PurchaseStoreItem_Chests(item_data)
    -- apply product
    local value = item_data:GetItemsData().chests
    local name = item_data.chest_type
    AddPlayerChest(name, value)
    local guarantee_epic = item_data:GetItemsData().guarantee_epic
    player.guarantee_epic[name] = player.guarantee_epic[name] + guarantee_epic
    bank.CommitGuaranteeEpic()
    player.epic_interval[name] = math.floor(GetChestCount(name) / player.guarantee_epic[name])

    local guarantee_legendary = item_data:GetItemsData().guarantee_legendary
    if guarantee_legendary then
        player.guarantee_legendary[name] = player.guarantee_legendary[name] + guarantee_legendary
        bank.CommitGuaranteeEpic()
        player.legendary_interval[name] = math.floor(GetChestCount(name) / player.guarantee_legendary[name])
    end

    -- display feedback
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.ChestsRewardFeedback(value, name)
    end, 0.125, true, true)

    -- start spend logging
    local items = { chests = string.format("%d", value) }
    local event = LogEvents.BuyItem(item_data, items)
    LogEvent(event)
end

function PurchaseStoreItem_EventChests(item_data)
    -- apply product
    local value = item_data:GetItemsData().chests
    local name = item_data.chest_type
    event_mode.AddChests(value)


    -- display feedback
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.EventChestsRewardFeedback(value, name)
    end, 0.125, true, true)

    -- start spend logging
    local items = { chests = string.format("%d", value) }
    local event = LogEvents.BuyItem(item_data, items)
    LogEvent(event)
end

function PurchaseStoreItem_AutoBattle(item_data)
    -- apply product
    local item = item_data:GetItemsData()
    AddGemAutoBattleTime(item.autobattle)

    -- display feedback
    ui.ShowFeedbackAddAutoBattle(item_data.label.store_value)

    -- start spend logging
    local items = { autobattle = item_data.label.store_value }
    local event = LogEvents.BuyItem(item_data, items)
    LogEvent(event)
end

function PurchaseStoreItem_SpeedUp(item_data)
    -- apply product
    local item = item_data:GetItemsData()
    AddSpeedx3Time(item.speedup)

    -- display feedback
    ui.ShowFeedbackAddSpeedUp(item_data.label.store_value)

    -- start spend logging
    local items = { speedup = item_data.label.store_value }
    local event = LogEvents.BuyItem(item_data, items)
    LogEvent(event)
end

function PurchaseStoreItem_Token(item_data)
    -- apply product
    local value = item_data:GetItemsData().token
    local first_time = CheckFirstTimeBonus(item_data.product_id)
    if first_time then
        value = value + value
        bank.CommitFirstTimeBonus(item_data.product_id)
        UpdateStoreTokenDisplay(shop.GetItemLayout(item_data.product_id))
    end
    AddPlayerOverflowGems(value)

    -- display feedback
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.TokenRewardFeedback(value)
    end, 0.125, true, true)

    -- start spend logging
    local items = { tokens = string.format("%d", value), first_time = first_time }
    local start_spend_log = true
    local event = LogEvents.BuyItem(item_data, items, start_spend_log)
    LogEvent(event)
end

function PurchaseStoreItem_Potion(item_data)
    -- apply product
    local value = item_data:GetItemsData().potion
    local first_time = CheckFirstTimeBonus(item_data.product_id)
    if first_time then
        value = value + value
        bank.CommitFirstTimeBonus(item_data.product_id)
        UpdateStoreTokenDisplay(shop.GetItemLayout(item_data.product_id))
    end
    event_mode.AddPotions(value)

    -- display feedback
    DCEI.TriggerAddTimerEventElapsed(function()
        ui.PotionRewardFeedback(value)
    end, 0.125, true, true)
end

function LogSocialLinkClickthrough(key)
    -- create a new entry if necessary
    player.shop_logging[key] = player.shop_logging[key] or 0

    -- increment shop entry count
    player.shop_logging[key] = player.shop_logging[key] + 1

    -- log user property
    local value = player.shop_logging[key]
    local name = "Open Social " .. key
    DCEI.Event.SetUserProperty(name, value)

    -- save to bank
    bank.CommitShopLogging()
end

function LogLeaderboardOpen()
    if not ui.leaderboard.log_open then
        ui.leaderboard.log_open = true
        local key = "leaderboard"

        -- create a new entry if necessary
        player.shop_logging[key] = player.shop_logging[key] or 0

        -- increment shop entry count
        player.shop_logging[key] = player.shop_logging[key] + 1

        local user_prop = UserProps.OpenLeaderboardCount()
        LogUserProperty(user_prop)

        -- save to bank
        bank.CommitShopLogging()
    end
end

function LogShopEntry(source)
    -- create a new shop entry source if necessary
    player.shop_logging[source] = player.shop_logging[source] or 0

    -- increment shop entry count
    player.shop_logging[source] = player.shop_logging[source] + 1

    local user_prop = UserProps.ShopEntryFrom(source)
    LogUserProperty(user_prop)

    -- save to bank
    bank.CommitShopLogging()
end

function SpendLog_Gold(value, spend_event)
    -- only log the event if we're actively logging spend events
    if spend_event and player.spend_log.log then
        local transaction = ShallowCopy(spend_event)
        local transaction_list = player.spend_log.log
        local count = #transaction_list
        local last_transaction = transaction_list[count]

        -- track total spend during for this spend log
        player.spend_log.spent_gold = player.spend_log.spent_gold and (player.spend_log.spent_gold + value) or value

        -- check if last transaction is the same is new transaction
        if
            last_transaction
            and spend_event.type == last_transaction.type
            and spend_event.target == last_transaction.target
        then
            -- if both are the same, combine them (ex Upgrade Hero Knight x3)
            last_transaction.count = last_transaction.count + 1
            last_transaction.value = last_transaction.value + tonumber(value)
            transaction = last_transaction
        else
            -- otherwise create a new transaction and add it to the list
            transaction.value = value
            transaction.count = 1
            transaction.currency = "gold"
            table.insert(transaction_list, transaction)
        end

        -- debug print transaction
        -- local print_type = transaction.type .. " " .. transaction.target .. " "
        -- local print_count = transaction.count > 1 and ("x" .. transaction.count .. " ") or ""
        -- local print_cost = "[" .. tostring(transaction.value) .. " gold]"
        -- DCEI.LogMessage(">> " .. print_type .. print_count .. print_cost)

        -- check for spend limit
        local spend_limit = tonumber(player.spend_log.shop_data.purchased_gold) or 0
        local spend_total = player.spend_log.spent_gold
        local spend_threshold = 0.75

        -- if spend threshold reached, end spend logging and send the event
        if spend_limit > 0 and spend_total >= (spend_limit * spend_threshold) then
            local trigger = "Spend"
            PushSpendLogging(trigger)

            -- return to avoid double committing spend log
            return
        end

        bank.CommitSpendLog()
    end
end

function SpendLog_Token(value, spend_event)
    -- only log the event if we're actively logging spend events
    if spend_event and player.spend_log.log then
        local transaction = ShallowCopy(spend_event)
        local transaction_list = player.spend_log.log
        local count = #transaction_list
        local last_transaction = transaction_list[count]

        -- track total spend during for this spend log
        player.spend_log.spent_token = player.spend_log.spent_token and (player.spend_log.spent_token + value) or value

        -- check if last transaction is the same is new transaction
        if
            last_transaction
            and spend_event.type == last_transaction.type
            and spend_event.target == last_transaction.target
        then
            -- if both are the same, combine them (ex Upgrade Defense AutoTurret x3)
            last_transaction.count = last_transaction.count + 1
            last_transaction.value = last_transaction.value + tonumber(value)
            transaction = last_transaction
        else
            -- otherwise create a new transaction and add it to the list
            transaction.value = value
            transaction.count = 1
            transaction.currency = "tokens"
            table.insert(transaction_list, transaction)
        end

        -- debug print transaction
        -- local print_type = transaction.type .. " " .. transaction.target .. " "
        -- local print_count = transaction.count > 1 and ("x" .. transaction.count .. " ") or ""
        -- local print_cost = "[" .. tostring(transaction.value) .. " tokens]"
        -- DCEI.LogMessage(">> " .. print_type .. print_count .. print_cost)

        -- check for spend limit
        local spend_limit = tonumber(player.spend_log.shop_data.purchased_token) or 0
        local spend_total = player.spend_log.spent_token
        local spend_threshold = 0.75

        -- if spend threshold reached, end spend logging and send the event
        if spend_limit > 0 and spend_total >= (spend_limit * spend_threshold) then
            local trigger = "Spend"
            PushSpendLogging(trigger)

            -- return to avoid double committing spend log
            return
        end

        bank.CommitSpendLog()
    end
end

function PushSpendLogging(trigger)
    -- don't send spend log if nothing was actually logged
    if #player.spend_log.log == 0 then
        return
    end

    -- construct spend log event
    local event = {}
    event.name = "Buy Item Spend"
    event.props = ShallowCopy(player.spend_log.shop_data)
    event.props.trigger = trigger

    local spend_types = {}
    event.props.spend_list = ""

    for index, transaction in ipairs(player.spend_log.log) do
        -- only log first 99 spend counts
        if index <= 99 then
            local type = tostring(transaction.type) .. " " .. tostring(transaction.target)
            local count = (tonumber(transaction.count) > 1) and ("x" .. transaction.count .. " ") or ""
            local cost = "[" .. tostring(transaction.value) .. " " .. tostring(transaction.currency) .. "]"

            -- format as two digits for readability in analytics
            local spend_index = "spend" .. string.format("%02d", index)
            event.props[spend_index] = type

            -- log type and details separately for easier comparison
            local spend_index = "spend_details" .. string.format("%02d", index)
            event.props[spend_index] = type .. " " .. count .. cost

            -- track each spend type
            spend_types[type] = true
        else
            break
        end
    end

    -- log each spend type in a collated string
    for k, v in pairs(spend_types) do
        local spend_tag = "[" .. tostring(k) .. "]"

        if event.props.spend_list then
            event.props.spend_list = event.props.spend_list .. " " .. spend_tag
        else
            event.props.spend_list = spend_tag
        end
    end

    -- log the event
    LogEvent(event)

    -- reset spend log
    player.spend_log = {}
    bank.CommitSpendLog()
end

function AttemptPushExistingSpendLog(trigger)
    -- push (and reset) existing log if log exists and has entries
    if player.spend_log.log and #player.spend_log.log > 0 then
        PushSpendLogging(trigger)
    end
end

function StartSpendLogging(event)
    -- send previous spend log if new spend log is started
    local trigger = "Buy Item"
    AttemptPushExistingSpendLog(trigger)

    -- create new spend log
    player.spend_log = {}
    player.spend_log.log = {}
    player.spend_log.shop_data = event and event.props

    -- save record of spend log opening, in case player quits before spend
    bank.CommitSpendLog()
end

-- STORE REDESIGN
function UpdateShopSpeedUpInfo(info_frame)
    local value = player.speed_up_timer and player.speed_up_timer:GetTimeRemaining() or 0

    if value and value > 0 then
        -- display 3x time remaining
        local text = DCEI.GetTranslationText(
            "shop/info/speed_up",
            { x = "<color=green>" .. ui.FormatTimeHours(value) .. "</color>" }
        )
        DCEI.SetTextFrameText(info_frame.InfoLabel, text)

        DCEI.SetFrameActive(info_frame.Info, true)
    else
        -- hide info frame if no 3x time
        DCEI.SetFrameActive(info_frame.Info, false)
    end
end

function ShowShopPip()
    ui.bottom_panel:ShowShopPip()

    -- also show pip on deals section tab, as all new items are in deals section for now
    -- don't show nav pip, as boosts are moved to gold/token
    --ui.shop.corner_navs[1]:ShowPip()
end

function ShowShopChestsPip()
    ui.bottom_panel:ShowShopPip()

    -- also show pip on deals section tab, as all new items are in deals section for now
    local layout = ui.shop.sections["chests"]
    layout:ShowPip()
end

function OpenShop(on_close)
    if not ui.shop.is_active then

        -- only open shop if user is online
        if DCEI.IsOnline() then

            -- iOS version shifts status UI down and looks bad with shop UI
            if PlayingiOSVersion() then
                DCEI.SetFrameActive(ui.top_panel.status_container, false)
            end

            -- mark shop as active
            ui.shop.on_close = on_close

            -- hide shop pip
            ui.bottom_panel:HideShopPip()

            CloseActivePrompts()
            MoveCamera(cameras.current, 0.5, {x_offset = -4})

            local game_state = ui.bottom_panel.state
            if game_state == "castle" then
                ui.loadout.Hide()
                --DCEI.SetFrameActive(ui.glue.upgrade_frame, false)
            elseif game_state == "town" then
                ui.loadout.town:Hide()
            elseif game_state == "battle" then
                ui.loadout.hero.Hide()
            end

            -- attempt to connect to server if haven't already
            if not server.day_fetched then
                CheckServerTimeOnOpenShop()
            end
            
            ui.shop:Show()
            DCEI.WaitForUiUpdateToComplete(function()
                -- update season pass
                if not event_mode.IsActive() then
                    DCEI.LogMessage(wave.data.GetCurrent())
                    if wave.data.GetMax() >= 60 then
                        ui.shop.sections["winter_pass"]:Show()
                        ui.shop.sections["autumn_pass"]:Show()
                        if not CheckSeasonEventExist("winter_event2022") then
                            DCEI.LogMessage("WINTER SEASON DOESNT EXIST")
                            ui.shop.sections["winter_pass"]:Hide()
                        end
                        if not CheckSeasonEventExist("autumn_event") then
                            DCEI.LogMessage("AUTUMN SEASON DOESNT EXIST")
                            ui.shop.sections["autumn_pass"]:Hide()
                        end
                    else
                        ui.shop.sections["winter_pass"]:Hide()
                        ui.shop.sections["autumn_pass"]:Hide()
                    end

                    local unlock_rank_reward = CheckUnlockHellModeRankReward()
                    DCEI.SetFrameActive(ui.shop.sections["hell_coin"].Frame, unlock_rank_reward)
                end
            end)
            SetGameState("shop")
        else

            ui.ShowFeedbackOffline()

        end
    end

end

function CloseShop()
    if ui.shop.is_active then
        ui.shop.is_active = false
        local duration = 0.3125
        MoveCamera(cameras.current, duration)

        local game_state = ui.bottom_panel.state
        if game_state == "castle" then
            ui.loadout.Show()
        elseif game_state == "town" then
            ui.loadout.town:Show()
        elseif game_state == "battle" then
            ui.loadout.hero.Show()
        end

        -- hide deals section pip
        -- ui.shop.corner_navs[1]:HidePip()
        -- ui.shop.navs[3]:HidePip()

        ui.shop:Close()

        CloseActivePrompts()

        if ui.shop.on_close then
            ui.shop.on_close()
        end

        -- show status UI in case hidden
        DCEI.SetFrameActive(ui.top_panel.status_container, true)
        SetGameState(ui.bottom_panel.state)
    end
end

function SetStoreEntrySource(source)
    source = source or "none"

    ui.shop.source = source

    -- log on delay to not potentially disrupt shop open animation
    DCEI.TriggerAddTimerEventElapsed(function()
        LogShopEntry(source)
    end, 0.5, false, true)
end

function OpenDealsStore(on_close)
    OpenShop(on_close)

    local page_key = "deals"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenGoldStore(on_close)
    OpenShop(on_close)

    local page_key = "gold"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenTokenStore(on_close)
    OpenShop(on_close)

    local page_key = "tokens"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenPotionStore(on_close)
    OpenShop(on_close)

    local page_key = "potions"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenChestStore(on_close)
    OpenShop(on_close)

    local page_key = "chests"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenAutoBattleStore(on_close)
    OpenShop(on_close)

    local page_key = "auto_battle"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenSpeedUpStore(on_close)
    OpenShop(on_close)

    local page_key = "speed_up"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

function OpenHellCoinShop(on_close)
    OpenShop(on_close)

    local page_key = "hell_coin"
    DCEI.WaitForUiUpdateToComplete(function()
        ui.shop.sections[page_key]:ScrollTo()
    end)
end

-- store enter
function OpenShopFromShopNav()
    if not wave.data.is_active and not ui.shop.is_active then
        local from = "shop button"

        -- log if shop button had pip when clicked
        if ui.bottom_panel.shop_pip then
            from = from .. " (pip)"
        end

        SetStoreEntrySource(from)

        CloseActivePrompts()

        OpenShop()

        local layout = ui.shop.sections["deals"]
        DCEI.WaitForUiUpdateToComplete(function()
            layout:ScrollTo()
        end)

        if TestB("test010920_massive_bundle") then
            CreateHugeBundlePopup()
        end
    end
end

function AttemptGoldStoreFromResourcePanel()
    if not wave.data.is_active and not ui.shop.is_active then
        local from = "glue gold"
        SetStoreEntrySource(from)
        CloseActivePrompts()
        OpenGoldStore()
    end
end

function AttemptTokenStoreFromResoucePanel()
    if not wave.data.is_active and not ui.shop.is_active then
        local from = "glue token"
        SetStoreEntrySource(from)

        CloseActivePrompts()
        if event_mode.IsActive() then
            OpenPotionStore()
        else
            OpenTokenStore()
        end
    end
end

-- !SECTION

-- SECTION Functions
-- DCOIN GET/SET
function SpendPlayerDCoin(value, skip_commit)
    DCEI.SetProtectedValue(data.safe.dcoin_spent, math.max(0, DCEI.GetProtectedValue(data.safe.dcoin_spent) + value))
    --ui.top_panel.dcoin:Update()

    if not skip_commit then
        bank.CommitDCoinSpent()
    end
end

function GetPlayerDcoinSpent()
    return DCEI.GetProtectedValue(data.safe.dcoin_spent)
end

function DevMode()
    return player.dev_mode
end

function AirshipTestMode()
    return player.airship_test_mode
end

ui.AirshipTestMode = AirshipTestMode

function UnlockEquipment()
    return wave.data.GetMax() >= 100 and not event_mode.IsActive()
        or #player.inventory.items > 0
        or GetMaterialCount("normal_material") > 0
        or GetMaterialCount("epic_material") > 0
end

function CloseActiveMenus()
    if ui.leaderboard then
        ui.leaderboard:Hide()
    end
    if ui.battlestats then
        ui.battlestats:Hide()
    end
    if ui.calendar then
        ui.calendar:Hide()
    end
    if ui.social then
       ui.social:Hide()
    end
    if ui.worker_menu then
        ui.worker_menu:Close()
    end
    if ui.vault_menu then
        ui.vault_menu:Close()
    end
    if ui.gear_menu then
        ui.gear_menu.Close()
    end
    if ui.talent_menu then
        ui.talent_menu.Close()
    end
    if ui.player_deck then
        ui.player_deck:Hide()
    end

    AttemptCloseHeroList()
    AttemptCloseTowerList()
    AttemptCloseDefenseList()
    AttemptClosePromoteList()
end

function CloseActivePrompts()
    ui.DestroyAdPrompt()

    if ui.prompt.tutorial then
        ui.prompt.tutorial.Close()
    end
end

function ShowSettingsMenu()
    if ui.settings_menu then
        ui.settings_menu:Close()
        return
    end
    SetGameSpeed(0)
    DCEI.SetBlurredScreenCaptureAsBackground(true, 0.5)
    DCEI.SetCameraFocus(100, 100, 0, false)
    ui.ActiveAllUI(false)
    ui.settings_menu = ui.CreateSettingsMenu()
    ui.settings_menu.on_close = function()
        DCEI.SetBlurredScreenCaptureAsBackground(false, 0.5)
        if wave.data.is_active then
            SetGameSpeed(player.speed)
        else
            ClearGameSpeed()
        end
        ui.ActiveAllUI(true)
        local duration = 0
        MoveCamera(cameras.current, duration)
        ui.settings_menu = nil
    end
    ui.settings_menu.load_savedata_callback = LoadBackupSaveDataPriority
end

-- !SECTION

-- SECTION UI INTEGRATION
function HookupAirShipUI()
    -- initialize colony menu
    ui.colony.Initialize(levels.colony, levels.seasonal_colony)
    ui.colony.tutorial_check = Tutorial07_Colony_StepC
    ui.challenge.tutorial_check = Tutorial09_Challenge_Check
    ui.raid_boss.tutorial_check = RaidBoss_Tutorial_Check

    -- initialize challenge menu
    ui.challenge.Initialize(levels.challenge)

    -- initialize raid_boss menu
    ui.raid_boss.Initialize(levels.raid_boss, player.raid_boss_log)

    DCEI.SetOnClickCallback(ui.colony.confirmation.button, StartColonyBattle)
    DCEI.SetOnClickCallback(ui.challenge.confirmation.button, StartChallengeBattle)
    DCEI.SetOnClickCallback(ui.raid_boss.confirmation.button, StartRaidBossBattle)
end

function HookupUI()
    -- share necessary functions with UI mod
    ui.GetPlayerGold = GetPlayerGold
    ui.GetPlayerElixir = GetPlayerElixir
    ui.GetMineContracts = GetMineContracts
    ui.GetTownContracts = GetTownContracts
    ui.GetPlayerChest = GetChestCount
    ui.GetPlayerMaterial = GetMaterialCount
    ui.GetPhilosopherStone = GetPhilosopherStone
    ui.GetRaidTurbo = GetRaidTurbo
    ui.GetHellCoin = GetHellCoin
    ui.MassDisenchant = MassDisenchantEquipment
    ui.MassSell = MassSellEquipment
    ui.GetCraftLevel = GetCraftLevel
    ui.UpgradeCraft = AttemptUpgradeCraft
    ui.OpenRenameMenu = OpenRenameMenu
    ui.OpenPlayerProfile = OpenPlayerProfile

    ui.ShowSettingMenu = ShowSettingsMenu

    ui.OpenGoldStoreFromPrompt = function(on_close)
        local from = "get gold prompt"
        SetStoreEntrySource(from)
        OpenGoldStore(on_close)
    end
    ui.OpenTokenStoreFromPrompt = function(on_close)
        local from = "get tokens prompt"
        SetStoreEntrySource(from)
        OpenTokenStore(on_close)
    end

    -- craft menu
    ui.craft.unlock_castle_level = data.unlock.unlock_craft_level
    ui.craft.max_level = CRAFT_MAX_LEVEL

    -- daily reward calendar
    ui.calendar.reward_data = data.calendar
    ui.calendar.player_data = player.calendar

    -- badges
    ui.badge_data = data.badges
    -- DCEI.SetOnClickCallback(
    --     ui.calendar.seasonpass_button,
    --     function()
    --         AttemptBuySeasonPass()
    --     end
    -- )

    -- store buttons
    DCEI.SetOnClickCallback(ui.top_panel.gold.frame, AttemptGoldStoreFromResourcePanel)
    DCEI.SetOnClickCallback(ui.top_panel.gem.frame, AttemptTokenStoreFromResoucePanel)

    -- hero loadout
    local loadout = ui.loadout.hero
    for id, select_box in ipairs(loadout.slots) do
        DCEI.SetOnClickCallback(select_box.button, function()
            ClickHeroSelectBox(select_box, id)
        end)
    end

    -- tower loadout
    loadout = ui.loadout.tower
    for id, select_box in ipairs(loadout.slots) do
        DCEI.SetOnClickCallback(select_box.button, function()
            select_box.Highlight()
            ClickTowerSelectBox(select_box, id)
        end)
    end

    -- defenses loadout
    loadout = ui.loadout.defense
    for id, select_box in ipairs(loadout.slots) do
        DCEI.SetOnClickCallback(select_box.button, function()
            select_box.Highlight()
            ClickDefenseSelectBox(select_box, id)
        end)
    end
    if not event_mode.IsActive() then
        for n = 1, data.unlock.gear_slot_max do
            local name = player.gear[n]
            local unit_data = data.gear[name]

            if unit_data then
                ui.loadout.gear:UpdateGear(unit_data, n)
            end
        end

        DCEI.SetOnClickCallback(ui.loadout.gear.OpenGearButton, function()
            NewGearMenu()
        end)

        -- town loadout
        loadout = ui.loadout.town
        
        DCEI.SetOnClickCallback(ui.glue.district_menu.frame, function()
            -- use last selected district index or 1
            local id = ui.loadout.town.select or 1
            ui.loadout.town.slots[id].Highlight()
            ClickTownSelectBox(id)
        end)

        DCEI.SetOnClickCallback(ui.glue.monster_lab_menu.frame, function()
            AttemptOpenMonsterLab()
        end)

        -- hidden town select boxes
        for id, select_box in ipairs(loadout.slots) do
            DCEI.SetOnClickCallback(select_box.button, function()
                select_box.Highlight()
                ClickTownSelectBox(id)
            end)
        end
        
        DCEI.SetOnClickCallback(loadout.vault.button, ClickVault)

        -- inventory menu
        DCEI.SetOnClickCallback(ui.glue.inventory_menu.frame, NewInventoryMenu)

        -- town hall
        DCEI.SetOnClickCallback(loadout.hall.button, AttemptOpenTownHallMenu)
    end
    DCEI.SetOnClickCallback(ui.loadout.town.dwarf.button, NewWorkerMenu)

    -- xp top bar
    UpdateExperienceBar()

    -- right glue
    DCEI.SetOnClickCallback(ui.glue.gear_menu.frame, NewGearMenu)
    DCEI.SetOnClickCallback(ui.glue.talent_menu.frame, ShowTalentMenu)
    ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())

    -- player deck button & reference
    DCEI.SetOnClickCallback(ui.glue.player_deck_menu.frame, ShowPlayerDeck)

    -- other buttons
    DCEI.SetOnClickCallback(ui.glue.other_buttons.frame, ShowSecondaryMenu)
    DCEI.SetOnClickCallback(ui.glue.leader_menu.frame, ShowLeaderBoard)
    DCEI.SetOnClickCallback(ui.glue.battlestats_menu.frame, ShowBattleStats)
    DCEI.SetOnClickCallback(ui.glue.calendar_menu.frame, ShowDailyRewardCalendar)
    --DCEI.SetOnClickCallback(ui.glue.social_menu.frame, ShowSocial) 	
    DCEI.SetFrameActive(ui.glue.social_menu.frame, false)
	

    DCEI.SetOnClickCallback(ui.glue.hell_mode_button.frame, ShowHellModeMenu)
    DCEI.SetOnClickCallback(ui.egg_hunt_banner.Button, AttemptOpenEggHuntGift)
    -- upgrades

    if event_mode.IsActive() then
        ui.HookUpHoldButtonFunctions(ui.glue.upgrade_archer, ui.glue.upgrade_archer.frame, AttemptUpgradeArcherEvent)
    else
        ui.HookUpHoldButtonFunctions(ui.glue.upgrade_archer, ui.glue.upgrade_archer.frame, AttemptUpgradeArcher)
    end

    UpdateArcherUpgradeButton()
    if event_mode.IsActive() then
        ui.HookUpHoldButtonFunctions(ui.glue.upgrade_castle, ui.glue.upgrade_castle.frame, AttemptUpgradeCastleLevelEvent)
    else
        ui.HookUpHoldButtonFunctions(ui.glue.upgrade_castle, ui.glue.upgrade_castle.frame, AttemptUpgradeCastleLevel)
    end
    
    UpdateCastleUpgradeButton()

    DCEI.SetOnClickCallback(ui.glue.floating_winter_event.Frame, OpenWinterEvent)
    DCEI.SetOnClickCallback(ui.glue.floating_autumn_event.Frame, OpenAutumnEvent)
    -- deals page
    DCEI.SetOnClickCallback(ui.glue.floating_deals.button, OpenStoreSpecials)
    -- event page
    DCEI.SetOnClickCallback(ui.glue.floating_event.Frame, OpenHellModeEvent)

    if not event_mode.IsActive() then
        -- airship menu
        DCEI.SetOnClickCallback(loadout.airship.button, AttemptOpenAirshipMenu)
    end

    -- MAIN NAVIGATION
    -- main > battle
    ui.bottom_panel.right_outer:SetOnClickCallbackForStates({ "castle", "town" }, function()
        local nextWave = NextWave()
        if TestB("wave_skip") and nextWave.rating > 25 then
            ShowWaveSkipWindow(nextWave)
        else
            WaveStartActions(nextWave)
        end
    end)
    -- main > replay
    ui.bottom_panel.right_inner:SetOnClickCallbackForStates({ "castle", "town" }, function()
        IncrementReplayCount()
        WaveStartActions(LastWave())
    end)
    -- main > town and main > castle
    ui.bottom_panel.left_inner:SetOnClickCallbackForStates({ "castle", "town" }, function()
        TownTransition()
    end)
    -- main > shop
    ui.bottom_panel.left_outer:SetOnClickCallbackForStates({ "castle", "town" }, function()
        OpenShopFromShopNav()
    end)

    DCEI.SetOnClickCallback(ui.shop.CloseButton, CloseShop)

    -- AIRSHIP NAVIGATION
    -- colony > back and challenge > back
    ui.bottom_panel.left_outer:SetOnClickCallbackForStates({ "colony", "challenge", "raid_boss" }, function()
        CloseAirshipMenu()
    end)
    -- colony, raid_boss > challenge
    ui.bottom_panel.right_inner:SetOnClickCallbackForStates({ "colony", "raid_boss" }, function()
        AttemptShowChallengeMenu()
    end)
    ui.bottom_panel.right_inner:SetOnClickCallbackForStates({ "challenge" }, function() end)
    -- challenge, raid_boss > colony
    ui.bottom_panel.left_inner:SetOnClickCallbackForStates({ "challenge", "raid_boss" }, function()
        ShowColonyMenu()
    end)
    -- colony > colony (press 3 times in dev mode to enable airship test)
    ui.bottom_panel.left_inner:SetOnClickCallbackForStates({ "colony" }, function()
        CheckUnlockAirshipTestMode()
    end)
    --colony, challenge > raid boss
    ui.bottom_panel.right_outer:SetOnClickCallbackForStates({ "colony", "challenge" }, function()
        AttemptShowRaidBossMenu()
    end)
    ui.bottom_panel.right_outer:SetOnClickCallbackForStates({ "raid_boss" }, function() end)

    -- BATTLE COMMANDS
    -- battle > back
    ui.bottom_panel.left_outer:SetOnClickCallbackForStates({ "battle" }, function()
        if not PlayerInTargetMode() then
            if not wave.data.ModeStandard() then
                CreateWaveExitConfirmMenu()
            else
                WaveEndActions()
            end
        end
    end)
    ui.bottom_panel.middle:SetOnClickCallbackForStates({ "battle" }, function()
        CheckExitTargetMode()
    end)

    DCEI.SetOnClickCallback(ui.bottom_panel.speed.x2.frame, function()
        if not PlayerInTargetMode() then
            AdvanceBattleSpeed()
        end
    end)
    DCEI.SetOnClickCallback(ui.bottom_panel.speed.x3.frame, function()
        if not PlayerInTargetMode() then
            AdvanceBattleSpeed()
        end
    end)

    -- initialize speedup values
    ui.bottom_panel.speed:Update(player.speed)
    ui.bottom_panel.speed:Set3xMode(player.speed_up)

    -- floating prompts (ignore input in target mode)
    DCEI.SetOnClickCallback(ui.battle.floating_autobattle.frame, function()
        if not PlayerInTargetMode() then
            CreateAutoBattleMenu()
        end
    end)
    DCEI.SetOnClickCallback(ui.battle.floating_speedup.frame, function()
        if not PlayerInTargetMode() then
            AttemptSpeedAdMenu()
        end
    end)

    -- unit loadout slots
    HookUpLoadoutSlots()

    if not event_mode.IsActive() then
        -- boost display data
        HookupBoostTimerData()
    end
	
	---add

	
	
	
end

function HookupBoostTimerData()
    -- move from game data
    for _, product in ipairs(store_product_list) do
        if product.type == "boosts" then
            local boost_data = {
                type = product.custom.boost.type,
                multiplier = product.custom.boost.multiplier,
            }
            local id = next(product:GetItemsData())

            data.boost_timers[id] = boost_data
        end
    end
    -- store references to boost data
    ui.boost_display.timers = player.timers
    ui.boost_display.boost_data = data.boost_timers
end

function HookUpLoadoutSlots()
    local options = {
        offset = { up = 1.75 },
        center_at_unit_origin = true,
    }
    local use_current_root = true

    -- hero slots
    for k, unit in ipairs(group_hero_slots) do
        local select_box = ui.loadout.hero.slots[k].frame
        DCEI.AttachFrameToUnit(select_box, unit, options, use_current_root)
    end

    -- hero skill bars
    options.offset = { up = 2.6 }
    for k, unit in ipairs(group_hero_slots) do
        local skill_bar = ui.loadout.hero.slots[k].skill_bar.frame
        DCEI.AttachFrameToUnit(skill_bar, unit, options, use_current_root)
    end

    -- tower slots
    options.offset = { up = 0.5 }
    for k, unit in ipairs(group_tower_slots) do
        local select_box = ui.loadout.tower.slots[k].frame
        DCEI.AttachFrameToUnit(select_box, unit, options, use_current_root)
    end

    -- defense slots
    options.offset = { up = 1.95 }
    for k, unit in ipairs(group_defense_slots) do
        local select_box = ui.loadout.defense.slots[k].frame
        DCEI.AttachFrameToUnit(select_box, unit, options, use_current_root)
    end

    options.offset = { up = 0.5 }
    DCEI.AttachFrameToUnit(ui.loadout.gear.Frame, gear_anchor, options, use_current_root)

    -- town slots
    options.offset = { up = 0.5 }
    for k, unit in ipairs(group_town_slots) do
        local select_box = ui.loadout.town.slots[k].frame
        DCEI.AttachFrameToUnit(select_box, unit, options, use_current_root)
    end

    -- dwarf workers
    options.offset = { up = -0.2 }
    local select_box = ui.loadout.town.dwarf.frame
    DCEI.AttachFrameToUnit(select_box, MINE_DEPOT, options, use_current_root)

    -- vault
    if VAULT and DCEI.UnitExists(VAULT) then
        local select_box = ui.loadout.town.vault.frame
        DCEI.AttachFrameToUnit(select_box, VAULT, options, use_current_root)
    end

    -- airship
    local select_box = ui.loadout.town.airship.frame
    DCEI.AttachFrameToUnit(select_box, AIRSHIP_DOCK, options, use_current_root)

    -- town hall
    local select_box = ui.loadout.town.hall.frame
    DCEI.AttachFrameToUnit(select_box, TOWN_HALL, options, use_current_root)
end

-- UI HELPERS
function AttemptCloseHeroList()
    if ui.list_hero then
        ui.list_hero.Close()
    end
end

function AttemptCloseTowerList()
    if ui.list_tower then
        ui.list_tower.Close()
    end
end

function AttemptCloseDefenseList()
    if ui.list_defense then
        ui.list_defense.Close()
    end
end

function AttemptClosePromoteList()
    if ui.list_promote then
        ui.list_promote.Close()
    end
end

function AttemptCloseSkipWave()
    if ui.skip_wave then
        ui.skip_wave.Close(true)
    end
end

function SetAdOnCooldown(ad_type, duration)
    duration = duration or ad_type.cooldown
    if ad_type.is_available then
        ad_type.is_available = false

        DCEI.TriggerAddTimerEventElapsed(function()
            ad_type.is_available = true
        end, duration, true, true)
    end
end

function AttemptGoldAd(is_defeat)
    ui.DestroyAdPrompt()
    if watch_ad_fail_count >= 2 then
        return
    end
    if TestB("test010715_daily_ads_limit") and player.remove_ad and player.daily_ads_limit <= 0 then
        return
    end
    local ad_type = ad_data.gold
    local unlock_gold_ad = 3
    local or_dev_mode = true

    if wave.data.GetMax() >= unlock_gold_ad and ad_type.is_available then
        if DCEI.IsOnline() and DCEI.IsAdsReady() then
            CreateGoldAd(is_defeat)
        end
    end
end

function CreateGoldAd(is_defeat, multiplier)
    local base_multiplier = 3.5
    local multiplier = multiplier or base_multiplier + math.random()
    local wave_value = is_defeat and wave.data.earned or wave.data.GetGoldValueWithBonus(wave.data.GetCurrent())
    local value = wave_value * multiplier

    if value < 1 then
        DCEI.LogMessage("> Ad Value too low")
        return
    end

    local function OnClose()
        -- LogEvent(LogEvents.exit_ad("Gold"))
    end

    value = math.floor(value)
    -- don't display remove ads banner before wave 10
    local display_remove_ad_banner = wave.data.GetMax() > 10
    ui.prompt.ad = ui.NewAdPrompt(value, OnClose, player.remove_ad, display_remove_ad_banner, is_defeat, multiplier)
    -- LogEvent(LogEvents.show_ad_gold(is_defeat, value))

    local function ResolveAds(success)
        -- trigger ads cooldown regardless of success or failure for now
        SetAdOnCooldown(ad_data.gold)

        if success then
            AdTrackLimit(ad_data.gold)
            AddPlayerGold(value)

            DCEI.TriggerAddTimerEventElapsed(function()
                ui.GoldRewardFeedback(value)
            end, 0.125, false, true)

            LogEvent(LogEvents.watch_ad("Gold", not is_defeat, value))
            player.ads_watched_gold = player.ads_watched_gold + 1

            CheckCampaignEvent_GoldAdx3()
            CheckCampaignEvent_GoldAdx5()

            bank.CommitAdWatchGold()
            bank.CommitGold()
        else
            DCEI.LogMessage("> Playing Ads Failed")
            LogEvent(LogEvents.watch_ad_failed("Gold", not is_defeat, value))
            watch_ad_fail_count = watch_ad_fail_count + 1
            -- LogEvent(LogEvents.fail_ad("Gold"))
        end
    end

    DCEI.SetOnClickCallback(ui.prompt.ad.button, function()
        ui.DestroyAdPrompt(true)
        DCEI.LogMessage("> Showing Ads")
        ShowAds(ResolveAds)
    end)
end

function ShowAds(resolve_ads, type, is_victory, gold_value)
    if DevMode() or player.remove_ad then
        local success = true
        resolve_ads(success)
    else
        -- local fallbackToUnityAds = not TestB("test010920_ironsource_ads")
        LogEvent(LogEvents.watch_ad_click(type, is_victory, gold_value))
        DCEI.ShowAds(resolve_ads, true)
    end
end

function CreateWaveDropResourceFlyAnimation()
    local wait_time = 0
    if TestB("winter_event2022") and ui.glue.floating_winter_event.is_active then
        local winter_coin = wave_drop_resource.winter_coin or 0
        if winter_coin > 0 then
            wait_time = 1
            BlockInput()
            local options = {
                stay_time = 0.8,
                fly_time = 0.675,
                start_h_offset = 100,
                h_curve = "Linear",
                v_curve = "OutCubic",
                icon = "general_icon_icecoin1",
                size = 60,
            }
            ui.FlyTo(ui.world, ui.glue.floating_winter_event.Frame, winter_coin, function()
                local current_coin = GetPlayerWinterCoin2022()
                local previous_coin = current_coin - winter_coin
                local previous_exp, required_exp = GetWinterPassExpFill(previous_coin)
                local current_exp = previous_exp + winter_coin

                local start_fill = previous_exp / required_exp
                local target_fill = current_exp / required_exp

                local animated = true
                ui.glue.floating_winter_event:UpdateFill(start_fill, target_fill, animated)
                local previous_level = player.winter_event2022.level
                CheckWinterEventUnlock()
                local current_level = player.winter_event2022.level
                if
                    target_fill >= 1
                    and previous_level < current_level
                    and tutorial.state == 0
                    and not ui.popup.is_active
                then
                    player.queued_deal_popup = false

                    ui.glue.floating_winter_event.play_upgrade_animation = true
                    DCEI.TriggerAddTimerEventElapsed(OpenWinterEvent, 0.5, true, true)
                else
                    UnblockInput()
                end
            end, options)
        end
    end
    if TestB("autumn_event") and ui.glue.floating_autumn_event.is_active then
        local autumn_coin = wave_drop_resource.autumn_coin or 0
        if autumn_coin > 0 then
            BlockInput()
            local options = {
                stay_time = 0.8,
                fly_time = 0.675,
                start_h_offset = 100,
                h_curve = "Linear",
                v_curve = "OutCubic",
                icon = "general_icon_coin_leaf",
                size = 60,
            }
            ui.FlyTo(ui.world, ui.glue.floating_autumn_event.Frame, autumn_coin, function()
                local current_coin = GetPlayerAutumnCoin()
                local previous_coin = current_coin - autumn_coin
                local previous_exp, required_exp = GetAutumnPassExpFill(previous_coin)
                local current_exp = previous_exp + autumn_coin

                local start_fill = previous_exp / required_exp
                local target_fill = current_exp / required_exp

                local animated = true
                ui.glue.floating_autumn_event:UpdateFill(start_fill, target_fill, animated)
                local previous_level = player.autumn_event.level
                CheckAutumnEventUnlock()
                local current_level = player.autumn_event.level
                if
                    target_fill >= 1
                    and previous_level < current_level
                    and tutorial.state == 0
                    and not ui.popup.is_active
                then
                    player.queued_deal_popup = false

                    ui.glue.floating_autumn_event.play_upgrade_animation = true
                    DCEI.TriggerAddTimerEventElapsed(OpenAutumnEvent, 0.5, true, true)
                else
                    UnblockInput()
                end
            end, options)
        end
    end
    if wave.data.egg_hunt_event then
        local egg = player.egg_hunt_event.raw or 0
        if egg > 0 then
            DCEI.TriggerAddTimerEventElapsed(function()
                local options = {
                    stay_time = 0.8,
                    fly_time = 0.675,
                    start_h_offset = 50,
                    start_v_offset = -50,
                    h_curve = "Linear",
                    v_curve = "OutCubic",
                    icon = "wildcastle_icon_easter_egg",
                    size = 64,
                }
                ui.FlyTo(ui.world, ui.egg_hunt_banner.Frame, egg, function()
                    local previous_egg = player.egg_hunt_event.egg
                    local current_egg = previous_egg + egg
                    local required_exp = EGG_UPGRADE_REQUIRE

                    local start_fill = previous_egg / required_exp
                    local target_fill = current_egg / required_exp
                    local animated = true

                    ui.egg_hunt_banner:UpdateFill(player.egg_hunt_event.gift, start_fill, target_fill, animated)
                    local fill_timer = 0
                    local repeat_count = 0
                    local display_gift = player.egg_hunt_event.gift
                    while current_egg > required_exp do
                        current_egg = current_egg - required_exp
                        player.egg_hunt_event.gift = player.egg_hunt_event.gift + 1
                        repeat_count = repeat_count + 1
                    end
                    player.egg_hunt_event.egg = current_egg
                    bank.CommitEventEggHunt()
                    if repeat_count > 0 then
                        for i = 1, repeat_count do
                            DCEI.TriggerAddTimerEventElapsed(function()
                                local fillamount = 1
                                if i == repeat_count then
                                    fillamount = current_egg / required_exp
                                end
                                ui.egg_hunt_banner:UpdateFill(display_gift + i, 0, fillamount, animated)
                            end, fill_timer + 0.5 * i, true, true)
                        end
                    end
                end, options)
            end, wait_time, true, true)
        end
        player.egg_hunt_event.raw = 0
    end
    wave_drop_resource = {}
end

function CreateCastleUI(map_start)
    local state = "castle"
    ui.bottom_panel:SetState(state)

    ui.battle.floating_autobattle:Hide()
    ui.battle.floating_speedup:Hide()

    if ui.new_monster_intro and DCEI.FrameExists(ui.new_monster_intro.Frame) then
        DCEI.DestroyFrame(ui.new_monster_intro.Frame)
        ui.new_monster_intro = nil
    end

    ui.loadout.Show()
    ui.loadout.tower.Show()
    ui.loadout.hero.Show()
    ui.loadout.defense.Show()
    ui.loadout.gear:Show()
    ui.loadout.town:Hide()

    ui.glue:Show()
    ui.town_glue:Hide()
    if event_mode.IsActive() then
        DCEI.SetFrameActive(ui.potion_lab.background, false)
        DCEI.SetFrameActive(ui.secret_lab.background, false)
    end

    ui.boost_display:Hide()

    CheckSystemUnlocks(map_start)

    if ui.download then
        ui.download:AnimateIn()
    end

    ShowBlessCrystal()
    if ui.devil_trail_ui then
        ui.devil_trail_ui:Update()
    end
    if ui.monster_bounty and DCEI.FrameExists(ui.monster_bounty.Frame) then
        DCEI.SetFrameActive(ui.monster_bounty.Frame, true)
    end
end

function CreateTownUI()
    local state = "town"
    ui.bottom_panel:SetState(state)
    ui.bottom_panel:HideTownPip()

    ui.loadout.Show()
    ui.loadout.hero.Hide()
    ui.loadout.tower.Hide()
    ui.loadout.defense.Hide()
    ui.loadout.gear:Hide()

    if not event_mode.IsActive() then
        ui.town_glue:Show()
    else
        DCEI.TriggerAddTimerEventElapsed(function()
            DCEI.SetFrameActive(ui.potion_lab.background, true)
            DCEI.AnimateFrameVerticalOffset(ui.potion_lab.background, -50, 280, 0.17, "Linear")
        end, 0.14)
        DCEI.TriggerAddTimerEventElapsed(function()
            DCEI.SetFrameActive(ui.secret_lab.background, true)
            DCEI.AnimateFrameVerticalOffset(ui.secret_lab.background, -230, 100, 0.17, "Linear")
        end, 0.14)
        
    end
    ui.glue:Hide()

    ui.boost_display:Hide()

    HideBlessCrystal()
    if ui.devil_trail_ui then
        DCEI.SetFrameActive(ui.devil_trail_ui.Frame, false)
    end
    if ui.monster_bounty and DCEI.FrameExists(ui.monster_bounty.Frame) then
        DCEI.SetFrameActive(ui.monster_bounty.Frame, false)
    end

    DCEI.TriggerAddTimerEventElapsed(function()
        if
            (
                cameras.current == cameras.town
                or cameras.current == cameras.mines
                or cameras.current == cameras.town_hall
                or cameras.current == cameras.vault
            ) and not ui.shop.is_active
        then
            --don't show town loadout if player navigated back to castle
            ui.loadout.town:Show()
            -- refresh vault progress bar
            local progress = GetVaultGold() / data.vault.capacity
            ui.loadout.town.vault.gold_indicator:Update(progress)
            if player.system.vault == system.locked then
                DCEI.SetFrameActive(ui.loadout.town.vault.frame, false)
            end
        end
    end, 0.25, false, true)

    CheckSystemUnlocks()
    UpdateTownObjective()
    CheckUpdateTownHallLabel()

    if ui.download then
        ui.download:AnimateOut()
    end

    -- check for town tutorial
    if tutorial.town_pip and not event_mode.IsActive() then
        ui.glue.district_menu:ShowPipAnimated()
    end
end

-- !SECTION

-- SECTION System checks
function CheckSystemUnlocks(map_start)
    CheckUnlockReplayWave()
    CheckUnlockTalent()
    CheckUnlockLeaderBoard()
    CheckUnlockGear()
    CheckUnlockSocial()
    CheckUnlockCalendar(map_start)
    CheckUnlockTown()
    CheckUnlockInventory()
    CheckUnlockDealsLink()
    CheckUnlockBattleStats()
    CheckLeaderboardTutorial()
    CheckUnlockTownSystemAirship()
    CheckUnlockPlayerDeck()
    CheckUnlockChallengeBattle()
    CheckUnlockRaidBossBattle()
    CheckMonsterLabUnlock()
    CheckHellModeUnlock()
    CheckHellModeEventPip()
    CheckHellModeEventLink()
    CheckOtherButtonPip()
    CheckTownBuildingUpgradable()
end

function UpdateTownObjective()
    local castle_level = GetCastleLevel()
    local max_unlock = data.unlock.max
    local next_castle_level

    -- hide by default
    ui.glue.town_objective:Hide()

    if castle_level < max_unlock then
        -- find next castle level that unlocks a town slot
        for level = castle_level + 1, max_unlock do
            if data.unlock[level] and data.unlock[level].town_slot then
                next_castle_level = level
                break
            end
        end

        if next_castle_level then
            local level_required = DCEI.GetTranslationText("general/level_number", { x = next_castle_level })
            local text = DCEI.GetTranslationText(
                "ui/town/unlock/castle",
                { x = "<color=green>" .. level_required .. "</color>" }
            )
            ui.glue.town_objective:Update(text)
        end
    end
end

function CreateBattleUI(wavestart)
    local prev_state = ui.bottom_panel.state
    local state = "battle"
    ui.bottom_panel:SetState(state)

    ui.loadout.Hide()
    ui.glue:Hide()
    ui.town_glue:Hide()
    ui.loadout.hero.Show()
    ui.loadout.hero.HideBorder()

    if ui.download and prev_state == "castle" then
        ui.download:AnimateOut()
    end

    if ui.monster_bounty and DCEI.FrameExists(ui.monster_bounty.Frame) then
        DCEI.SetFrameActive(ui.monster_bounty.Frame, false)
    end

    -- only enable boosts in standard and replay mode
    if wave.data.ModeStandard() and wavestart then
        ui.boost_display:Show()
        StartBoostTimer()
        -- only show bless in standard and replay mode
        ShowBlessCrystal()
        BlessUnitPromptup()
    end
end

-- System Unlock UIs
function CheckUnlockReplayWave()
    local unlock_wave = 1
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local frame = ui.bottom_panel.right_inner.frame

    DCEI.SetFrameActive(frame, unlock_check)
end

function CheckUnlockTalent()
    local unlock_wave = 1
    local unlock_check = wave.data.GetMax() >= unlock_wave
    ui.glue.talent_menu:SetEnabled(unlock_check)

    -- check for talent tutorial
    if not tutorial.talent and unlock_check and GetTalentPoints() > 0 and data.talents.Spent() == 0 then
        tutorial.talent = true
        ui.glue.talent_menu:ShowPipAnimated()
    end
end

function CheckUnlockLeaderBoard()
    local unlock_wave = 3
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local button = ui.glue.leader_menu

    button:SetEnabled(unlock_check)
    DCEI.SetFrameActive(button.frame, unlock_check)
end

function CheckUnlockGear()
    local unlock_wave = 15
    local unlock_check = wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()
    ui.glue.gear_menu:SetEnabled(unlock_check)
    ui.loadout.gear:UnlockGear(unlock_check)
    -- check for gear tutorial
    if not tutorial.gear and unlock_check and data.GetOwnedGearCount() == 0 then
        tutorial.gear = true
        ui.glue.gear_menu:ShowPipAnimated()
    end
end

function CheckUnlockSocial()
    local unlock_wave = 5
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local button = ui.glue.social_menu

    button:SetEnabled(unlock_check)
end

function CheckUnlockTown()
    local unlock_wave = 5
    local unlock_level = 20
    local castle_level_check = (GetCastleLevel() >= unlock_level)
    local unlock_check = (wave.data.GetMax() >= unlock_wave) or castle_level_check or event_mode.IsActive()

    local frame = ui.bottom_panel.left_inner.frame
    DCEI.SetFrameActive(frame, unlock_check)

    -- enable town building button when town buildings unlock
    ui.glue.district_menu:SetEnabled(castle_level_check)
end

function CheckTownBuildingUpgradable()
    if not player.town_building_pip and wave.data.GetMax() >= 20 and GetCastleLevel() >= 20 and not event_mode.IsActive() then
        local current_gold = GetPlayerGold()
        for _, town_building in ipairs(data.town) do
            if town_building.is_purchased then
                local cost = town_building:GetUpgradeCost()
                local level = town_building:GetLevel()
                local max_level = town_building.max_level
                if level < max_level and current_gold >= cost then
                    if ui.bottom_panel.state == "town" then
                        ui.glue.district_menu:ShowPipAnimated()
                        player.town_building_pip = true
                    else
                        ui.bottom_panel:ShowTownPip()
                    end
                    return true
                end
            end
        end
    end
end

function CheckMineUpgradable()
    local current_gold = GetPlayerGold()
    local worker_data = data.dwarf_workers
    local cost = worker_data:GetCost(GetDwarfLevel() + 1)
    local contract_needed = 1

    if GetDwarfLevel() < worker_data.max_level and GetMineContracts() >= contract_needed and current_gold >= cost then
        if ui.bottom_panel.state ~= "town" then
            ui.bottom_panel:ShowTownPip()
        end
        if not ui.loadout.town.dwarf.pip_is_active then
            ui.loadout.town.dwarf:ShowPipAnimated()
        end
    else
        ui.loadout.town.dwarf:HidePip()
    end
end

function CheckUnlockInventory()
    local unlock_wave = 5
    local unlock_level = 20
    local unlock_check = ((wave.data.GetMax() >= unlock_wave) or (GetCastleLevel() >= unlock_level)) and not event_mode.IsActive()
    local button = ui.glue.inventory_menu

    button:SetEnabled(unlock_check)
end

function CheckUnlockBattleStats()
    local unlock_wave = 10
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local button = ui.glue.battlestats_menu

    button:SetEnabled(unlock_check)
    DCEI.SetFrameActive(button.frame, unlock_check)
end

function CheckUnlockCalendar(map_start)
    local unlock_wave = 10
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local button = ui.glue.calendar_menu

    button:SetEnabled(unlock_check)
    DCEI.SetFrameActive(button.frame, unlock_check)

    -- enable the pip if calendar is hidden so pip is shown when calendar unlocks
    if unlock_check == false then
        button:ShowPip()
    end

    -- to avoid conflict with CheckServerTimeOnGameStart on map start
    if not map_start and wave.data.GetMax() == unlock_wave then
        local function ResultCallbcak(time_stamp)
            ScheduleNotification(time_stamp)
        end

        AttemptGetServerTime(ResultCallbcak)
    end
end

function CheckUnlockTownSystemAirship()
    local state = player.system.airship
    local unlock_wave = 20
    local unlock_check = wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()

    if (state == system.locked or state == system.unlocked) and unlock_check then
        player.system.airship = system.active
        bank.CommitSystemState()
    end
end

function CheckLeaderboardTutorial()
    local unlock_wave = 25
    local unlock_check = wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()
    local username_check = player.username and not player.username.name

    if not tutorial.username and username_check and unlock_check then
        tutorial.username = true
        ui.glue.leader_menu:ShowPip()
    end
end

function CheckHellModeEventPip()
    if ui.leaderboard.season_index then
        local season_id = "HM_season" .. ui.leaderboard.season_index
        local season_data = player.hell_mode[season_id]
        local player_score = season_data and season_data.total or 0
        local claim_list = player.hell_mode_event_claim_list

        for index, item in ipairs(data.hell_mode_event_reward) do
            local key = "reward" .. index
            if player_score >= item.required_score and not claim_list[key] then
                ui.glue.leader_menu:ShowPip()
                ui.leaderboard.event_reward_button:ShowPip()
                ui.CheckOtherButtonPip()
                ui.glue.floating_event:ShowPip()
                break
            end
        end
    end
end

function CheckOtherButtonPip()
    local unlock_wave = 3
    local unlock_check = wave.data.GetMax() >= unlock_wave
    local button = ui.glue.other_buttons

    button:SetEnabled(unlock_check)
    DCEI.SetFrameActive(button.frame, unlock_check)

    ui.CheckOtherButtonPip()
    if not player.system.check_seceondary_menu or player.system.check_seceondary_menu == system.locked then
        button:ShowPipAnimated()
    end
    if not unlock_check then
        button:HidePip()
    end
end

function IsChallengeBattleLocked()
    local unlock_wave = GetChallengeUnlockWave()
    return wave.data.GetMax() < unlock_wave and not tutorial.challenge and not event_mode.IsActive()
end

function CheckUnlockChallengeBattle()
    local game_state = ui.bottom_panel.state
    local nav = ui.bottom_panel.right_inner
    local nav_state = nav.states["colony"]

    if nav_state then
        nav_state.locked = IsChallengeBattleLocked()
    end

    nav_state = nav.states["raid_boss"]

    if nav_state then
        nav_state.locked = IsChallengeBattleLocked()
    end
end

function IsRaidBossBattleLocked()
    local unlock_wave = GetRaidBossUnlockWave()
    return wave.data.GetMax() < unlock_wave and not event_mode.IsActive()
end

function CheckRaidBossUnlock()
    local unlock_wave = GetRaidBossUnlockWave()
    local tutorial_raid_boss = player.system.raid_boss_tutorial
    return (tutorial_raid_boss == system.unlocked) and (wave.data.GetMax() >= unlock_wave) and not event_mode.IsActive()
end

ui.UnlockRaidBossBattle = CheckRaidBossUnlock

function CheckUnlockRaidBossBattle()
    local game_state = ui.bottom_panel.state
    local nav = ui.bottom_panel.right_outer
    local nav_state = nav.states["colony"]

    if nav_state then
        nav_state.locked = IsRaidBossBattleLocked()
    end

    nav_state = nav.states["challenge"]

    if nav_state then
        nav_state.locked = IsRaidBossBattleLocked()
    end
end

function RaidBoss_Tutorial_Check()
    if CheckRaidBossUnlock() then
        local offset = -500
        local text = tutorial.message.raid_boss
        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    end
end

function CheckUnlockTalentsTier2()
    local unlock_level = 100
    local unlock_check = (GetPlayerLevel() >= unlock_level)

    if unlock_check and ui.talent_menu.initialized then
        ui.talent_menu.UnlockTabTier2()
    end
end

function CheckUnlockParagonTalents()
    local unlock_level = 481
    local unlock_check = (GetPlayerLevel() >= unlock_level)

    if unlock_check and ui.talent_menu.initialized then
        ui.talent_menu.UnlockParagonTab()
    end
end

-- check if player deck should be unlocked
function CheckUnlockPlayerDeck()
    local unlock_wave = 100
    local unlock_check = wave.data.GetMax() >= unlock_wave or DevMode()

    DCEI.SetFrameActive(ui.glue.player_deck_menu.frame, unlock_check)
    if wave.data.GetMax() == 100 then
        ui.glue.player_deck_menu:ShowPip()
    end
end

function CheckUpdateTownHallLabel()
    local colonies_completed = #levels.colony.complete
    local unlock_check = colonies_completed >= 1

    if TestB("deferred_town_hall") then
        unlock_check = wave.data.GetMax() >= 50
    end

    if unlock_check then
        local text = DCEI.GetTranslationText("ui/town/building/town_hall")
        ui.loadout.town.hall:SetTitle(text)
        ui.loadout.town.hall:HideSubtitle()
    end
    if event_mode.IsActive() then
        DCEI.SetFrameActive(ui.loadout.town.hall.frame, false)
    end
end

function UpdateAirshipLabel()
    local state = player.system.airship

    if state == system.active then
        local text = DCEI.GetTranslationText("ui/town/building/airship")
        ui.loadout.town.airship:SetTitle(text)
        ui.loadout.town.airship:HideSubtitle()
    end

    DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
    if event_mode.IsActive() then
        DCEI.SetFrameActive(ui.loadout.town.airship.frame, false)
    end
end

function CheckUnlockDealsLink()
    local unlock_wave = 10
    local unlock_check = wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()
    -- local platform_check = PlayingMobileVersion()

    if unlock_check then
        DCEI.SetFrameActive(ui.glue.floating_deals.frame, true)
    end
end

function CheckHellModeEventLink()
    local unlock_check = ui.leaderboard.unlock_hellmode and ui.leaderboard.season_index
    local anmiation = not ui.glue.floating_event.active and unlock_check
    ui.glue.floating_event.active = unlock_check
    DCEI.SetFrameActive(ui.glue.floating_event.Frame, unlock_check)
    if anmiation then
        DCEI.AnimateFrameScale(ui.glue.floating_event.Frame, { x = 0.2, y = 0.2 }, { x = 1, y = 1 }, 0.5, "OutBack")
    end
end

-- !SECTION

-- SECTION In battle round buttons
function AttemptShowSpeedButton()
    if watch_ad_fail_count >= 2 then
        return
    end
    local unlock_speed_up = 25
    local ad_type = ad_data.speed

    -- first check if speed up is unlocked (post wave 25)
    if wave.data.GetMax() >= unlock_speed_up then
        -- then check if player is online or has speed_up time remaining
        if (DCEI.IsOnline() and DCEI.IsAdsReady() and not AdLimitReached(ad_type)) or player.speed_up then
            ui.battle.floating_speedup:Show()
        end
    end
end

function ShowAutoBattleButton(show_pip)
    local wave_instance = wave.data.instance
    local show_instance = ui.battle.floating_autobattle.instance + 1

    ui.battle.floating_autobattle:Show(show_pip)

    if player.autobattle.is_active then
        -- remove autobattle button countdown animation
        DCEI.RemoveBehaviorAsync(CASTLE_UNIT, DCEI.Behavior("System AutoBattleButton Duration"))

        ui.battle.floating_autobattle:Activate()
    else
        -- display autobattle button countdown animation
        local duration = 4 * player.speed
        DCEI.ApplyBehaviorWithDurationAsync(CASTLE_UNIT, DCEI.Behavior("System AutoBattleButton Duration"), duration)

        -- hide auto battle button after duration if inactive
        DCEI.TriggerAddTimerEventElapsed(function()
            if wave_instance == wave.data.instance and show_instance == ui.battle.floating_autobattle.instance then
                ui.battle.floating_autobattle:Hide()
            end
        end, duration, false, true)
    end
end

function CreateAutoBattlePrompt()
    local is_auto = true
    local level = wave.data.GetMax() + 1
    local cost = data.autobattle:GetCost(level)
    local auto_time = player.gem_auto_time
    local auto_data = player.autobattle

    ResetCastle()

    -- auto battle prompt uses global mask, leftover from victory message
    local function HideGlobalMask()
        ui.GlobalFadeIn(0, 0.8)
    end

    local function ShowGlobalMask()
        ui.GlobalFadeOut(0, 0.8)
    end

    local function OnCancel()
        ui.autobattle.prompt = nil
        HideGlobalMask()

        DisableAutoBattle()
        CreateCastleUI()
        CreateWaveDropResourceFlyAnimation()
        CheckToQueueDealsPopup()
        CheckForQueuedDealPopup()
        AttemptGoldAd()
        DwarfWorkers_UseUnits()
        RestoreBuildingsOnWaveEnd()
        TutorialNewGearOnboarding_Check()
        CheckRaceEventUnlock()
    end

    local function OnContinueGold()
        local gold = GetPlayerGold()

        if cost <= gold then
            SpendPlayerGold(cost)

            player.autobattle_gold = player.autobattle_gold or 0
            player.autobattle_gold = player.autobattle_gold + 1
            bank.CommitAutoBattleGoldCount()
            -- LogEvent(LogEvents.StartGoldAutoBattle(cost, true))
            HideGlobalMask()
            WaveStartActions(NextWave(), is_auto, "gold")
        else
            OnCancel()
        end
    end

    local function OnContinueGem()
        if auto_time > 0 then
            player.autobattle_gem = player.autobattle_gem or 0
            player.autobattle_gem = player.autobattle_gem + 1
            bank.CommitAutoBattleGemCount()
            -- LogEvent(LogEvents.StartGemAutoBattle(true))
            HideGlobalMask()
            WaveStartActions(NextWave(), is_auto, "gem")
            StartGemAutoBattleTimer()
        else
            OnCancel()
        end
    end

    local function OnStore()
        local on_close = function()
            DCEI.SetFrameActive(ui.autobattle.prompt.frame, true)
            ui.autobattle.prompt:ResetProgressBar()

            ui.battle:Show()
            ShowGlobalMask()
            ClearGameSpeed()

            -- show boost display
            ui.boost_display:Show()
        end

        -- log store opening from ab prompt
        local from = "autobattle prompt (continue)"
        SetStoreEntrySource(from)

        HideGlobalMask()
        DCEI.SetFrameActive(ui.autobattle.prompt.frame, false)
        ui.battle:Hide()

        -- SetGameSpeed(0)
        OpenAutoBattleStore(on_close)
        
        local item_time = shop.GetItemLazyLoadTimeRemaining() or 0
        DCEI.TriggerAddTimerEventElapsed(
            function()
                SetGameSpeed(0)
            end,
            item_time,
            true,
            true
        )

        -- hide boost display
        ui.boost_display:Hide()
    end

    local callbacks = {
        OnContinueGold = OnContinueGold,
        OnContinueGem = OnContinueGem,
        OnCancel = OnCancel,
        OnStore = OnStore,
    }

    local max_gems = 50 + player.attributes.token_cap
    if player.autobattle.auto_stop_when_token_full and DCEI.GetProtectedValue(data.safe.gems) >= max_gems then
        OnCancel()
    else
        -- immediately hide screen mask created by autobattle prompt
        ui.autobattle.prompt = ui.NewAutoBattlePrompt(callbacks, cost, auto_time, auto_data)
        ui.screen_mask.Hide()
    end
end

function AttemptSpeedAdMenu()
    if watch_ad_fail_count >= 2 then
        return
    end
    if TestB("test010715_daily_ads_limit") and player.remove_ad and player.daily_ads_limit <= 0 then
        return
    end
    local ad_type = ad_data.speed

    if not AdLimitReached(ad_type) then
        if DCEI.IsOnline() and DCEI.IsAdsReady() then
            CreateSpeedAdMenu()
        else
            ui.ShowFeedbackOffline()
        end
    end
end

function CreateSpeedAd()
    local function ResolveAds(success)
        if success then
            DCEI.TriggerAddTimerEventElapsed(function()
                AddSpeedx3Time(ad_data.speed.duration)
                SetGameSpeed(player.speed)
            end, 0.125, false, true)
            AdTrackLimit(ad_data.speed)
            AddSpeedx3AdCooldown(ad_data.speed.limit_duration)

            LogEvent(LogEvents.watch_ad("Speed"))
            player.ads_watched_speed = player.ads_watched_speed + 1

            bank.CommitAdWatchSpeed()
        else
            DCEI.LogMessage("> Playing Ads Failed")
            LogEvent(LogEvents.watch_ad_failed("Speed"))
            watch_ad_fail_count = watch_ad_fail_count + 1
            -- LogEvent(LogEvents.fail_ad("Speed"))
        end

        -- resume boost timer when battle resumes
        ResumeBoostTimer()

        SetGameSpeed(player.speed)
    end

    ShowAds(ResolveAds)
end

function CreateSpeedAdMenu()
    local function OnCancel()
        -- LogEvent(LogEvents.exit_ad("Speed"))
        SetGameSpeed(player.speed)

        -- resume boost timer when battle resumes
        ResumeBoostTimer()
    end

    local callbacks = {
        OnCancel = OnCancel,
        OnContinue = CreateSpeedAd,
        remove_ad = player.remove_ad,
    }

    -- LogEvent(LogEvents.show_ad_speed)
    ui.speedup.menu = ui.NewSpeedUpMenu(callbacks)
    SetGameSpeed(0)

    -- pause boost timer while in menus
    StopBoostTimer()
end

-- NOTE: this is a UI hookup stuff
function OpenStoreSpecials()
    local from = "glue deals"
    SetStoreEntrySource(from)

    OpenDealsStore()
end

-- !SECTION

-- SECTION Events functions
function AttemptClaimWinterEvent(item_layout, reward, index, pass_reward)
    if pass_reward and not player.winter_event2022.purchase_pass then
        local pass_name = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro")
        ui.ShowFeedbackNeedSeasonEventPass(pass_name)
        return
    end
    local prefix = pass_reward and "Pass" or "Free"
    local level_key = pass_reward and "pass_level_" or "free_level_"
    level_key = level_key .. index
    if item_layout[prefix] == "locked" then
        local coin_name = "<sprite name=general_icon_icecoin1>"
        ui.ShowFeedbackNeedSeasonEventCoin(coin_name)
    elseif item_layout[prefix] == "unlocked" and not player.winter_event2022[level_key] then
        player.winter_event2022[level_key] = true
        item_layout:Claimed(pass_reward)
        ClaimSeasonEventReward(reward)
        bank.CommitSeasonEvent("winter_event2022")
        local type = pass_reward and "pass" or "free"
        local event = LogEvents.LogClaimSeasonEventReward("Winter Event", type, index)
        LogEvent(event)
        PlayGoldSpendSound()
        CheckWinterEventRewardAllClear()
    end
end

function AttemptClaimAutumnEvent(item_layout, reward, index, pass_reward)
    if pass_reward and not player.autumn_event.purchase_pass then
        local pass_name = DCEI.GetTranslationText("ui/menu/autumn_event/pass")
        ui.ShowFeedbackNeedSeasonEventPass(pass_name)
        return
    end
    local prefix = pass_reward and "Pass" or "Free"
    local level_key = pass_reward and "pass_level_" or "free_level_"
    level_key = level_key .. index
    if item_layout[prefix] == "locked" then
        local coin_name = "<sprite name=general_icon_coin_leaf>"
        ui.ShowFeedbackNeedSeasonEventCoin(coin_name)
    elseif item_layout[prefix] == "unlocked" and not player.autumn_event[level_key] then
        player.autumn_event[level_key] = true
        item_layout:Claimed(pass_reward)
        ClaimSeasonEventReward(reward)
        bank.CommitSeasonEvent("autumn_event")
        local type = pass_reward and "pass" or "free"
        local event = LogEvents.LogClaimSeasonEventReward("Autumn Event", type, index)
        LogEvent(event)
        PlayGoldSpendSound()
        CheckAutumnEventRewardAllClear()
    end
end

function ClaimSeasonEventReward(reward)
    for reward_name, value in pairs(reward) do
        if reward_name == "gold" then
            local wave_gold = wave.data.GetGoldValue(wave.data.GetMax())
            local gold_value = math.floor(value * wave_gold)
            AddPlayerGold(gold_value)
            bank.CommitGold()
            ui.GoldRewardFeedback(gold_value)
        elseif reward_name == "token" then
            AddPlayerOverflowGems(value)
            ui.TokenRewardFeedback(value)
        elseif reward_name == "auto_battle" then
            AddGemAutoBattleTime(value * 60)
            local icon = "shop_autobattle_01"
            local icon_size = 110
            local text = "<size=120%>+" .. value .. "</size> " .. DCEI.GetTranslationText("general/minutes")
            text = "<color=green>" .. text .. "</color>"
            local offset = {
                spacing = -20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        elseif reward_name == "speed" then
            AddSpeedx3Time(value * 60)
            local icon = "shop_speedup_01"
            local icon_size = 110
            local text = "<size=120%>+" .. value .. "</size> " .. DCEI.GetTranslationText("general/minutes")
            text = "<color=green>" .. text .. "</color>"
            local offset = {
                spacing = -20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        elseif reward_name == "mine_contract" then
            AddPlayerContract_Mine(value)
            local icon = "icon_catalyst_main_item2_orange_03"
            local icon_size = 120
            local text = "x " .. value
            text = "<color=green>" .. text .. "</color>"
            local offset = {
                spacing = -20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        elseif reward_name == "gold_boost_200" then
            local type = "gold_boost_200"
            local duration = value * 60
            player.timers[type] = player.timers[type] or 0
            player.timers[type] = player.timers[type] + duration
            bank.CommitTimers()
            local icon = "icon_passability_bg_ice"
            local icon_size = 90
            local text = "<size=120%>+" .. value .. "</size> " .. DCEI.GetTranslationText("general/minutes")
            text = "<color=yellow>" .. text .. "</color>"
            local offset = {
                spacing = 20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            local layout = ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            local hstack = DCEI.CreateHStackFrame(layout.image)
            local label = DCEI.CreateTextFrame(hstack)
            DCEI.SetTextFrameFontSizeMax(label, 30)
            DCEI.SetTextFrameFontSizeMin(label, 20)
            DCEI.SetTextFrameColor(label, { r = 1, g = 0.92, b = 0.016 })
            DCEI.SetFrameMaxSize(label, 40, 70)
            DCEI.SetTextFrameText(label, "2" .. "<size=75%>x")
            local icon = DCEI.CreateFrame(hstack)
            DCEI.SetFrameImage(icon, "icon_coin00")
            DCEI.SetFrameSize(icon, 28, 28)
        elseif reward_name == "token_boost_500" then
            local type = "token_boost_500"
            local duration = value * 60
            player.timers[type] = player.timers[type] or 0
            player.timers[type] = player.timers[type] + duration
            bank.CommitTimers()
            local icon = "icon_passability_bg_ice"
            local icon_size = 90
            local text = "<size=120%>+" .. value .. "</size> " .. DCEI.GetTranslationText("general/minutes")
            text = "<color=yellow>" .. text .. "</color>"
            local offset = {
                spacing = 20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            local layout = ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            local hstack = DCEI.CreateHStackFrame(layout.image)
            local label = DCEI.CreateTextFrame(hstack)
            DCEI.SetTextFrameFontSizeMax(label, 30)
            DCEI.SetTextFrameFontSizeMin(label, 20)
            DCEI.SetTextFrameColor(label, { r = 1, g = 0.92, b = 0.016 })
            DCEI.SetFrameMaxSize(label, 40, 70)
            DCEI.SetTextFrameText(label, "5" .. "<size=75%>x")
            local icon = DCEI.CreateFrame(hstack)
            DCEI.SetFrameImage(icon, "icon_coin_dungeon00")
            DCEI.SetFrameSize(icon, 28, 28)
        elseif reward_name == "equipment" then
            local equipment_data = Deepcopy(value)
            equipment_data.level = 1
            table.insert(player.inventory.items, equipment_data)
            bank.CommitEquipments()
            local icon = value.type.icon[value.rarity.name]
            local icon_size = 100
            local text = ""
            local offset = {
                spacing = 6,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        elseif reward_name == "hero_wolfhuntress" then
            local hero_data = data.heroes["Hero WolfHuntress"]
            PurchaseHero(hero_data)
            UpdateHeroList()
            CreateClaimNewHeroPopup(hero_data)
        elseif reward_name == "hero_windknight" then
            local hero_data = data.heroes["Hero WindKnight"]
            PurchaseHero(hero_data)
            UpdateHeroList()
            CreateClaimNewHeroPopup(hero_data)
        else
            if reward_name == "raid_key" then
                AddPlayerRaidKeyExtra(value)
            elseif reward_name == "raid_turbo" then
                AddPlayerRaidTurbo(value)
            elseif string.match(reward_name, "chest") then
                AddPlayerChest(reward_name, value)
            end
            local layout_data = data.winter_event_reward.item_layout[reward_name]
            if layout_data then
                local icon = layout_data.icon
                local icon_size = 100
                local text = "<size=120%>+" .. value .. "</size>"
                local offset = {
                    spacing = 6,
                    x = 0,
                    y1 = -20,
                    y2 = 60,
                }
                ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            end
        end
    end
end

function CreateClaimNewHeroPopup(hero_data)
    local layout = gmui.Layout:New({
        parent = ui.super,
        name = "SeasonEvent/PassMenu/HeroClaim",
    })

    local title = DCEI.GetTranslationText("ui/menu/challenge/unlocked") .. DCEI.GetTranslationText("ui/challenge/hero")
    DCEI.SetTextFrameText(layout.Title, title)
    DCEI.SetFrameImage(layout.HeroPortrait, hero_data.portrait)

    function layout:Close()
        if layout and DCEI.FrameExists(layout.Frame) then
            DCEI.DestroyFrame(layout.Frame)
        end
    end

    DCEI.SetOnClickCallback(layout.CloseButton, function()
        layout:Close()
    end)

    DCEI.SetOnClickCallback(layout.GoButton, function()
        layout:Close()
        if ui.winter_event_menu then
            ui.winter_event_menu:Hide()
        end
        if ui.autumn_event_menu then
            ui.autumn_event_menu:Hide()
        end
        local unit_index = GetHeroGalleryIndex(hero_data)
        ui.loadout.hero.select = GetEmptyHeroSlot()
        ui.loadout.current = ui.loadout.hero
        CreateHeroList(unit_index)
        local unlock_animation = true
        ui.CreateNewHeroGlow(ui.list_hero.items[unit_index], unlock_animation)
    end)
end

function AttemptClaimWinterEventBonusChest()
    if not player.winter_event2022.purchase_pass then
        CreateWinterPassPopup()
        return
    end
    local max_level = #data.winter_event_level_exp

    local required_coin = data.winter_event_level_exp[max_level]
    local current_coin = GetPlayerWinterCoin2022()
    current_coin = current_coin - data.winter_event_level_exp.total_require[max_level - 1]
    if player.winter_event2022.level >= max_level and current_coin >= required_coin then
        -- remove 600 coin after claiming
        AddPlayerWinterCoin2022(-required_coin)
        -- claim reward
        local new_equipment = GetEquipmentFromChest("rare_chest_l")

        local icon = new_equipment.type.icon[new_equipment.rarity.name]
        local icon_size = 120
        local text = "<size=120%>+1</size>"
        local offset = {
            spacing = 6,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)

        local event = LogEvents.LogClaimSeasonEventReward("Winter Event", "Bonus Chest")
        LogEvent(event)
        PlayGoldSpendSound()
        CheckWinterEventRewardAllClear()
        -- update page
        UpdateWinterEventMenu()
    else
        local coin_name = "<sprite name=general_icon_icecoin1>"
        ui.ShowFeedbackNeedSeasonEventCoin(coin_name)
    end
end

function AttemptClaimAutumnEventBonusChest()
    if not player.autumn_event.purchase_pass then
        CreateAutumnPassPopup()
        return
    end
    local max_level = #data.autumn_event_level_exp

    local required_coin = data.autumn_event_level_exp[max_level]
    local current_coin = GetPlayerAutumnCoin()
    current_coin = current_coin - data.autumn_event_level_exp.total_require[max_level - 1]
    if player.autumn_event.level >= max_level and current_coin >= required_coin then
        -- remove 600 coin after claiming
        AddPlayerAutumnCoin(-required_coin)
        -- claim reward
        local new_equipment = GetEquipmentFromChest("rare_chest_l")

        local icon = new_equipment.type.icon[new_equipment.rarity.name]
        local icon_size = 120
        local text = "<size=120%>+1</size>"
        local offset = {
            spacing = 6,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)

        local event = LogEvents.LogClaimSeasonEventReward("Autumn Event", "Bonus Chest")
        LogEvent(event)
        PlayGoldSpendSound()
        CheckAutumnEventRewardAllClear()
        -- update page
        UpdateAutumnEventMenu()
    else
        local coin_name = "<sprite name=general_icon_coin_leaf>"
        ui.ShowFeedbackNeedSeasonEventCoin(coin_name)
    end
end

function AttemptBuyWinterPass()
    local product_id = "com.wildsky.castle.iap.winterpass.01"
    AttemptAnyItemPurchase(product_id)

    if ui.winter_pass_popup then
        ui.winter_pass_popup:Close()
    end
    UpdateWinterEventMenu()
end

function AttemptBuyAutumnPass()
    local product_id = "com.wildsky.castle.iap.autumnpass.01"
    AttemptAnyItemPurchase(product_id)

    if ui.autumn_pass_popup then
        ui.autumn_pass_popup:Close()
    end
    UpdateAutumnEventMenu()
end

function CreateWinterPassPopup()
    if ui.winter_pass_popup then
        ui.winter_pass_popup:Close()
    end
    local layout_name = "WinterEvent/Popup"
    local call_backs = {
        purchase_pass = AttemptBuyWinterPass,
    }
    local pass_name = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro")
    ui.winter_pass_popup = ui.CreateSeasonEventPassPopup(layout_name, call_backs, pass_name)
    if player.winter_event2022_timer then
        player.winter_event2022_timer:Tick()
    end

    local product_id = "com.wildsky.castle.iap.winterpass.01"
    local localized_price = shop.GetProductPriceLocalized(product_id)
    local default_price = "¥" .. shop.GetProductPriceUSD(product_id)
    local price_text = localized_price or default_price
    DCEI.SetTextFrameText(ui.winter_pass_popup.PriceLabel, price_text)
end

function CreateAutumnPassPopup()
    if ui.autumn_pass_popup then
        ui.autumn_pass_popup:Close()
    end
    local layout_name = "AutumnEvent/Popup"
    local call_backs = {
        purchase_pass = AttemptBuyAutumnPass,
    }
    local pass_name = DCEI.GetTranslationText("ui/menu/autumn_event/pass")
    ui.autumn_pass_popup = ui.CreateSeasonEventPassPopup(layout_name, call_backs, pass_name)
    if player.autumn_event_timer then
        player.autumn_event_timer:Tick()
    end

    local product_id = "com.wildsky.castle.iap.autumnpass.01"
    local localized_price = shop.GetProductPriceLocalized(product_id)
    local default_price = "¥" .. shop.GetProductPriceUSD(product_id)
    local price_text = localized_price or default_price
    DCEI.SetTextFrameText(ui.autumn_pass_popup.PriceLabel, price_text)
end

function AttemptStartWinterEventTimer(time_stamp)
    if TestB("winter_event2022") then
        if player.autumn_event_timer then
            local new_hero = data.heroes.gallery[#data.heroes.gallery]
            if not new_hero.is_purchased then
                for index, hero in ipairs(data.heroes.gallery) do
                    if hero.name == "Hero WolfHuntress" then
                        table.remove(data.heroes.gallery, #data.heroes.gallery)
                        break
                    end
                end
            end
            -- don't trigger winter event if player haven't finish last
            return
        end
        if player.winter_event2022_timer then
            player.winter_event2022_timer:Destroy()
        end
        -- sometims player's connection failed and get 0 time stamp, so get the time again if the expire_time is 20 years ago
        if
            not player.winter_event2022.expire_time
            or (player.winter_event2022.expire_time and player.winter_event2022.expire_time < 946684800)
        then
            player.winter_event2022.expire_time = time_stamp + 30 * 24 * 3600
            bank.CommitSeasonEvent("winter_event2022")
        end
        local expire_time = player.winter_event2022.expire_time
        local time_remaining = expire_time - time_stamp
        if time_remaining > 0 then
            local timer = core.OsTimer:New()
            player.winter_event2022_timer = timer

            timer:AddDuration(time_remaining)

            function timer:Tick()
                local time = self:GetTimeRemaining()
                ui.glue.floating_winter_event:TimerUpdate(time)
                if ui.winter_event_menu and ui.winter_event_menu.is_active then
                    ui.winter_event_menu:TimerUpdate(time)
                end
                if ui.winter_pass_popup and ui.winter_pass_popup.is_active then
                    ui.winter_pass_popup:TimerUpdate(time)
                end
            end

            function timer:Action()
                WinterEventExpire()
                player.winter_event2022_timer = nil
            end

            timer:Tick()
            local animate_pip = true
            CheckWinterEventUnlock(animate_pip)

            -- update exp fill
            local current_exp, required_exp = GetWinterPassExpFill()
            local fillamount = current_exp / required_exp
            ui.glue.floating_winter_event:UpdateFill(0, fillamount)
        else
            LogSeasonEnds("winter_event2022")
        end
    end
end

function AttemptStartAutumnEventTimer(time_stamp)
    if TestB("autumn_event") then
        if player.summer_event_timer then
            local new_hero = data.heroes.gallery[#data.heroes.gallery]
            if not new_hero.is_purchased then
                for index, hero in ipairs(data.heroes.gallery) do
                    if hero.name == "Hero WindKnight" then
                        table.remove(data.heroes.gallery, #data.heroes.gallery)
                        break
                    end
                end
            end
            -- don't trigger autumn event if player haven't finish summner event yet
            return
        end
        if player.autumn_event_timer then
            player.autumn_event_timer:Destroy()
        end
        -- sometims player's connection failed and get 0 time stamp, so get the time again if the expire_time is 20 years ago
        if
            not player.autumn_event.expire_time
            or (player.autumn_event.expire_time and player.autumn_event.expire_time < 946684800)
        then
            player.autumn_event.expire_time = time_stamp + 30 * 24 * 3600
            bank.CommitSeasonEvent("autumn_event")
        end
        local expire_time = player.autumn_event.expire_time
        local time_remaining = expire_time - time_stamp
        if time_remaining > 0 then
            local timer = core.OsTimer:New()
            player.autumn_event_timer = timer

            timer:AddDuration(time_remaining)

            function timer:Tick()
                local time = self:GetTimeRemaining()
                ui.glue.floating_autumn_event:TimerUpdate(time)
                if ui.autumn_event_menu and ui.autumn_event_menu.is_active then
                    ui.autumn_event_menu:TimerUpdate(time)
                end
                if ui.autumn_pass_popup and ui.autumn_pass_popup.is_active then
                    ui.autumn_pass_popup:TimerUpdate(time)
                end
            end

            function timer:Action()
                AutumnEventExpire()
                player.autumn_event_timer = nil
            end

            timer:Tick()
            local animate_pip = true
            CheckAutumnEventUnlock(animate_pip)

            -- update exp fill
            local current_exp, required_exp = GetAutumnPassExpFill()
            local fillamount = current_exp / required_exp
            ui.glue.floating_autumn_event:UpdateFill(0, fillamount)
        else
            LogSeasonEnds("autumn_event")
        end
    end
end

function CheckWinterEventRewardAllClear(animate_pip)
    local unclaimed_reward = false
    for i = player.winter_event2022.level, 1, -1 do
        local free_key = "free_level_" .. i
        local pass_key = "pass_level_" .. i
        if not player.winter_event2022[free_key] then
            unclaimed_reward = true
            break
        end
        if player.winter_event2022.purchase_pass and not player.winter_event2022[pass_key] then
            unclaimed_reward = true
            break
        end
    end

    local max_level = #data.winter_event_level_exp
    local winter_coin = GetPlayerWinterCoin2022()

    if
        player.winter_event2022.level >= max_level
        and (
            winter_coin - data.winter_event_level_exp.total_require[max_level]
            >= data.winter_event_level_exp[max_level]
        )
    then
        unclaimed_reward = true
    end

    if unclaimed_reward then
        ui.glue.floating_winter_event:ShowPip(animate_pip)
    else
        ui.glue.floating_winter_event:HidePip()
    end

    return unclaimed_reward
end

function CheckAutumnEventRewardAllClear(animate_pip)
    local unclaimed_reward = false
    for i = player.autumn_event.level, 1, -1 do
        local free_key = "free_level_" .. i
        local pass_key = "pass_level_" .. i
        if not player.autumn_event[free_key] then
            unclaimed_reward = true
            break
        end
        if player.autumn_event.purchase_pass and not player.autumn_event[pass_key] then
            unclaimed_reward = true
            break
        end
    end

    local max_level = #data.autumn_event_level_exp
    local autumn_coin = GetPlayerAutumnCoin()

    if
        player.autumn_event.level >= max_level
        and (
            autumn_coin - data.autumn_event_level_exp.total_require[max_level]
            >= data.autumn_event_level_exp[max_level]
        )
    then
        unclaimed_reward = true
    end

    if unclaimed_reward then
        ui.glue.floating_autumn_event:ShowPip(animate_pip)
    else
        ui.glue.floating_autumn_event:HidePip()
    end

    return unclaimed_reward
end

function CheckSeasonEventExist(event_name)
    if event_mode.IsActive() then
        return false
    end
    local event_key = event_name .. "_timer"
    return player[event_key] and player[event_key]:GetTimeRemaining() > 0
end

function CheckWinterEventUnlock(animate_pip, do_not_update_level)
    local unlock_wave = 60
    wave.data.winter_event = false
    if CheckSeasonEventExist("winter_event2022") then
        if wave.data.GetMax() >= unlock_wave then
            wave.data.winter_event = true
        end
        if wave.data.GetMax() > unlock_wave then
            ui.glue.floating_winter_event:Show()
            if not player.winter_event2022.purchase_pass then
                ui.shop.sections["winter_pass"]:Show()
            end
            if not player.system.first_winter_event or player.system.first_winter_event == system.locked then
                ui.glue.floating_winter_event:ShowPip(true)
                player.system.first_winter_event = system.unlocked
                bank.CommitSystemState()
            end
            if not do_not_update_level then
                -- update level and reward pip
                local winter_coin = GetPlayerWinterCoin2022()
                for i = 1, #data.winter_event_level_exp.total_require do
                    player.winter_event2022.level = i
                    if winter_coin < data.winter_event_level_exp.total_require[i] then
                        break
                    end
                end
            end
            local unclaimed_reward = CheckWinterEventRewardAllClear(animate_pip)
        end
    end
end

function CheckAutumnEventUnlock(animate_pip, do_not_update_level)
    local unlock_wave = 60
    wave.data.autumn_event = false
    if CheckSeasonEventExist("autumn_event") then
        if wave.data.GetMax() >= unlock_wave then
            wave.data.autumn_event = true
        end
        if wave.data.GetMax() > unlock_wave then
            ui.glue.floating_autumn_event:Show()
            if not player.autumn_event.purchase_pass then
                ui.shop.sections["autumn_pass"]:Show()
            end
            if not player.system.first_autumn_event or player.system.first_autumn_event == system.locked then
                ui.glue.floating_autumn_event:ShowPip(true)
                player.system.first_autumn_event = system.unlocked
                bank.CommitSystemState()
            end
            if not do_not_update_level then
                -- update autumn level and reward pip
                local autumn_coin = GetPlayerAutumnCoin()
                for i = 1, #data.autumn_event_level_exp.total_require do
                    player.autumn_event.level = i
                    if autumn_coin < data.autumn_event_level_exp.total_require[i] then
                        break
                    end
                end
            end
            local unclaimed_reward = CheckAutumnEventRewardAllClear(animate_pip)
        end
    end
end

function LogSeasonEnds(event_key)
    local season_table = player[event_key]
    if not season_table.log_end then
        local season_level = season_table.level or 0
        local user_prop = UserProps.SeasonEventLevel(event_key, season_level)
        LogUserProperty(user_prop)
        local event = LogEvents.SeasonEventEnds(event_key, season_level, season_table.purchase_pass)
        LogEvent(event)
        season_table.log_end = true
        bank.CommitSeasonEvent(event_key)
    end
end

function WinterEventExpire()
    ui.glue.floating_winter_event:Hide()
    ui.shop.sections["winter_pass"]:Hide()
    wave.data.winter_event = false
    LogSeasonEnds("winter_event2022")
end

function AutumnEventExpire()
    ui.glue.floating_autumn_event:Hide()
    ui.shop.sections["autumn_pass"]:Hide()
    wave.data.autumn_event = false
    LogSeasonEnds("autumn_event")
end

function GetWinterPassExpFill(current_exp)
    current_exp = current_exp or GetPlayerWinterCoin2022()
    player.winter_event2022.level = player.winter_event2022.level or 1
    if player.winter_event2022.level > 1 then
        current_exp = current_exp - data.winter_event_level_exp.total_require[player.winter_event2022.level - 1]
    end
    local required_exp = data.winter_event_level_exp[player.winter_event2022.level]

    return current_exp, required_exp
end

function GetAutumnPassExpFill(current_exp)
    current_exp = current_exp or GetPlayerAutumnCoin()
    player.autumn_event.level = player.autumn_event.level or 1
    if player.autumn_event.level > 1 then
        current_exp = current_exp - data.autumn_event_level_exp.total_require[player.autumn_event.level - 1]
    end
    local required_exp = data.autumn_event_level_exp[player.autumn_event.level]

    return current_exp, required_exp
end

function UpdateWinterEventMenu()
    local current_exp, required_exp = GetWinterPassExpFill()

    local wave_gold = wave.data.GetGoldValue(wave.data.GetMax())
    ui.winter_event_menu:Update(player.winter_event2022, current_exp, required_exp, wave_gold)

    local fillamount = current_exp / required_exp
    ui.glue.floating_winter_event:UpdateFill(0, fillamount)

    if ui.glue.floating_winter_event.play_upgrade_animation then
        ui.glue.floating_winter_event.play_upgrade_animation = false
        BlockInput()
        local animation_duration = ui.winter_event_menu:UpgradeAnimation(player.winter_event2022, current_exp, required_exp)
        DCEI.TriggerAddTimerEventElapsed(function()
            UnblockInput()
        end, animation_duration, true, true)
    end
end

function UpdateAutumnEventMenu()
    local current_exp, required_exp = GetAutumnPassExpFill()

    local wave_gold = wave.data.GetGoldValue(wave.data.GetMax())
    ui.autumn_event_menu:Update(player.autumn_event, current_exp, required_exp, wave_gold)

    local fillamount = current_exp / required_exp
    ui.glue.floating_autumn_event:UpdateFill(0, fillamount)

    if ui.glue.floating_autumn_event.play_upgrade_animation then
        ui.glue.floating_autumn_event.play_upgrade_animation = false
        BlockInput()
        local animation_duration = ui.autumn_event_menu:UpgradeAnimation(player.autumn_event, current_exp, required_exp)
        DCEI.TriggerAddTimerEventElapsed(function()
            UnblockInput()
        end, animation_duration, true, true)
    end
end

function OpenWinterEvent()
    UnblockInput()
    ui.DestroyBigHand()
    if not ui.winter_event_menu then
        local call_backs = {
            claim_callback = AttemptClaimWinterEvent,
            claim_bonus_chest = AttemptClaimWinterEventBonusChest,
            event_pass_popup = CreateWinterPassPopup,
        }
        local event_info = {
            layout_name = "WinterEvent/PassMenu",
            level_item_layout = "SeasonEvent/PassMenu/Item",
            pass_label = DCEI.GetTranslationText("ui/menu/winter_event/pass/pro"),
            event_description = DCEI.GetTranslationText("ui/menu/event/pass/description", {
                x = "<sprite name=general_icon_icecoin1>",
                y = DCEI.GetTranslationText("ui/menu/winter_event/title"),
            }),
            bonus_chest_description = DCEI.GetTranslationText(
                "ui/menu/bonus_chest/description",
                { x = "<color=yellow>" .. DCEI.GetTranslationText("ui/menu/winter_event/pass/pro") .. "</color>" }
            ),
        }
        ui.winter_event_menu = ui.CreateSeasonEventMenu(event_info, data.winter_event_reward, call_backs)
    end
    ui.winter_event_menu:Show()
    UpdateWinterEventMenu()
    if player.winter_event2022_timer then
        player.winter_event2022_timer:Tick()
    end
end

function OpenAutumnEvent()
    UnblockInput()
    ui.DestroyBigHand()
    if not ui.autumn_event_menu then
        local call_backs = {
            claim_callback = AttemptClaimAutumnEvent,
            claim_bonus_chest = AttemptClaimAutumnEventBonusChest,
            event_pass_popup = CreateAutumnPassPopup,
        }
        local event_info = {
            layout_name = "AutumnEvent/PassMenu",
            level_item_layout = "SeasonEvent/PassMenu/Item",
            pass_label = DCEI.GetTranslationText("ui/menu/autumn_event/pass"),
            event_description = DCEI.GetTranslationText("ui/menu/event/pass/description", {
                x = "<sprite name=general_icon_coin_leaf>",
                y = DCEI.GetTranslationText("ui/menu/autumn_event/title"),
            }),
            bonus_chest_description = DCEI.GetTranslationText(
                "ui/menu/bonus_chest/description",
                { x = "<color=yellow>" .. DCEI.GetTranslationText("ui/menu/autumn_event/pass") .. "</color>" }
            ),
        }
        ui.autumn_event_menu = ui.CreateSeasonEventMenu(event_info, data.autumn_event_reward, call_backs)
    end
    ui.autumn_event_menu:Show()
    UpdateAutumnEventMenu()
    if player.autumn_event_timer then
        player.autumn_event_timer:Tick()
    end
end

function CreateSeasonEventMonsterIntroIcon(index)
    if not ui.new_monster_intro then
        local intro = gmui.Layout:New({
            parent = ui.world,
            name = "WinterEvent/NewMonsterIcon",
        })
        DCEI.AnimateFrameScale(intro.Frame, { x = 0.4, y = 0.4 }, { x = 0.8, y = 0.8 }, 0.5, "OutBack")
        local has_event = CheckSeasonEventExist("winter_event2022")
        local portrait = has_event and data.winter_event_monsters[index].portrait
            or data.autumn_event_monsters[index].portrait
        DCEI.SetFrameImage(intro.Portrait, portrait)
        DCEI.SetFrameBottomAlignmentInParent(intro.Frame)
        DCEI.SetFrameVerticalOffsetInParent(intro.Frame, 800)
        DCEI.SetFrameRightAlignmentInParent(intro.Frame)
        DCEI.SetFrameHorizontalOffsetInParent(intro.Frame, -140)
        DCEI.SetOnClickCallback(intro.Frame, function()
            SetGameSpeed(0)
            CreateSeasonEventMonsterDescription(index)
            if ui.new_monster_intro and DCEI.FrameExists(ui.new_monster_intro.Frame) then
                DCEI.DestroyFrame(ui.new_monster_intro.Frame)
                ui.new_monster_intro = nil
                local tutorial_key = has_event and "winter_event_monster_" .. index
                    or "autumn_event_monster_" .. index
                player.system[tutorial_key] = system.unlocked
                bank.CommitSystemState()
            end
        end)
        ui.new_monster_intro = intro
    end
end

function CreateSeasonEventMonsterDescription(index)
    if not ui.new_monster_description then
        local layout = gmui.Layout:New({
            parent = ui.upper,
            name = "WinterEvent/MonsterDescription",
        })
        local has_event = CheckSeasonEventExist("winter_event2022")
        local monster_data = has_event and data.winter_event_monsters[index] or data.autumn_event_monsters[index]
        DCEI.SetTextFrameText(layout.Title, monster_data.display_name or "")
        DCEI.SetTextFrameText(layout.Description, '<align="left"><line-height=110%>' .. monster_data.description)
        DCEI.SetFrameImage(layout.Portrait, monster_data.portrait)
        for i = 1, monster_data.rarity do
            local icon = DCEI.CreateFrame(layout.SnowCoinRarity)
            DCEI.SetFrameImage(icon, has_event and "general_icon_icecoin1" or "general_icon_coin_leaf")
            DCEI.SetFrameSize(icon, 60, 60)
        end
        DCEI.SetOnClickCallback(layout.OKButton, function()
            SetGameSpeed(player.speed)
            if ui.new_monster_description and DCEI.FrameExists(ui.new_monster_description.Frame) then
                DCEI.DestroyFrame(ui.new_monster_description.Frame)
                ui.new_monster_description = nil
            end
        end)
        ui.new_monster_description = layout
    end
end

function AttemptStartEggHuntTimer(time_stamp)
    if not time_stamp then
        return
    end
    if not player.egg_hunt_event.expire_time then
        player.egg_hunt_event.expire_time = time_stamp + EGG_HUNT_DURATION
    end
    if time_stamp == 0 or time_stamp > player.egg_hunt_event.expire_time then
        return
    end

    if player.egg_hunt_timer then
        player.egg_hunt_timer:Destroy()
    end
    local time_remaining = player.egg_hunt_event.expire_time - time_stamp
    if time_remaining > 0 then
        local timer = core.OsTimer:New()
        player.egg_hunt_timer = timer

        timer:AddDuration(time_remaining)

        function timer:Tick()
            local time = self:GetTimeRemaining()
            ui.egg_hunt_banner:TimerUpdate(time)
        end

        function timer:Action()
            EggHuntExpire()
            player.egg_hunt_timer = nil
        end

        timer:Tick()

        if player.egg_hunt_event.raw > 0 then
            local previous_egg = player.egg_hunt_event.egg
            local current_egg = previous_egg + player.egg_hunt_event.raw
            player.egg_hunt_event.raw = 0
            local required_exp = EGG_UPGRADE_REQUIRE
            while current_egg > required_exp do
                current_egg = current_egg - required_exp
                player.egg_hunt_event.gift = player.egg_hunt_event.gift + 1
            end
            player.egg_hunt_event.egg = current_egg
            bank.CommitEventEggHunt()
        end
        -- update exp fill
        local current_exp = player.egg_hunt_event.egg
        local fillamount = current_exp / EGG_UPGRADE_REQUIRE
        ui.egg_hunt_banner:UpdateFill(player.egg_hunt_event.gift, 0, fillamount)
        CheckEggHuntUnlock()
    end
end

function EggHuntExpire()
    ui.egg_hunt_banner:Hide()
    wave.data.egg_hunt_event = false
end

function CheckEggHuntUnlock()
    local unlock_wave = 30
    wave.data.egg_hunt_event = false
    if player.egg_hunt_timer and player.egg_hunt_timer:GetTimeRemaining() > 0 then
        if wave.data.GetMax() >= unlock_wave then
            wave.data.egg_hunt_event = true
        end
        if wave.data.GetMax() > unlock_wave then
            ui.egg_hunt_banner:Show()
        end
    end
end

function AttemptOpenEggHuntGift()
    if player.egg_hunt_event.gift > 0 then
        if ui.egg_hung_spin and DCEI.FrameExists(ui.egg_hung_spin.Frame) then
            ui.egg_hung_spin:Close()
        end
        local callbacks = {
            count = player.egg_hunt_event.gift,
            pool_data = data.egg_hunt_gift_pool,
            update_reward_pool = UpdateEggHuntGiftRewardPool,
            spin_reward = EggHuntSpinReward,
            get_spin_reward = EggHuntGetSpinReward,
        }
        ui.egg_hung_spin = ui.CreateEggHuntSpin(callbacks)
        UpdateEggHuntGiftRewardPool()
    else
        local text = DCEI.GetTranslationText("ui/menu/not_enough_egg")
        ShowFeedbackMessage("<size=80%>" .. text)
    end
end

function UpdateEggHuntGiftRewardPool()
    if ui.egg_hung_spin and DCEI.FrameExists(ui.egg_hung_spin.Frame) then
        data.egg_hunt_gift_pool[1].gold_value = data.egg_hunt_gift_pool[1].Value(wave.data.GetMax())
        local spin_pool = {}
        for i = 1, 6 do
            local rd = math.random(1, data.egg_hunt_gift_pool.total_weight)
            local weight = 0
            for k, item in ipairs(data.egg_hunt_gift_pool) do
                local get_result = false
                for _, v in ipairs(item.pool) do
                    weight = weight + v.weight
                    if rd <= weight then
                        get_result = true
                        local new_item = {
                            type = item.type,
                            icon = item.icon,
                            color = item.color,
                            icon_scale = item.icon_scale,
                            value = v.value or v.multiplier * item.gold_value,
                            weight = v.weight,
                            rare = item.rare,
                        }
                        table.insert(spin_pool, new_item)
                        break
                    end
                end
                if get_result then
                    break
                end
            end
        end

        ui.egg_hung_spin:Update(spin_pool)
    end
end

function EggHuntSpinReward(spin_pool)
    if player.egg_hunt_event.gift <= 0 then
        ui.egg_hung_spin:Close()
        return
    end
    BlockInput()
    player.egg_hunt_event.gift = player.egg_hunt_event.gift - 1
    for i = 1, #spin_pool do
        local item_data = spin_pool[i]
        DCEI.SetFrameRotation3D(item_data.slot.frame, 0, 0, -item_data.slot.angle)
    end

    -- roll for reward
    local rd = math.random(1, #spin_pool)
    ui.egg_hung_spin:Spin(0, 0.25, 0, spin_pool[rd])

    local current_exp = player.egg_hunt_event.egg
    local fillamount = current_exp / EGG_UPGRADE_REQUIRE
    ui.egg_hunt_banner:UpdateFill(player.egg_hunt_event.gift, 0, fillamount)
end

function EggHuntGetSpinReward(item)
    UnblockInput()
    DCEI.PlaySound(DCEI.Sound("Gacha"))
    player.egg_hunt_event.open_gift = player.egg_hunt_event.open_gift + 1
    bank.CommitEventEggHunt()
    local event = LogEvents.SpinEggHuntGift(player.egg_hunt_event.open_gift)
    LogEvent(event)

    if item.type == "medal" then
        -- for medals/tokens/"gems"
        AddPlayerOverflowGems(item.value)
        ui.TokenRewardFeedback(item.value)
    elseif item.type == "gold" then
        -- for gold
        local value = math.floor(item.value)
        AddPlayerGold(value)
        bank.CommitGold()
        ui.GoldRewardFeedback(value)
    elseif item.type == "autobattle" then
        -- for auto battle time
        AddGemAutoBattleTime(item.value)
        local icon = "shop_autobattle_01"
        local icon_size = 110
        local minutes = math.floor(item.value / 60)
        local text = "<size=120%>" .. minutes .. "</size>" .. " " .. DCEI.GetTranslationText("general/min")
        text = "<color=green>" .. text .. "</color>"
        local offset = {
            spacing = -20,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
    elseif item.type == "steel_chest" then
        AddPlayerChest(item.type, item.value)
        local icon = item.icon
        local icon_size = 100
        local text = item.color .. "\n<size=120%>+" .. item.value .. "</size></color>"
        local offset = {
            spacing = 6,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
    elseif item.type == "speed" then
        AddSpeedx3Time(item.value)
        local icon = "shop_speedup_01"
        local icon_size = 110
        local minutes = math.floor(item.value / 60)
        local text = "<size=120%>" .. minutes .. "</size>" .. " " .. DCEI.GetTranslationText("general/min")
        text = "<color=green>" .. text .. "</color>"
        local offset = {
            spacing = -20,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
    end
end

function CheckRaceEventUnlock()
    local unlock_wave = 150
    if wave.data.GetMax() >= unlock_wave then
        AttemptStartRaceEventTimer(os.time())
    end
end

function AttemptStartRaceEventTimer(time_stamp)
    -- don't start race event if egg hunt event exists
    if wave.data.egg_hunt_event or event_mode.IsActive() then
        return
    end
    -- check restart race event every 5 minutes
    if player.race_event.event_end and player.race_event.expire_time then
        if time_stamp - player.race_event.expire_time >= RACE_EVENT_COOLDOWN then
            local race_event_count = player.race_event.cum_count or 1
            player.race_event = {}
            player.race_event.cum_count = race_event_count + 1
        end
    end
    -- don't start race event if there already is, or the event ends
    if player.race_event_timer or player.race_event.event_end then
        return
    end
    if not time_stamp then
        return
    end

    if not player.race_event.expire_time then
        player.race_event.expire_time = time_stamp + RACE_EVENT_DURATION
        bank.CommitRaceEvent()
        local event = LogEvents.RaceEvent("Trigger", 0)
        LogEvent(event)
    end
    if time_stamp == 0 or time_stamp > player.race_event.expire_time then
        player.race_event.event_end = true
        bank.CommitRaceEvent()
        return
    end

    local time_remaining = player.race_event.expire_time - time_stamp
    if time_remaining > 0 then
        ui.race_event_banner = CreateRaceEventBanner()
        local timer = core.OsTimer:New()
        player.race_event_timer = timer

        timer:AddDuration(time_remaining)

        function timer:Tick()
            local time = self:GetTimeRemaining()
            ui.race_event_banner:TimerUpdate(time)
            if ui.race_event_menu then
                ui.race_event_menu:TimerUpdate(time)
            end
        end

        function timer:Action()
            RaceEventEnd()
            player.race_event_timer = nil
        end

        timer:Tick()
        AttemptFetchRaceEventData()
    end
end

function RaceEventEnd()
    ui.race_event_banner:Hide()
    player.race_event.event_end = true
    bank.CommitRaceEvent()
    if CheckRaceDataReady() then
        local event = LogEvents.RaceEvent("End", player.race_event.players[1].rank)
        LogEvent(event)
    end
end

function ClaimRaceEventReward()
    local chest_num = 3
    local reward_chests = {
        "uncommon_chest",
        "massive_chest",
        "steel_chest",
    }
    local rank_chests_icon = {
        "hero_chest_t3",
        "hero_chest_t5",
        "hero_chest_t2",
    }
    if player.race_event.rank_reward and not player.race_event.reward_claimed then
        local chest_type = reward_chests[player.race_event.rank_reward]
        AddPlayerChest(chest_type, chest_num)
        local icon = rank_chests_icon[player.race_event.rank_reward]
        local icon_size = 100
        local text = "<size=120%>+" .. chest_num .. "</size>"
        local offset = {
            spacing = 6,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        player.race_event.event_set = true
        player.race_event.reward_claimed = true
        bank.CommitRaceEvent()
        if ui.race_event_menu and DCEI.FrameExists(ui.race_event_menu.Frame) then
            DCEI.SetFrameActive(ui.race_event_menu.ClaimButton, false)
            ui.race_event_menu:Close()
        end
    end
end

function OpenRaceEventMenu()
    if not ui.race_event_menu then
        local cum_count = player.race_event.cum_count or 1
        -- if not accept, display start panel, else, display event panel
        local layout_name = (cum_count % 2 == 0) and "RacePortraitEvent/Start" or "RaceEvent/Start"
        if player.race_event.accept then
            layout_name = (cum_count % 2 == 0) and "RacePortraitEvent/Menu" or "RaceEvent/Menu"
        end

        local layout = gmui.Layout:New({
            parent = ui.super,
            name = layout_name,
        })

        ui.race_event_menu = layout

        function layout:TimerUpdate(time)
            local text = ui.FormatTimeShort(time)
            DCEI.SetTextFrameText(self.TimeLabel, text)
        end

        function layout:Close()
            if player.race_event.accept then
                self:Hide()
                if player.race_event.event_set and player.race_event.reward_claimed then
                    RaceEventEnd()
                    ui.race_event_menu = nil
                    DCEI.DestroyFrame(layout.Frame)
                end
            else
                DCEI.DestroyFrame(self.Frame)
                ui.race_event_menu = nil
            end
        end

        local text2 = DCEI.GetTranslationText("ui/menu/sky_race/description", { x = RACE_EVENT_TOTAL_PROGRESS })
        if not player.race_event.accept then
            local text1 =
                DCEI.GetTranslationText("ui/menu/sky_race/start", { x = DCEI.GetTranslationText("ui/menu/sky_race") })
            DCEI.SetTextFrameText(layout.Description, text1 .. " " .. text2)
            DCEI.SetOnClickCallback(layout.StartButton, function()
                layout:Close()
                player.race_event.accept = true
                bank.CommitRaceEvent()
                AttemptCreateRaceEventData()
                OpenRaceEventMenu()
                local event = LogEvents.RaceEvent("Accept", 0)
                LogEvent(event)
            end)
        else
            DCEI.SetTextFrameText(layout.Description, text2)
        end

        DCEI.SetOnClickCallback(layout.CloseButton, function()
            layout:Close()
        end)
    end

    if player.race_event.accept then
        ui.race_event_menu:Show()
        UpdateRaceMenu()
    end
end

function CheckRaceDataReady()
    local data_ready = true
    local player_count = 5
    if player.race_event.players then
        for i = 1, player_count do
            if player.race_event.players[i] then
                local player_data = player.race_event.players[i]
                if player_data.name and player_data.prev_score and player_data.cur_score and player_data.progress then
                    data_ready = true
                else
                    data_ready = false
                    break
                end
            else
                data_ready = false
                break
            end
        end
    else
        data_ready = false
    end

    return data_ready
end

function UpdateRaceMenu()
    local layout = ui.race_event_menu
    if ui.race_event_banner and layout and DCEI.FrameExists(layout.Frame) then
        if ui.race_event_banner.fetching then
            DCEI.SetFrameActive(layout.LoadingFrame, true)
            return
        else
            -- check if all data is ready
            if not CheckRaceDataReady() then
                DCEI.SetFrameActive(layout.LoadingFrame, true)
                AttemptCreateRaceEventData()
                return
            end
        end
        DCEI.SetFrameActive(layout.LoadingFrame, false)
        DCEI.SetFrameActive(layout.RewardFrame, false)
        layout.display_reward = nil

        if not ui.race_event_banner.menu_updated then
            BlockInput()
            local cum_count = player.race_event.cum_count or 1
            -- populate player data
            local animation_delay = 0.5
            local animation_duration = 1
            local RACE_LENGTH = 420
            local GOAL_LENGTH = 560
            local portrait = cum_count % 2 == 0
            if portrait then
                RACE_LENGTH = 500
                GOAL_LENGTH = 500
            end
            local has_change = false
            -- list for sort and record progress
            for i = 1, 5 do
                local player_data = player.race_event.players[i]
                local name_key = "PlayerName" .. i
                local medal_key = "Medal" .. i
                local ship_key = "Ship" .. i
                local progress_key = "WaveProgress" .. i
                DCEI.SetFrameActive(layout[medal_key], false)
                DCEI.SetTextFrameText(layout[name_key], player_data.name)
                DCEI.SetTextFrameText(layout[progress_key], player_data.progress)
                local h_offset = player_data.progress / RACE_EVENT_TOTAL_PROGRESS * RACE_LENGTH
                if portrait then
                    DCEI.SetFrameVerticalOffsetInParent(layout[ship_key], h_offset)
                else
                    DCEI.SetFrameHorizontalOffsetInParent(layout[ship_key], h_offset)
                end
                local progress_increment = player_data.cur_score - player_data.prev_score
                if progress_increment > 0 then
                    has_change = true
                end
                local cur_progress = player_data.progress + progress_increment
                local target_offset = cur_progress / RACE_EVENT_TOTAL_PROGRESS * RACE_LENGTH
                if cur_progress >= RACE_EVENT_TOTAL_PROGRESS then
                    cur_progress = RACE_EVENT_TOTAL_PROGRESS
                    target_offset = GOAL_LENGTH
                end
                -- animation
                DCEI.TriggerAddTimerEventElapsed(function()
                    local ease_type = cur_progress >= RACE_EVENT_TOTAL_PROGRESS and "InQuad" or "InOutSine"
                    if portrait then
                        DCEI.AnimateFrameVerticalOffset(
                            layout[ship_key],
                            h_offset,
                            target_offset,
                            animation_duration,
                            ease_type
                        )
                    else
                        DCEI.AnimateFrameHorizontalOffset(
                            layout[ship_key],
                            h_offset,
                            target_offset,
                            animation_duration,
                            ease_type
                        )
                    end
                    local increment = cur_progress - player_data.progress
                    if increment > 0 then
                        local text_change_step = 0.9 * animation_duration / increment
                        for f = 1, increment do
                            DCEI.TriggerAddTimerEventElapsed(function()
                                local display_progress = player_data.progress + f
                                DCEI.SetTextFrameText(layout[progress_key], display_progress)
                            end, text_change_step * f, true, true)
                        end
                    end
                end, animation_delay, true, true)
            end
            if not has_change then
                animation_delay = 0
                animation_duration = 0
            end
            DCEI.TriggerAddTimerEventElapsed(function()
                UnblockInput()
                ui.race_event_banner.menu_updated = true
                -- ranking
                local rank_chests_icon = {
                    "hero_chest_t3",
                    "hero_chest_t5",
                    "hero_chest_t2",
                }
                local reward_show = {
                    {
                        "wildcastle_icon_sword_green",
                        "wildcastle_icon_robe_blue",
                        "wildcastle_icon_bow_purple",
                    },
                    {
                        "wildcastle_icon_armor_plate_green",
                        "wildcastle_icon_bow_white",
                        "wildcastle_icon_hammer_plate_blue",
                    },
                    {
                        "wildcastle_icon_staff_green",
                        "wildcastle_icon_sword_white",
                        "wildcastle_icon_armor_leather_blue",
                    },
                }
                for i = 1, 5 do
                    local rank = player.race_event.players[i].rank
                    local medal_key = "Medal" .. i
                    if rank <= 3 then
                        DCEI.SetFrameActive(layout[medal_key], true)
                        DCEI.AnimateFrameScale(layout[medal_key], { x = 0, y = 0 }, { x = 1, y = 1 }, 0.5, "OutBack")
                        DCEI.SetFrameImage(layout[medal_key], "icon_wildcastle_rank" .. rank)
                        DCEI.SetOnClickCallback(layout[medal_key], function()
                            if not player.race_event.event_set then
                                if layout.display_reward == rank then
                                    DCEI.SetFrameActive(layout.RewardFrame, false)
                                    layout.display_reward = nil
                                else
                                    DCEI.SetFrameActive(layout.RewardFrame, true)
                                    DCEI.AnimateFrameScale(
                                        layout.RewardFrame,
                                        { x = 0.3, y = 0.6 },
                                        { x = 1, y = 1 },
                                        0.5,
                                        "OutBack"
                                    )
                                    local medal_position = DCEI.GetFramePosition3D(layout[medal_key])
                                    if portrait then
                                        local h_offset = medal_position.x - 370
                                        DCEI.SetFrameHorizontalOffsetInParent(layout.RewardFrame, h_offset)
                                    else
                                        local v_offset = medal_position.y - 443
                                        DCEI.SetFrameVerticalOffsetInParent(layout.RewardFrame, v_offset)
                                    end
                                    layout.display_reward = rank
                                end
                                DCEI.SetFrameImage(layout.RewardIcon, rank_chests_icon[rank])
                                for r = 1, 3 do
                                    local reward_show_key = "RewardShow" .. r
                                    DCEI.SetFrameImage(layout[reward_show_key], reward_show[rank][r])
                                end
                            end
                        end)
                    end
                end
                -- settle player records
                for i = 1, 5 do
                    local player_data = player.race_event.players[i]
                    local progress_increment = player_data.cur_score - player_data.prev_score
                    local cur_progress = player_data.progress + progress_increment
                    if cur_progress >= RACE_EVENT_TOTAL_PROGRESS then
                        cur_progress = RACE_EVENT_TOTAL_PROGRESS
                    end
                    player_data.prev_score = player_data.cur_score
                    player_data.progress = cur_progress
                end
                -- if player finish 50 waves and has reward
                if player.race_event.rank_reward and not player.race_event.reward_claimed then
                    DCEI.SetFrameActive(layout.ClaimButton, true)
                    DCEI.SetOnClickCallback(layout.ClaimButton, ClaimRaceEventReward)
                end
                if player.race_event.event_set then
                    DCEI.SetFrameActive(layout.RaceOverFrame, true)
                    DCEI.AnimateFrameScale(layout.RaceOverText, { x = 0, y = 0 }, { x = 1, y = 1 }, 0.5, "OutBack")
                    -- if event is set but player doesn't get rank 1~3 then event ends
                    if not player.race_event.rank_reward then
                        player.race_event.reward_claimed = true
                    end
                end

                bank.CommitRaceEvent()
            end, animation_delay + animation_duration)
        end
    end
end

function CreateRaceEventBanner()
    local layout_name = "RaceEvent/Banner"
    local cum_count = player.race_event.cum_count or 1
    if cum_count % 2 == 0 then
        layout_name = "RacePortraitEvent/Banner"
    end
    local banner = gmui.Layout:New({
        parent = ui.glue.frame,
        name = layout_name,
        on_show_animation = gmui.Animations.BackIn,
        on_hide_animation = gmui.Animations.BackOut,
    })
    DCEI.SetFrameRightAlignmentInParent(banner.Frame)
    DCEI.SetFrameVerticalOffsetInParent(banner.Frame, 0)
    if ui.PlayingiOSVersion() then
        DCEI.SetFrameVerticalOffsetInParent(banner.Frame, -60)
    end
    DCEI.SetFrameHorizontalOffsetInParent(banner.Frame, -100)

    function banner:TimerUpdate(time)
        local text = ui.FormatTimeShort(time)
        DCEI.SetTextFrameText(self.TimeLabel, text)
    end

    -- directly accept during this test
    DCEI.SetOnClickCallback(banner.Button, function()
        -- first time directly accept
        local cum_count = player.race_event.cum_count or 1
        if cum_count <= 1 and not player.race_event.accept then
            player.race_event.accept = true
            bank.CommitRaceEvent()
            AttemptCreateRaceEventData()
            local event = LogEvents.RaceEvent("Accept", 0)
            LogEvent(event)
        end
        OpenRaceEventMenu()
    end)
    banner:Show()

    return banner
end

function AttemptCreateRaceEventData()
    if ui.race_event_banner and not player.race_event.event_end then
        ui.race_event_banner.fetching = true
        local type = "default"
        local limit = 25

        local function ResultCallback(result)
            if result then
                ui.race_event_banner.fetching = false
                local player_list = {}
                for n, user in ipairs(result.users) do
                    if user.uuid ~= DCEI.GetUserId() then
                        local payload = user.payload

                        if StringIsEncodedTable(payload) then
                            -- new format is table (as of 0.0.66)
                            local decoded = json.decode(payload)

                            payload = decoded
                        elseif payload and payload ~= "" then
                            -- old format is string

                            payload = { username = payload }
                        end
                        local player_data = {
                            name = payload.username or user.uuid,
                            uuid = user.uuid,
                            prev_score = math.floor(user.score),
                            cur_score = math.floor(user.score),
                            time = user.time,
                        }
                        table.insert(player_list, player_data)
                    end
                end
                local function sort_by_time(a, b)
                    return a.time > b.time
                end

                table.sort(player_list, sort_by_time)
                player.race_event.players = {}
                -- player 1 is player self
                local self_data = {
                    name = (player.username and player.username.name) or ui.PlayerHandle(),
                    prev_score = wave.data.GetMax(),
                    cur_score = wave.data.GetMax(),
                    progress = 0,
                    rank = 1,
                }
                table.insert(player.race_event.players, self_data)
                local player_count = 5
                -- select top 4 active player
                for i = 1, (player_count - 1) do
                    local list_player_data = player_list[i]
                    local other_player_data = {
                        name = list_player_data.name,
                        uuid = list_player_data.uuid,
                        prev_score = list_player_data.prev_score,
                        cur_score = list_player_data.cur_score,
                        progress = 0,
                        rank = i + 1,
                    }
                    table.insert(player.race_event.players, other_player_data)
                end
                bank.CommitRaceEvent()
                UpdateRaceMenu()
            end
        end

        DCEI.GetLeaderBoardSurroundingUsers(type, limit, ResultCallback)
    end
end

function AttemptFetchRaceEventData()
    if not CheckRaceDataReady() then
        return
    end
    if
        ui.race_event_banner
        and player.race_event.accept
        and not player.race_event.event_set
        and not player.race_event.event_end
    then
        -- if player accept race or after each wave win, fetch data from server
        ui.race_event_banner.fetching = true
        ui.race_event_banner.menu_updated = false
        -- store data in race event table as actual data
        -- when open race event menu, check display data and actual data, show aniamtion
        local uuid_list = {}
        local player_count = 5
        for i = 2, player_count do
            table.insert(uuid_list, player.race_event.players[i].uuid)
        end
        local function ResultCallback(result)
            if result then
                -- update player's data, only when player has network
                local self_data = player.race_event.players[1]
                self_data.cur_score = wave.data.GetMax()
                -- fake some data
                -- self_data.cur_score = self_data.cur_score + math.random(30, 40)
                for n, user in ipairs(result.users) do
                    for i = 2, player_count do
                        local other_player = player.race_event.players[i]
                        if other_player.uuid == user.uuid then
                            other_player.cur_score = math.floor(user.score)
                            -- fake some data
                            -- other_player.cur_score = other_player.cur_score + math.random(20, 30)
                            break
                        end
                    end
                end

                local temp_list = {}
                local start_rank = 0
                local finish_num = 0
                local total_progress = 0
                for i = 1, player_count do
                    local player_data = player.race_event.players[i]
                    if player_data.finish then
                        start_rank = start_rank + 1
                        finish_num = finish_num + 1
                    else
                        local progress_increment = player_data.cur_score - player_data.prev_score
                        local cur_progress = player_data.progress + progress_increment

                        if cur_progress >= RACE_EVENT_TOTAL_PROGRESS then
                            cur_progress = RACE_EVENT_TOTAL_PROGRESS
                            player_data.finish = true
                            finish_num = finish_num + 1
                        end

                        local temp_data = {
                            id = i,
                            progress = cur_progress,
                        }
                        table.insert(temp_list, temp_data)
                    end
                end
                if #temp_list > 0 then
                    local function sort_by_progress(a, b)
                        if a.progress == b.progress then
                            return a.id < b.id
                        else
                            return a.progress > b.progress
                        end
                    end

                    table.sort(temp_list, sort_by_progress)
                    for i = 1, #temp_list do
                        local id = temp_list[i].id
                        player.race_event.players[id].rank = start_rank + i
                    end
                end

                if self_data.finish and self_data.rank <= 3 then
                    player.race_event.rank_reward = self_data.rank
                    player.race_event.event_set = true
                end
                if finish_num >= 3 then
                    player.race_event.event_set = true
                end

                bank.CommitRaceEvent()
            end
            ui.race_event_banner.fetching = false
        end

        DCEI.LogMessage("Fetching Race event data")
        DCEI.GetLeaderBoardUserScores("default", uuid_list, ResultCallback)
    end
end

-- !SECTION

-- NOTE: this is a hell mode function
function OpenHellModeEvent()
    ui.leaderboard.hell_mode = true
    ShowLeaderBoard()
    ui.HellModeEventRewardFrame()
end

-- SECTION Auto battle functions
function CreateAutoBattleMenu()
    local level = wave.data.GetCurrent()
    local gold_cost = data.autobattle:GetCost(level)
    local auto_time = player.gem_auto_time
    local auto_data = player.autobattle
    auto_data.mode = nil

    -- advance instance to prevent hiding button while menu is open
    ui.battle.floating_autobattle:AdvanceInstance()

    local function OnCancel()
        SetGameSpeed(player.speed)
        ShowAutoBattleButton()

        -- resume boost timer when battle resumes
        ResumeBoostTimer()
    end

    local function OnContinueGold()
        AttemptPurchaseGoldAutoBattle(gold_cost)
    end

    local function OnContinueGem()
        AttemptStartGemAutoBattle(auto_time)
    end

    local function OnStore()
        local on_close = function()
            DCEI.SetFrameActive(ui.autobattle.menu.frame, true)
            ui.screen_mask.upper:Show()

            ShowHeroSkillBars()
            ui.battle:Show()

            -- show boost display
            ui.boost_display:Show()
        end

        -- log store opening from ab prompt
        local from = "autobattle prompt (start)"
        SetStoreEntrySource(from)

        ui.screen_mask.Hide()
        DCEI.SetFrameActive(ui.autobattle.menu.frame, false)
        HideHeroSkillBars()
        ui.battle:Hide()

        OpenAutoBattleStore(on_close)
        SetGameSpeed(player.speed)
        local load_time = shop.GetItemLazyLoadTimeRemaining() or 0
        DCEI.TriggerAddTimerEventElapsed(
            function()
                SetGameSpeed(0)
            end,
            load_time,
            true,
            true
        )

        -- hide boost display
        ui.boost_display:Hide()
    end

    local function OnStoreGold()
        local on_close = function()
            DCEI.SetFrameActive(ui.autobattle.menu.frame, true)
            ui.screen_mask.upper:Show()

            ShowHeroSkillBars()
            ui.battle:Show()

            -- show boost display
            ui.boost_display:Show()
        end

        -- log store opening from ab prompt
        local from = "autobattle prompt (start)"
        SetStoreEntrySource(from)

        ui.screen_mask.Hide()
        DCEI.SetFrameActive(ui.autobattle.menu.frame, false)
        HideHeroSkillBars()
        ui.battle:Hide()

        OpenGoldStore(on_close)

        -- hide boost display
        ui.boost_display:Hide()
    end

    local callbacks = {
        OnCancel = OnCancel,
        OnContinueGold = OnContinueGold,
        OnContinueGem = OnContinueGem,
        OnStore = OnStore,
        OnStoreGold = OnStoreGold,
        autobattle_setting = player.autobattle,
    }

    ui.autobattle.menu = ui.NewAutoBattleMenu(callbacks, gold_cost, auto_time, auto_data)
    SetGameSpeed(0)

    -- pause boost timer while in menus
    StopBoostTimer()

    player.autobattle_open = player.autobattle_open or 0
    player.autobattle_open = player.autobattle_open + 1
    bank.CommitAutoBattleOpenCount()
end

function AttemptStartGemAutoBattle(auto_time)
    if auto_time > 0 then
        EnableAutoBattle("gem")
        StartGemAutoBattleTimer()
        -- LogEvent(LogEvents.StartGemAutoBattle())
        player.autobattle_gem = player.autobattle_gem or 0
        player.autobattle_gem = player.autobattle_gem + 1
        bank.CommitAutoBattleGemCount()
    else
        ui.ShowFeedbackNotEnoughAutoBattle()
    end

    ui.autobattle.menu.Close()
end

function AttemptPurchaseGoldAutoBattle(cost)
    local gold = GetPlayerGold()

    if cost <= gold then
        SpendPlayerGold(cost)

        EnableAutoBattle("gold")
        -- LogEvent(LogEvents.StartGoldAutoBattle(cost))
        player.autobattle_gold = player.autobattle_gold or 0
        player.autobattle_gold = player.autobattle_gold + 1
        bank.CommitAutoBattleGoldCount()
    else
        ui.ShowFeedbackNotEnoughGold()
    end

    ui.autobattle.menu.Close()
end

-- !SECTION

-- NOTE Create Effect Help Function
function CreateEffect(effect, x, y)
    DCEI.CreateEffect(effect, CASTLE_UNIT, x, y)
end

-- SECTION DAILY REWARDS
function AdvanceCalendarRewardIndex()
    local cycle_length = #data.calendar
    local current_index = player.calendar.reward_index
    local next_index = (current_index % cycle_length) + 1

    player.calendar.reward_index = next_index
end

function AdvanceBoostAvailableIndex()
    local cycle_length = #data.boosts.array
    local current_index = player.calendar.boost_index
    local next_index = (current_index % cycle_length) + 1

    player.calendar.boost_index = next_index
end

function InitializeBoostAvailableIndex()
    local day_one_boost = data.boosts["boost_gold"]
    player.calendar.boost_index = day_one_boost.index
end

-- SERVER TIME
function UpdateBoostsOnNewDay()
    if player.calendar.boost_index and player.calendar.boost_index ~= 0 then
        AdvanceBoostAvailableIndex()

        local index = player.calendar.boost_index
        local key = data.boosts.array[index].key
        ClearPurchaseDataForCollection(key)

        ShowShopPip()
    end
end

function UpdateRaidTurboBundleOnNewWeek()
    local key = "raid_turbo_pack"
    ClearPurchaseDataForCollection(key)

    local items_list = GetStoreCollectionData(key)
    ui.shop:RemoveItems(items_list)

    if wave.data.GetMax() >= 100 then
        local options = { init = true }
        AddItemsToShopDealsFromCollectionKey(key, options)
    end
end

function AddBoostItemsToStore(show_pip)
    -- initialize boost calendar data if non exists
    if not player.calendar.boost_index or player.calendar.boost_index and player.calendar.boost_index == 0 then
        InitializeBoostAvailableIndex()
    end

    -- clear existing boosts from store (if any)
    RemoveCurrentBoostsFromDealsPage()

    -- add items to store from current key
    local index = player.calendar.boost_index

    -- clamp index within array bounds
    index = Clamp(index, 1, #data.boosts.array)
    local key = data.boosts.array[index].key
    local options = { init = true }
    AddItemsToShopDealsFromCollectionKey(key, options)

    -- show shop pip when boost added to store (if above wave 10)
    local unlock_wave = 10
    if show_pip and (wave.data.GetMax() >= unlock_wave) then
        ShowPipsForCollectionKey(key)
    end
end

function ShowPipsForCollectionKey(key)
    for _, product_id in pairs(store_propers.collections[key]) do
        if type(product_id) == "string" then
            local item = shop.GetItemLayout(product_id)
            if item and shop.GetItemStockAvailable(product_id) ~= 0 then
                item:ShowPip()
                ShowShopPip()
            end
        end
    end
end

function ClearPurchaseDataForCollection(key)
    if ValidStoreCollectionKey(key) then
        for _, product_id in ipairs(store_propers.collections[key]) do
            shop.ClearItemPurchaseData(product_id)
        end
    end
end

function UpdateCalendarOnNewDay(game_age, hide_calendar_pip)
    -- advance calendar reward index when new reward becomes available (unclaimed rewards aren't skipped)
    if not player.calendar.reward_available then
        AdvanceCalendarRewardIndex()
        player.calendar.reward_available = true
    end

    -- show calendar pip (skip calendar pip on game start in case of auto-claim rewards calendar)
    if not hide_calendar_pip then
        ui.glue.calendar_menu:ShowPip()
    end
end

function OnGetServerTimeUpdate(time_stamp, hide_calendar_pip)
    if event_mode.IsActive() then
        return
    end

    local day_length = 86400
    local week_length = day_length * 7
    local game_birth = 1580389200
    local game_age_in_seconds = time_stamp - game_birth
    local game_age = math.floor(game_age_in_seconds / day_length)
    local game_week = math.floor(game_age_in_seconds / week_length)
    next_daily_reward_time = (game_age + 1) * day_length + game_birth

    player.calendar.week = player.calendar.week or 0
    -- DEBUG
    -- player.calendar.day = 200

    local show_boost_pip

    -- check if new day
    if game_age > player.calendar.day then
        -- store new calendar day
        player.calendar.day = game_age

        -- update daily rewards calendar
        UpdateCalendarOnNewDay(game_age, hide_calendar_pip)

        -- update daily boosts
        UpdateBoostsOnNewDay()
        show_boost_pip = true
        if game_week > player.calendar.week then
            player.calendar.week = game_week
            UpdateRaidTurboBundleOnNewWeek()
        end

        -- refill raid key
        AddPlayerRaidKeyDaily(RAID_KEY_DAILY_COUNT)
        player.raid_key_extra_purchase_count = 0
        bank.CommitRaidKeyExtraPurchaseCount()
        if ui.raid_boss then
            ui.raid_boss.UpdateRaidKeys()
            ui.raid_boss.UpdateRaidKeyExtraCost()
        end
        -- can accumulate at most 6 times
        local free_count = GetHellModeFreeCount()
        local total_count = math.min(free_count + 3, 6)
        SetHellModeFreeCount(total_count)
        bank.CommitHellModeFreeCount()
        if ui.hell_mode then
            ui.hell_mode:UpdateBattleButton(GetHellModeFreeCount())
        end

        player.daily_ads_limit = DAILY_ADS_LIMIT
        bank.CommitDailyAdsLimit()
    end

    -- add current boost items to store
    AddBoostItemsToStore(show_boost_pip)

    -- hookup timer displays
    HookupCalendarTimerDisplay(time_stamp, next_daily_reward_time)

    -- commit calendar data
    bank.CommitCalendarData()
end

function AttemptGetServerTime(callback, fail_on_purpose)
    local function ResultCallback(time_stamp)
        server.is_connecting = false

        -- in case we want to debug GetServerTime failing
        if fail_on_purpose then
            DCEI.LogWarning("> GetServerTime failed on purpose")
            time_stamp = nil
        end

        if time_stamp then
            server.day_fetched = true
            DCEI.LogMessage("> GetServerTime Succeeded")
        end

        if callback then
            callback(time_stamp)
        end
    end

    -- prevent GetServerTime result from being spammed
    if not server.is_connecting then
        server.is_connecting = true
        DCEI.GetServerTime(ResultCallback)
    end
end

-- SERVER TIME CHECKS
function CheckServerTimeOnTimerExpire()
    local function ResultCallback(time_stamp)
        if time_stamp then
            -- updates daily reward calendar and boosts
            OnGetServerTimeUpdate(time_stamp)

            -- claim daily reward if player connects to server while calendar is open
            if player.calendar.reward_available and ui.calendar.is_visible then
                ClaimDailyReward()
            end
        end
    end

    AttemptGetServerTime(ResultCallback)
end

function CheckServerTimeOnOpenShop()
    local function ResultCallback(time_stamp)
        if time_stamp then
            -- updates daily reward calendar and boosts
            OnGetServerTimeUpdate(time_stamp)
        end
    end

    AttemptGetServerTime(ResultCallback)
end

function CheckServerTimeOnOpenDailyRewardCalendar()
    local function ResultCallback(time_stamp)
        if time_stamp then
            -- updates daily reward calendar and boosts
            OnGetServerTimeUpdate(time_stamp)

            -- claim daily reward if player connects to server while calendar is open
            if player.calendar.reward_available and ui.calendar.is_visible then
                ClaimDailyReward()
            end
        end
    end

    AttemptGetServerTime(ResultCallback)
end

function RotateHellCoinShop(time_stamp)
    player.hell_coin_shop.badge_index = player.hell_coin_shop.badge_index + 1
    if player.hell_coin_shop.badge_index > #data.badges then
        player.hell_coin_shop.badge_index = 1
    end
    player.hell_coin_shop.rotate_time = time_stamp + HELL_COIN_SHOP_ROTATE_DURATION
    bank.CommitHellCoinShopData()
    local product_id = "com.wildsky.castle.hell_coin.philosopher_stone.01"
    shop.ClearItemPurchaseData(product_id)

    InitializeHellCoinPage()
end

function CheckServerTimeOnHellCoinShopExpire()
    local function ResultCallback(time_stamp)
        if time_stamp then
            RotateHellCoinShop(time_stamp)
            AttemptStartHellCoinShopTimer(time_stamp)
        end
    end

    AttemptGetServerTime(ResultCallback)
end

-- !SECTION

-- SECTION Mail, send player missing item after IAP purchase
local ProcessClaimEmailItem = {
    ["remove_ad"] = function()
        local product_id = "com.wildsky.castle.iap.remove_ad.01"
        OnAnyItemPurchase(product_id)
    end,
    ["autumn_pass"] = function()
        local product_id = "com.wildsky.castle.iap.autumnpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["summer_pass"] = function()
        local product_id = "com.wildsky.castle.iap.summerpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["royal_pass"] = function()
        local product_id = "com.wildsky.castle.iap.royalpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["spring_pass"] = function()
        local product_id = "com.wildsky.castle.iap.springpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["winter_pass1"] = function()
        local product_id = "com.wildsky.castle.iap.winterpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["winter_pass2"] = function()
        local product_id = "com.wildsky.castle.iap.winterpass.02"
        OnAnyItemPurchase(product_id)
    end,
    ["winter_pass3"] = function()
        local product_id = "com.wildsky.castle.iap.winterpass.03"
        OnAnyItemPurchase(product_id)
    end,
    ["seasonpass"] = function()
        local product_id = "com.wildsky.castle.iap.seasonpass.01"
        OnAnyItemPurchase(product_id)
    end,
    ["seasonpass14days"] = function()
        local product_id = "com.wildsky.castle.iap.seasonpass.02"
        OnAnyItemPurchase(product_id)
    end,
    ["bundle_huge"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.huge.01"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["raid_turbo"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.raid_turbo.01"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_mine01"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.mine.01"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_mine02"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.mine.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_starter02"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.starter.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_starter03"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.starter.03"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle02"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle03"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.03"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle_new02"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.new.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle_new03"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.new.03"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle_new04"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.new.04"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_autobattle_new05"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.autobattle.new.05"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_token01"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.token.01"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["bundle_token02"] = function()
        local product_id = "com.wildsky.castle.iap.bundle.token.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["boost_gold30"] = function()
        local product_id = "com.wildsky.castle.iap.boost.gold.30"
        OnAnyItemPurchase(product_id)
    end,
    ["boost_gold60"] = function()
        local product_id = "com.wildsky.castle.iap.boost.gold.60"
        OnAnyItemPurchase(product_id)
    end,
    ["boost_gold90x2"] = function()
        local product_id = "com.wildsky.castle.iap.boost.gold.90x2"
        OnAnyItemPurchase(product_id)
    end,
    ["boost_token30"] = function()
        local product_id = "com.wildsky.castle.iap.boost.token.30"
        OnAnyItemPurchase(product_id)
    end,
    ["boost_gold60x2"] = function()
        local product_id = "com.wildsky.castle.iap.boost.gold.60x2"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_gold01"] = function()
        local product_id = "com.wildsky.castle.iap.gold.01"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_gold02"] = function()
        local product_id = "com.wildsky.castle.iap.gold.02"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_gold03"] = function()
        local product_id = "com.wildsky.castle.iap.gold.03"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_gold04"] = function()
        local product_id = "com.wildsky.castle.iap.gold.04"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_gold05"] = function()
        local product_id = "com.wildsky.castle.iap.gold.05"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_gold06"] = function()
        local product_id = "com.wildsky.castle.iap.gold.06"
        local item_data = { item_data = shop.GetProductData(product_id) }
        if item_data.item_data then
            GenerateGoldValue(item_data)
            OnAnyItemPurchase(product_id)
        end
    end,
    ["iap_token_new01"] = function()
        local product_id = "com.wildsky.castle.iap.token.new.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_token01"] = function()
        local product_id = "com.wildsky.castle.iap.token.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_token02"] = function()
        local product_id = "com.wildsky.castle.iap.token.02"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_token03"] = function()
        local product_id = "com.wildsky.castle.iap.token.03"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_token_new04"] = function()
        local product_id = "com.wildsky.castle.iap.token.new.04"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_token04"] = function()
        local product_id = "com.wildsky.castle.iap.token.04"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_chest01"] = function()
        local product_id = "com.wildsky.castle.iap.chests.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_chest02"] = function()
        local product_id = "com.wildsky.castle.iap.chests.02"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_chest03"] = function()
        local product_id = "com.wildsky.castle.iap.chests.03"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_chest04"] = function()
        local product_id = "com.wildsky.castle.iap.chests.04"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle01"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle02"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.02"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle03"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.03"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle04"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.04"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle_new01"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.new.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle_new02"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.new.02"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle_new03"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.new.03"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_autobattle_new04"] = function()
        local product_id = "com.wildsky.castle.iap.autobattle.new.04"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_speed01"] = function()
        local product_id = "com.wildsky.castle.iap.speed.01"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_speed02"] = function()
        local product_id = "com.wildsky.castle.iap.speed.02"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_speed03"] = function()
        local product_id = "com.wildsky.castle.iap.speed.03"
        OnAnyItemPurchase(product_id)
    end,
    ["iap_speed04"] = function()
        local product_id = "com.wildsky.castle.iap.speed.04"
        OnAnyItemPurchase(product_id)
    end,
    ["gold"] = function(value)
        AddPlayerGold(value)
        bank.CommitGold()
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.GoldRewardFeedback(value)
            PlayGoldSpendSound()
        end, 0.125, true, true)
    end,
    ["token"] = function(value)
        AddPlayerOverflowGems(value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.TokenRewardFeedback(value)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["potion"] = function(value)
        event_mode.AddPotions(value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.PotionRewardFeedback(value)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["chest_l"] = function(value)
        local chest_type = "rare_chest_l"
        AddPlayerChest(chest_type, value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.ChestsRewardFeedback(value, chest_type)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["chest"] = function(value)
        local chest_type = "rare_chest"
        AddPlayerChest(chest_type, value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.ChestsRewardFeedback(value, chest_type)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["leg_chest_l"] = function(value)
        local chest_type = "legendary_chest_l"
        AddPlayerChest(chest_type, value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.ChestsRewardFeedback(value, chest_type)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["leg_chest"] = function(value)
        local chest_type = "legendary_chest"
        AddPlayerChest(chest_type, value)
        -- display feedback
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.ChestsRewardFeedback(value, chest_type)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["auto_battle"] = function(value)
        AddGemAutoBattleTime(value * 3600)

        local icon = "shop_autobattle_01"
        local icon_size = 110
        local text = "+" .. value .. " " .. DCEI.GetTranslationText("general/hours")
        text = "<color=green>" .. text .. "</color>"
        local offset = {
            spacing = -20,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["speed"] = function(value)
        AddSpeedx3Time(value * 24 * 60 * 60)

        local icon = "shop_speedup_01"
        local icon_size = 110
        local day_text = value == 1 and DCEI.GetTranslationText("general/day")
            or DCEI.GetTranslationText("general/days")
        local text = "+" .. value .. " " .. day_text
        text = "<color=green>" .. text .. "</color>"
        local offset = {
            spacing = -20,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["mine_contract"] = function(value)
        AddPlayerContract_Mine(value)

        local icon = "icon_catalyst_main_item2_orange_03"
        local icon_size = 120
        local text = "x " .. value
        text = "<color=green>" .. text .. "</color>"
        local offset = {
            spacing = -20,
            x = 0,
            y1 = -20,
            y2 = 60,
        }
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            PlayGoldSpendSound()
        end, 0.625, true, true)
    end,
    ["gold_boost_200"] = function(value)
        local type = "gold_boost_200"
        player.timers[type] = player.timers[type] or 0
        player.timers[type] = player.timers[type] + value
        PlayGoldSpendSound()
        bank.CommitTimers()
    end,
    ["gold_boost_150"] = function(value)
        local type = "gold_boost_150"
        player.timers[type] = player.timers[type] or 0
        player.timers[type] = player.timers[type] + value
        PlayGoldSpendSound()
        bank.CommitTimers()
    end,
    ["token_boost_500"] = function(value)
        local type = "token_boost_500"
        player.timers[type] = player.timers[type] or 0
        player.timers[type] = player.timers[type] + value
        PlayGoldSpendSound()
        bank.CommitTimers()
    end,
}

function EmailFetchCallback(fetch)
    -- create claimed mail's key dictionary
    local claimed_mail_dictionary = {}
    for _, id in ipairs(player.claimed_mails) do
        claimed_mail_dictionary[id] = true
    end

    local fetched_mail_ids = {}
    local mails_table = fetch.mails
    for _, mail in ipairs(mails_table) do
        local instance_id = mail.instance_id
        fetched_mail_ids[instance_id] = true

        -- only claim mails that haven't claimed yet
        if not claimed_mail_dictionary[instance_id] then
            for _, item in ipairs(mail.items) do
                ProcessClaimEmailItem[item.key](item.value)
            end

            table.insert(player.claimed_mails, instance_id)
            bank.CommitClaimedMails()
        end
    end

    for _, id in ipairs(player.claimed_mails) do
        -- only delete mails that still exist on server
        if fetched_mail_ids[id] then
            DCEI.DeleteMail(id)
        end
    end
end

-- !SECTION

-- NOTE CheckServerTimeOnGameStart
function CheckServerTimeOnGameStart()
    local function ResultCallback(time_stamp)
        if time_stamp then
            -- updates daily reward calendar and boosts
            local hide_calendar_pip = true
            OnGetServerTimeUpdate(time_stamp, hide_calendar_pip)

            -- attempt on game start popups
            local game_start_popup_delay = 0.625
            if player.calendar.reward_available then
                -- show/claim rewards calendar if new reward is available
                DCEI.TriggerAddTimerEventElapsed(AttemptAutoOpenRewardCalendar, game_start_popup_delay, false, true)
            else
                -- else attempt to show deals popup
                DCEI.TriggerAddTimerEventElapsed(AttemptOnStartDealsPopup, game_start_popup_delay, false, true)
            end

            ScheduleNotification(time_stamp)

            if not GetPlayerTest("pause_offline_gold") and player.system.vault == system.active then
                if time_stamp > data.vault.time_record then
                    if GetVaultGold() < data.vault.capacity then
                        local time_diff = time_stamp - data.vault.time_record
                        local offline_gold = math.floor(time_diff / data.vault.interval) * data.vault:GetVaultIncome()

                        GenerateVaultGold(offline_gold)
                    else
                        ui.bottom_panel:ShowTownPip()
                    end
                end

                data.vault.time_record = time_stamp
                bank.CommitVaultRecord()

                ScheduleVaultNotification()
            end

            -- try to claim items on game start
            DCEI.TriggerAddTimerEventElapsed(function()
                local forceFetch = true
                DCEI.GetMail(EmailFetchCallback, forceFetch)
            end, 3, true, true)

            -- try to claim last season's ranking reward on game start
            DCEI.TriggerAddTimerEventElapsed(function()
                HookupLeaderboardTimerDisplay(time_stamp)
                if not event_mode.IsActive() then
                    FetchHellModeSeasonRewardOnGameStart(time_stamp)
                    AttemptStartAutumnEventTimer(time_stamp)
                    AttemptStartWinterEventTimer(time_stamp)
                end
            end, 1, true, true)
        end
    end

    -- DEBUG
    -- local fail_on_purpose = true

    AttemptGetServerTime(ResultCallback, fail_on_purpose)
end

-- SECTION Daily rewards again
function AttemptAutoOpenRewardCalendar()
    local unlock_wave = 10

    -- avoid opening popup if player is in combat
    if wave.data.GetMax() >= unlock_wave then
        ShowDailyRewardCalendar()
    end
end

function ShowDailyRewardCalendar()
    CloseActivePrompts()

    -- prevent calendar from showing if wave started or already open
    if wave.data.is_active or ui.shop.is_active or ui.calendar.is_visible then
        return
    end
    if ui.glue.other_button_panel.opening or ui.glue.other_button_panel.closing then
        return
    end
    ui.glue.other_button_panel:Close()

    -- attempt to connect to server if haven't already
    if not server.day_fetched then
        CheckServerTimeOnOpenDailyRewardCalendar()
    end

    -- update and show calendar
    ui.calendar.max_wave = wave.data.GetMax()
    ui.calendar.seasonpass_timer = player.seasonpass_timer
    ui.calendar:Show()

    -- claim daily reward if available
    if player.calendar.reward_available then
        ClaimDailyReward()
    end

    local has_seasonpass = player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() > 0

    if not has_seasonpass then
        DCEI.SetFrameActive(ui.calendar.seasonpass_container, false)
    end
    -- if not ui.calendar.displayed_pip and not has_seasonpass then
    --     ui.calendar.displayed_pip = true
    --     ui.calendar:ShowPipAnimated()
    -- end
end

function ClaimDailyReward()
    local has_seasonpass = player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() > 0
    -- under some unknown cases {player.calendar.reward_index} returns an invalid calendar item as of 0.0.114
    player.calendar.reward_index = Clamp(player.calendar.reward_index, 1, #data.calendar)

    local index = player.calendar.reward_index
    local item = data.calendar[index]
    local value = has_seasonpass and item.value and item.value * 3 or item.value

    ui.calendar:Claim(index)
    -- make more obvious when claiming daily rewards
    DCEI.TriggerAddTimerEventElapsed(function()
        if item.type == "medal" then
            -- for medals/tokens/"gems"
            AddPlayerOverflowGems(value)
            ui.TokenRewardFeedback(value)
        elseif item.type == "gold" then
            -- for gold
            if not GetPlayerTest("pause_offline_gold") then
                local type = "gold_boost_150"
                local duration = has_seasonpass and 1200 * 3 or 1200
                player.timers[type] = player.timers[type] or 0
                player.timers[type] = player.timers[type] + duration
                bank.CommitTimers()
                local icon = "icon_passability_bg_ice"
                local icon_size = 90
                local minutes = math.floor(duration / 60)
                local text = "<size=120%>+" .. minutes .. "</size> " .. DCEI.GetTranslationText("general/minutes")
                text = "<color=yellow>" .. text .. "</color>"
                local offset = {
                    spacing = 20,
                    x = 0,
                    y1 = -20,
                    y2 = 60,
                }
                local layout = ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
                local hstack = DCEI.CreateHStackFrame(layout.image)
                local label = DCEI.CreateTextFrame(hstack)
                DCEI.SetTextFrameFontSizeMax(label, 30)
                DCEI.SetTextFrameFontSizeMin(label, 20)
                DCEI.SetTextFrameColor(label, { r = 1, g = 0.92, b = 0.016 })
                DCEI.SetFrameMaxSize(label, 40, 70)
                DCEI.SetTextFrameText(label, "1.5" .. "<size=75%>x")
                local icon = DCEI.CreateFrame(hstack)
                DCEI.SetFrameImage(icon, "icon_coin00")
                DCEI.SetFrameSize(icon, 28, 28)
            else
                value = item.Value(wave.data.GetMax())
                value = has_seasonpass and value * 3 or value
                AddPlayerGold(value)
                bank.CommitGold()
                ui.GoldRewardFeedback(value)
            end
        elseif item.type == "autobattle" then
            -- for auto battle time
            AddGemAutoBattleTime(value)
            local icon = "shop_autobattle_01"
            local icon_size = 90
            local minutes = math.floor(value / 60)
            local text = "<size=120%>" .. minutes .. "</size>" .. " " .. DCEI.GetTranslationText("general/min")
            text = "<color=green>" .. text .. "</color>"
            local offset = {
                spacing = -20,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        elseif item.type == "rare_chest" then
            AddPlayerChest(item.type, value)
            local icon = item.icon
            local icon_size = 100
            local text = item.color .. "\n<size=120%>+" .. value .. "</size></color>"
            local offset = {
                spacing = 6,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
        end

        -- update calendar data
        player.calendar.reward_available = false
        player.calendar.total_rewards_claimed = player.calendar.total_rewards_claimed or 0
        player.calendar.total_rewards_claimed = player.calendar.total_rewards_claimed + 1

        -- update UI
        ui.glue.calendar_menu:HidePip()
        ui.CheckOtherButtonPip()

        CheckNotificationsWhenClaimDailyReward()
        bank.CommitCalendarData()
    end, 0.5625, false, true)
end

-- NOTE Only affect abtest players have calendar season pass, can remove purchase button first, and remove related functions
function AttemptBuySeasonPass()
    local product_id = "com.wildsky.castle.iap.seasonpass.02"
    ui.calendar:HidePip()
    if not ui.calendar.preview_seasonpass_reward then
        ui.calendar.click_instance = ui.calendar.click_instance or 0
        ui.calendar.click_instance = ui.calendar.click_instance + 1
        local instance = ui.calendar.click_instance
        -- preview season pass reward and price
        -- LogEvent(LogEvents.seasonpass_clickthrough())
        ui.calendar.preview_seasonpass_reward = true
        local item_data = shop.GetProductData(product_id)
        local localized_price = shop.GetProductPriceLocalized(product_id)
        local default_price = "¥" .. shop.GetProductPriceUSD(product_id)
        local price_text = localized_price or default_price
        if player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() > 0 then
            price_text = price_text .. "\n<color=green>+14</color> " .. DCEI.GetTranslationText("general/days")
        end
        DCEI.SetTextFrameText(ui.calendar.seasonpass_label, price_text)

        DCEI.AnimateFrameScale(ui.calendar.seasonpass_button, { x = 1, y = 0.5 }, { x = 1, y = 1 }, 0.5, "OutBack")
        local seasonpass = true
        ui.calendar:Update(seasonpass)
        DCEI.TriggerAddTimerEventElapsed(function()
            if ui.calendar.preview_seasonpass_reward and instance == ui.calendar.click_instance then
                DCEI.AnimateFrameScale(
                    ui.calendar.seasonpass_button,
                    { x = 1, y = 0.5 },
                    { x = 1, y = 1 },
                    0.5,
                    "OutBack"
                )
                ui.calendar.preview_seasonpass_reward = false
                ui.calendar:UpdateSeasonPassTime(player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining())
                ui.calendar:Update(player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() > 0)
            end
        end, 3, true, true)
    else
        -- buy season pass
        AttemptAnyItemPurchase(product_id)
    end
end

function HookupCalendarTimerDisplay(time_stamp, reset_time)
    local remaining_seconds = reset_time - time_stamp

    -- create new timer in ui.calendar.timer
    if not ui.calendar.timer then
        ui.calendar.timer = core.OsTimer:New()
    end

    ui.calendar.timer:AddDuration(remaining_seconds)

    -- convert time_in_seconds to 00:00:00 format
    local function GetCalenderTimerDisplayText(time_in_seconds)
        return os.date("!%H:%M:%S", time_in_seconds)
    end

    -- check to update various timer displays each tick
    function ui.calendar.timer:Tick()
        -- update calendar timer display if calendar is open
        if ui.calendar.is_visible then
            -- pass time remaining to timer display
            local time_remaining = ui.calendar.timer:GetTimeRemaining()
            local text = GetCalenderTimerDisplayText(time_remaining)
            ui.calendar:UpdateTimer(text)
        end

        -- update hell mode free count timer if hell mode panel is open
        if ui.hell_mode then
            local time_remaining = ui.calendar.timer:GetTimeRemaining()
            local text = GetCalenderTimerDisplayText(time_remaining)
            ui.hell_mode:UpdateTimer(text)
        end

        if ui.raid_boss and ui.raid_boss.open then
            local time_remaining = ui.calendar.timer:GetTimeRemaining()
            local text = GetCalenderTimerDisplayText(time_remaining)
            ui.raid_boss.UpdateDailyKeyRefillTimer(text)
        end
    end

    -- check server time when server day resets
    function ui.calendar.timer:Action()
        ui.calendar.timer = nil
        server.day_fetched = false

        ui.calendar:UpdateTimer("00:00:00")

        if ui.hell_mode then
            ui.hell_mode:UpdateTimer("00:00:00")
        end

        if ui.raid_boss then
            ui.raid_boss.UpdateDailyKeyRefillTimer("00:00:00")
        end

        -- DEBUG
        -- player.calendar.day = 200

        -- using timer as a safety buffer
        DCEI.TriggerAddTimerEventElapsed(CheckServerTimeOnTimerExpire, 2, true, true)
    end

    -- call tick function to immediately update displays
    ui.calendar.timer:Tick()
end

-- !SECTION

-- SECTION AIRSHIP MENU
function AttemptOpenAirshipMenu()
    local state = player.system.airship

    if state == system.locked then
        -- do nothing
    elseif state == system.active then
        local mode = ui.bottom_panel.airship_mode or "colony"

        -- open airship menu
        OpenAirshipMenu(mode)
        ui.loadout.town.airship:HidePip()
        ui.loadout.town.seasonal_colony_pip = false
    end
end

function BuildAirshipSequence(skip_sequence)
    -- prevent repeating this sequence in case of player quitting during colony tutorial
    if AIRSHIP_LANDED then
        return
    end

    if not skip_sequence then
        CreateEffect("Build FX 150%", 25.75, 10.2)
        CreateEffect("Build FX 100%", 25.4, 10.7)
        CreateEffect("Build FX 100%", 25.4, 9.2)
        CreateEffect("Build FX 100%", 26.2, 9.4)
        DCEI.Wait(0.0625)
    end

    DCEI.LoadPropLayer("Town_Airship_Props", true)
    DCEI.ChangePropLayerDisplay("Town_Airship_Props", true)

    DCEI.UnloadPropLayer("Town_Airship_Locked")
    DCEI.ChangePropLayerDisplay("Town_Airship_Locked", false)
    DCEI.ApplyBehaviorAsync(AIRSHIP_DOCK, DCEI.Behavior("Town Airship Landing SpawnAirship"), 1)

    if not skip_sequence then
        DCEI.Wait(0.875)
        DCEI.AnimateFrameAlpha(ui.loadout.town.airship.frame, 0, 1, 0.25, "InOutCirc")
    end

    UpdateAirshipLabel()

    AIRSHIP_LANDED = true
end

function OpenAirshipMenu(type)
    CloseActivePrompts()
    MoveCamera(cameras.airship)

    if not ui.create_airship_menu then
        ui.create_airship_menu = true
        ui.CreateAirShipMenu()
        HookupAirShipUI()
    end

    ui.town_glue:Hide()
    ui.loadout.town:Hide()
    ui.bottom_panel:SetState(type)

    -- show appropriate UI for menu type
    if type == "colony" then
        ui.colony.FadeIn()
        Tutorial07_Colony_StepB_End()
    elseif type == "challenge" then
        ui.challenge.FadeIn()
    elseif type == "raid_boss" then
        ui.raid_boss.FadeIn(levels.raid_boss, player.raid_boss_log)
    end

    if ui.loadout.town.seasonal_colony_pip then
        ui.colony.switch_button:ShowPip(true)
    end
end

function CloseAirshipMenu()
    MoveCamera(cameras.town)

    CloseActivePrompts()
    DCEI.SetFrameActive(ui.bottom_panel.right_inner.pip, false)
    DCEI.SetFrameActive(ui.bottom_panel.right_outer.pip, false)

    ui.town_glue:Show()
    ui.loadout.town:Show()

    -- attempt start airship timer here (if not started on map start)
    AttemptStartAirshipTimer()

    -- town upgrade tutorial check
    Tutorial08_TownUpgrade_Check()

    local previous_state = ui.bottom_panel.airship_mode

    if previous_state == "colony" then
        ui.colony.FadeOut()
    elseif previous_state == "challenge" then
        ui.challenge.FadeOut()
    elseif previous_state == "raid_boss" then
        ui.raid_boss.FadeOut()
    end

    local state = "town"
    ui.bottom_panel:SetState(state)

    -- for dev mode colony test sequence
    DEV.airship_mode_tracking = 0

    -- check raid turbo boost popup
    if player.system.popup_raid_turbo_boost == system.unlocked then
        player.system.popup_raid_turbo_boost = system.active
        bank.CommitSystemState()
        local key = "raid_turbo_pack"
        if AnyBundleItemsAreAvailable(key) then
            local function on_close()
                -- add raid turbo bundle to deals page on close
                local key = "raid_turbo_pack"
                local items_list = GetStoreCollectionData(key)
                if AnyBundleItemsAreAvailable(key) then
                    ui.shop:RemoveItems(items_list)
                    AddItemsToShopDealsFromCollectionKey(key)
                end
            end

            CloseActivePrompts()
            local from = "raid turbo popup"
            SetStoreEntrySource(from)
            local text = DCEI.GetTranslationText("shop/section/deals")
            OpenDealsPopup(key, text, on_close)
        end
    end
end

function ShowColonyMenu()
    ui.challenge.Hide()
    ui.raid_boss.Hide()
    ui.colony.Show()

    local state = "colony"
    ui.bottom_panel:SetState(state)

    -- for dev mode colony test sequence
    CheckUnlockAirshipTestMode()
end

function CheckUnlockAirshipTestMode()
    if DevMode() then
        DEV.airship_mode_tracking = DEV.airship_mode_tracking + 1

        if DEV.airship_mode_tracking >= 3 then
            bank.CommitAirshipTestMode()
        end
    end
end

function AttemptShowChallengeMenu()
    local nav = ui.bottom_panel.right_inner
    local nav_state = nav.states["colony"]
    local unlock_wave = GetChallengeUnlockWave()

    if nav_state.locked then
        local offset = -441
        local text = DCEI.GetTranslationText(
            "ui/town/building/airship/req",
            { x = "<color=green>" .. unlock_wave .. "</color>" }
        )
        text = core.StripChars(text, "-")

        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    else
        ShowChallengeMenu()
    end
end

function ShowChallengeMenu()
    ui.challenge.Show()
    ui.colony.Hide()
    ui.raid_boss.Hide()

    local state = "challenge"
    ui.bottom_panel:SetState(state)

    -- for dev mode colony test sequence
    DEV.airship_mode_tracking = 0
end

function AttemptShowRaidBossMenu()
    local nav = ui.bottom_panel.right_outer
    local nav_state = nav.states["colony"]
    local nav_state_challenge = nav.states["challenge"]
    local unlock_wave = GetRaidBossUnlockWave()

    if nav_state.locked or nav_state_challenge.locked then
        local offset = -441
        local text = DCEI.GetTranslationText(
            "ui/town/building/airship/req",
            { x = "<color=green>" .. unlock_wave .. "</color>" }
        )
        text = core.StripChars(text, "-")

        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    else
        ShowRaidBossMenu()
    end
end

function ShowRaidBossMenu()
    ui.challenge.Hide()
    ui.colony.Hide()
    ui.raid_boss.Show(levels.raid_boss, player.raid_boss_log)

    local state = "raid_boss"
    ui.bottom_panel:SetState(state)

    -- for dev mode colony test sequence
    DEV.airship_mode_tracking = 0
end

function StartChallengeBattle()
    local id = ui.challenge.select
    local index = ui.challenge.items[id].level_index
    local challenge = levels.challenge[index]

    levels.challenge.current = challenge

    -- check for challenge tutorial
    Tutorial09_Challenge_StepC()

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    ui.GlobalFadeOut()

    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.castle, duration)

        -- take advantage of fade out perform potentially costly memory unloading
        UnloadQueuedChunks()
        AttempToLoadMoreCastleChunks()

        CreateBattleScene()
        ui.GlobalFadeIn()

        -- battle start actions
        ui.challenge.Hide()
        ui.challenge.confirmation:Hide()

        local state = "battle"
        ui.bottom_panel:SetState(state)

        local delay = 0.25
        DCEI.TriggerAddTimerEventElapsed(function()
            WaveStartActions(challenge)
        end, delay, false, true)
    end, duration, false, true)
end

function GetCurrentRaidKeyExtraCost()
    local count = math.max(1, player.raid_key_extra_purchase_count)
    count = math.min(count, #raid_key_extra_purchase_cost)
    return raid_key_extra_purchase_cost[count]
end

function AttemptPurchaseRaidkeyExtra()
    local cost = GetCurrentRaidKeyExtraCost()
    local gems = GetPlayerGems()
    if cost <= gems then
        local spend_event = { type = "buy", target = "raid key" }
        SpendPlayerGems(cost, spend_event)
        player.raid_key_extra_purchase_count = player.raid_key_extra_purchase_count + 1
        bank.CommitRaidKeyExtraPurchaseCount()
        LogPurchaseRaidKeyExtra(cost)
        PlayGoldSpendSound()
        AddPlayerRaidKeyExtra(1)

        if ui.raid_boss then
            ui.raid_boss.UpdateRaidKeys()
            ui.raid_boss.UpdateRaidKeyExtraCost()
        end
    else
        local function OnStore()
            CloseAirshipMenu()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

function LogPurchaseRaidKeyExtra(cost)
    player.raid_key_extra_purchase_count_total = player.raid_key_extra_purchase_count_total + 1
    player.raid_key_extra_purchase_count_spend = player.raid_key_extra_purchase_count_spend + cost
    LogEvent(LogEvents.PurchaseRaidKey)
    bank.CommitPurchaseRaidKeyLog()
end

function StartRaidBossBattle()
    -- check if enough raid boss key
    local raid_key_daily = GetRaidKeyDailyCount()
    local raid_key_extra = GetRaidKeyExtraCount()
    if raid_key_daily <= 0 and raid_key_extra <= 0 then
        ui.ShowFeedbackNotEnoughRaidKey()
        return
    end
    local id = ui.raid_boss.select
    local index = ui.raid_boss.items[id].level_index
    local raid_boss = levels.raid_boss[index]

    levels.raid_boss.current = raid_boss
    wave.data.mode = "raid_boss"
    --local key = raid_boss.reward.name

    if player.system.raid_boss_tutorial ~= system.active then
        player.system.raid_boss_tutorial = system.active
        bank.CommitSystemState()
    end

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    ui.GlobalFadeOut()
    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.castle, duration)

        -- take advantage of fade out perform potentially costly memory unloading
        UnloadQueuedChunks()
        AttempToLoadMoreCastleChunks()

        CreateBattleScene()
        ui.GlobalFadeIn()

        -- battle start actions
        ui.raid_boss.Hide()
        ui.raid_boss.confirmation:Hide()

        local state = "battle"
        ui.bottom_panel:SetState(state)

        StartRaidBossWave(raid_boss)
    end, duration, false, true)
end

function StartRaidBossWave(wave_data)
    local dx, dy = 0, -1
    local x, y = 24, 35
    local unit = DCEI.CreateUnitAsync(-1, -1, wave_data.boss, x, y, dx, dy)
    local health_stacks = (wave_data.boss_health - 1) * 100
    local attack_stacks = wave_data.boss_attack * 100
    wave_data.health_modifier = 1.25
    if wave_data.growth then
        local key = "Raid Boss " .. wave_data.index
        local victory_count = player.raid_boss_log[key] or 0
        local stack = victory_count
        wave_data.rating_growth = math.floor(victory_count * wave_data.growth.rating_step)
        -- use a clamp to avoid behavior stacks overflow
        wave_data.rating_growth = math.min(wave_data.rating_growth, 585100)
        if wave_data.peak then
            local dynamic_rating = wave.data.GetMax() + 100
            local actual_rating = math.max(wave_data.rating + wave_data.rating_growth, dynamic_rating)
            wave_data.rating_growth = actual_rating - wave_data.rating
            wave_data.rating_growth = math.min(wave_data.rating_growth, 585100)
            stack = math.floor(wave_data.rating_growth / wave_data.growth.rating_step)
        end
        health_stacks = health_stacks + wave_data.growth.health * stack * 100
        attack_stacks = attack_stacks + wave_data.growth.attack * stack * 100
    end

    if not wave.data.FAST_BATTLE then
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Monster Attribute HealthPower"), health_stacks)
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Monster Attribute AttackPower"), attack_stacks)
    end

    DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Colony Castle Start Invulnerability"), 1)

    wave_data.objective = unit
    wave.data.is_boss_spawned = true

    local delay = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        WaveStartActions(wave_data)
    end, delay, false, true)
end

function ColonyTargetingCheck(unit, wave_data)
    if wave_data.mode == "colony" or wave_data.mode == "raid_boss" then
        DCEI.ApplyBehaviorAsync(unit, DCEI.Behavior("System Hero InCombat PeriodicRetarget"))
    end
end

function RestoreDefaultScene()
    SetPropLayer("_Default_SceneLower", true)
    SetPropLayer("_Default_SceneUpper", true)

    for _, set in ipairs(SCENE) do
        SetPropLayer(set, false)
    end
    UpdatePropLayer_AndHandleLoading()
end

function StartTrainingBattle(duration)
    local dx, dy = 0, -1
    local x, y = 24, 28
    local name = DCEI.Unit("System Training Dummy")
    local unit = DCEI.CreateUnitAsync(-1, -1, name, x, y, dx, dy)

    local wave_data = {}
    wave_data.objective = unit
    wave_data.mode = "training"

    ui.CreateBossBar(unit, "TRAINING DUMMY")
    table.insert(wave.data.units, unit)

    local state = "battle"
    ui.bottom_panel:SetState(state)

    WaveStartActions(wave_data)

    DCEI.Wait(duration)
    WaveEndActions()
end

function StartColonyBattle()
    local id = ui.colony.select

    if id == 0 then
        return
    end

    local index = ui.colony.items[id].level_index
    local colony_set = levels.colony
    if index > #levels.colony then
        index = index - #levels.colony
        colony_set = levels.seasonal_colony
    end
    local colony = colony_set[index]
    if colony.seasonal then
        if ui.colony.seasonal_region.free_remaining <= 0 then
            local cost = 5
            local gems = GetPlayerGems()
            if cost <= gems then
                local spend_event = { type = "buy", target = "seasonal colony" }
                SpendPlayerGems(cost, spend_event)
                PlayGoldSpendSound()
            else
                local function OnStore()
                    CloseAirshipMenu()
                end

                local offset = -356
                ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
                return
            end
        end
    end
    levels.colony.current = colony

    -- check for colony tutorial
    Tutorial07_Colony_StepE()

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    ui.GlobalFadeOut()
    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.castle, duration)

        -- take advantage of fade out perform potentially costly memory unloading
        UnloadQueuedChunks()
        AttempToLoadMoreCastleChunks()

        CreateColonyBattleSet(colony)
        ui.GlobalFadeIn()

        -- battle start actions
        ui.colony.Hide()
        ui.colony.confirmation:Hide()

        local state = "battle"
        ui.bottom_panel:SetState(state)

        StartColonyWave(colony)
    end, duration, false, true)
end

function StartColonyWave(wave_data)
    local dx, dy = 0, 1
    local x, y = 24, 31
    local spawn_height = 32
    local colony_name = DCEI.Unit("Colony Castle Controller")
    local unit = DCEI.CreateUnitAsync(-1, -1, colony_name, x, y, dx, dy)
    local health = wave_data.health or levels.GetColonyHealth(wave_data.rating)
    wave_data.health_modifier = levels.GetColonyHealthModifier(wave_data.rating)

    if not wave.data.FAST_BATTLE then
        local stacks = math.floor(health - 100)
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Monster Attribute HealthPower"), stacks)
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Colony Castle Start Invulnerability"), 1)
    end

    local text = DCEI.GetTranslationText("general/wild_castle")
    ui.CreateBossBar(unit, text)

    table.insert(wave.data.units, unit)
    DCEI.SetUnitPosition3DAsync(unit, x, 0.2, y)

    wave_data.objective = unit
    wave_data.objective_height = spawn_height

    local delay = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        WaveStartActions(wave_data)
    end, delay, false, true)
end

-- change background props for challenge and raid boss
function CreateBattleScene()
    if game.special_scene then
        SCENE = {}
        SetPropLayer("_Default_SceneLower", false)
        SetPropLayer("_Default_SceneUpper", false)

        local set = game.special_scene.set.main
        SetPropLayer(set, true)
        table.insert(SCENE, set)

        UpdatePropLayer_AndHandleLoading()
    end
end

function CreateColonyBattleSet(wave_data)
    if wave_data.scene and wave_data.colony then
        SCENE = {}
        SetPropLayer("_Default_SceneLower", false)
        SetPropLayer("_Default_SceneUpper", false)

        local set = wave_data.scene.set.main
        SetPropLayer(set, true)
        table.insert(SCENE, set)

        local backdrop = wave_data.scene.set.back
        if backdrop then
            SetPropLayer(backdrop, true)
            table.insert(SCENE, backdrop)
        end

        local colony = wave_data.colony
        SetPropLayer(colony, true)
        table.insert(SCENE, colony)

        UpdatePropLayer_AndHandleLoading()
    end
end

function ExitColonyBattle(is_victory)
    local colony = levels.colony.current
    levels.colony.current = nil

    ui.loadout.hero.Hide()
    ui.battle.floating_speedup:Hide()
    ui.battle.floating_autobattle:Hide()

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    -- fade out using default time/end alpha, but 0.6 start due to challenge mode special case
    ui.GlobalFadeOut(nil, nil, 0.6)

    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.airship, duration)

        -- take advantage of screen transition to do costly GC
        PreloadNextTownChunks()
        RestoreDefaultScene()
        GarbageCollect()

        ui.GlobalFadeIn()

        -- battle exit actions
        if ui.prompt.wave_result then
            DCEI.SetFrameActive(ui.prompt.wave_result.frame, false)
        end

        if is_victory then
            CompleteColony(colony)
        end

        -- show colony menu must occur after CompleteColony
        ui.colony.Show()
        ui.screen_mask.under:Show()

        if is_victory then
            -- check for colony complete tutorial
            Tutorial07_Colony_StepF(colony)
        end

        local state = "colony"
        ui.bottom_panel:SetState(state)
    end, duration, false, true)
end

function ExitChallengeBattle(is_victory)
    local challenge = levels.challenge.current
    levels.challenge.current = nil

    ui.loadout.hero.Hide()
    ui.battle.floating_speedup:Hide()
    ui.battle.floating_autobattle:Hide()

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    -- fade out using default time/end alpha, but 0.6 start due to challenge mode special case
    ui.GlobalFadeOut(nil, nil, 0.6)

    if ui.prompt.wave_result then
        ui.prompt.wave_result:AnimateOut()
    end

    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.airship, duration)

        -- take advantage of screen transition to do costly GC
        RestoreDefaultScene()
        GarbageCollect()

        ui.GlobalFadeIn()

        -- battle exit actions
        if ui.prompt.wave_result then
            DCEI.SetFrameActive(ui.prompt.wave_result.frame, false)
        end

        if is_victory then
            CompleteChallenge(challenge)
            if challenge.id == "challenge02" then
                -- enter gear tutorial
                Tutorial10_EquipGear_Check()
            end
        end

        -- show challenge menu must occur after CompleteChallenge
        ui.challenge.Show()
        ui.screen_mask.under:Show()

        local state = "challenge"
        ui.bottom_panel:SetState(state)
    end, duration, false, true)
end

function ExitRaidBossBattle()
    local raid_boss = levels.raid_boss.current
    levels.raid_boss.current = nil

    ui.loadout.hero.Hide()
    ui.battle.floating_speedup:Hide()
    ui.battle.floating_autobattle:Hide()

    -- fade out/fade in sequence in case of frame drops from dynamic load/unload of prop layers
    -- fade out using default time/end alpha, but 0.6 start due to challenge mode special case
    ui.GlobalFadeOut(nil, nil, 0.6)

    if ui.prompt.wave_result then
        ui.prompt.wave_result:AnimateOut()
    end

    local duration = 0.25
    DCEI.TriggerAddTimerEventElapsed(function()
        local duration = 0
        MoveCamera(cameras.airship, duration)

        -- take advantage of screen transition to do costly GC
        RestoreDefaultScene()
        GarbageCollect()

        ui.GlobalFadeIn()

        -- battle exit actions
        if ui.prompt.wave_result then
            DCEI.SetFrameActive(ui.prompt.wave_result.frame, false)
        end

        -- show challenge menu must occur after CompleteChallenge
        ui.raid_boss.Show(levels.raid_boss, player.raid_boss_log)
        ui.screen_mask.under:Show()

        if not player.system.first_raid_boss or player.system.first_raid_boss == system.locked then
            player.system.first_raid_boss = system.active
            ShowShopChestsPip()
            bank.CommitSystemState()
        end

        local state = "raid_boss"
        ui.bottom_panel:SetState(state)
    end, duration, false, true)
end

function ShowChallengeRewardFrame(on_close)
    local challenge = levels.challenge.current
    ui.NewChallengeRewardFrame(challenge, data, on_close)
end

function CompleteChallenge(challenge, skip_commit)
    -- check if challenge exists for forwards compatability
    if challenge then
        local reward_type = challenge.reward.type
        local reward_id = challenge.reward.name
        local unit_data = nil
        local purchase_func = nil

        challenge:Complete()

        if reward_type == "heroes" then
            unit_data = data.heroes[reward_id]
            unit_data = data.heroes[reward_id]
            unit_data = data.heroes[reward_id]
            purchase_func = PurchaseHero

            unit_data.locked = false
        elseif reward_type == "gear" then
            unit_data = data.gear[reward_id]
            purchase_func = PurchaseGear

            unit_data.locked = false
        elseif reward_type == "talents" then
            unit_data = data.talents[reward_id]
            purchase_func = UnlockTalent

            unit_data.locked = false
        end

        if not skip_commit then
            -- when actually completing challenge, not through load
            ui.challenge.Update()
            bank.CommitChallengeState()

            if unit_data and purchase_func then
                purchase_func(unit_data)
            end

            if reward_type == "gear" then
                ui.glue.gear_menu:ShowPip()
            end
        end
    end
end

function UnlockTalent(unit_data)
    if unit_data.layout then
        unit_data.layout.Unlock()
    end
    ui.glue.talent_menu:ShowPip()
end

function CompleteColony(colony, skip_commit)
    -- check if colony exists for forwards compatability
    if colony then
        colony:Complete()

        -- skip_commit occurs when loading from bank during initialization
        if not skip_commit then
            ui.colony.Update()
            if colony.seasonal then
                bank.CommitSeasonalColonyState()
            else
                bank.CommitColonyState()
                -- check if change to next map
                ChangeMapCheck()
            end
        end

        -- completing first colony can unlock town hall (test A)
        CheckUpdateTownHallLabel()
    end
end

ui.CompleteColony = CompleteColony

function InitializeSeasonalColonies(time_stamp)
    local need_commit = false
    if not levels.seasonal_colony.reset_time then
        levels.seasonal_colony.reset_time = time_stamp + SEASONAL_COLONIES_RESET_TIME
        need_commit = true
    end
    if time_stamp > levels.seasonal_colony.reset_time then
        for k, colony in ipairs(levels.seasonal_colony) do
            colony.complete = false
        end
        local time_difference = time_stamp - levels.seasonal_colony.reset_time
        if time_difference > SEASONAL_COLONIES_RESET_TIME then
            local mod = time_difference % SEASONAL_COLONIES_RESET_TIME
            levels.seasonal_colony.reset_time = time_stamp - mod + SEASONAL_COLONIES_RESET_TIME
        else
            levels.seasonal_colony.reset_time = levels.seasonal_colony.reset_time + SEASONAL_COLONIES_RESET_TIME
        end
        need_commit = true
    end
    if need_commit then
        -- decide this season's colony start rating
        levels.seasonal_colony.start_rating = math.max(400, wave.data.GetMax())
        levels.seasonal_colony.start_rating = math.min(30000, levels.seasonal_colony.start_rating)
        bank.CommitSeasonalColonyState()
        ui.bottom_panel:ShowTownPipAnimated()
        ui.loadout.town.airship:ShowPipAnimated()
        ui.loadout.town.seasonal_colony_pip = true
    end
    if ui.colony then
        ui.colony.Update()
    end

    local time_remaining = levels.seasonal_colony.reset_time - time_stamp
    if time_remaining > 0 then
        local timer = core.OsTimer:New()
        ui.seasonal_colony_timer = timer

        timer:AddDuration(time_remaining)

        function timer:Tick()
            local time = self:GetTimeRemaining()
            if ui.colony and ui.colony.open then
                ui.colony:SeasonalTimerUpdate(time)
            end
        end

        function timer:Action()
            DCEI.TriggerAddTimerEventElapsed(function()
                InitializeSeasonalColonies(os.time())
            end, 2, true, true)
            ui.seasonal_colony_timer = nil
        end
    end

    --assign scene set
    local seasonal_scene = levels.seasonal_colony[1].scene
    local colony_castle_set = {}
    for k, colony in ipairs(levels.colony) do
        if colony.scene.map_index == levels.colony.current_map_id then
            local current_scene = colony.scene
            seasonal_scene.set = current_scene.set
            levels.seasonal_colony.monster_set = colony.monsters
            colony_castle_set[colony.colony] = true
        elseif colony.scene.map_index > levels.colony.current_map_id then
            break
        end
    end
    local castle_list = {}
    for castle, v in pairs(colony_castle_set) do
        table.insert(castle_list, castle)
    end
    for index, colony in ipairs(levels.seasonal_colony) do
        colony.monsters = levels.seasonal_colony.monster_set
        colony.colony = castle_list[math.random(#castle_list)]
        colony.rating = levels.seasonal_colony.start_rating + levels.seasonal_colony.rating_growth[index]
        colony.value = levels.seasonal_colony:GenerateColonyValue(colony.rating, index)
    end
end

-- !SECTION

-- SECTION TOWN HALL MENU
function HookupTownHallMenuData()
    ui.town_menu.data = {
        max_level = data.town_upgrade.max_upgrade_level,
    }

    -- get upgrade level of town
    function ui.town_menu.data.GetUpgradeLevel()
        return GetTownUpgradeLevel()
    end

    -- get cost of next upgrade
    function ui.town_menu.data.GetNextUpgradeCost()
        local upgrade_level = GetTownUpgradeLevel() + 1
        return data.town_upgrade:GetCost(upgrade_level)
    end

    -- get current elixir income
    function ui.town_menu.data.GetElixirIncome()
        return data.town_upgrade:GetElixirIncome()
    end

    -- get current elixir cap
    function ui.town_menu.data.GetElixirCap()
        return data.town_upgrade:GetElixirCap()
    end

    -- get current town level
    function ui.town_menu.data.GetLevel()
        return data.town_upgrade:GetVisualLevel()
    end

    -- get max town level
    function ui.town_menu.data.GetMaxLevel()
        return data.town_upgrade.level.max
    end

    function ui.town_menu.data.GetMaxUpgradeLevel()
        return data.town_upgrade.max_upgrade_level
    end

    -- get required upgrade count for next town level
    function ui.town_menu.data.GetNextLevelRequirement()
        local level = data.town_upgrade:GetVisualLevel()
        return data.town_upgrade:GetLevelUpgradeRequirement(level)
    end

    -- get required upgrade count for last town level
    function ui.town_menu.data.GetLastLevelRequirement()
        local level = data.town_upgrade:GetVisualLevel() - 1
        return data.town_upgrade:GetLevelUpgradeRequirement(level)
    end

    -- update contract item count
    function ui.town_menu.data.UpdateContractDisplay()
        local item_count = GetTownContracts()
        ui.town_menu.item_component:Update(item_count)
    end
end

function AttemptOpenTownHallMenu()
    local colonies_completed = #levels.colony.complete
    local unlock_check = colonies_completed >= 1

    if TestB("deferred_town_hall") then
        unlock_check = wave.data.GetMax() >= 50
    end

    if unlock_check then
        OpenTownHallMenu()
        Tutorial08_TownUpgrade_StepB()
    end
end

function OpenTownHallMenu()
    -- prevent menu from opening when wave started
    if wave.data.is_active then
        return
    end

    if not ui.town_menu then
        ui.town_menu = ui.NewTownHallMenu()
    end

    local function OnClose(skip_camera)
        if not skip_camera then
            MoveCamera(cameras.town)
            DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, true)
            DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
            DCEI.SetFrameActive(ui.loadout.town.vault.frame, true)
        end

        Tutorial08_1_Mine_Vault_Check()
    end

    local callbacks = {
        upgrade = AttemptUpgradeTownHall,
        monster_lab = AttemptOpenMonsterLab,
        on_close = OnClose,
    }

    -- initialize menu callbacks and data
    if not ui.town_menu.data then
        HookupTownHallMenuData()
    end

    if not ui.town_menu.callbacks then
        ui.town_menu.callbacks = callbacks
    end

    CloseActivePrompts()
    MoveCamera(cameras.town_hall)
    DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.airship.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.vault.frame, false)

    ui.town_menu:Show()
end

function AttemptUpgradeTownHall()
    local max_level = data.town_upgrade.max_upgrade_level
    local level = GetTownUpgradeLevel()
    local cost = data.town_upgrade:GetCost(level + 1)
    local gold = GetPlayerGold()
    local contract_needed = 1
    local success = false

    if level == max_level then
        ui.ShowFeedbackMaxLevel()
    elseif contract_needed > GetTownContracts() then
        ui.ShowFeedbackNotEnoughTownContracts()
    elseif cost <= gold then
        local spend_event = { type = "upgrade", target = "TownHall" }
        SpendPlayerGold(cost, spend_event)
        AddPlayerContract_Town(-1)
        PlayGoldSpendSound()

        UpgradeTownHall()
        ui.town_menu:Update()

        success = true
    else
        local function OnStore()
            -- skip worker menu close camera transition to preserve store camera transition
            local skip_camera = true
            ui.town_menu:Hide(skip_camera)
        end

        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local offset = -214
        ui.ShowFeedbackNeedMoreGold(offset, ui.upper, OnStore, OnStoreClose)
    end

    Tutorial08_TownUpgrade_StepC()

    return success
end

-- SECTION Monster lab
function CheckMonsterLabUnlock()
    local unlock_check = wave.data.GetMax() >= 80
    ui.glue.monster_lab_menu:SetEnabled(unlock_check)
    if
        unlock_check
        and (not player.system.first_open_monster_lab or player.system.first_open_monster_lab == system.locked)
    then
        ui.bottom_panel:ShowTownPipAnimated()
        ui.glue.monster_lab_menu:ShowPipAnimated()
    end
end

function AttemptOpenMonsterLab()
    local lock = wave.data.GetMax() < 80
    if lock then
        local offset = -441
        local unlock_wave = 80
        local text = DCEI.GetTranslationText(
            "ui/town/building/airship/req",
            { x = "<color=green>" .. unlock_wave .. "</color>" }
        )
        text = core.StripChars(text, "-")

        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    else
        if ui.town_menu then
            ui.town_menu:Hide(true)
        end

        if ui.monster_lab then
            DCEI.DestroyFrame(ui.monster_lab.Frame)
        end

        ui.bottom_panel:HideTownPip()
        ui.glue.monster_lab_menu:HidePip()
        if not player.system.first_open_monster_lab or player.system.first_open_monster_lab == system.locked then
            player.system.first_open_monster_lab = system.active
            bank.CommitSystemState()
            local event = LogEvents.FirstOpenMonsterLab()
            LogEvent(event)
        end

        ui.monster_lab = ui.NewMonsterLabMenu(player.monster_data)

        local function OnClose(skip_camera)
            if not skip_camera then
                MoveCamera(cameras.town)
                DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, true)
                DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
                DCEI.SetFrameActive(ui.loadout.town.vault.frame, true)
            end
        end

        local callbacks = {
            on_close = OnClose,
            purchase_weak = AttemptPurchaseMonsterWeakPoint,
            use_weak = UseMonsterWeakPoint,
            open_monster_card = OpenMonsterCard,
        }
        ui.monster_lab.callbacks = callbacks
    end
end

function OpenMonsterCard(monster)
    local kill_count = player.monster_data[monster.name] and player.monster_data[monster.name].kill or 0
    local unlocked = kill_count >= monster.unlock_kill
    if unlocked and not player.monster_data[monster.name].unlocked then
        player.monster_data[monster.name].unlocked = true
        local event = LogEvents.UnlockMonsterInfo(monster)
        LogEvent(event)
        bank.CommitMonsterData()
    end
    return ui.NewMonsterCard(monster)
end

function AttemptPurchaseMonsterWeakPoint(monster, index)
    local cost = monster.weak_point[index].cost
    local gems = GetPlayerGems()

    if cost <= gems then
        local spend_event = { type = "buy", target = "monster weak" }
        SpendPlayerGems(cost, spend_event)

        player.monster_lab_token = player.monster_lab_token + cost
        LogEvent(LogEvents.PurchaseMonsterWeakPoint)
        bank.CommitMonsterLabSpendToken()
        local event = LogEvents.PurchaseMonsterWeak(monster, index, player.monster_lab_token)
        LogEvent(event)

        PlayGoldSpendSound()
        player.monster_data[monster.name].weak[index] = true
        local skip_commit = true
        UseMonsterWeakPoint(monster, index, skip_commit)
        bank.CommitMonsterData()

        if ui.monster_card then
            ui.monster_card:UpdateWeakPoint()
        end
    else
        local function OnStore()
            ui.monster_card:Close()
            local skip_camera = true
            ui.monster_lab:Close(skip_camera)
        end

        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local offset = -470
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore, OnStoreClose)
    end
end

function UseMonsterWeakPoint(monster, index, skip_commit)
    player.monster_data[monster.name].select_weak = index

    if ui.monster_card then
        ui.monster_card:UpdateWeakPoint()
    end

    if not skip_commit then
        bank.CommitMonsterData()
    end
end

-- !SECTION
function UpgradeTownHall()
    local town_upgrades = data.town_upgrade.GetUpgradeLevel()
    local old_town_level = data.town_upgrade:GetVisualLevel(town_upgrades)
    local new_town_level = data.town_upgrade:GetVisualLevel(town_upgrades + 1)

    AddTownUpgradeLevel(1)

    -- start elixir timer on first upgrade
    if GetTownUpgradeLevel() == 1 then
        AttemptStartElixirTimer()
    end

    -- only change town layers if town levels up
    if new_town_level ~= old_town_level then
        local animate = true
        BuildTown(new_town_level, animate)

        -- add elixir gain on town level up for instant player feedback
        GenerateElixir()

        -- check for unloadable prop layers
        AttemptQueueUnloadTownChunk(new_town_level)
    end

    bank.CommitTownLevel()
end

-- SECTION VAULT, only affect old players have abtest offline gold, but can't remove
function ClickVault()
    if GetVaultGold() > 0 then
        CollectVaultGold()
    else
        NewVaultMenu()
    end
end

function NewVaultMenu()
    if wave.data.is_active or event_mode.IsActive() then
        -- prevent worker menu from opening when wave started
        return
    end

    if tutorial.state == 8.1 then
        tutorial.state = 0
        ui.DestroyBigHand()
        ui.prompt.tutorial.Close()
        UnblockInput()
        AttemptStartVaultTimer()
        -- add vault gain instant player feedback
        GenerateVaultGold()
    end

    local function OnClose(skip_camera)
        -- close NotEnoughGoldPrompt on menu close
        CloseActivePrompts()

        -- return camera to town
        if not skip_camera then
            MoveCamera(cameras.town)
        end

        DCEI.SetFrameActive(ui.loadout.town.hall.frame, true)
        DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
        DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, true)

        -- remove worker menu reference
        ui.vault_menu = nil
    end

    local callbacks = {
        vault_gold = GetVaultGold,
        claim_vault_gold = CollectVaultGold,
        upgrade = AttemptUpgradeVault,
        on_close = OnClose,
    }

    CloseActivePrompts()
    MoveCamera(cameras.vault)
    DCEI.SetFrameActive(ui.loadout.town.hall.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.airship.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, false)

    data.vault:CheckUpgradable(wave.data.GetMax())
    ui.vault_menu = ui.NewVaultMenu(data.vault, callbacks)
end

function AttemptUpgradeVault()
    local cost = data.vault:GetCost()
    local token = GetPlayerGems()

    if GetVaultLevel() == data.vault.max_level then
        ui.ShowFeedbackMaxLevel()
    elseif not data.vault.upgradable then
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
        if ui.prompt.feedback then
            ui.prompt.feedback.Close()
        end

        local offset = -441
        unlock_wave = data.vault:GetNextUpgradeUnlockWave()
        local text = DCEI.GetTranslationText(
            "ui/town/building/airship/req",
            { x = "<color=green>" .. unlock_wave .. "</color>" }
        )
        text = core.StripChars(text, "-")

        ui.prompt.tutorial = ui.NewPromptMessageClosable(text, offset)
    elseif cost <= token then
        local spend_event = { type = "upgrade", target = "vault" }
        SpendPlayerGems(cost, spend_event)

        UpgradeVault()
        PlayGoldSpendSound()

        ui.vault_menu:Update()
    else
        local function OnStore()
            -- skip worker menu close camera transition to preserve store camera transition
            local skip_camera = true
            ui.vault_menu:Close(skip_camera)
        end

        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local offset = -260
        ui.ShowFeedbackNeedMoreTokens(offset, ui.upper, OnStore, OnStoreClose)
    end
end

function UpgradeVault()
    local worker_data = data.vault
    AddVaultLevel(1)

    data.vault:CheckUpgradable(wave.data.GetMax())
    if ui.vault_menu then
        ui.vault_menu:Update()
    end

    bank.CommitVaultLevel()
end

function CollectVaultGold()
    local id = DCEI.CompileActorAction('{"playAnimation":{"clipId":6}}')
    DCEI.SendActorAction(VAULT, id)
    if not ui.loadout.town.vault.collecting then
        ui.loadout.town.vault.collecting = true

        if not DCEI.IsOnline() then
            ui.ShowFeedbackOffline()
        end

        local vault_gold = GetVaultGold()
        if vault_gold > 0 then
            AddPlayerGold(vault_gold)
            bank.CommitGold()
            ui.GoldRewardFeedback(math.floor(vault_gold), nil, ui.loadout.town.vault.gold_indicator.progress)
            PlayGoldSpendSound()
            SetVaultGold(0)
        end

        bank.CommitVaultGold()

        ScheduleVaultNotification()

        if ui.vault_menu then
            ui.vault_menu:UpdateProgress()
        end
        ui.loadout.town.vault.gold_indicator:Update(0)
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.loadout.town.vault.collecting = false
        end, 1, false, true)
    end
end

function AttemptStartVaultTimer()
    -- don't create vault timer if one already exists
    if data.vault.timer then
        return
    end

    -- don't start vault timer if wave progress too low
    if wave.data.GetMax() < 5 or player.system.vault == system.locked then
        return
    end

    local timer = core.RealTimer:New()
    data.vault.timer = timer
    data.vault.colony = levels.colony
    data.vault.timer.interval = data.vault.interval
    data.vault.timer.duration = 0

    function timer:Tick()
        self.duration = self.duration + GLOBAL_TICK_RATE

        if self.duration >= self.interval then
            AttemptHandleVaultIncome()

            -- reset duration
            self.duration = 0
        end
    end
end

function AttemptHandleVaultIncome()
    local timer = data.vault.timer

    if timer and GetVaultGold() < data.vault.capacity then
        local income = data.vault:GetVaultIncome() * (timer.duration / data.vault.interval)
        timer.duration = 0
        GenerateVaultGold(income)
    end
end

function GenerateVaultGold(value)
    -- default value is vault income/min
    value = value or data.vault:GetVaultIncome()

    AddVaultGold(value)

    if not wave.data.is_active then
        local progress = GetVaultGold() / data.vault.capacity
        ui.loadout.town.vault.gold_indicator:Update(progress)
        if progress > 0.95 then
            ui.bottom_panel:ShowTownPip()
        end

        if ui.vault_menu then
            ui.vault_menu:UpdateProgress()
        end

        local function ResultCallbcak(time_stamp)
            data.vault.time_record = time_stamp or data.vault.time_record
            bank.CommitVaultRecord()
        end

        AttemptGetServerTime(ResultCallbcak)
        bank.CommitVaultGold()
    end
end

-- !SECTION
-- !SECTION

-- SECTION Craft stuff, in the inventory menu
function AttemptStartCraftTimer()
    if ui.craft.refreshing then
        return
    end

    -- when player start game agian for a long time
    if player.craft_refresh < os.time() then
        local diff = os.time() - player.craft_refresh
        if diff > CRAFT_AUTO_REFRESH_TIME then
            local mod = diff % CRAFT_AUTO_REFRESH_TIME
            player.craft_refresh = os.time() - mod + CRAFT_AUTO_REFRESH_TIME
        else
            player.craft_refresh = player.craft_refresh + CRAFT_AUTO_REFRESH_TIME
        end
        math.randomseed(os.time())
        GenerateRandomCraftTypes()
    end

    -- don't create craft timer if one already exists
    if ui.craft.timer then
        return
    end

    local timer = core.OsTimer:New()
    ui.craft.timer = timer
    ui.craft.timer:AddDuration(player.craft_refresh - os.time())

    -- convert time_in_seconds to 00:00:00 format
    local function GetCraftTimerDisplayText(time_in_seconds)
        return os.date("!%H:%M:%S", time_in_seconds)
    end

    -- check to update various timer displays each tick
    function ui.craft.timer:Tick()
        -- update craft timer display if craft menu is open
        if ui.craft.active then
            -- pass time remaining to timer display
            local time_remaining = ui.craft.timer:GetTimeRemaining()
            local text = GetCraftTimerDisplayText(time_remaining)
            ui.craft:UpdateTimer(text)
        end
    end

    function ui.craft.timer:Action()
        ui.craft.timer = nil

        ui.craft:UpdateTimer("00:00:00")

        ui.craft.refreshing = true
        ui.craft:DisplayButton(false)

        GenerateRandomCraftTypes()
        player.craft_refresh = os.time() + CRAFT_AUTO_REFRESH_TIME
        ui.craft:UpdateCraftList(player.craft_types)

        -- using timer as a safety buffer
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.craft.refreshing = false
            AttemptStartCraftTimer()
            ui.craft:DisplayButton(true)
        end, 2, true, true)
    end

    -- call tick function to immediately update displays
    ui.craft.has_legendary = player.craft_legendary
    ui.craft:UpdateCraftList(player.craft_types)
    ui.craft.timer:Tick()

    bank.CommitCraft()
end

function GenerateRandomCraftTypes(has_legendary)
    player.craft_legendary = has_legendary
        or (GetCraftLevel() == CRAFT_MAX_LEVEL and math.random() <= LEGENDARY_REFRESH_CHANCE)
    ui.craft.has_legendary = player.craft_legendary
    for i = 1, 3 do
        player.craft_types[i] = math.random(1, #data.equipments.type)
    end
end

function RefreshCraftList(index, has_legendary)
    if index then
        player.craft_types[index] = math.random(1, #data.equipments.type)
        player.craft_legendary = false
        if index == 1 and GetCraftLevel() == CRAFT_MAX_LEVEL then
            player.craft_legendary = math.random() <= LEGENDARY_REFRESH_CHANCE
            ui.craft.has_legendary = player.craft_legendary
            if player.craft_legendary and player.craft_types[1] > 4 then
                player.craft_types[1] = math.random(1, 4)
            end
        end
        bank.CommitCraftTypes()
    else
        if ui.craft.timer then
            ui.craft.timer:Destroy()
            ui.craft.timer = nil
        end

        GenerateRandomCraftTypes(has_legendary)
        player.craft_refresh = os.time() + CRAFT_AUTO_REFRESH_TIME
        AttemptStartCraftTimer()
    end
    ui.craft:UpdateCraftList(player.craft_types)
end

ui.RefreshCraftList = RefreshCraftList

function AttemptRefreshCraftListGold()
    if ui.craft.refreshing then
        return
    end

    local cost = math.floor(wave.data.GetGoldValue(wave.data.GetMax()) * 5.5)
    local gold = GetPlayerGold()

    if cost <= gold then
        local spend_event = { type = "Refresh", target = "Craft" }
        SpendPlayerGold(cost, spend_event)
        PlayGoldSpendSound()

        RefreshCraftList()
        IncrementCraftRefreshCount("gold")
    else
        local function OnStore()
            ui.craft:Close()
            ui.CloseUnitCard()
            ui.item_menu:Close()
        end

        local offset = -530
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)
    end
end

ui.RefreshCraftGold = AttemptRefreshCraftListGold

function AttemptRefreshCraftListToken()
    if ui.craft.refreshing then
        return
    end

    local cost = 10
    local gems = GetPlayerGems()

    if cost <= gems then
        local spend_event = { type = "Refresh", target = "Craft" }
        SpendPlayerGems(cost, spend_event)
        PlayGoldSpendSound()

        RefreshCraftList()
        IncrementCraftRefreshCount("tokens")
    else
        local function OnStore()
            ui.craft:Close()
            ui.CloseUnitCard()
            ui.item_menu:Close()
        end

        local offset = -530
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

ui.RefreshCraftToken = AttemptRefreshCraftListToken

function AttemptUpgradeCraft()
    local level = GetCraftLevel()
    local cost = data.equipments.craft_system[level].upgrade_cost
    local max_level = CRAFT_MAX_LEVEL
    local gems = GetPlayerGems()

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
    elseif cost <= gems then
        local spend_event = { type = "upgrade", target = "CraftSystem" }
        SpendPlayerGems(cost, spend_event)

        --UpgradeCarft
        AddCraftLevel(1)
        LogEvent(LogEvents.UpgradeCraft(GetCraftLevel()))
        bank.CommitCraft()
        -- guarantee a legnedary recipie when first upgrade to level 5
        local has_legendary = GetCraftLevel() == 5
        RefreshCraftList(nil, has_legendary)

        PlayGoldSpendSound()
    else
        local function OnStore()
            ui.craft:Close()
            ui.CloseUnitCard()
            ui.item_menu:Close()
        end

        local offset = -530
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

-- !SECTION

-- SECTION WORKER MENU
function NewWorkerMenu()
    if wave.data.is_active then
        -- prevent worker menu from opening when wave started
        return
    end

    local function OnClose(skip_camera)
        -- close NotEnoughGoldPrompt on menu close
        CloseActivePrompts()

        -- return camera to town
        if not skip_camera then
            MoveCamera(cameras.town)
        end
        
        DCEI.SetFrameActive(ui.loadout.town.hall.frame, true)
        DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
        DCEI.SetFrameActive(ui.loadout.town.vault.frame, true)

        if not event_mode.IsActive() then
            -- show mine bundle after finishing mine tutorial
            Tutorial05_Mine_SurfaceMineBundle()
        else
            DCEI.TriggerAddTimerEventElapsed(function()
                DCEI.SetFrameActive(ui.potion_lab.background, true)
            end, 0.3)
            DCEI.TriggerAddTimerEventElapsed(function()
                DCEI.SetFrameActive(ui.secret_lab.background, true)
            end, 0.3)
        end

        -- spawn
        -- remove worker menu reference
        ui.worker_menu = nil
    end

    local callbacks = {
        upgrade = AttemptUpgradeDwarves,
        on_close = OnClose,
    }

    CloseActivePrompts()
    MoveCamera(cameras.mines)
    DCEI.SetFrameActive(ui.loadout.town.hall.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.airship.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.vault.frame, false)

    ui.worker_menu = ui.NewWorkerMenu(data.dwarf_workers, callbacks)
    ui.worker_menu.item_component:Update(GetMineContracts())
    -- LogEvent(LogEvents.menu_open("Dwarf Worker Menu"))
    if not event_mode.IsActive() then
        Tutorial05_Mine_StepC()
    else
        DCEI.SetFrameActive(ui.potion_lab.background, false)
        DCEI.SetFrameActive(ui.secret_lab.background, false)
    end
end

function AttemptUpgradeDwarves()
    local worker_data = data.dwarf_workers
    local cost = worker_data:GetCost(GetDwarfLevel() + 1)
    local gold = GetPlayerGold()
    local contract_needed = 1
    local success = false

    if GetDwarfLevel() == worker_data.max_level then
        ui.ShowFeedbackMaxLevel()
    elseif contract_needed > GetMineContracts() then
        ui.ShowFeedbackNotEnoughMineContracts()
    elseif cost <= gold then
        local spend_event = { type = "upgrade", target = "Mine" }
        SpendPlayerGold(cost, spend_event)

        AddPlayerContract_Mine(-1)
        UpgradeDwarves()
        PlayGoldSpendSound()

        ui.worker_menu:Update()
        ui.worker_menu.item_component:Update(GetMineContracts())

        success = true
    else
        local function OnStore()
            -- skip worker menu close camera transition to preserve store camera transition
            local skip_camera = true
            ui.worker_menu:Close(skip_camera)
        end

        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local offset = -142
        ui.ShowFeedbackNeedMoreGold(offset, ui.upper, OnStore, OnStoreClose)
    end
    if not event_mode.IsActive() then
        Tutorial05_Mine_StepD()
    end

    return success
end

function UpgradeDwarves()
    local worker_data = data.dwarf_workers
    AddDwarfLevel(1)

    if GetDwarfLevel() <= worker_data.max then
        SpawnWorkerDwarf_AtDepot()
    end
    if event_mode.IsActive() then
        bank.CommitEventDwarfLevel()
    else
        bank.CommitDwarfLevel()
    end
end

-- !SECTION

-- SECTION Main screen UI
function UpdateArcherUpgradeButton()
    local level = GetArcherLevel()
    local cost = GetArcherUpgradeCost()
    local atk = GetArcherDamageValue()
    ui.glue.upgrade_archer.Update(level, cost, atk)
    ui.glue.upgrade_archer.UpdatePotionCost()
end

function UpdateCastleUpgradeButton()
    local level = GetCastleLevel()
    local cost = GetCastleUpgradeCost()
    ui.glue.upgrade_castle.Update(level, cost)
    ui.glue.upgrade_castle.UpdatePotionCost()
end

function NewGearMenu()
    if wave.data.is_active or event_mode.IsActive() then
        -- prevent new gear menu from opening when wave started
        return
    end
    local function OnClose()
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
    end

    local callbacks = {
        purchase = AttemptPurchaseGear,
        upgrade = AttemptUpgradeGear,
        equip = ProcessEquipGear,
        gear_card_tutorial = Tutorial10_EquipGear_StepC,
        on_close = OnClose,
    }

    CloseActivePrompts()
    ui.DestroyBigHand()
    UnblockInput()
    if tutorial.gear then
        -- show gear tutorial message
        local offset = -500
        local text = tutorial.message.gear
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)

        ui.glue.gear_menu:HidePip()
    end

    if ui.gear_menu then
        ui.gear_menu:ShowGear()
    else
        ui.gear_menu = ui.NewGearMenu(data.gear, callbacks)
    end
    -- LogEvent(LogEvents.menu_open("Gear Menu"))

    for n = 1, data.unlock.gear_slot_max do
        local name = player.gear[n]
        local unit_data = data.gear[name]

        if unit_data then
            ui.UpdateGearEquipSlot(unit_data, n)
            ui.SetGearToEquipped(unit_data, true)
        end
    end

    gear_menu_click = gear_menu_click + 1
    bank.CommitGearMenuClick()
    Tutorial10_EquipGear_StepB()
end

-- SECTION Inventory
function NewInventoryMenu(hero_unit_data, last_card_callbacks)
    if wave.data.is_active then
        -- prevent new gear menu from opening when wave started
        return
    end

    if not ui.item_menu then
        ui.item_menu = ui.NewInventoryMenu()
        DCEI.SetOnClickCallback(ui.item_menu.sort_button, SortInventory)
        DCEI.SetOnClickCallback(ui.quick_open_chest.button, function()
            if ui.quick_open_chest.chest_name and ui.quick_open_chest.ready then
                OpenEquipmentChest(ui.quick_open_chest.chest_name)
            end
        end)
    end

    local function OnClose()
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
    end

    local function OpenGoldMine()
        TownTransition()
        NewWorkerMenu()
    end

    local function OpenTownHall()
        TownTransition()
        AttemptOpenTownHallMenu()
    end

    local function OpenGearMenu()
        NewGearMenu()
    end

    local function OpenRaidBoss()
        BlockInput()
        TownTransition()
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.bottom_panel.airship_mode = "raid_boss"
            AttemptOpenAirshipMenu()
            UnblockInput()
        end, 0.5, true, true)
    end

    local callbacks = {
        on_close = OnClose,
        hero_data = hero_unit_data,
        hero_card_callbacks = last_card_callbacks,
        open_gold_mine = OpenGoldMine,
        open_town_hall = OpenTownHall,
        open_gear_menu = OpenGearMenu,
        open_raid_boss = OpenRaidBoss,
        open_hell_coin_shop = OpenHellCoinShop,
        open_chest = OpenEquipmentChest,
        open_all_chest = OpenAllChests,
        process_equip_equipment = ProcessEquipEquipment,
        hero_list = NewEquipHeroList,
        upgrade = AttemptUpgradeEquipment,
        sell = SellEquipment,
        disenchant = DisenchantEquipment,
        enchant = EnchantEquipment,
        reset_enchant = ResetEquipmentEnchant,
        unlock_equipment = UnlockEquipment,
    }
    CloseActivePrompts()
    local colonies_completed = #levels.colony.complete
    local unlock_town_check = colonies_completed >= 1
    local _data = {
        unlock_town = unlock_town_check,
        mine_contract = GetMineContracts(),
        town_contract = GetTownContracts(),
        elixir = GetPlayerElixir(),
        item_data = player.inventory,
    }
    CheckInventoryPip()
    ui.item_menu:ShowItem()
    ui.item_menu.data = _data
    ui.item_menu.callbacks = callbacks
    UpdateItemMenu()
    if not hero_unit_data then
        DCEI.SetFrameActive(ui.quick_craft.frame, UnlockEquipment())
        ui.craft.locked = GetCastleLevel() < data.unlock.unlock_craft_level
        ui.quick_craft:UpdateStatus()
    end

    for i = 1, 3 do
        if ui.glue.inventory_menu.resource_pip[i] then
            ui.item_menu:ShowResourcePip(i)
            ui.glue.inventory_menu.resource_pip[i] = false
        end
    end
end

function InventoryHasChests()
    local total_chests = 0
    for _, name in ipairs(ui.item_data.chest_order) do
        local chest_count = ui.GetPlayerChest(name)
        total_chests = total_chests + chest_count
    end
    return total_chests > 0
end

function CheckInventoryPip()
    if not InventoryHasChests() or event_mode.IsActive() then
        ui.glue.inventory_menu:HidePip()
    else
        if not DCEI.IsFrameActive(ui.glue.inventory_menu.pip) then
            ui.glue.inventory_menu:ShowPipAnimated()
        end
    end
end
ui.CheckInventoryPip = CheckInventoryPip

function SortInventory()
    local function equipment_compare(a, b)
        if a.rarity ~= b.rarity then
            return a.rarity.index > b.rarity.index
        elseif a.type.index ~= b.type.index then
            return a.type.index < b.type.index
        else
            return a.level > b.level
        end
    end

    if not ui.item_menu.sorting then
        ui.item_menu.sorting = true
        DCEI.TriggerAddTimerEventElapsed(function()
            ui.item_menu.sorting = false
        end, 0.5, false, true)

        table.sort(player.inventory.items, equipment_compare)
        UpdateItemMenu()
    end
end

function NewEquipHeroList(item_data)
    local list_data = data.heroes.gallery

    local hero_can_equip = {}
    for _, hero in ipairs(list_data) do
        if hero.is_purchased then
            if item_data.type.type == "armor" then
                table.insert(hero_can_equip, hero)
            elseif hero.weapon_slot == item_data.type.name then
                table.insert(hero_can_equip, hero)
            end
        end
    end

    local function EquipFromInventory(hero, item_data)
        EquipEquipment(hero, item_data)
        ui.equip_hero_list.Close()
        UpdateItemMenu()
    end

    local callbacks = {
        equipment_list = true,
        equip_item = EquipFromInventory,
        processing_equipment = item_data,
    }

    -- hide unit list UI
    ui.CloseUnitCard()
    ui.item_menu.Show(false)

    local function OnClose()
        ui.item_menu.Show(true)
        ShowParentScreenMask(ui.upper)
        DCEI.DestroyFrame(ui.equip_hero_list.frame)
        ui.equip_hero_list = nil
    end

    local text = DCEI.GetTranslationText("ui/menu/heroes")
    ui.equip_hero_list = ui.NewList(hero_can_equip, OnClose, callbacks, ui.upper)
    ui.equip_hero_list.SetTitle(text)
    local sort = false
    ui.equip_hero_list.PopulateWithUnitData(sort)
end

function UnequipEquipment(hero, slot_type, skip_commit)
    if hero.equipment[slot_type] then
        DCEI.LogMessage("Hero: " .. hero.name .. " Unequip equipment " .. slot_type)
        local equiping_item = hero.equipment[slot_type]
        equiping_item.equiping_hero = nil
        hero.equipment[slot_type] = nil

        ProcessEquipmentAttributesToHero(hero, equiping_item, false)

        if not skip_commit then
            bank.CommitEquipments()
        end

        UpdateHeroList()
    end
end

function EquipEquipment(hero, item_data, skip_commit)
    -- unequip equipment if hero has this type of equipment current
    UnequipEquipment(hero, item_data.type.type, true)

    DCEI.LogMessage("Hero: " .. hero.name .. " Equip equipment: " .. item_data.type.name)
    local root_data = hero:GetParentData() or hero
    item_data.equiping_hero = root_data.name
    hero.equipment[item_data.type.type] = item_data

    ProcessEquipmentAttributesToHero(hero, item_data, true)

    if not skip_commit then
        bank.CommitEquipments()
    end

    UpdateHeroList()
end

function ProcessEquipEquipment(hero, item_data, close)
    if item_data.equiping_hero then
        UnequipEquipment(hero, item_data.type.type)
        if close then
            ui.unit_card:Close()
        end
    else
        EquipEquipment(hero, item_data)
    end
    UpdateItemMenu()
end

function ProcessEquipmentAttributesToHero(hero, item_data, is_addition)
    local sign = is_addition and 1 or -1

    for _, attribute in ipairs(item_data.attributes) do
        local hero_attribute = attribute.type.attribute
        local behavior_name = attribute.type.behavior
        local upgrade_behavior_name = attribute.type.upgrade_behavior
        local attribute_value = attribute.base + (item_data.level - 1) * attribute.type.values.growth
        if hero_attribute then
            if hero.equipment[hero_attribute] then
                hero.equipment[hero_attribute] = hero.equipment[hero_attribute] + sign * attribute_value
            elseif sign > 0 then
                hero.equipment[hero_attribute] = attribute_value
            end
        end
        local slot = hero.slot
        if slot and slot > 0 then
            if behavior_name then
                local unit = player.heroes[slot]
                local stack_count = math.floor(attribute.base / attribute.type.values.random_step)
                local count = sign * stack_count
                if count > 0 then
                    DCEI.ApplyBehaviorCountAsync(unit, behavior_name, count)
                else
                    DCEI.RemoveBehaviorCountAsync(unit, behavior_name, -1 * count)
                end
                if upgrade_behavior_name then
                    local upgrade_stack_count = sign * (item_data.level - 1)
                    if upgrade_stack_count > 0 then
                        DCEI.ApplyBehaviorCountAsync(unit, upgrade_behavior_name, upgrade_stack_count)
                    elseif upgrade_stack_count < 0 then
                        DCEI.RemoveBehaviorCountAsync(unit, upgrade_behavior_name, -1 * upgrade_stack_count)
                    end
                end
            end
        end
    end
    if item_data.epic_attributes then
        for _, attribute in ipairs(item_data.epic_attributes) do
            local behavior_name = attribute.type.behavior
            local gear_effect = attribute.type.gear_effect
            local slot = hero.slot
            if slot and slot > 0 then
                if behavior_name then
                    local unit = player.heroes[slot]
                    if is_addition then
                        DCEI.ApplyBehaviorAsync(unit, behavior_name)
                    else
                        DCEI.RemoveBehaviorAsync(unit, behavior_name)
                    end
                end
                if gear_effect then
                    local unit_data = data.gear[gear_effect]
                    ProcessEquipBonuses(unit_data, is_addition)
                end
            end
        end
    end
end

function ProcessEquipmentBehaviorToHero(hero_data, unit, is_addition)
    local equipments = { hero_data.equipment.weapon, hero_data.equipment.armor }
    for k, equipment in pairs(equipments) do
        for _, attribute in ipairs(equipment.attributes) do
            local behavior_name = attribute.type.behavior
            local upgrade_behavior_name = attribute.type.upgrade_behavior
            if behavior_name then
                local stack_count = math.floor(attribute.base / attribute.type.values.random_step)
                DCEI.ApplyBehaviorCountAsync(unit, behavior_name, stack_count)
                if upgrade_behavior_name then
                    local upgrade_stack_count = equipment.level - 1
                    if upgrade_stack_count > 0 then
                        DCEI.ApplyBehaviorCountAsync(unit, upgrade_behavior_name, upgrade_stack_count)
                    end
                end
            end
        end
        if equipment.epic_attributes then
            for _, attribute in ipairs(equipment.epic_attributes) do
                local behavior_name = attribute.type.behavior
                if behavior_name then
                    if is_addition then
                        DCEI.ApplyBehaviorAsync(unit, behavior_name)
                    else
                        DCEI.RemoveBehavior(unit, behavior_name)
                    end
                end
                local gear_effect = attribute.type.gear_effect
                if gear_effect then
                    local unit_data = data.gear[gear_effect]
                    ProcessEquipBonuses(unit_data, is_addition)
                end
            end
        end
    end
end

-- for legendary equipment aura attributes like power up
function ProcessEquipmentEquipBehaviorToHero(hero_data, unit, is_addition)
    local equipments = { hero_data.equipment.weapon, hero_data.equipment.armor }
    for k, equipment in pairs(equipments) do
        if equipment.epic_attributes then
            for _, attribute in ipairs(equipment.epic_attributes) do
                if attribute.type.equip_behavior then
                    local behavior_name = attribute.type.behavior
                    if behavior_name then
                        if is_addition then
                            DCEI.ApplyBehaviorAsync(unit, behavior_name)
                        else
                            DCEI.RemoveBehavior(unit, behavior_name)
                        end
                    end
                end
            end
        end
    end
end

function SellEquipment(item_data)
    if ui.unit_card then
        ui.unit_card:Close()
    end

    if item_data.equiping_hero then
        return
    end

    local sell_value = item_data.rarity.initial_value
    -- calculate upgrade value * 0.5
    local upgrace_cost = data.equipments.GetTotalUpgradeCost(item_data)
    sell_value = sell_value + upgrace_cost * 0.5

    AddPlayerGold(sell_value)
    bank.CommitGold()

    -- delete equipment
    for index, value in ipairs(player.inventory.items) do
        if value == item_data then
            table.remove(player.inventory.items, index)
            break
        end
    end

    bank.CommitEquipments()

    UpdateItemMenu()

    PlayGoldSpendSound()
    local icon = "icon_coin00"
    local icon_size = 56
    local text = "<color=yellow>+" .. ui.FormatWithCommas(sell_value) .. "</color>"
    local offset = {
        spacing = 4,
        x = 140,
        y1 = -280,
        y2 = -160,
    }
    ui.CreateFloatingMessage(icon, icon_size, text, offset, 1.25)
end

function MassSellEquipment(item_list, price)
    AddPlayerGold(price)
    bank.CommitGold()

    -- delete equipment
    for _, slot in pairs(item_list) do
        slot.item_type.item_data.delete = true
    end

    local new_item_list = {}
    for _, item in ipairs(player.inventory.items) do
        if not item.delete then
            table.insert(new_item_list, item)
        end
    end

    player.inventory.items = new_item_list

    bank.CommitEquipments()

    UpdateItemMenu()

    PlayGoldSpendSound()
    local icon = "icon_coin00"
    local icon_size = 56
    local text = "<color=yellow>+" .. ui.FormatWithCommas(price) .. "</color>"
    local offset = {
        spacing = 4,
        x = 140,
        y1 = -280,
        y2 = -160,
    }
    ui.CreateFloatingMessage(icon, icon_size, text, offset, 1.25)
end

function DisenchantEquipment(item_data)
    if ui.unit_card then
        ui.unit_card:Close()
    end

    if item_data.equiping_hero then
        return
    end

    local material_type = item_data.rarity.disenchant_material
    local material_value = item_data.rarity.disenchant_value

    AddPlayerMaterial(material_type, material_value)

    if item_data.rarity.index >= data.equipments.rarity.Epic.index then
        LogUserProperty(UserProps.epic_material_count)
    else
        LogUserProperty(UserProps.normal_material_count)
    end

    -- delete equipment
    for index, value in ipairs(player.inventory.items) do
        if value == item_data then
            table.remove(player.inventory.items, index)
            break
        end
    end

    bank.CommitEquipments()

    UpdateItemMenu()

    DCEI.PlaySound(DCEI.Sound("treants_spell_cast"))
    local icon = data.items[material_type].icon
    local text = "+" .. material_value
    local icon_size = 86
    local offset = {
        spacing = 2,
        x = 140,
        y1 = -280,
        y2 = -160,
    }
    ui.CreateFloatingMessage(icon, icon_size, text, offset, 1.25)
end

function MassDisenchantEquipment(item_list, value)
    local material_type = "normal_material"
    AddPlayerMaterial(material_type, value)

    LogUserProperty(UserProps.normal_material_count)

    -- delete equipment
    for _, slot in pairs(item_list) do
        slot.item_type.item_data.delete = true
    end

    local new_item_list = {}
    for _, item in ipairs(player.inventory.items) do
        if not item.delete then
            table.insert(new_item_list, item)
        end
    end

    player.inventory.items = new_item_list

    bank.CommitEquipments()

    UpdateItemMenu()

    DCEI.PlaySound(DCEI.Sound("treants_spell_cast"))
    local icon = data.items[material_type].icon
    local text = "+" .. value
    local icon_size = 86
    local offset = {
        spacing = 2,
        x = 140,
        y1 = -280,
        y2 = -160,
    }
    ui.CreateFloatingMessage(icon, icon_size, text, offset, 1.25)
end

function ConfirmEnchantEnequipment(item_data, new_attribute)
    local equiping_hero = item_data.equiping_hero
    local hero_data = nil
    if equiping_hero then
        local current_class = data.heroes[equiping_hero].current_class
        hero_data = data.heroes[current_class]
    end
    if hero_data then
        ProcessEquipmentAttributesToHero(hero_data, item_data, false)
    end

    if item_data.enchant then
        -- use number to represent enchanted attributes index, if it's epic attribute, use index + 10
        local epic = item_data.enchant > 10
        if epic then
            local index = item_data.enchant - 10
            if item_data.epic_attributes then
                item_data.epic_attributes[index] = new_attribute
            end
        else
            item_data.attributes[item_data.enchant] = new_attribute
        end
    end

    if hero_data then
        ProcessEquipmentAttributesToHero(hero_data, item_data, true)
    end

    bank.CommitEquipments()

    DCEI.PlaySound(DCEI.Sound("healing_v2"))

    if ui.unit_card and ui.unit_card.confirm_bar then
        ui.unit_card.confirm_bar:Close()
    end
end

function EnchantEquipment(item_data, attribute, enchent_index, cost_data)
    -- roll 3 different attributes
    local new_attributes = {}
    local is_epic_attribute = (attribute.type.rarity == "Epic" or attribute.type.rarity == "Legendary")
    enchent_index = (is_epic_attribute and (enchent_index - #item_data.attributes)) or enchent_index
    if is_epic_attribute then
        for i = 1, 3 do
            local reroll = true
            while reroll do
                reroll = false
                local random_attribute = math.random(1, #data.equipments.epic_attribute)
                local new_attribute_type = data.equipments.epic_attribute[random_attribute]
                -- check if there is duplicate attribute
                for j = 1, i - 1 do
                    if new_attributes[j].type == new_attribute_type then
                        reroll = true
                        break
                    end
                end
                -- check minimum rarity for legendary attributes
                local minimum_rarity = item_data.rarity.index
                if minimum_rarity < data.equipments.rarity[new_attribute_type.rarity].index then
                    reroll = true
                end
                if not reroll then
                    if
                        not new_attribute_type.limit
                        or (new_attribute_type.limit and new_attribute_type.limit[item_data.type.name])
                    then
                        local new_attribute = {}
                        new_attribute.type = new_attribute_type
                        table.insert(new_attributes, new_attribute)
                    else
                        reroll = true
                    end
                end
            end
        end
    else
        -- create enchanting pool and calculate enchanting weight
        local enchant_pool = {}
        local total_weight = 0
        for _, v in ipairs(data.equipments.attribute) do
            if not v.limit or (v.limit and v.limit[item_data.type.name]) then
                table.insert(enchant_pool, v)
                total_weight = total_weight + v.enchant_weight
            end
        end

        for i = 1, 3 do
            local reroll = true
            while reroll do
                reroll = false
                local new_attribute_type
                local random_weight = math.random(total_weight)
                local weight = 0
                for _, item in ipairs(enchant_pool) do
                    weight = weight + item.enchant_weight
                    if random_weight <= weight then
                        new_attribute_type = item
                        break
                    end
                end
                -- don't check minimum rarity because only rare and epic can enchant

                -- check if there is duplicate attribute
                for j = 1, i - 1 do
                    if new_attributes[j].type == new_attribute_type then
                        reroll = true
                        break
                    end
                end
                if not reroll then
                    local new_attribute = {}
                    new_attribute.type = new_attribute_type
                    -- roll base
                    local min = item_data.rarity.attribute_range[1]
                    local max = item_data.rarity.attribute_range[2]

                    local max_step = (new_attribute_type.values[max] - new_attribute_type.values[min])
                        / new_attribute_type.values.random_step
                    local random_fraction = math.random(0, math.floor(max_step))
                    new_attribute.base = new_attribute_type.values[min]
                        + random_fraction * new_attribute_type.values.random_step

                    table.insert(new_attributes, new_attribute)
                end
            end
        end
    end

    -- use number to represent enchanted attributes index, if it's epic attribute, use index + 10
    if not item_data.enchant then
        item_data.enchant = (is_epic_attribute and (enchent_index + 10)) or enchent_index
    end

    item_data.enchant_count = item_data.enchant_count or 0
    item_data.enchant_count = item_data.enchant_count + 1

    local spend_event = { type = "enchant", target = item_data.rarity.name .. " " .. item_data.type.type }
    SpendPlayerGold(cost_data.gold)

    local mat_type = "normal_material"
    AddPlayerMaterial(mat_type, -cost_data.normal_material)
    LogUserProperty(UserProps.normal_material_count)
    if cost_data.epic_material then
        local mat_type = "epic_material"
        AddPlayerMaterial(mat_type, -cost_data.epic_material)
        LogUserProperty(UserProps.epic_material_count)
        IncrementEnchantEpicCount()
    else
        IncrementEnchantNonEpicCount()
    end

    bank.CommitEquipments()

    PlayGoldSpendSound()

    UpdateItemMenu()

    return ui.NewEnchantResultList(item_data, attribute, new_attributes, ConfirmEnchantEnequipment)
end

function ResetEquipmentEnchant(item_data)
    item_data.enchant = false
    AddPlayerPhilosopherStone(-1)
    bank.CommitEquipments()
    UpdateItemMenu()
end

function RollEquipmentAttributes(new_equipment)
    new_equipment.attributes = {}
    local attribute_number = 2
    if new_equipment.rarity.name == "Common" then
        attribute_number = 1
    elseif new_equipment.rarity.name == "Epic" then
        if new_equipment.type.type == "armor" then
            attribute_number = 3
        else
            attribute_number = 2
        end
    elseif new_equipment.rarity.name == "Legendary" then
        if new_equipment.type.type == "armor" then
            attribute_number = 4
        else
            attribute_number = 3
        end
    end

    for i = 1, attribute_number do
        local reroll = true
        local new_attribute = {}
        while reroll do
            reroll = false
            local random_attribute = math.random(1, #data.equipments.attribute)
            local attribute = data.equipments.attribute[random_attribute]
            -- check minimum rarity
            local minimum_rarity = attribute.rarity
            if new_equipment.rarity.index < data.equipments.rarity[minimum_rarity].index then
                reroll = true
            end
            if i > 1 and not reroll then
                -- check if there is duplicate attribute
                for j = 1, i - 1 do
                    if new_equipment.attributes[j].type == attribute then
                        reroll = true
                        break
                    end
                end
            end
            if not reroll then
                if not attribute.limit or (attribute.limit and attribute.limit[new_equipment.type.name]) then
                    new_attribute.type = attribute
                    -- roll base
                    local min = new_equipment.rarity.attribute_range[1]
                    local max = new_equipment.rarity.attribute_range[2]

                    local max_step = (attribute.values[max] - attribute.values[min]) / attribute.values.random_step
                    local random_fraction = math.random(0, math.floor(max_step))
                    new_attribute.base = attribute.values[min] + random_fraction * attribute.values.random_step

                    table.insert(new_equipment.attributes, new_attribute)
                else
                    reroll = true
                end
            end
        end
    end
    for _, attri in ipairs(new_equipment.attributes) do
        attri.type = attri.type.index
    end
    if
        new_equipment.type.type == "weapon"
        and (new_equipment.rarity.name == "Epic" or new_equipment.rarity.name == "Legendary")
    then
        local reroll = true
        new_equipment.epic_attributes = {}
        local new_epic_attribute = {}
        while reroll do
            reroll = false
            local random_attribute = math.random(1, #data.equipments.epic_attribute)
            local attribute = data.equipments.epic_attribute[random_attribute]
            -- check minimum rarity
            local minimum_rarity = attribute.rarity
            if new_equipment.rarity.index < data.equipments.rarity[minimum_rarity].index then
                reroll = true
            end
            if not reroll then
                if not attribute.limit or (attribute.limit and attribute.limit[new_equipment.type.name]) then
                    new_epic_attribute.type = attribute.index
                    table.insert(new_equipment.epic_attributes, new_epic_attribute)
                else
                    reroll = true
                end
            end
        end
    end
end

function GetEquipmentFromChest(rarity_type)
    local new_equipment = {}
    local first_rare = (player.system.open_first_common_chest == system.unlocked)
    if first_rare and rarity_type == "common_chest" then
        -- first common chest from wave 100 boss, get fixed attribute sword
        player.system.open_first_common_chest = system.active
        new_equipment.rarity = data.equipments.rarity.Uncommon
        new_equipment.type = data.equipments.type.Sword
        new_equipment.attributes = {}
        new_equipment.attributes[1] = {
            type = data.equipments.attribute["Attribute Damage Plus"].index,
            base = 45,
        }
        new_equipment.attributes[2] = {
            type = data.equipments.attribute["Attribute Critical Chance"].index,
            base = 0.02,
        }

        bank.CommitSystemState()

        LogEvent(LogEvents.open_first_common_chest)
    else
        -- normal chest
        local random_rarity = math.random()
        local weight = 0
        for rarity, rate in pairs(data.items[rarity_type].rarity_rate) do
            weight = weight + rate
            if random_rarity < weight then
                new_equipment.rarity = data.equipments.rarity[rarity]
                break
            end
        end

        local iap_chests = {
            "rare_chest_l",
            "rare_chest",
            "legendary_chest_l",
            "legendary_chest",
        }
        for _, chest in ipairs(iap_chests) do
            if rarity_type == chest then
                if new_equipment.rarity.name ~= "Epic" and player.guarantee_epic[chest] > 0 then
                    -- if repeatly get none epic chest, then give a guaranteed epic chest when none epic chest number reaches epic interval
                    player.non_epic_chest[chest] = player.non_epic_chest[chest] + 1
                    if player.non_epic_chest[chest] >= player.epic_interval[chest] then
                        new_equipment.rarity = data.equipments.rarity["Epic"]
                    end
                    bank.CommitChestNonEpic()
                end
                -- guarantee legendary priority so it get overwrite the previous
                if new_equipment.rarity.name ~= "Legendary" and player.guarantee_legendary[chest] > 0 then
                    -- if repeatly get none legendary chest, then give a guaranteed legendary chest when none legendary chest number reaches legendary interval
                    player.non_legendary_chest[chest] = player.non_legendary_chest[chest] + 1
                    if player.non_legendary_chest[chest] >= player.legendary_interval[chest] then
                        new_equipment.rarity = data.equipments.rarity["Legendary"]
                    end
                    bank.CommitChestNonEpic()
                end

                -- clear none legendary chest if rarity is legendary
                if new_equipment.rarity.name == "Legendary" then
                    player.non_legendary_chest[chest] = 0
                    player.guarantee_legendary[chest] = math.max(player.guarantee_legendary[chest] - 1, 0)
                    player.legendary_interval[chest] = player.guarantee_legendary[chest] == 0 and math.huge
                        or math.floor(GetChestCount(chest) / player.guarantee_legendary[chest])

                    bank.CommitGuaranteeEpic()
                    bank.CommitChestNonEpic()
                elseif new_equipment.rarity.name == "Epic" then
                    -- clear none epic chest if rarity is epic
                    player.non_epic_chest[chest] = 0
                    player.guarantee_epic[chest] = math.max(player.guarantee_epic[chest] - 1, 0)
                    player.epic_interval[chest] = player.guarantee_epic[chest] == 0 and math.huge
                        or math.floor(GetChestCount(chest) / player.guarantee_epic[chest])

                    bank.CommitGuaranteeEpic()
                    bank.CommitChestNonEpic()
                end

                break
            end
        end

        local total_type = #data.equipments.type
        local random_type = math.random(1, total_type)
        new_equipment.type = data.equipments.type[random_type]
        RollEquipmentAttributes(new_equipment)
    end

    new_equipment.type = new_equipment.type.index
    new_equipment.rarity = new_equipment.rarity.index
    new_equipment.level = 1
    table.insert(player.inventory.items, data.GetEquipment(new_equipment))

    bank.CommitEquipments()
    return new_equipment
end

function OpenChestFunction(rarity_type)
    -- check inventory is full
    if player.inventory.full then
        ui.ShowFeedbackInventoryFull()
        return false
    end

    if GetChestCount(rarity_type) > 0 then
        AddPlayerChest(rarity_type, -1)

        local new_equipment = GetEquipmentFromChest(rarity_type)

        if ui.item_menu then
            ui.item_menu:FilterSelectAll()
        end
        UpdateItemMenu(true)
        return true
    end

    return false
end

function OpenEquipmentChest(rarity_type)
    if GetChestCount(rarity_type) == 0 then
        ui.CloseUnitCard()
        ui.item_menu:Close()

        local from = "inventory chest"
        SetStoreEntrySource(from)
        OpenChestStore()
        return false
    end

    local success = OpenChestFunction(rarity_type)
    if success then
        DCEI.PlaySound(DCEI.Sound("chest_open"))

        if ui.unit_card then
            ui.unit_card:Close()
        end
        ui.item_menu.OpenNewEquipment()
        --DCEI.AnimateFrameAlpha(ui.unit_card.frame, 0, 1, 0.5, "InSine")
        local start_scale = { x = 0.5, y = 0.5, z = 0 }
        local end_scale = { x = 1, y = 1, z = 0 }
        DCEI.AnimateFrameScale(ui.unit_card.frame, start_scale, end_scale, 0.3, "OutBack")
        DCEI.AnimateFrameScale(ui.unit_card.additional_buttons, start_scale, end_scale, 0.3, "OutBack")

        -- show quick open button
        local rest_count = GetChestCount(rarity_type)
        if rest_count > 0 then
            ui.quick_open_chest:UpdateChest(rarity_type, data.items[rarity_type].icon, rest_count)
        else
            ui.quick_open_chest:Hide()
        end

        return true
    end

    return false
end

function OpenAllChestsProcess(rarity_type)
    local success = OpenChestFunction(rarity_type)
    if success then
        DCEI.TriggerAddTimerEventElapsed(function()
            OpenAllChestsProcess(rarity_type)
        end, 0, true, true)
    else
        UnblockInput()
    end
end

function OpenAllChests(rarity_type)
    local success = OpenChestFunction(rarity_type)
    if success then
        BlockInput()
        DCEI.PlaySound(DCEI.Sound("chest_open"))
        if ui.unit_card then
            ui.unit_card:Close()
        end
        DCEI.TriggerAddTimerEventElapsed(function()
            OpenAllChestsProcess(rarity_type)
        end, 0, true, true)
    end
end

function CraftEquipment(rarity, type, cost_data)
    local new_equipment = {}
    new_equipment.rarity = data.equipments.rarity[rarity]
    new_equipment.type = data.equipments.type[type]
    RollEquipmentAttributes(new_equipment)

    new_equipment.type = new_equipment.type.index
    new_equipment.rarity = new_equipment.rarity.index
    new_equipment.level = 1
    table.insert(player.inventory.items, data.GetEquipment(new_equipment))

    local mat_type = "normal_material"
    AddPlayerMaterial(mat_type, -cost_data.normal_material)
    LogUserProperty(UserProps.normal_material_count)
    if cost_data.epic_material then
        local mat_type = "epic_material"
        AddPlayerMaterial(mat_type, -cost_data.epic_material)
        LogUserProperty(UserProps.epic_material_count)
    end
    IncrementCraftCount(rarity)

    bank.CommitEquipments()
    UpdateItemMenu()
    if ui.unit_card then
        ui.unit_card:Close()
    end
    ui.item_menu.OpenNewEquipment()
    local start_scale = { x = 0.5, y = 0.5, z = 0 }
    local end_scale = { x = 1, y = 1, z = 0 }
    DCEI.AnimateFrameScale(ui.unit_card.frame, start_scale, end_scale, 0.3, "OutBack")
    DCEI.AnimateFrameScale(ui.unit_card.additional_buttons, start_scale, end_scale, 0.3, "OutBack")
end

ui.CraftEquipment = CraftEquipment

function AttemptUpgradeEquipment(item_data)
    local cost_data = data.equipments.GetUpgradeCost(item_data)
    local gold = GetPlayerGold()
    local elixir = GetPlayerElixir()
    local success = true

    if gold >= cost_data.gold and elixir >= cost_data.elixir then
        local spend_event = { type = "upgrade", target = "Equipment " .. tostring(item_data.rarity.name) }
        SpendPlayerGold(cost_data.gold, spend_event)
        SpendPlayerElixir(cost_data.elixir)

        UpgradeEquipment(item_data)
        PlayGoldSpendSound()

        -- update card display
        if ui.unit_card and ui.unit_card.Update then
            ui.unit_card:Update()
        end

        -- update inventory menu
        if item_data.slot then
            item_data.slot:Update()
        end
    elseif gold < cost_data.gold then
        -- need more gold prompt
        local function OnStore()
            ui.CloseUnitCard()
            ui.item_menu:Close()
            AttemptCloseHeroList()
            ui.CloseUnitCard()
            AttemptCloseHeroList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)

        success = false
    elseif elixir < cost_data.elixir then
        -- need more elixir
        ui.ShowFeedbackNotEnoughElixir()

        success = false
    end

    return success
end

function UpgradeEquipment(item_data)
    local equiping_hero = item_data.equiping_hero
    local hero_data = nil
    if equiping_hero then
        local current_class = data.heroes[equiping_hero].current_class
        hero_data = data.heroes[current_class]
    end
    if hero_data then
        ProcessEquipmentAttributesToHero(hero_data, item_data, false)
    end

    data.equipments.LevelUp(item_data)

    if hero_data then
        ProcessEquipmentAttributesToHero(hero_data, item_data, true)
    end

    bank.CommitEquipments()

    IncrementEquipmentUpgradeCount()
end

-- !SECTION

-- !SECTION
function ShowTalentMenu()
    if wave.data.is_active then
        -- prevent talent menu from opening when wave started
        return
    end

    CloseActivePrompts()

    if not ui.talent_menu then
        local function OnClose()
            ui.screen_mask.Hide()

            if ui.prompt.tutorial then
                ui.prompt.tutorial.Close()
            end
        end

        local callbacks = {
            upgrade = AttemptUpgradeTalent,
            on_close = OnClose,
            input = DevModeLogInput,
            TalentSpent = data.talents.Spent,
            GetTalentPoints = GetTalentPoints,
            GetParagonPoints = GetParagonPoints,
            GetParagonLevel = GetParagonLevel,
            reset_paragon = ResetParagonPoints,
            reset_normal = ResetNormalTalentPoints,
            reset_check = AttemptResetTalent,
        }

        ui.talent_menu = ui.CreateTalentMenu(data.talents, callbacks)
    end

    if tutorial.talent then
        if tutorial.state == 4.1 then
            tutorial.state = 0
            UnblockInput()
            ui.DestroyBigHand()
        end

        -- show talent tutorial message
        local offset = -470
        local text = tutorial.message.talent
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
        ui.glue.talent_menu:StopAnimating()
    else
        -- hide talent pip if not tutorial
        ui.glue.talent_menu:HidePip()
    end

    ui.talent_menu.UpdatePoints(GetTalentPoints(), GetParagonPoints())

    ui.talent_menu:Show()

    CheckUnlockTalentsTier2()
    CheckUnlockParagonTalents()

    --CheckDevModeSendNotification()
end

function CheckDevModeSendNotification()
    -- for dev mode notification test sequence
    if DevMode() then
        DEV.test_send_notification_tracking = DEV.test_send_notification_tracking + 1

        if DEV.test_send_notification_tracking >= 4 then
            DEV.test_send_notification_tracking = 0
            -- send a dev test notification
            DCEI.ScheduleNotification("notification title", "notification body", 10)
            -- push back create date so that log event will trigger
            if player.tests.create_time then
                player.tests.create_time = player.tests.create_time - 86400
                DCEI.ShowFeedbackMessage("Push back create date!")
                bank.CommitTests()
            end
        end
    end
end

-- SECTION PLAYER DECK
function ShowPlayerDeck()
    -- prevents player deck from opening when wave started
    if wave.data.is_active then
        return
    end
    if ui.glue.other_button_panel.opening or ui.glue.other_button_panel.closing then
        return
    end
    ui.glue.other_button_panel:Close()
    ui.glue.player_deck_menu:HidePip()

    if not ui.player_deck then
        ui.player_deck = ui.NewPlayerDeckFrame()

        ui.player_deck:InitiailizePlayerDeckWithReference(player.player_deck)

        -- each player deck slot buttons
        for id, deck in ipairs(ui.player_deck.deck_slots) do
            DCEI.SetOnClickCallback(deck.button, function()
                PlayerDeckSlotSelected(id)
            end)
        end

        DCEI.SetOnClickCallback(ui.player_deck.rename_layout.confirm_button, function()
            AttemptPlayerDeckRename(player.player_deck.current_id)
        end)
        -- deck reset confirmation
        DCEI.SetOnClickCallback(ui.player_deck.reset_layout.confirm_button, ResetPlayerDeckSlots)
    end

    CloseActivePrompts()
    ui.screen_mask.upper:Show()
    ui.player_deck:Show()
end

function PlayerDeckSlotSelected(id)
    -- exit if new slot is alread selected
    if id == player.player_deck.current_id then
        return
    end

    -- validation check on player decks
    if not player.player_deck.decks then
        return
    end

    -- store previous deck
    player.player_deck.decks[player.player_deck.current_id] = EncodeCurrentDeckToString()

    -- set current deck id to new id
    player.player_deck.current_id = id
    ui.player_deck:SelectCurrentDeckSlot(id)

    UnequipAllPlayerDeckItemSlots()
    EquipAllItemsFromEncodedDeck(id)

    for index, equipment in ipairs(player.inventory.items) do
        if equipment["deck" .. id] then
            -- unequip if this equipment is on other hero
            if equipment.equiping_hero then
                local hero_name = equipment.equiping_hero
                local class = data.heroes[hero_name].current_class
                local unit_data = data.heroes[class] or data.heroes[hero_name]
                UnequipEquipment(unit_data, equipment.type.type, true)
            end
            local hero_name = equipment["deck" .. id]
            local class = data.heroes[hero_name].current_class
            local unit_data = data.heroes[class] or data.heroes[hero_name]
            EquipEquipment(unit_data, equipment, true)
        end
    end

    -- commit deck last
    bank.CommitPlayerDeckData()
    bank.CommitEquipments()
end

function EquipAllItemsFromEncodedDeck(deck_index)
    local deck = player.player_deck.decks[deck_index]

    if not deck or deck == "empty" or deck == "" then
        return
    end

    local decoded_deck = json.decode(deck)
    -- iterate through decoded HERO deck and equipt to appropriate slot and unit_data
    for i, hero_name in ipairs(decoded_deck.heroes) do
        if hero_name ~= "" then
            local class = data.heroes[hero_name].current_class
            local unit_data = data.heroes[class] or data.heroes[hero_name]
            EquipHero(unit_data, i)
        end
    end
    -- iterate through decoded TOWER deck and assign
    for j, tower_name in ipairs(decoded_deck.towers) do
        if tower_name ~= "" then
            local class = data.towers[tower_name].current_class
            local unit_data = data.towers[class] or data.towers[tower_name]

            EquipTower(unit_data, j)
        end
    end
    -- iterate through decoded DEFENSE deck and assign
    for k, defense_name in ipairs(decoded_deck.defenses) do
        if defense_name ~= "" then
            local unit_data = data.defenses[defense_name]
            EquipDefense(unit_data, k)
        end
    end
    -- iterate through decoded TOWN deck and assign
    for l, town_name in ipairs(decoded_deck.town) do
        if town_name ~= "" then
            local unit_data = data.town[town_name]
            EquipTown(unit_data, l)
        end
    end
    -- iterate through decoded GEAR deck and assign
    for m, gear_name in ipairs(decoded_deck.gear) do
        if gear_name ~= "" then
            local unit_data = data.gear[gear_name]
            EquipGear(unit_data, m)
        end
    end
end

function UnequipAllPlayerDeckItemSlots()
    -- player.heroes
    local slot_count = data.unlock.hero_slot_max

    for n = 1, slot_count do
        UnequipHeroSlot(n)
    end
    -- player.towers
    slot_count = data.unlock.tower_slot_max
    for n = 1, slot_count do
        UnequipTowerSlot(n)
    end
    -- player.defenses
    slot_count = data.unlock.defense_slot_max
    for n = 1, slot_count do
        UnequipDefenseSlot(n)
    end
    -- player.town
    slot_count = data.unlock.town_slot_max
    for n = 1, slot_count do
        UnequipTownSlot(n)
    end
    -- player.gear
    slot_count = data.unlock.gear_slot_max
    for n = 1, slot_count do
        UnequipGearSlot(n)
    end
end

function ResetPlayerDeckSlots()
    local deck_index = player.player_deck.current_id
    local text = DCEI.GetTranslationText("ui/deck_number", { x = deck_index })

    player.player_deck.decks[deck_index] = "empty"

    -- get resetted slot name while resetting player deck slot ui
    ui.player_deck.deck_slots[deck_index]:UpdateSlotName(text)
    player.player_deck.names[deck_index] = text
    ui.player_deck.reset_layout:Close()

    UnequipAllPlayerDeckItemSlots()
    bank.CommitPlayerDeckData()

    for index, equipment in ipairs(player.inventory.items) do
        if equipment["deck" .. deck_index] then
            equipment["deck" .. deck_index] = nil
        end
    end
    bank.CommitEquipments()
end

function AttemptPlayerDeckRename(deck_index)
    local new_name = DCEI.GetInputFrameText(ui.player_deck.rename_layout.input)
    local character_limit = 15

    if new_name:len() > character_limit then
        -- warn if name too long
        ui.ShowFeedbackNameTooLong()
    elseif new_name and new_name ~= "" then
        -- set name if no issues
        SetPlayerDeckRename(deck_index)
    else
        -- warn if invalid name
        ui.ShowFeedbackNameInvalid()
    end
end

function SetPlayerDeckRename(deck_index)
    -- store renamed deck name to player data
    player.player_deck.names[deck_index] = ui.player_deck:GetPlayerDeckRenameInput()

    -- update it in ui
    ui.player_deck:RenamePlayerDeckSlotName(deck_index)

    -- commit new deck name to bank
    bank.CommitPlayerDeckData()
end

function EncodeCurrentDeckToString()
    local castle_level = GetCastleLevel()
    local unlock = data.GetUnlockedSlotCount(castle_level)

    local deck = {
        heroes = {},
        towers = {},
        defenses = {},
        town = {},
        gear = {},
    }

    local deck_id = player.player_deck.current_id

    -- process equipment
    for index, equipment in ipairs(player.inventory.items) do
        if equipment.equiping_hero then
            equipment["deck" .. deck_id] = equipment.equiping_hero
        else
            equipment["deck" .. deck_id] = nil
        end
    end
    -- transcribe heroes to deck
    for n = 1, unlock.hero_slots do
        local unit = player.heroes[n]
        local name = ""

        -- save base class name to deck since deck doesn't care about current promotion
        if unit and DCEI.UnitExists(unit) then
            local key = DCEI.GetUnitType(unit)
            local unit_data = data.heroes[key]
            local root_data = unit_data:GetParentData() or unit_data

            name = root_data.name
        end

        deck.heroes[n] = name
    end

    -- transcribe towers to deck
    for n = 1, unlock.tower_slots do
        local unit = player.towers[n]
        local name = ""

        -- save base class name to deck since deck doesn't care about current promotion
        if unit and DCEI.UnitExists(unit) then
            local key = DCEI.GetUnitType(unit)
            local unit_data = data.towers[key]
            local root_data = unit_data:GetParentData() or unit_data

            name = root_data.name
        end

        deck.towers[n] = name
    end

    -- transcribe defenses to deck
    for n = 1, unlock.defense_slots do
        local unit = player.defenses[n]
        local name = ""

        -- can simply save name for defenses as they don't have promotions
        if unit and DCEI.UnitExists(unit) then
            name = DCEI.GetUnitType(unit)
        end

        deck.defenses[n] = name
    end

    -- transcribe town to deck
    for n = 1, unlock.town_slots do
        local unit = player.town[n]
        local name = ""

        -- can simply save name for town as they don't have promotions
        if unit and DCEI.UnitExists(unit) then
            name = DCEI.GetUnitType(unit)
            -- get base name for rank 2 and 3
            name = data.GetTownBaseName(name)
        end

        deck.town[n] = name
    end

    -- transcribe gear to deck
    for n = 1, unlock.gear_slots do
        deck.gear[n] = player.gear[n] or ""
    end

    -- encode deck table to string
    return json.encode(deck)
end

-- !SECTION
-- DEV MODE
function DevModeLogInput(id)
    ui.talent_menu.sequence = ui.talent_menu.sequence .. id
end

function DevModeCheckInput()
    local dmode_sequence = "221010575734"
    local input_sequence = ui.talent_menu.sequence

    if input_sequence == dmode_sequence then
        player.dev_mode = true
        bank.CommitDevMode()

        DCEI.ShowFeedbackMessage("success")

        local dev_time = 20 * 60
        AddSpeedx3Time(dev_time)
        AddGemAutoBattleTime(dev_time)
    end
end

function UpgradeTalent(talent_data, cost)
    if talent_data.paragon then
        AddParagonPoints(-cost)
    else
        AddTalentPoints(-cost)
    end
    DCEI.SetFrameActive(ui.talent_menu.reset_paragon.frame, true)
    talent_data.level = talent_data.level + 1

    ProcessTalentAttributes(talent_data)
    ProcessTalentBehaviors(talent_data)

    ui.talent_menu.UpdatePoints(GetTalentPoints(), GetParagonPoints())
    ui.glue.talent_menu:Update(GetTalentPoints(), GetParagonPoints())

    if tutorial.talent then
        -- end talent tutorial on talent spend
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end

        tutorial.talent = false
        ui.glue.talent_menu:HidePip()
    end
    if event_mode.IsActive() then
        bank.CommitEventTalent(talent_data)
    else
        bank.CommitTalent(talent_data)
    end
end

function AttemptUpgradeTalent(talent_data)
    local max_level = talent_data.max_level
    local level = talent_data.level
    local cost = 1
    if talent_data.cost_curve then
        for _, v in ipairs(talent_data.cost_curve) do
            if level >= v.start_level then
                cost = v.cost
                break
            end
        end
    end
    local talent_points = talent_data.paragon and GetParagonPoints() or GetTalentPoints()
    local success = false

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
    elseif cost <= talent_points then
        UpgradeTalent(talent_data, cost)

        --check dev for dev mode sequence
        DevModeCheckInput()
        success = true
    else
        ui.ShowFeedbackNotEnoughTalent(talent_data, cost)
    end

    return success
end

function UpdateExperienceBar()
    local max_level = data.GetPlayerLevelMax()
    local level = GetPlayerLevel()
    local paragon_level = GetParagonLevel()
    local xp_bar = ui.top_panel.xp_bar
    xp_bar.level = xp_bar.level or 0

    if xp_bar.level ~= (level + paragon_level) then
        xp_bar.level = level + paragon_level
        if level >= max_level then
            data.paragon_level_xp = data.paragon_level_xp or data.UpdateParagonLevelXP(paragon_level)
            xp_bar.min = 0
            xp_bar.max = data.paragon_level_xp
        else
            xp_bar.min = data.GetPlayerLevelXP(level)
            xp_bar.max = data.GetPlayerLevelXP(level + 1) - xp_bar.min
        end

        xp_bar:SetLevel(level, paragon_level, level == max_level)
    end

    -- if level == max_level and not ui.top_panel.xp_bar.complete then
    --     -- display MAX if player is max level
    --     xp_bar:SetMaxed(level)
    --     return
    -- end
    local current_xp = 0
    local current_xp = level >= max_level and player.paragon_xp or player.player_xp
    local value = current_xp - xp_bar.min
    value = math.max(0, value / xp_bar.max)
    xp_bar:SetProgress(value)
end

function UpdateGemDisplay()
    local value = math.floor(DCEI.GetProtectedValue(data.safe.gems))
    local overflow = math.floor(DCEI.GetProtectedValue(data.safe.gems_overflow))

    ui.top_panel.gem:Update(value, overflow)
end

function UpdateGoldDisplay()
    if not ui.top_panel.gold.update_this_frame then
        ui.top_panel.gold.update_this_frame = true

        local simple = true
        DCEI.TriggerAddTimerEventElapsed(function()
            local value = GetPlayerGold()
            ui.top_panel.gold:Update(value)
            ui.top_panel.gold.update_this_frame = false
        end, 0.0625, false, simple)
    end
end

-- !SECTION

-- SECTIOn Loot drops
function AssignRareMonsterDrops(wave_id)
    wave.data.rare_monster_drop = nil
    if wave.data.ModeStandard() then
        for _, drop in ipairs(levels.rare_monster_drops) do
            if wave_id >= drop.rating then
                wave.data.rare_monster_drop = drop
            end
        end
    end
    -- if wave.data.rare_monster_drop then
    --     DCEI.LogMessage("AssignRareMonsterDrops "..wave.data.rare_monster_drop.drops[1])
    -- end
end

function ProcessRareMonsterDrops(unit)
    if wave.data.rare_monster_drop and DCEI.GetUnitBehaviorStackCount(unit, "Unit Tag Rare Monster") > 0 then
        local rd = math.random()
        if rd < wave.data.rare_monster_drop.drop_rate then
            -- for now only has one drop: equipment chest
            local drop_item = wave.data.rare_monster_drop.drops[1]
            -- don't commit to bank during wave for performanc reasons (commit after wave end)
            AddPlayerChest(drop_item, 1, true)

            -- show loot drop feedback
            local icon = data.items[drop_item].icon
            ui.CreateFloatingLootDisplay(unit, icon, 1)
        end
    end
end

function ProcessLootDrops(unit)
    local token_behavior = DCEI.Behavior("Loot DropToken")
    local token_stacks = DCEI.GetUnitBehaviorStackCount(unit, token_behavior)

    local mine_behavior = DCEI.Behavior("Loot DropMineContract")
    local mine_stacks = DCEI.GetUnitBehaviorStackCount(unit, mine_behavior)

    local elixir_behavior = DCEI.Behavior("Loot DropElixir")
    local elixir_stacks = DCEI.GetUnitBehaviorStackCount(unit, elixir_behavior)

    -- don't commit to bank during wave for performanc reasons (bank will be saved when wave ends anyway)
    local skip_commit = true

    if token_stacks > 0 then
        local allow_overflow = true
        AddPlayerOverflowGems(token_stacks, skip_commit)

        -- show loot drop feedback
        local icon = "icon_coin_dungeon00"
        ui.CreateFloatingLootDisplay(unit, icon, token_stacks)
    end

    if mine_stacks > 0 then
        AddPlayerContract_Mine(mine_stacks, skip_commit)

        -- show loot drop feedback
        local icon = "icon_catalyst_main_item2_orange_03"
        ui.CreateFloatingLootDisplay(unit, icon, mine_stacks)
    end

    if elixir_stacks > 0 then
        AddPlayerElixir(elixir_stacks, skip_commit)

        -- show loot drop feedback
        local icon = "icon_ingame_blackpotion"
        ui.CreateFloatingLootDisplay(unit, icon, elixir_stacks)
    end

    if wave.data.egg_hunt_event then
        local egg_hunt_behavior = DCEI.Behavior("Loot Egg Hunt")
        local egg_hunt_stacks = DCEI.GetUnitBehaviorStackCount(unit, egg_hunt_behavior)
        if egg_hunt_stacks > 0 then
            -- show loot drop feedback
            local icon = "wildcastle_icon_easter_egg"
            ui.CreateFloatingLootDisplay(unit, icon, egg_hunt_stacks)

            player.egg_hunt_event.raw = player.egg_hunt_event.raw + egg_hunt_stacks
        end
    end
end

function ProcessSeasonEventDrops(unit)
    local drop_behavior = "Loot DropSeasonEventCoin"
    local drop_stacks = DCEI.GetUnitBehaviorStackCount(unit, drop_behavior)
    local skip_commit = true

    if drop_stacks > 0 then
        if CheckSeasonEventExist("winter_event2022") then
            AddPlayerWinterCoin2022(drop_stacks, skip_commit)
            -- show loot drop feedback
            local icon = "general_icon_icecoin1"
            ui.CreateFloatingLootDisplay(unit, icon, drop_stacks)
            wave_drop_resource.winter_coin = wave_drop_resource.winter_coin or 0
            wave_drop_resource.winter_coin = wave_drop_resource.winter_coin + drop_stacks
        elseif CheckSeasonEventExist("autumn_event") then
            AddPlayerAutumnCoin(drop_stacks, skip_commit)
            -- show loot drop feedback
            local icon = "general_icon_coin_leaf"
            ui.CreateFloatingLootDisplay(unit, icon, drop_stacks)
            wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin or 0
            wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin + drop_stacks
        end
    end
end

-- SECTION Boosts
function InitializeBoostGetters()
    -- allow these functions get referenced in wave generator
    wave.data.GetPlayerBoostMultiplier_Gold = GetPlayerBoostMultiplier_Gold
    wave.data.GetPlayerBoostModifier_Token = GetPlayerBoostModifier_Token
    wave.data.GetPlayerBoostModifier_Mine = GetPlayerBoostModifier_Mine
    wave.data.GetPlayerLootModifier_Elixir = GetPlayerLootModifier_Elixir
end

function GetPlayerBoostMultiplier_Gold()
    local multiplier = 0

    if wave.data.mode == "standard" then
        for k, v in pairs(player.timers) do
            local boost = data.boost_timers[k]
            if boost and boost.type == "gold" then
                multiplier = multiplier + tonumber(boost.multiplier)
            end
        end
    end

    return 1 + multiplier
end

function GetPlayerBoostModifier_Token()
    local modifier = 0

    if wave.data.mode == "standard" then
        for k, v in pairs(player.timers) do
            local boost = data.boost_timers[k]
            if boost and boost.type == "token" then
                modifier = modifier + tonumber(boost.multiplier)
            end
        end
    end

    return modifier
end

function GetPlayerBoostModifier_Mine()
    local modifier = 0

    if wave.data.mode == "standard" then
        for k, v in pairs(player.timers) do
            local boost = data.boost_timers[k]
            if boost and boost.type == "mine" then
                modifier = modifier + tonumber(boost.multiplier)
            end
        end
    end

    return modifier
end

function GetPlayerLootModifier_Elixir()
    local bonus_elixir = data.town_upgrade:GetElixirIncome()

    bonus_elixir = math.max(10, bonus_elixir * 0.5)

    return math.ceil(player.attributes.monster_elixir * bonus_elixir)
end

function StartBoostTimer()
    local timer = core.RealTimer:New()

    -- pass time to boost display each tick
    function timer:Tick()
        local time_elapsed = self:GetTimeElapsed()
        ui.boost_display:PassTime(time_elapsed)
    end

    ui.boost_display.timer = timer
end

function ResumeBoostTimer()
    StartBoostTimer()

    ui.boost_display:ClearBoosts()
    ui.boost_display:CreateBoosts()
end

function StopBoostTimer()
    if ui.boost_display.timer then
        -- consume minumum of 1 second of time to avoid potential cheesing
        local time_consumed = ui.boost_display.timer:GetTimeElapsed()
        time_consumed = math.max(time_consumed, 1)

        -- consume time from timers
        for key, timer in pairs(player.timers) do
            player.timers[key] = timer - time_consumed
        end

        -- update UI timer data
        ui.boost_display.timer:Destroy()
        ui.boost_display.timer = nil

        -- save to bank
        bank.CommitTimers()
    end
end

function StartGemAutoBattleTimer()
    local value = player.gem_auto_time
    local timer = core.RealTimer:New()
    player.gem_auto_timer = timer

    -- wave data instance is set by parameter since the battle timer starts
    timer.wave_instance = wave.data.instance
    timer:AddDuration(value)

    function timer:Condition()
        if self.wave_instance ~= wave.data.instance then
            return true
        elseif self:GetTimeRemaining() <= 0 then
            return true
        end

        return false
    end

    function timer:Action()
        local time = self:GetTimeElapsed()
        time = math.ceil(time)

        DisableAutoBattle()

        player.gem_auto_time = player.gem_auto_time - time
        bank.CommitAutoBattleTime()
    end
end

function AddGemAutoBattleTime(value)
    player.gem_auto_time = player.gem_auto_time + value

    -- layout.frame is set to nil OnContinue or OnCanel in autobattle.menu/prompt layout
    if ui.autobattle.menu and ui.autobattle.menu.frame then
        ui.autobattle.menu.UpdateTimeStack(player.gem_auto_time)
    end
    if ui.autobattle.prompt and ui.autobattle.prompt.frame then
        ui.autobattle.prompt.UpdateTimeStack(player.gem_auto_time)
    end

    bank.CommitAutoBattleTime()
end

function AddSpeedx3AdCooldown(value, skip_commit)
    if player.speed_ad_timer and player.speed_ad_timer:GetTimeRemaining() > 0 then
        -- add time to timer if it exists and isn't expired
        player.speed_ad_timer = player.speed_ad_timer
    else
        -- otherwise create new timer
        player.speed_ad_timer = core.OsTimer:New()
    end

    player.speed_ad_timer:AddDuration(value)

    function player.speed_ad_timer:Tick()
        -- only show speed ad cooldown if player not sped up
        if not player.speed_up then
            local is_cooldown = true
            local time_remaining = player.speed_ad_timer:GetTimeRemaining()
            ui.battle.floating_speedup:PassTime(time_remaining, is_cooldown)
        end
    end

    function player.speed_ad_timer:Action()
        ui.battle.floating_speedup:Deactivate()
    end

    if not skip_commit then
        local key = bank.SpeedAdTimerKey()
        bank.CommitTimerExpiration(key, value)
    end
end

function PassSpeedx3TimeToStore(time)
    if ui.shop and ui.shop.is_active then
        local page_key = "speed_up"
        local shop_info = ui.shop.sections[page_key].info_frame
        if shop_info then
            UpdateShopSpeedUpInfo(shop_info)
        end
    end
end

function AddSpeedx3Time(value, skip_commit)
    if player.speed_up_timer and player.speed_up_timer:GetTimeRemaining() > 0 then
        -- add time to timer if it exists and isn't expired
        player.speed_up_timer = player.speed_up_timer
    else
        -- otherwise create new timer
        player.speed_up_timer = core.OsTimer:New()
    end

    player.speed_up_timer:AddDuration(value)

    -- each tick, pass time remaining to relevant UI
    function player.speed_up_timer:Tick()
        local time_remaining = player.speed_up_timer:GetTimeRemaining()

        ui.battle.floating_speedup:PassTime(time_remaining)
        PassSpeedx3TimeToStore(time_remaining)
    end

    function player.speed_up_timer:Action()
        SpeedUpExpireActions()
    end

    -- execute an update tick immediately if time remains
    if player.speed_up_timer:GetTimeRemaining() > 0 then
        ui.battle.floating_speedup:Activate()
        player.speed_up_timer:Tick()

        SetSpeedx3(true)
    end

    if not skip_commit then
        local key = bank.SpeedUpTimerKey()
        bank.CommitTimerExpiration(key, value)

        -- only set player speed if not set from bank.LoadTimers
        player.speed = 3
        ui.bottom_panel.speed:Update(player.speed)
    end
end

function SpeedUpExpireActions()
    SetSpeedx3(false)

    if player.speed_ad_timer and player.speed_ad_timer:GetTimeRemaining() > 0 then
        ui.battle.floating_speedup:Deactivate()
        player.speed_ad_timer:Tick()
    else
        ui.battle.floating_speedup:Deactivate()
    end
end

function SetSpeedx3(enabled)
    player.speed_up = enabled

    ui.bottom_panel.speed:Set3xMode(enabled)

    if not enabled then
        player.speed = math.min(player.speed, 2)

        if wave.data.is_active then
            ui.bottom_panel.speed:Update(player.speed)
            SetGameSpeed(player.speed)
        end
    end
end

-- !SECTION

function AddSeasonPassTime(value, skip_commit)
    if player.seasonpass_timer and player.seasonpass_timer:GetTimeRemaining() > 0 then
        -- add time to timer if it exists and isn't expired
        player.seasonpass_timer = player.seasonpass_timer
    else
        -- otherwise create new timer
        player.seasonpass_timer = core.OsTimer:New()
    end

    player.seasonpass_timer:AddDuration(value)

    -- each tick, pass time remaining to relevant UI
    function player.seasonpass_timer:Tick()
        if ui.calendar and ui.calendar.is_visible then
            ui.calendar:UpdateSeasonPassTime(player.seasonpass_timer:GetTimeRemaining())
        end
    end

    function player.seasonpass_timer:Action()
        if ui.calendar and ui.calendar.is_visible then
            ui.calendar:UpdateSeasonPassTime(0)
            ui.calendar:Update()
            DCEI.SetFrameActive(ui.calendar.seasonpass_container, false)
        end
    end

    -- execute an update tick immediately if time remains
    if player.seasonpass_timer:GetTimeRemaining() > 0 then
        player.seasonpass_timer:Tick()
        if ui.calendar and ui.calendar.is_visible then
            ui.calendar:Update(true)
        end
    end

    if not skip_commit then
        local key = bank.SeasonPassKey()
        bank.CommitTimerExpiration(key, value)
    end
end

-- SECTION Another Functions
function AdvanceBattleSpeed()
    local max_speed = player.speed_up and 3 or 2
    player.speed = (player.speed % max_speed) + 1

    if max_speed == 2 and player.speed == 2 then
        AttemptShowSpeedButton()
    elseif not player.speed_up then
        ui.battle.floating_speedup:Hide()
    end

    SetGameSpeed(player.speed)
    ui.bottom_panel.speed:Update(player.speed)
end

function SetGameSpeed(speed)
    local max_speed = 3
    speed = math.min(speed, max_speed)

    if speed == 3 then
        DCEI.SetUpdateFrequency(2)
        DCEI.SetSpeedFactor(1.5)
    elseif speed == 2 then
        DCEI.SetUpdateFrequency(2)
        DCEI.SetSpeedFactor(1)
    else
        DCEI.SetUpdateFrequency(1)
        DCEI.SetSpeedFactor(speed)
    end
end

function ClearGameSpeed()
    local default = 1

    DCEI.SetSpeedFactor(default)
    DCEI.SetUpdateFrequency(default)
end

-- !SECTION
function InitTowerList()
    CreateTowerList()
    ui.list_tower:Hide()
end

function InitHeroList()
    CreateHeroList()
    ui.list_hero:Hide()
end
-- SECTION Click select box
function ClickDefenseSelectBox(select_box, select_index)
    if select_box.locked then
        if select_box.unlock_level then
            ui.ShowFeedbackNeedCastelLevel(select_box.unlock_level)
        end
    else
        local unit_index = nil

        if player.defenses[select_index] then
            local name = DCEI.GetUnitType(player.defenses[select_index])
            unit_index = GetUnitDataIndex(data.defenses[name], data.defenses)
        end

        ui.loadout.defense.select = select_index
        ui.loadout.current = ui.loadout.defense
        CloseActivePrompts()

        if not ui.list_defense or (ui.list_defense and not ui.list_defense.is_active) then
            -- create defense list if one doesn't exist
            CreateDefenseList(unit_index)
        elseif unit_index then
            -- try to scroll to selected unit
            ui.list_defense.ScrollTo(unit_index)
        end

        ui.list_defense.Highlight(unit_index)
    end
end

function ClickTowerSelectBox(select_box, select_index)
    if select_box.locked then
        if select_box.unlock_level then
            ui.ShowFeedbackNeedCastelLevel(select_box.unlock_level)
        end
    elseif select_box.token_locked then
        if ui.buy_extra_tower_menu then
            ui.buy_extra_tower_menu:Close()
        end
        local callbacks = {
            purchase_slot = AttemptPurchaseExtraTowerSlot,
            cost = data.unlock.token_unlock_cost.tower_slot[select_index] or 20,
        }
        ui.buy_extra_tower_menu = ui.CreateBuyExtraTowerSlotMenu(select_box, select_index, callbacks)
    else
        local unit_index = nil

        if player.towers[select_index] then
            local name = DCEI.GetUnitType(player.towers[select_index])
            local unit_data = data.towers[name]
            unit_index = GetTowerGalleryIndex(unit_data)
        end

        ui.loadout.tower.select = select_index
        ui.loadout.current = ui.loadout.tower
        CloseActivePrompts()

        if not ui.list_tower or (ui.list_tower and not ui.list_tower.is_active) then
            -- create town list if one doesn't exist
            CreateTowerList(unit_index)
        elseif unit_index then
            -- try to scroll to selected unit
            ui.list_tower.ScrollTo(unit_index)
        end

        ui.list_tower.Highlight(unit_index)
    end
end

function ClickTownSelectBox(select_index)
    local unit_index = nil

    if player.town[select_index] then
        local name = DCEI.GetUnitType(player.town[select_index])
        name = name:gsub(" Rank 2", "")
        name = name:gsub(" Rank 3", "")
        name = name:gsub(" Skin Winter", "")
        unit_index = GetUnitDataIndex(data.town[name], data.town)
    end

    ui.loadout.town.select = select_index
    ui.loadout.current = ui.loadout.town

    if not ui.list_town or (ui.list_town and not ui.list_town.is_active) then
        -- create town list if one doesn't exist
        CreateTownList(unit_index)
    elseif unit_index then
        -- try to scroll to selected unit
        ui.list_town.ScrollTo(unit_index)
    end

    ui.list_town.Highlight(unit_index)
end

function ClickHeroSelectBox(select_box, id)
    if ui.loadout.hero.is_active then
        local unit = player.heroes[id]

        if wave.data.is_active then
            -- attempt cast hero skill

            if unit then
                if not event_mode.IsActive() then
                    Tutorial01_Hero_StepB(unit)
                end
                AttemptHeroCastSkill(unit)
            end
        else
            if select_box.locked then
                if select_box.unlock_level then
                    ui.ShowFeedbackNeedCastelLevel(select_box.unlock_level)
                end
            elseif select_box.token_locked then
                if ui.buy_extra_tower_menu then
                    ui.buy_extra_tower_menu:Close()
                end
                local callbacks = {
                    purchase_slot = AttemptPurchaseLockedHeroSlot,
                    cost = data.unlock.token_unlock_cost.hero_slot[id] or 20,
                }
                ui.buy_extra_tower_menu = ui.CreateBuyExtraTowerSlotMenu(select_box, id, callbacks)
            else
                -- open hero list

                select_box.Highlight()
                LoadoutSelectHero(id)
                if not event_mode.IsActive() then
                    Tutorial06_Promote_StepB()
                end
            end
        end
    end
end

function LoadoutSelectHero(select_index)
    local unit_index = nil
    ui.loadout.hero.select = select_index
    ui.loadout.current = ui.loadout.hero
    CloseActivePrompts()

    if player.heroes[select_index] then
        local name = DCEI.GetUnitType(player.heroes[select_index])
        local unit_data = data.heroes[name]
        unit_index = GetHeroGalleryIndex(unit_data)
    end

    if not ui.list_hero or (ui.list_hero and not ui.list_hero.is_active) then
        -- create hero list if one doesn't exist
        CreateHeroList(unit_index)
    elseif unit_index then
        -- try to scroll to selected unit
        ui.list_hero.ScrollTo(unit_index)
    end

    -- update list item highlighting
    ui.list_hero.Highlight(unit_index)

    if wave.data.GetMax() >= 60 and not player.autumn_event_timer then
        if not player.system.hero_wolfhuntress_onboard then
            BlockInput()
            local new_hero_index = GetHeroGalleryIndex(data.heroes["Hero WolfHuntress"])
            ui.CreateNewHeroGlow(ui.list_hero.items[new_hero_index])
            DCEI.TriggerAddTimerEventElapsed(function()
                UnblockInput()
                local animation = true
                ui.list_hero.ScrollTo(new_hero_index, animation)
            end, 0.5, true, true)
            player.system.hero_wolfhuntress_onboard = system.unlocked
            bank.CommitSystemState()
        end
    end
end

-- !SECTION

-- SECTION Update lists
function UpdateHeroList()
    if ui.list_hero then
        ui.list_hero.Update()
    end
    if ui.list_promote then
        ui.list_promote.Update()
    end
end

function UpdateTowerList()
    if ui.list_tower then
        ui.list_tower.Update()
    end
    if ui.list_promote then
        ui.list_promote.Update()
    end
end

function UpdateDefenseList()
    if ui.list_defense then
        ui.list_defense.Update()
    end
end

function UpdateTownList()
    if ui.list_town then
        ui.list_town.Update()
    end
end

function UpdateItemMenu(scroll_to_bottom)
    if ui.item_menu and ui.item_menu.is_active then
        ui.item_menu:Update(scroll_to_bottom)
    end
end

-- !SECTION

-- SECTION Hero list and purchase, promote
function CreateHeroList(unit_index)
    local list_data = ""
    if event_mode.IsActive() then
        list_data = data.heroes.event_gallery
    else
        list_data = data.heroes.gallery
    end

    unit_index = unit_index or 1

    local callbacks = {
        unlock_equipment = UnlockEquipment,
        purchase = AttemptPurchaseHero,
        upgrade = AttemptUpgradeHero,
        equip = ProcessEquipHero,
        promote = CreateHeroPromoteList,
        process_equip_equipment = ProcessEquipEquipment,
        upgrade_equipment = AttemptUpgradeEquipment,
        sell = SellEquipment,
        disenchant = DisenchantEquipment,
        enchant = EnchantEquipment,
        reset_enchant = ResetEquipmentEnchant,
        open_inventory = NewInventoryMenu,
        check_hero_level = CheckHeroLevelTutorial,
        promote_tutorial_check = Tutorial06_Promote_StepC_Check,
        promote_tutorial = Tutorial06_Promote_StepC,
    }

    AttemptCloseTowerList()
    AttemptCloseDefenseList()

    ui.loadout.tower.FadeOut()
    ui.loadout.defense.FadeOut()
    ui.loadout.gear:Hide()

    local function OnClose()
        DIAGNOSTIC.ul_close_time = REAL_TIMER_TIME

        ui.loadout.tower.FadeIn()
        ui.loadout.defense.FadeIn()
        ui.loadout.gear:Show()
        ui.loadout.hero.ClearHighlight()
    end

    if ui.list_hero then
        ui.list_hero:ShowListLayout()
    else
        local text = DCEI.GetTranslationText("ui/menu/heroes")
        ui.list_hero = ui.NewList(list_data, OnClose, callbacks)
        ui.list_hero.SetTitle(text)
        local sort = true
        ui.list_hero.PopulateWithUnitData(sort)
    end
    ui.list_hero.Update()
    ui.list_hero.ScrollTo(unit_index)
    -- LogEvent(LogEvents.menu_open("Hero List"))
end

function CreateTowerList(unit_index)
    local list_data = data.towers.gallery
    unit_index = unit_index or 1

    local callbacks = {
        purchase = AttemptPurchaseTower,
        upgrade = AttemptUpgradeTower,
        equip = ProcessEquipTower,
        promote = CreateTowerPromoteList,
    }

    AttemptCloseHeroList()
    AttemptCloseDefenseList()

    MoveCamera(cameras.towers)
    ui.loadout.hero.FadeOut()
    ui.loadout.defense.FadeOut()
    ui.loadout.gear:Hide()

    local function OnClose()
        ui.loadout.hero.FadeIn()
        ui.loadout.defense.FadeIn()
        ui.loadout.gear:Show()
        ui.loadout.tower.ClearHighlight()
        MoveCamera(cameras.castle)
    end

    if ui.list_tower then
        ui.list_tower:ShowListLayout()
    else
        local text = DCEI.GetTranslationText("ui/menu/towers")
        ui.list_tower = ui.NewList(list_data, OnClose, callbacks)
        ui.list_tower.SetTitle(text)
        local sort = true
        ui.list_tower.PopulateWithUnitData(sort)
    end
    ui.list_tower.ScrollTo(unit_index)
    -- LogEvent(LogEvents.menu_open("Tower List"))
end

function CreateDefenseList(unit_index)
    local list_data = data.defenses
    unit_index = unit_index or 1

    local callbacks = {
        purchase = AttemptPurchaseDefense,
        upgrade = AttemptUpgradeDefense,
        equip = ProcessEquipDefense,
    }

    AttemptCloseHeroList()
    AttemptCloseTowerList()

    MoveCamera(cameras.defenses)
    ui.loadout.hero.FadeOut()
    ui.loadout.tower.FadeOut()
    ui.loadout.gear:Hide()

    local function OnClose()
        ui.loadout.hero.FadeIn()
        ui.loadout.tower.FadeIn()
        ui.loadout.gear:Show()
        ui.loadout.defense.ClearHighlight()
        MoveCamera(cameras.castle)
    end

    if ui.list_defense then
        ui.list_defense:ShowListLayout()
    else
        local text = DCEI.GetTranslationText("ui/menu/defenses")
        ui.list_defense = ui.NewList(list_data, OnClose, callbacks)
        ui.list_defense.SetTitle(text)
        local sort = false
        ui.list_defense.PopulateWithUnitData(sort)
    end
    ui.list_defense.ScrollTo(unit_index)
    -- LogEvent(LogEvents.menu_open("Defense List"))
end

function CreateTowerPromoteList(unit_data)
    local promote_tree = unit_data:GetPromoteTree()

    -- hide unit list UI
    ui.CloseUnitCard()
    ui.list_tower.Show(false)

    local callbacks = {}
    if event_mode.IsActive() then
        callbacks = {
            upgrade = AttemptUpgradeTower,
            equip = ProcessEquipTower,
            promote = AttemptPromoteTower,
            --promote = event_mode.AttemptPromoteTower,
            reset = ChangeTowerClassAndUpdateUI,
        }
    else
        callbacks = {
            upgrade = AttemptUpgradeTower,
            equip = ProcessEquipTower,
            promote = AttemptPromoteTower,
            reset = ChangeTowerClassAndUpdateUI,
        }
    end

    local function OnClose()
        ui.list_promote = nil

        if ui.list_tower then
            ui.list_tower.Show(true)
            ui.screen_mask.lower:Show()
        end
    end

    local text = DCEI.GetTranslationText("ui/menu/promote")
    --local no_scroll = true
    ui.list_promote = ui.NewPromote(promote_tree, OnClose, callbacks, ui.upper)
    ui.list_promote.unit_data = unit_data
    ui.list_promote.PopulateWithPromoteData()
    ui.list_promote.SetTitle(text)
end

function AttemptPromoteTower(from_unit_data, unit_data)
    local cost = unit_data:GetPromoteCost(from_unit_data)
    local gems = GetPlayerGems()

    local level = unit_data:GetLevel()
    local required_level = unit_data:GetPromoteLevel()

    if level < required_level then
        ui.ShowFeedbackTooLowLevel()
    elseif gems < cost then
        local function OnStore()
            ui.CloseUnitCard()

            AttemptClosePromoteList()
            AttemptCloseTowerList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    else
        local spend_event = { type = "promote", target = unit_data.name }
        SpendPlayerGems(cost, spend_event)

        -- promote hero
        ChangeTowerClassAndUpdateUI(from_unit_data, unit_data)
        PlayGoldSpendSound()
    end
end

function CreateTownList(unit_index)
    local list_data = data.town
    unit_index = unit_index or 1

    local callbacks = {
        purchase = AttemptPurchaseTown,
        equip = ProcessEquipTown,
        upgrade = AttemptUpgradeTown,
    }

    local function OnClose(skip_camera)
        ui.loadout.town:SetAlpha(0)
        ui.loadout.town:ClearHighlight()
        ui.loadout.town:HighlightUnequiptSlot()

        CloseActivePrompts()

        if not skip_camera then
            MoveCamera(cameras.town)
            DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, true)
            DCEI.SetFrameActive(ui.loadout.town.airship.frame, true)
            DCEI.SetFrameActive(ui.loadout.town.hall.frame, true)
            DCEI.SetFrameActive(ui.loadout.town.vault.frame, true)
        end
    end

    CloseActivePrompts()
    MoveCamera(cameras.buildings)
    ui.loadout.town:FadeIn()

    DCEI.SetFrameActive(ui.loadout.town.dwarf.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.airship.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.hall.frame, false)
    DCEI.SetFrameActive(ui.loadout.town.vault.frame, false)

    -- hide district button pip
    ui.glue.district_menu:StopAnimating()
    ui.glue.district_menu:HidePip()
    tutorial.town_pip = false

    -- check for town tutorial message
    if tutorial.town then
        -- local offset = -491
        -- local text = tutorial.message.town_building
        -- ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
    end

    local owned_count = data.GetOwnedTownCount()
    if owned_count > 0 and owned_count <= 2 and not data.OneTownIsUpgrade() then
        local offset = -491
        local text = tutorial.message.town_building
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
    end

    if ui.list_town then
        ui.list_town:ShowListLayout()
    else
        local text = DCEI.GetTranslationText("ui/menu/town")
        ui.list_town = ui.NewList(list_data, OnClose, callbacks)
        ui.list_town.SetTitle(text)
        local sort = false
        ui.list_town.PopulateWithUnitData(sort)
    end
    ui.list_town.ScrollTo(unit_index)

    -- LogEvent(LogEvents.menu_open("Town List"))
end

function CreateHeroPromoteList(hero_data)
    local promote_tree = hero_data:GetPromoteTree()

    if not ui.list_hero then
        DIAGNOSTIC.ul_create_time = REAL_TIMER_TIME
        BugTracking_CreateHeroPromoteList(hero_data)
        return
    end

    -- hide unit list UI
    ui.CloseUnitCard()
    ui.list_hero.Show(false)

    local callbacks = {}
    if event_mode.IsActive() then
        callbacks = {
            upgrade = AttemptUpgradeHero,
            equip = ProcessEquipHero,
            promote = event_mode.AttemptPromoteHero,
            awaken = event_mode.AttemptAwakenHero3rdPromotion,
            reset = ChangeHeroClassAndUpdateUI,
            promote_tutorial = Tutorial06_Promote_StepE,
        }
    else 
        callbacks = {
            upgrade = AttemptUpgradeHero,
            equip = ProcessEquipHero,
            promote = AttemptPromoteHero,
            awaken = AttemptAwakenHero3rdPromotion,
            reset = ChangeHeroClassAndUpdateUI,
            promote_tutorial = Tutorial06_Promote_StepE,
        }
    end

    if hero_data:GetMaxLevel() == 9999 then
        callbacks.hero_3rd_promotion = true
    end

    local function OnClose()
        ui.list_promote = nil

        if ui.list_hero then
            ui.list_hero.Show(true)
            ui.screen_mask.lower:Show()
        end

        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
    end

    local text = DCEI.GetTranslationText("ui/menu/promote")
    --local no_scroll = true
    ui.list_promote = ui.NewPromote(promote_tree, OnClose, callbacks, ui.upper)
    ui.list_promote.unit_data = hero_data
    ui.list_promote.PopulateWithPromoteData()
    ui.list_promote.SetTitle(text)

    if not event_mode.IsActive() then
        Tutorial06_Promote_StepD()
    end
end

function AttemptPromoteHero(from_unit_data, unit_data)
    if unit_data.awaken_cost and not unit_data.is_awaken then
        return
    end
    local cost = unit_data:GetPromoteCost(from_unit_data)
    local gems = GetPlayerGems()

    local level = unit_data:GetLevel()
    local required_level = unit_data:GetPromoteLevel()

    if level < required_level then
        ui.ShowFeedbackTooLowLevel()
    elseif gems < cost then
        local function OnStore()
            ui.CloseUnitCard()

            AttemptClosePromoteList()
            AttemptCloseHeroList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    else
        local spend_event = { type = "promote", target = unit_data.name }
        SpendPlayerGems(cost, spend_event)

        -- promote hero
        ChangeHeroClassAndUpdateUI(from_unit_data, unit_data)
        PlayGoldSpendSound()

        -- finish promote tutorial on hero promotion
        if player.system.promote_tutorial ~= system.active then
            player.system.promote_tutorial = system.active
            bank.CommitSystemState()
        end
    end
    if not event_mode.IsActive() then
        Tutorial06_Promote_StepF()
    end
end

function AttemptAwakenHero3rdPromotion(unit_data, card)
    local material_require = unit_data.awaken_cost.normal_material
    local material = GetMaterialCount("normal_material")
    if material < material_require then
        ui.ShowFeedbackNotEnoughMaterial("normal_material")
        return
    end
    local token_require = unit_data.awaken_cost.token
    local token = GetPlayerGems()
    if token < token_require then
        local function OnStore()
            ui.CloseUnitCard()
            AttemptClosePromoteList()
            AttemptCloseHeroList()
        end

        local offset = -200
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
        return
    end
    local spend_event = { type = "awaken", target = unit_data.name }
    SpendPlayerGems(token_require, spend_event)
    AddPlayerMaterial("normal_material", -material_require)
    PlayGoldSpendSound()
    BlockInput()
    if ui.opeing_promote_node and DCEI.FrameExists(ui.opeing_promote_node.seal) then
        DCEI.DestroyFrame(ui.opeing_promote_node.seal)
    end
    DCEI.DestroyFrame(card.awaken_panel.Frame)
    DCEI.SetFrameImage(card.promote.button, "btn_yellow")
    unit_data.is_awaken = true
    bank.CommitHeroAwaken(unit_data)
    DCEI.AnimateFrameAlpha(card.seal_mask, 1, 0, 3, "InCubic")
    DCEI.AnimateFrameScale(card.seal_icon, { x = 1, y = 1 }, { x = 1.8, y = 1.8 }, 2.5, "InBounce")
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.PlaySound(DCEI.Sound("awaken_sucessed"))
        DCEI.AnimateFrameAlpha(card.seal_icon, 1, 0, 0.8, "InCubic")
    end, 1.8, true, true)
    DCEI.TriggerAddTimerEventElapsed(function()
        UnblockInput()
        DCEI.DestroyFrame(card.seal_mask)
        DCEI.DestroyFrame(card.seal_icon)
    end, 2.5, true, true)
end

-- NOTE Change Hero Class
function ChangeHeroClassAndUpdateUI(old_class, new_class)
    ChangeHeroClass(old_class, new_class)
    LogEvent(LogEvents.HeroSwitchClass(old_class, new_class))
    if event_mode.IsActive() then
        bank.CommitEventUnitData(new_class, bank.bank_data.prop.class, new_class.name)
    else
        bank.CommitUnitData(new_class, bank.bank_data.prop.class, new_class.name)
    end
    if ui.autobattle_button then
        DCEI.DestroyFrame(ui.autobattle_button.frame)
        ui.autobattle_button = nil
    end

    -- close promote UI
    ui.CloseUnitCard()
    DCEI.DestroyFrame(ui.list_promote.frame)
    ui.list_promote = nil


    DIAGNOSTIC.ul_close_time = REAL_TIMER_TIME
    DCEI.DestroyFrame(ui.list_hero.frame)
    ui.list_hero = nil

    CreateHeroList()
    ui.list_hero:Close()

    -- destroy and recreate hero list with new class
    local unit_index = GetHeroGalleryIndex(new_class)

    CreateHeroList(unit_index)
    ui.list_hero.Open(unit_index)
    -- only update highlight if unit is actually equipped
    if new_class.slot > 0 then
        ui.list_hero.Highlight(unit_index)
    end
end

function ChangeHeroClass(old_class, new_class)
    local slot = old_class.slot
    local root_data = old_class:GetParentData()
    root_data.current_class = new_class.name

    new_class.equipment = old_class.equipment

    -- replace hero if currently equipped
    if slot > 0 then
        UnequipHeroSlot(slot)
        EquipHero(new_class, slot)
    end

    -- update gallery data
    data.Replace(data.heroes.gallery, old_class, new_class)
    if event_mode.IsActive() then
        data.Replace(data.heroes.event_gallery, old_class, new_class)
    end
end

function AttemptPurchaseHero(unit_data)
    if unit_data.event_lock then
        if CheckSeasonEventExist(unit_data.event_lock) then
            ui.CloseUnitCard()
            AttemptCloseHeroList()
            if unit_data.event_lock == "winter_event2022" then
                OpenWinterEvent()
                ui.winter_event_menu:SetRewardScroll(1)
            elseif unit_data.event_lock == "autumn_event" then
                OpenAutumnEvent()
                ui.autumn_event_menu:SetRewardScroll(1)
            end
        else
            local token_cost = unit_data.token_cost
            local gems = GetPlayerGems()

            if token_cost <= gems then
                local spend_event = { type = "buy", target = unit_data.name }
                SpendPlayerGems(token_cost, spend_event)

                ActuallyPurchaseHero(unit_data)
            else
                local function OnStore()
                    ui.CloseUnitCard()
                    AttemptCloseHeroList()
                end

                local offset = -356
                ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
            end
        end
    else
        local cost = unit_data.cost_adjust or unit_data.cost
        local gold = GetPlayerGold()
        if cost <= gold then
            local spend_event = { type = "buy", target = unit_data.name }
            SpendPlayerGold(cost, spend_event)

            ActuallyPurchaseHero(unit_data)
        else
            local function OnStore()
                ui.CloseUnitCard()
                AttemptCloseHeroList()
            end

            local offset = -356
            ui.ShowFeedbackNeedMoreGold(offset, ui.upper, OnStore)
        end
    end
end

function ActuallyPurchaseHero(unit_data)
    PurchaseHero(unit_data)
    PlayGoldSpendSound()
    EquipHero(unit_data)

    if ui.unit_card then
        DCEI.DestroyFrame(ui.unit_card.purchase.button)

        -- don't show promote button when hero level-up tutorial is active
        local unlock_equipment = UnlockEquipment()
        local display_equipment_slot = true
        local hide_promote_state = false
        if CheckHeroLevelTutorial() then
            hide_promote_state = true
            display_equipment_slot = false
        end

        ui.GenerateUnitCardButtons(ui.unit_card, hide_promote_state, display_equipment_slot, unlock_equipment)

        ui.UpdateUnitCard()
    end
end

function AttemptUpgradeHero(unit_data)
    local cost = unit_data:GetUpgradeCost()
    local level = unit_data:GetLevel()
    local max_level = unit_data.max_level
    local gold = GetPlayerGold()
    local success = true

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
        success = false
    elseif cost <= gold then
        local spend_event = { type = "upgrade", target = unit_data.name }
        SpendPlayerGold(cost, spend_event)

        UpgradeHero(unit_data)
        PlayGoldSpendSound()

        if ui.unit_card then
            ui.UpdateUnitCard()
        end
        UpdateHeroList()
    else
        local function OnStore()
            ui.CloseUnitCard()

            -- need to check to close both lists as units can be upgraded from promote list
            AttemptClosePromoteList()
            AttemptCloseHeroList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)
        success = false
    end


    return success
end

function ProcessEquipHero(unit_data)
    local slot = ui.loadout.hero.select

    if slot == unit_data.slot then
        UnequipHero(unit_data)
    else
        EquipHero(unit_data)
    end

    ui.unit_card.equip.Update()
end

function AttemptPurchaseLockedHeroSlot(select_box, select_index)
    if not select_box.token_locked then
        return
    end
    local cost = data.unlock.token_unlock_cost.hero_slot[select_index] or 20
    local gem = GetPlayerGems()

    if cost <= gem then
        local spend_event = { type = "buy", target = "extra tower slot" }
        SpendPlayerGems(cost, spend_event)

        select_box.token_locked = false
        DCEI.SetFrameActive(select_box.lock_icon, false)
        if select_box.token_label then
            DCEI.DestroyFrame(select_box.token_label.Frame)
        end
        bank.CommitHeroSlotTokenUnlock(select_index)
        PlayGoldSpendSound()

        if ui.buy_extra_tower_menu then
            ui.buy_extra_tower_menu:Close()
        end

        ClickHeroSelectBox(select_box, select_index)
    else
        local function OnStore()
            AttemptCloseHeroList()
            if ui.buy_extra_tower_menu then
                ui.buy_extra_tower_menu:Close()
            end
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

-- !SECTION

-- SECTION TOWER purchase and equip
function AttemptPurchaseExtraTowerSlot(select_box, select_index)
    if not select_box.token_locked then
        return
    end
    local cost = data.unlock.token_unlock_cost.tower_slot[select_index] or 20
    local gem = GetPlayerGems()

    if cost <= gem then
        local spend_event = { type = "buy", target = "extra tower slot" }
        SpendPlayerGems(cost, spend_event)

        select_box.token_locked = false
        DCEI.SetFrameActive(select_box.lock_icon, false)
        if select_box.token_label then
            DCEI.DestroyFrame(select_box.token_label.Frame)
        end
        bank.CommitExtraTowerSlots(select_index)
        PlayGoldSpendSound()

        if ui.buy_extra_tower_menu then
            ui.buy_extra_tower_menu:Close()
        end

        ClickTowerSelectBox(select_box, select_index)
    else
        local function OnStore()
            AttemptCloseTowerList()
            if ui.buy_extra_tower_menu then
                ui.buy_extra_tower_menu:Close()
            end
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

function ChangeTowerClassAndUpdateUI(old_class, new_class)
    ChangeTowerClass(old_class, new_class)
    if event_mode.IsActive() then
        bank.CommitEventUnitData(new_class, bank.bank_data.prop.class, new_class.name)
    else
        bank.CommitUnitData(new_class, bank.bank_data.prop.class, new_class.name)
    end

    -- close promote UI
    ui.CloseUnitCard()
    DCEI.DestroyFrame(ui.list_promote.frame)
    ui.list_promote = nil

    -- destroy and recreate hero list with new class
    local unit_index = GetUnitDataIndex(new_class, data.towers.gallery)

    DCEI.DestroyFrame(ui.list_tower.frame)
    ui.list_tower = nil

    CreateTowerList(unit_index)
    ui.list_tower.Open(unit_index)

    -- only update highlight if unit is actually equipped
    if new_class.slot > 0 then
        ui.list_tower.Highlight(unit_index)
    end
end

function ChangeTowerClass(old_class, new_class)
    local slot = old_class.slot
    local root_data = old_class:GetParentData()
    root_data.current_class = new_class.name

    -- replace tower if currently equipped
    if slot > 0 then
        UnequipTowerSlot(slot)
        EquipTower(new_class, slot)
    end

    -- update gallery data
    data.Replace(data.towers.gallery, old_class, new_class)
end

function AttemptPurchaseTower(unit_data)
    local cost = unit_data.cost_adjust or unit_data.cost
    local gold = GetPlayerGold()

    if cost <= gold then
        local spend_event = { type = "buy", target = unit_data.name }
        SpendPlayerGold(cost, spend_event)

        PurchaseTower(unit_data)
        PlayGoldSpendSound()
        EquipTower(unit_data)

        if ui.unit_card then
            DCEI.DestroyFrame(ui.unit_card.purchase.button)
            ui.GenerateUnitCardButtons(ui.unit_card)

            ui.UpdateUnitCard()
        end
    else
        local function OnStore()
            ui.CloseUnitCard()
            AttemptCloseTowerList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.upper, OnStore)
    end
end

function PurchaseTower(unit_data)
    unit_data.is_purchased = true
    if event_mode.IsActive() then
        bank.CommitEventUnitData(unit_data, bank.bank_data.prop.level, 1)
        bank.CommitEventUnitData(unit_data, bank.bank_data.prop.class, unit_data.name, true)
    else
        bank.CommitUnitData(unit_data, bank.bank_data.prop.level, 1)
        bank.CommitUnitData(unit_data, bank.bank_data.prop.class, unit_data.name, true)
    end

    LogUserProperty(UserProps.towers_owned)
end

function AttemptUpgradeTower(unit_data)
    local cost = unit_data:GetUpgradeCost()
    local level = unit_data:GetLevel()
    local max_level = unit_data.max_level
    local gold = GetPlayerGold()
    local success = true

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
        success = false
    elseif cost <= gold then
        local spend_event = { type = "upgrade", target = unit_data.name }
        SpendPlayerGold(cost, spend_event)

        UpgradeTower(unit_data)
        PlayGoldSpendSound()

        if ui.unit_card then
            ui.UpdateUnitCard()
        end
        UpdateTowerList()
    else
        local function OnStore()
            ui.CloseUnitCard()

            -- need to check to close both lists as units can be upgraded from promote list
            AttemptClosePromoteList()
            AttemptCloseTowerList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)
        success = false
    end

    return success
end

function EquippedTower(unit_data)
    local slot = unit_data.slot
    if slot and slot > 0 then
        return player.towers[slot]
    end
end

function UpgradeTower(unit_data)
    local unit = EquippedTower(unit_data)

    if unit_data.slot > 0 then
        ProcessSkillEquipBonuses(unit_data, false)
    end
    unit_data:LevelUp()

    if event_mode.IsActive() then
        bank.CommitEventUnitData(unit_data, bank.bank_data.prop.level, unit_data:GetLevel())
    else
        bank.CommitUnitData(unit_data, bank.bank_data.prop.level, unit_data:GetLevel())
    end

    if unit_data.slot > 0 then
        ProcessSkillEquipBonuses(unit_data, true)
    end

    if unit and DCEI.UnitExists(unit) then
        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end
end

function EquipTower(unit_data, slot, skip_commit)
    slot = slot or ui.loadout.tower.select
    local unit_type = unit_data.name
    local slot_unit = group_tower_slots[slot]

    -- unequip unit currently in the slot, if any
    UnequipTowerSlot(slot)

    -- unequip unit if currently equipped in another slot
    UnequipTower(unit_data)

    DCEI.LogMessage("> equipping " .. unit_type .. " to slot " .. slot)

    local dx, dy = 0, 1
    local pos = DCEI.GetUnitPosition2D(slot_unit)
    local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, dx, dy)
    player.towers[slot] = unit
    unit_data.slot = slot
    unit_data.unit = unit

    if not skip_commit then
        if event_mode.IsActive() then
            bank.CommitEventUnitSlot(unit_data, slot)
        else
            bank.CommitUnitSlot(unit_data, slot)
        end
    end

    ApplyUpgradeDataToUnit(data.tower_upgrades, unit)
    -- DCEI.LogMessage("LIST TOWER UPGRADES (" .. #data.tower_upgrades .. ")")
    -- for k, v in ipairs(data.tower_upgrades) do
    --     DCEI.LogMessage(k .. ". " .. v)
    -- end

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ProcessSkillEquipBonuses(unit_data, true)
    UpdateBlessBonusUnits(unit_data, true)
    UpdateTowerList()

    if ui.list_tower then
        local unit_index = GetTowerGalleryIndex(unit_data)
        ui.list_tower.Highlight(unit_index)
    end
end

function UnequipTower(unit_data)
    local unit_type = unit_data.name
    local slot = unit_data.slot
    local unit = player.towers[slot]

    if unit and DCEI.UnitExists(unit) then
        DCEI.LogMessage("> unequipping " .. unit_type .. " from slot " .. slot)

        if event_mode.IsActive() then
            bank.ClearEventUnitSlot(bank.bank_data.units.tower, slot)
        else
            bank.ClearUnitSlot(bank.bank_data.units.tower, slot)
        end

        DCEI.RemoveUnit(unit)
        ProcessSkillEquipBonuses(unit_data, false)
        UpdateBlessBonusUnits(unit_data, false)

        unit_data.slot = 0
        unit_data.unit = nil
        player.towers[slot] = nil
    end

    -- update UI
    UpdateTowerList()
    if ui.list_tower then
        ui.list_tower.Highlight(0)
    end
end

function UnequipTowerSlot(slot)
    local unit = player.towers[slot]

    if not (unit and DCEI.UnitExists(unit)) or DCEI.GetUnitType(unit) == DCEI.Unit("Castle Hero Slot") then
        return
    end

    local unit_type = DCEI.GetUnitType(unit)
    local unit_data = GetUnitDataFromUnitType(unit_type, data.towers)

    UnequipTower(unit_data)
end

function ProcessEquipTower(unit_data)
    local slot = ui.loadout.tower.select

    if slot == unit_data.slot then
        UnequipTower(unit_data)
    else
        EquipTower(unit_data)
    end

    ui.unit_card.equip.Update()
end

-- !SECTION

-- SECTION DEFENSES
function AttemptPurchaseDefense(unit_data)
    local cost = unit_data.cost_adjust or unit_data.cost
    local gems = GetPlayerGems()

    if cost <= gems then
        local spend_event = { type = "buy", target = unit_data.name }
        SpendPlayerGems(cost, spend_event)

        PlayGoldSpendSound()
        PurchaseDefense(unit_data)
        EquipDefense(unit_data)
        bank.CommitGold()

        if ui.unit_card then
            DCEI.DestroyFrame(ui.unit_card.purchase.button)
            ui.GenerateUnitCardButtons(ui.unit_card)

            ui.UpdateUnitCard()
        end
    else
        local function OnStore()
            ui.CloseUnitCard()
            ui.list_defense.Close()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    end
end

function PurchaseDefense(unit_data)
    unit_data.is_purchased = true
    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, 1)

    LogUserProperty(UserProps.defenses_owned)
end

function AttemptUpgradeDefense(unit_data)
    local cost = unit_data:GetUpgradeCost()
    local level = unit_data:GetLevel()
    local max_level = unit_data.max_level
    local gems = GetPlayerGems()
    local success = true

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
        success = false
    elseif cost <= gems then
        local spend_event = { type = "upgrade", target = unit_data.name }
        SpendPlayerGems(cost, spend_event)

        UpgradeDefense(unit_data)
        PlayGoldSpendSound()

        if ui.unit_card then
            ui.UpdateUnitCard()
        end
        UpdateDefenseList()
    else
        local function OnStore()
            ui.CloseUnitCard()
            AttemptCloseDefenseList()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
        success = false
    end

    return success
end

function EquippedDefense(unit_data)
    local slot = unit_data.slot
    if slot and slot > 0 then
        return player.defenses[slot]
    end
end

function UpgradeDefense(unit_data)
    local unit = EquippedDefense(unit_data)

    if unit_data.slot > 0 then
        ProcessSkillEquipBonuses(unit_data, false)
    end
    unit_data:LevelUp()

    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, unit_data:GetLevel())
    bank.CommitGold()

    if unit_data.slot > 0 then
        ProcessSkillEquipBonuses(unit_data, true)
    end

    if unit and DCEI.UnitExists(unit) then
        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end
end

function EquipDefense(unit_data, slot, skip_commit)
    slot = slot or ui.loadout.defense.select
    local unit_type = unit_data.name
    local slot_unit = group_defense_slots[slot]

    -- unequip unit currently in the slot, if any
    UnequipDefenseSlot(slot)

    -- unequip unit if currently equipped in another slot
    UnequipDefense(unit_data)

    DCEI.LogMessage("> equipping " .. unit_type .. " to slot " .. slot)
    local dx, dy = 0, 1
    local pos = DCEI.GetUnitPosition2D(slot_unit)
    local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, dx, dy)
    player.defenses[slot] = unit
    unit_data.slot = slot
    unit_data.unit = unit

    if not skip_commit then
        bank.CommitUnitSlot(unit_data, slot)
    end

    -- castle defenses are improved by tower upgrades
    ApplyUpgradeDataToUnit(data.tower_upgrades, unit)

    ApplyBaseBehaviorsToUnit(unit_data, unit)
    ProcessSkillEquipBonuses(unit_data, true)
    UpdateBlessBonusUnits(unit_data, true)
    UpdateDefenseList()

    if ui.list_defense then
        local unit_index = GetUnitDataIndex(unit_data, data.defenses)
        ui.list_defense.Highlight(unit_index)
    end
end

function UnequipDefense(unit_data)
    local unit_type = unit_data.name
    local slot = unit_data.slot
    local unit = player.defenses[slot]

    if unit and DCEI.UnitExists(unit) then
        DCEI.LogMessage("> unequipping " .. unit_type .. " from slot " .. slot)

        bank.ClearUnitSlot(bank.bank_data.units.defense, slot)


        unit_data.slot = 0
        unit_data.unit = nil
        player.defenses[slot] = nil

        DCEI.RemoveUnit(unit)
        ProcessSkillEquipBonuses(unit_data, false)
        UpdateBlessBonusUnits(unit_data, false)
    end

    -- update UI
    UpdateDefenseList()
    if ui.list_defense then
        ui.list_defense.Highlight(0)
    end
end

function UnequipDefenseSlot(slot)
    local unit = player.defenses[slot]

    if not (unit and DCEI.UnitExists(unit)) or DCEI.GetUnitType(unit) == DCEI.Unit("Castle Hero Slot") then
        return
    end

    local unit_type = DCEI.GetUnitType(unit)
    local unit_data = GetUnitDataFromUnitType(unit_type, data.defenses)

    UnequipDefense(unit_data)
end

function ProcessEquipDefense(unit_data)
    local slot = ui.loadout.defense.select

    if slot == unit_data.slot then
        UnequipDefense(unit_data)
    else
        EquipDefense(unit_data)
    end

    ui.unit_card.equip.Update()
end

-- !SECTION

-- SECTION TOWN
function AttemptPurchaseTown(unit_data)
    local cost = unit_data.cost_adjust or unit_data.cost
    local gold = GetPlayerGold()

    if cost <= gold then
        local spend_event = { type = "buy", target = unit_data.name }
        SpendPlayerGold(cost, spend_event)

        PurchaseTown(unit_data)
        PlayGoldSpendSound()
        EquipTown(unit_data)

        if ui.unit_card then
            DCEI.DestroyFrame(ui.unit_card.purchase.button)
            ui.GenerateUnitCardButtons(ui.unit_card)

            ui.UpdateUnitCard()
        end
    else
        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local function OnStore()
            ui.CloseUnitCard()

            local skip_camera = true
            ui.list_town.Close(skip_camera)
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.upper, OnStore, OnStoreClose)
    end
end

function PurchaseTown(unit_data)
    unit_data.is_purchased = true
    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, 1)

    -- end tutorial on first purchase
    if tutorial.town then
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
        tutorial.town = false
    end

    -- log waves of first 4 town buildings purchased
    local town_owned = data.GetOwnedTownCount()
    if town_owned <= 4 then
        local user_prop = UserProps.TownOwnedWave(town_owned)
        LogUserProperty(user_prop)
    end

    LogUserProperty(UserProps.town_owned)
    ui.glue.district_menu:HidePip()
end

function GetTownPositionData(unit_data, slot)
    local dx, dy = 0, 1

    -- original town slots positions
    local pos_x, pos_y = { 21.9, 26.1, 22, 26 }, { 5.8, 5.8, 3.2, 3.2 }

    local data = {
        x = pos_x[slot],
        y = pos_y[slot],
        dx = dx,
        dy = dy,
    }

    -- determine if slot is top/bot, left/right
    local is_top_side = (slot == 1 or slot == 2)
    local is_left_side = (slot == 1 or slot == 3)

    -- calculated offsets based on slot position
    local y_offset = is_top_side and 0.4 or 0.2
    local x_offset = is_left_side and 0.1 or -0.1
    x_offset = x_offset * (is_top_side and 1 or -1)

    -- building-specific data
    if unit_data.name == "Town Building Brewery" then
        -- 180 degress
        data.dy = -1
    elseif unit_data.name == "Town Building Workshop" then
        -- 270 degrees
        data.dy = 0
        data.dx = -1
    elseif unit_data.name == "Town Building Blacksmith" then
        -- 90 degrees
        data.dy = 0
        data.dx = 1
    elseif unit_data.name == "Town Building Temple" then
        -- 315 degrees
        data.dy = 0.7
        data.dx = -0.7
    end

    -- apply calculated offsets
    data.y = data.y + y_offset
    data.x = data.x + x_offset
    return data
end

function UpdateTownDistrict(unit_data, index)
    local build_func = BUILD.town_slot[index]

    if unit_data then
        local hide = DCEI.Behavior("System Hide")
        local slot = group_town_slots[index]
        DCEI.ApplyBehaviorAsync(slot, hide)
    else
        local hide = DCEI.Behavior("System Hide")
        local slot = group_town_slots[index]
        DCEI.RemoveBehaviorAsync(slot, hide)
    end

    if build_func then
        build_func()
    end
end

function EquipTown(unit_data, slot)
    slot = slot or ui.loadout.town.select
    local unit_type = unit_data.name
    for name, v in pairs(unit_data.skins) do
        if v == 2 then
            unit_type = unit_data.skin_unit[name] or unit_type
        end
    end
    local level = unit_data.level
    if level > 1 then
        unit_type = unit_type .. " Rank " .. level
    end
    -- unequip unit currently in the slot, if any
    UnequipTownSlot(slot)

    -- unequip unit if currently equipped in another slot
    UnequipTown(unit_data)

    DCEI.LogMessage("> equipping " .. unit_type .. " to slot " .. slot)

    -- use scripted position data for new town visuals
    local pos = GetTownPositionData(unit_data, slot)
    local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, pos.dx, pos.dy)
    player.town[slot] = unit
    unit_data.slot = slot
    ui.loadout.town.slots[slot].equipped = true
    DCEI.ApplyBehaviorAsync(unit, "Builder Birth")
    UpdateTownDistrict(unit_data, slot)
    UpdatePropLayer()

    bank.CommitUnitSlot(unit_data, slot)
    ProcessEquipBonuses(unit_data, true)
    UpdateTownList()

    if ui.list_town then
        local unit_index = GetUnitDataIndex(unit_data, data.town)
        ui.list_town.Highlight(unit_index)
    end
end

function UnequipTown(unit_data)
    local unit_type = unit_data.name
    local slot = unit_data.slot
    local unit = player.town[slot]

    if unit and DCEI.UnitExists(unit) then
        DCEI.LogMessage("> unequipping " .. unit_type .. " Rank " .. unit_data.level .. " from slot " .. slot)

        -- revert default layers
        UpdateTownDistrict(nil, slot)
        UpdatePropLayer()

        bank.ClearUnitSlot(bank.bank_data.units.town, slot)

        unit_data.slot = 0
        player.town[slot] = nil
        ui.loadout.town.slots[slot].equipped = false
        DCEI.RemoveUnit(unit)
        ProcessEquipBonuses(unit_data, false)
    end

    -- update UI
    UpdateTownList()
    if ui.list_town then
        ui.list_town.Highlight(0)
    end
end

function UnequipTownSlot(slot)
    local unit = player.town[slot]

    if not (unit and DCEI.UnitExists(unit)) or DCEI.GetUnitType(unit) == TOWN_SLOT_NAME then
        return
    end

    local unit_type = DCEI.GetUnitType(unit)
    unit_type = data.GetTownBaseName(unit_type)
    local unit_data = GetUnitDataFromUnitType(unit_type, data.town)

    UnequipTown(unit_data)
end

function ProcessEquipTown(unit_data)
    local slot = ui.loadout.town.select

    if slot == unit_data.slot then
        UnequipTown(unit_data)
    else
        EquipTown(unit_data)
    end

    ui.unit_card.equip.Update()
end

function AttemptUpgradeTown(unit_data)
    local cost = unit_data:GetUpgradeCost()
    local level = unit_data:GetLevel()
    local max_level = unit_data.max_level
    local gold = GetPlayerGold()
    local success = true

    if max_level and level == max_level then
        ui.ShowFeedbackMaxLevel()
        success = false
    elseif cost <= gold then
        local spend_event = { type = "upgrade", target = unit_data.name }
        SpendPlayerGold(cost, spend_event)

        UpgradeTown(unit_data)
        PlayGoldSpendSound()

        if ui.unit_card then
            ui.UpdateUnitCard()
        end
        UpdateTownList()
    else
        local function OnStoreClose()
            -- return camera to town
            MoveCamera(cameras.town)
        end

        local function OnStore()
            ui.CloseUnitCard()

            local skip_camera = true
            ui.list_town.Close(skip_camera)
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore, OnStoreClose)
        success = false
    end

    return success
end

function UpgradeTown(unit_data)
    local slot = unit_data.slot
    if slot > 0 then
        -- if the building is equiped, unequip it first to unload building effect
        UnequipTown(unit_data)
    end

    if ui.prompt.tutorial then
        ui.prompt.tutorial.Close()
    end

    unit_data:LevelUp()

    if slot > 0 then
        -- the building is equiped, update building unit
        EquipTown(unit_data, slot)
    end

    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, unit_data:GetLevel())
end

-- NOTE remove town units to improve battle performance
function RemoveBuildingsOnWaveStart()
    local slot_count = data.unlock.town_slot_max
    for slot = 1, slot_count do
        local unit = player.town[slot]
        if unit and DCEI.UnitExists(unit) then
            player.town[slot] = nil
            DCEI.RemoveUnit(unit)
        end
    end

    if VAULT and DCEI.UnitExists(VAULT) then
        DCEI.RemoveUnit(VAULT)
        VAULT = nil
    end
end

function RestoreBuildingsOnWaveEnd()
    local slot_count = data.unlock.town_slot_max
    for slot = 1, slot_count do
        -- restore from save data
        local key = bank.UnitSlotKey(bank.bank_data.units.town, slot)
        local name = DCEI.Save.Get(key)

        if Valid(name) then
            local unit_data = data.town[name]
            local unit_type = unit_data.name
            for name, v in pairs(unit_data.skins) do
                if v == 2 then
                    unit_type = unit_data.skin_unit[name] or unit_type
                end
            end
            local level = unit_data.level
            if level > 1 then
                unit_type = unit_type .. " Rank " .. level
            end

            local pos = GetTownPositionData(unit_data, slot)
            local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, pos.dx, pos.dy)
            player.town[slot] = unit
            unit_data.slot = slot
        end
    end
    -- restore vault
    if not VAULT then
        -- move the vault outside the screen if vault is not active
        local x, y = 23.3, -20
        if not GetPlayerTest("pause_offline_gold") and player.system.vault == system.active then
            x, y = 23.3, 11.2
        end
        local vault_unit_name = DCEI.Unit("Town Vault")
        VAULT = DCEI.CreateUnitAsync(1, 1, vault_unit_name, x, y, 0, -1)
        local options = {
            offset = { up = -0.2 },
            center_at_unit_origin = true,
        }
        local use_current_root = true
        local select_box = ui.loadout.town.vault.frame
        DCEI.AttachFrameToUnit(select_box, VAULT, options, use_current_root)
    end
end

function ChangeTownBuildingSkin(unit_data, skin_name)
    for name, v in pairs(unit_data.skins) do
        unit_data.skins[name] = 1
    end
    if skin_name then
        unit_data.skins[skin_name] = 2
    end
    if unit_data.slot > 0 then
        local slot = unit_data.slot
        local unit = player.town[slot]
        if unit and DCEI.UnitExists(unit) then
            DCEI.RemoveUnit(unit)
            local unit_type = unit_data.name
            for name, v in pairs(unit_data.skins) do
                if v == 2 then
                    unit_type = unit_data.skin_unit[name] or unit_type
                end
            end
            local level = unit_data.level
            if level > 1 then
                unit_type = unit_type .. " Rank " .. level
            end

            local pos = GetTownPositionData(unit_data, slot)
            local unit = DCEI.CreateUnitAsync(1, 1, unit_type, pos.x, pos.y, pos.dx, pos.dy)
            player.town[slot] = unit
        end
    end
    bank.CommitUnitData(unit_data, bank.bank_data.prop.skin, unit_data.skins)
end

-- !SECTION

-- SECTION Process equip
function ProcessEquipBonuses(unit_data, is_addition)
    local sign = is_addition and 1 or -1
    local level = unit_data.level or 0
    local growth_level = level - 1

    if unit_data.attributes then
        for k, attribute in ipairs(unit_data.attributes) do
            local value = unit_data.values[k]
            local name = attribute.name
            local bonus = value.base

            if value.growth then
                bonus = bonus + (value.growth * growth_level)
            end

            if value.max then
                bonus = math.min(bonus, value.max)
            end

            bonus = sign * bonus

            DCEI.LogMessage("> ATTRIBUTE " .. name .. " " .. bonus .. " from " .. unit_data.name)
            player.attributes[name] = player.attributes[name] + bonus

            if name == "archer_damage" or name == "all_damage" then
                UpdateArcherUpgradeButton()
            end
        end
    end

    if unit_data.behaviors then
        for _, behavior in pairs(unit_data.behaviors) do
            local behavior_name = behavior.name
            local target = behavior.target

            DCEI.LogMessage("apply " .. sign .. " " .. behavior_name .. " to " .. target)
            AddBehaviorUpgrade(target, behavior_name, sign)

            -- apply upgrade behaviors for upgraded gears
            -- 11/11 add upgrade behaviors for upgrade town buildings
            local upgrade_behavior = behavior.upgrade
            if upgrade_behavior and growth_level ~= 0 then
                DCEI.LogMessage("apply " .. (sign * growth_level) .. " " .. upgrade_behavior .. " to " .. target)
                AddBehaviorUpgrade(target, upgrade_behavior, sign * growth_level)
            end
        end
    end

    if unit_data.summons then
        for _, summon in pairs(unit_data.summons) do
            AddSummonData(summon, unit_data, sign)
        end
    end
end

function ProcessSkillEquipBonuses(unit_data, is_addition)
    local root_data = unit_data:GetParentData()
    local level = root_data.level or 0
    local growth_level = level - 1
    local sign = is_addition and 1 or -1

    if not unit_data.skill then
        -- early out if unit has no skills (ie most towers)
        return
    end

    for _, skill_data in ipairs(unit_data.skill) do
        -- process hero attributes on equip
        if skill_data.attributes then
            for k, attribute in ipairs(skill_data.attributes) do
                local name = attribute.name
                local value = skill_data.values[k]
                local bonus = value.base

                if value.growth then
                    bonus = bonus + (value.growth * growth_level)
                end
                if value.max then
                    bonus = math.min(bonus, value.max)
                end
                bonus = sign * bonus

                DCEI.LogMessage(
                    "> ATTRIBUTE "
                        .. name
                        .. " "
                        .. bonus
                        .. " from "
                        .. unit_data.name
                        .. "("
                        .. skill_data.name
                        .. ")"
                )
                player.attributes[name] = player.attributes[name] + bonus

                if name == "archer_damage" or name == "all_damage" then
                    UpdateArcherUpgradeButton()
                end
            end
        end

        -- process targeted upgrade behaviors
        if skill_data.behaviors then
            for k, behavior in ipairs(skill_data.behaviors) do
                local behavior_name = behavior.name
                local count = behavior.count or (behavior.count_is_level and growth_level) or 0
                local target = behavior.target

                if count ~= 0 then
                    DCEI.LogMessage("apply " .. (sign * count) .. " " .. behavior_name .. " to " .. target)
                    AddBehaviorUpgrade(target, behavior_name, sign * count)
                end
            end
        end

        -- process tower progress
        if skill_data.progress_key and skill_data.progress_max_key then
            local progress = player.attributes[skill_data.progress_key] or 0
            local max_progress = player.attributes[skill_data.progress_max_key] or 1
            local percentage = progress / max_progress
            local slot = unit_data.slot
            if slot > 0 then
                ui.loadout.tower.UpdateProgress(slot, percentage, is_addition)
            end

            local data = {
                slot = slot,
                max = max_progress,
            }
            player.wave_win_increment[skill_data.progress_key] = is_addition and data or nil
        end
    end
end

-- for single hero
function PrcessHeroEquipBebaviors(unit, unit_data, equip)
    if unit_data.skill then
        for _, skill_data in ipairs(unit_data.skill) do
            if skill_data.equip_behavior then
                for _, behavior in ipairs(skill_data.equip_behavior) do
                    DCEI.RemoveBehaviorAsync(unit, behavior)

                    if equip then
                        -- apply behavior in next frame to avoid some stack mistacks
                        DCEI.TriggerAddTimerEventElapsed(function()
                            DCEI.ApplyBehaviorAsync(unit, behavior)
                        end, 0, false, true)
                    end
                end
            end
        end
    end
    -- legendary attributes
    ProcessEquipmentEquipBehaviorToHero(unit_data, unit, false)
    if equip then
        -- apply behavior in next frame to avoid some stack mistacks
        DCEI.TriggerAddTimerEventElapsed(function()
            ProcessEquipmentEquipBehaviorToHero(unit_data, unit, true)
        end, 0, false, true)
    end
end

-- process passive skills like add damage/atk speed/cooldown etc.
function RefreshHeroesEquipBehaviors()
    -- go through all equipped heroes, remove passive skills first then apply again, to ignore the effect of order or position to equip heroes
    for _, unit in pairs(player.heroes) do
        if unit and DCEI.UnitExists(unit) then
            local unit_type = DCEI.GetUnitType(unit)
            local unit_data = GetHeroDataFromUnitType(unit_type)

            PrcessHeroEquipBebaviors(unit, unit_data, true)
        end
    end
end

-- !SECTION

-- SECTION GEAR
function AttemptUpdateCardElixirDisplay(elixir)
    -- update gear card item component if active
    if ui.unit_card and ui.unit_card.unit_data then
        local unit_data = ui.unit_card.unit_data

        if ui.unit_card.is_upgrade_mode and ui.unit_card.item_component then
            local cost_data = unit_data:GetUpgradeCostData()

            ui.unit_card.item_component:Update(elixir, cost_data.elixir)
        end
    end
end

function AttemptUpdateItemMenuElixirDisplay(elixir)
    if ui.item_menu and ui.item_menu.is_active then
        ui.item_menu:UpdateItemMenuElixirDisplay(elixir)
    end
    if ui.unit_card and ui.unit_card.is_elixir then
        ui.unit_card:UpdateCount(elixir)
    end
end

function AttemptUpgradeGear(unit_data)
    local cost_data = unit_data:GetUpgradeCostData()
    local gold = GetPlayerGold()
    local elixir = GetPlayerElixir()
    local success = true

    if gold >= cost_data.gold and elixir >= cost_data.elixir then
        local spend_event = { type = "upgrade", target = "gear " .. tostring(unit_data.name) }
        SpendPlayerGold(cost_data.gold, spend_event)
        SpendPlayerElixir(cost_data.elixir)

        gear_upgrade_gold = gear_upgrade_gold + cost_data.gold
        bank.CommitGearUpgradeGold()

        UpgradeGear(unit_data)
        PlayGoldSpendSound()

        -- update card display
        if ui.unit_card and ui.unit_card.Update then
            ui.unit_card:Update()
        end

        -- update gear menu
        local index = unit_data.index
        ui.gear_menu.gear_slots[index]:Update(unit_data)

        if unit_data.slot and unit_data.slot > 0 then
            ui.gear_menu.equip_slots[unit_data.slot]:Update(unit_data)
            ui.loadout.gear:UpdateGearLevel(unit_data, unit_data.slot)
        end

        -- for token cap gear
        UpdateGemDisplay()
    elseif gold < cost_data.gold then
        -- need more gold prompt
        local function OnStore()
            ui.CloseUnitCard()
            ui.gear_menu.Close()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)

        success = false
    elseif elixir < cost_data.elixir then
        -- need more elixir
        ui.ShowFeedbackNotEnoughElixir()

        success = false
    end

    return success
end

function UpgradeGear(unit_data)
    if unit_data.slot > 0 then
        ProcessEquipBonuses(unit_data, false)
    end

    unit_data:LevelUp()

    if unit_data.slot > 0 then
        ProcessEquipBonuses(unit_data, true)
    end

    local level = unit_data:GetLevel()
    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, level)

    LogUserProperty(UserProps.gear_upgrades)
end

function AttemptPurchaseGear(unit_data)
    if not unit_data.is_purchased then
        if unit_data.unlock_wave then
            local message = DCEI.GetTranslationText(
                "campaign/unlock",
                { x = DCEI.GetTranslationText("campaign/wave/standard", { x = unit_data.unlock_wave }) }
            )
            ShowFeedbackMessage(message)
            return
        end
        if unit_data.locked then
            ShowFeedbackMessage(DCEI.GetTranslationText("ui/feedback/challenge_battle"))
            return
        end
    end
    local cost = unit_data.cost_adjust or unit_data.cost
    local gold = GetPlayerGold()

    if cost <= gold then
        local spend_event = { type = "buy", target = "gear " .. tostring(unit_data.name) }
        SpendPlayerGold(cost, spend_event)

        PurchaseGear(unit_data)
        PlayGoldSpendSound()
        EquipGear(unit_data)

        DCEI.DestroyFrame(ui.unit_card.purchase.button)
        ui.GenerateGearCardButtons(ui.unit_card)
    else
        local function OnStore()
            ui.CloseUnitCard()
            ui.gear_menu.Close()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreGold(offset, ui.super, OnStore)
    end
end

function PurchaseGear(unit_data)
    unit_data.is_purchased = true
    unit_data.level = unit_data.level or 1

    -- end tutorial on first purchase
    if tutorial.gear then
        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
        tutorial.gear = false
    end

    bank.CommitUnitData(unit_data, bank.bank_data.prop.level, unit_data.level)
    LogUserProperty(UserProps.gear_owned)
end

function EquipGear(unit_data, slot)
    slot = slot or FirstOpenGearSlot(unit_data)
    local gear_id = unit_data.name

    -- unequip gear currently in the slot, if any
    UnequipGearSlot(slot)

    -- unequip gear if currently equipped in another slot
    UnequipGear(unit_data)

    DCEI.LogMessage("> equipping " .. unit_data.name .. " to slot " .. slot)
    player.gear[slot] = gear_id
    unit_data.slot = slot

    bank.CommitUnitSlot(unit_data, slot)
    ProcessEquipBonuses(unit_data, true)

    ui.SetGearToEquipped(unit_data, true)
    ui.UpdateGearEquipSlot(unit_data, slot)
    ui.loadout.gear:UpdateGear(unit_data, slot)

    -- for token cap gear
    UpdateGemDisplay()

    Tutorial10_EquipGear_StepD()
end

function UnequipGear(unit_data)
    local gear_id = unit_data.name
    local slot = unit_data.slot
    local equipped_gear = player.gear[slot]

    if equipped_gear and equipped_gear ~= "" then
        DCEI.LogMessage("> unequipping " .. unit_data.name .. " from slot " .. slot)

        bank.ClearUnitSlot(bank.bank_data.units.gear, slot)

        unit_data.slot = 0
        player.gear[slot] = nil
        ProcessEquipBonuses(unit_data, false)
    end

    ui.SetGearToEquipped(unit_data, false)
    ui.UpdateGearEquipSlot(nil, slot)
    ui.loadout.gear:UpdateGear(nil, slot)

    -- for token cap gear
    UpdateGemDisplay()
end

function UnequipGearSlot(slot)
    local gear_id = player.gear[slot]

    if gear_id and gear_id ~= "" then
        UnequipGear(data.gear[gear_id])
    end
end

function ProcessEquipGear(unit_data)
    if unit_data.slot ~= 0 then
        UnequipGear(unit_data)
    else
        EquipGear(unit_data)
    end
    ui.unit_card.equip:Update()
end

function FirstOpenGearSlot(unit_data)
    if unit_data.type == "weapon" then
        -- weapons are always slot 1
        return 1
    else
        -- look for first open slot
        for n = 2, data.unlock.gear_slot_max do
            if not player.gear[n] then
                return n
            end
        end
    end

    -- if no open slots, always return final slot
    return data.unlock.gear_slot_max
end

-- !SECTION

-- SECTION Leaderboard
-- SECTION Player profile and rename
function OpenPlayerProfile()
    if ui.player_profile and DCEI.FrameExists(ui.player_profile.Frame) then
        DCEI.DestroyFrame(ui.player_profile.Frame)
    end
    local callbacks = {
        commit_player_profile = function()
            bank.CommitHellCoinShopData()
        end,
        set_favourite = function(id)
            player.hell_coin_shop.owned.favourite = id
        end,
    }
    ui.player_profile = ui.NewPlayerProfileFrame(callbacks)
end

-- RENAME
function OpenRenameMenu()
    if not ui.rename then
        ui.rename = ui.NewRenameFrame()
        DCEI.SetOnClickCallback(ui.rename.confirm_button, AttemptRename)
    end
    ui.rename.Open()
end

function AttemptRename()
    local new_name = DCEI.GetInputFrameText(ui.rename.input)
    if ui.leaderboard.username then
        if new_name == ui.leaderboard.username.name then
            return
        end
    end
    local character_limit = 15
    local cost = player.username.count and player.username.count > 0 and 5 or 0
    local gems = GetPlayerGems()

    function OnNameChange(count)
        UnblockInput()
        -- this only works when live
        if count then
            OnSuccess(count)
        else
            OnFailure()
        end
    end

    function OnSuccess(count)
        bank.CommitUsername(new_name, count)
        ui.leaderboard.username = player.username

        ui.leaderboard:ShowPlayerProfile()
        ui.rename.Close()

        -- success feedback
        ui.ShowFeedbackNameChangeSuccess()

        -- end username tutorial on username change
        if tutorial.username then
            if ui.prompt.tutorial then
                ui.prompt.tutorial.Close()
            end

            tutorial.username = false
            ui.glue.leader_menu:HidePip()
            ui.CheckOtherButtonPip()
        end

        -- update existing player usename labels on current scoreboard
        --local text = ui.NAME_SPACING .. new_name
        for _, item in ipairs(ui.leaderboard.player_name) do
            DCEI.SetUserInputTextFrameText(item.user_label, " " .. new_name)
        end
        if ui.player_profile and DCEI.FrameExists(ui.player_profile.Frame) then
            DCEI.SetUserInputTextFrameText(ui.player_profile.user_label, new_name)
        end

        if cost > 0 then
            local spend_event = { type = "rename", target = "rename" }
            SpendPlayerGems(cost, spend_event)
            PlayGoldSpendSound()
        end
    end

    function OnFailure()
        ui.ShowFeedbackOffline()
    end

    if gems < cost then
        local function OnStore()
            ui.rename.Close()
            if ui.player_profile and DCEI.FrameExists(ui.player_profile.Frame) then
                ui.player_profile:Close()
            end
            ui.leaderboard:Close()
            ui.loadout.Hide()
        end

        local offset = -356
        ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
    elseif new_name:len() > character_limit then
        -- warn if name too long
        ui.ShowFeedbackNameTooLong()
    elseif new_name and new_name ~= "" then
        BlockInput()
        -- set name if no issues
        DCEI.SetUserName(new_name, OnNameChange)
    else
        -- warn if invalid name
        ui.ShowFeedbackNameInvalid()
    end
end

-- !SECTION

-- SECTION leaderboard
-- LEADERBOARD SCORING
function WaveScoreWithTimeBonus()
    -- score bonus based off of time, gives about ~82 seconds of precision
    local score_bonus = 82000000 / (os.time() - 1500000000)
    score_bonus = math.floor(score_bonus * 1000000) / 1000000

    player.score = wave.data.GetMax() + score_bonus

    return player.score
end

function PackLeaderBoardPayload()
    local payload = {}

    payload.username = (player.username and player.username.name) or ui.PlayerHandle()
    payload.badge = GenerateHellCoinBadgeTable()
    -- DebugPayload(payload, ">")

    local encode = json.encode(payload)
    return encode
end

function StringIsEncodedTable(string)
    if string then
        local len = string:len()
        local first = string:sub(1, 1)
        local last = string:sub(len, len)

        -- encoded table strings look like `{"key":value}`
        return first == "{" and last == "}"
    end

    return false
end

function RepackLeaderBoardList(list)
    for n, user in ipairs(list.users) do
        -- user.payload = GenerateDummyPayload()
        local payload = user.payload

        if StringIsEncodedTable(payload) then
            -- new format is table (as of 0.0.66)
            local decoded = json.decode(payload)

            user.payload = decoded
        elseif payload and payload ~= "" then
            -- old format is string

            user.payload = { username = payload }
        end
        -- -- fake some badge data
        -- user.payload = {}
        -- user.payload.badge = {}
        -- user.payload.badge.fv = math.random( 1, 12 )
        -- user.payload.badge.list = {}
        -- for i = 1, 12 do
        --     local insert = math.random( 1, 100 ) > 50
        --     if insert then
        --         local key = "a"..i
        --         user.payload.badge.list[key] = true
        --     end
        -- end
    end

    return list
end

-- Cheater Detection
function TalentPointPossible()
    return GetPlayerLevel() - 1
end

function TalentPointTotal()
    local points = GetTalentPoints() + GetParagonPoints()
    for _, talent in ipairs(data.talents) do
        local upgrade_cost = GetParagonTalenUpgradeCost(talent)
        points = points + upgrade_cost
    end
    return points
end

function CheaterSaveFile()
    -- check talent points
    local possible_talents = TalentPointPossible()
    local paragon_level = GetParagonLevel()
    local actual_talents = TalentPointTotal()

    if actual_talents ~= (possible_talents + paragon_level) then
        bank.CommitCheater()
        return true
    end
end

function PostLeaderBoardScore(score, show_update, hell_mode_key)
    local type = hell_mode_key or "default"
    local payload = PackLeaderBoardPayload()
    local spent_dcoin = GetPlayerDcoinSpent()
    local required_wave = 3

    if score < required_wave then
        return
    end

    if CheaterSaveFile() then
        -- don't post score if save file is impossible
        return
    end

    local function ResultCallback(result)
        if result then
            -- update rank and store change
            ui.leaderboard.new_rank = result.rank
            ui.leaderboard.max_wave = hell_mode_key and score or wave.data.GetMax()

            if show_update then
                -- delay used for victory sequence pacing (also  so Wait() can be used in animation function)
                local delay = 0.25
                local simple = true
                DCEI.TriggerAddTimerEventElapsed(function()
                    ui.delay_display_leaderboard_update = false
                    if not ui.battlestats.is_active then
                        ui.ShowLeaderBoardUpdate(hell_mode_key)
                    else
                        ui.delay_display_leaderboard_update = true
                    end
                end, delay, true, simple)
            end
        end
    end

    -- attempt to get old ranking
    DCEI.GetLeaderBoardScore(type, function(result)
        if result then
            ui.leaderboard.old_rank = result.rank
        end

        -- then attempt to post score and update ranking
        DCEI.PostLeaderBoardScore(type, score, payload, spent_dcoin, ResultCallback)
    end)
end

function UpdateLeaderBoard_FetchCurrent(result_callback)
    local type = "default"
    local limit = 25

    DCEI.GetLeaderBoardSurroundingUsers(type, limit, result_callback)
end

function UpdateLeaderBoard_FetchTop()
    local type = "default"
    local skip = 0
    local limit = 50

    local function ResultCallback(result)
        if result then
            -- update leaderboard if fetch successful
            data.leaderboard.top = RepackLeaderBoardList(result)
            CheckLeaderBoardUpdate()
            ui.leaderboard.fetch_wave = true

            -- LogEvent(LogEvents.menu_open("LeaderBoard Fetch01 Success"))
        else
            -- LogEvent(LogEvents.menu_open("LeaderBoard Fetch01 Failure"))
        end
    end

    -- fetch top 50 players
    DCEI.GetLeaderBoardScoreList(type, skip, limit, ResultCallback)
end

function UpdateLeaderBoard_FetchHellModeCurrent(result_callback)
    local type = ui.leaderboard.season_id or "empty"
    local limit = 25

    DCEI.GetLeaderBoardSurroundingUsers(type, limit, result_callback)
end

function UpdateLeaderBoard_FetchHellModeTop()
    local type = ui.leaderboard.season_id or "empty"
    local skip = 0
    local limit = 50

    local function ResultCallback(result)
        if result then
            -- update leaderboard if fetch successful
            data.leaderboard.hell_mode_top = RepackLeaderBoardList(result)
            CheckLeaderBoardUpdate()
            ui.leaderboard.fetch_hellmode = true

            -- LogEvent(LogEvents.menu_open("LeaderBoard Fetch01 Success"))
        else
            -- LogEvent(LogEvents.menu_open("LeaderBoard Fetch01 Failure"))
        end
    end

    -- fetch top 50 players
    DCEI.GetLeaderBoardScoreList(type, skip, limit, ResultCallback)
end

function CheckLeaderBoardUpdate()
    local top = data.leaderboard.top
    local current = data.leaderboard.current
    local hell_mode_top = data.leaderboard.hell_mode_top
    local hell_mode_current = data.leaderboard.hell_mode_current

    if ui.leaderboard.hell_mode then
        if hell_mode_current and ui.leaderboard.selected_tab == 3 then
            -- generate leaderboard items
            ui.UpdateLeaderBoard(hell_mode_current.users, 3)
        elseif hell_mode_top and ui.leaderboard.selected_tab == 4 then
            ui.UpdateLeaderBoard(hell_mode_top.users, 4)
        end
    else
        if current and ui.leaderboard.selected_tab == 1 then
            -- generate leaderboard items
            ui.UpdateLeaderBoard(current.users, 1)
        elseif top and ui.leaderboard.selected_tab == 2 then
            ui.UpdateLeaderBoard(top.users, 2)
        end
    end
    -- show leaderboard data on success
    ui.leaderboard:ShowResults()
    -- track how many times users open leaderboard
    LogLeaderboardOpen()
end

function FetchLeaderBoardWave()
    -- load leaderboard data
    local function ResultCallback(result)
        if result then
            data.leaderboard.current = RepackLeaderBoardList(result)
            UpdateLeaderBoard_FetchTop()
        end
    end

    ui.leaderboard:ShowLoading()
    UpdateLeaderBoard_FetchCurrent(ResultCallback)
end

function FetchLeaderBoardHellMode()
    -- load leaderboard data
    local function ResultCallback(result)
        if result then
            data.leaderboard.hell_mode_current = RepackLeaderBoardList(result)
            UpdateLeaderBoard_FetchHellModeTop()
        end
    end

    ui.leaderboard:ShowLoading()
    UpdateLeaderBoard_FetchHellModeCurrent(ResultCallback)
end

function ShowSecondaryMenu()
    ui.glue.other_button_panel:Open()
    if not player.system.check_seceondary_menu or player.system.check_seceondary_menu == system.locked then
        player.system.check_seceondary_menu = system.active
        bank.CommitSystemState()
    end
end

function ShowLeaderBoard()
    if wave.data.is_active then
        -- prevent opening leaderboard after wave has started
        return
    end
    if ui.glue.other_button_panel.opening or ui.glue.other_button_panel.closing then
        return
    end
    ui.glue.other_button_panel:Close()

    -- hide other UI when leaderboard shown
    ui.loadout.Hide()
    ui.screen_mask.lower:Show()
    CloseActivePrompts()

    -- show other UI when leaderboard hide
    ui.leaderboard.on_hide = function()
        ui.loadout.Show()

        if ui.prompt.tutorial then
            ui.prompt.tutorial.Close()
        end
    end

    if tutorial.username then
        -- show username tutorial
        local offset = -513
        local text = tutorial.message.username
        ui.prompt.tutorial = ui.NewPromptMessage(text, offset)
    end

    ui.leaderboard.log_open = false
    ui.leaderboard.fetch_wave = false
    ui.leaderboard.fetch_hellmode = false

    DCEI.SetFrameActive(ui.leaderboard.mode_tabs[2].button, ui.leaderboard.unlock_hellmode)
    ui.leaderboard.player_name = {}
    if ui.leaderboard.hell_mode then
        FetchLeaderBoardHellMode()
    else
        FetchLeaderBoardWave()
    end

    DCEI.SetOnClickCallback(ui.leaderboard.mode_tabs[1].button, function()
        ui.leaderboard.hell_mode = false
        if not ui.leaderboard.fetch_wave then
            FetchLeaderBoardWave()
        end
        ui.leaderboard:ShowCurrentTab()
        CheckLeaderBoardUpdate()
    end)

    DCEI.SetOnClickCallback(ui.leaderboard.mode_tabs[2].button, function()
        ui.leaderboard.hell_mode = true
        if not ui.leaderboard.fetch_hellmode then
            FetchLeaderBoardHellMode()
        end
        ui.leaderboard:ShowCurrentTab()
        CheckLeaderBoardUpdate()
    end)

    local season_index = ui.leaderboard.season_index or 0
    local season_id = "HM_season" .. season_index
    local season_data = player.hell_mode[season_id]
    ui.leaderboard.hell_mode_event_callbacks = {
        player_score = season_data and season_data.total or 0,
        claim_list = player.hell_mode_event_claim_list,
        event_time = ui.leaderboard.timer and ui.leaderboard.timer:GetTimeRemaining() or 0,
        claim_event_reward = ClaimHellModeEventReward,
    }
    if not ui.leaderboard.season_index or ui.leaderboard.season_not_ready then
        ui.leaderboard.hell_mode_event_callbacks.not_ready = true
        ui.leaderboard.hell_mode_event_callbacks.event_time = ui.leaderboard.hell_mode_event_callbacks.event_time
    end
    -- LogEvent(LogEvents.menu_open("LeaderBoard Open"))
    ui.leaderboard.username = player.username
    ui.leaderboard:Show()

    local unlock_rank_reward = CheckUnlockHellModeRankReward()
    DCEI.SetFrameActive(ui.leaderboard.shop_shortcut.Frame, unlock_rank_reward)
    DCEI.SetFrameActive(ui.leaderboard.rank_reward.Frame, unlock_rank_reward)
end

-- !SECTION
-- !SECTION

-- SECTION BATTLE STATS
function UpdateBattleData()
    if wave.data.unfinished then
        ResetBattleData()
    else
        ui.battlestats.data = ProcessBattleData()
        ui.UpdateBattleStats()
    end
end

function GetUnitNameIcon(name)
    local icon = "icon_enhance_potiondealer_unidentified_potion"
    local words = {}
    for word in name:gmatch("%w+") do
        table.insert(words, word)
    end
    local type = words[1]

    if name == DCEI.Unit("Castle Archer") then
        -- hardcode archer icon
        icon = "talent_archer_damage"
    elseif type == "Hero" then
        -- get hero icon
        local hero_data = GetHeroCurrentClassDataFromBaseClassName(name)
        icon = hero_data.portrait
    elseif type == "Tower" then
        -- get tower icon
        local tower_data = GetTowerCurrentClassDataFromBaseClassName(name)
        icon = tower_data.portrait
    elseif type == "Defense" then
        -- get defense icon
        icon = data.defenses[name].portrait
    elseif type == "Summon" then
        icon = data.summons[name].icon
    end

    return icon
end

function ResetBattleData()
    ui.battlestats.data = nil
    ui.UpdateBattleStats()
    DCEI.ResetCardDamageStats()
end

function ProcessBattleData()
    local stats = DCEI.GetCardDamageStats()
    local damage_total = 0
    local battle_data = {}

    for unit_name, damage_value in pairs(stats) do
        damage_total = damage_total + damage_value

        table.insert(battle_data, { name = unit_name, damage = damage_value })
    end

    DCEI.ResetCardDamageStats()
    function sort_by_damage(a, b)
        return a.damage > b.damage
    end

    table.sort(battle_data, sort_by_damage)

    for k, unit_data in ipairs(battle_data) do
        local damage = unit_data.damage
        unit_data.damage_percent = damage / damage_total
        unit_data.icon = GetUnitNameIcon(unit_data.name)
        -- DCEI.LogMessage(">> " .. k .. ". " .. unit_data.name .. ": " .. unit_data.damage)
    end

    -- added earned data to battle_data that stores wave.data.earned
    battle_data.earned = math.floor(wave.data.earned)
    battle_data.mine = wave.data.mine_income
    local colony_income = levels.colony:Income()
    local colony_cooldown = levels.colony:Cooldown()
    battle_data.colony = math.floor(colony_income / colony_cooldown * wave.data.elapsed)
    battle_data.total_income = battle_data.earned + battle_data.mine + battle_data.colony

    battle_data.tip = CheckDisplayTipOnBattleStats()
    return battle_data
end

function ShowBattleStats()
    if wave.data.is_active then
        -- prevent battle stats opening if wave started
        return
    end
    if ui.glue.other_button_panel.opening or ui.glue.other_button_panel.closing then
        return
    end
    ui.glue.other_button_panel:Close()

    -- hide other UI when leaderboard shown
    ui.loadout.Hide()
    ui.screen_mask.lower:Show()
    CloseActivePrompts()

    -- show other UI when leaderboard hide
    ui.battlestats.on_hide = function()
        ui.loadout.Show()
    end

    -- LogEvent(LogEvents.menu_open("BattleStats Open"))
    ui.battlestats:Show()
end

-- !SECTION

-- SECTION Cross Promote menu
function ShowSocial()
    if wave.data.is_active then
        -- prevent social menu from opening if wave started
        -- return
        HideHeroSkillBars()
        SetGameSpeed(0)
    end
    ui.loadout.Hide()

    if not ui.social then
        ui.social = ui.NewSocialMenu()
        --HookupSocialLinks()
        -- show loadout UI on hide
        ui.social.on_hide = function()
            ui.loadout.Show()
            if wave.data.is_active then
               -- ShowHeroSkillBars()
                SetGameSpeed(player.speed)
            else
                ClearGameSpeed()
            end
        end
    end

    ui.social:Show()
    -- log social menu open
    LogEvent(LogEvents.social_menu_open)
end

function HookupSocialLinks()
    DCEI.SetOnClickCallback(ui.social.buttons["discord"], function()
        LogSocialLinkClickthrough("discord")
        DCEI.OpenUrl("https://discord.gg/HfWgMgV")
    end)
    DCEI.SetOnClickCallback(ui.social.buttons["instagram"], function()
        LogSocialLinkClickthrough("instagram")
        DCEI.OpenUrl("https://www.instagram.com/wildcastle3d/")
    end)
    DCEI.SetOnClickCallback(ui.social.buttons["reddit"], function()
        LogSocialLinkClickthrough("reddit")
        DCEI.OpenUrl("https://reddit.com/r/WildCastle/")
    end)
    DCEI.SetOnClickCallback(ui.social.buttons["facebook"], function()
        LogSocialLinkClickthrough("facebook")
        DCEI.OpenUrl("https://www.facebook.com/wildcastle3d/")
    end)

    DCEI.SetOnClickCallback(ui.social.buttons["android"], function()
        LogSocialLinkClickthrough("android")
        DCEI.OpenUrl("https://play.google.com/store/apps/details?id=com.wildsky.wildcastle")
    end)

    DCEI.SetOnClickCallback(ui.social.buttons["wild sky"], function()
        LogSocialLinkClickthrough("wild sky")
        LogEvent(LogEvents.wild_sky_clickthrough)
        DCEI.OpenUrl(
            "https://play.google.com/store/apps/details?id=com.wildsky.td&referrer=utm_source%3Dwildcastle%26utm_medium%3Dsocial"
        )
    end)
end

-- !SECTION

-- SECTION Hell mode
function SetHellModeAttributeText(layout, attri)
    local text = layout.value
    if attri.multiplier then
        text = layout.value * attri.multiplier
    end
    if attri.percent then
        text = text .. "%"
    end
    DCEI.SetTextFrameText(layout.Number, text)
end

function UpdateAttributeDescriptionText(layout, attri)
    local text = ""
    local value = layout.value
    if attri.multiplier then
        value = layout.value * attri.multiplier
    end
    if attri.description then
        text = attri.description
        text = text:gsub("VAL", value)
    elseif attri.description_func then
        text = attri.description_func(layout.value)
    end

    text = text:gsub("%[%[", "<color=green>")
    text = text:gsub("%]%]", "</color>")
    text = text:gsub("%[", "<color=yellow>")
    text = text:gsub("%]", "</color>")

    DCEI.SetTextFrameText(ui.hell_mode.Description, text)
end

function HookupHellModeAttribute(layout, attri)
    local Item = gmui.Layout:New({
        frame = layout[attri.name],
    })

    Item.value = attri.default or 0
    Item.min = attri.min or 0

    DCEI.SetFrameSize(Item.Fill, Item.value / attri.max * attri.max_width, 40)
    SetHellModeAttributeText(Item, attri)
    DCEI.SetFrameImage(Item.Icon, attri.icon)
    if attri.sub_icon then
        DCEI.SetFrameActive(Item.SubIcon, true)
        DCEI.SetFrameImage(Item.SubIcon, attri.sub_icon)
        if attri.sub_icon_offset_h then
            DCEI.SetFrameHorizontalOffsetInParent(Item.SubIcon, attri.sub_icon_offset_h)
        end
    end

    ui.HookUpHoldButtonFunctions(Item, Item.Dec, function()
        local increment = Item.tick_count == 1 and 1 or (attri.increment + Item.tick_count)
        local success = true
        for i = 1, increment do
            if Item.value <= Item.min then
                Item.value = Item.min
                success = false
                break
            end
            Item.value = Item.value - 1
            local cost = math.huge
            if attri.cost.increment_cost then
                cost = attri.cost.start_cost + Item.value * attri.cost.increment_cost
            else
                for _, v in ipairs(attri.cost) do
                    if Item.value >= v.start_level then
                        cost = v.cost
                        break
                    end
                end
            end
            layout.gem = layout.gem + cost
        end

        layout:UpdateGem()
        local fraction = Item.value / attri.max * attri.max_width
        DCEI.SetFrameSize(Item.Fill, fraction, 40)
        SetHellModeAttributeText(Item, attri)
        UpdateAttributeDescriptionText(Item, attri)
        return success
    end, true)

    ui.HookUpHoldButtonFunctions(Item, Item.Inc, function()
        local increment = Item.tick_count == 1 and 1 or (attri.increment + Item.tick_count)
        local success = true
        for i = 1, increment do
            if Item.value >= attri.max then
                Item.value = attri.max
                success = false
                break
            end
            local cost = math.huge
            if attri.cost.increment_cost then
                cost = attri.cost.start_cost + Item.value * attri.cost.increment_cost
            else
                for _, v in ipairs(attri.cost) do
                    if Item.value >= v.start_level then
                        cost = v.cost
                        break
                    end
                end
            end
            if cost <= layout.gem then
                Item.value = Item.value + 1
                layout.gem = layout.gem - cost
            else
                local text = DCEI.GetTranslationText(
                    "ui/feedback/need_generic",
                    { x = "<size=90%><sprite name=gears></size>" }
                ) .. ": <color=green>" .. cost
                ui.prompt.tutorial = ui.NewPromptMessageClosable(text, -490)
                success = false
                break
            end
        end

        layout:UpdateGem()
        local fraction = Item.value / attri.max * attri.max_width
        DCEI.SetFrameSize(Item.Fill, fraction, 40)
        SetHellModeAttributeText(Item, attri)
        UpdateAttributeDescriptionText(Item, attri)
        return success
    end, true)

    DCEI.SetOnClickCallback(Item.Icon, function()
        UpdateAttributeDescriptionText(Item, attri)
    end)

    attri.item = Item
end

function SetHellModeAttributes()
    -- Reset hero tower and defense level
    local level = data.hell_mode.attributes["HeroLevel"].item.value
    for _, unit in pairs(player.towers) do
        local name = DCEI.GetUnitType(unit)

        local unit_data = data.towers[name]
        if level > unit_data.max_level then
            unit_data:LevelUp(unit_data.max_level)
        else
            unit_data:LevelUp(level)
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    for _, unit in pairs(player.heroes) do
        local name = DCEI.GetUnitType(unit)

        local unit_data = data.heroes[name]
        if level > unit_data.max_level then
            unit_data:LevelUp(unit_data.max_level)
        else
            unit_data:LevelUp(level)
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    for _, unit in pairs(player.defenses) do
        local name = DCEI.GetUnitType(unit)

        local unit_data = data.defenses[name]
        if level > unit_data.max_level then
            unit_data:LevelUp(unit_data.max_level)
        else
            unit_data:LevelUp(level)
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    -- Castle
    if not ui.hell_mode.battle_active then
        local behavior = DCEI.Behavior("Player Castle Upgrade")
        local count = GetCastleLevel() - 1
        AddCastleUpgrade(behavior, -count)
        local behavior = DCEI.Behavior("Player Castle Basic Hell Mode")
        AddCastleUpgrade(behavior, 1)
    end

    local hp = DCEI.GetHealth(CASTLE_UNIT)
    local mana = DCEI.GetMana(CASTLE_UNIT)
    local behavior = DCEI.Behavior("Player Castle Upgrade Hell Mode")
    if data.castle_upgrades[behavior] and data.castle_upgrades[behavior].count > 0 then
        AddCastleUpgrade(behavior, -data.castle_upgrades[behavior].count)
    end
    if data.hell_mode.attributes["CastleLevel"].item.value > 1 then
        AddCastleUpgrade(behavior, data.hell_mode.attributes["CastleLevel"].item.value - 1)
    end
    if ui.hell_mode.battle_active then
        DCEI.SetHealthValueAsync(CASTLE_UNIT, hp)
        DCEI.SetManaValueAsync(CASTLE_UNIT, mana)
    end

    -- Archer
    local behavior = DCEI.Behavior("Archer AttackPower")
    local attack_upgrade = data.archer_upgrades[behavior]
    local attack_value = attack_upgrade and attack_upgrade.count or 0
    AddArcherUpgrade(behavior, -attack_value)

    local new_val = GetArcherBaseAttackPower(data.hell_mode.attributes["ArcherLevel"].item.value)
    if new_val > 0 then
        AddArcherUpgrade(behavior, math.floor(new_val))
    end

    -- bonus gold
    player.attributes.hellmode_gold = data.hell_mode.attributes["GoldBonus"].item.value
        * data.hell_mode.attributes["GoldBonus"].multiplier
        * 0.01

    -- element attributes
    for i = 7, #data.hell_mode.attributes do
        local behavior = data.hell_mode.attributes[i].behavior
        player.attributes.hellmode_attributes[behavior] = data.hell_mode.attributes[i].item.value
    end
end

function StartHellModeBattle()
    if
        player.system.first_play_hell_mode
        and player.system.first_play_hell_mode == system.active
        and not ui.hell_mode.battle_active
    then
        if GetHellModeFreeCount() > 0 then
            AddHellModeFreeCount(-1)
            bank.CommitHellModeFreeCount()
        else
            local cost = 30
            local gems = GetPlayerGems()

            if cost <= gems then
                local spend_event = { type = "buy", target = "Hell Mode Count" }
                SpendPlayerGems(cost, spend_event)
                PlayGoldSpendSound()
            else
                local function OnStore()
                    ui.hell_mode:Close()
                    ui.loadout.Hide()
                end

                local offset = -520
                ui.ShowFeedbackNeedMoreTokens(offset, ui.super, OnStore)
                return
            end
        end
    end
    HideBlessCrystal()
    player.attributes.hellmode_attributes = {}
    ui.hell_mode:Hide()
    if not ui.hell_mode.battle_active then
        local wave_data = {
            mode = "hell_mode",
            rating = 2000,
        }
        WaveStartActions(wave_data)
        ui.hell_mode.battle_active = true
        ui.hell_mode.gem_kill = 0
        ui.CreateHellModeBattleUI()
        ui.HellModePanel.CASTLE_UNIT = CASTLE_UNIT
        DCEI.SetOnDownCallback(ui.HellModePanel.info_button, function()
            SetGameSpeed(0)
            DCEI.SetFrameActive(ui.HellModePanel.info_panel.Frame, true)
        end)

        DCEI.SetOnUpCallback(ui.HellModePanel.info_button, function()
            SetGameSpeed(player.speed)
            DCEI.SetFrameActive(ui.HellModePanel.info_panel.Frame, false)
        end)
        SetHellModeScore(0)
        local text = DCEI.GetTranslationText("ui/hell_mode/score") .. ": " .. GetHellModeScore()
        DCEI.SetTextFrameText(ui.HellModePanel.score, text)

        if not player.system.first_play_hell_mode or player.system.first_play_hell_mode ~= system.active then
            TutorialHellMode04Battle()
        end
    end
end

function ExitHellModeBattle()
    ui.DestoryHellModeBattleUI()
    -- Reset hero tower and defense level
    for _, unit in pairs(player.towers) do
        local name = DCEI.GetUnitType(unit)
        local unit_data = data.towers[name]

        local key = bank.UnitDataKey(unit_data, bank.bank_data.prop.level)
        local level = DCEI.Save.Get(key)

        if level then
            if level > unit_data.max_level then
                unit_data:LevelUp(unit_data.max_level)
            else
                unit_data:LevelUp(level)
            end
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    for _, unit in pairs(player.heroes) do
        local name = DCEI.GetUnitType(unit)
        local unit_data = data.heroes[name]

        local key = bank.UnitDataKey(unit_data, bank.bank_data.prop.level)
        local level = DCEI.Save.Get(key)

        if level then
            if level > unit_data.max_level then
                unit_data:LevelUp(unit_data.max_level)
            else
                unit_data:LevelUp(level)
            end
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    for _, unit in pairs(player.defenses) do
        local name = DCEI.GetUnitType(unit)
        local unit_data = data.defenses[name]

        local key = bank.UnitDataKey(unit_data, bank.bank_data.prop.level)
        local level = DCEI.Save.Get(key)

        if level then
            if level > unit_data.max_level then
                unit_data:LevelUp(unit_data.max_level)
            else
                unit_data:LevelUp(level)
            end
        end

        ApplyBaseBehaviorsToUnit(unit_data, unit)
    end

    -- Castle
    local behavior = DCEI.Behavior("Player Castle Upgrade Hell Mode")
    if data.hell_mode.attributes["CastleLevel"].item.value > 1 then
        AddCastleUpgrade(behavior, -(data.hell_mode.attributes["CastleLevel"].item.value - 1))
    end
    local behavior = DCEI.Behavior("Player Castle Basic Hell Mode")
    AddCastleUpgrade(behavior, -1)
    local behavior = DCEI.Behavior("Player Castle Upgrade")
    local count = GetCastleLevel() - 1
    AddCastleUpgrade(behavior, count)

    -- Archer
    local attack_value = GetArcherBaseAttackPower()
    local new_val = GetArcherBaseAttackPower(data.hell_mode.attributes["ArcherLevel"].item.value)

    if new_val > 0 then
        AddArcherUpgrade("Archer AttackPower", -math.floor(new_val))
    end
    AddArcherUpgrade("Archer AttackPower", math.floor(attack_value))

    player.attributes.hellmode_gold = 0

    if ui.hell_mode.atk_up then
        RemoveHellModeItemAttackUp()
    end
    if ui.hell_mode.def_up then
        RemoveHellModeItemDefenseUp()
    end

    if ui.hell_mode and DCEI.FrameExists(ui.hell_mode.Frame) then
        DCEI.DestroyFrame(ui.hell_mode.Frame)
        ui.hell_mode = nil
    end

    ui.glue.hell_mode_button:HidePip()
    if not player.system.first_play_hell_mode or player.system.first_play_hell_mode ~= system.active then
        player.system.first_play_hell_mode = system.active
        bank.CommitSystemState()
        local event = LogEvents.FirstPlayHellMode()
        LogEvent(event)
    end
end

function ShowHellModeMenu()
    if wave.data.is_active then
        -- prevent menu from opening if wave started
        return
    end
    -- hide loadout UI on show
    ui.loadout.Hide()

    if ui.hell_mode and DCEI.FrameExists(ui.hell_mode.Frame) then
        DCEI.DestroyFrame(ui.hell_mode.Frame)
    end
    ui.hell_mode = ui.NewHellModeMenu()
    ui.hell_mode.gem = data.hell_mode.start_gem
    DCEI.SetTextFrameText(ui.hell_mode.Description, "")
    for _, attri in ipairs(data.hell_mode.attributes) do
        HookupHellModeAttribute(ui.hell_mode, attri)
    end
    ui.hell_mode:UpdateGem()

    DCEI.SetOnClickCallback(ui.hell_mode.Battle, StartHellModeBattle)

    DCEI.SetOnClickCallback(ui.hell_mode.LeaderboardShortcutButton, function()
        if ui.hell_mode.battle_active then
            return
        end
        ui.hell_mode:Close()
        ui.leaderboard.hell_mode = true
        ShowLeaderBoard()
    end)

    ui.hell_mode.on_hide = function()
        for _, attri in ipairs(data.hell_mode.attributes) do
            attri.item.min = attri.item.value
        end
        SetHellModeAttributes()
        for behavior, count in pairs(player.attributes.hellmode_attributes) do
            for _, monster in pairs(wave.data.units) do
                local stack = DCEI.GetUnitBehaviorStackCount(monster, behavior)
                if stack > 0 then
                    DCEI.RemoveBehavior(monster, behavior, stack)
                end
                if count > 0 then
                    DCEI.ApplyBehaviorToSelf(monster, behavior, count)
                end
            end
        end
        if ui.HellModePanel then
            if ui.hell_mode.gem > 0 then
                DCEI.SetFrameActive(ui.HellModePanel.gem_icon.ItemCount, true)
                DCEI.SetTextFrameText(ui.HellModePanel.gem_icon.ItemCount, ui.hell_mode.gem)
            else
                DCEI.SetFrameActive(ui.HellModePanel.gem_icon.ItemCount, false)
            end
        end
        SetGameSpeed(player.speed)
    end
    -- show loadout UI on hide
    ui.hell_mode.on_close = function()
        ui.loadout.Show()
        ClearGameSpeed()
    end

    ui.hell_mode.on_show = function()
        if ui.hell_mode.battle_active then
            DCEI.SetFrameActive(ui.hell_mode.LeaderboardShortcut, false)
            DCEI.SetFrameActive(ui.hell_mode.FreeCountRefreshContainer, false)
            local hide_cost = true
            ui.hell_mode:UpdateBattleButton(GetHellModeFreeCount(), hide_cost)
            if not player.system.first_play_hell_mode or player.system.first_play_hell_mode ~= system.active then
                TutorialHellMode06BattleEnd()
            end
        end
        SetGameSpeed(0)
    end

    ui.hell_mode:Show()
    ui.hell_mode:UpdateBattleButton(GetHellModeFreeCount())
    local time_remaining = ui.calendar.timer and ui.calendar.timer:GetTimeRemaining() or 0
    ui.hell_mode:UpdateTimer(os.date("!%H:%M:%S", time_remaining))

    if not player.system.first_play_hell_mode or player.system.first_play_hell_mode == system.locked then
        player.system.first_play_hell_mode = system.unlocked
        TutorialHellMode01()
        local hide_cost = true
        ui.hell_mode:UpdateBattleButton(GetHellModeFreeCount(), hide_cost)
        DCEI.SetFrameActive(ui.hell_mode.FreeCountRefreshContainer, false)
    end
end

function CreateWaveExitConfirmMenu()
    SetGameSpeed(0)
    ui.screen_mask.super:Show()
    local menu = gmui.Layout:New({
        parent = ui.super,
        name = "HellMode/ExitConfirm",
    })
    DCEI.AnimateFrameScale(menu.Frame, { x = 0.5, y = 0.5, z = 0.5 }, { x = 1, y = 1, z = 1 }, 0.3, "OutBack")
    DCEI.SetTextFrameText(menu.Title, DCEI.GetTranslationText("ui/hell_mode/pause"))
    DCEI.SetTextFrameText(menu.ResumeLabel, DCEI.GetTranslationText("ui/hell_mode/continue"))
    DCEI.SetTextFrameText(menu.ExitLabel, DCEI.GetTranslationText("ui/exit"))

    function menu:Close()
        ui.screen_mask.Hide()
        DCEI.DestroyFrame(menu.Frame)
        ui.wave_exit_confirm = nil
    end

    function menu:Resume()
        menu:Close()
        SetGameSpeed(player.speed)
    end

    function menu:Exit()
        menu:Close()
        if wave.data.mode == "hell_mode" then
            WaveEndActions(false, true)
        else
            WaveEndActions()
        end
    end

    DCEI.SetOnClickCallback(menu.ResumeButton, function()
        menu:Resume()
    end)

    DCEI.SetOnClickCallback(menu.ExitButton, function()
        menu:Exit()
    end)
    ui.wave_exit_confirm = menu
end

function HellModeKillGemCooldown()
    if ui.hell_mode then
        ui.hell_mode.gem_kill = ui.hell_mode.gem_kill
            + 1
                * (1 + data.hell_mode.attributes["GemCD"].item.value * data.hell_mode.attributes["GemCD"].multiplier * 0.01)
        if ui.hell_mode.gem_kill >= data.hell_mode.gem_refresh_kill then
            ui.hell_mode.gem_kill = ui.hell_mode.gem_kill - data.hell_mode.gem_refresh_kill
            ui.hell_mode.gem = ui.hell_mode.gem
                + data.hell_mode.gem_refresh_amount
                + data.hell_mode.attributes["GemPlus"].item.value * data.hell_mode.attributes["GemPlus"].multiplier
            DCEI.SetFrameActive(ui.HellModePanel.gem_icon.ItemCount, true)
            DCEI.SetTextFrameText(ui.HellModePanel.gem_icon.ItemCount, ui.hell_mode.gem)
            ui.hell_mode:UpdateGem()
        end
        local frac = ui.hell_mode.gem_kill / data.hell_mode.gem_refresh_kill
        DCEI.SetFrameImageFillAmount(ui.HellModePanel.gem_icon.Countdown, frac)
    end
end

function HellModeAddScore(value)
    if ui.HellModePanel then
        AddHellModeScore(value)
        local text = DCEI.GetTranslationText("ui/hell_mode/score") .. ": " .. GetHellModeScore()
        DCEI.SetTextFrameText(ui.HellModePanel.score, text)
    end
end

function RemoveHellModeItemAttackUp()
    local behavior = DCEI.Behavior("Hellmode Attack Up")
    ApplyArcherBehavior(behavior, -1)
    AddSummonUpgrade(behavior, -1)
    for k, unit in pairs(group_summons) do
        ModifyUnitBehavior(unit, behavior, -1)
    end
    ApplyTowerBehavior(behavior, -1)
    ApplyHeroBehavior(behavior, -1)
    for _, unit in pairs(player.defenses) do
        ModifyUnitBehavior(unit, behavior, -1)
    end
end

function HellModeItemAttackUp(duration)
    ui.hell_mode.atk_up = true
    local wave_instance = wave.data.instance
    local behavior = DCEI.Behavior("Hellmode Attack Up")
    ApplyArcherBehavior(behavior, 1)
    AddSummonUpgrade(behavior, 1)
    for k, unit in pairs(group_summons) do
        ModifyUnitBehavior(unit, behavior, 1)
    end
    ApplyTowerBehavior(behavior, 1)
    ApplyHeroBehavior(behavior, 1)
    for _, unit in pairs(player.defenses) do
        ModifyUnitBehavior(unit, behavior, 1)
    end
    DCEI.TriggerAddTimerEventElapsed(function()
        if wave_instance == wave.data.instance then
            if ui.hell_mode then
                ui.hell_mode.atk_up = false
            end
            RemoveHellModeItemAttackUp()
        end
    end, duration, false, true)
end

function RemoveHellModeItemDefenseUp()
    local behavior = "Hellmode Defense Up"
    ApplyCastleWallBehavior(behavior, -1)
end

function HellModeItemDefenseUp(duration)
    ui.hell_mode.def_up = true
    local wave_instance = wave.data.instance
    local behavior = "Hellmode Defense Up"
    ApplyCastleWallBehavior(behavior, 1)
    DCEI.TriggerAddTimerEventElapsed(function()
        if wave_instance == wave.data.instance then
            if ui.hell_mode then
                ui.hell_mode.def_up = false
            end
            RemoveHellModeItemDefenseUp()
        end
    end, duration, false, true)
end

function HellModeItemResetCooldown()
    for _, unit in pairs(player.heroes) do
        local unit_data = GetHeroDataFromUnitType(DCEI.GetUnitType(unit))
        ReduceHeroSkillCooldown(unit_data, 9999)
    end
end

function HellModeMonsterFreeze()
    core.CreateEffect("Hero BlueMage Skill NorthWind FX Impact", 23.5, 24)
    DCEI.PlaySound(DCEI.Sound("iceshard_launch"))
    DCEI.TriggerAddTimerEventElapsed(function()
        core.CreateEffect("Hero BlueMage Skill NorthWind FX Impact", 25, 27.5)
    end, 0.125, false, true)
    DCEI.TriggerAddTimerEventElapsed(function()
        core.CreateEffect("Hero BlueMage Skill NorthWind FX Impact", 23, 31)
    end, 0.25, false, true)
    DCEI.TriggerAddTimerEventElapsed(function()
        core.CreateEffect("Hero BlueMage Skill NorthWind FX Impact", 24.5, 34.5)
        local behavior_name = DCEI.Behavior("Hero BlueMage Skill Blizzard Freeze")
        for _, unit in pairs(wave.data.units) do
            DCEI.ApplyBehaviorToSelf(unit, behavior_name)
            if DCEI.GetUnitBehaviorStackCount(unit, behavior_name) > 0 then
                DCEI.SetBehaviorDuration(unit, behavior_name, 15)
            end
        end
    end, 0.375, false, true)
end

function HellModeMonsterKillAll()
    for i = 24, 33, 3 do
        local x1 = 22 + math.random() * 4
        local x2 = 22 + math.random() * 4
        local x_offset1 = 1 - math.random() * 2
        local x_offset2 = 1 - math.random() * 2
        core.CreateEffect("HellMode Thunder FX Impact", x1, i + x_offset1)
        core.CreateEffect("HellMode Thunder FX Impact", x2, i + x_offset2)
        DCEI.Wait(0.125)
    end
    local units = {}
    for _, unit in pairs(wave.data.units) do
        table.insert(units, unit)
    end
    -- create effect
    for _, unit in pairs(units) do
        DCEI.KillUnit(unit)
    end
end

function HellModeCastleFullHP()
    local full_health = DCEI.GetMaxHealth(CASTLE_UNIT)
    DCEI.AddHealth(CASTLE_UNIT, full_health)
    CASTLE_UNIT_INVINC = true
    DCEI.TriggerAddTimerEventElapsed(function()
        CASTLE_UNIT_INVINC = false
    end, 0.0125, false, true)
end

function HellModeCastleFullMana()
    local full_mana = DCEI.GetMaxMana(CASTLE_UNIT)
    DCEI.AddMana(CASTLE_UNIT, full_mana)
end

data.hell_mode.item_functions = {
    ["atk_up"] = HellModeItemAttackUp,
    ["defense_up"] = HellModeItemDefenseUp,
    ["cooldown"] = HellModeItemResetCooldown,
    ["freeze"] = HellModeMonsterFreeze,
    ["thunder"] = HellModeMonsterKillAll,
    ["hp"] = HellModeCastleFullHP,
    ["mana"] = HellModeCastleFullMana,
}

-- SECTION Hell mode leaderboard
function GetHellModeSeasonId(time_stamp)
    local time_elapsed = time_stamp - HELL_MODE_SEASON_START
    if time_elapsed > 0 then
        local season_count = math.floor(time_elapsed / SEASON_DURATION)
        return "HM_season" .. season_count, season_count
    end
    return nil, nil
end

function GetHellModeSeasonNameString(season_id)
    local season_start_date = HELL_MODE_SEASON_START + season_id * SEASON_DURATION
    local season_end_date = season_start_date + SEASON_DURATION - SEASON_INTERVAL
    local season_name = "Season_"
        .. os.date("%m/%d/%Y", season_start_date)
        .. "_"
        .. os.date("%m/%d/%Y", season_end_date)

    return season_name
end

function CheckHellModeSeasonScore()
    local score = GetHellModeScore()

    player.hell_mode.count = player.hell_mode.count or 0
    player.hell_mode.count = player.hell_mode.count + 1

    player.hell_mode.best = player.hell_mode.best or 0
    if score > player.hell_mode.best then
        player.hell_mode.best = score
    end
    bank.CommitHellModeData()

    -- log player deck if score is high enough
    if score > 1000 then
        LogEvent(LogEvents.hell_mode_end)
    end

    -- check season id
    local function ResultCallback(time_stamp)
        if time_stamp then
            if ui.leaderboard.season_not_ready then
                -- DCEI.LogError("season_not_ready")
                return
            end

            local key, season_id = GetHellModeSeasonId(time_stamp)
            if key ~= nil then
                if not player.hell_mode[key] then
                    player.hell_mode[key] = {
                        id = season_id,
                        no_log = true,
                        best = 0,
                        total = 0,
                    }
                end
                -- change hell mode season reward so need to be ture in initialize
                -- for players that update game late, they will get chance to claim season reward
                if season_id > 12 then
                    player.hell_mode[key].not_claim = true
                end
                if score > player.hell_mode[key].best then
                    player.hell_mode[key].best = score
                end
                player.hell_mode[key].total = player.hell_mode[key].total + score
                bank.CommitHellModeData()
                local show_update = true
                PostLeaderBoardScore(player.hell_mode[key].total, show_update, key)
            end
        end
    end

    AttemptGetServerTime(ResultCallback)
end

function FetchHellModeSeasonRewardOnGameStart(time_stamp)
    if time_stamp then
        if time_stamp < HELL_MODE_SEASON_START then
            return
        end
        AttemptClaimHellModeSeasonReward(time_stamp)
    end
end

function GetHellModeSeasonReward(result)
    local rank = result.rank
    local reward
    for _, item in ipairs(data.hell_mode_seasonal_reward) do
        if item.rank_max then
            if rank <= item.rank_max then
                reward = item.reward
                break
            end
        else
            -- last rank
            reward = item.reward
        end
    end
    if reward then
        AddHellCoin(reward.count)
        bank.CommitHellCoin()
        ui.CreateHellModeSeasonRewardPrompt(rank, reward)
    end
end

function AttemptClaimHellModeSeasonReward(time_stamp)
    -- + SEASON_INTERVAL as season end on Tuesday and next season starts on Wednesday
    local current_season_key, current_season_id = GetHellModeSeasonId(time_stamp + SEASON_INTERVAL)
    if current_season_key and current_season_id then
        for key, value in pairs(player.hell_mode) do
            if key ~= "best" and key ~= "count" and key ~= "daycount" then
                if key ~= current_season_key and value.not_claim then
                    DCEI.GetLeaderBoardScore(key, function(result)
                        if result then
                            GetHellModeSeasonReward(result)
                            value.not_claim = false
                            -- push season properties
                            if value.no_log and value.id then
                                local season_name = GetHellModeSeasonNameString(value.id)
                                local user_prop = UserProps.HellModeRecord(season_name, value.total)
                                LogUserProperty(user_prop)
                                value.no_log = false
                            end
                            bank.CommitHellModeData()
                        end
                    end)
                end
            end
        end
    end
end

function ClaimHellModeEventReward(index)
    if ui.leaderboard.season_index then
        local season_id = "HM_season" .. ui.leaderboard.season_index
        local player_score = player.hell_mode[season_id] and player.hell_mode[season_id].total or 0
        local event_item = data.hell_mode_event_reward[index]
        local required_score = event_item.required_score
        local key = "reward" .. index
        if player_score >= required_score and not player.hell_mode_event_claim_list[key] then
            player.hell_mode_event_claim_list[key] = true
            ui.hellmode_event_reward:UpdateRewardList()
            local reward_name = event_item.reward.name
            local reward_count = event_item.reward.count
            local item_data = data.items[reward_name]
            local icon = item_data.icon
            local icon_size = 100
            if reward_name == "raid_turbo" then
                AddPlayerRaidTurbo(reward_count)
            elseif reward_name == "philosopher_stone" then
                AddPlayerPhilosopherStone(reward_count)
                icon_size = 120
            end

            local text = "<size=120%>+" .. reward_count .. "</size>"
            local offset = {
                spacing = 6,
                x = 0,
                y1 = -20,
                y2 = 60,
            }
            ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            DCEI.PlaySound(DCEI.Sound("Gacha"))
            bank.CommitHellModeEventClaimList()

            ui.leaderboard.event_reward_button:HidePip()
            ui.glue.floating_event:HidePip()
            ui.glue.leader_menu:HidePip()
            ui.CheckOtherButtonPip()
            CheckHellModeEventPip()
            local season_name = GetHellModeSeasonNameString(ui.leaderboard.season_index)
            local event = LogEvents.ClaimHellModeReward(season_name, index, player_score)
            LogEvent(event)
        end
    end
end

function CheckUnlockHellModeRankReward()
    local unlock_wave = 150
    return wave.data.GetMax() >= unlock_wave
        and player.hell_coin_shop.start_rotate
        and ui.leaderboard.season_index
        and ui.leaderboard.season_index >= NEW_RANK_REWARD_SEASON
end

function CheckServerTimeOnLeaderBoardReset()
    local function ResultCallback(time_stamp)
        if time_stamp then
            AttemptClaimHellModeSeasonReward(time_stamp)
            HookupLeaderboardTimerDisplay(time_stamp)
        end
    end

    AttemptGetServerTime(ResultCallback)
end

function GenerateHellCoinBadgeTable()
    local badge = {}
    badge.fv = player.hell_coin_shop.owned.favourite
    badge.list = {}
    for key, v in pairs(player.hell_coin_shop.owned.owned_badges) do
        badge.list[key] = true
    end

    return badge
end

ui.GenerateHellCoinBadgeTable = GenerateHellCoinBadgeTable

function HookupLeaderboardTimerDisplay(time_stamp)
    ui.leaderboard.season_not_ready = false
    ui.leaderboard.season_id, ui.leaderboard.season_index = GetHellModeSeasonId(time_stamp)
    if ui.leaderboard.season_index then
        DCEI.LogMessage("Hell Mode Season Id: " .. ui.leaderboard.season_index)
    end
    local prev_season_index = math.floor((time_stamp - (HELL_MODE_SEASON_START - SEASON_INTERVAL)) / SEASON_DURATION)
    local next_season_start_time = HELL_MODE_SEASON_START + prev_season_index * SEASON_DURATION
    if time_stamp < next_season_start_time then
        ui.leaderboard.season_not_ready = true
    end
    local season_index = ui.leaderboard.season_index
    if ui.leaderboard.season_index == nil then
        season_index = 0
        ui.leaderboard.season_not_ready = true
    end

    local reset_time = HELL_MODE_SEASON_START + (season_index + 1) * SEASON_DURATION - SEASON_INTERVAL
    local remaining_seconds = reset_time - time_stamp
    if ui.leaderboard.season_not_ready then
        remaining_seconds = next_season_start_time - time_stamp
    end

    if ui.leaderboard.season_index then
        if
            ui.leaderboard.season_index > 10
            and (
                not player.hell_mode_event_claim_list.season_index
                or (
                    player.hell_mode_event_claim_list.season_index
                    and player.hell_mode_event_claim_list.season_index < ui.leaderboard.season_index
                )
            )
        then
            for i = 1, #data.hell_mode_event_reward do
                player.hell_mode_event_claim_list["reward" .. i] = false
            end
            player.hell_mode_event_claim_list.season_index = ui.leaderboard.season_index
            bank.CommitHellModeEventClaimList()
        end
    end

    CheckHellModeEventLink()
    CheckHellModeEventPip()
    -- create new timer in ui.calendar.timer
    if not ui.leaderboard.timer then
        ui.leaderboard.timer = core.OsTimer:New()
    end

    ui.leaderboard.timer:AddDuration(remaining_seconds)

    -- check to update various timer displays each tick
    function ui.leaderboard.timer:Tick()
        -- update calendar timer display if calendar is open
        if ui.leaderboard.is_active then
            -- pass time remaining to timer display
            local time_remaining = ui.leaderboard.timer:GetTimeRemaining()
            local text = ui.FormatTimeDHM(time_remaining)
            ui.leaderboard:UpdateTimer(text)
        end

        if ui.glue.floating_event.active then
            local time_remaining = ui.leaderboard.timer:GetTimeRemaining()
            local text = ui.FormatTimeShort(time_remaining)
            ui.glue.floating_event:UpdateTimer(text)
        end
    end

    -- check server time when server day resets
    function ui.leaderboard.timer:Action()
        ui.leaderboard.timer = nil

        ui.leaderboard:UpdateTimer("00:00:00")

        -- DEBUG
        -- player.calendar.day = 200

        -- using timer as a safety buffer
        DCEI.TriggerAddTimerEventElapsed(CheckServerTimeOnLeaderBoardReset, 2, true, true)
    end

    ui.leaderboard.timer:Tick()

    -- hook up hell coin shop
    local unlock_check = IsHellModeUnlock()
    if
        time_stamp
        and unlock_check
        and ui.leaderboard.season_index
        and ui.leaderboard.season_index >= NEW_RANK_REWARD_SEASON
        and not player.hell_coin_shop.start_rotate
    then
        player.hell_coin_shop.start_rotate = true
        player.hell_coin_shop.badge_index = 1
        player.hell_coin_shop.rotate_time = time_stamp + HELL_COIN_SHOP_ROTATE_DURATION
        bank.CommitHellCoinShopData()
    end
    if player.hell_coin_shop.start_rotate then
        if player.hell_coin_shop.rotate_time < time_stamp then
            RotateHellCoinShop(time_stamp)
        end
    end
    AttemptStartHellCoinShopTimer(time_stamp)
end

function IsHellModeUnlock()
    local unlock_wave = 150
    return wave.data.GetMax() >= unlock_wave and not event_mode.IsActive()
end

function CheckHellModeUnlock()
    local unlock_check = IsHellModeUnlock()
    DCEI.SetFrameActive(ui.glue.hell_mode_button.frame, unlock_check)
    ui.leaderboard.unlock_hellmode = unlock_check
    if
        unlock_check
        and (not player.system.first_play_hell_mode or player.system.first_play_hell_mode == system.locked)
    then
        ui.glue.hell_mode_button:ShowPipAnimated()
    end
    -- start hell coin shop rotate
    if
        unlock_check
        and ui.leaderboard.season_index
        and ui.leaderboard.season_index >= NEW_RANK_REWARD_SEASON
        and not player.hell_coin_shop.start_rotate
    then
        local function ResultCallback(time_stamp)
            if time_stamp then
                player.hell_coin_shop.start_rotate = true
                player.hell_coin_shop.badge_index = 1
                player.hell_coin_shop.rotate_time = time_stamp + HELL_COIN_SHOP_ROTATE_DURATION
                bank.CommitHellCoinShopData()
                AttemptStartHellCoinShopTimer(time_stamp)
            end
        end

        AttemptGetServerTime(ResultCallback)
    end
end

function AttemptStartHellCoinShopTimer(time_stamp)
    if not player.hell_coin_shop.start_rotate or player.hell_coin_shop.rotate_time == 0 then
        return
    end
    if player.hell_coin_shop_timer then
        player.hell_coin_shop_timer:Destroy()
    end
    local expire_time = player.hell_coin_shop.rotate_time
    local timer = core.OsTimer:New()
    player.hell_coin_shop_timer = timer

    timer:AddDuration(expire_time - time_stamp)

    function timer:Tick()
        local time = self:GetTimeRemaining()
        if ui.shop.sections["hell_coin"].timer_display then
            ui.shop.sections["hell_coin"].timer_display:TimerUpdate(time)
        end
    end

    function timer:Action()
        if ui.shop.sections["hell_coin"].timer_display then
            ui.shop.sections["hell_coin"].timer_display:TimerUpdate(0)
        end
        player.hell_coin_shop_timer = nil
        DCEI.TriggerAddTimerEventElapsed(CheckServerTimeOnHellCoinShopExpire, 2, true, true)
    end
end

-- !SECTION
-- !SECTION

-- SECTION MONSTER BOUNTY
function CheckStartMonsterBounty(wave_id)
    -- if there is already a bounty not complete, don't show next bounty
    if not player.running_monster_bounty then
        for k, item in ipairs(data.monster_bounty) do
            if wave_id >= item.wave and not item.complete then
                StartMonsterBounty(k)
                break
            end
        end
    end
end

function StartMonsterBounty(id)
    local bounty_data = data.monster_bounty[id]
    player.running_monster_bounty = bounty_data

    local current_time = os.time()
    if not bounty_data.expire_time then
        bounty_data.expire_time = current_time + 3 * 24 * 3600
    end

    local call_back = {
        claim_reward = ClaimMonsterBountyReward,
        create_bountylist = CreateBountyList,
    }
    ui.monster_bounty = ui.CreateMonsterBountyBar(bounty_data, call_back, data.monster_bounty)

    local current_kill = bounty_data.progress or 0
    ui.monster_bounty:UpdateProgress(current_kill)

    if current_kill == 0 then
        local event = LogEvents.MonsterBountyStart(id)
        LogEvent(event)
    end

    local time_remaining = bounty_data.expire_time - current_time
    if time_remaining > 0 then
        AddMonsterBountyTime(time_remaining)
    end

    DCEI.TriggerAddTimerEventElapsed(function()
        bank.CommitMonsterBounty()
    end, 0, true, true)
end

function ClaimMonsterBountyReward()
    if player.running_monster_bounty then
        BlockInput()
        PlayGoldSpendSound()
        local wave_id = player.running_monster_bounty.wave
        for k, item in ipairs(data.monster_bounty) do
            if wave_id == item.wave then
                local event = LogEvents.MonsterBountyClaim(k)
                LogEvent(event)
                break
            end
        end

        if player.running_monster_bounty.reward then
            local reward = player.running_monster_bounty.reward
            local value = reward.count
            if reward.item == "token" then
                AddPlayerOverflowGems(value)
                ui.TokenRewardFeedback(value)
            elseif reward.item == "gold" then
                AddPlayerGold(value)
                ui.GoldRewardFeedback(value)
            elseif reward.item == "auto_battle" then
                AddGemAutoBattleTime(value)
                local icon = reward.icon
                local icon_size = 160
                local hour = value / 3600
                local key = hour > 1 and "general/hours_number" or "general/hour_number"
                hour = string.format("%.1f", hour)
                local text = DCEI.GetTranslationText(key, { x = "<size=120%>+" .. hour .. "</size>" })
                text = "<color=green>" .. text .. "</color>"
                local offset = {
                    spacing = -20,
                    x = 0,
                    y1 = -20,
                    y2 = 60,
                }
                ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            elseif reward.item == "chest" then
                local chest_type = reward.type
                AddPlayerChest(chest_type, value)
                local icon = reward.icon
                local icon_size = 100
                local color = reward.color or ""
                local text = color .. "<size=120%>+" .. value .. "</size>"
                local offset = {
                    spacing = 6,
                    x = 0,
                    y1 = -20,
                    y2 = 60,
                }
                ui.CreateFloatingMessage(icon, icon_size, text, offset, 1)
            end
        end
        DCEI.AnimateFrameScale(ui.monster_bounty.Frame, { x = 1, y = 1 }, { x = 0, y = 0 }, 0.5, "InBack")
        DCEI.TriggerAddTimerEventElapsed(function()
            EndMonsterBounty()
            CheckStartMonsterBounty(wave.data.GetMax())
            UnblockInput()
        end, 0.625, true, true)
    end
end

function EndMonsterBounty()
    if player.running_monster_bounty then
        local bounty_data = player.running_monster_bounty
        bounty_data.complete = true
        if ui.monster_bounty and DCEI.FrameExists(ui.monster_bounty.Frame) then
            DCEI.DestroyFrame(ui.monster_bounty.Frame)
            ui.monster_bounty = nil
        end
        player.running_monster_bounty = nil

        bank.CommitMonsterBounty()
    end
end

function AddMonsterBountyTime(duration)
    if player.monster_bounty_timer then
        player.monster_bounty_timer:Destroy()
    end
    -- otherwise create new timer
    player.monster_bounty_timer = core.OsTimer:New()

    player.monster_bounty_timer:AddDuration(duration)

    -- each tick, pass time remaining to relevant UI
    function player.monster_bounty_timer:Tick()
        local time = self:GetTimeRemaining()
        if not wave.data.is_active and ui.monster_bounty and DCEI.FrameExists(ui.monster_bounty.Frame) then
            ui.monster_bounty:TimerUpdate(time)
        end
    end

    function player.monster_bounty_timer:Action()
        if player.running_monster_bounty then
            local wave_id = player.running_monster_bounty.wave
            for k, item in ipairs(data.monster_bounty) do
                if wave_id == item.wave then
                    EndMonsterBounty()
                    break
                end
            end
        end
    end

    -- execute an update tick immediately if time remains
    if player.monster_bounty_timer:GetTimeRemaining() > 0 then
        player.monster_bounty_timer:Tick()
    end
end

-- !SECTION

-- SECTION PLATFORM help functions
function GetPlatform()
    return (PlayingEditorMode() and DEBUG.simulate_platform) or DCEI.Platform
end

function PlayingEditorMode()
    return DCEI.Platform == "WindowsPlayer" and DCEI.GetUserId() == "dev"
end

function PlayingWebVersion(or_dev_mode)
    return GetPlatform() == "WebGLPlayer"
end

function PlayingMobileVersion(or_dev_mode)
    return PlayingAndroidVersion() or PlayingiOSVersion()
end

function PlayingAndroidVersion()
    return GetPlatform() == "Android"
end

function PlayingiOSVersion()
    return GetPlatform() == "IPhonePlayer"
end

-- platform specific UI hookups need to happen before ui mod is executed
ui.PlayingWebVersion = PlayingWebVersion
ui.PlayingMobileVersion = PlayingMobileVersion
ui.PlayingAndroidVersion = PlayingAndroidVersion
ui.PlayingiOSVersion = PlayingiOSVersion
-- !SECTION

-- SECTION ENEMIES
function SendEnemyTowardsCastle(unit)
    local pos = DCEI.GetUnitPosition2D(unit)
    local name = DCEI.GetUnitType(unit)

    local castle_y = 18

    -- DCEI.LogMessage("> " .. name .. " at " .. pos.x .. ", " .. pos.y)
    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.MoveAttack(unit, pos.x, castle_y)
    end, 0, false, simple)
end

function SeasonEventEnemyPassBy(unit, run_away)
    local pos = DCEI.GetUnitPosition2D(unit)

    local castle_y = 18
    local runaway_y = 40
    local stay_time = 4

    if run_away then
        stay_time = 6
        DCEI.TriggerAddUnitDamageEvent(unit, function()
            local u = DCEI.TriggeringUnit
            local health_frac = DCEI.GetHealth(u) / DCEI.GetMaxHealth(u)
            local count_stack = DCEI.GetUnitBehaviorStackCount(u, DCEI.Behavior("Monster Season Event Lost Hp Stack"))
            local frac_ceil = math.ceil(health_frac * 10)
            local winter_event = CheckSeasonEventExist("winter_event2022")
            if count_stack < (10 - frac_ceil) then
                local drop_count = 10 - frac_ceil - count_stack
                local skip_commit = true
                for i = 1, drop_count do
                    local coin_num = math.random(2, 4)
                    local icon = "general_icon_icecoin1"
                    if winter_event then
                        AddPlayerWinterCoin2022(coin_num, skip_commit)
                        wave_drop_resource.winter_coin = wave_drop_resource.winter_coin or 0
                        wave_drop_resource.winter_coin = wave_drop_resource.winter_coin + coin_num
                        -- show loot drop feedback
                    else
                        AddPlayerAutumnCoin(coin_num, skip_commit)
                        wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin or 0
                        wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin + coin_num
                        -- show loot drop feedback
                        icon = "general_icon_coin_leaf"
                    end
                    ui.CreateFloatingLootDisplay(unit, icon, coin_num)
                    DCEI.ApplyBehaviorToSelf(u, DCEI.Behavior("Monster Season Event Lost Hp Stack"), 1)
                end
            end
        end)
    end

    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        DCEI.Move(unit, pos.x, castle_y)
    end, 0, false, simple)
    DCEI.TriggerAddTimerEventElapsed(function()
        if unit and DCEI.UnitExists(unit) then
            local ability_list = DCEI.GetUnitAbilityList(unit)
            local rd = math.random(1, #ability_list)
            DCEI.CastAbilityAtUnit(ability_list[rd], unit, unit)
            DCEI.TriggerAddTimerEventElapsed(function()
                if unit and DCEI.UnitExists(unit) then
                    DCEI.Move(unit, pos.x, runaway_y)
                end
            end, 1.5, false, simple)
        end
    end, stay_time, false, simple)
    local function CheckSeasonEventEnemyPosition(unit)
        if unit and DCEI.UnitExists(unit) then
            local position = DCEI.GetUnitPosition2D(unit)
            if position.y > 38 then
                DCEI.RemoveUnit(unit)
                CheckIfWaveIsOver()
            else
                DCEI.TriggerAddTimerEventElapsed(function()
                    CheckSeasonEventEnemyPosition(unit)
                end, 2, false, simple)
            end
        end
    end

    CheckSeasonEventEnemyPosition(unit)
end

-- !SECTION

-- SECTION CASTLE CREATION
function InitializeCastle()
    local behavior = DCEI.Behavior("Player Castle Upgrade")
    local count = GetCastleLevel() - 1

    SpawnCastleUnit()
    SpawnCastleChunks()

    if count > 0 then
        AddUpgradeData(data.castle_upgrades, behavior, count)
        ApplyUpgradeDataToUnit(data.castle_upgrades, CASTLE_UNIT)

        local level = GetCastleLevel()
        local is_instant = true
        for n = 1, level do
            CheckCastleUnlocks(n, is_instant)
        end
    end
    PreviewNextUnlockSlot()
end

function SpawnCastleUnit()
    local dx, dy = 0, 1
    local x, y = 24, 17

    CASTLE_UNIT = DCEI.CreateUnitAsync(1, 1, CASTLE_UNIT_NAME, x, y, dx, dy)
    ui.HookupCastleStatusBars(CASTLE_UNIT)
end

function SpawnCastleChunks()
    local unit_type = DCEI.Unit("Player Castle Chunk")
    local y = 19.5

    for x = 21, 27, 1 do
        local unit = DCEI.CreateUnitAsync(1, 1, unit_type, x, y)
        table.insert(group_castle_wall, unit)

        -- apply castle wall upgrades
        ApplyUpgradeDataToUnit(data.castle_wall_upgrades, unit)
    end
end

-- UPGRADE CASTLE
function AttemptUpgradeCastleLevel()
    local gold = GetPlayerGold()
    local cost = GetCastleUpgradeCost()
    local success = true

    if cost <= gold then
        local spend_event = { type = "upgrade", target = "Castle" }
        SpendPlayerGold(cost, spend_event)

        PlayGoldSpendSound()

        local upgrade_castle = DCEI.GetTranslationText("ui/upgrade/castle")
        local upgrade_hp_mp = DCEI.GetTranslationText("ui/upgrade/castle/stats")
        local text = (GetCastleLevel() < data.unlock.max) and upgrade_castle or upgrade_hp_mp
        ui.glue.upgrade_castle:AnimateText(text)
        UpgradeCastle()
    else
        local offset = -214
        ui.ShowFeedbackNeedMoreGold(offset)
        success = false
    end
    if not event_mode.IsActive() then
        Tutorial03_Castle_StepB()
    end

    return success
end

-- UPGRADE CASTLE EVENT
function AttemptUpgradeCastleLevelEvent()
    local gold = GetPlayerGold()
    local gold_cost = GetCastleUpgradeCost()
    local level = GetCastleLevel()
    local potions = event_mode.GetPlayerPotions()
    local potion_cost = level
    
    local success = true

    if (level % 10) == 0 then -- every 10 levels add a potion cost
        if gold_cost <= gold and potion_cost <= potions then
            local spend_event = { type = "upgrade", target = "Castle" }
            SpendPlayerGold(gold_cost, spend_event)
            event_mode.SpendPlayerPotions(potion_cost)

            PlayGoldSpendSound()

            local upgrade_castle = DCEI.GetTranslationText("ui/upgrade/castle")
            local upgrade_hp_mp = DCEI.GetTranslationText("ui/upgrade/castle/stats")
            local text = (GetCastleLevel() < data.unlock.max) and upgrade_castle or upgrade_hp_mp
            ui.glue.upgrade_castle:AnimateText(text)
            UpgradeCastle()
        else
            if gold_cost > gold then
                local offset = -214
                ui.ShowFeedbackNeedMoreGold(offset)
            else
                event_mode.ShowFeedbackNotEnoughPowerPotions()
            end
            success = false
        end
    else
        if gold_cost <= gold then
            local spend_event = { type = "upgrade", target = "Castle" }
            SpendPlayerGold(gold_cost, spend_event)

            PlayGoldSpendSound()

            local upgrade_castle = DCEI.GetTranslationText("ui/upgrade/castle")
            local upgrade_hp_mp = DCEI.GetTranslationText("ui/upgrade/castle/stats")
            local text = (GetCastleLevel() < data.unlock.max) and upgrade_castle or upgrade_hp_mp
            ui.glue.upgrade_castle:AnimateText(text)
            UpgradeCastle()
        else
            local offset = -214
            ui.ShowFeedbackNeedMoreGold(offset)
            success = false
        end
    end

    return success
end

function UpgradeCastle()
    -- apply upgrade behavior
    local upgrade = DCEI.Behavior("Player Castle Upgrade")
    AddCastleUpgrade(upgrade, 1)

    -- set protected level variable
    AddCastleLevel(1)

    -- update upgrade button and check unlocks
    local level = GetCastleLevel()
    UpdateCastleUpgradeButton()
    CheckCastleUnlocks()
    PreviewNextUnlockSlot()

    -- upgrade castle visuals and check for unloadable prop layers
    local animate = true
    DCEI.TriggerAddTimerEventElapsed(function()
        BuildCastle(level, animate)
    end, 0, false, false)
    AttemptQueueUnloadCastleChunk(level)

    PreloadNextCastleChunks(level)

    -- save bank
    if event_mode.IsActive() then
        bank.CommitEventCastle()
    else
        bank.CommitCastle()
    end
end

-- !SECTION

-- SECTION SUMMON MANAGEMENT
function AddSummonData(summon, unit_data, sign)
    -- stores summon to summon table
    local name = summon.name
    local count = (summon.count or 1) * sign
    local summon_data = data.summons[name]

    if summon_data then
        summon_data.count = summon_data.count + count
    else
        local new_summon = {
            name = name,
            count = count,
            use_archer_upgrades = summon.use_archer_upgrades,
            upgrade = summon.upgrade,
            icon = summon.icon,
        }
        table.insert(data.summons, new_summon)
        data.summons[name] = new_summon
    end

    -- pass growth level
    data.summons[name].unit_data = unit_data
end

function SpawnSummons(wave_data)
    local index = 1
    local summon_count = 0
    local slots = {}
    local slots_odd = {
        { x = 24, y = 22 },
        { x = 22, y = 22 },
        { x = 26, y = 22 },
        { x = 23, y = 23.5 },
        { x = 25, y = 23.5 },
    }
    local slots_even = {
        { x = 23.25, y = 22 },
        { x = 24.75, y = 22 },
        { x = 21.75, y = 22 },
        { x = 26.25, y = 22 },
    }

    for _, summon in ipairs(data.summons) do
        summon_count = summon_count + ((summon.count > 0) and 1 or 0)
    end

    if summon_count % 2 == 0 then
        slots = slots_even
    else
        slots = slots_odd
    end

    local dx, dy = 0, 1
    for _, summon in ipairs(data.summons) do
        if summon.count > 0 then
            local x = slots[index].x
            local y = slots[index].y

            -- create summon
            local unit = DCEI.CreateUnitAsync(1, 1, summon.name, x, y, dx, dy)
            table.insert(group_summons, unit)

            -- apply base upgrades
            if summon.use_archer_upgrades then
                local skip_source_all = true
                ApplyUpgradeDataToUnit(data.archer_upgrades, unit, skip_source_all)
            end

            ApplyUpgradeDataToUnit(data.summon_upgrades, unit)

            -- apply level upgrades
            local upgrade = summon.upgrade
            local growth_level = summon.unit_data:GetLevel() - 1
            if upgrade and growth_level > 0 then
                DCEI.ApplyBehaviorCountAsync(unit, upgrade, growth_level)
            end

            ColonyTargetingCheck(unit, wave_data)

            index = math.min(index + 1, #slots)
        end
    end
end

-- !SECTION

-- SECTION UPGRADE BEHAVIOR MANAGEMENT
function AddUpgradeData(data_set, behavior, count, source)
    -- stores upgrade behavior/counts in data
    local upgrade_data = data_set[behavior]
    if upgrade_data then
        upgrade_data.count = upgrade_data.count + count
    else
        local new_upgrade = {
            name = behavior,
            count = count,
            source = source,
        }
        table.insert(data_set, new_upgrade)
        data_set[behavior] = new_upgrade
    end
end

function ApplyUpgradeDataToUnit(data_set, unit, skip_source_all)
    -- apply all data_set upgrades to unit
    local unit_upgrades = data_set
    if unit_upgrades then
        for _, upgrade in ipairs(unit_upgrades) do
            local behavior = upgrade.name
            local count = upgrade.count
            local source = upgrade.source

            if skip_source_all and source == "all" then
                -- skip this case
            elseif count ~= 0 then
                DCEI.ApplyBehaviorCountAsync(unit, behavior, count)
            end
        end
    end
end

function ModifyUnitBehavior(unit, behavior, count)
    if count > 0 then
        DCEI.ApplyBehaviorCountAsync(unit, behavior, count)
    else
        DCEI.RemoveBehaviorCountAsync(unit, behavior, -1 * count)
    end
end

-- hero
function AddHeroUpgrade(behavior, count, source)
    AddUpgradeData(data.hero_upgrades, behavior, count, source)
    ApplyHeroBehavior(behavior, count)
end

function ApplyHeroBehavior(behavior, count)
    for _, unit in pairs(player.heroes) do
        ModifyUnitBehavior(unit, behavior, count)
    end
end

-- tower
function AddTowerUpgrade(behavior, count, source)
    AddUpgradeData(data.tower_upgrades, behavior, count, source)
    ApplyTowerBehavior(behavior, count)
end

function ApplyTowerBehavior(behavior, count)
    for _, unit in pairs(player.towers) do
        ModifyUnitBehavior(unit, behavior, count)
    end
end

-- monster
function AddMonsterUpgrade(behavior, count)
    AddUpgradeData(data.monster_upgrades, behavior, count)
end

-- summon
function AddSummonUpgrade(behavior, count, source)
    AddUpgradeData(data.summon_upgrades, behavior, count, source)
end

-- castle controller
function AddCastleUpgrade(behavior, count)
    AddUpgradeData(data.castle_upgrades, behavior, count)
    ApplyCastleBehavior(behavior, count)
end

function ApplyCastleBehavior(behavior, count)
    local unit = CASTLE_UNIT
    if unit then
        ModifyUnitBehavior(unit, behavior, count)
    end
end

-- castle wall
function AddCastleWallUpgrade(behavior, count)
    AddUpgradeData(data.castle_wall_upgrades, behavior, count)
    ApplyCastleWallBehavior(behavior, count)
end

function ApplyCastleWallBehavior(behavior, count)
    for _, unit in pairs(group_castle_wall) do
        ModifyUnitBehavior(unit, behavior, count)
    end
end

-- archer
function AddArcherUpgrade(behavior, count, source)
    AddUpgradeData(data.archer_upgrades, behavior, count, source)
    ApplyArcherBehavior(behavior, count)

    UpdateArcherUpgradeButton()
end

function ApplyArcherBehavior(behavior, count)
    for _, unit in pairs(group_archers) do
        ModifyUnitBehavior(unit, behavior, count)
    end
end

-- !SECTION

-- SECTION Archer
-- UPGRADE ARCHER
function AttemptUpgradeArcher()
    local player_id = 1
    local gold = GetPlayerGold()
    local cost = GetArcherUpgradeCost()
    local success = true
    local next_level = GetArcherLevel() + 1

    if cost <= gold then
        local spend_event = { type = "upgrade", target = "Archer" }
        SpendPlayerGold(cost, spend_event)

        PlayGoldSpendSound()

        local upgrade_archer = DCEI.GetTranslationText("ui/upgrade/archer")
        local upgrade_damage = DCEI.GetTranslationText("ui/upgrade/archer/damage")
        local text = (GetArcherLevel() < POINTS.archer.max) and upgrade_archer or upgrade_damage
        ui.glue.upgrade_archer:AnimateText(text)
        UpgradeArcher()
    else
        local offset = -214
        ui.ShowFeedbackNeedMoreGold(offset)
        success = false
    end

    if not event_mode.IsActive() then
        Tutorial02_Archer_StepB()
    end

    return success
end

-- UPGRADE ARCHER
function AttemptUpgradeArcherEvent()
    local player_id = 1
    local gold = GetPlayerGold()
    local cost = GetArcherUpgradeCost()
    local success = true
    local next_level = GetArcherLevel() + 1
    local level = GetCastleLevel()
    local potions = event_mode.GetPlayerPotions()
    local potion_cost = level

    if (level % 10) == 0 then -- every 10 levels add a potion cost
        if cost <= gold and potion_cost <= potions then
            local spend_event = { type = "upgrade", target = "Archer" }
            SpendPlayerGold(cost, spend_event)
            event_mode.SpendPlayerPotions(potion_cost)

            PlayGoldSpendSound()

            local upgrade_archer = DCEI.GetTranslationText("ui/upgrade/archer")
            local upgrade_damage = DCEI.GetTranslationText("ui/upgrade/archer/damage")
            local text = (GetArcherLevel() < POINTS.archer.max) and upgrade_archer or upgrade_damage
            ui.glue.upgrade_archer:AnimateText(text)
            UpgradeArcher()
        else
            if cost > gold then
                local offset = -214
                ui.ShowFeedbackNeedMoreGold(offset)
            else
                event_mode.ShowFeedbackNotEnoughPowerPotions()
            end
            success = false
        end
    else
        if cost <= gold then
            local spend_event = { type = "upgrade", target = "Archer" }
            SpendPlayerGold(cost, spend_event)

            PlayGoldSpendSound()

            local upgrade_archer = DCEI.GetTranslationText("ui/upgrade/archer")
            local upgrade_damage = DCEI.GetTranslationText("ui/upgrade/archer/damage")
            local text = (GetArcherLevel() < POINTS.archer.max) and upgrade_archer or upgrade_damage
            ui.glue.upgrade_archer:AnimateText(text)
            UpgradeArcher()
        else
            local offset = -214
            ui.ShowFeedbackNeedMoreGold(offset)
            success = false
        end
    end

    return success
end

-- UPGRADE ARCHER
function AttemptUpgradeArcherLevelCap()
    local player_id = 1
    local gold = GetPlayerGold()
    local cost = GetArcherUpgradeCost()
    local success = true
    local next_level = GetArcherLevel() + 1

    if cost <= gold then
        local spend_event = { type = "upgrade", target = "Archer" }
        SpendPlayerGold(cost, spend_event)

        PlayGoldSpendSound()

        local upgrade_archer = DCEI.GetTranslationText("ui/upgrade/archer")
        local upgrade_damage = DCEI.GetTranslationText("ui/upgrade/archer/damage")
        local text = (GetArcherLevel() < POINTS.archer.max) and upgrade_archer or upgrade_damage
        ui.glue.upgrade_archer:AnimateText(text)
        UpgradeArcher()
    else
        local offset = -214
        ui.ShowFeedbackNeedMoreGold(offset)
        success = false
    end
    if not event_mode.IsActive() then
        Tutorial02_Archer_StepB()
    end


    return success
end

function InitializeArcher()
    local max_count = POINTS.archer.max
    local attack_power = GetArcherBaseAttackPower()
    local archer_count = math.min(GetArcherLevel(), max_count)
    local behavior = DCEI.Behavior("Archer AttackPower")

    -- value is floored as upgrade data works with ints
    AddUpgradeData(data.archer_upgrades, behavior, math.floor(attack_power))

    -- upgrade behaviors are applied here
    for n = 1, archer_count do
        SpawnArcher(n)
    end

    -- apply archer visual
    local archer_level = GetArcherLevel()
    local visual_table = data.upgrades.archer.visual_update
    for i = #visual_table, 1, -1 do
        local change_level = visual_table[i].level
        if archer_level > change_level then
            local count = math.floor((archer_level - change_level) / visual_table[i].upgrade)
            count = math.min(math.max(count, 0), 20)
            local behavior_name = "Castle Archer Visual " .. change_level
            for n = 1, count do
                local archer_unit = group_archers[n]
                DCEI.ApplyBehaviorAsync(archer_unit, behavior_name)
            end
            if count < 20 and i > 1 then
                local change_level = visual_table[i - 1].level
                local behavior_name = "Castle Archer Visual " .. change_level
                -- apply rest archers the lower level visual
                for n = count + 1, 20 do
                    local archer_unit = group_archers[n]
                    DCEI.ApplyBehaviorAsync(archer_unit, behavior_name)
                end
            end
            break
        end
    end
    UpdateArcherUpgradeButton()
end

function UpgradeArcher()
    local max_purchase = POINTS.archer.max
    AddArcherLevel(1)

    if GetArcherLevel() <= max_purchase then
        SpawnArcher()
    else
        local attack_upgrade = data.archer_upgrades["Archer AttackPower"]
        local attack_value = attack_upgrade and attack_upgrade.count or 0
        local expected_val = GetArcherBaseAttackPower()
        local count = expected_val - attack_value

        -- value is floored but will still roll-over
        AddArcherUpgrade("Archer AttackPower", math.floor(count))
        local visual_update = data.upgrades.archer.visual_update_table[GetArcherLevel()]
        if visual_update then
            local update_archer = group_archers[visual_update.index]
            local behavior_name = "Castle Archer Visual " .. visual_update.level
            DCEI.ApplyBehaviorAsync(update_archer, behavior_name)
            DCEI.ApplyBehaviorCountAsync(update_archer, DCEI.Behavior("Castle Archer Birth FX"), 1)
        end
    end

    UpdateArcherUpgradeButton()
    if event_mode.IsActive() then
        bank.CommitEventArchers()
    else
        bank.CommitArchers()
    end
end

function GeneratePointFormation(formation_data)
    local formation = {}

    local height = formation_data.offset.y * (formation_data.count.y - 1)
    local width = formation_data.offset.x * (formation_data.count.x - 1)

    local start_x = formation_data.mid.x - (width / 2)
    local start_y = formation_data.mid.y - (height / 2)

    if formation_data.y_to_x then
        for x = start_x, start_x + ((formation_data.count.x - 1) * formation_data.offset.x), formation_data.offset.x do
            for y = start_y, start_y + ((formation_data.count.y - 1) * formation_data.offset.y), formation_data.offset.y do
                table.insert(formation, { x = x, y = y })
            end
        end
    else
        for y = start_y, start_y + ((formation_data.count.y - 1) * formation_data.offset.y), formation_data.offset.y do
            for x = start_x, start_x + ((formation_data.count.x - 1) * formation_data.offset.x), formation_data.offset.x do
                table.insert(formation, { x = x, y = y })
            end
        end
    end

    return formation
end

function GenerateArcherPoints()
    POINTS.archer = {}
    local formation_data = {
        mid = { x = 24, y = 14.125 },
        offset = { x = 0.45, y = -0.625 },
        count = { x = 10, y = 2 },
    }

    POINTS.archer[1] = GeneratePointFormation(formation_data)
    POINTS.archer[2] = GeneratePointFormation(formation_data)
    POINTS.archer[3] = GeneratePointFormation(formation_data)

    POINTS.archer.current = POINTS.archer[1]
    POINTS.archer.max = 20

    -- create gear unity helper
    GEAR_UNITY_HELPER = DCEI.CreateUnit(
        1,
        1,
        DCEI.Unit("Gear Unity Display Archer Critical Damage Helper"),
        formation_data.mid.x,
        formation_data.mid.y + 1.75,
        0,
        1
    )
end

function SpawnArcher(n)
    local play_fx = (n == nil)
    n = n or #group_archers + 1
    local dx, dy = 0, 1
    local unit_type = DCEI.Unit("Castle Archer")
    local unit = nil

    local point = POINTS.archer.current[n]
    if point then
        unit = DCEI.CreateUnitAsync(1, 1, unit_type, point.x, point.y, dx, dy)
        table.insert(group_archers, unit)
        ApplyUpgradeDataToUnit(data.archer_upgrades, unit)
    end

    -- no archer index means the archer was spawned from upgrade and should have build anim
    if unit and play_fx then
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Castle Archer Birth FX"), 1)
    end
end

function MigrateArchers()
    local archers = { table.unpack(group_archers) }
    group_archers = {}

    for k, unit in ipairs(archers) do
        DCEI.RemoveUnit(unit)
        SpawnArcher(k)
    end
end

-- !SECTION

-- SECTION Castle Unlocks
function CheckCastleUnlocks(level, is_instant)
    level = level or GetCastleLevel()

    if level <= data.unlock.max then
        if data.unlock[level] then
            local hero_slot = data.unlock[level].hero_slot
            local tower_slot = data.unlock[level].tower_slot
            local defense_slot = data.unlock[level].defense_slot
            local town_slot = data.unlock[level].town_slot
            local unlock_craft = data.unlock[level].unlock_craft

            if hero_slot then
                UnlockHeroSlot(hero_slot)
            end
            if tower_slot then
                UnlockTowerSlot(tower_slot)
            end
            if defense_slot and not event_mode.IsActive() then
                UnlockDefenseSlots()
            end
            if town_slot then
                UnlockTownSlot(town_slot, is_instant)
            end
            if unlock_craft and not is_instant then
                if UnlockEquipment() then
                    ui.glue.inventory_menu:ShowPipAnimated()
                end
            end
        end
    end
end

function GetCastleUpgradeCost()
    local level = GetCastleLevel()
    return data.upgrades.castle:GetUpgradeCost(level)
end

function GetArcherUpgradeCost()
    local level = GetArcherLevel()
    return data.upgrades.archer:GetUpgradeCost(level)
end

function GetArcherBaseAttackPower(_level)
    local level = _level or GetArcherLevel()
    local base = 30
    local ranges = {
        {
            key_value = base,
            level = 1,
            growth = 0,
        },
        {
            key_value = base,
            level = 20,
            growth = 1,
        },
        {
            key_value = base + (180 * 1),
            level = 200,
            growth = 1.25,
        },
    }

    for k, range in ipairs(ranges) do
        local next_range = ranges[k + 1]

        if next_range then
            if level >= range.level and level < next_range.level then
                local delta = level - range.level
                return range.key_value + (delta * range.growth)
            end
        else
            local delta = level - range.level
            return range.key_value + (delta * range.growth)
        end
    end
end

function GetArcherDamageValue(atk_power)
    local attack_upgrade = data.archer_upgrades["Archer AttackPower"]
    local attack_value = attack_upgrade and attack_upgrade.count or 0
    attack_value = atk_power and atk_power or attack_value
    local multiplier = (1 + player.attributes.archer_damage) * (1 + player.attributes.all_damage)

    return (attack_value / 10) * multiplier + player.attributes.archer_damage_flat
end

function PreviewNextUnlockSlot()
    local castle_level = GetCastleLevel()
    if castle_level >= 19 then
        for i = castle_level + 1, data.unlock.max do
            if data.unlock[i] then
                local hero_slot = data.unlock[i].hero_slot
                local tower_slot = data.unlock[i].tower_slot
                local defense_slot = data.unlock[i].defense_slot
                if hero_slot or tower_slot or defense_slot then
                    if hero_slot then
                        PreviewNextUnlockHeroSlot(hero_slot, i)
                    end
                    if tower_slot then
                        PreviewNextUnlockTowerSlot(tower_slot, i)
                    end
                    if defense_slot and not event_mode.IsActive() then
                        PreviewNextUnlockDefenseSlot(i)
                    end
                    break
                end
            end
        end
    end
    DCEI.TriggerAddTimerEventElapsed(function()
        core.SendActorMessage(group_tower_slots[5], "setVisibility", castle_level >= 150)
        core.SendActorMessage(group_tower_slots[6], "setVisibility", castle_level >= 200)
    end, 0, true, true)
end

function PreviewNextUnlockHeroSlot(index, unlock_castle_level)
    local slot = ui.loadout.hero.slots[index]
    slot.locked = true
    slot.unlock_level = unlock_castle_level
    DCEI.SetFrameActive(slot.button, true)
    DCEI.SetFrameActive(slot.select, false)
    DCEI.SetFrameActive(slot.lock_icon, true)
end

function PreviewNextUnlockTowerSlot(index, unlock_castle_level)
    local slot = ui.loadout.tower.slots[index]
    slot.locked = true
    slot.unlock_level = unlock_castle_level
    DCEI.SetFrameActive(slot.button, true)
    DCEI.SetFrameActive(slot.select, false)
    DCEI.SetFrameActive(slot.lock_icon, true)
end

function PreviewNextUnlockDefenseSlot(unlock_castle_level)
    for index = 1, 2 do
        local slot = ui.loadout.defense.slots[index]
        slot.locked = true
        slot.unlock_level = unlock_castle_level
        DCEI.SetFrameActive(slot.button, true)
        DCEI.SetFrameActive(slot.select, false)
        DCEI.SetFrameActive(slot.lock_icon, true)
    end
end

function UnlockHeroSlot(index)
    local slot = ui.loadout.hero.slots[index]
    slot.locked = false
    DCEI.SetFrameActive(slot.button, true)
    DCEI.SetFrameActive(slot.select, true)
    DCEI.SetFrameActive(slot.lock_icon, false)
end

function UnlockTowerSlot(index)
    local slot = ui.loadout.tower.slots[index]
    slot.locked = false
    DCEI.SetFrameActive(slot.button, true)
    DCEI.SetFrameActive(slot.select, true)
    DCEI.SetFrameActive(slot.lock_icon, false)
    if index >= 5 then
        local key = "extra_tower_unlock|" .. index
        if not player[key] then
            slot.token_locked = true
            DCEI.SetFrameActive(slot.lock_icon, true)
            if not slot.token_label then
                slot.token_label = gmui.Layout:New({
                    parent = slot.lock_icon,
                    name = "UnlockExtraTower/TokenLabel",
                })
                DCEI.SetTextFrameText(slot.token_label.UnlockCost, data.unlock.token_unlock_cost.tower_slot[index])
            end
        else
            slot.token_locked = false
            DCEI.SetFrameActive(slot.lock_icon, false)
            if slot.token_label then
                DCEI.DestroyFrame(slot.token_label.Frame)
                slot.token_label = nil
            end
        end
    end
end

function UnlockDefenseSlots()
    for index = 1, 2 do
        local slot = ui.loadout.defense.slots[index]
        slot.locked = false
        DCEI.SetFrameActive(slot.button, true)
        DCEI.SetFrameActive(slot.select, true)
        DCEI.SetFrameActive(slot.lock_icon, false)

        local unit = group_defense_slots[index]
        DCEI.RemoveBehaviorAsync(unit, DCEI.Behavior("System Hide"))
    end
end

function UnlockTownSlot(index, on_init)
    if player.town[index] == nil then
        local hide = DCEI.Behavior("System Hide")
        local slot = group_town_slots[index]
        DCEI.RemoveBehaviorAsync(slot, hide)
    end

    -- don't show campfire if slot is unlocked
    SetPropLayer(district_layers[index], false)
    district_layers[index] = ""

    -- provide quick reference for town build system
    BUILD.town.slot_count = index

    -- update prop layers to match
    local build_func = BUILD.town_slot[index]
    if build_func then
        build_func()
        UpdatePropLayer()
    end

    -- enable town slot select box
    local slot = ui.loadout.town.slots[index]
    slot.locked = false
    DCEI.SetFrameActive(slot.button, true)
    DCEI.SetFrameActive(slot.lock_icon, false)

    CheckUnlockTown()

    -- show town pip and district button pip
    if not on_init then
        ui.bottom_panel:ShowTownPipAnimated()
        ui.glue.district_menu:ShowPip()
        ui.loadout.town:HighlightUnequiptSlot()
    end

    -- check for town tutorial
    if not tutorial.town and data.GetOwnedTownCount() == 0 then
        tutorial.town = true
        tutorial.town_pip = true
    end
end

-- !SECTION

-- SECTION TOWN MINE / MINE WORKERS
function WorkerDwarfIncome(unit)
    -- previous remainder gets rolled over
    local income = data.dwarf_workers:GetIncome()
    income = income + data.dwarf_workers.remainder

    -- store new remainder
    local floored_income = math.floor(income)
    data.dwarf_workers.remainder = income - floored_income

    AddPlayerGold(floored_income, true)

    -- don't show this in cinematic mode
    if ui.CINEMATIC_MODE then
        return 0
    end

    if unit and data.dwarf_workers.mode == "unit" then
        -- only show gold floating text not during wave
        local pos = DCEI.GetUnitPosition3D(unit)
        local text = string.format("%i", floored_income)
        local formatted_text = "<color=yellow>" .. "+" .. text
        local anim_id = 0
        local show_duration = 0.5
        local anim_duration = 1

        DCEI.ShowFloatingText(pos, formatted_text, show_duration, anim_id, anim_duration, {
            offset = {
                up = 0.5 + (math.random(-4, 4) * 0.05),
                right = 0,
                front = -0.8,
            },
        })
    end

    return floored_income
end

function InitializeWorkerDwarfData()
    group_worker_dwarf = {
        count = 0,
        { x = 22.05, interval = 6.0 },
        { x = 21.90, interval = 3.0 },
        { x = 22.20, interval = 2.0 },
        { x = 21.75, interval = 1.5 },
        { x = 22.35, interval = 1.2 },
        { x = 22.05, interval = 1.0 },
        { x = 22.20, interval = 0.857 },
        { x = 21.90, interval = 0.75 },
        { x = 22.35, interval = 0.667 },
        { x = 21.75, interval = 0.5625 },
    }
end

function SpawnWorkerDwarf_AtDepot(n)
    n = n or group_worker_dwarf.count + 1

    local unit_data = group_worker_dwarf[n]
    local x, y = unit_data.x, 10.5

    local unit = DCEI.CreateUnitAsync(1, 1, DCEI.Unit("Town Worker Dwarf"), x, y)
    DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker Index"), n)
    DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker Birth FX"), 1)

    group_worker_dwarf[n].unit = unit
    group_worker_dwarf.count = group_worker_dwarf.count + 1

    DCEI.TriggerAddTimerEventElapsed(function()
        if n > 1 then
            DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker ToMines"), 1)
        else
            -- first trip of first dwarf is faster to showcase gold gain
            local duration = 1.125
            DCEI.ApplyBehaviorWithDurationAsync(unit, DCEI.Behavior("Town Worker ToMines"), duration)
        end
    end, 0.1875, false, true)
end

function SpawnWorkerDwarf_InMines(n)
    n = n or group_worker_dwarf.count + 1

    local unit_data = group_worker_dwarf[n]
    local x, y = unit_data.x, 12.25
    local delay = (n - 1) * 0.6

    DCEI.TriggerAddTimerEventElapsed(function()
        local unit = DCEI.CreateUnitAsync(1, 1, DCEI.Unit("Town Worker Dwarf"), x, y)
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker Index"), n)
        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker ToDepot"), 1)

        group_worker_dwarf[n].unit = unit
        group_worker_dwarf.count = group_worker_dwarf.count + 1
    end, delay, false, true)
end

function WorkerDwarf_ToMines(unit)
    local y = 12.25
    local index = DCEI.GetUnitBehaviorStackCount(unit, "Town Worker Index")
    local x = group_worker_dwarf[index].x
    DCEI.Move(unit, x, y)
end

function WorkerDwarf_ToDepot(unit)
    local y = 10.5
    local index = DCEI.GetUnitBehaviorStackCount(unit, "Town Worker Index")

    local x = group_worker_dwarf[index].x
    DCEI.Move(unit, x, y)
end

function CreateTownUnits()
    local x, y = 22, 10.25
    local mine_depot_name = DCEI.Unit("Town Mechanic Mine Depot")
    local airship_dock_name = DCEI.Unit("Town Airship Landing")
    local vault_unit_name = DCEI.Unit("Town Vault")

    -- town visuals TEST
    -- BuildTown(GetTownUpgradeLevel())

    -- town visuals
    BuildTown(data.town_upgrade:GetVisualLevel())

    -- town hall
    TOWN_HALL = DCEI.FindUnit("Town System TownHall")

    -- depot unit is used for tutorial helper attachment
    MINE_DEPOT = DCEI.CreateUnitAsync(1, 1, mine_depot_name, x, y)
    InitializeWorkerDwarfData()
    DwarfWorkers_UseUnits()

    -- airship
    x, y = 25.75, 10
    AIRSHIP_DOCK = DCEI.CreateUnitAsync(1, 1, airship_dock_name, x, y, 0, -1)

    local skip_sequence = true
    if player.system.airship == system.active then
        BuildAirshipSequence(skip_sequence)
    end

    -- vault
    if not player.system.vault then
        player.system.vault = system.locked
    end

    -- move the vault outside the screen if vault is not active
    x, y = 23.3, -20
    if not GetPlayerTest("pause_offline_gold") and player.system.vault == system.active then
        x, y = 23.3, 11.2
    end
    VAULT = DCEI.CreateUnitAsync(1, 1, vault_unit_name, x, y, 0, -1)
end

function DwarfWorkers_UseUnits()
    if data.dwarf_workers.mode ~= "unit" then
        data.dwarf_workers.instance = data.dwarf_workers.instance + 1
        data.dwarf_workers.mode = "unit"

        if GetDwarfLevel() > 0 then
            -- remove timer
            if data.dwarf_workers.timer then
                data.dwarf_workers.timer:Destroy()
            end

            -- create dwarf units
            local count = math.min(GetDwarfLevel(), data.dwarf_workers.max)
            local instance = data.dwarf_workers.instance

            for n = 1, count do
                local unit_data = group_worker_dwarf[n]
                local x, y = unit_data.x, 12.25
                local delay = (n - 1) * 0.6

                local simple = true
                DCEI.TriggerAddTimerEventElapsed(function()
                    if instance == data.dwarf_workers.instance and not wave.data.is_active then
                        local unit = DCEI.CreateUnitAsync(1, 1, DCEI.Unit("Town Worker Dwarf"), x, y)
                        DCEI.ApplyBehaviorCountAsync(unit, "Town Worker Index", n)
                        DCEI.ApplyBehaviorCountAsync(unit, DCEI.Behavior("Town Worker ToDepot"), 1)

                        group_worker_dwarf[n].unit = unit
                        group_worker_dwarf.count = group_worker_dwarf.count + 1
                    end
                end, delay, false, simple)
            end
        end
    end
end

function DwarfWorkers_UseTimers()
    wave.data.mine_income = 0
    if data.dwarf_workers.mode ~= "timer" then
        data.dwarf_workers.instance = data.dwarf_workers.instance + 1
        data.dwarf_workers.mode = "timer"
        if GetDwarfLevel() > 0 then
            -- create dwarf income timer
            local timer = core.GameTimer:New()
            local count = math.min(GetDwarfLevel(), data.dwarf_workers.max)
            local interval = group_worker_dwarf[count].interval
            data.dwarf_workers.timer = timer
            data.dwarf_workers.timer.interval = interval
            data.dwarf_workers.timer.duration = 0

            function timer:Tick()
                self.duration = self.duration + GLOBAL_TICK_RATE

                if self.duration >= self.interval then
                    self.duration = self.duration - self.interval

                    local mine_income = WorkerDwarfIncome()
                    wave.data.mine_income = wave.data.mine_income + mine_income
                end
            end

            -- remove all current dwarf worker units
            for _, worker in ipairs(group_worker_dwarf) do
                if worker.unit then
                    DCEI.RemoveUnit(worker.unit)
                    worker.unit = nil
                end
            end
            group_worker_dwarf.count = 0
        end
    end
end

function GetElixirInterval()
    return wave.data.is_active and data.town_upgrade.wave_interval or data.town_upgrade.base_interval
end

function AttemptStartElixirTimer()
    -- don't create elixir timer if one already exists
    if data.town_upgrade.timer then
        return
    end

    -- don't create elixir timer if town level too low
    if GetTownUpgradeLevel() <= 0 then
        return
    end

    -- create elixir income timer
    local timer = core.GameTimer:New()
    data.town_upgrade.timer = timer
    data.town_upgrade.timer.interval = GetElixirInterval()
    data.town_upgrade.timer.duration = 0

    function timer:Tick()
        self.duration = self.duration + GLOBAL_TICK_RATE

        if self.duration >= self.interval then
            AttemptHandleElixirIncome()

            -- reset duration
            self.duration = 0
        end
    end
end

function AttemptHandleElixirIncome()
    if event_mode.IsActive() then
        return
    end
    local timer = data.town_upgrade.timer

    if timer then
        -- get income_per_last_interval from income_per_minute
        local income = data.town_upgrade:GetElixirIncome() * (timer.duration / 60)
        timer.duration = 0

        -- update interval
        timer.interval = GetElixirInterval()

        -- previous remainder gets rolled over
        income = income + data.town_upgrade.remainder

        -- store new remainder
        local floored_income = math.floor(income)
        data.town_upgrade.remainder = income - floored_income

        GenerateElixir(floored_income)
    end
end

function GenerateElixir(value)
    -- default value is elixir income/min
    value = value or data.town_upgrade:GetElixirIncome()

    -- elixir saving is handled by AutoSaveTimer and WaveEndActions
    local skip_commit = true
    AddPlayerElixir(value, skip_commit)

    -- don't show this in cinematic mode
    if ui.CINEMATIC_MODE then
        return
    end

    if not wave.data.is_active then
        -- only display elixir gain when not capped
        local elixir_cap = data.town_upgrade:GetElixirCap()
        local elixir = GetPlayerElixir()
        if elixir < elixir_cap and value > 0 then
            -- display elixir gain on town hall building
            local icon = "icon_ingame_blackpotion"
            local loot_display = ui.CreateFloatingLootDisplay(TOWN_HALL, icon, value)
            DCEI.MoveFrameToBottom(loot_display.frame)

            AttemptUpdateCardElixirDisplay(elixir)
            AttemptUpdateItemMenuElixirDisplay(elixir)
        end
    end
end

-- !SECTION

-- SECTION AIRSHIP DELIVERY
function AttemptStartSeasonalAirshipTimer()
    -- don't create airship timer if one already exists
    if levels.seasonal_colony.timer then
        return
    end

    -- don't create airship timer if no colony income
    if levels.seasonal_colony:Income() <= 0 then
        return
    end

    local timer = core.GameTimer:New()
    levels.seasonal_colony.timer = timer
    levels.seasonal_colony.timer.interval = levels.seasonal_colony:Cooldown()
    levels.seasonal_colony.timer.duration = 0

    function timer:Tick()
        self.duration = self.duration + GLOBAL_TICK_RATE

        if self.duration >= self.interval then
            self.duration = self.duration - self.interval

            if levels.seasonal_colony:Income() > 0 then
                CreateColonyAirshipDelivery(true)
            end
        end
    end
end

function AttemptStartAirshipTimer()
    AttemptStartSeasonalAirshipTimer()
    -- don't create airship timer if one already exists
    if levels.colony.timer then
        return
    end

    -- don't create airship timer if no colony income
    if levels.colony:Income() <= 0 then
        return
    end

    local timer = core.GameTimer:New()
    levels.colony.timer = timer
    levels.colony.timer.interval = levels.colony:Cooldown()
    levels.colony.timer.duration = 0

    function timer:Tick()
        self.duration = self.duration + GLOBAL_TICK_RATE

        if self.duration >= self.interval then
            self.duration = self.duration - self.interval

            CreateColonyAirshipDelivery()
        end
    end
end

function CreateColonyAirshipDelivery(seasonal)
    local x1, y1 = 16, 9
    local x2, y2 = 32, 9
    local dx, dy = 0, -1
    local unit_type = DCEI.Unit("System Colony Delivery")

    local income_delay = 3.625
    local unit = DCEI.CreateUnitAsync(1, 1, unit_type, x1, y1, dx, dy)

    DCEI.Move(unit, x2, y2)

    -- need to wrap this in a timer rather than using wait because it's being called from global timer tick (engine timer optimization 6/12)
    local simple = true
    DCEI.TriggerAddTimerEventElapsed(function()
        local income = seasonal and levels.seasonal_colony:Income() or levels.colony:Income()
        local floored_income = math.floor(income)
        AddPlayerGold(floored_income, true)

        if not wave.data.is_active then
            bank.CommitGold()
            bank.CommitElixir()
            bank.CommitVaultGold()
        end

        -- don't show this in cinematic mode
        if ui.CINEMATIC_MODE then
            return
        end

        if unit then
            local pos = DCEI.GetUnitPosition3D(unit)
            local text = ui.FormatWithCommas(floored_income)
            local formatted_text = "<color=yellow>" .. "<size=120%>" .. "+" .. text
            local anim_id = 0
            local show_duration = 0.5
            local anim_duration = 1

            DCEI.ShowFloatingText(pos, formatted_text, show_duration, anim_id, anim_duration, {
                offset = {
                    up = 0,
                    right = 0.0,
                    front = 1.1,
                },
            })
        end
    end, income_delay, false, simple)
end

-- !SECTION

-- SECTION MISC functions
function MoveCamera(target_index, duration, options)
    target_index = target_index or ((cameras.current % 2) + 1)
    cameras.current = target_index

    local duration = duration or 0.5
    -- local camera = cameras[target_index]
    local camera = ShallowCopy(cameras[target_index])
    local use_game_time = true

    camera.x = camera.x + (options and options.x_offset or 0)
    camera.y = camera.y + (options and options.y_offset or 0)

    DCEI.SetCameraFocus(camera.x, camera.y, duration, use_game_time)
    DCEI.SetCameraDistance(camera.distance, duration, use_game_time)
end

function RemoveUnitGroupUnits(unit_group)
    for _, unit in ipairs(unit_group) do
        DCEI.RemoveUnit(unit)
    end
    unit_group = {}
end

-- TIMER
function RestartAutoSaveTimer()
    -- don't use auto save for devs, since it overwrites local edits
    if player.is_dev then
        return
    end

    -- destroy existing timer
    if player.auto_save_timer then
        player.auto_save_timer:Destroy()
    end

    -- create new timer
    local timer = core.OsTimer:New()
    player.auto_save_timer = timer

    -- commit resources every minute
    local duration = 60
    timer:AddDuration(duration)

    function timer:Action()
        if not wave.data.is_active then
            -- commit resources and restart timer (restart occurs in CommitGold())
            bank.CommitGold() 
        end
    end
end

function StartWaveTimer()
    local timer = core.GameTimer:New()
    timer.wave_instance = wave.data.instance

    function timer:Condition()
        return self.wave_instance ~= wave.data.instance
    end

    function timer:Action()
        local time = self:GetTimeElapsed()
        wave.data.elapsed = time
        DCEI.LogMessage("Timer [" .. self.id .. "] Elapsed: " .. time)
        UpdateBattleData()
    end
end

function LogPlayerAttributes()
    local print = "PLAYER ATTRIBUTES"
    local keys = {}
    local n = 0

    for k, v in pairs(player.attributes) do
        print = "\n" .. k .. ": " .. v
        DCEI.LogMessage(print)
    end
    return print
end

function LogEvent(event)
    -- check if logging is enabled
    if LogEvents.logging then
        -- check if log event is valid
        if event then
            local key = event.name or event.Name()
            local props = event.props or event.Props()

            if not key then
                DCEI.LogError("error: invalid event key [" .. tostring(key) .. "]")
            elseif not props then
                DCEI.LogError("error: invalid event props [" .. tostring(props) .. "]")
            end

            -- if event.use_wave_cap and wave.data.GetMax() > LogEvents.wave_cap then
            --     -- log less things after wave 100
            --     return
            -- end

            if key == "Buy Item" and event.start_spend_log then
                -- start spend logging after shop purchase
                StartSpendLogging(event)
            end

            if event.use_wave_cap and props.Max_wave then
                local wave = props.Max_wave

                if
                    (wave < 500 and (wave == 1 or wave == 3 or (wave % 5 == 0)))
                    or (wave >= 500 and (wave % 100 == 0))
                then
                    -- only log the specified waves
                elseif props.Mode == "standard" then
                    return
                end

                -- if (
                --     wave == 1 or
                --     wave == 3 or
                --     wave == 5 or
                --     wave == 10 or
                --     wave == 20 or
                --     wave == 25 or
                --     wave == 30 or
                --     wave == 35 or
                --     wave == 40 or
                --     wave == 50 or
                --     wave == 60 or
                --     wave == 80 or
                --     (wave % 100 == 0)
                -- ) then
                --     -- only log the specified waves
                -- else
                --     return
                -- end
            end

            DCEI.Event.Log(key, props)

            -- debug option for printing log events & event properties
            if DEBUG.print_events then
                local print = "> LOGGING EVENT [" .. key .. "]\n"
                for k, v in pairs(props) do
                    print = print .. k .. " = " .. tostring(v) .. "\n"
                end
                DCEI.LogMessage(print)
            end
        else
            DCEI.LogError("error: attempted to log invalid event")
        end
    end
end

bank.LogEvent = LogEvent

function LogUserProperty(prop)
    if prop then
        local key = prop.name or prop.Name()
        local value = prop.value or prop.Value()
        DCEI.Event.SetUserProperty(key, value)

        -- debug option for printing user properties
        if DEBUG.print_user_props then
            DCEI.LogMessage("> LOGGING [" .. key .. "] = " .. value)
        end
    else
        DCEI.LogError("error: attempted to log invalid user property")
    end
end

bank.LogUserProperty = LogUserProperty

function LogUserProperty_GearEquipped()
    LogUserProperty(UserProps.gear_equipped)
    LogUserProperty(UserProps.equipment_equipped)
end

bank.LogUserProperty_GearEquipped = LogUserProperty_GearEquipped

function SetGameState(state_name)
    state_name = state_name:gsub("%_", " ")
    -- DCEI.LogError("SetGameState: " .. state_name)
    DCEI.Event.SetGameState(state_name)
end

-- SECTION Help functions
function Valid(value)
    return value and value ~= "null"
end

-- UTILITY
function Clamp(value, min, max)
    if value < min then
        return min
    elseif value > max then
        return max
    end
    return value
end

-- http://lua-users.org/wiki/CopyTable
function ShallowCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in pairs(orig) do
            copy[orig_key] = orig_value
        end
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

function Deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == "table" then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[Deepcopy(orig_key)] = Deepcopy(orig_value)
        end
        setmetatable(copy, Deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

-- BUG TRACKING
function BugTracking_CreateHeroPromoteList(hero_data)
    local unit_data_name = hero_data and hero_data.name or "invalid"
    local list_show_time = tostring(DIAGNOSTIC.ul_create_time)
    local list_close_time = tostring(DIAGNOSTIC.ul_close_time)

    local event = LogEvents.BugTracking_CreateHeroPromoteList(unit_data_name, list_show_time, list_close_time)
    LogEvent(event)
end

-- !SECTION

-- SECTION Notifications
function ScheduleVaultNotification()
    if vault_full_notification_id then
        DCEI.CancelNotification(vault_full_notification_id)
    end
    if GetVaultGold() < data.vault.capacity then
        local title = DCEI.GetTranslationText("notification/vault/title")
        local body = DCEI.GetTranslationText("notification/vault/body")

        local time =
            math.floor((data.vault.capacity - GetVaultGold()) / data.vault:GetVaultIncome() * data.vault.interval)

        vault_full_notification_id = DCEI.ScheduleNotification(title, body, time)
        -- DCEI.LogMessage("ScheduleVaultNotification: "..time)
        -- DCEI.LogMessage(title)
        -- DCEI.LogMessage(body)
    end
end

function ScheduleNotification(time_stamp)
    local unlock_wave = 10
    if not time_stamp or wave.data.GetMax() < unlock_wave then
        return
    end

    -- cancel exist notifications
    for _, noti in ipairs(notifications) do
        DCEI.CancelNotification(noti.id)
    end
    notifications = {}
    local notification_count = 3

    if player.tests.create_time then
        local time_from_create = os.time() - player.tests.create_time
        local day_from_create = math.floor(time_from_create / 86400)
        notification_count = 3 - day_from_create
        if day_from_create >= 3 then
            -- push notification to all new players in first 3 days
            return
        end
    else
        -- stop push notifications to old players during test
        return
    end

    local reset_time_local_date = os.date("*t", next_daily_reward_time)
    local date = os.date("*t")
    date.hour = 0
    date.min = 0
    date.sec = 0
    local date_start = os.time(date)
    local time_elapsed = os.time() - date_start
    local eight_hours = 28800
    local eighteen_hours = 64800
    local day_length = 86400
    local notification_hour = 8
    local first_notification_sec = eight_hours - time_elapsed

    if reset_time_local_date.hour > 8 and reset_time_local_date.hour <= 18 then
        first_notification_sec = eighteen_hours - time_elapsed
        notification_hour = 18
    end

    if
        first_notification_sec < 0
        or reset_time_local_date.day ~= date.day
        or reset_time_local_date.hour > notification_hour
    then
        -- schedule notification tomorrow if current time > today's notification time, or today's daily reward has been claimed, or reset hour > notification hour
        first_notification_sec = first_notification_sec + day_length
    end

    local new_notifications_time =
        { first_notification_sec, first_notification_sec + day_length, first_notification_sec + day_length * 2 }

    -- if DevMode() then
    --     new_notifications_time = {30, 60, 90}
    --     DCEI.ShowFeedbackMessage("new_notifications_time")
    -- end

    DCEI.LogMessage("Schedule next [" .. notification_count .. "] days notification")
    for i = 1, notification_count do
        local time = new_notifications_time[i]
        local text_id = 1
        local new_schedule_id =
            DCEI.ScheduleNotification(notification_texts[text_id].title, notification_texts[text_id].body, time)
        table.insert(notifications, { id = new_schedule_id, time = time + os.time() })
    end
end

function CheckNotificationsWhenClaimDailyReward()
    local next_notification = notifications[1]
    if next_notification then
        local date = os.date("*t")
        local date_check = os.date("*t", next_notification.time)
        if date_check.day == date.day then
            -- next notification is today but today's reward has been claimed
            DCEI.CancelNotification(next_notification.id)
            table.remove(notifications, 1)
        end
    end
end

function CheckOSDaysElapsed(from_focus)
    local last_notification_id = DCEI.GetLastNotificationId()
    local from_notification = (last_notification_id ~= nil)
    if DevMode() then
        DCEI.TriggerAddTimerEventElapsed(function()
            local message = from_notification and "From notification" or "Not from notification"
            DCEI.ShowFeedbackMessage(message)
        end, 3, true, true)
    end

    if player.tests.create_time then
        local time_from_create = os.time() - player.tests.create_time
        local day_from_create = math.floor(time_from_create / 86400)

        if not player.os_days or player.os_days < day_from_create then
            player.os_days = day_from_create
            DCEI.LogMessage(player.os_days .. " days from create")
            local event = LogEvents.OpenGame(from_notification, from_focus)
            bank.CommitOSDays()
            LogEvent(event)
        end
    end
end

-- !SECTION

-- NOTE Load player back up save data
function LoadBackupSaveDataPriority()
    if not loading_backup_savedata then
        loading_backup_savedata = true
        local callback_func = function(result)
            if result then
                local index = 1
                local max_wave = 0
                if result.saves then
                    for id, entry in ipairs(result.saves) do
                        -- DCEI.LogMessage("Save data time:"..entry.time)
                        if entry.save and entry.save.max_wave then
                            -- DCEI.LogMessage("Max wave:"..entry.save.max_wave)
                            if entry.save.max_wave > max_wave then
                                index = id
                                max_wave = entry.save.max_wave
                            end
                        end
                    end
                    if max_wave > 0 then
                        -- DCEI.LogMessage("Restore wave:"..max_wave.."  id: "..index)
                        DCEI.TriggerAddTimerEventElapsed(function()
                            DCEI.ResetSavedMapDataByIndex(index)
                        end, 0, true, true)
                    end
                end
            end
        end
        DCEI.GetSaveDataHistory(10, callback_func)
    end
end

-- NOTE android back button, for google feature
function PressBackButton()
    if ui.load_confirm_message then
        ui.load_confirm_message:Close()
        return true
    end
    if ui.settings_menu then
        ui.settings_menu:Close()
        return
    end
    -- close hell mode menu
    if ui.hell_mode and tutorial.state == 0 then
        if ui.hell_mode.battle_active then
            ui.hell_mode:Hide()
        else
            ui.hell_mode:Close()
        end
    end
    -- exit battle if in battle
    if wave.data.is_active then
        if PlayerInTargetMode() then
            ExitTargetMode()
        elseif ui.wave_exit_confirm then
            ui.wave_exit_confirm:Resume()
        else
            if not wave.data.ModeStandard() then
                CreateWaveExitConfirmMenu()
            else
                WaveEndActions()
            end
        end
        return true
    end
    -- close pop up
    if ui.popup.is_active then
        ui.popup:Hide()
        return true
    end
    -- exit colony
    if
        ui.bottom_panel.state == "colony"
        or ui.bottom_panel.state == "challenge"
        or ui.bottom_panel.state == "raid_boss"
    then
        CloseAirshipMenu()
        return true
    end
    -- close shop
    if ui.shop.is_active then
        -- close shop info pop up
        if ui.shop.info_popup.is_active then
            ui.shop.info_popup:Hide()
            return true
        end
        CloseShop()
        return true
    end
    -- close unit card
    if ui.unit_card then
        ui.CloseUnitCard()
        return true
    end
    -- close list if any
    if ui.equip_hero_list then
        ui.equip_hero_list:Close()
        return true
    end
    -- close inventory/craft
    if ui.craft and ui.craft.active then
        ui.craft:Close()
        return true
    end
    if ui.item_menu and ui.item_menu.is_active then
        ui.item_menu:Close()
        return true
    end
    -- close town hall/town menu/monster lab/gold mine/colony
    if ui.town_menu and ui.town_menu.is_active then
        ui.town_menu:Hide()
        return true
    end
    if ui.list_town and ui.list_town.is_active then
        ui.list_town:Close()
        return true
    end
    if ui.monster_lab then
        ui.monster_lab:Close()
        return true
    end

    -- close deck menu
    -- close leaderboard/battle stats/calender/community
    -- close gear
    -- close talent
    if ui.leaderboard and ui.leaderboard.is_active then
        ui.leaderboard:Close()
        return true
    end
    if ui.battlestats and ui.battlestats.is_active then
        ui.battlestats:Close()
        return true
    end
    if ui.calendar and ui.calendar.is_visible then
        ui.calendar:Hide()
        return true
    end
    if ui.social and ui.social.is_active then
        ui.social:Hide()
        return true
    end
    if ui.worker_menu then
        ui.worker_menu:Close()
        return true
    end
    if ui.vault_menu then
        ui.vault_menu:Close()
        return true
    end
    if ui.gear_menu and ui.gear_menu.is_active then
        ui.gear_menu.Close()
        return true
    end
    if ui.talent_menu and ui.talent_menu.is_active then
        ui.talent_menu.Close()
        return true
    end
    if ui.player_deck and ui.player_deck.is_active then
        ui.player_deck:Hide()
        return true
    end

    if ui.list_hero and ui.list_hero.is_active then
        ui.list_hero.Close()
        return true
    end

    if ui.list_tower and ui.list_tower.is_active then
        ui.list_tower.Close()
        return true
    end

    if ui.list_defense and ui.list_defense.is_active then
        ui.list_defense.Close()
        return true
    end

    if ui.list_promote and ui.list_promote.is_active then
        ui.list_promote.Close()
        return true
    end
    ShowSettingsMenu()
    return true
end

-- NOTE Initialization
BehaviorCallbacks = {
    ["Town Worker Proc Income"] = WorkerDwarfIncome,
    ["Town Worker ToDepot"] = WorkerDwarf_ToDepot,
    ["Town Worker ToMines"] = WorkerDwarf_ToMines,
    ["Defense IceFieldGenerator Aura Enable"] = ProcessIceFieldGenerator,
}

InBattleBehaviorCallbacks = {
    ["Player Castle Chunk HitResponse Trigger"] = ProcessCastleDamage,
    ["Hero DarkScout Skill LeechSoul Trigger"] = ProcessHeroSkillLeechSoul,
    ["Hero DarkScout Skill Leech Trigger"] = ProcessHeroSkillLeech,
    ["Hero Lifesteal Trigger"] = ProcessHeroLifesteal,
    ["Paragon Lifesteal Trigger"] = ProcessParagonHeroLifesteal,
    ["Hero DwarfSmith Skill RoyalRoar Trigger"] = ProcessRoyalRoarLifesteal,
    ["Summon Gear Golem Arcane ManaSteal Trigger"] = ProcessSummonArcaneGolemManaSteal,
    ["Hero Mason 2A Skill Repair Trigger"] = ProcessHeroSkillRepair,
    ["Hero WhiteMage Skill Channel Buff"] = ProcessHeroSkillChannel,
    ["Hero BlackMage Skill Curse Debuff"] = ProcessHeroSkillCurse,
    ["Hero BlackMage Skill Curse 2 Debuff"] = ProcessHeroSkillCurse,
    ["Hero BlackMage Skill Decrepify Debuff"] = ProcessHeroSkillDecrepify,
    ["Hero BlackMage Skill Curse Visual Effect Trigger"] = function()
        core.CreateEffect("Hero BlackMage Skill Curse Visual Effect", 24, 26)
    end,

    ["Hero Preacher Skill Inspire ArcherAttackDamage Buff"] = ProcessHeroSkillInspire,
    ["Hero Preacher Skill RousingWords HeroAttackSpeed Buff"] = ProcessHeroSkillRousingWords,
    ["Hero Marksman Skill Volley ArcherAttackSpeed Buff"] = ProcessHeroSkillVolley,
    ["Hero Lancer Skill Reinforce 2 Poison Troop DoT DamageProc"] = function(u)
        local hero_data = data.heroes["Hero Lancer 3AB"]
        ApplyHeroAttackDamage(hero_data, u, 0.5)
    end,

    ["Hero BabyDragon Skill Dragon Allies Poison Troop DoT DamageProc"] = function(u)
        local hero_data = data.heroes["Hero BabyDragon 2A"]
        ApplyHeroAttackDamage(hero_data, u, 0.5)
    end,

    ["Castle Archer Gear AcidArrow Debuff"] = ProcessAcidArrowUpgrade,
    ["Castle Archer Gear IceArrow Slow"] = ProcessIceArrowUpgrade,
    ["Gear Unity Hero Reduce Cooldown Trigger"] = ProcessUnityHeroCooldownReduce,
    ["RaidBoss Golem Gold Lightning Stun Debuff"] = function(u)
        ProcessHeroStun(u, true)
    end,
    ["RaidBoss Ult Dragon Fire Prison Stun Debuff"] = function(u)
        ProcessHeroStun(u, true)
    end,
    ["Hero Stun Recover"] = function(u)
        ProcessHeroStun(u, false)
    end,
    ["Monster Goblin Warship Summon"] = wave.ProcessMonsterGoblinWarshipSummon,
    ["Polymorphed Monster"] = function(u)
        wave.ProcessMorphedMonster(u)
        SendEnemyTowardsCastle(u)
        -- apply upgrades to monster units
        ApplyUpgradeDataToUnit(data.monster_upgrades, u)
    end,
    ["RaidBoss Golem Green Summon"] = wave.ProcessRaidBossGreenSummons,
    ["RaidBoss Golem Purple Summon"] = wave.ProcessRaidBossPurpleSummons,
    ["RaidBoss Golem Gold Summon"] = wave.ProcessRaidBossGoldSummons,
    ["RaidBoss Ult Dragon Summon"] = wave.ProcessRaidBossUltDragonSummons,
    ["Hero DrowMage Skill Time Lock Visual Effect Trigger"] = function()
        core.CreateEffect("Hero DrowMage Skill Time Lock Visual Effect", 24, 26)
    end,
    ["Hero DrowMage Skill Chronosphere Visual Effect Trigger"] = function()
        core.CreateEffect("Hero DrowMage Skill Chronosphere Visual Effect", 24, 26)
    end,
    ["Hero Season Event Skill Coin Drop Trigger"] = function(u)
        local name = DCEI.TriggeringBehaviorName
        local stacks = DCEI.GetUnitBehaviorStackCount(u, name)
        local skip_commit = true
        local icon = "general_icon_icecoin1"
        if CheckSeasonEventExist("winter_event2022") then
            AddPlayerWinterCoin2022(stacks, skip_commit)
            wave_drop_resource.winter_coin = wave_drop_resource.winter_coin or 0
            wave_drop_resource.winter_coin = wave_drop_resource.winter_coin + stacks
        else
            AddPlayerAutumnCoin(stacks, skip_commit)
            icon = "general_icon_coin_leaf"
            wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin or 0
            wave_drop_resource.autumn_coin = wave_drop_resource.autumn_coin + stacks
        end
        ui.CreateFloatingLootDisplay(u, icon, stacks)
    end,
    ["Devil Trail Negative Castle HP Lose Trigger"] = ProcessDevilTrailCastleHPLose,
}

ProcessSummonedUnitCallbacks = {
    -- lancer summons
    ["Hero Lancer Skill Reinforce Troop"] = ProcessHeroSkillReinforce,
    ["Hero Lancer Skill Reinforce 2 Troop"] = ProcessHeroSkillReinforce,
    ["Hero Lancer Skill Reinforce 2 Poison Troop"] = ProcessHeroSkillReinforce,
    -- knight summons
    ["Hero Knight Skill Defend Troop"] = ProcessHeroSkillDefend,
    ["Hero Knight Skill Defend Fire Troop"] = ProcessHeroSkillDefend,
    ["Hero Knight Skill Defend Arcane Troop"] = ProcessHeroSkillDefend,
    ["Hero Knight Skill Defend 2 Troop"] = ProcessHeroSkillDefend,
    ["Hero Knight Skill Defend 2 Cold Troop"] = ProcessHeroSkillDefend,
    ["Hero Knight Skill Defend 2 Lightning Troop"] = ProcessHeroSkillDefend,
    -- angel summons
    ["Hero Angel Skill Host Troop"] = ProcessHeroSkillAngelicHost,
    ["Hero Angel Skill Host 2 Troop"] = ProcessHeroSkillAngelicHost,
    ["Hero Angel Skill Chthonic Host Troop"] = ProcessHeroSkillAngelicHost,
    -- dragon summons
    ["Hero BabyDragon Skill Dragon Allies Fire Troop"] = ProcessHeroSkillBabyDragon,
    ["Hero BabyDragon Skill Dragon Allies Poison Troop"] = ProcessHeroSkillBabyDragon,
    ["Hero BabyDragon Skill Dragon Allies Ice Troop"] = ProcessHeroSkillBabyDragon,
    ["Hero BabyDragon Skill Dragon Boss Arcane Troop"] = ProcessHeroSkillBabyDragon,
    ["Hero BabyDragon Skill Dragon Boss Lightning Troop"] = ProcessHeroSkillBabyDragon,
    -- wolf huntress summons 
    ["Hero WolfHuntress Skill Troop Wolf"] = ProcessHeroSkillWolf,
    -- goblin 3A mines
    ["Hero Goblin Skill GobBomb Mine"] = ProcessHeroSkillGoblin3AmineMove,
    -- barrack summons
    ["Tower Barracks Troop"] = ProcessTowerBarracksSummon,
    ["Tower Barracks 2A Troop"] = ProcessTowerBarracksSummon,
    ["Tower Barracks 3A Troop"] = ProcessTowerBarracksSummon,
    -- stone guard summons
    ["Tower Monolith Troop"] = ProcessTowerMonolithSummon,
    ["Tower Monolith 2A Troop"] = ProcessTowerMonolithSummon,
    ["Tower Monolith 3A Troop"] = ProcessTowerMonolithSummon,
    -- legendary weapon necromancer summons
    ["Legendary Summon SkeletonArcher"] = ProcessLegendarySummonSkeleton,
}

OnMapStart()

if DEBUG.run_tests then
    DCEI.TriggerAddTimerEventElapsed(RunTests, 0, false, true)
end

local SIMPLE = true
DCEI.TriggerAddUnitSpawnEvent(DCEI.UnitAny, OnUnitSpawn, SIMPLE)
DCEI.TriggerAddUnitDiedEvent(DCEI.UnitAny, OnUnitDied, SIMPLE)
DCEI.TriggerAddUnitRemovedEvent(DCEI.UnitAny, OnUnitRemoved, SIMPLE)
DCEI.TriggerAddBehaviorAddEvent(DCEI.UnitAny, OnUnitBehaviorAdd, SIMPLE)
DCEI.TriggerAddCastAbilityEvent(DCEI.UnitAny, OnAbilityCast, SIMPLE)
DCEI.TriggerAddUnitAbilityOnOffEvent(DCEI.UnitAny, OnAbilityCooldownChange, SIMPLE)
DCEI.TriggerAddTextCommandEvent(OnTextCommand)
DCEI.TriggerOnFocusChange(OnFocusChange)

return game
